(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Lists
 * Listes
 * Listas
 *
 *)

open BairePervasives


(** Utilitaires *)
(** Utils *)
(** tiles *)


let moduleName = "BaireList"

let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type


(** Oprations catgorielles *)
(** Categorial operations *)
(** Operaciones categoriales *)


let injection = function x -> [x]
  
(* Evaluation order is forced like in the Caml standard library *)
(* On impose un ordre d'valuation comme dans la librairie standard de Caml *)
(* Se impone orden de evaluacin como en la librera estandar de Caml *)

let rec map f = function 
  | [] -> []
  | head :: tail -> let r = (f head) in r :: map f tail

let rec plus = fun a b ->
  match a with
    | [] -> b
    | head :: tail -> head :: plus tail b

let zero () = []

let (@) = plus

let rec join = function
  | [] -> []
  | head :: tail -> head @ join tail

let rec make_rec x = function
  | 0 -> []
  | k -> x :: make_rec x (k - 1)
      
let make = fun n x ->
  if (n < 0) then invalid "make" Negative
  else make_rec x n

let from_list = id

let to_list = id
    
let rec get = fun list n ->
  match list with
    | [] -> raise (Out_of_bounds n)
    | head :: _  when (n = 0) -> head
    | _ :: tail -> get tail (n - 1)

let rec set = fun list n value ->
  match list with
    | [] -> raise (Out_of_bounds n)
    | _ :: tail when (n = 0) -> value :: tail
    | head :: tail -> head :: set tail (n - 1) value


(** Reversals *)      
(** Renversements *)
(** Inversiones *)


let rec reverse_append = fun list result ->
  match list with
    | [] -> result
    | head :: tail -> reverse_append tail (head :: result)

let reverse = function list -> reverse_append list []

let rec append_reverse_rec = fun reversed a b ->
  match (a, b) with
    | ([], []) -> reversed
    | ([], _) -> reverse_append b reversed
    | (_, []) -> a @ reversed
    | (x :: xs, y :: ys) -> x :: append_reverse_rec (y :: reversed) xs ys

let append_reverse = fun a b -> append_reverse_rec [] a b

let rec reverse_map_append f result = function
  | [] -> result
  | head :: tail -> reverse_map_append f (f head :: result) tail

let reverse_map f = function list -> reverse_map_append f [] list

let rec reverse_take_rec result k = function 
  | [] -> raise (Out_of_bounds k)
  | head :: _ when k = 1 -> head :: result 
  | head :: tail -> reverse_take_rec (head :: result) (k - 1) tail

let reverse_take = fun n list -> reverse_take_rec [] n list

let rec reverse_unzip_rec left right = function
  | [] -> (left, right)
  | (x, y) :: tail -> reverse_unzip_rec (x :: left) (y :: right) tail

let reverse_unzip = function list -> reverse_unzip_rec [] [] list


(** Acces *)
(** Accs *)
(** Acceso *)


let cons = fun x list -> x :: list

let rec snoc = fun list x ->
  match list with
    | [] -> [x]
    | head :: tail -> head :: snoc tail x

let init = fun n f ->
  let rec init_rec = function
    | k when (k = n) -> []
    | k -> f k :: init_rec (k + 1)
  in
    if (n < 0) then invalid "init" Negative
    else init_rec 0

let rec filter f = function
  | [] -> []
  | head :: tail when (f head) -> head :: filter f tail
  | _ :: tail -> filter f tail

let rec take = fun n list ->
  match list with
    | _ when (n = 0) -> []
    | [] -> raise (Out_of_bounds n)
    | tete :: tail -> tete :: take (n - 1) tail

let rec drop = fun n list ->
  match list with
    | _ when n = 0 -> list
    | [] -> raise (Out_of_bounds n)
    | _ :: tail -> drop (n - 1) tail

let rec split_rec k left = function
  | right when k = 0 -> (reverse left, right)
  | head :: tail -> split_rec (k - 1) (head :: left) tail
  | _  -> raise (Out_of_bounds k)

let split_at = fun n list -> split_rec n [] list

let rec take_while f = function
  | head :: tail when (f head) -> head :: take_while f tail
  | _ -> []

let rec drop_while f = function
  | head :: tail when (f head) -> drop_while f tail
  | list -> list

let rec split_rec f left = function
  |  head :: tail when f head -> split_rec f (head :: left) tail
  |  right -> (reverse left, right)

let split_while = fun f list -> split_rec f [] list

let sub = fun dbut fin liste -> 
  let rec sub_rec n = function
    | [] when n = fin -> []
    | [] -> raise (Out_of_bounds fin)
    | _ :: tail when n < dbut -> sub_rec (n + 1) tail
    | _ when n > fin -> []
    | head :: tail -> head :: sub_rec (n + 1) tail 
  in
    if (dbut < 0) then invalid "sub" Negative
    else if (dbut > fin) then invalid "sub" OutOfBounds
    else sub_rec 0 liste


(** View *)
(** Vues *)
(** Vistas *)


let rec partition_rec f trueL falseL = function
  | [] -> (reverse trueL, reverse falseL)
  | head :: tail when (f head) -> partition_rec f (head :: trueL) falseL tail
  | head :: tail -> partition_rec f trueL (head :: falseL) tail

let partition f = function list -> partition_rec f [] [] list

let left_view = function
  | [] -> raise Empty
  | head :: tail -> (head, tail)

let left_head = function
  | [] -> raise Empty
  | head :: _ -> head

let left_tail = function
  | [] -> raise Empty
  | _ :: tail -> tail

let rec right_view_rec result = function
  | [] -> raise Empty
  | [x] -> (reverse result, x)
  | head :: tail -> right_view_rec (head :: result) tail

let right_view = function list -> right_view_rec [] list

let rec right_head = function
  | [] -> raise Empty
  | [x] -> x
  | _ :: tail -> right_head tail

let rec right_tail = function
  | [] -> raise Empty
  | [_] -> []
  | head :: tail -> head :: right_tail tail


(** Observers *)
(** Observateurs *)
(** Observadores *)


let is_empty = function
  | [] -> true
  | _ -> false

let rec size_rec k = function
  | [] -> k
  | _ :: tail -> size_rec (k + 1) tail

let size = function list -> size_rec 0 list

let rec exists f = function 
  | [] -> false
  | head :: tail -> f head || exists f tail

let rec for_all f = function
  | [] -> true
  | head :: tail -> f head && for_all f tail

let rec look_for f = function
  | [] -> raise Not_found
  | head :: _ when f head -> head
  | _ :: tail -> look_for f tail


(** Traversals *)
(** Parcours *)
(** Recorridos *)


let rec iter f = function
  | [] -> ()
  | head :: tail -> f head ; iter f tail

let rec fold_right f e = function
  | [] -> e
  | head :: tail -> f head (fold_right f e tail)

let rec fold_left f e = function
  | [] -> e
  | head :: tail -> fold_left f (f e head) tail


(** Miscellaneous *)
(** Divers *)
(** Micelaneos *)


let rec bind = fun list f ->
  match list with
    | [] -> []
    | head :: tail -> f head @ bind tail f

let rec concat_map f = function
  | [] -> []
  | head :: tail -> f head @ concat_map f tail


(** Index based operations *)
(** Operations indices *)
(** Operaciones con indices *)


let rec index_rec x k = function
  | [] -> raise Not_found
  | head :: _ when head = x -> k
  | _ :: tail -> index_rec x (k + 1) tail

let index x = function list -> index_rec x 0 list

let rec mapi_rec f k = function
  | [] -> []
  | head :: tail -> let r = f k head in r :: mapi_rec f (k + 1) tail

let map_with_index f = function list -> mapi_rec f 0 list

let rec iteri_rec f k = function
  | [] -> ()
  | head :: tail -> f k head ; iteri_rec f (k + 1) tail

let iter_with_index f = function list -> iteri_rec f 0 list

let rec look_fori_rec f k = function
  | [] -> raise Not_found
  | head :: _ when f k head -> head
  | _ :: tail -> look_fori_rec f (k + 1) tail

let look_for_with_index f = function list -> look_fori_rec f 0 list

let rec filteri_rec f k = function
  | [] -> []
  | head :: tail when (f k head) -> head :: filteri_rec f (k + 1) tail
  | _ :: tail -> filteri_rec f (k + 1) tail

let filter_with_index f = function list -> filteri_rec f 0 list

let rec foldri f e k = function
  | [] -> e
  | head :: tail -> f k head (foldri f e (k + 1) tail)

let fold_right_with_index f e = function list -> foldri f e 0 list

let rec foldli f e k = function
  | [] -> e
  | head :: tail -> foldli f (f k e head) (k + 1) tail

let fold_left_with_index f e = function list -> foldli f e 0 list

let fill = fun list dbut fin motif ->
  let rec fill_rec = function
    | (_, []) -> []
    | (k, _ :: tail) when (k >= dbut) && (k < fin) -> 
	motif :: fill_rec (k + 1, tail)
    | (k, x :: tail) -> x :: fill_rec (k + 1, tail)
  in 
    if (dbut < 0) then invalid "fill" Negative
    else if (dbut > fin) then invalid "fill" OutOfBounds
    else fill_rec (0, list)

let blit = fun list_a indiceA list_b indiceB taille ->
  let rec 
    decale = function
      | (_, _, [], _) -> raise (Out_of_bounds indiceA)
      | (_, _, _, []) -> raise (Out_of_bounds indiceB)
      | (0, 0, list_a, list_b) -> copie (taille, list_a, list_b)
      | (n, 0, _ :: tail_a, list) -> decale (n - 1, 0, tail_a, list)
      | (0, m, list, head_b :: tail_b) -> 
	  head_b :: decale (0, m - 1, list, tail_b)
      | (n, m, _ :: tail_a, head_b :: tail_b) -> 
	  head_b :: decale (n - 1, m - 1, tail_a, tail_b)
  and
    copie = function
      | (0, _, list_b) -> list_b 
      | (k, head_a :: tail_a, _ :: tail_b) 
	-> head_a :: copie (k - 1, tail_a, tail_b)
      | _ -> raise (Out_of_bounds taille)
  in
    decale (indiceA, indiceB, list_a, list_b)


(* Combinaisons *)


(* Ne lve pas d'exception si les listes ne sont pas de la meme taille *)
let rec unsafe_zip = fun list_a list_b ->
  match (list_a, list_b) with
    | ([], _) -> []
    | (_, []) -> []
    | (head_a :: tail_a, head_b :: tail_b) -> 
	(head_a, head_b) :: unsafe_zip tail_a tail_b

let rec zip = fun list_a list_b ->
  match (list_a, list_b) with
    | ([], []) -> []
    | (head_a :: tail_a, head_b :: tail_b) ->
	(head_a, head_b) :: zip tail_a tail_b
    | _ -> invalid "zip" SameLength

let rec unzip_rec list_a list_b = function
  | [] -> (reverse list_a, reverse list_b)
  | (x,y) :: tail -> unzip_rec (x :: list_a) (y :: list_b) tail

let unzip = function list -> unzip_rec [] [] list


(* Operations utilises par d'autres librairies *)


let rec iter2 = fun f a b ->
  match (a, b) with
    | ([], []) -> ()
    | (head_a :: tail_a, head_b :: tail_b) -> 
	f head_a head_b ; iter2 f tail_a tail_b
    | _ -> invalid "iter2" SameLength

let rec map2 f = fun a b ->
  match (a, b) with
    | ([], []) -> []
    | (head_a :: tail_a, head_b :: tail_b) ->
	f head_a head_b :: map2 f tail_a tail_b
    | _ -> invalid "map2" SameLength

let reverse_map2 = fun f a b ->
  let rec map2_rec result a b =
    match (a, b) with
      | ([], []) -> result
      | (head_a :: tail_a, head_b :: tail_b) -> 
	  map2_rec (f head_a head_b :: result) tail_a tail_b
      | _ -> invalid "reverse_map" SameLength
  in
    map2_rec [] a b

let rec fold_left2 f = fun result a b ->
  match (a, b) with
    | ([], []) -> result
    | (head_a :: tail_a, head_b :: tail_b) ->
	fold_left2 f (f result head_a head_b) tail_a tail_b
    | _ -> invalid "fold_left2" SameLength

let rec fold_right2 f e = fun a b ->
    match (a, b) with
      | ([], []) -> e
      | (head_a :: tail_a, head_b :: tail_b) ->
	  f head_a head_b (fold_right2 f e tail_a tail_b)
      | _ -> invalid "fold_right2" SameLength
 
let rec for_all2 = fun f a b ->
  match (a, b) with
    | ([], []) -> true
    | (head_a :: tail_a, head_b :: tail_b) -> 
	f head_a head_b && for_all2 f tail_a tail_b
    | _ -> invalid "for_all2" SameLength

let rec exists2 = fun f a b ->
  match (a, b) with
    | ([], []) -> false
    | (head_a :: tail_a, head_b :: tail_b) -> 
	f head_a head_b || exists2 f tail_a tail_b
    | _ -> invalid "exits2" SameLength


(** Aliases *)


let empty = zero
let append = plus
let (@) = plus


(* Oprations sur les collections *)


let insert = cons
       
let insert_list = plus

let union = plus

let rec delete x = function
  | [] -> []
  | head :: tail when head = x -> tail
  | head :: tail -> head :: delete x tail

let rec delete_q x = function
  | [] -> []
  | head :: tail when head == x -> tail
  | head :: tail -> head :: delete x tail
      
let rec delete_all x = function
  | [] -> []
  | head :: tail when head = x -> delete_all x tail
  | head :: tail -> head :: delete_all x tail
      
let rec delete_all_q x = function
  | [] -> []
  | head :: tail when head = x -> delete_all_q x tail
  | head :: tail -> head :: delete_all_q x tail
      
let rec member x = function
  | [] -> false
  | head :: _ when head = x -> true
  | _ :: tail -> member x tail
      
let rec member_q x = function
  | [] -> false
  | head :: _ when head == x -> true
  | _ :: tail -> member_q x tail
      
let rec count_rec x n = function
  | [] -> n
  | head :: tail when (head = x) -> count_rec x (n + 1) tail
  | _ :: tail -> count_rec x n tail
      
let count x = function list -> count_rec x 0 list
  
let rec count_q_rec x n = function
  | [] -> n
  | head :: tail when (head == x) -> count_q_rec x (n + 1) tail
  | _ :: tail -> count_q_rec x n tail
      
let count_q x = function list -> count_q_rec x 0 list
  
let fold = fold_left


(* Oprations sur les collections ordonnes *)


let rec min_element_rec min = function
  | [] -> min
  | head :: tail when head < min -> min_element_rec head tail
  | _ :: tail -> min_element_rec min tail

let min_element = function
  | [] -> raise Empty
  | head :: tail -> min_element_rec head tail

let rec max_element_rec max = function
  | [] -> max
  | head :: tail when head > max -> max_element_rec head tail
  | _ :: tail -> max_element_rec max tail

let max_element = function
  | [] -> raise Empty
  | head :: tail -> max_element_rec head tail

let min_view = function
  | [] -> raise Empty
  | list -> let min = min_element list in (min, delete_all min list)

let max_view = function
  | [] -> raise Empty
  | list -> let max = max_element list in (max, delete_all max list)

let delete_min = function
  | [] -> raise Empty
  | list -> delete_all (min_element list) list

let delete_max = function
  | [] -> raise Empty
  | list -> delete_all (max_element list) list

(* Oprations sur les listes d'associations 
 * Elles sont un suffixes par _assoc 
 * Le module ListAssociation permet un accs non suffix
 *)
	     
type ('k, 'a) assoc = ('k * 'a) list
			
let injection_assoc = fun key x -> [(key, x)]
  
let rec map_assoc f = function
  | [] -> []
  | (k, x) :: tail -> (k, f x) :: map_assoc f tail
      
let rec bind_assoc = fun list f ->
  match list with
    | [] -> []
    | (_, x) :: tail -> f x @ bind_assoc tail f
	
let zero_assoc = zero
		   
let plus_assoc = plus
		   
let insert_assoc = fun key x list -> (x, key) :: list
  
let insert_list_assoc = plus
			  
let union_assoc = plus
		    
(* ne lve pas d'exception si l'lment n'a pas t trouv *)
		    
let rec delete_rec_assoc key result = function
  | [] -> reverse result
  | (k, _) :: tail when (k = key) -> reverse_append result tail
  | head :: tail -> delete_rec_assoc key (head :: result) tail
      
let delete_assoc = fun list key -> delete_rec_assoc key [] list
  
let rec delete_q_rec_assoc key result = function
  | [] -> reverse result
  | (k, _) :: tail when (k == key) -> reverse_append result tail
  | head :: tail -> delete_q_rec_assoc key (head :: result) tail
      
let delete_q_assoc = fun list key -> delete_q_rec_assoc key [] list
  
let rec delete_all_rec_assoc key result = function
  | [] -> reverse result
  | (k, _) :: tail when (k = key) -> delete_all_rec_assoc key result tail
  | head :: tail -> delete_all_rec_assoc key (head :: result) tail
      
let delete_all_assoc = fun list key -> delete_all_rec_assoc key [] list
  
let rec delete_all_q_rec_assoc key result = function
  | [] -> reverse result
  | (k, _) :: tail when (k == key) -> 
      delete_all_q_rec_assoc key result tail
  | head :: tail -> delete_all_q_rec_assoc key (head :: result) tail
      
let delete_all_q_assoc = fun list key -> delete_all_q_rec_assoc key [] list
  
let is_empty_assoc = is_empty
		       
let size_assoc = size
		   
let rec member_assoc = fun list key ->
  match list with
    | [] -> false
    | (k, _) :: _ when (k = key) -> true
    | _ :: tail -> member_assoc tail key
	
let rec member_q_assoc = fun list key ->
  match list with
    | [] -> false
    | (k, _) :: _ when (k == key) -> true
    | _ :: tail -> member_assoc tail key
	
let rec count_rec_assoc key n = function
  | [] -> n
  | (k, _) :: tail when (k = key) -> count_rec_assoc key (n + 1) tail
  | _ :: tail -> count_rec_assoc key n tail
      
let count_assoc = fun list key -> count_rec_assoc key 0 list
  
let rec count_q_rec_assoc key n = function
  | [] -> n
  | (k, _) :: tail when (k = key) -> count_q_rec_assoc key (n + 1) tail
  | _ :: tail -> count_q_rec_assoc key n tail
      
let count_q_assoc = fun list key -> count_q_rec_assoc key 0 list
  
let rec lookup_assoc = fun list key ->
  match list with
    | [] -> raise Not_found
    | (k, x) :: _ when (k = key) -> x
    | _ :: tail -> lookup_assoc tail key
	
let rec lookup_q_assoc = fun list key ->
  match list with
    | [] -> raise Not_found
    | (k, x) :: _ when (k == key) -> x
    | _ :: tail -> lookup_q_assoc tail key
	
let rec lookup_all_assoc = fun list key ->
  match list with
    | [] -> []
    | (k, x) :: tail when (k = key) -> x :: lookup_all_assoc tail key
    | _ :: tail -> lookup_all_assoc tail key
	
let rec lookup_all_q_assoc = fun list key ->
  match list with
    | [] -> []
    | (k, x) :: tail when (k == key) -> x :: lookup_all_q_assoc tail key
    | _ :: tail -> lookup_all_q_assoc tail key
	
let rec elements_assoc = function
  | [] -> []
  | (_, x) :: tail -> x :: elements_assoc tail
      
let rec adjust_assoc f key = function
  | [] -> []
  | (k, x) :: tail when (k = key) -> (key, f x) :: tail
  | head :: tail -> head :: adjust_assoc f key tail
      
let rec adjust_all_assoc f key = function
  | [] -> []
  | (k, x) :: tail when (k = key) ->
	(key, f x) :: adjust_all_assoc f key tail
  | head :: tail -> head :: adjust_all_assoc f key tail

let rec fold_assoc f e = function
  | [] -> e
  | (_, x) :: tail -> fold_assoc f (f e x) tail

let rec fold_with_key_assoc f e = function
  | [] -> e
  | (k, x) :: tail -> fold_with_key_assoc f (f e k x) tail
      
let rec filter_keys_assoc f = function
  | [] -> []
  | (k, _) :: tail when f k -> filter_keys_assoc f tail
  | head :: tail -> head :: filter_keys_assoc f tail
      
let rec filter_elements_assoc f = function
  | [] -> []
  | (k, x) as e :: tail when (f x) -> e :: filter_elements_assoc f tail
  | _ :: tail -> filter_elements_assoc f tail
      
let delete_list_assoc = fun list to_delete ->
  let rec make_deletion_list f = function
    | [] -> f
    | head :: tail when f head -> make_deletion_list f tail
    | head :: tail -> make_deletion_list
	(fun x -> x <> head && f x) tail
  in
  let filter_function = make_deletion_list (fun x -> true) to_delete
  in 
    filter_keys_assoc filter_function list
      
let partition_keys_assoc = fun f list ->
  let rec part_rec = function
    | ([], left, right) -> (left, right)
    | ((k, _) as head :: tail, left, right) when (f k) ->
	part_rec (tail, head :: left, right)
    | (head :: tail, left, right) ->
	part_rec (tail, left, head :: right)
  in
    part_rec (list, [], [])
      
let partition_elements_assoc = fun f list ->
  let rec part_rec = function
    | ([], left, right) -> (left, right)
    | ((k, x) :: tail, left, right) when (f x) ->
	part_rec (tail, (k, x) :: left, right)
    | ((k, x) :: tail, left, right) ->
	part_rec (tail, left, (k, x) :: right)
  in
    part_rec (list, [], [])

let rec map_with_key_assoc f = function
  | [] -> []
  | (k, x) :: tail -> (k, f k x) :: map_with_key_assoc f tail

let from_list_assoc = id


(* Oprations sur les ensembles 
 * Elles sont suffixes par _set
 * Le module ListSet permet un accs non suffix
 *)

let rec khi_rec_set f = function
  | [] -> f
  | head :: tail when f head -> khi_rec_set f tail
  | head :: tail -> khi_rec_set (function x -> x = head || f x) tail

let khi_set = function list -> khi_rec_set (function x -> false) list

let rec from_list_rec_set f result = function
  | [] -> result
  | head :: tail when f head -> from_list_rec_set f result tail
  | head :: tail -> from_list_rec_set (fun x -> x = head || f x) 
      (head :: result) tail

let from_list_set = function list -> 
  from_list_rec_set (fun x -> false) [] list

let injection_set = injection
let map_set = map
let join_set = join

let rec union_rec_set f result = function
  | [] -> (f, result)
  | head :: tail when f head -> union_rec_set f result tail
  | head :: tail -> 
      union_rec_set (function x -> x = head || f x) (head :: result) tail

let union_set = fun list_a list_b ->
  let (khi, list) = union_rec_set (function x -> false) [] list_a in
    from_list_rec_set khi list list_b

let plus_set = union_set

let zero_set = zero

let insert_set x = function ensemble ->
  if member x ensemble then ensemble
  else x :: ensemble

let rec insert_list_set = fun list ensemble ->
  from_list_set (list @ ensemble)

let delete_set x = function ensemble -> delete x ensemble

let delete_list_set l = function x -> raise (Not_working "delete_list_set")

let is_empty_set = is_empty

let size_set = size

let member_set = member

let count_set = count

let filter_set = filter

let partition_set = partition

let fold_set = fold_left

let intersection_set = fun list_a list_b ->
  let khi = khi_set list_a in
  let rec inter_rec f = function
    | [] -> []
    | head :: tail when khi head && not (f head) ->
	head :: inter_rec (function x -> x = head || f x) tail
    | _ :: tail -> inter_rec f tail
  in
    inter_rec (function x -> false) list_b

let difference_set = fun list_a list_b ->
  let
    khi = khi_set list_a and
    psi = khi_set list_b
  in
    filter (fun x -> not (psi x)) list_a @ filter (fun x -> not (khi x)) list_b

let subset_set = fun list_a list_b ->
  let khi = khi_set list_b in for_all khi list_a
