(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Graph represended by adjacency lists
 * Graphes reprsents par listes d'adjacence
 * Grafos representados por listas de adjacencia
 *)

open BairePervasives

let moduleName = "AdjacencyListGraph"

type node = int
type 'a edge = node * 'a * node

type 'a graph = (node * 'a) list array


(** Utils **)
(** Utilitaires **)
(** tiles **)


let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type


(** Oprations catgorielles **)
(** Categorial operiations **)
(** Operaciones de categora **)


let injection = fun i v j -> 
  let graph = Array.make (1 + max i j) [] in
    graph.(i) <- [(j, v)] ;
    graph

let rec insert_to_all result i = function
  | [] -> result
  | (j, v) :: tail -> insert_to_all ((i, v, j) :: result) i tail

let rec to_list_rec graph result = function
  | n when n < 0 -> result
  | n -> to_list_rec graph (insert_to_all result n graph.(n)) (n - 1)

let to_list = function graph ->
  let n = Array.length graph in
    to_list_rec graph [] n
    
let rec max_list m = function
  | [] -> m
  | (i, _, j) :: tail ->
      let mx = (max i j) in 
	if mx > m then max_list mx tail else max_list m tail

let rec from_list_rec graph = function
  | [] -> graph
  | (i, v, j) :: tail -> 
      graph.(i) <- OrderedListMap.insert j v graph.(i) ;
      from_list_rec graph tail

let from_list = function list ->
  let m = max_list 0 list in
  let graph = Array.make (1 + m) [] in
    from_list_rec graph list

let map f = function graph -> from_list (List.map f (to_list graph))

let plus = fun a b ->
  let size = max (Array.length a) (Array.length b) in
  let graph = Array.make size [] in
    for i = 0 to size - 1 do
      graph.(i) <- OrderedListMap.plus a.(i) b.(i)
    done ;
    graph

let zero = function () -> [||]


(** Oprations binaires **)
(** Binairy operations **)
(** Operaciones binrias **)


let union = plus

let intersection = fun a b ->
  let size = min (Array.length a) (Array.length b) in
  let graph = Array.make size [] in
    for i = 0 to size - 1 do
      graph.(i) <- OrderedListMap.intersection a.(i) b.(i)
    done ;
    graph

let difference = fun a b ->
  let 
    size_a = Array.length a and
    size_b = Array.length b
  in
  let graph = Array.make size_a [] in
    for i = 0 to size_a - 1 do
      if i < size_b then graph.(i) <- OrderedListMap.difference a.(i) b.(i)
      else graph.(i) <- a.(i)
    done ;
    graph

let symetric_difference = fun a b ->
  let (min_g, min_s, max_g, max_s) =
    let size_a = Array.length a and size_b = Array.length b in
      if size_a > size_b then (b, size_b, a, size_a) 
      else (a, size_a, b, size_b)
  in
  let graph = Array.make max_s [] in
    for i = 0 to max_s - 1 do
      if i < min_s then 
	graph.(i) <- OrderedListMap.difference a.(i) b.(i)
      else graph.(i) <- max_g.(i)
    done ;
    graph


(** Observateurs **)
(** Observers **)
(** Observadores **)


let is_empty = function
  | [||] -> true
  | _ -> false

let rec nodes_rec total graph = function 
  | n when n < 0 -> total
  | n when graph.(n) = [] -> nodes_rec total graph (n - 1)
  | n -> nodes_rec (1 + total) graph (n - 1)

let nodes = function graph -> nodes_rec 0 graph (Array.length graph - 1)

let rec nodes_list_rec result graph = function 
  | n when n < 0 -> result
  | n when graph.(n) = [] -> nodes_list_rec result graph (n - 1)
  | n -> nodes_list_rec (n :: result) graph (n - 1)

let nodes_list = function graph -> 
  nodes_list_rec [] graph (Array.length graph - 1)

let rec list_size total = function
  | [] -> total
  | _ :: tail -> list_size (1 + total) tail

let rec edges_rec total graph = function
  | n when n < 0 -> total
  | n -> edges_rec (list_size total graph.(n)) graph (n - 1)

let edges = function graph -> edges_rec 0 graph (Array.length graph - 1)

let rec edges_rec i result = function
  | [] -> result
  | (j, v) :: tail -> edges_rec i ((i, v, j) :: result) tail

let rec edges_list_rec result graph = function
  | n when n < 0 -> result
  | n -> edges_list_rec (edges_rec n result graph.(n)) graph (n - 1)

let edges_list = function graph -> 
  edges_list_rec [] graph (Array.length graph - 1)

let rec labels_rec result = function
  | [] -> result
  | (_, v) :: tail -> labels_rec (v :: result) tail

let rec labels_list_rec result graph = function
  | n when n < 0 -> result
  | n -> labels_list_rec (labels_rec result graph.(n)) graph (n - 1)

let labels_list = function graph -> 
  labels_list_rec [] graph (Array.length graph - 1)

let exterior_degree = fun x graph -> List.length graph.(x)

let rec interior_degree_rec x graph total = function
  | n when n < 0 -> total
  | n -> 
      if OrderedListMap.member x graph.(n)
      then interior_degree_rec x graph (1 + total) (n - 1)
      else interior_degree_rec x graph total (n - 1)

let rec interior_degree = fun x graph -> 
  interior_degree_rec x graph 0 (Array.length graph - 1)

let degree = fun x graph -> 
  (interior_degree x graph, exterior_degree x graph)
  
let rec degrees_list_rec graph result = function
  | n when n < 0 -> result
  | n -> 
      degrees_list_rec graph 
      ((n, interior_degree n graph, List.length graph.(n)) :: result)
      (n - 1)

let degrees_list = function graph -> 
  degrees_list_rec graph [] (Array.length graph - 1)


(** Miscellaneous **)
(** Divers **)
(** Micelaneos **)


let union_with f = fun a b ->
  let size = max (Array.length a) (Array.length b) in
  let graph = Array.make size [] in
    for i = 0 to size - 1 do
      graph.(i) <- OrderedListMap.union_with f a.(i) b.(i)
    done ;
    graph

let intersection_with f = fun a b ->
  let size = min (Array.length a) (Array.length b) in
  let graph = Array.make size [] in
    for i = 0 to size - 1 do
      graph.(i) <- OrderedListMap.intersection_with f a.(i) b.(i)
    done ;
    graph


(** Operations on nodes **)
(** Oprations sur les noeuds **)
(** Operaciones sobre los nudos **)


let rec successors_rec result = function
  | [] -> result
  | (j, _) :: tail -> successors_rec (j :: result) tail
      
let successors = fun i graph -> successors_rec [] graph.(i)

let rec predecessors_rec i graph result = function
  | n when n < 0 -> result
  | n ->
      if OrderedListMap.member i graph.(n) 
      then predecessors_rec i graph (n :: result) (n - 1)
      else predecessors_rec i graph result (n - 1)

let predecessors = fun i graph -> 
  predecessors_rec i graph [] (Array.length graph - 1)

let adjacents = fun i graph -> (predecessors i graph, successors i graph)

let exterior_edges = fun i graph -> insert_to_all [] i graph.(i)

let rec interior_edges_rec i graph result = function
  | n when n < 0 -> result
  | n ->
      try
	let v = OrderedListMap.lookup i graph.(n) in
	  interior_edges_rec i graph ((n, v, i) :: result) (n - 1)
      with Not_found -> interior_edges_rec i graph result (n - 1)

let interior_edges = fun i graph ->
  interior_edges_rec i graph [] (Array.length graph - 1)

let adjacent_edges = fun i graph -> 
  (interior_edges i graph, exterior_edges i graph)

let member_node = fun n graph -> n >= 0 && n < Array.length graph

let rec exists_node_rec f graph = function
  | n when n < 0 -> false
  | n -> f n || exists_node_rec f graph (n - 1)

let exists_node = fun f graph -> 
  exists_node_rec f graph (Array.length graph - 1)

let rec for_all_nodes_rec f graph = function
  | n when n < 0 -> true
  | n -> f n && for_all_nodes_rec f graph (n - 1)

let for_all_nodes = fun f graph ->
  for_all_nodes_rec f graph (Array.length graph - 1)

let rec iter_nodes_rec f graph = function
  | n when n < 0 -> ()
  | n -> f n ; iter_nodes_rec f graph (n - 1)

let iter_nodes = fun f graph -> iter_nodes_rec f graph (Array.length graph - 1)


(** Operations on edges **)
(** Operations sur les arcs **)
(** Operaciones sobre los arcos **)


let insert_edge = fun i v j graph -> 
  graph.(i) <- OrderedListMap.insert j v graph.(i) ;
  graph

let delete_edge = fun i j graph -> 
  graph.(i) <- OrderedListMap.delete j graph.(i) ;
  graph

let member_edge = fun i j graph -> OrderedListMap.member j graph.(i)

let rec look_for_edge_rec f graph = function
  | n when n < 0 -> raise Not_found
  | n ->
      try
	let (j, v) = OrderedListMap.look_for (fun j v -> f n v j) graph.(n) 
	in (n, v, j)
      with Not_found -> look_for_edge_rec f graph (n - 1)

let look_for_edge = fun f graph -> 
  look_for_edge_rec f graph (Array.length graph - 1)

