(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Functional graphs
 * Graphes fonctionnels
 * Grafos funcionales
 *
 *)

open BairePervasives

let moduleName = "FunctionalHashGraph"

(* This graph is equivalent to
 * type 'a graph = (node, ('a, node) map) map
 *)

(* Cette structure de donnes est quivalente 
 * type 'a graph = (node, ('a, node) map) map
 *)

(* Esta estructura de datos es equivalente a
 * type 'a graph = (node, ('a, node) map) map
 *)

type node = int
type 'a edge = node * 'a * node

type 'a edge_list = E | N of node * 'a * node * 'a edge_list
type 'a graph = { mutable size : int ; mutable data : 'a edge_list array }


(** Utilitaires **)
(** Utils **)
(** tiles **)


(* the hashtable size should be a prime or a number having few small divisors.
 * We hope being able to provide this improvment in a future version
 *)

(* la taille de la table de hachage devrait tre un nombre premier, ou du moins
 * avec peu de petits diviseurs. Nous esprons dans une version future pouvoir
 * implmenter cela
 *)

(* el tamao de la tabla de hach debera ser un nmero primo, o por lo menos
 * con pocos pequeos divisores. Esperamos poder implementar sta mejora
 * en una proxima versin
 *)

let new_size = function x -> x

let hash = Hashtbl.hash

(* all the following functions are come from [OrderedListMap], but have
 * been specialized to the bidimensional case, that is
 * 
 * 
 * type 'a map = E | N of int * 'a * int * 'a map
 *
 * instead of (using 'a list notation)
 * 
 * type 'a map = [] | :: of (int * 'a * int) * 'a map
 *)

(* toutes les fonctions qui suivent sont extraites de [OrderedListMap], 
 * et spcialises pour le cas bidimensionnel 
 * 
 * type 'a map = E | N of int * 'a * int * 'a map
 *
 * au lieu de (on utilise la notation de 'a list)
 * 
 * type 'a map = [] | :: of (int * 'a * int) * 'a map
 *)

(* todas las funciones suiguientes viene de [OrderedListMap], pero has sido
 * especializadas para el caso bidimensional, es decir
 * 
 * type 'a map = E | N of int * 'a * int * 'a map
 *
 * en vez de (usando la notacin de 'a list)
 * 
 * type 'a map = [] | :: of (int * 'a * int) * 'a map
 *)

let injection_list = fun i v j -> N (i, v, j, E)

let rec insert_list' = fun i v j list ->
  match list with
    | E -> injection_list i v j
    | N (i', v', j', tail) ->
	if j < j' then N (i, v, j, list)
	else N (i', v', j', insert_list' i v j tail)

let rec insert_list i v j = function
  | E -> injection_list i v j
  | N (i', v', j', tail) as list ->
      if i < i' then N (i, v, j, list)
      else insert_list' i v j list

let rec size_list total = function
  | E -> total
  | N (_, _, _, tail) -> size_list (1 + total) tail

(* edges_to_list == to_list_list *)

let rec edges_to_list result = function
  | E -> result
  | N (i, v, j, tail) -> edges_to_list ((i, v, j) :: result) tail

let rec lookup_list i v = function
  | E -> raise Not_found
  | N (i', v', j, tail) ->
      match compare i i' with
	| n when n < 0 -> lookup_list i v tail
	| n when n > 0 -> raise Not_found
	| _ -> 
	    match compare v v' with
	      | n when n < 0 -> lookup_list i v tail
	      | n when n > 0 -> raise Not_found
	      | _ -> j

let rec look_for_list f = function
  | E -> raise Not_found
  | N (i, v, j, _) when f i v j -> (i, v, j)
  | N (_, _, _, tail) -> look_for_list f tail

let rec member_list i j = function
  | E -> false
  | N (i', _, j', tail) ->
      match compare i i' with
	| n when n < 0 -> member_list i j tail
	| n when n > 0 -> false
	| _ ->
	    match compare j j' with
	      | n when n < 0 -> member_list i j tail
	      | n when n > 0 -> false
	      | _ -> true

(* normalement, si le taux de collisions est faible, la version purement
 * fonctionnelle des listes devrait se montrer plus rapide que leur quivalent
 * impratif (en raison de leur plus faible occupation mmoire)
 *)


(** Oprations catgorielles **)
(** Categorial operiations **)
(** Operaciones de categora **)


let injection = fun i v j -> { size = 1; data = [| N (i, v, j, E) |] }

let rec to_list_rec data result = function
  | n when n < 0 -> result
  | n -> to_list_rec data (edges_to_list result data.(n)) (n - 1)

let to_list = function graph -> to_list_rec graph.data [] (graph.size)

let zero = function () -> { size = 0 ; data = [||] }

let insert = fun i v j data size ->
  let index = (hash (i, v)) mod size in
    data.(index) <- insert_list i v j data.(index)

let rec insert_list size data = function
  | [] -> ()
  | (i, v, j) :: tail ->
      insert i v j data size ;
      insert_list size data tail
      
let from_list = function list ->
  let size = new_size (List.length list) in
  let data = Array.make size E in
    insert_list size data list ;
    { size = size ; data = data }

let rec insert_edge_list size data = function
  | E -> ()
  | N (i, v, j, tail) ->
      insert i v j data size ;
      insert_edge_list size data tail

let rec resize_rec data size new_array = function
  | n when n < 0 -> ()
  | n ->
      insert_edge_list size new_array data.(n) ;
      resize_rec data size new_array (n - 1)

let rec resize = fun new_size graph ->
  let new_array = Array.make new_size E in
    resize_rec graph.data graph.size new_array (graph.size - 1);
    graph.data <- new_array ;
    graph.size <- new_size


(** Observateurs **)
(** Observers **)
(** Observadores **)


let is_empty = function graph -> graph.size = 0

let rec edges_rec data total = function
  | n when n < 0 -> total
  | n -> edges_rec data (size_list total data.(n)) (n - 1)

let edges = function graph -> edges_rec graph.data 0 (graph.size - 1)

let rec nodes_rec set = function
  | E -> set
  | N (i, _, j, tail) -> nodes_rec (AvlSet.insert i (AvlSet.insert j set)) tail

let rec nodes_list_rec set data = function
  | n when n < 0 ->  set
  | n -> nodes_list_rec (nodes_rec set data.(n)) data (n - 1)

let nodes_list = function graph -> 
  AvlSet.to_list (nodes_list_rec (AvlSet.zero ()) graph.data (graph.size - 1))

let nodes = function graph ->
  AvlSet.size (nodes_list_rec (AvlSet.zero ()) graph.data (graph.size - 1))

let rec labels_list_rec result = function
  | E -> result
  | N (_, v, _, tail) -> labels_list_rec (v :: result) tail

let rec labels_rec data result = function
  | n when n < 0 -> result
  | n -> labels_rec data (labels_list_rec result data.(n)) (n - 1)

let labels_list = function graph -> labels_rec graph.data [] (graph.size - 1)

let edges_list = to_list


(** Miscellaneous **)
(** Divers **)
(** Micelaneos **)


let next = fun i v graph ->
  let index = (hash (i, v)) mod graph.size in
    lookup_list i v graph.data.(index)

let rec next_list_rec data size i = function
  | [] -> i
  | v :: tail -> 
      next_list_rec data size 
      (lookup_list i v data.((hash (i, v)) mod size)) tail
	
let next_list = fun node list graph ->
  next_list_rec graph.data graph.size node list


(** Operations on nodes **)
(** Oprations sur les noeuds **)
(** Operaciones sobre los nudos **)


let rec delete_node_list x = function
  | E -> E
  | N (i, v, j, tail) ->
      if i = x || j = x then delete_node_list x tail
      else N (i, v, j, delete_node_list x tail)

let rec delete_node_rec x data = function
  | n when n < 0 -> ()
  | n ->
      data.(n) <- delete_node_list x data.(n) ;
      delete_node_rec x data (n - 1)

let delete_node = fun x graph -> 
  delete_node_rec x graph.data (graph.size - 1) ;
  graph

let rec delete_node_list_rec data size = function
  | [] -> ()
  | node :: tail ->
      delete_node_rec node data size ;
      delete_node_list_rec data size tail

let delete_node_list = fun list graph ->
  delete_node_list_rec graph.data (graph.size - 1) list;
  graph

let rec successors_list x result = function
  | E -> result
  | N (i, _, j, tail) ->
      match compare x i with
	| n when n < 0 -> successors_list x result tail
	| n when n > 0 -> result
	| _ -> successors_list x (j :: result) tail

let rec successors_rec x result data = function
  | n when n < 0 -> result
  | n -> successors_rec x (successors_list x result data.(n)) data (n - 1)

let successors = fun x graph -> successors_rec x [] graph.data (graph.size - 1)

let rec predecessors_list x result = function
  | E -> result
  | N (i, _, j, tail) ->
      match compare x j with
	| 0 -> predecessors_list x (i :: result) tail
	| _ -> predecessors_list x result tail

let rec predecessors_rec x result data = function
  | n when n < 0 -> result
  | n -> predecessors_rec x (successors_list x result data.(n)) data (n - 1)

let predecessors = fun x graph -> 
  predecessors_rec x [] graph.data (graph.size - 1)

let adjacents = fun x graph -> (predecessors x graph, successors x graph)

let rec exterior_list x result = function
  | E -> result
  | N (i, v, j, tail) ->
      match compare x i with
	| n when n < 0 -> exterior_list x result tail
	| n when n > 0 -> result
	| _ -> exterior_list x ((i, v, j) :: result) tail

let rec exterior_rec x result data = function
  | n when n < 0 -> result
  | n -> exterior_rec x (exterior_list x result data.(n)) data (n - 1)

let exterior_edges = fun x graph -> 
  exterior_rec x [] graph.data (graph.size - 1)

let rec interior_list x result = function
  | E -> result
  | N (i, v, j, tail) ->
      match compare x j with
	| 0 -> interior_list x ((i, v, j) :: result) tail
	| _ -> interior_list x result tail

let rec interior_rec x result data = function
  | n when n < 0 -> result
  | n -> interior_rec x (exterior_list x result data.(n)) data (n - 1)

let interior_edges = fun x graph -> 
  interior_rec x [] graph.data (graph.size - 1)

let adjacent_edges = fun x graph -> 
  (interior_edges x graph, exterior_edges x graph)

let rec exterior_degree_list x total = function
  | E -> total
  | N (i, _, _, tail) ->
      match compare x i with
	| n when n < 0 -> exterior_degree_list x total tail
	| n when n > 0 -> total
	| _ -> exterior_degree_list x (1 + total) tail

let rec exterior_degree_rec x total data = function
  | n when n < 0 -> total
  | n -> 
      exterior_degree_rec x 
      (exterior_degree_list x total data.(n)) data (n - 1)

let exterior_degree = fun x graph -> 
  exterior_degree_rec x 0 graph.data (graph.size - 1)

let rec interior_degree_list x total = function
  | E -> total
  | N (_, _, j, tail) ->
      match compare x j with
	| 0 -> interior_degree_list x (1 + total) tail
	| _ -> interior_degree_list x total tail

let rec interior_degree_rec x total data = function
  | n when n < 0 -> total
  | n -> 
      interior_degree_rec x 
      (exterior_degree_list x total data.(n)) data (n - 1)

let interior_degree = fun x graph -> 
  interior_degree_rec x 0 graph.data (graph.size - 1)

let degree = fun x graph -> (interior_degree x graph, exterior_degree x graph)

let rec member_node_list x = function
  | E -> false
  | N (i, _, j, tail) -> 
      if i = x || j = x then true else member_node_list x tail

let rec member_node_rec x data = function
  | n when n < 0 -> false
  | n -> member_node_list x data.(n) || member_node_rec x data (n - 1)

let member_node = fun x graph -> member_node_rec x graph.data (graph.size - 1)

let exists_node = fun f graph -> List.exists f (nodes_list graph)

let for_all_nodes = fun f graph -> List.for_all f (nodes_list graph)

let iter_nodes = fun f graph -> List.iter f (nodes_list graph)

let fold_right_nodes = fun f graph e -> List.fold_right f (nodes_list graph) e

let fold_nodes = fold_right_nodes

let fold_left_nodes = fun f e graph -> List.fold_left f e (nodes_list graph)


(** Operations on edges **)
(** Operations sur les arcs **)
(** Operaciones sobre los arcos **)


let insert_edge = fun i v j graph ->
  insert i v j graph.data graph.size ;
  graph

let rec insert_edge_list = fun list graph ->
  insert_list graph.size graph.data list ;
  graph

let rec member_rec data i j = function
  | n when n < 0 -> false
  | n -> member_list i j data.(n) || member_rec data i j (n - 1)

let rec member_edge = fun i j graph -> 
  member_rec graph.data i j (graph.size - 1)

let rec look_for_rec f data = function
  | n when n < 0 -> raise Not_found
  | n ->
      try look_for_list f data.(n) 
      with Not_found -> look_for_rec f data (n - 1)

let look_for_edge = fun f graph -> look_for_rec f graph.data (graph.size - 1)

let rec choose_rec data = function
  | n when n < 0 -> raise Empty
  | n ->
      match data.(n) with
	| E -> choose_rec data (n - 1)
	| N (i, v, j, _) -> (i, v, j)

let choose_edge = function graph -> choose_rec graph.data (graph.size - 1)

