(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Non determinist graphs
 * Graphes non deterministes
 * Grafos no deterministas
 *)

open BairePervasives
open StreamHeader

let moduleName = "ListRangeGraph"

(** Non determinist graphs implemented as range AVL trees
  This data structure is equivalent to
  type 'a graph = (node, ('a, node list) map) map
*)

(* Graphes non dterminstes reprsents par des arbres AVL
 * Cette structure de donnes est quivalente 
 * type 'a graph = (node, ('a, node list) map) map
 *)

(* Grafos no deterministas representados con rboles AVL
 * Esta estructura de datos es equivalente a
 * type 'a graph = (node, ('a, node list) map) map
 *)

type node = int
type 'a two = ETwo | NTwo of 'a two * 'a * node list * 'a two * int
type 'a graph = E | N of 'a graph * node * 'a graph * 'a two * int


(** Utils *)
(** Utilitaires *)
(** tiles *)


let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type

let makeTree = fun l v r s -> 
  let h = match (l, r) with
    | (E, E) -> 1
    | (E, N (_, _, _, _, n)) | (N (_, _, _, _, n), E) -> 1 + n
    | (N (_, _, _, _, m), N (_, _, _, _, n)) -> 1 + max m n
  in
    N (l, v, r, s, h)

let rotate_L = fun l v r s ->
  match r with
    | N (rl, rv, rr, rs, _) -> makeTree (makeTree l v rl s) rv rr rs
    | _ -> makeTree l v r s
	
let rotate_R = fun l v r s ->
  match l with
    | N (ll, lv, lr, ls, _) -> makeTree ll lv (makeTree lr v r s) ls
    | _ -> makeTree l v r s

let rotate_RL = fun l v r s ->
  match r with
    | N (N (rll, rlv, rlr, rls, _), rv, rr, rs, _) -> 
	makeTree (makeTree l v rll s) rlv (makeTree rlr rv rr rs) rls
    | _ -> makeTree l v r s

let rotate_LR = fun l v r s ->
  match l with
    | N (ll, lv, N (lrl, lrv, lrr, lrs, _), ls, _) ->
	makeTree (makeTree ll lv lrl ls) lrv (makeTree lrr v r s) lrs
    | _ -> makeTree l v r s

let height = function
  | E -> 0
  | N (_, _, _, _, n) -> n

let makeBTree = fun l v r s ->
  let hl = height l and hr = height r in
    if hl > hr + 1 then
      (match l with
	 | E -> invalid "makeBTree" UnexpectedPattern
         | N (ll, _, lr, _, _) ->
             if height ll >= height lr then rotate_R l v r s
	     else rotate_LR l v r s
      )
    else if hr > hl + 1 then
      (match r with
	 | E -> invalid "makeBTree" UnexpectedPattern
         | N (rl, _, rr, _, _) ->
             if height rr >= height rl then rotate_L l v r s
	     else rotate_RL l v r s
      )
    else makeTree l v r s
      
let rec makeSTree = fun l v r s ->
  let hl = height l and hr = height r in
    if hl > hr + 2 then
      (match l with
	 | E -> invalid "makeSTree" UnexpectedPattern
	 | N (ll, lv, lr, ls, _) -> makeBTree ll lv (makeSTree lr v r s) ls
      )
    else if hr > hl + 2 then
      (match r with
	 | E -> invalid "makeSTree" UnexpectedPattern
	 | N (rl, rv, rr, rs, _) -> makeBTree (makeSTree l v rl s) rv rr rs
      )
    else makeBTree l v r s

let rec append = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l1, v1, r1, s1, _), N (l2, v2, r2, s2, _)) ->
        makeBTree l1 v1 (makeBTree (append r1 l2) v2 r2 s2) s1

let rec appendS = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l1, v1, r1, s1, _), N (l2, v2, r2, s2, _)) ->
        makeSTree l1 v1 (makeSTree (appendS r1 l2) v2 r2 s2) s1

let makeTwo = fun l k v r -> 
  let h = match (l, r) with
    | (ETwo, ETwo) -> 1
    | (ETwo, NTwo (_, _, _, _, n)) | (NTwo (_, _, _, _, n), ETwo) -> 1 + n
    | (NTwo (_, _, _, _, m), NTwo (_, _, _, _, n)) -> 1 + max m n
  in
    NTwo (l, k, v, r, h)

let rotate_L_two = fun l k v r ->
  match r with
    | NTwo (rl, rk, rv, rr, _) -> makeTwo (makeTwo l k v rl) rk rv rr
    | _ -> makeTwo l k v r
	
let rotate_R_two = fun l k v r ->
  match l with
    | NTwo (ll, lk, lv, lr, _) -> makeTwo ll lk lv (makeTwo lr k v r)
    | _ -> makeTwo l k v r

let rotate_RL_two = fun l k v r ->
  match r with
    | NTwo (NTwo (rll, rlk, rlv, rlr, _), rk, rv, rr, _) -> 
	makeTwo (makeTwo l k v rll) rlk rlv (makeTwo rlr rk rv rr)
    | _ -> makeTwo l k v r

let rotate_LR_two = fun l k v r ->
  match l with
    | NTwo (ll, lk, lv, NTwo (lrl, lrk, lrv, lrr, _), _) ->
	makeTwo (makeTwo ll lk lv lrl) lrk lrv (makeTwo lrr k v r)
    | _ -> makeTwo l k v r

let height_two = function
  | ETwo -> 0
  | NTwo (_, _, _, _, n) -> n

let makeBTwo = fun l k v r ->
  let hl = height_two l and hr = height_two r in
    if hl > hr + 1 then
      (match l with
	 | ETwo -> invalid "makeBTwo" UnexpectedPattern
         | NTwo (ll, _, _, lr, _) ->
             if height_two ll >= height_two lr then rotate_R_two l k v r 
	     else rotate_LR_two l k v r
      )
    else if hr > hl + 1 then
      (match r with
	 | ETwo -> invalid "makeBTwo" UnexpectedPattern
         | NTwo (rl, _, _, rr, _) ->
             if height_two rr >= height_two rl then rotate_L_two l k v r 
	     else rotate_RL_two l k v r
      )
    else makeTwo l k v r
      
let rec makeSTwo = fun l k v r ->
  let hl = height_two l and hr = height_two r in
    if hl > hr + 2 then
      (match l with
	 | ETwo -> invalid "makeSTwo" UnexpectedPattern
	 | NTwo (ll, lk, lv, lr, _) -> makeBTwo ll lk lv (makeSTwo lr k v r)
      )
    else if hr > hl + 2 then
      (match r with
	 | ETwo -> invalid "makeSTwo" UnexpectedPattern
	 | NTwo (rl, rk, rv, rr, _) -> makeBTwo (makeSTwo l k v rl) rk rv rr
      )
    else makeBTwo l k v r

let rec append_two = fun a b ->
  match (a, b) with
    | (ETwo, _) -> b
    | (_, ETwo) -> a
    | (NTwo (l1, k1, v1, r1, _), NTwo (l2, k2, v2, r2, _)) ->
        makeBTwo l1 k1 v1 (makeBTwo (append_two r1 l2) k2 v2 r2)

let rec appendS_two = fun a b ->
  match (a, b) with
    | (ETwo, _) -> b
    | (_, ETwo) -> a
    | (NTwo (l1, k1, v1, r1, _), NTwo (l2, k2, v2, r2, _)) ->
        makeSTwo l1 k1 v1 (makeSTwo (appendS_two r1 l2) k2 v2 r2)

let rec map_tree f = function
  | E -> E
  | N (l, k, r, s, _) -> makeTree (map_tree f l) k (map_tree f r) (f s)


(** Categorial operations *)
(** Oprations catgorielles *)
(** Operaciones categoriales *)


let injection = fun x v y -> makeTree E x E (makeTwo ETwo v [y] ETwo)

let zero = function () -> E

let rec to_list_two x result = function
  | ETwo -> result
  | NTwo (l, v, y, r, _) -> 
      to_list_two x (
	List.fold_right 
		       (fun y result -> (x, v, y) :: result) y 
		       (to_list_two x result r)
      ) l

let rec to_list_one result = function
  | E -> result
  | N (l, k, r, s, _) -> 
      to_list_one (to_list_two k (to_list_one result r) s) l 

let to_list = function tree -> to_list_one [] tree

let rec insert_two v y = function
  | ETwo -> makeTwo ETwo v [y] ETwo
  | NTwo (l, v', y', r, _) ->
      match compare v v' with
	| n when n < 0 -> makeBTwo (insert_two v y l) v' y' r
	| n when n > 0 -> makeBTwo l v' y' (insert_two v y r)
	| _ -> makeBTwo l v' (OrderedListSet.insert y y') r

let rec insert_raise x v y = function 
  | E -> raise Not_found
  | N (l, k, r, s, _) ->
      match compare x k with
	 | n when n < 0 -> makeTree (insert_raise x v y l) k r s
	 | n when n > 0 -> makeTree l k (insert_raise x v y r) s
	 | _ -> makeTree l k r (insert_two v y s)

let insert = fun x v y tree ->
  try 
    insert_raise x v y tree 
  with Found -> tree

let rec insert_node_raise id = function
  | E -> makeTree E id E ETwo
  | N (l, v, r, s, _) ->
      match compare id v with
	| n when n < 0 -> makeBTree (insert_node_raise id l) v r s
	| n when n > 0 -> makeBTree l v (insert_node_raise id r) s
	| _ -> raise Found

let insert_node = fun id graph ->
  try
    insert_node_raise id graph
  with Found -> graph

let rec from_list_rec result = function
  | [] -> result
  | (i, v, j) :: tail -> 
      from_list_rec (insert i v j (insert_node i (insert_node j result))) tail

let from_list = function list -> from_list_rec E list

let map f = function tree -> from_list (List.map f (to_list tree)) 


(** Binary operations *)
(** Oprations binaires *)
(** Operaciones binrias *)


(** Observateurs *)
(** Observers *)
(** Observadores *)


let is_empty = function
  | E -> true
  | N _ -> false

let rec nodes_rec total = function
  | E -> total
  | N (l, _, r, _, _) -> nodes_rec (nodes_rec (1 + total) l) r

let nodes = function graph -> nodes_rec 0 graph

let rec size_two total = function
  | ETwo -> total
  | NTwo (l, _, y, r, _) -> 
      size_two (size_two (List.length y + total) l) r

let rec size_one total = function
  | E -> total
  | N (l, _, r, s, _) -> size_one (size_one (size_two total s) l) r

let edges = function graph -> size_one 0 graph

let rec labels_two result = function
  | ETwo -> result
  | NTwo (l, v, _, r, _) -> labels_two (labels_two (v :: result) l) r

let rec labels_one result = function
  | E -> result
  | N (l, _, r, s, _) -> labels_one (labels_one (labels_two result s) l) r

let labels_list = function graph -> labels_one [] graph

let rec nodes_one result = function
  | E -> result
  | N (l, k, r, _, _) -> nodes_one (nodes_one (k :: result) l) r

let rec to_list_tree result = function
  | Leaf -> result
  | Node (l, v, r) -> to_list_tree (to_list_tree (v :: result) l) r

let rec nodes_two result = function
  | ETwo -> result
  | NTwo (l, _, k, r, _) -> 
      nodes_two (
	nodes_two (List.rev_append k result) l
      ) r

let nodes_list = function graph -> nodes_one [] graph

let rec edges_rec result = function
  | E -> result
  | N (l, v, r, s, _) ->
      edges_rec (
	edges_rec (
	  to_list_two v result s
	) l
      ) r

let edges_list = function graph -> edges_rec [] graph

let rec member_two y = function
  | ETwo -> false
  | NTwo (l, _, k, r, _) -> 
      OrderedListSet.member y k || member_two y l || member_two y r

let rec interior_degree_rec total x = function
  | E -> total
  | N (l, v, r, s, _) ->
      interior_degree_rec (
	interior_degree_rec (
	  if member_two x s then (1 + total) else total
	) x l
      ) x r
			     
let interior_degree = fun id graph -> interior_degree_rec 0 id graph

let degrees_list = function graph ->
  let rec degrees_rec result = function
    | E -> raise Not_found
    | N (l, v, r, s, _) ->
	degrees_rec (
	  degrees_rec (
	    (v, interior_degree_rec 0 v graph, size_two 0 s) :: result
	  ) l
	) r
  in
    degrees_rec [] graph
  
let rec labels_two result = function
  | ETwo -> result
  | NTwo (l, v, _, r, _) ->
      labels_two (labels_two (UnbalancedSet.insert v result) l) r

let rec labels_one result = function
  | E -> result
  | N (l, _, r, s, _) -> labels_one (labels_one (labels_two result s) l) r

let labels_tree = function graph -> labels_one Leaf graph

let rec nodes_tree = function
  | E -> Leaf
  | N (l, v, r, _, _) -> Node (nodes_tree l, v, nodes_tree r)

let rec edges_tree_two x result = function
  | ETwo -> result
  | NTwo (l, v, y, r, _) ->
      edges_tree_two x (
	edges_tree_two x (
	  List.fold_right (
	    fun y l -> UnbalancedSet.insert (x, y, v) l) y result
	) l
      ) r

let rec edges_tree_one result = function
  | E -> result
  | N (l, v, r, s, _) ->
      edges_tree_one (
	edges_tree_one (
	  edges_tree_two v result s
	) l
      ) r
  
let edges_tree = function graph -> edges_tree_one Leaf graph

let degrees_tree = function graph ->
  let rec degrees_rec result = function
    | E -> raise Not_found
    | N (l, v, r, s, _) ->
	degrees_rec (
	  degrees_rec (
	    UnbalancedSet.insert 
			 (v, interior_degree_rec 0 v graph, size_two 0 s)
			 result
	  ) l
	) r
  in
    degrees_rec Leaf graph


(** Miscellaneous *)
(** Divers *)
(** Micelaneos *)


let rec lookup_two v = function
  | ETwo -> raise Not_found
  | NTwo (l, v', k', r, _) ->
      match compare v v' with
	| n when n < 0 -> lookup_two v l
	| n when n > 0 -> lookup_two v r
	| _ -> k'

let rec next x v = function
  | E -> raise Not_found
  | N (l, k, r, s, _) ->
      match compare x k with
	| n when n < 0 -> next x v l
	| n when n > 0 -> next x v r
	| _ -> lookup_two v s

let rec lookup_two' v = function
  | ETwo -> []
  | NTwo (l, v', k', r, _) ->
      match compare v v' with
	| n when n < 0 -> lookup_two' v l
	| n when n > 0 -> lookup_two' v r
	| _ -> k'

let rec next' x v = function
  | E -> []
  | N (l, k, r, s, _) ->
      match compare x k with
	| n when n < 0 -> next' x v l
	| n when n > 0 -> next' x v r
	| _ -> lookup_two' v s

let rec zip = fun symbols_list tail stack ->
  match tail with
    | [] -> stack
    | node :: node_list -> 
	zip symbols_list node_list ((node, symbols_list) :: stack)

let rec next_first = fun stack graph ->
  match stack with
    | [] -> []
    | (state, symbol_list) :: stack_tail -> 
	next_list_rec state stack_tail symbol_list graph

and next_list_rec = fun state stack symbol_list graph ->
  match symbol_list with
    | [] -> state :: (next_first stack graph)
    | v :: symbol_tail ->
	match next' state v graph with
	  | [] -> next_first stack graph
	  | [x] -> next_list_rec x stack symbol_tail graph
	  | node :: node_list -> 
	      next_list_rec node 
	      (zip symbol_tail node_list stack) symbol_tail graph

let next_list_df = fun node list graph -> next_list_rec node [] list graph

let rec next_first_lazy = fun stack graph ->
  match stack with
    | [] -> Nil
    | (state, symbol_list) :: stack_tail -> 
	next_list_lazy_rec state stack_tail symbol_list graph

and next_list_lazy_rec = fun state stack symbol_list graph ->
  match symbol_list with
    | [] -> Cons (state, lazy (next_first_lazy stack graph))
    | v :: symbol_tail ->
	match next' state v graph with
	  | [] -> next_first_lazy stack graph
	  | [x] -> next_list_lazy_rec x stack symbol_tail graph
	  | node :: node_list -> 
	      next_list_lazy_rec node 
	      (zip symbol_tail node_list stack) symbol_tail graph

let next_list_df_lazy = fun node list graph -> 
  next_list_lazy_rec node [] list graph

let rec next_first = fun head tail graph ->
  match head with
    | [] -> 
	(match tail with
	   | [] -> []
	   | _ -> next_first (List.rev tail) [] graph
	)
    | (state, symbol_list) :: stack_tail ->
	next_list_rec state stack_tail tail symbol_list graph

and next_list_rec = fun state head tail symbol_list graph ->
  match symbol_list with
    | [] -> state :: (next_first head tail graph)
    | v :: symbol_tail ->
	match next' state v graph with
	  | [] -> next_first head tail graph
	  | [x] -> next_list_rec x head tail symbol_tail graph
	  | node :: node_list -> 
	      next_list_rec node 
	      (zip symbol_tail node_list head) tail symbol_tail graph

let next_list_bf = fun node list graph -> next_list_rec node [] [] list graph

let rec next_first_lazy = fun head tail graph ->
  match head with
    | [] -> 
	(match tail with
	   | [] -> Nil
	   | _ -> next_first_lazy (List.rev tail) [] graph
	)
    | (state, symbol_list) :: stack_tail ->
	next_list_lazy_rec state stack_tail tail symbol_list graph

and next_list_lazy_rec = fun state head tail symbol_list graph ->
  match symbol_list with
    | [] -> Cons (state, lazy (next_first_lazy head tail graph))
    | v :: symbol_tail ->
	match next' state v graph with
	  | [] -> next_first_lazy head tail graph
	  | [x] -> next_list_lazy_rec x head tail symbol_tail graph
	  | node :: node_list -> 
	      next_list_lazy_rec node 
	      (zip symbol_tail node_list head) tail symbol_tail graph

let next_list_bf_lazy = fun node list graph -> 
  next_list_lazy_rec node [] [] list graph

(** Operations on node *)
(** Oprations sur les noeuds *)
(** Operaciones sobre los nudos *)


let rec member_node x = function
  | E -> false
  | N (l, v, r, _, _) ->
      match compare x v with
	| n when n < 0 -> member_node x l
	| n when n > 0 -> member_node x r
	| _ -> true

let rec new_node_rec = fun n graph ->
  if member_node n graph then new_node_rec (n + 1) graph
  else n

let new_node = function graph -> new_node_rec 0 graph

let rec new_nodes_rec n graph = function
  | 0 -> []
  | k when member_node n graph -> new_nodes_rec (n + 1) graph k
  | k -> n :: new_nodes_rec (n + 1) graph (k - 1)

let new_nodes = fun n graph -> new_nodes_rec 0 graph n

let rec insert_node_list = fun list graph ->
  match list with
    | [] -> graph
    | node :: tail -> insert_node_list tail (insert_node node graph)

let rec delete_two_raise y = function
  | ETwo -> raise Not_found
  | NTwo (l, v, k, r, _) ->
      match OrderedListSet.delete_raise y k with
	| [] -> append_two l r
	| k' -> makeBTwo (delete_two_raise y l) v k' (delete_two_raise y r)

let delete_two = fun y tree ->
  try
    delete_two_raise y tree
  with Not_found -> tree

let rec delete_node_raise id = function
  | E -> raise Not_found
  | N (l, v, r, s, _) ->
      match compare id v with
	| n when n < 0 -> 
	    makeBTree (delete_node_raise id l) v 
	    (map_tree (delete_two id) r) (delete_two id s)
	| n when n > 0 ->
	    makeBTree (map_tree (delete_two id) l) v 
	    (delete_node_raise id r) (delete_two id s)
	| _ -> append (map_tree (delete_two id) l) (map_tree (delete_two id) r)

let delete_node = fun id graph ->
  try
    delete_node_raise id graph
  with Not_found -> graph

let rec delete_node_unsafe_raise id = function
  | E -> raise Not_found
  | N (l, v, r, s, _) ->
      match compare id v with
	| n when n < 0 -> makeBTree (delete_node_unsafe_raise id l) v r s
	| n when n > 0 -> makeBTree l v (delete_node_unsafe_raise id r) s
	| _ -> append l r

let delete_node_unsafe = fun id graph ->
  try
    delete_node_unsafe_raise id graph
  with Not_found -> graph

let rec delete_node_list = fun list graph ->
  match list with
    | [] -> graph
    | node :: tail -> delete_node_list tail (delete_node node graph)

let rec successors_rec result x = function
  | E -> raise Not_found
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> successors_rec result x l
	| n when n > 0 -> successors_rec result x r
	| _ -> nodes_two result s

let rec successors x = function
  | E -> raise Not_found
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> successors x l
	| n when n > 0 -> successors x r
	| _ -> nodes_two [] s

let rec predecessors_rec result x = function
  | E -> result
  | N (l, v, r, s, _) ->
      predecessors_rec (
	predecessors_rec (
	  if member_two x s then v :: result else result
	) x l
      ) x r

let predecessors = fun x graph -> predecessors_rec [] x graph

let rec adjacents_rec (pred, succ) x = function
  | E -> (pred, succ)
  | N (l, v, r, s, _) ->
      adjacents_rec (
	adjacents_rec (
	  ((if member_two x s then v :: pred else pred), 
	   if x = v then nodes_two succ s else succ)
	) x l
      ) x r

let adjacents = fun x graph -> adjacents_rec ([], []) x graph

let rec exterior_edges_rec result x = function
  | E -> raise Not_found
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> exterior_edges_rec result x l
	| n when n > 0 -> exterior_edges_rec result x r
	| _ -> to_list_two x result s

let rec exterior_edges x = function
  | E -> raise Not_found
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> exterior_edges x l
	| n when n > 0 -> exterior_edges x r
	| _ -> to_list_two x [] s

let rec inverse_lookup y = function
  | ETwo -> raise Not_found
  | NTwo (_, v, k, _, _) when OrderedListSet.member y k -> v
  | NTwo (l, _, _, r, _) ->
      try inverse_lookup y l 
      with Not_found -> inverse_lookup y r

let rec interior_edges_rec result x = function
  | E -> result
  | N (l, v, r, s, _) ->
      interior_edges_rec (
	interior_edges_rec (
	  try
	    (v, inverse_lookup x s, x) :: result
	  with Not_found -> result
	) x l
      ) x r

let interior_edges = fun x graph -> interior_edges_rec [] x graph

let rec adjacent_edges_rec (pred, succ) x = function
  | E -> (pred, succ)
  | N (l, v, r, s, _) ->
      adjacent_edges_rec (
	adjacent_edges_rec (
	  ((try 
	      (v, inverse_lookup x s, x) :: pred 
	    with Not_found -> pred), 
	   if x = v then to_list_two x succ s else succ)
	) x l
      ) x r

let adjacent_edges = fun x graph -> adjacent_edges_rec ([], []) x graph

let rec exterior_degree x = function
  | E -> raise Not_found
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> exterior_degree x l
	| n when n > 0 -> exterior_degree x r
	| _ -> size_two 0 s

let rec interior_degree_rec total x = function
  | E -> total
  | N (l, v, r, s, _) ->
      interior_degree_rec (
	interior_degree_rec (
	  if member_two x s then (1 + total) else total
	) x l
      ) x r
			     
let interior_degree = fun id graph -> interior_degree_rec 0 id graph

let degree = fun x graph -> (interior_degree x graph, exterior_degree x graph)

let rec look_for_node f = function
  | E -> raise Not_found
  | N (_, v, _, _, _) when f v -> v
  | N (l, _, r, _, _) -> 
      try look_for_node f l 
      with Not_found -> look_for_node f r

let choose_node = function
  | E -> raise Empty
  | N (_, v, _, _, _) -> v

let rec min_node = function
  | E -> raise Not_found
  | N (E, v, _, _, _) -> v
  | N (l, _, _, _, _) -> min_node l

let rec max_node = function
  | E -> raise Not_found
  | N (_, v, E, _, _) -> v
  | N (_, _, r, _, _) -> max_node r

let rec next_node x = function
  | E -> raise Not_found
  | N (l, v, r, _, _) ->
      match compare x v with
	| n when n < 0 -> (try next_node x l with Not_found -> v)
	| n when n > 0 -> next_node x r
	| _ -> min_node r (* [Not_found] *)

let rec previous_node x = function
  | E -> raise Not_found
  | N (l, v, r, _, _) ->
      match compare x v with
	| n when n < 0 -> previous_node x l
	| n when n > 0 -> (try previous_node x r with Not_found -> v)
	| _ -> max_node l (* [Not_found] *)
	    
let rec ceil_node x = function
  | E -> raise Not_found
  | N (l, v, r, _, _) ->
      match compare x v with
	| n when n < 0 -> (try ceil_node x l with Not_found -> v)
	| n when n > 0 -> ceil_node x r
	| _ -> x

let rec floor_node x = function
  | E -> raise Not_found
  | N (l, v, r, _, _) ->
      match compare x v with
	| n when n < 0 -> floor_node x l
	| n when n > 0 -> (try floor_node x r with Not_found -> v)
	| _ -> x

let rec for_all_nodes f = function
  | E -> true
  | N (l, v, r, _, _) -> f v && for_all_nodes f l && for_all_nodes f r

let rec exists_node f = function
  | E -> false
  | N (l, v, r, _, _) -> f v || exists_node f l || exists_node f r

let rec iter_nodes f = function
  | E -> ()
  | N (l, v, r, _, _) -> iter_nodes f l ; f v ; iter_nodes f r

let rec fold_left_nodes f e = function
  | E -> e
  | N (l, v, r, _, _) -> fold_left_nodes f (f (fold_left_nodes f e l) v) r

let rec fold_right_nodes = fun f set e ->
  match set with
    | E -> e 
    | N (l, v, r, _, _) -> fold_right_nodes f l (f v (fold_right_nodes f r e))

let fold_nodes = fold_right_nodes


(** Operations on edges *)
(** Operations sur les arcs *)
(** Operaciones sobre los arcos *)


let insert_edge = insert

let insert_edge_raise = insert_raise

let insert_edge_list = fun list graph -> from_list_rec graph list

let rec delete_edge_raise x y = function
  | E -> raise Not_found
  | N (l, k, r, s, _) ->
      match compare x k with
	| n when n < 0 -> makeTree (delete_edge_raise x y l) k r s
	| n when n > 0 -> makeTree l k (delete_edge_raise x y r) s
	| _ -> makeTree l k r (delete_two_raise y s)

let delete_edge = fun x y tree ->
  try 
    delete_edge_raise x y tree
  with Not_found -> tree

let rec delete_edge_list = fun list set ->
  match list with
    | [] -> set
    | (x, y) :: tail -> delete_edge_list tail (delete_edge x y set)

let rec edge_label x y = function
  | E -> raise Not_found
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> edge_label x y l
	| n when n > 0 -> edge_label x y r
	| _ -> inverse_lookup y s

let rec member_edge x y = function
  | E -> false
  | N (l, k, r, s, _) ->
      match compare x k with
	| n when n < 0 -> member_edge x y l
	| n when n > 0 -> member_edge x y r
	| _ -> member_two y s

let rec look_for_two f = function
  | ETwo -> raise Not_found
  | NTwo (l, v, k, r, _) ->
      try (v, OrderedListSet.look_for (f v) k)
      with Not_found -> try look_for_two f l 
      with Not_found -> look_for_two f r

let rec look_for_edge f = function
  | E -> raise Not_found
  | N (l, k, r, s, _) ->
      try let (v, k') = look_for_two (f k) s in (k, v, k')
      with Not_found ->
	try look_for_edge f l 
	with Not_found -> look_for_edge f r
	  
let choose_edge = function graph -> look_for_edge (fun i v j -> true) graph

(* k cannot be the empty list *)

let rec min_element_two = function
  | ETwo -> raise Not_found
  | NTwo (ETwo, v, k, _, _) -> (List.hd k, v)
  | NTwo (l, _, _, _, _) -> min_element_two l

let rec min_edge = function
  | E -> raise Not_found
  | N (E, x, _, s, _) -> let (y, v) = min_element_two s in (x, v, y)
  | N (l, _, _, _, _) -> min_edge l

(* k cannot be the empty list *)

let rec max_element_two = function
  | ETwo -> raise Not_found
  | NTwo (_, v, k, ETwo, _) -> (OrderedListSet.max_element k, v)
  | NTwo (_, _, _, r, _) -> max_element_two r

let rec max_edge = function
  | E -> raise Not_found
  | N (_, x, E, s, _) -> let (y, v) = max_element_two s in (x, v, y)
  | N (_, _, l, _, _) -> max_edge l

let rec exists_two f = function
  | ETwo -> false
  | NTwo (l, v, y, r, _) -> 
      OrderedListSet.exists (f v) y || exists_two f l || exists_two f r

let rec exists_edge f = function
  | E -> false
  | N (l, x, r, s, _) -> 
      exists_two (f x) s || exists_edge f l || exists_edge f r

let rec for_all_two f = function
  | ETwo -> true
  | NTwo (l, v, y, r, _) -> 
      OrderedListSet.for_all (f v) y && for_all_two f l && for_all_two f r

let rec for_all_edges f = function
  | E -> true
  | N (l, x, r, s, _) -> 
      for_all_two (f x) s && for_all_edges f l && for_all_edges f r

let rec iter_two f = function
  | ETwo -> ()
  | NTwo (l, v, y, r, _) -> 
      iter_two f l ; 
      OrderedListSet.iter (f v) y ; 
      iter_two f r

let rec iter_edges f = function
  | E -> ()
  | N (l, x, r, s, _) -> iter_edges f l ; iter_two (f x) s ; iter_edges f r

let rec fold_left_two f e = function
  | ETwo -> e
  | NTwo (l, v, k, r, _) -> 
      fold_left_two f (
	OrderedListSet.fold_left (fun e -> f e v)(fold_left_two f e l) k 
      ) r

let rec fold_left_edges f e = function
  | E -> e
  | N (l, x, r, s, _) -> 
      fold_left_edges 
      f (fold_left_two (fun e a y -> f e x a y) (fold_left_edges f e l) s) r

let rec fold_right_two = fun f set e ->
  match set with
    | ETwo -> e 
    | NTwo (l, v, k, r, _) -> 
	fold_right_two f l (
	  List.fold_right (f v) k (fold_right_two f r e)
	)

let rec fold_right_edges = fun f set e ->
  match set with
    | E -> e
    | N (l, x, r, s, _) -> 
	fold_right_edges f l (fold_right_two (f x) s (fold_right_edges f r e))

let fold_edges = fold_right_edges


(** Views *)
(** Vues *)
(** Vistas *)



(** Graph traversals *)
(** Parcours de graphes *)
(** Recorridos de grafos *)


let rec dfs f e graph = function
  | [] -> e
  | x :: tail ->
      let reduced = delete_node_unsafe x graph in
	try
	  dfs f (f x e) reduced (successors_rec tail x graph)
	with Not_found -> dfs f e reduced tail

let rec dfs' f e graph = function
  | [] -> e
  | (i, v, j) :: tail ->
      let reduced = delete_node_unsafe j graph in
	try 
	  dfs' f (f i v j e) reduced (exterior_edges_rec tail j graph)
	with Not_found -> dfs' f e reduced tail
	  
let depth_first_fold = fun f e graph x -> dfs f e graph [x]

let depth_first_fold' = fun f e graph x -> 
  dfs' f e graph (exterior_edges x graph)

let rec dfs_list result graph = function
  | [] -> result
  | x :: tail ->
      let reduced = delete_node_unsafe x graph in
	try
	  dfs_list (x :: result) reduced (successors_rec tail x graph)
	with Not_found -> dfs_list result reduced tail
	  
let rec dfs_list' result graph = function
  | [] -> result
  | (_, _, j) as head :: tail ->
      let reduced = delete_node_unsafe j graph in
	try
	  dfs_list' (head :: result) reduced (exterior_edges_rec tail j graph)
	with Not_found -> dfs_list' result reduced tail

let depth_first_list = fun graph x -> List.rev (dfs_list [] graph [x])

let depth_first_list' = fun graph x -> 
  List.rev (dfs_list' [] 
	      (delete_node_unsafe x graph) 
	      (List.rev (exterior_edges x graph))
	   )

let rec 
  bfs_notail f e graph = function
    | [] -> e
    | x :: tail ->
	let reduced = delete_node_unsafe x graph in
	  try
	    bfs_tail f (f x e) reduced tail (successors x graph)
	  with Not_found -> bfs_notail f e reduced tail
and
  bfs_tail f e = fun graph hd tl ->
    match hd with
      | [] ->
	  (match tl with
	     | [] -> e
	     | _ -> bfs_notail f e graph (List.rev tl)
	  )
      | x :: tail ->
	  let reduced = delete_node_unsafe x graph in
	    try
	      bfs_tail f (f x e) reduced tail (successors_rec tl x graph)
	    with Not_found -> bfs_tail f e reduced tail tl

let rec 
  bfs_notail' f e graph = function
    | [] -> e
    | (i, v, j) :: tail ->
	let reduced = delete_node_unsafe j graph in
	  try
	    bfs_tail' f (f i v j e) reduced tail (exterior_edges j graph)
	  with Not_found -> bfs_notail' f e reduced tail
and
  bfs_tail' f e = fun graph hd tl ->
    match hd with
      | [] ->
	  (match tl with
	     | [] -> e
	     | _ -> bfs_notail' f e graph (List.rev tl)
	  )
      | (i, v, j) :: tail ->
	  let reduced = delete_node_unsafe j graph in
	    try
	      bfs_tail' f (f i v j e) reduced tail 
		(exterior_edges_rec tl j graph)
	    with Not_found -> bfs_tail' f e reduced tail tl

let breadth_first_fold = fun f e graph x -> bfs_notail f e graph [x]

let breadth_first_fold' = fun f e graph x -> 
  bfs_notail' f e (delete_node_unsafe x graph) (exterior_edges x graph)

let rec 
  bfs_list_notail result graph = function
    | [] -> result
    | x :: tail ->
	let reduced = delete_node_unsafe x graph in
	  try
	    bfs_list_tail (x :: result) reduced tail (successors x graph)
	  with Not_found -> bfs_list_notail result reduced tail
and
  bfs_list_tail result = fun graph hd tl ->
    match hd with
      | [] ->
	  (match tl with
	     | [] -> result
	     | _ -> bfs_list_notail result graph (List.rev tl)
	  )
      | x :: tail ->
	  let reduced = delete_node_unsafe x graph in
	    try
	      bfs_list_tail (x :: result) reduced tail 
		(successors_rec tl x graph)
	    with Not_found -> bfs_list_tail result reduced tail tl

let rec 
  bfs_list_notail' result graph = function
    | [] -> result
    | (_, _, j) as head :: tail ->
	let reduced = delete_node_unsafe j graph in
	  try
	    bfs_list_tail' (head :: result) reduced tail 
	      (exterior_edges j graph)
	  with Not_found -> bfs_list_notail' result reduced tail
and
  bfs_list_tail' result = fun graph hd tl ->
    match hd with
      | [] ->
	  (match tl with
	     | [] -> result
	     | _ -> bfs_list_notail' result graph (List.rev tl)
	  )
      | (_, _, j) as head :: tail ->
	  let reduced = delete_node_unsafe j graph in
	    try
	      bfs_list_tail' (head :: result) reduced tail 
		(exterior_edges_rec tl j graph)
	    with Not_found -> bfs_list_tail' result reduced tail tl

let breadth_first_list = fun graph x -> List.rev (bfs_list_notail [] graph [x])

let breadth_first_list' = fun graph x -> 
  List.rev (bfs_list_notail' [] 
	      (delete_node_unsafe x graph) 
	      (exterior_edges x graph)
	   )

exception Cyclic

let rec topological_rec = function
  | E -> raise Not_found
  | N (_, v, _, ETwo, _) -> v
  | N (l, _, r, _, _) ->
      try topological_rec l with Not_found -> topological_rec r

let rec topological_list_rec result = function
  | E -> result
  | graph ->
      try
	let x = topological_rec graph in
	  topological_list_rec (x :: result) (delete_node x graph)
      with Not_found -> raise Cyclic

let topological_list = function graph -> topological_list_rec [] graph

let rec topological_fold f e = function
  | E -> e
  | graph ->
      try
	let x = topological_rec graph in
	  topological_fold f (f x e) (delete_node x graph)
      with Not_found -> raise Cyclic


(* let compare = fun a b -> compare_rec [a] [b] *)


(** Debug *)



