(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 *
 * Graphes implemented with maps AVL
 * Graphes reprsents par des dictionnaires AVL
 * Grafos representados por diccionrios AVL
 * 
 *)

open BairePervasives

let moduleName = "MapGraph"

type node = int
type 'a edge = node * 'a * node

type 'a tree = E | N of 'a tree * node * 'a * 'a tree * int
type 'a graph = ('a tree * 'a tree) tree


(** Utilitaires **)
(** Utils **)
(** tiles **)


let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type

let makeTree = fun l k v r -> 
  let h = match (l, r) with
    | (E, E) -> 1
    | (E, N (_, _, _ , _, n)) | (N (_, _, _, _, n), E) -> 1 + n
    | (N (_, _, _, _, m), N (_, _, _, _, n)) -> 1 + max m n
  in
    N (l, k, v, r, h)

let rotate_L = fun l k v r ->
  match r with
    | N (rl, rk, rv, rr, _) -> makeTree (makeTree l k v rl) rk rv rr
    | _ -> makeTree l k v r
	
let rotate_R = fun l k v r ->
  match l with
    | N (ll, lk, lv, lr, _) -> makeTree ll lk lv (makeTree lr k v r)
    | _ -> makeTree l k v r

let rotate_RL = fun l k v r ->
  match r with
    | N (N (rll, rlk, rlv, rlr, _), rk, rv, rr, _) -> 
	makeTree (makeTree l k v rll) rlk rlv (makeTree rlr rk rv rr)
    | _ -> makeTree l k v r

let rotate_LR = fun l k v r ->
  match l with
    | N (ll, lk, lv, N (lrl, lrk, lrv, lrr, _), _) ->
	makeTree (makeTree ll lk lv lrl) lrk lrv (makeTree lrr k v r)
    | _ -> makeTree l k v r

let height = function
  | E -> 0
  | N (_, _, _, _, n) -> n

let makeBTree = fun l k v r ->
  let hl = height l and hr = height r in
    if hl > hr + 1 then
      (match l with
	 | E -> invalid "makeBTree" UnexpectedPattern
         | N (ll, _, _, lr, _) ->
             if height ll >= height lr then rotate_R l k v r 
	     else rotate_LR l k v r
      )
    else if hr > hl + 1 then
      (match r with
	 | E -> invalid "makeBTree" UnexpectedPattern
         | N (rl, _, _, rr, _) ->
             if height rr >= height rl then rotate_L l k v r 
	     else rotate_RL l k v r
      )
    else makeTree l k v r

let rec makeSTree = fun l k v r ->
  let hl = height l and hr = height r in
    if hl > hr + 2 then
      (match l with
	 | E -> invalid "makeSTree" UnexpectedPattern
	 | N (ll, lk, lv, lr, _) -> makeBTree ll lk lv (makeSTree lr k v r)
      )
    else if hr > hl + 2 then
      (match r with
	 | E -> invalid "makeSTree" UnexpectedPattern
	 | N (rl, rk, rv, rr, _) -> makeBTree (makeSTree l k v rl) rk rv rr
      )
    else makeBTree l k v r

let rec append = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l1, k1, v1, r1, _), N (l2, k2, v2, r2, _)) ->
        makeBTree l1 k1 v1 (makeBTree (append r1 l2) k2 v2 r2)

let rec appendS = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l1, k1, v1, r1, _), N (l2, k2, v2, r2, _)) ->
        makeSTree l1 k1 v1 (makeSTree (appendS r1 l2) k2 v2 r2)


(** Oprations catgorielles **)
(** Categorial operations **)
(** Operaciones categoriales **)


let injection_tree = fun k v -> makeTree E k v E

let rec insert_raise_tree k v = function
  | E -> injection_tree k v
  | N (l, k', v', r, _) ->
      match compare k k' with
	| n when n < 0 -> makeBTree (insert_raise_tree k v l) k' v' r
	| n when n > 0 -> makeBTree l k' v' (insert_raise_tree k v r)
	| _ when v <> v' -> makeTree l k v r
	| _ -> raise Found

let rec insert_tree k v = function tree ->
  try 
    insert_raise_tree k v tree 
  with Found -> tree

let injection = fun i v j ->
  let (min, max) = if i < j then (i, j) else (j, i) in
    makeTree E min (injection_tree max v, E)
      (makeTree E max (E, injection_tree min v) E)

let rec insert_succ i v j = function
  | E -> raise Not_found
  | N (l, x, ((pred, succ) as k), r, _) ->
      match compare i x with
	| n when n < 0 -> makeTree (insert_succ i v j l) x k r
	| n when n > 0 -> makeTree l x k (insert_succ i v j r)
	| _ -> makeTree l x (pred, insert_tree j v succ) r

let rec insert_pred i v j = function
  | E -> raise Not_found
  | N (l, x, ((pred, succ) as k), r, _) ->
      match compare j x with
	| n when n < 0 -> makeTree (insert_pred i v j l) x k r
	| n when n > 0 -> makeTree l x k (insert_pred i v j r)
	| _ -> makeTree l x (insert_tree i v pred, succ) r

let rec insert_edge_raise i v j = function
  | E -> raise Not_found
  | N (l, x, (pred, succ), r, _) when x = i ->
      if x < j then 
	makeTree l x (pred, insert_tree j v succ) (insert_pred i v j r)
      else if x = j then 
	makeTree l x (insert_tree x v pred, insert_tree x v succ) r
      else (* x > j *)
	makeTree (insert_pred i v j l) x (pred, insert_tree j v succ) r
  | N (l, x, (pred, succ), r, _) when x = j ->
      if x < i then 
	makeTree l x (insert_tree i v pred, succ) (insert_succ i v j r)
      else (* x > i *)
	makeTree (insert_succ i v j l) x (insert_tree i v pred, succ) r
  | N (l, x, k, r, _) ->
      if x < i then
	if x < j then makeTree l x k (insert_edge_raise i v j r)
	else makeTree (insert_pred i v j l) x k (insert_succ i v j r)
      else (* x > i *)
	if x < j then makeTree (insert_succ i v j l) x k (insert_pred i v j r)
	else makeTree (insert_edge_raise i v j l) x k r

let insert_edge = fun i v j graph ->
  try insert_edge_raise i v j graph with
    | Found -> graph
    | Not_found -> failwith "one of the nodes is not in the graph"

let rec insert_node_raise x = function
  | E -> makeTree E x (E, E) E
  | N (l, k, v, r, _) ->
      match compare x k with
	| n when n < 0 -> makeBTree (insert_node_raise x l) k v r
	| n when n > 0 -> makeBTree l k v (insert_node_raise x r)
	| _ -> raise Found      

let insert_node = fun node graph ->
  try
    insert_node_raise node graph
  with Found -> graph

let rec from_list_rec result = function
  | [] -> result
  | (i, v, j) :: tail ->
      from_list_rec
      (insert_edge i v j (insert_node i (insert_node j result)))
      tail

let from_list = function list -> from_list_rec E list

let rec to_list_tree i result = function
  | E -> result
  | N (l, j, v, r, _) ->
      to_list_tree i (
	to_list_tree i (
	  (i, v, j) :: result
	) l
      ) r

let rec to_list_rec result = function
  | E -> result
  | N (l, i, (_, succ), r, _) ->
      to_list_rec (
	to_list_rec (
	  to_list_tree i result succ
	) l
      ) r

let to_list = function graph -> to_list_rec [] graph

let map = fun f graph -> from_list (List.map f (to_list graph))

let rec filter_lt x = function
  | E -> E
  | N (l, k, v, r, _) ->
      match compare x k with
	| n when n < 0 -> filter_lt x l
	| n when n > 0 -> makeSTree l k v (filter_lt x r)
	| _ -> l

let rec filter_gt x = function
  | E -> E
  | N (l, k, v, r, _) ->
      match compare x k with
	| n when n < 0 -> makeSTree (filter_gt x l) k v r
	| n when n > 0 -> filter_gt x r
	| _ -> r

let rec cut = fun lo tree hi ->
  match tree with
    | N (_, k, _, r, _) when compare k lo <= 0 -> cut lo r hi
    | N (l, k, _, _, _) when compare k hi >= 0 -> cut lo l hi
    | _ -> (lo, tree, hi)

let rec cut_lo = fun tree hi ->
  match tree with
    | N (l, k, _, _, _) when compare k hi >= 0 -> cut_lo l hi
    | _ -> (tree, hi)

let rec cut_hi = fun lo tree ->
  match tree with
    | N (_, k, _, r, _) when compare k lo <= 0 -> cut_hi lo r
    | _ -> (lo, tree)

let rec uni = fun a (lo, b, hi) ->
  match (a, b) with
    | (_, E) -> a
    | (E, N (l, k, v, r, _)) -> makeSTree (filter_gt lo l) k v (filter_lt hi r)
    | (N (l, k, v, r, _), N _) ->
	makeSTree (uni l (cut lo b k)) k v (uni r (cut k b hi))  

let rec uni_lo = fun a (b, hi) ->
  match (a, b) with
    | (_, E) -> a
    | (E, N (l, k, v, r, _)) -> makeSTree l k v (filter_lt hi r)
    | (N (l, k, v, r, _), N _) ->
	makeSTree (uni_lo l (cut_lo b k)) k v (uni r (cut k b hi))

let rec uni_hi = fun a (lo, b) ->
  match (a, b) with
    | (_, E) -> a
    | (E, N (l, k, v, r, _)) -> makeSTree (filter_gt lo l) k v r
    | (N (l, k, v, r, _), N _) -> 
	makeSTree (uni l (cut lo b k)) k v (uni_hi r (cut_hi k b)) 

let plus_tree = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l, k, v, r, _), N _) ->
	makeSTree (uni_lo l (cut_lo b k)) k v (uni_hi r (cut_hi k b)) 

let rec partition_tree_inf_sup pivot = function
  | E -> (E, None, E)
  | N (l, k, v, r, _) as node ->
      match compare pivot k with
	| n when n < 0 ->
	    let (inf, value, sup) = partition_tree_inf_sup pivot l in
	      (inf, value, makeSTree sup k v r)
	| n when n > 0 ->
	    let (inf, value, sup) = partition_tree_inf_sup pivot r in
	      (makeSTree l k v inf, value, sup)
	| _ -> (l, Some (k, v), r)

let rec union_tree_with f = fun a b ->
  match (a, b) with
    | (E, _) -> a
    | (_, E) -> b
    | (N (l, k, v, r, _), N _) ->
	match partition_tree_inf_sup k b with
	  | (l', None, r') -> 
	      makeSTree (union_tree_with f l l') k v (union_tree_with f r r')
	  | (l', Some (_, v'), r') ->
	      makeSTree (union_tree_with f l l') k (f v v') 
	      (union_tree_with f r r')

let rec plus = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | _ -> union_tree_with 
	(fun (pred, succ) (pred', succ') -> 
	   (plus_tree pred pred', plus_tree succ succ'))
	a b

let zero = function () -> E


(** Oprations binaires **)
(** Binairy operations **)
(** Operaciones binrias **)


let union = plus

let rec intersection_tree = fun a b ->
  match (a, b) with
    | (E, _) | (_, E) -> E
    | (N (l, k, v, r, _), N _) ->
	match partition_tree_inf_sup k b with
	  | (l', None, r') -> 
	      appendS (intersection_tree l l') (intersection_tree r r')
	  | (l', Some _, r') -> 
	      makeSTree (intersection_tree l l') k v (intersection_tree r r')

let rec intersection_tree_with f = fun a b ->
  match (a, b) with
    | (E, _) | (_, E) -> E
    | (N (l, k, v, r, _), N _) ->
	match partition_tree_inf_sup k b with
	  | (l', None, r') -> 
	      appendS (intersection_tree_with f l l') 
	      (intersection_tree_with f r r')
	  | (l', Some (_, v'), r') ->
	      makeSTree (intersection_tree_with f l l') k (f v v')
	      (intersection_tree_with f r r')

let intersection = fun a b ->
  intersection_tree_with
  (fun (pred, succ) (pred', succ') -> 
     (intersection_tree pred pred', intersection_tree succ succ')
  )
  a b


(** Observateurs **)
(** Observers **)
(** Observadores **)


let is_empty = function
  | E -> true
  | N _ -> false

let rec size_tree_rec result = function
  | E -> result
  | N (l, _, _, r, _) -> size_tree_rec (size_tree_rec (1 + result) r) l

let size_tree = function map -> size_tree_rec 0 map

let nodes = function graph -> size_tree_rec 0 graph

let rec edges_rec result = function
  | E -> result
  | N (l, _, (_, succ), r, _) ->
      edges_rec (
	edges_rec (
	  size_tree_rec result succ
	) l
      ) r

let edges = function graph -> edges_rec 0 graph

let rec labels_tree result = function
  | E -> result
  | N (l, _, v, r, _) -> labels_tree (labels_tree (v :: result) l) r

let rec labels_rec result = function
  | E -> result
  | N (l, _, (_, succ), r, _) -> 
      labels_rec (labels_rec (labels_tree result succ) l) r

let labels_list = function graph -> labels_rec [] graph

let rec nodes_rec result = function
  | E -> result
  | N (l, i, _, r, _) -> nodes_rec (nodes_rec (i :: result) l) r

let nodes_list = function graph -> nodes_rec [] graph

let edges_list = to_list

let rec degrees_list_rec result = function
  | E -> result
  | N (l, k, (pred, succ), r, _) ->
      degrees_list_rec (
	degrees_list_rec (
	  (k, size_tree pred, size_tree succ) :: result
	) l
      ) r

let degrees_list = function graph -> degrees_list_rec [] graph


(** Operations on nodes **)
(** Oprations sur les noeuds **)
(** Operaciones sobre los nudos **)


let rec member_node k = function
  | E -> false
  | N (l, k', _, r, _) ->
      match compare k k' with
	| n when n < 0 -> member_node k l
	| n when n > 0 -> member_node k r
	| _ -> true

let rec new_node_rec = fun n graph ->
  if member_node n graph then new_node_rec (n + 1) graph
  else n

let new_node = function graph -> new_node_rec 0 graph

let rec new_nodes_rec n graph = function
  | 0 -> []
  | k when member_node n graph -> new_nodes_rec (n + 1) graph k
  | k -> n :: new_nodes_rec (n + 1) graph (k - 1)

let new_nodes = fun n graph -> new_nodes_rec 0 graph n

let rec insert_node_list = fun list graph ->
  match list with
    | [] -> graph
    | node :: tail -> insert_node_list tail (insert_node node graph)

let rec delete_raise_tree x = function
  | E -> raise Not_found
  | N (l, k, v, r, _) ->
      match compare x k with
	| n when n < 0 -> makeBTree (delete_raise_tree x l) k v r
	| n when n > 0 -> makeBTree l k v (delete_raise_tree x r)
	| _ -> append l r

let delete_tree x = function map -> 
  try 
    delete_raise_tree x map 
  with Not_found -> map

let rec map_images f = function
  | E -> E
  | N (l, k, v, r, h) -> N (map_images f l, k, f v, map_images f r, h)

let rec delete_node_rec f x = function
  | E -> raise Not_found
  | N (l, v, (pred, succ), r, _) ->
      match compare x v with
	| n when n < 0 -> 
	    makeBTree (delete_node_rec f x l) 
	    v (delete_tree x pred, delete_tree x succ) (map_images f r)
	| n when n > 0 -> 
	    makeBTree (map_images f l)
	    v (delete_tree x pred, delete_tree x succ) 
	    (delete_node_rec f x r) 
	| _ -> append (map_images f l) (map_images f r)

let delete_node = fun x graph ->
  let f = function (pred, succ) -> (delete_tree x pred, delete_tree x succ) in
    delete_node_rec f x graph

let rec delete_node_raise x = function
  | E -> raise Not_found
  | N (l, v, (pred, succ), r, _) ->
      match compare x v with
	| n when n < 0 -> 
	    makeBTree (delete_node_raise x l) 
	    v (delete_tree x pred, delete_tree x succ) 
	    (map_images 
	       (function (pred, succ) -> 
		  (delete_tree x pred, delete_tree x succ)
	       )
	       r)
	| n when n > 0 -> 
	    makeBTree 
	    (map_images 
	       (function (pred, succ) -> 
		  (delete_tree x pred, delete_tree x succ)
	       )
	       l)
	    v (delete_tree x pred, delete_tree x succ) 
	    (delete_node_raise x r) 
	| _ -> 
	    let f = function (pred, succ) -> 
	      (delete_tree x pred, delete_tree x succ)
	    in
	      append (map_images f l) (map_images f r)

let rec delete_node_list = fun list graph ->
  match list with
    | [] -> graph
    | node :: tail -> delete_node_list tail (delete_node node graph)

let rec keys_rec result = function
  | E -> result
  | N (l, k, _, r, _) -> keys_rec (k :: keys_rec result r) l

let keys = function tree -> keys_rec [] tree

let rec successors x = function
  | E -> raise Not_found
  | N (l, v, (_, succ), r, _) ->
      match compare x v with
	| n when n < 0 -> successors x l
	| n when n > 0 -> successors x r
	| _ -> keys_rec [] succ

let rec successors_rec result x = function
  | E -> raise Not_found
  | N (l, v, (_, succ), r, _) ->
      match compare x v with
	| n when n < 0 -> successors_rec result x l
	| n when n > 0 -> successors_rec result x r
	| _ -> keys_rec result succ

let rec predecessors x = function
  | E -> raise Not_found
  | N (l, v, (pred, _), r, _) ->
      match compare x v with
	| n when n < 0 -> predecessors x l
	| n when n > 0 -> predecessors x r
	| _ -> keys_rec [] pred

let rec predecessors_rec result x = function
  | E -> raise Not_found
  | N (l, v, (pred, _), r, _) ->
      match compare x v with
	| n when n < 0 -> predecessors_rec result x l
	| n when n > 0 -> predecessors_rec result x r
	| _ -> keys_rec result pred

let rec adjacents x = function
  | E -> raise Not_found
  | N (l, v, (pred, succ), r, _) ->
      match compare x v with
	| n when n < 0 -> adjacents x l
	| n when n > 0 -> adjacents x r
	| _ -> (keys_rec [] pred, keys_rec [] succ)

let rec exterior_edges x = function
  | E -> raise Not_found
  | N (l, v, (_, succ), r, _) ->
      match compare x v with
	| n when n < 0 -> exterior_edges x l
	| n when n > 0 -> exterior_edges x r
	| _ -> to_list_tree v [] succ

let rec exterior_edges_rec result x = function
  | E -> raise Not_found
  | N (l, v, (_, succ), r, _) ->
      match compare x v with
	| n when n < 0 -> exterior_edges_rec result x l
	| n when n > 0 -> exterior_edges_rec result x r
	| _ -> to_list_tree v result succ

let rec interior_tree j result = function
  | E -> result
  | N (l, i, v, r, _) ->
      interior_tree j (
	interior_tree j (
	  (i, v, j) :: result
	) l
      ) r

let rec interior_edges x = function
  | E -> raise Not_found
  | N (l, v, (pred, _), r, _) ->
      match compare x v with
	| n when n < 0 -> interior_edges x l
	| n when n > 0 -> interior_edges x r
	| _ -> interior_tree v [] pred

let rec interior_edges_rec result x = function
  | E -> raise Not_found
  | N (l, v, (pred, _), r, _) ->
      match compare x v with
	| n when n < 0 -> interior_edges_rec result x l
	| n when n > 0 -> interior_edges_rec result x r
	| _ -> interior_tree v result pred

let rec adjacent_edges x = function
  | E -> raise Not_found
  | N (l, v, (pred, succ), r, _) ->
      match compare x v with
	| n when n < 0 -> adjacent_edges x l
	| n when n > 0 -> adjacent_edges x r
	| _ -> (interior_tree v [] pred, to_list_tree v [] succ)

let rec exterior_degree x = function
  | E -> raise Not_found
  | N (l, v, (_, succ), r, _) ->
      match compare x v with
	| n when n < 0 -> exterior_degree x l
	| n when n > 0 -> exterior_degree x r
	| _ -> size_tree succ

let rec interior_degree x = function
  | E -> raise Not_found
  | N (l, v, (pred, _), r, _) ->
      match compare x v with
	| n when n < 0 -> interior_degree x l
	| n when n > 0 -> interior_degree x r
	| _ -> size_tree pred

let rec degree x = function
  | E -> raise Not_found
  | N (l, v, (pred, succ), r, _) ->
      match compare x v with
	| n when n < 0 -> degree x l
	| n when n > 0 -> degree x r
	| _ -> (size_tree pred, size_tree succ)

let rec filter_nodes_rec f result = function
  | E -> result
  | N (l, k, v, r, _) ->
      filter_nodes_rec f 
      (filter_nodes_rec f 
	 (if f k then insert_tree k v result else result) 
	 l) 
      r

let filter_nodes = fun f graph -> filter_nodes_rec f E graph

let rec filter_nodes_le x = function
  | E -> E
  | N (l, k, v, r, _) ->
      match compare x k with
	| n when n < 0 -> filter_nodes_le x l
	| _ -> makeSTree l k v (filter_nodes_le x r)

let rec filter_nodes_ge x = function
  | E -> E
  | N (l, k, v, r, _) ->
      match compare x k with
	| n when n <= 0 -> makeSTree (filter_nodes_ge x l) k v r
	| _ -> filter_nodes_ge x r

let filter_nodes_gt = filter_gt
let filter_nodes_lt = filter_lt

let rec range_nodes min max = function
  | E -> E
  | N (l, k, v, r, _) ->
      if compare k min < 0 then range_nodes min max r
      else if compare k max > 0 then range_nodes min max l
      else makeSTree (filter_nodes_ge min l) k v (filter_nodes_le max r)

let rec member_nodes_min_bound b k = function
  | E -> compare b k = 0
  | N (l, k', _, r, _) ->
      match compare k k' with
	| v when v < 0 -> member_nodes_min_bound b k l
	| _ -> member_nodes_min_bound k' k r

let rec member_nodes' k = function
  | E -> false
  | N (l, k', _, r, _) ->
      match compare k k' with
	| n when n < 0 -> member_nodes' k l
	| _ -> member_nodes_min_bound k' k r

let rec look_for_node f = function
  | E -> raise Not_found
  | N (_, k, _, _, _) when f k -> k
  | N (l, _, _, r, _) -> 
      (try look_for_node f l with Not_found -> look_for_node f r)

let choose_node = function
  | E -> raise Empty
  | N (_, k, _, _, _) -> k

let rec min_node = function
  | E -> raise Not_found
  | N (E, k, _, _, _) -> k
  | N (l, _, _, _, _) -> min_node l

let rec max_node = function
  | E -> raise Not_found
  | N (_, k, _, E, _) -> k
  | N (_, _, _, r, _) -> max_node r

let rec next_node x = function
  | E -> raise Not_found
  | N (l, k, _, r, _) ->
      match compare x k with
	| n when n < 0 -> (try next_node x l with Not_found -> k)
	| n when n > 0 -> next_node x r
	| _ -> min_node r (* raises [Not_found] *)

let rec previous_node x = function
  | E -> raise Not_found
  | N (l, k, _, r, _) ->
      match compare x k with
	| n when n < 0 -> previous_node x l
	| n when n > 0 -> (try previous_node x r with Not_found -> k)
	| _ -> max_node l (* raises [Not_found] *)

let rec ceil_node x = function
  | E -> raise Not_found
  | N (l, k, _, r, _) ->
      match compare x k with
	| n when n < 0 -> (try ceil_node x l with Not_found -> k)
	| n when n > 0 -> ceil_node x r
	| _ -> k

let rec floor_node x = function
  | E -> raise Not_found
  | N (l, k, _, r, _) ->
      match compare x k with
	| n when n < 0 -> floor_node x l
	| n when n > 0 -> (try floor_node x r with Not_found -> k)
	| _ -> k

let rec for_all_nodes f = function
  | E -> true
  | N (l, k, _, r, _) -> f k && for_all_nodes f l && for_all_nodes f r

let rec exists_node f = function
  | E -> false
  | N (l, k, _, r, _) -> f k || exists_node f l || exists_node f r


(** Operations on edges **)
(** Operations sur les arcs **)
(** Operaciones sobre los arcos **)


let insert_edge_list = fun list graph -> from_list_rec graph list

let rec delete_succ i j = function
  | E -> raise Not_found
  | N (l, x, ((pred, succ) as k), r, _) ->
      match compare i x with
	| n when n < 0 -> makeTree (delete_succ i j l) x k r
	| n when n > 0 -> makeTree l x k (delete_succ i j r)
	| _ -> makeTree l x (pred, delete_tree j succ) r

let rec delete_pred i j = function
  | E -> raise Not_found
  | N (l, x, ((pred, succ) as k), r, _) ->
      match compare j x with
	| n when n < 0 -> makeTree (delete_pred i j l) x k r
	| n when n > 0 -> makeTree l x k (delete_pred i j r)
	| _ -> makeTree l x (delete_tree i pred, succ) r

let rec delete_edge_raise i j = function
  | E -> raise Not_found
  | N (l, x, (pred, succ), r, _) when x = i ->
      if x < j then 
	makeTree l x (pred, delete_tree j succ) (delete_pred i j r)
      else if x = j then 
	makeTree l x (delete_tree x pred, delete_tree x succ) r
      else (* x > j *)
	makeTree (delete_pred i j l) x (pred, delete_tree j succ) r
  | N (l, x, (pred, succ), r, _) when x = j ->
      if x < i then 
	makeTree l x (delete_tree i pred, succ) (delete_succ i j r)
      else (* x > i *)
	makeTree (delete_succ i j l) x (delete_tree i pred, succ) r
  | N (l, x, k, r, _) ->
      if x < i then
	if x < j then makeTree l x k (delete_edge_raise i j r)
	else makeTree (delete_pred i j l) x k (delete_succ i j r)
      else (* x > i *)
	if x < j then makeTree (delete_succ i j l) x k (delete_pred i j r)
	else makeTree (delete_edge_raise i j l) x k r

let delete_edge = fun i j graph ->
  try
    delete_edge_raise i j graph
  with Not_found -> graph

let rec delete_edge_list = fun list graph ->
  match list with
    | [] -> graph
    | (i, j) :: tail -> delete_edge_list tail (delete_edge i j graph)

let rec lookup x = function
  | E -> raise Not_found
  | N (l, k, v, r, _) ->
      match compare x k with
	| n when n < 0 -> lookup x l
	| n when n > 0 -> lookup x r
	| _ -> v

let rec edge_label i j = function
  | E -> raise Not_found
  | N (l, k, (_, succ), r, _) ->
      match compare i k with
	| n when n < 0 -> edge_label i j l
	| n when n > 0 -> edge_label i j r
	| _ -> lookup j succ

let rec map_edge_labels f = function
  | E -> E
  | N (l, k, (pred, succ), r, h) ->
      N (map_edge_labels f l, k, 
	 (map_images f pred, map_images f succ), 
	 map_edge_labels f r, h)

let rec member_edge i j = function
  | E -> raise Not_found
  | N (l, k, (_, succ), r, _) ->
      match compare k i with
	| n when n < 0 -> member_edge i j l
	| n when n > 0 -> member_edge i j r
	| _ -> member_node j succ

let rec look_for f = function
  | E -> raise Not_found
  | N (_, k, v, _, _) when f k v -> (k, v)
  | N (l, _, _, r, _) -> (try look_for f l with Not_found -> look_for f r)

let rec look_for_edge f = function
  | E -> raise Not_found
  | N (l, i, (_, succ), r, _) ->
      try 
	let (j, v) = look_for (fun j v -> f i v j) succ in
	  (i, v, j)
      with Not_found -> try look_for_edge f l 
      with Not_found -> look_for_edge f r

let choose_edge = function graph -> look_for_edge (fun i v j -> true) graph

let rec exists f = function
  | E -> false
  | N (l, k, v, r, _) -> f k v || exists f l || exists f r

let rec exists_edge f = function
  | E -> false
  | N (l, i, (_, succ), r, _) ->
      exists (fun j v -> f i v j) succ || exists_edge f l || exists_edge f r

let rec for_all f = function
  | E -> true
  | N (l, k, v, r, _) -> f k v && for_all f l && for_all f r

let rec for_all_edges f = function
  | E -> true
  | N (l, i, (_, succ), r, _) -> 
      for_all (fun j v -> f i v j) succ && 
      for_all_edges f l && 
      for_all_edges f r

let rec iter f = function
  | E -> ()
  | N (l, k, v, r, _) -> iter f l ; f k v ; iter f r

let rec iter_edges f = function
  | E -> ()
  | N (l, i, (_, succ), r, _) -> 
      iter_edges f l ; 
      iter (fun j v -> f i v j) succ ;
      iter_edges f r

let rec fold_left f e = function
  | E -> e
  | N (l, k, v, r, _) -> fold_left f (f (fold_left f e l) k v) r

let rec fold_left_edges f e = function
  | E -> e
  | N (l, i, (_, succ), r, _) ->
      fold_left_edges f (
	fold_left (fun e j v -> f e i v j) (fold_left_edges f e l) succ
      ) r

let rec fold_right = fun f set e ->
  match set with
    | E -> e 
    | N (l, k, v, r, _) -> fold_right f l (f k v (fold_right f r e))

let rec fold_right_edges = fun f tree e ->
  match tree with
    | E -> e
    | N (l, i, (_, succ), r, _) ->
	fold_right_edges f l
	(fold_right (fun j v e -> f i v j e) succ (fold_right_edges f r e))

let fold_edges = fold_right_edges



(** Traversals **)
(** Parcours **)
(** Recorridos **)


let rec iter_nodes f = function
  | E -> ()
  | N (l, k, _, r, _) -> iter_nodes f l ; f k ; iter_nodes f r

let rec fold_left_nodes f e = function
  | E -> e
  | N (l, k, _, r, _) -> fold_left_nodes f (f (fold_left_nodes f e l) k) r

let rec fold_right_nodes = fun f set e ->
  match set with
    | E -> e 
    | N (l, k, _, r, _) -> fold_right_nodes f l (f k (fold_right_nodes f r e))

let fold_nodes = fold_right_nodes


(** Graph traversals **)
(** Parcours de graphes **)
(** Recorridos de grafos **)


let rec delete_node_unsafe x = function
  | E -> raise Not_found
  | N (l, k, v, r, _) ->
      match compare x k with
	| n when n < 0 -> makeBTree (delete_node_unsafe x l) k v r
	| n when n > 0 -> makeBTree l k v (delete_node_unsafe x r)
	| _ -> append l r

let rec dfs f e graph = function
  | [] -> e
  | x :: tail ->
      let reduced = delete_node_unsafe x graph in
	try
	  dfs f (f x e) reduced (successors_rec tail x graph)
	with Not_found -> dfs f e reduced tail

let rec dfs' f e graph = function
  | [] -> e
  | (i, v, j) :: tail ->
      let reduced = delete_node_unsafe j graph in
	try 
	  dfs' f (f i v j e) reduced (exterior_edges_rec tail j graph)
	with Not_found -> dfs' f e reduced tail
	  
let depth_first_fold = fun f e graph x -> dfs f e graph [x]

let depth_first_fold' = fun f e graph x -> 
  dfs' f e graph (exterior_edges x graph)

let rec dfs_list result graph = function
  | [] -> result
  | x :: tail ->
      let reduced = delete_node_unsafe x graph in
	try
	  dfs_list (x :: result) reduced (successors_rec tail x graph)
	with Not_found -> dfs_list result reduced tail
      
let rec dfs_list' result graph = function
  | [] -> result
  | (_, _, j) as head :: tail ->
      let reduced = delete_node_unsafe j graph in
	try
	  dfs_list' (head :: result) reduced (exterior_edges_rec tail j graph)
	with Not_found -> dfs_list' result reduced tail

let depth_first_list = fun graph x -> List.rev (dfs_list [] graph [x])

let depth_first_list' = fun graph x -> 
  List.rev (dfs_list' [] 
	      (delete_node_unsafe x graph) 
	      (List.rev (exterior_edges x graph)) 
	   )

let rec 
  bfs_notail f e graph = function
    | [] -> e
    | x :: tail ->
	let reduced = delete_node_unsafe x graph in
	  try
	    bfs_tail f (f x e) reduced tail (successors x graph)
	  with Not_found -> bfs_notail f e reduced tail
and
  bfs_tail f e = fun graph hd tl ->
    match hd with
      | [] ->
	  (match tl with
	     | [] -> e
	     | _ -> bfs_notail f e graph (List.rev tl)
	  )
      | x :: tail ->
	  let reduced = delete_node_unsafe x graph in
	    try
	      bfs_tail f (f x e) reduced tail (successors_rec tl x graph)
	    with Not_found -> bfs_tail f e reduced tail tl

let rec 
  bfs_notail' f e graph = function
    | [] -> e
    | (i, v, j) :: tail ->
	let reduced = delete_node_unsafe j graph in
	  try
	    bfs_tail' f (f i v j e) reduced tail (exterior_edges j graph)
	  with Not_found -> bfs_notail' f e reduced tail
and
  bfs_tail' f e = fun graph hd tl ->
    match hd with
      | [] ->
	  (match tl with
	     | [] -> e
	     | _ -> bfs_notail' f e graph (List.rev tl)
	  )
      | (i, v, j) :: tail ->
	  let reduced = delete_node_unsafe j graph in
	    try
	      bfs_tail' f (f i v j e) reduced tail 
		(exterior_edges_rec tl j graph)
	    with Not_found -> bfs_tail' f e reduced tail tl

let breadth_first_fold = fun f e graph x -> bfs_notail f e graph [x]

let breadth_first_fold' = fun f e graph x -> 
  bfs_notail' f e (delete_node_unsafe x graph) (exterior_edges x graph)

let rec 
  bfs_list_notail result graph = function
    | [] -> result
    | x :: tail ->
	let reduced = delete_node_unsafe x graph in
	  try
	    bfs_list_tail (x :: result) reduced tail (successors x graph)
	  with Not_found -> bfs_list_notail result reduced tail
and
  bfs_list_tail result = fun graph hd tl ->
    match hd with
      | [] ->
	  (match tl with
	     | [] -> result
	     | _ -> bfs_list_notail result graph (List.rev tl)
	  )
      | x :: tail ->
	  let reduced = delete_node_unsafe x graph in
	    try
	      bfs_list_tail (x :: result) reduced tail 
		(successors_rec tl x graph)
	    with Not_found -> bfs_list_tail result reduced tail tl

let rec 
  bfs_list_notail' result graph = function
    | [] -> result
    | (_, _, j) as head :: tail ->
	let reduced = delete_node_unsafe j graph in
	  try
	    bfs_list_tail' (head :: result) reduced tail 
	      (exterior_edges j graph)
	  with Not_found -> bfs_list_notail' result reduced tail
and
  bfs_list_tail' result = fun graph hd tl ->
    match hd with
      | [] ->
	  (match tl with
	     | [] -> result
	     | _ -> bfs_list_notail' result graph (List.rev tl)
	  )
      | (_, _, j) as head :: tail ->
	  let reduced = delete_node_unsafe j graph in
	    try
	      bfs_list_tail' (head :: result) reduced tail 
		(exterior_edges_rec tl j graph)
	    with Not_found -> bfs_list_tail' result reduced tail tl

let breadth_first_list = fun graph x -> 
  List.rev (bfs_list_notail [] graph [x])

let breadth_first_list' = fun graph x -> 
  List.rev (bfs_list_notail' [] 
	      (delete_node_unsafe x graph)
	      (exterior_edges x graph)
	   )

exception Cyclic

let rec make_heap heap = function
  | E -> heap
  | N (l, v, (_, s), r, _) ->
      make_heap (
	make_heap (
	  CartesianHeap.insert v (size_tree_rec 0 s) heap
	) l
      ) r

let rec reduce_heap heap = function
  | [] -> heap
  | node :: tail -> reduce_heap (CartesianHeap.substract_one node heap) tail
  
let rec topological_list_rec = fun graph heap list ->
  if CartesianHeap.is_empty heap then list
  else
    match CartesianHeap.min_priority_view heap with
      | ((node, 0), new_heap) ->
	  topological_list_rec graph 
	  (reduce_heap new_heap (predecessors node graph))
	  (node :: list)
      | _ -> raise Cyclic

let topological_list = function graph -> 
  let heap = make_heap (CartesianHeap.zero ()) graph in
    topological_list_rec graph heap []

let rec topological_rec = function
  | E -> raise Not_found
  | N (_, v, (_, E), _, _) -> v
  | N (l, _, _, r, _) ->
      try topological_rec l with Not_found -> topological_rec r

let rec topological_list_rec result = function
  | E -> result
  | graph ->
     try
       let x = topological_rec graph in
	 topological_list_rec (x :: result) (delete_node x graph)
     with Not_found -> raise Cyclic

let topological_list' = function graph -> topological_list_rec [] graph

let rec topological_fold f e = function
  | E -> e
  | graph ->
      try
	let x = topological_rec graph in
	  topological_fold f (f x e) (delete_node x graph)
      with Not_found -> raise Cyclic

