(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Graphs represented by arrays of tries
 * Graphes reprsents par des tableaux de tries
 * Grafos representados por tableros de tries
 * 
 *)

open BairePervasives

type node = int

(* autre reprsentation possible
 *
 * type 'a tree = 
 *    | E 
 *    | L of int 
 *    | N of 'a tree * 'a * node * 'a tree * 'a tree
 * 
 * Elle vite les duplications de code mais augmente la taille mmoire
 * de la structure de donnes et oblige certaines fonctions  passer
 * des paramtres supplmentaires
 * 
 * De surcrot, on ne sait trop comment interprter un arbre qui s'achverait
 * sur un constructeur de type L et non E
 *)

type 'a tree = 
  | E 
  | L of 'a tree * char * node * 'a tree
  | N of 'a tree * char * node * 'a tree * 'a tree

type 'a pointer = 
  | Pointer of node * 'a list * 'a tree
  | Trie of int * 'a tree 

type 'a graph = 'a pointer array


(** Utils **)
(** Utilitaires **)
(** tiles **)


let moduleName = "TrieGraph"

let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type

let makeNode = fun l v n r s -> N (l, v, n, r, s)
let makeLeaf = fun l v n r -> L (l, v, n, r)

(** Oprations catgorielles **)
(** Categorial operiations **)
(** Operaciones de categora **)
      

let rec insert_trie_rec value node tree = function
  | E -> makeNode E value node E tree
  | L (l, v, k, r) -> 
      (match compare value v with
	 | n when n < 0 -> makeLeaf (insert_trie_rec value node tree l) v k r
	 | n when n > 0 -> makeLeaf l v k (insert_trie_rec value node tree r)
	 | _ -> raise Not_found
      )
  | N (l, v, k, r, s) ->
      match compare value v with
	| n when n < 0 -> makeNode (insert_trie_rec value node tree l) v k r s
	| n when n > 0 -> makeNode l v k (insert_trie_rec value node tree r) s
	| _ -> makeNode l v k r tree

let rec insert_trie = fun value node tree path trie ->
  match path with
    | [] -> insert_trie_rec value node tree trie
    | x :: tail ->
	match tree with
	  | E -> makeNode E value node E tree
	  | L (l, v, k, r) ->
	      (match compare x v with
		 | n when n < 0 -> 
		     makeLeaf (insert_trie value node tree path l) v k r
		 | n when n > 0 ->
		     makeLeaf l v k (insert_trie value node tree path r)
		 | _ -> raise Not_found
	      )
	  | N (l, v, k, r, s) ->
	      match compare x v with
		| n when n < 0 -> 
		    makeNode (insert_trie value node tree path l) v k r s
		| n when n > 0 -> 
		    makeNode l v k (insert_trie value node tree path r) s
		| _ -> makeNode l v k r (insert_trie value node tree tail s)

let rec insert_tree = fun index value node tree path graph ->
  match graph.(index) with
    | Pointer (root, tail_path, _) -> 
	  insert_tree root value node tree (path @ tail_path) graph
    | Trie (count, t) -> 
	let trie = insert_trie value node tree path t in
	  graph.(index) <- Trie (count, trie) ;
	  (value :: path, trie)

let remove_level = function
  | E | L _ -> invalid "remove_level" UnexpectedPattern
  | N (_, _, _, _, s) -> s

let insert = fun i v j graph ->
  graph.(j) <- (match graph.(j) with
		  | Trie (0, t) ->
		      let (path, trie) = insert_tree i v j t [] graph in
			Pointer (i, path, remove_level trie)
		  | Trie (n, t) ->
		      ignore (insert_tree i v j E [] graph) ;
		      Trie (n + 1, t)
		  | Pointer (root, path, t) ->
		      ignore (insert_tree root v j E path graph) ;
		      Trie (2, t)
	       );
  graph

let rec insert_list = fun list graph ->
  match list with
    | [] -> graph
    | (i, v, j) :: tail -> insert_list tail (insert i v j graph)

let rec to_list_trie i result = function
  | E -> result
  | L (l, v, j, r) -> to_list_trie i (to_list_trie i ((i, v, j) :: result) l) r
  | N (l, v, j, r, s) ->
      to_list_trie i(
	to_list_trie i (
	  to_list_trie j ((i, v, j) :: result) s
	) l
      ) r

let rec to_list_rec graph result = function
  | n when n < 0 -> result
  | n ->
      match graph.(n) with
	| Pointer _ -> to_list_rec graph result (n - 1)
	| Trie (_, t) -> to_list_rec graph (to_list_trie n result t) (n - 1)

let rec to_list = function graph -> 
  to_list_rec graph [] (Array.length graph - 1)

let zero = function () -> [||]

let new_node = function graph -> Array.length graph

let new_nodes = fun n graph -> n + Array.length graph

let insert_node = fun n graph ->
  let size = max n (Array.length graph - 1) in
  let new_array = Array.make size (Trie (0, E)) in
    Array.blit graph 0 new_array 0 size ;
    new_array

let rec max_list mx = function
  | [] -> mx
  | x :: tail when x > mx -> max_list x tail
  | _ :: tail -> max_list mx tail

let insert_node_list = fun list graph ->
  let n = max_list 0 list in
  let size = max n (Array.length graph - 1) in
  let new_array = Array.make size (Trie (0, E)) in
    Array.blit graph 0 new_array 0 size ;
    new_array

let make_graph = function n -> Array.make n (Trie (0, E))
  
let rec next_rec = fun x path trie graph ->
  match trie with
    | E -> raise Not_found
    | L (l, v, n, r) ->
	(match compare x v with
	   | k when k < 0 -> next_rec x path l graph
	   | k when k > 0 -> next_rec x path r graph
	   | _ -> next_list path graph
	)
    | N (l, v, n, r, s) ->
	match compare x v with
	  | k when k < 0 -> next_rec x path l graph
	  | k when k > 0 -> next_rec x path r graph
	  | _ -> next_list_rec path trie graph

and next_list = fun node path graph ->
  match graph.(node) with
    | Pointer (_, _, t) -> next_list_rec path t graph
    | Trie (_, t) -> next_list_rec path t graph

and next_list_rec = fun path trie graph ->
  match path with
    | [] -> 
	(match trie with
	   | E -> raise Not_found
	   | L (_, _, n, _) | N (_, _, n, _, _) -> n
	)
    | x :: tail -> next_rec x tail trie graph
