(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Leftist heap
 * Tas gauche
 * Monton izquierdo
 *
 *)

open BairePervasives

let moduleName = "LeftistHeap"

type 'a heap =  E | N of int * 'a heap * 'a * 'a heap


(** Utilitaires **)
(** Utils **)
(** tiles **)


let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type

let makeNode = fun a x b ->
  match (a, b) with
    | (_, E) -> N (1, a, x, E)
    | (E, _) -> N (1, b, x, E)
    | (N (m, _, _, _) , N (n, _, _, _)) when m <= n -> N (m + 1, b, x, a)
    | (_, N (n, _, _, _)) -> N (n + 1, a, x, b)


(** Categorial operations **)
(** Oprations catgorielles **)
(** Operaciones categoriales **)


let injection = function x -> N (1, E, x, E)

let rec plus = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (_, ax, x, bx), N (_, ay, y, by))  ->
	match compare x y with
	  | n when n < 0 -> makeNode ax x (plus bx b)
	  | _ -> makeNode ay y (plus a by)
	      
let rec to_list = function
  | E -> []
  | N (_, a, x, b) -> x :: to_list (plus a b)

let rec from_list_rec result = function
  | x :: y :: tail -> from_list_rec (plus x y :: result) tail
  | [x] -> from_list_rec (x :: result) []
  | [] -> 
      match result with
	| [] -> E
	| [x] -> x
	| _ -> from_list_rec [] result

let from_list = function liste -> from_list_rec [] (List.map injection liste)

let map f = function heap -> from_list (List.map f (to_list heap))

let rec join_rec result = function
  | E -> result
  | N (_, a, x, b) -> join_rec (join_rec (plus x result) a) b 

let join = function tree -> join_rec E tree

let zero = function () -> E

let rec make_rec a x b = function
  | 1 -> makeNode a x b
  | n when n mod 2 = 0 -> make_rec a x (makeNode b x b) (n / 2)
  | n -> make_rec (makeNode a x b) x (makeNode b x b) ((n - 1) / 2)

let make = fun n x -> 
  if n < 0 then invalid "make" Negative
  else make_rec E x E n


(** View **)
(** Vues **)
(** Vistas **)


let min_view = function
  | N (_, a, x, b) -> (x, plus a b)
  | _ -> raise Empty

let rec max_view = function
  | E -> raise Empty
  | N (_, E, x, _) -> (E, x)
  | N (_, a, x, E) -> let (a', y) = max_view a in (N (1, a', x, E), y)
  | N (_, a, x, b) ->
      let 
	(a', y) = max_view a and
	(b', z) = max_view b
      in
	match compare y z with
	  | n when n < 0 -> (makeNode a x b', z)
	  | _ -> (makeNode a' x b, y)    

let rec partition f = function
  | E -> (E, E)
  | N (_, a, x, b) -> 
      let 
	(a', a'') = partition f a and
	(b', b'') = partition f b
      in 
	if f x then (makeNode a' x b', plus a'' b'')
	else (plus a' b', makeNode a'' x b'')


(** Acces **)
(** Accs **)
(** Acceso **)


let rec insert x = function
  | E -> injection x
  | N (_, a, y, b) as tree -> 
      match compare x y with
	| n when n < 0 -> N (1, tree, x, E)
	| _ -> makeNode a y (insert x b)
	    
let insert_list list = function heap -> plus (from_list list) heap

let rec delete_rec x = function
  | E -> raise Exit
  | N (n, a, y, b) ->
      match compare x y with
	| 0 -> plus a b
	| _ ->
	    try N (n, a, y, delete_rec x b) 
	    with Exit -> N (n, delete_rec x a, y, b)

let rec delete x = function heap -> 
  try delete_rec x heap 
  with Exit -> heap

let rec delete_all_rec x = function
  | E -> raise Exit
  | N (n, a, y, b) as node ->
      let 
	left = try delete_all_rec x a with Exit -> a and
	right = try delete_all_rec x b with Exit -> b
      in
	match compare x y with
	  | 0 -> plus left right
	  | _ when a == left && b == right -> node
	  | _ -> N (n, right, y, left)

let delete_all x = function heap ->
  try delete_all_rec x heap 
  with Exit -> heap

let delete_min = function
  | E -> E
  | N (_, a, _, b) -> plus a b

let rec find_max m = function
  | E -> m
  | N (_, E, x, _) -> 
      (match compare x m with
	 | n when n < 0 -> m
	 | _ -> x
      )
  | N (_, a, x, b) -> find_max (find_max m a) b

let rec find_max_leaf = function
  | E -> raise Empty
  | N (_, E, x, _) -> x
  | N (_, a, _, b) -> find_max (find_max_leaf a) b

let max_element = function
  | E -> raise Empty
  | N (_, E, x, _) -> x
  | N (_, a, _, b) -> find_max (find_max_leaf a) b
	  
let delete_max = function heap -> 
  try 
    delete (max_element heap) heap
  with Empty -> E

let rec filter f = function
  | E -> E
  | N (_, a, x, b) when f x -> makeNode (filter f a) x (filter f b)
  | N (_, a, _, b) -> plus (filter f a) (filter f b)


(** Observers **)
(** Observateurs **)
(** Observadores **)


let is_empty = function
  | E -> true
  | N _ -> false

let rec size_rec total = function
  | E -> total
  | N (_, a, _, b) -> size_rec (size_rec (total + 1) b) a

let size = function heap -> size_rec 0 heap

let rec member x = function
  | E -> false
  | N (_, a, y, b) ->
      match compare x y with
	| n when n < 0 -> false
	| 0 -> true
	| _ -> member x b || member x a

let rec count_rec total x = function
  | E -> total
  | N (_, a, y, b) ->
      match compare x y with
	| n when n < 0 -> total
	| 0 -> count_rec (1 + count_rec total x a) x b
	| _ -> count_rec (count_rec total x a) x b

let count x = function heap -> count_rec 0 x heap

let min_element = function
  | N (_, _, x, _) -> x
  | _ -> raise Empty

let rec exists f = function
  | E -> false
  | N (_, a, x, b) -> f x || exists f a || exists f b

let rec for_all f = function
  | E -> true
  | N (_, a, x, b) -> f x && for_all f a && for_all f b


(** Traversals **)
(** Parcours **)
(** Recorridos **)


let rec fold_right = fun f heap e ->
  match heap with
  | E -> e
  | N (_, a, x, b) -> f x (fold_right f (plus a b) e)

let rec fold_left f e = function
  | E -> e
  | N (_, a, x, b) -> fold_left f (f e x) (plus a b)

let fold = fun f heap e -> fold_right f heap e

let rec iter f = function
  | E -> ()
  | N (_, a, x, b) -> iter f a ; f x ; iter f b


(** Miscellaneous **)
(** Divers **)
(** Micelaneos **)


let rec unsafe_map_monotonic f = function
  | E -> E
  | N (n, a, x, b) -> 
      N (n, unsafe_map_monotonic f a, f x , unsafe_map_monotonic f b)


(** Debug **)

let rec depth = function
  | E -> 0
  | N (_, a, _, b) -> 1 + max (depth a) (depth b)
