(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Pairing heap
 * Tas couplants
 * Montones apareados
 * 
 *)

open BairePervasives

let moduleName = "PairingHeap"

type 'a heap =  E | N of 'a * 'a heap list


(** Utilitaires **)
(** Utils **)
(** tiles **)


let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type


(** Categorial operations **)
(** Oprations catgorielles **)
(** Operaciones categoriales **)


let injection = function x -> N (x, [])

let plus = fun h1 h2 ->
  match (h1, h2) with
    | (_, E) -> h1
    | (E, _) -> h2
    | (N (x, c1), N (y, c2)) ->
	match compare x y with
	  | n when n < 0 -> N (x, h2 :: c1)
	  | _ -> N (y, h1 :: c2)

let zero = function () -> E

let rec plus_pairs = function
  | [] -> E
  | [h] -> h
  | h1 :: h2 :: hs -> plus (plus h1 h2) (plus_pairs hs)

let rec join = function
  | E -> E
  | N (x, hs) -> plus x (join (plus_pairs hs))

let rec to_list = function
  | E -> []
  | N (x, hs) -> x :: to_list (plus_pairs hs)

let from_list = function liste ->
  plus_pairs (List.map injection liste)

let map f = function heap -> from_list (List.map f (to_list heap))


(** Acces **)
(** Accs **)
(** Acceso **)


let insert x = function
  | E -> injection x
  | heap -> plus (injection x) heap

let insert_list liste = function heap -> plus (from_list liste) heap

let rec delete_rec x = function
    | E -> raise Not_found
    | N (y, hs) when compare x y = 0 -> plus_pairs hs
    | N (y, hs) -> N (y, delete_list x hs)
and
  delete_list x = function
    | [] -> raise Not_found
    | head :: tail ->
	try 
	  (match delete_rec x head with
	     | E -> tail
	     | other -> other :: tail
	  )
	with Not_found -> head :: delete_list x tail

let delete x = function heap ->
  try
    delete_rec x heap
  with Not_found -> heap

(* How to improve ?
 * let x = from_list [1;2;3] in delete 6 x = x evaluates to true
 * let x = from_list [1;2;3] in delete 6 x == x evaluates to true
 * whereas
 * let x = from_list [1;2;3] in delete_all 6 x = x evaluates to true
 * let x = from_list [1;2;3] in delete_all 6 x == x evaluates to false
 *)

let rec delete_all_rec x = function
    | E -> raise Not_found
    | N (y, hs) when compare x y = 0 -> plus_pairs hs
    | N (y, hs) -> N (y, delete_all_list x hs)
and
  delete_all_list x = function
    | [] -> []
    | head :: tail ->
	try
	  (match delete_all_rec x head with
	     | E -> delete_all_list x tail
	     | other -> other :: delete_all_list x tail
	  )
	with Not_found -> head :: delete_list x tail

let delete_all x = function heap ->
  try
    delete_all_rec x heap
  with Not_found -> heap

let delete_min = function
  | E -> raise Empty
  | N (_, hs) -> plus_pairs hs

let rec max_element_rec m = function
  | E -> invalid "max_element_rec" UnexpectedPattern
  | N (x, hs) ->
      match compare x m with
	| n when n < 0 -> List.fold_left max_element_rec m hs
	| _ -> List.fold_left max_element_rec x hs
 
let max_element = function
  | E -> raise Empty
  | N (x, []) -> x
  | N (x, hs) -> List.fold_left max_element_rec x hs

let delete_max = function heap -> 
  try
    delete (max_element heap) heap
  with Empty -> heap

let rec filter_rec f = function
    | [] -> []
    | head :: tail ->
	match filter f head with
	  | E -> filter_rec f tail
	  | other -> other :: filter_rec f tail
and
  filter f = function
    | E -> E
    | N (x, hs) when f x -> N (x, filter_rec f hs)
    | N (_, hs) -> plus_pairs (List.map (filter f) hs)


(** View **)
(** Vues **)
(** Vistas **)


let min_view = function
  | E -> raise Empty
  | N (x, hs) -> (x, plus_pairs hs)

let rec max_view = function
  | E -> raise Empty
  | N (x, []) -> (E, x)
  | heap -> let m = max_element heap in (delete m heap, m)

let rec partition_rec f (left, right) = function
  | E -> (left, right)
  | N (x, hs) when f x -> 
      List.fold_left (partition_rec f) (insert x left, right) hs
  | N (x, hs) -> 
      List.fold_left (partition_rec f) (left, insert x right) hs

let partition f = function heap -> partition_rec f (E, E) heap


(** Observers **)
(** Observateurs **)
(** Observadores **)


let is_empty = function
  | E -> true
  | _ -> false

let rec size_rec total = function
  | E -> total
  | N (_, hs) -> 1 + List.fold_left size_rec total hs

let size = function heap -> size_rec 0 heap

let rec member x = function
  | E -> false
  | N (y, hs) ->
      match compare x y with
	| n when n < 0 -> false
	| 0 -> true
	| _ -> List.exists (member x) hs

let rec count_rec x total = function
  | E -> total
  | N (y, hs) ->
      match compare x y with
	| n when n < 0 -> total
	| 0 -> 1 + List.fold_left (count_rec x) total hs
	| _ -> List.fold_left (count_rec x) total hs

let count x = function heap -> count_rec x 0 heap

let min_element = function
  | E -> raise Empty
  | N (x, _) -> x

let rec exists f = function
  | E -> false
  | N (x, hs) -> f x || List.exists (exists f) hs

let rec for_all f = function
  | E -> true
  | N (x, hs) -> f x && List.for_all (for_all f) hs

let rec normalize = function
  | [] -> []
  | E :: tail -> normalize tail
  | head :: tail -> head :: normalize tail


(** Traversals **)
(** Parcours **)
(** Recorridos **)


let rec iter f = function
  | E -> ()
  | N (x, hs) -> f x ; List.iter (iter f) hs

let rec fold_right = fun f heap e ->
  match heap with
    | E -> e
    | N (x, hs) -> List.fold_left (sym (fold_right f)) (f x e) hs

let rec fold_left f e = function
  | E -> e
  | N (x, hs) -> List.fold_left (fold_left f) (f e x) hs

let fold = fold_right


(** Miscellaneous **)
(** Divers **)
(** Micelaneos **)


let rec inc_map f = function
  | E -> E
  | N (x, hs) ->  N (f x, List.map (inc_map f) hs)
