(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Simple heap
 * Tas simple
 * Monton simple
 *
 *)

open BairePervasives

let moduleName = "SimpleHeap"

type 'a heap = E | N of 'a heap * 'a * 'a heap


(** Utilitaires **)
(** Utils **)
(** tiles **)


let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type


(** Categorial operations **)
(** Oprations catgorielles **)
(** Operaciones categoriales **)


let injection = function x -> N (E, x, E)

let rec plus_rec = fun h a x b h' ->
  match h' with
    | E -> h
    | N (c, y, d) ->
	match compare x y with
	  | n when n > 0 -> N (plus_rec h a x b d, y, c)
	  | _ -> N (plus_rec h' c y d b, x, a)

let plus = fun h h' ->
  match h with
    | E -> h'
    | N (a, x, b) -> plus_rec h a x b h'

let zero = function () -> E

let rec to_list = function
  | E -> []
  | N (a, x, b) -> x :: to_list (plus a b)

let rec from_list_rec result = function
  | x :: y :: tail -> from_list_rec (plus x y :: result) tail
  | [x] -> from_list_rec (x :: result) []
  | [] -> 
      match result with
	| [x] -> x
	| _ -> from_list_rec [] result

let from_list = function liste -> from_list_rec [] (List.map injection liste)

let map f = function heap -> from_list (List.map f (to_list heap))

let rec join_rec result = function
  | E -> result
  | N (a, x, b) -> join_rec (join_rec (plus x result) a) b 

let join = function tree -> join_rec E tree


(** Acces **)
(** Accs **)
(** Acceso **)


let rec make_rec a x b = function
  | 1 -> N (a, x, b)
  | n when n mod 2 = 0 -> make_rec a x (N (b, x, b)) (n / 2)
  | n -> make_rec (N (a, x, b)) x (N (b, x, b)) ((n - 1) / 2)

let make = fun n x ->
  if n < 0 then invalid "make" Negative
  else make_rec E x E n

let rec insert x = function
  | E -> injection x
  | N (a, y, b) as node->
      match compare y x with
	| n when n > 0 -> N (node, x, E)
	| _ -> N (insert x b, y, a)

let insert_list list = function heap -> plus (from_list list) heap

let rec delete_rec x = function
  | E -> raise Exit
  | N (a, y, b) ->
      match compare x y with
	| n when n < 0 -> raise Exit
	| 0 -> plus a b
	| _ -> 
	    try N (a, y, delete_rec x b) 
	    with Exit -> N (delete_rec x a, y, b)
	      
let delete x = function
  | E -> E
  | heap -> try delete_rec x heap with Exit -> heap

let rec delete_all_rec x = function
  | E -> raise Exit
  | N (a, y, b) as node ->
      let 
	left = try delete_all_rec x a with Exit -> a and
	right = try delete_all_rec x b with Exit -> b
      in
	match compare x y with
	  | 0 -> plus left right
	  | _ when a == left && b == right -> node
	  | _ -> N (right, y, left)

let delete_all x = function heap ->
  try delete_all_rec x heap 
  with Exit -> heap

let delete_min = function
  | E -> raise Empty
  | N (a, x, b) -> plus a b

let rec filter f = function
  | E -> E
  | N (a, x, b) when f x -> N (filter f a, x, filter f b)
  | N (a, _, b) -> plus (filter f a) (filter f b)


(** View **)
(** Vues **)
(** Vistas **)


let min_view = function
  | E -> raise Empty
  | N (a, x, b) -> (x, plus a b)

let rec max_view = function
  | E -> raise Empty
  | N (E, x, E) -> (E, x)
  | N (a, x, E) -> let (a', y) = max_view a in (N (a', x, E), y)
  | N (E, x, a) -> let (a', y) = max_view a in (N (a', x, E), y)
  | N (a, x, b) ->
      let 
	(a', y) = max_view a and
	(b', z) = max_view b
      in
	match compare z y with
	  | n when n > 0 -> (N (a', x, b), y)
	  | _ -> (N (a, x, b'), z)

let delete_max = function heap ->
  try let (max_heap, _) = max_view heap in max_heap
  with Empty -> E

let rec partition f = function
  | E -> (E, E)
  | N (a, x, b) ->
      let 
	(a0, a1) = partition f a and
	(b0, b1) = partition f b
      in
	if f x then (N (a0, x, b0), plus a1 b1)
	else  (plus a0 b0, N (a1, x, b1))


(** Observers **)
(** Observateurs **)
(** Observadores **)


let is_empty = function
  | E -> true
  | N _ -> false

let rec size_rec n = function
  | E -> n
  | N (a, _, b) -> size_rec (size_rec (n + 1) b) a

let size = function heap -> size_rec 0 heap

let rec member x = function
  | E -> false
  | N (a, y, b) ->
      match compare x y with
	| n when n < 0 -> false
	| 0 -> true
	| _ -> member x a || member x b

let rec count_rec total x = function
  | E -> total
  | N (a, y, b) ->
      match compare x y with
	| n when n < 0 -> total
	| 0 -> count_rec (1 + count_rec total x a) x b
	| _ -> count_rec (count_rec total x a) x b

let count x = function heap -> count_rec 0 x heap

let rec exists f = function
  | E -> false
  | N (a, x, b) -> f x || exists f a || exists f b

let rec for_all f = function
  | E -> true
  | N (a, x, b) -> f x && for_all f a && for_all f b

let min_element = function
  | E -> raise Empty
  | N (_, x, _) -> x

let rec find_max m = function 
  | E -> m
  | N (E, x, E) ->
      (match compare x m with
	 | n when n > 0 -> x
	 | _ -> m)
  | N (a, x, E) -> find_max m a
  | N (E, x, a) -> find_max m a
  | N (a, x, b) -> find_max (find_max m b) a

let rec find_max_leaf = function
  | E -> raise Empty
  | N (E, x, E) -> x
  | N (a, x, E) -> find_max_leaf a
  | N (E, x, a) -> find_max_leaf a
  | N (a, x, b) -> find_max (find_max_leaf a) b

let rec max_element = function
  | E -> raise Empty
  | N (E, x, E) -> x
  | N (a, x, E) -> max_element a
  | N (E, x, a) -> max_element a
  | N (a, x, b) -> find_max (find_max_leaf a) b


(** Traversals **)
(** Parcours **)
(** Recorridos **)


let rec fold_right = fun f heap e ->
  match heap with
    | E -> e
    | N (a, x, b) -> f x (fold_right f (plus a b) e)

let rec fold_left f e = function
  | E -> e
  | N (a, x, b) -> fold_left f (f e x) (plus a b)

let rec inc_map f = function
  | E -> E
  | N (a, x, b) -> N (inc_map f a, f x, inc_map f b)

let rec iter f = function
  | E -> ()
  | N (a, x, b) -> iter f a ; f x ; iter f b


(** Miscellaneous **)
(** Divers **)
(** Micelaneos **)


let unsafe_insert_min = fun x h -> N (h, x, E)
let rec unsafe_append = fun h h' ->
  match h with
    | E -> h'
    | N (a, x, b) -> N (unsafe_append b h, x, a)

