(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Splay heaps
 * Tas splay
 * Montones splay
 *
 *)

open BairePervasives

let moduleName = "PairingHeap"

type 'a heap = E | N of 'a heap * 'a * 'a heap


(** Utilitaires **)
(** Utils **)
(** tiles **)


let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type

let rec partition_inf_sup pivot = function
  | E -> (E, None, E)
  | N (l, v ,r) as tree ->
      match compare pivot v with
	| n when n < 0 ->
	    (match l with
	       | E -> (E, None, tree)
	       | N (ll, lv, lr) ->
		   match compare pivot lv with
		     | n when n < 0 -> 
			 let (inf, m, sup) = partition_inf_sup pivot ll in 
			   (inf, m, N (N (sup, lv, lr), v, r))
		     | n when n > 0 -> 
			 let (inf, m, sup) = partition_inf_sup pivot lr in
			   (N (ll, lv, inf), m, N (sup, v, r))
		     | _ -> (ll, Some lv, N (lr, v, r))
	    )
	| n when n > 0 ->
	    (match r with
	       | E -> (tree, None, E)
	       | N (rl, rv, rr) ->
		   match compare pivot rv with
		     | n when n < 0 ->
			 let (inf, m, sup) = partition_inf_sup pivot rl in
			   (N (l, v, inf), m, N (sup, rv, rr))
		     | n when n > 0 ->
			 let (inf, m, sup) = partition_inf_sup pivot rr in
			   (N (N (l, v, rl), rv, inf), m, sup)
		     | _ -> (N (l, v, rl), Some rv, rr)
	    )		     
	| _ -> (l, Some v, r)


(** Categorial operations **)
(** Oprations catgorielles **)
(** Operaciones categoriales **)

		  
let injection = function x -> N (E, x, E)

let rec plus = fun h1 h2 ->
  match h1 with
    | E -> h2
    | N (a, x, b) -> 
	match partition_inf_sup x h2 with
	  | (a2, None, b2) -> N (plus a a2, x, plus b b2)
	  | (a2, _, b2) -> N (plus a a2, x, N (E, x, plus b b2))

let zero = function () -> E

let rec to_list_rec result = function
  | E -> result
  | N (a, x, b) -> to_list_rec (x :: to_list_rec result b) a

let to_list = function heap -> to_list_rec [] heap

let rec from_list_rec rsultat = function
  | x :: y :: queue -> from_list_rec (plus x y :: rsultat) queue
  | [x] -> from_list_rec (x :: rsultat) []
  | [] -> 
      match rsultat with
	| [x] -> x
	| _ -> from_list_rec [] rsultat

let from_list = function liste -> from_list_rec [] (List.map injection liste)

let map f = function heap -> from_list (List.map f (to_list heap))

let rec join_rec result = function
  | E -> result
  | N (a, x, b) -> join_rec (join_rec (plus x result) a) b 

let join = function heap -> join_rec E heap

let rec make_rec a x b = function
  | 1 -> N (a, x, b)
  | n when n mod 2 = 0 -> make_rec a x (N (b, x, b)) (n / 2)
  | n -> make_rec (N (a, x, b)) x (N (b, x, b)) ((n - 1) / 2)

let make = fun n x -> 
  if n < 0 then invalid "make" Negative
  else make_rec E x E n


(** View **)
(** Vues **)
(** Vistas **)


let rec min_view_rec = function
  | (E, x, b) ->  (x, b)
  | (N (E, x, b), y, c) -> (x, N (b, y, c))
  | (N (N (a, x, b), y, c), z, d) ->
      let (w, ab) = min_view_rec (a, x, b) in (w, N (ab, y, N (c, z, d)))

let min_view = function
  | E -> raise Empty
  | N (a, x, b) -> min_view_rec (a, x, b)

let rec max_view_rec a x = function
  | E -> (a, x)
  | N (b, y, E) -> (N (a, x, b), y)
  | N (b, y, N (c, z, d)) -> 
      let (cd, w) = max_view_rec c z d in (N (N (a, x, b), y, cd), w)

let max_view = function
  | E -> raise Empty
  | N (a, x, b) -> max_view_rec a x b

let unsafe_append = fun a b ->
  try
    let (a', x) = max_view a in N (a', x, b)
  with Empty -> b

let rec partition f = function
  | E -> (E, E)
  | N (a, x, b) ->
      let 
	(a0, a1) = partition f a and
	(b0, b1) = partition f b
      in
	if f x then (N (a0, x, b0), unsafe_append a1 b1)
	else (unsafe_append a0 b0, N (a1, x, b1))


(** Acces **)
(** Accs **)
(** Acceso **)


let insert x = function tree -> 
  match partition_inf_sup x tree with
    | (_, Some _, _) -> tree
    | (a, _, b) ->  N (a, x, b)

let insert_list liste = function heap -> plus (from_list liste) heap

let delete x = function heap ->
  match partition_inf_sup x heap with
    | (_, None, _) -> heap
    | (E, _, b) -> b
    | (a, _, b) -> unsafe_append a b
  
let rec delete_all x = function heap ->
  match partition_inf_sup x heap with
    | (_, None, _) -> heap
    | (l, _, r) ->
	let 
	  left = delete_all x l and
	  right = delete_all x r 
	in
	  unsafe_append left right

let rec delete_min = function
  | E -> E
  | N (E, x, b) -> b
  | N (a, x, b) -> N (delete_min a, x, b)

let rec delete_max = function
  | E -> E
  | N (a, x, E) -> a
  | N (a, x, b) -> N (a, x, delete_max b)

let rec filter f = function
  | E -> E
  | N (a, x, b) when f x -> N (filter f a, x, filter f b)
  | N (a, _, b) -> unsafe_append (filter f a) (filter f b)


(** Observers **)
(** Observateurs **)
(** Observadores **)


let is_empty = function
  | E -> true
  | N _ -> false

let rec size_rec n = function
  | E -> n
  | N (a, _, b) -> size_rec (size_rec (n + 1) a) b

let size = function heap -> size_rec 0 heap

let rec member x = function
  | E -> false
  | N (a, y, b) ->
      match compare x y with
	| n when n < 0 -> member x a
	| n when n > 0 -> member x b
	| _ -> true

let rec count_rec n x = function
  | E -> n
  | N (a, y, b) ->
      match compare x y with
	| n when n < 0 -> count_rec n x a
	| n when n > 0 -> count_rec n x b
	| _ -> count_rec (count_rec (1 + n) x a) x b

let count x = function heap -> count_rec 0 x heap

let rec min_element = function
  | N (E, x, _) -> x
  | N (a, _, _) -> min_element a
  | _ -> raise Empty

let rec max_element = function
  | N (_, x, E) -> x
  | N (_, _, a) -> max_element a
  | _ -> raise Empty

let rec exists f = function
  | E -> false
  | N (a, x, b) -> f x || exists f a || exists f b

let rec for_all f = function
  | E -> true
  | N (a, x, b) -> f x && for_all f a && for_all f b


(** Traversals **)
(** Parcours **)
(** Recorridos **)


let rec fold_left f e = function
  | E -> e
  | N (a, x, b) -> fold_left f (f (fold_left f e a) x) b

let rec fold_right = fun f heap e ->
  match heap with
    | E -> e 
    | N (a, x, b) -> fold_right f a (f x (fold_right f b e))

let fold = fun f heap e -> fold_right f heap e

let rec iter f = function
  | E -> ()
  | N (a, x, b) -> iter f a ; f x ; iter f b
