(*
 * Join list
 *
 *)

open BairePervasives

type 'a seq =
  | E
  | L of 'a
  | N of 'a seq * 'a seq

  (* Invariant :
   * Un noeud ne contient aucun descendant vide 
   *)
      
let rec to_list_rec result = function
  | E -> result
  | L x -> x :: result
  | N (a, b) -> to_list_rec (to_list_rec result b) a
      
let to_list = function list -> to_list_rec [] list
  
(* la fonction from_list cre un arbre balanc *)
  
let rec from_list_rec k = function
  | [] -> E
  | [x] -> L x 
  | [x ; y] -> N (L x, L y)
  | list ->
      let i = k / 2 in
      let (front, rear) = List.split_at i list in
	N (from_list_rec i front, from_list_rec (k - i) rear)
	  
let from_list = function list -> from_list_rec (List.size list) list
  
let unit = function x -> L x
  
let rec map f = function
  | E -> E
  | L x -> L (f x)
  | N (a, b) -> N (map f a, map f b)
      
let rec join_rec result = function
  | E -> result
  | L x -> N (x, result)
  | N (a, b) -> join_rec (join_rec E b) a
      
let rec join = function list -> join_rec E list

let rec bind = fun list f ->
  match list with
    | E -> E
    | L x -> f x
    | N (a, b) -> N (bind a f, bind b f)
	
let zero = function () -> E
	     
let plus = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | _ -> N (a, b)

let rec make_rec x = function
  | 1 -> L x
  | n when (n mod 2 = 0) -> let a = make_rec x (n / 2) in N (a, a)
  | n -> let a = make_rec x (n / 2) in N (L x, N (a, a))
				       
let make = fun n x ->
  if (n < 0) then invalid_arg "JoinList.make : argument ngatif"
  else if (n = 0) then E
  else make_rec x n

let rec get = fun list n ->
  match list with
    | E -> raise (Out_of_bounds n)
    | L x when (n = 0) -> x
    | L _ -> raise (Out_of_bounds (n - 1))
    | N (a, b) ->
	try get a n
	with Out_of_bounds k -> get b n

(* amliorations possibles *)

let rec set_rec k x = function
  | E when (k = 1) -> (0, L x)
  | E -> (1, E)
  | L _ when (k = 1) -> (0, L x)
  | L _ as feuille -> (k - 1, feuille)
  | N (a, b) ->
      match set_rec k x a with
	| (0, arbre_droit) -> (0, N (arbre_droit, b))
	| (n, _) ->
	    let (m, arbre_gauche) = set_rec n x b in
	      (m, N (a, arbre_gauche))
	      
let set = fun list n x ->
  match set_rec (n + 1) x list with
    | (0, result) -> result 
    | _ -> raise (Out_of_bounds n)
	  
(* la fonction init cre une chane *)

let rec init_rec f result = function
  | 0 -> N (L (f 0), result)
  | k -> init_rec f (N (L (f k), result)) (k - 1)
      
let init = fun n f ->
  if (n < 0) then invalid_arg "JoinList.init : argument ngatif"
  else if (n = 0) then E
  else init_rec f E (n - 1)
    
let cons x = function
  | E -> L x
  | list -> N (L x, list)
      
let snoc = fun list x ->
  match list with
    | E -> L x
    | _ -> N (list, L x)
	
let rec left_view_rec = fun a b ->
  match (a, b) with
    | (E, _) -> raise Empty
    | (L x, _) -> (x, b)
    | (N (x, y), _) -> left_view_rec x (N (y, b))

let left_view = function
  | E -> raise Empty
  | L x -> (x, E)
  | N (a, b) -> left_view_rec a b
      
let rec left_head = function
  | E -> raise Empty
  | L x -> x
  | N (a, b) -> left_head a
      
let left_tail = function
  | E -> raise Empty
  | L _ -> E
  | N (a, b) -> let (_, tail) = left_view_rec a b in tail

let rec right_view_rec = fun a b ->
  match (a, b) with
    | (_, N (b, N (c, d))) -> 
	right_view_rec (N (a, N (b, c))) d
    | (_, N (b, L x)) -> (N (a, b), x)
    | (_, L x) -> (a, x)
    | _ -> raise (Unexpected_pattern "JoinList.right_view_rec")

let right_view = function
  | E -> raise Empty
  | L x -> (E, x)
  | N (a, b) -> right_view_rec a b
      
let rec right_head = function
  | E -> raise Empty
  | L x -> x
  | N (a, b) -> right_head a
      
let right_tail = function
  | E -> raise Empty
  | L _ -> E
  | N (a, b) -> let (tail, _) = right_view_rec a b in tail
							   
let is_empty = function
  | E -> true
  | _ -> false
      
let rec size_rec k = function
  | E -> k
  | L _ -> k + 1
  | N (a, b) -> size_rec (size_rec k a) b
      
let size = function list -> size_rec 0 list
  
let rec reverse = function
  | N (a, b) -> N (reverse b, reverse a)
  | _ as other -> other
      
let reverse_append = fun to_rev to_append -> N (reverse to_rev, to_append)

let reverse_onto = fun to_append to_rev -> N (reverse to_rev, to_append)
  
let concat_map = fun f list -> bind list f

let rec fold_left f e = function
  | E -> e
  | L x -> f e x
  | N (a, b) -> fold_left f (fold_left f e a) b
      
let rec fold_right f e = function
  | E -> e
  | L x -> f x e
  | N (a, b) -> fold_right f (fold_right f e b) a

let rec iter f = function
  | E -> ()
  | L x -> f x
  | N (a, b) -> iter f a ; iter f b
      
let split_at n = function list ->
  let (a, b) = List.split_at n (to_list list) in
    (from_list a, from_list b)
    
let sub = fun dbut fin list -> 
  from_list (List.sub dbut fin (to_list list))

let rec for_all f = function
  | E -> true
  | L x -> f x
  | N (a, b) -> for_all f a && for_all f b

let rec exists f = function
  | E -> false
  | L x -> f x
  | N (a, b) -> exists f a || exists f b

let clean = function
  | N (E, E) -> E
  | N (L x, E) -> L x
  | N (E, L x) -> L x
  | _ as other -> other

let rec filter f = function
  | N (a, b) -> clean (N (filter f a, filter f b))
  | L x as feuille when f x -> feuille
  | _ -> E

let partition f = function list ->
  let (a, b) = List.partition f (to_list list) in
    (from_list a, from_list b)

let take_while f = function list ->
  from_list (List.take_while f (to_list list))

let drop_while f = function list ->
  from_list (List.drop_while f (to_list list))

let split_while f = function list ->
  let (a, b) = List.split_while f (to_list list) in
    (from_list a, from_list b)

let map_with_index = fun f list -> 
  from_list (List.map_with_index f (to_list list))

let iter_with_index = fun f list -> List.iter_with_index f (to_list list)

let filter_with_index = fun f list ->
  from_list (List.filter_with_index f (to_list list))

let fold_right_with_index = fun f e list -> 
  List.fold_right_with_index f e (to_list list)

let fold_left_with_index = fun f e list ->
  List.fold_left_with_index f e (to_list list)

let fill = fun list dbut fin motif ->
  from_list (List.fill (to_list list) dbut fin motif)
  
let blit = fun listA indiceA listB indiceB taille ->
  from_list 
  (List.blit (to_list listA) indiceA (to_list listB) indiceB taille)

let zip = fun listA listB -> 
  from_list (List.zip (to_list listA) (to_list listB))

let unzip = fun list ->
  let (a, b) = List.unzip (to_list list) in
    (from_list a, from_list b)

(** Aliases **)

let empty = zero
let append = plus 
