(*
 * Mutable lists
 *
 *)

open BairePervasives

type 
  'a cell = { mutable content : 'a ; next : 'a seq } and
  'a seq = 'a cell option

(* Les listes mutables se comportent exactement comme les listes
 * natives pour les fonctions usuelles (style fonctionnel)
 *)

(* fonctions utiles *)

let rec copy = function
  | None -> None
  | Some cell -> Some { content = cell.content ; next = copy cell.next }

(* Signature monadique *)

let unit = function x -> Some { content = x ; next = None }
  
let rec map f = function
  | None -> None
  | Some cell -> Some { content = f cell.content ; next =  map f cell.next }

let rec plus = fun a b ->
  match a with
    | None -> copy b
    | Some cell -> Some { content = cell.content ; next = plus cell.next b }

let rec join = function
  | None -> None
  | Some cell -> plus cell.content (join cell.next)

let bind = fun liste f -> join (map f liste)

let zero = None

let rec make_rec x = function
  | 0 -> None
  | k -> Some { content = x ; next = make_rec x (k - 1) }

let make = fun n x -> make_rec x n

let get = fun liste n ->
  let rec get_rec k = function
    | None -> raise (Out_of_bounds n)
    | Some cell when (k = 0) -> cell.content
    | Some cell -> get_rec (k - 1) cell.next
  in
    if (n < 0) then invalid_arg "MutableList.get : argument ngatif"
    else get_rec n liste

let set = fun liste n x ->
  let rec set_rec k = function
    | None -> raise (Out_of_bounds n)
    | Some cell when (k = 0) -> Some { content = x ; next = copy cell.next }
    | Some cell ->
	Some { content = cell.content ; next = set_rec (k - 1) cell.next }
  in
    if (n < 0) then invalid_arg "MutableList.set : argument ngatif"
    else set_rec n liste

(* Constructeurs *)

let cons x = function
  | None -> unit x
  | list -> Some { content = x ; next = copy list }

let rec snoc = fun liste x ->
  match liste with
    | None -> unit x
    | Some cell -> Some { content = cell.content ; next = snoc cell.next x }

let rec from_list = function
  | [] -> None
  | tete :: queue -> Some { content = tete ; next = from_list queue }

let init = fun n f ->
  let rec init_rec = function
    | k when k = n -> None
    | k -> Some { content =  f k ; next = init_rec (k + 1) }
  in
    if (n < 0) then invalid_arg "MutableList.init : argument ngatif"
    else init_rec 0

(* Renversements *)

let rec reverse_onto rsultat = function
  | None -> rsultat
  | Some cell -> 
      reverse_onto 
      (Some { content = cell.content ; next =  rsultat }) 
      cell.next

let rec reverse_append = fun liste rsultat -> reverse_onto rsultat liste

let reverse = function liste -> reverse_onto None liste

let rec reverse_map_rec f rsultat = function
  | None -> rsultat
  | Some cell -> 
      reverse_map_rec f (Some { content = f cell.content ; next = rsultat })
      cell.next

let reverse_map f = function liste -> reverse_map_rec f None liste


(* Destructeurs *)

let left_view = function
  | None -> raise Empty
  | Some cell -> (cell.content, cell.next)

let left_head = function
  | None -> raise Empty
  | Some cell -> cell.content

let left_tail = function
  | None -> raise Empty
  | Some cell -> cell.next

let rec right_rec rsultat = function
  | Some cell when cell.next = None -> (reverse rsultat, cell.content)
  | Some cell -> 
      right_rec (Some { content = cell.content ; next = rsultat }) cell.next
  | _ -> raise Empty

let right_view = function liste -> right_rec None liste

let rec right_head = function
  | None -> raise Empty
  | Some cell when cell.next = None -> cell.content 
  | Some cell -> right_head cell.next

let rec right_tail = function liste ->
  let (rsultat, _) = right_rec None liste in
    rsultat

(* Observateurs *)

let is_empty = function
  | None -> true
  | _ -> false

let rec size_rec k = function
  | None -> k
  | Some cell -> size_rec (k + 1) cell.next

let size = function liste -> size_rec 0 liste

let rec to_list = function
  | None -> []
  | Some cell -> cell.content :: to_list cell.next

(* Folds et maps *)

let rec concat_map f = function
  | None -> None
  | Some cell -> plus (f cell.content) (concat_map f cell.next) 
      

let rec fold_right f e = function
  | None -> e
  | Some cell -> f cell.content (fold_right f e cell.next)

let rec fold_left f e = function
  | None -> e
  | Some cell -> fold_left f (f e cell.content) cell.next

let reverse_fold_left = fun f e liste -> 
  fold_left f e (reverse_onto None liste)

let reverse_fold_right = fun f e liste -> 
  fold_right f e (reverse_onto None liste)

let rec iter f = function
  | None -> ()
  | Some cell -> f cell.content ; iter f cell.next

(* Sous squences *)

let take n = function liste ->
  let rec take_rec k = function
    | None -> raise (Out_of_bounds n)
    | Some cell when k = 0 -> None
    | Some cell ->
	Some { content = cell.content ; next = take_rec (k - 1) cell.next }
  in
    if n < 0 then invalid_arg "MutableList.take : argument ngatif"
    else take_rec n liste

let drop n = function liste ->
  let rec drop_rec k = function
    | None when k = 0 -> None
    | None -> raise (Out_of_bounds n)
    | Some _ as liste when k = 0 -> copy liste
    | Some cell -> drop_rec (k - 1) cell.next
  in
    if n < 0 then invalid_arg "MutableList.drop : argument ngatif"
    else drop_rec n liste

let split_at n = function liste ->
  let rec split_rec = function
    | (0, droite, gauche) -> (reverse droite, gauche)
    | (k, droite, Some cell) -> 
	split_rec 
	(k - 1, Some { content = cell.content ; next = droite }, 
	 cell.next)
    | _ -> raise (Out_of_bounds n)
  in
    split_rec (n, None, liste)

let sub = fun dbut fin liste ->
  let rec sub_rec n = function
    | None -> None
    | Some cell when (n < dbut) -> sub_rec (n + 1) cell.next
    | _ when (n >= fin) -> None
    | Some cell ->
	Some { content = cell.content ; next = sub_rec (n + 1) cell.next }
  in
    if (dbut < 0) then invalid_arg "MutableList.sub : argument ngatif"
    else if (dbut > fin) then invalid_arg "MutableList.sub : dbut > fin"
    else sub_rec 0 liste

(* Prdicats *)

let rec exists f = function
  | None -> false
  | Some cell -> (f cell.content) || (exists f cell.next)

let rec for_all f = function
  | None -> true
  | Some cell -> (f cell.content) && (for_all f cell.next)

let rec find f = function
  | None -> raise Not_found
  | Some cell when f cell.content -> cell.content
  | Some cell -> find f cell.next

let rec filter f = function
  | None -> None
  | Some cell when (f cell.content) -> 
      Some { content = cell.content ; next =  filter f cell.next }
  | Some cell -> filter f cell.next

let rec partition_rec f vrai faux = function
  | None -> (reverse vrai, reverse faux)
  | Some cell when (f cell.content) ->
      partition_rec f 
      (Some { content = cell.content ; next = vrai }) faux cell.next
  | Some cell ->
      partition_rec f vrai 
      (Some { content = cell.content ; next = faux }) cell.next
      
let partition f = function liste -> partition_rec f None None liste

let rec take_while f = function
  | Some cell when (f cell.content) -> 
      Some { content = cell.content ; next = take_while f cell.next }
  | _ -> None

let rec drop_while f = function
  | Some cell when not (f cell.content) -> drop_while f cell.next
  | liste -> copy liste

let rec split_rec f gauche = function
  | Some cell when f cell.content ->
      split_rec f (Some { content = cell.content ; next = gauche }) cell.next
  | droite -> (reverse gauche, droite)

let split_while = fun f liste -> split_rec f None liste

(* Fonctions indexes *)

let rec mapi_rec f k = function
  | None -> None
  | Some cell -> Some { content = f k cell.content ;
			next = mapi_rec f (k + 1) cell.next }

let map_with_index = fun f liste -> mapi_rec f 0 liste

let rec iteri_rec f k = function
  | None -> ()
  | Some cell -> f k cell.content ; iteri_rec f (k + 1) cell.next

let iter_with_index = fun f liste -> iteri_rec f 0 liste

let rec filteri_rec f k = function
  | None -> None
  | Some cell when f k cell.content -> 
      Some { content = cell.content ; next = filteri_rec f (k + 1) cell.next }
  | Some cell -> filteri_rec f (k + 1) cell.next

let filter_with_index = fun f liste -> filteri_rec f 0 liste

let rec findi_rec f k = function
  | None -> raise Not_found
  | Some cell when f k cell.content -> cell.content
  | Some cell -> findi_rec f (k + 1) cell.next

let find_with_index = fun f liste -> findi_rec f 0 liste

let rec foldri f e k = function
  | None -> e
  | Some cell -> f k cell.content (foldri f e (k + 1) cell.next)

let fold_right_with_index f e = function liste -> foldri f e 0 liste

let rec foldli f e k = function
  | None -> e
  | Some cell -> foldli f (f k e cell.content) (k + 1) cell.next

let fold_left_with_index f e = function liste -> foldli f e 0 liste

let fill = fun liste dbut fin motif ->
  let rec fill_rec k = function
    | None -> None
    | Some cell when (k >= dbut) && (k < fin) -> 
	Some { content = motif ; next = fill_rec (k + 1) cell.next }
    | Some cell ->
	Some { content = cell.content ; next  = fill_rec (k + 1) cell.next }
  in 
    if (dbut < 0) then invalid_arg "MutableList.fill : argument ngatif"
    else if (dbut > fin) then invalid_arg "MutableList.fill : dbut > fin"
    else fill_rec 0 liste

let blit = fun listeA indiceA listeB indiceB taille ->
  let rec 
    decale = function
      | (_, _, None, _) -> raise (Out_of_bounds indiceA)
      | (_, _, _, None) -> raise (Out_of_bounds indiceB)
      | (0, 0, listeA, listeB) -> copie (taille, listeA, listeB)
      | (n, 0, Some cellA, liste) ->
	  decale (n - 1, 0, cellA.next, liste)
      | (0, m, liste, Some cellB) -> 
          Some ({ content = cellB.content ; 
		  next = decale (0, m - 1, liste, cellB.next)})
      | (n, m, Some cellA, Some cellB) -> 
          Some ({ content = cellB.content ;
		  next = decale (n - 1, m - 1, cellA.next, cellB.next)})
  and
    copie = function
      | (0, _, listeB) -> listeB 
      | (k, Some cellA, Some cellB) 
        -> Some ({ content = cellA.content ;
		   next = copie (k - 1, cellA.next, cellB.next)})
      | _ -> raise (Out_of_bounds taille)
  in
    decale (indiceA, indiceB, listeA, listeB)

      
(* Combinaisons *)
      
(* Ne lve pas d'exception si les listes ne sont pas de la meme taille *)

let rec unsafe_zip = fun listeA listeB ->
  match (listeA, listeB) with
    | (None, _) -> None
    | (_, None) -> None
    | (Some cellA, Some cellB) -> 
	Some { content = (cellA.content, cellB.content) ; 
	       next = unsafe_zip cellA.next cellB.next }
	
let rec zip = fun listeA listeB ->
  match (listeA, listeB) with
    | (None, None) -> None
    | (Some cellA, Some cellB) ->
	Some { content = (cellA.content, cellB.content) ;
	       next = zip cellA.next cellB.next }
    | _ -> 
	invalid_arg "MutableList.zip : les squences doivent tre de meme taille"

let rec unzip_rec listeA listeB = function
  | None -> (reverse listeA, reverse listeB)
  | Some cell ->
      let (x, y) = cell.content in
	unzip_rec 
	  (Some { content = x ; next = listeA }) 
	  (Some { content = y ; next = listeB }) 
	  cell.next

let unzip = function liste -> unzip_rec None None liste


(** Aliases **)

let empty = zero
let append = plus


(* Oprations sur les collections *)

let insert = cons

let insert_list = plus

let union = plus

let rec delete x = function
  | None -> None
  | Some cell when cell.content = x -> copy cell.next
  | Some cell -> Some { content = cell.content ; next = delete x cell.next }
      
let rec delete_q x = function
  | None -> None
  | Some cell when cell.content == x -> copy cell.next
  | Some cell -> Some { content = cell.content ; 
			next = delete_q x cell.next }
      
let rec delete_all x = function
  | None -> None
  | Some cell when cell.content = x -> delete_all x cell.next
  | Some cell -> Some { content = cell.content ; 
			next = delete_all x cell.next }

let rec delete_all_q x = function
  | None -> None
  | Some cell when cell.content = x -> delete_all_q x cell.next
  | Some cell -> Some { content = cell.content ;
			next = delete_all_q x cell.next } 

let rec member x = function
  | None -> false
  | Some cell when cell.content = x -> true
  | Some cell -> member x cell.next

let rec member_q x = function
  | None -> false
  | Some cell when cell.content == x -> true
  | Some cell -> member_q x cell.next

let rec count_rec x n = function
  | None -> n
  | Some cell when cell.content = x -> count_rec x (n + 1) cell.next
  | Some cell -> count_rec x n cell.next

let count x = function liste -> count_rec x 0

let rec count_q_rec x n = function
  | None -> n
  | Some cell when cell.content == x -> count_q_rec x (n + 1) cell.next
  | Some cell -> count_q_rec x n cell.next

let count_q x = function liste -> count_q_rec x 0

let rec delete_repeated_rec rsultat f = function
  | None -> reverse rsultat
  | Some cell when f cell.content -> delete_repeated_rec rsultat f cell.next
  | Some cell -> delete_repeated_rec 
      (Some { content = cell.content ; next = rsultat}) 
      (function x -> x = cell.content || f x) cell.next

let delete_repeated = function liste -> 
  delete_repeated_rec None (fun n -> false) liste

let fold = fold_left

(* Oprations sur les tables d'associations *)

type ('k, 'a) assoc = ('k * 'a) cell option

let unit_assoc = fun cl x -> Some { content = (cl, x) ; next = None }

let rec map_assoc f = function
  | None -> None
  | Some ({ content = (k, x) } as cell) -> 
      Some { content = (k, f x) ; next = map_assoc f cell.next }

let zero_assoc = None

let plus_assoc = plus

let insert_assoc = fun cl x liste -> 
  Some ({ content = (x, cl) ; next = liste })

let rec insert_list_assoc = fun liste collection ->
  match liste with
    | [] -> collection
    | tete :: queue -> Some { content = tete ; 
			      next = insert_list_assoc queue collection }
let union_assoc = plus
		    
(* ne lve pas d'exception si l'lment n'a pas t trouv *)

let rec delete_assoc_rec cl rsultat = function
  | None -> reverse rsultat
  | Some ({ content = (k, _) } as cell) when (k = cl) -> 
      reverse_onto cell.next rsultat
  | Some cell -> delete_assoc_rec cl 
      (Some ({ content = cell.content ; next = rsultat})) cell.next

let delete_assoc = fun liste cl -> delete_assoc_rec cl None liste
  
let rec delete_q_assoc_rec cl rsultat = function
  | None -> reverse rsultat
  | Some ({ content = (k, _) } as cell) when (k == cl) -> 
      reverse_onto cell.next rsultat
  | Some cell -> delete_q_assoc_rec cl 
      (Some ({ content = cell.content ; next = rsultat})) cell.next

let delete_q_assoc = fun liste cl -> delete_q_assoc_rec cl None liste

let rec delete_all_assoc_rec cl rsultat = function
  | None -> reverse rsultat
  | Some ({ content = (k, _) } as cell) when (k = cl) -> 
      delete_all_assoc_rec cl rsultat cell.next
  | Some cell -> delete_all_assoc_rec cl 
      (Some ({ content = cell.content ; next = rsultat})) cell.next

let delete_all_assoc = fun liste cl -> delete_all_assoc_rec cl None liste

let rec delete_all_q_assoc_rec cl rsultat = function
  | None -> reverse rsultat
  | Some ({ content = (k, _) } as cell) when (k == cl) -> 
      delete_all_q_assoc_rec cl rsultat cell.next
  | Some cell -> delete_all_q_assoc_rec cl 
      (Some ({ content = cell.content ; next = rsultat})) cell.next

let is_empty_assoc = is_empty

let size_assoc = size

let rec member_assoc = fun liste cl ->
  match liste with
    | None -> false
    | Some { content = (k, _) } when k = cl -> true
    | Some cell -> member_assoc cell.next cl

let rec member_q_assoc = fun liste cl ->
  match liste with
    | None -> false
    | Some { content = (k, _) } when k == cl -> true
    | Some cell -> member_q_assoc cell.next cl

let rec count_assoc_rec cl n = function
  | None -> n
  | Some ({ content = (k, _) } as cell) when k = cl ->
      count_assoc_rec cl (n + 1) cell.next
  | Some cell -> count_assoc_rec cl n cell.next

let count_assoc = fun liste cl -> count_assoc_rec cl 0 liste

let rec count_q_assoc_rec cl n = function
  | None -> n
  | Some ({ content = (k, _) } as cell) when k == cl ->
      count_q_assoc_rec cl (n + 1) cell.next
  | Some cell -> count_q_assoc_rec cl n cell.next

let count_q_assoc = fun liste cl -> count_q_assoc_rec cl 0 liste

let rec lookup_assoc = fun liste cl ->
  match liste with
    | None -> raise Not_found
    | Some { content = (k, x) } when (k = cl) -> x
    | Some cell -> lookup_assoc cell.next cl

let rec lookup_q_assoc = fun liste cl ->
  match liste with
    | None -> raise Not_found
    | Some { content = (k, x) } when (k == cl) -> x
    | Some cell -> lookup_q_assoc cell.next cl

let rec lookup_all_assoc = fun liste cl ->
  match liste with
    | None -> raise Not_found
    | Some ({ content = (k, x) } as cell) when (k = cl) -> 
	Some { content = x ; next = lookup_all_assoc cell.next cl }
    | Some cell -> lookup_all_assoc cell.next cl

let rec lookup_all_q_assoc = fun liste cl ->
  match liste with
    | None -> raise Not_found
    | Some ({ content = (k, x) } as cell) when (k = cl) -> 
	Some { content = x ; next = lookup_all_q_assoc cell.next cl }
    | Some cell -> lookup_all_q_assoc cell.next cl

let rec elements_assoc = function
  | None -> None
  | Some ({ content = (_, x) } as cell) -> 
      Some { content =  x ; next = elements_assoc cell.next }

let rec adjust_assoc f cl = function
  | None -> None
  | Some ({ content = (k, x) } as cell) when (k = cl) ->
      Some { content = (cl, f x) ; next = copy cell.next }
  | Some cell -> Some { content = cell.content ; 
			next = adjust_assoc f cl cell.next }

let rec adjust_all_assoc f cl = function
  | None -> None
  | Some ({ content = (k, x) } as cell) when (k = cl) ->
      Some { content = (cl, f x) ; next = adjust_all_assoc f cl cell.next }
  | Some cell -> Some { content = cell.content ; 
			next = adjust_all_assoc f cl cell.next }

let rec fold_assoc f e = function
  | None -> e
  | Some ({ content = (_, x) } as cell) -> fold_assoc f (f e x) cell.next

let rec filter_keys_assoc f = function
  | None -> None
  | Some ({ content = (k, _) } as cell) when f k -> 
      filter_keys_assoc f cell.next
  | Some cell -> Some { content = cell.content ; 
			next = filter_keys_assoc f cell.next }

let rec filter_elements_assoc f = function
  | None -> None
  | Some ({ content = (_, x)} as cell) when (f x) ->
      Some { content = cell.content ; 
	     next = filter_elements_assoc f cell.next }
  | Some cell -> filter_elements_assoc f cell.next

let delete_list_assoc = fun liste to_delete ->
  let rec make_deletion_list f = function
    | [] -> f
    | tete :: queue when f tete -> make_deletion_list f queue
    | tete :: queue -> make_deletion_list
        (fun x -> x <> tete && f x) queue
  in
  let filter_function = make_deletion_list (fun x -> true) to_delete
  in 
    filter_keys_assoc filter_function liste

let partition_keys_assoc = fun f liste ->
  let rec part_rec = function
    | (None, gauche, droite) -> (gauche, droite)
    | (Some ({ content = (k, x) } as cell), gauche, droite) when f k ->
	part_rec (cell.next, 
		 Some { content = (k, x) ; next = gauche}, 
		 droite)
    | (Some cell, gauche, droite) ->
	part_rec (cell.next, 
		 gauche, 
		 Some { content = cell.content ; next = droite})
  in
    part_rec (liste, None, None)


(* Messages d'erreur *)

let librairie_alternative = ["SimplyLinkedCircularList.ml" ; "?.ml"]
let sparateur = "or"
let invite = "try"

let list_to_string = function
  | "" -> 
      let rec to_string = function
	| [] -> ""
	| tete :: queue -> tete ^ to_string queue
      in
	to_string
  | sep ->
      let rec to_string = function
	| [] -> ""
	| [tete] -> tete
	| tete :: queue -> tete ^ " " ^ sep ^ " " ^ to_string queue
      in
	to_string

let message_erreur = 
  match List.size librairie_alternative with
    | 0 -> ""
    | 1 -> " : " ^ invite ^ " " ^ List.get librairie_alternative 0
    | _ ->
	let append_strings = list_to_string sparateur in
	  " : " ^ invite ^ " " ^ append_strings librairie_alternative

let not_implemented = function nom -> 
  raise (Not_working (nom ^ message_erreur))

let zero_side = function () -> None
  
let set_side n x = function liste ->
  let rec set_rec k = function
    | None -> raise (Out_of_bounds n)
    | Some cell when k = 0 -> cell.content <- x
    | Some cell -> set_rec (k - 1) cell.next
  in
    if (n < 0) then invalid_arg "MutableList.set : argument ngatif"
    else set_rec n liste ; 
    liste
      
let rec copy_into = fun a b ->
  match (a, b) with
    | (None, None) -> ()
    | (Some cellA, Some cellB) -> 
	cellA.content <- cellB.content ; copy_into cellA.next cellB.next
    | _ -> 
	invalid_arg "MutableList.copy_into : les listes doivent tre de mme taille"

let reverse_side = function liste ->
  let temp = reverse_onto None liste in
    copy_into liste temp ;
    liste

let rec adjust_side f = function liste ->
  match liste with
    | None -> liste
    | Some cell -> cell.content <- f cell.content ; adjust_side f cell.next

let rec adjusti_rec_side f k = function
  | None -> ()
  | Some cell -> cell.content <- f k cell.content ; 
      adjusti_rec_side f (k + 1) cell.next

let adjust_with_index_side f = function liste -> 
  adjusti_rec_side f 0 liste ; liste

(* Ne lve aucune exception *)

let fill_side = fun liste dbut fin motif ->
  let rec fill_rec k = function
    | Some cell when k < dbut -> fill_rec (k + 1) cell.next
    | Some cell when (k > dbut) && (k < fin) -> 
	cell.content <- motif ;
	fill_rec (k + 1) cell.next
    | _ -> ()
  in
    fill_rec 0 liste ;
    liste

let rec adjust_assoc_side f cl = function assoc ->
  match assoc with
    | None -> assoc
    | Some ({ content = (k, x) } as cell) when (k = cl) ->
	cell.content <- (cl, f x) ; assoc
    | Some cell -> adjust_assoc_side f cl cell.next

let rec adjust_all_assoc_side f cl = function assoc ->
  match assoc with
    | None -> assoc
    | Some ({ content = (k, x) } as cell) when (k = cl) ->
	cell.content <- (cl, f x) ; adjust_all_assoc_side f cl cell.next
    | Some cell -> adjust_all_assoc_side f cl cell.next

