(*
 * Listes mutables
 *
 *)

open BairePervasives

type 'a seq =
  | Vide
  | Noeud of 'a ref * 'a seq

let unit = function x -> Noeud (ref x, Vide)
  
let rec map f = function
  | Vide -> Vide
  | Noeud (x, suivant) -> Noeud (ref (f !x), map f suivant)

let rec plus = fun a b ->
  match a with
    | Vide -> b
    | Noeud (x, suivant) -> Noeud (x, plus suivant b)

let rec join = function
  | Vide -> Vide
  | Noeud (x, suivant) -> plus !x (join suivant)

let bind = fun liste f -> join (map f liste)

let rec make_rec x = function
  | 0 -> Vide
  | k -> Noeud (ref x, make_rec x (k - 1))

let make = fun n x -> make_rec x n

let get n = function liste ->
  let rec get_rec k = function
    | Vide -> raise (Out_of_bounds n)
    | Noeud (x, _) when k = 0 -> !x
    | Noeud (_, suivant) -> get_rec (k - 1) suivant
  in
    if (n < 0) then invalid_arg "get : argument ngatif"
    else get_rec n liste

let set n x = function liste ->
  let rec set_rec k = function
    | Vide -> raise (Out_of_bounds n)
    | Noeud (_, suivant) when k = 0 -> Noeud (ref x, suivant)
    | Noeud (v, suivant) -> Noeud (v, set_rec (k - 1) suivant)
  in
    if (n < 0) then invalid_arg "set : argument ngatif"
    else set_rec n liste

let cons x = function
  | Vide -> Noeud (ref x, Vide)
  | Noeud _ as noeud -> Noeud (ref x, noeud)

let rec snoc = fun liste x ->
  match liste with
    | Vide -> Noeud (ref x, Vide)
    | Noeud (v, suivant) -> Noeud (v, snoc suivant x)

let rec from_list = function
  | [] -> Vide
  | tete :: queue -> Noeud (ref tete, from_list queue)

let init = fun n f ->
  let rec init_rec = function
    | k when k = n -> Vide
    | k -> Noeud (ref (f k), init_rec (k - 1))
  in
    if (n < 0) then invalid_arg "init : argument ngatif"
    else init_rec 0

let rec reverse_onto rsultat = function
  | Vide -> rsultat
  | Noeud (x, suivant) -> reverse_onto (Noeud (x, rsultat)) suivant

let reverse = function liste -> reverse_onto Vide liste

let reverse_append = fun rev app -> reverse_onto app rev

let left_view = function
  | Vide -> raise Empty
  | Noeud (x, suivant) -> (!x, suivant)

let left_head = function
  | Vide -> raise Empty
  | Noeud (x, _) -> !x

let left_tail = function
  | Vide -> raise Empty
  | Noeud (_, suivant) -> suivant

let rec right_rec rsultat = function
  | Noeud (x, Vide) -> (reverse rsultat, !x)
  | Noeud (x, suivant) -> right_rec (Noeud (x, rsultat)) suivant
  | _ -> raise Empty

let right_view = function liste -> right_rec Vide liste

let rec right_head = function
  | Vide -> raise Empty
  | Noeud (x, Vide) -> !x
  | Noeud (_, suivant) -> right_head suivant

let rec right_tail = function liste ->
  let (rsultat, _) = right_rec Vide liste in
    rsultat

let is_empty = function
  | Vide -> true
  | _ -> false

let rec size_rec k = function
  | Vide -> k
  | Noeud (_, suivant) -> size_rec (k + 1) suivant

let size = function liste -> size_rec 0 liste

let rec to_list = function
  | Vide -> []
  | Noeud (x, suivant) -> !x :: to_list suivant

let rec concat_map f = function
  | Vide -> Vide
  | Noeud (x, suivant) -> plus (f !x) (concat_map f suivant) 

let rec fold_right f e = function
  | Vide -> e
  | Noeud (x, suivant) -> f !x (fold_right f e suivant)

let rec fold_left f e = function
  | Vide -> e
  | Noeud (x, suivant) -> fold_left f (f e !x) suivant

let rec iter f = function
  | Vide -> ()
  | Noeud (x, suivant) -> f !x ; iter f suivant

let split_at n = function liste ->
  let rec split_rec = function
    | (0, droite, gauche) -> (reverse droite, gauche)
    | (k, droite, Noeud (x, suivant)) -> 
	split_rec (k - 1, Noeud (x, droite), suivant)
    | _ -> raise (Out_of_bounds n)
  in
    split_rec (n, Vide, liste)

let sub = fun dbut fin liste ->
  let rec sub_rec = function
    | (_, Vide) -> Vide
    | (n, Noeud (_, suivant)) when (n < dbut) -> sub_rec (n + 1, suivant)
    | (n, _) when (n >= fin) -> Vide
    | (n, Noeud (x, suivant)) -> Noeud (x, sub_rec (n + 1, suivant))
  in
    if (dbut < 0) then invalid_arg "sub : argument ngatif"
    else if (dbut > fin) then invalid_arg "sub : dbut > fin"
    else sub_rec (0, liste)

let rec exists f = function
  | Vide -> false
  | Noeud (x, suivant) -> (f !x) || (exists f suivant)

let rec for_all f = function
  | Vide -> true
  | Noeud (x, suivant) -> (f !x) && (for_all f suivant)

let rec filter f = function
  | Vide -> Vide
  | Noeud (x, suivant) when (f !x) -> Noeud (x, filter f suivant)
  | Noeud (_, suivant) -> filter f suivant

let rec partition_rec f vrai faux = function
  | Vide -> (reverse vrai, reverse faux)
  | Noeud (x, suivant) when (f !x) -> 
      partition_rec f (Noeud (x, vrai)) faux suivant
  | Noeud (x, suivant) -> partition_rec f vrai (Noeud (x, faux)) suivant
      
let partition f = function liste -> partition_rec f Vide Vide liste

let rec take_while f = function
  | Noeud (x, suivant) when (f !x) -> Noeud (x, take_while f suivant)
  | _ -> Vide

let rec drop_while f = function
  | Noeud (x, suivant) when not (f !x) -> drop_while f suivant
  | liste -> liste

let rec split_rec f gauche = function
  | Noeud (x, suivant) when f !x -> split_rec f (Noeud (x, gauche)) suivant
  | droite -> (reverse gauche, droite)

let split_while = fun f liste -> split_rec f Vide liste

let rec mapi_rec f k = function
  | Vide -> Vide
  | Noeud (x, suivant) -> Noeud (f k !x, mapi_rec f (k + 1) suivant)

let map_with_index = fun f liste -> mapi_rec f 0 liste

let rec iteri_rec f k = function
  | Vide -> ()
  | Noeud (x, suivant) -> f k !x ; iteri_rec f (k + 1) suivant

let iter_with_index = fun f liste -> iteri_rec f 0 liste

let rec filteri_rec f k = function
  | Vide -> Vide
  | Noeud (x, suivant) when f k !x -> Noeud (x, filteri_rec f (k + 1) suivant)
  | Noeud (_, suivant) -> filteri_rec f (k + 1) suivant

let filter_with_index = fun f liste -> filteri_rec f 0 liste

let rec foldri f e k = function
  | Vide -> e
  | Noeud (x, suivant) -> f k !x (foldri f e (k + 1) suivant)

let fold_right_with_index f e = function liste -> foldri f e 0 liste

let rec foldli f e k = function
  | Vide -> e
  | Noeud (x, suivant) -> foldli f (f k e !x) (k + 1) suivant

let fold_left_with_index f e = function liste -> foldli f e 0 liste

let fill = fun liste dbut fin motif ->
  let rec fill_rec = function
    | (_, Vide) -> Vide
    | (k, Noeud (_ , suivant)) when (k >= dbut) && (k < fin) -> 
	Noeud (motif, fill_rec (k + 1, suivant))
    | (k, Noeud (x, suivant)) -> Noeud (x, fill_rec (k + 1, suivant))
  in 
    if (dbut < 0) then invalid_arg "fill : argument ngatif"
    else if (dbut > fin) then invalid_arg "fill : dbut > fin"
    else fill_rec (0, liste)
      
let blit = fun listeA indiceA listeB indiceB taille ->
  let rec 
    decale = function
      | (_, _, Vide, _) -> raise (Out_of_bounds indiceA)
      | (_, _, _, Vide) -> raise (Out_of_bounds indiceB)
      | (0, 0, listeA, listeB) -> copie (taille, listeA, listeB)
      | (n, 0, Noeud (_, suivantA), liste) -> 
	  decale (n - 1, 0, suivantA, liste)
      | (0, m, liste, Noeud (teteB, suivantB)) -> 
	  Noeud (teteB, decale (0, m - 1, liste, suivantB))
      | (n, m, Noeud (_, suivantA), Noeud (teteB, suivantB)) -> 
	  Noeud (teteB, decale (n - 1, m - 1, suivantA, suivantB))
  and
    copie = function
      | (0, _, listeB) -> listeB 
      | (k, Noeud (teteA, suivantA), Noeud (_, suivantB)) -> 
	  Noeud (teteA, copie (k - 1, suivantA, suivantB))
      | _ -> raise (Out_of_bounds taille)
  in
    decale (indiceA, indiceB, listeA, listeB)

(* Combinaisons *)
      
(* Ne lve pas d'exception si les listes ne sont pas de la meme taille *)

let rec unsafe_zip = fun listeA listeB ->
  match (listeA, listeB) with
    | (Vide, _) -> Vide
    | (_, Vide) -> Vide
    | (Noeud (teteA, suivantA), Noeud (teteB, suivantB)) -> 
	Noeud (ref (!teteA, !teteB), unsafe_zip suivantA suivantB)
	
let rec zip = fun listeA listeB ->
  match (listeA, listeB) with
    | (Vide, Vide) -> Vide
    | (Noeud (teteA, suivantA), Noeud (teteB, suivantB)) ->
	Noeud (ref (!teteA, !teteB), zip suivantA suivantB)
    | _ ->
	invalid_arg "zip : les squences doivent tre de mme taille"

let rec unzip_rec listeA listeB = function
  | Vide -> (reverse listeA, reverse listeB)
  | Noeud (valeur, suivant) ->
      let (x, y) = !valeur in
	unzip_rec (Noeud (ref x, listeA)) (Noeud (ref y, listeB)) suivant

let unzip = function liste -> unzip_rec Vide Vide liste
    
