(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * AVL mutable map
 * Ensemble AVL mutable
 * Conjunto AVL mutable
 *
 *)

open BairePervasives
   
(* AVL trees are binary search trees in which the balance information
 * is encoded by the height of the tree. The difference of height between
 * the two subtrees of a node must not be more than 1
 * The total height of the tree is then inferior to 1.44 log n
 * 
 * AVL trees can be seen as a specialisation of cartesian trees with
 * the given height function (confer cartesianTree.ml)
 * 
 * let rec height = function
 *   | E -> 0
 *   | N (a, x, b) -> 1 + max (height a) (height b)
 * 
 * This implementation follows in many points the implementation by Xavier
 * Leroy in the Caml standard library. Some optimisations have been eliminated
 * (mostly function in lining) for more readabilty and some algorithms
 * have been improved.
 *)

(* Les arbres AVL sont des arbres dont l'information de rquilibrage
 * encode la hauteur de l'arbre. La diffrence de hauteur entre
 * les sous-arbres gauche et droit ne peut dpasser 1
 * La hauteur de l'arbre est donc majore par 1.44 log n
 *
 * Les arbres AVL peuvent tre obtenus comme une spcialisation des arbres
 * cartsiens avec pour hauteur associe (voir cartesianTree.ml)
 * 
 * let rec hauteur = function 
 *   | E -> 0 
 *   | N (a, x, b) -> 1 + max (hauteur a) (hauteur b)
 * 
 * Cette implmentation est en grande partie reprise de celle de Xavier Leroy
 * dans la librairie standard de Caml. Cependant, un certain nombre 
 * d'optimisations ont t limines (expansion en ligne de fonctions) 
 * pour plus de clart et certains algorithmes ont t amliors
 *)

(* Los rboles AVL son rboles binarios de busqueda cuya informacin de 
 * equilbrio se ha codificado en la altura del rbol. La diferencia de
 * altura entre los subrboles izquierdo y derecho de cada nudo no debe
 * ser mayor que 1
 * Razn por la cual la altura total del rbol no es mayor que 1.44 log n
 * 
 * Los rboles AVL pueden ser obtenidos especializando los rboles cartesianos
 * con la funcin de altura siguiente (ver cartesianTree.ml)
 * 
 * let rec altura = function
 *   | E -> 0
 *   | N (a, x, b) -> 1 + max (altura a) (altura b)
 * 
 * sta implementacin se inspira grandemente de la de Xavier Leroy en la
 * librera estndar de Caml. Nobstante, ciertas optimizaciones has sido
 * eliminadas (esencialmente la expansin en lnea de algunas funciones)
 * para mayor lisibilidad, ademas ciertos algortmos han sido mejorados.
 *)

type 
  ('a, 'b) map = E | N of ('a, 'b) cell and
  ('a, 'b) cell = 
    { 
      mutable l : ('a, 'b) map ; 
      mutable k : 'a ;
      mutable v : 'b ;
      mutable r : ('a, 'b) map ;
      mutable h : int
    }


(** Utilitaires **)
(** Utils **)
(** tiles **)


let moduleName = "AvlMutableSet"

let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type

(* The aim of this  smart constructor  is to hide the tree decorations 
 * and allow manipulation of trees as if they were usual (unbalanced)
 * binary search trees
 *)

(* Ce contructeur est destin  occulter les dcorations de l'arbre afin
 * de pouvoir manipuler les donnes comme des arbres binaires classiques
 *)

(* El propsito de ste constructor es ocultar las decoraciones del rbol
 * afin de poder manipular los datos como si fuesen un rboles binarios comnes
 *)

let makeTree = fun l k v r -> 
  let h = match (l, r) with
    | (E, E) -> 1
    | (E, N { h = n }) | (N { h = n}, E) -> 1 + n
    | (N a, N b) -> 1 + max a.h b.h
  in
    N { l = l ; k = k ; v = v ; r = r ; h = h}

(* Note on the implementation of rotations :
 * The symetric order is conserved by rotations
 * 
 * examples :
 *  rotate_L : a x (ba y bb) -> (a x ba) y bb
 *  rotate_R : (aa y ab) x b -> aa y (ab x b)
 *  rotate_RL : a x ((baa y bab) z bb) -> (a x baa) y (bab z bb)
 *  rotate_LR : (aa x (aba y abb)) z b -> (aa x aba) y (abb z b) 
 *
 * This property allows a fast code correction verification
 *)

(* Note sur l'implmentation des rotations :
 * La notation infixe est drive de l'invariant des arbres de recherche 
 * donc l'ordre des termes (dit ordre symtrique) se conserve lors des 
 * rotations
 *
 * exemples : 
 *  rotate_L : a x (ba y bb) -> (a x ba) y bb
 *  rotate_R : (aa y ab) x b -> aa y (ab x b)
 *  rotate_RL : a x ((baa y bab) z bb) -> (a x baa) y (bab z bb)
 *  rotate_LR : (aa x (aba y abb)) z b -> (aa x aba) y (abb z b) 
 *
 * Cette proprit permet de vrifier immdiatement la correction
 * des algorithmes
 *)

(* Nota sobre la implementacin de las rotaciones :
 * El rden simtrico es conservado por las rotaciones
 * 
 * ejemplos :
 *  rotate_L : a x (ba y bb) -> (a x ba) y bb
 *  rotate_R : (aa y ab) x b -> aa y (ab x b)
 *  rotate_RL : a x ((baa y bab) z bb) -> (a x baa) y (bab z bb)
 *  rotate_LR : (aa x (aba y abb)) z b -> (aa x aba) y (abb z b) 
 *
 * sta propriedad permite verificar de manera imediata la correccin
 * de los algortmos 
 *)

let rotate_L = fun l k v r ->
  match r with
    | N t -> makeTree (makeTree l k v t.l) t.k t.v t.r
    | _ -> makeTree l k v r
	
let rotate_R = fun l k v r ->
  match l with
    | N t -> makeTree t.l t.k t.v (makeTree t.r k v r)
    | _ -> makeTree l k v r

let rotate_RL = fun l k v r ->
  match r with
    | N ({ l = N tl } as t) -> 
	makeTree (makeTree l k v tl.l) tl.k tl.v (makeTree tl.r t.k t.v t.r)
    | _ -> makeTree l k v r

let rotate_LR = fun l k v r ->
  match l with
    | N ({ r = N tr } as t) ->
	makeTree (makeTree t.l t.k t.v tr.l) tr.k tr.v (makeTree tr.r k v r)
    | _ -> makeTree l k v r

let height = function
  | E -> 0
  | N { h = n } -> n

(* [makeBTree] suposes that :
 * - the symetric order is respected l < v < r
 * - the two subtrees are well formed
 * - the height difference between l and r is not greater than 2
 *)

(* la fonction [makeBTree] suppose :
 * - le respect de l'ordre symtrique l < v < r
 * - les sous-arbres l et r tre bien forms
 * - la diffrence de hauteur entre l et r ne pas excder 2
 *)

(* la funcin [makeBTree] supone :
 * - el respeto del orden simtrico l < v < r
 * - los subrboles l y r estar bien formados
 * - la diferencia de altura entre l y r no valer ms de 2
 *)

let makeBTree = fun l k v r ->
  let hl = height l and hr = height r in
    if hl > hr + 1 then
      (match l with
	 | E -> invalid "makeBTree" UnexpectedPattern
         | N t  ->
             if height t.l >= height t.r then rotate_R l k v r 
	     else rotate_LR l k v r
      )
    else if hr > hl + 1 then
      (match r with
	 | E -> invalid "makeBTree" UnexpectedPattern
         | N t ->
             if height t.r >= height t.l then rotate_L l k v r 
	     else rotate_RL l k v r
      )
    else makeTree l k v r
      
(* makeSTree function suposes that :
 * - the symetric order is respected l < v < r
 * - the two subtrees are well formed
 *)

(* la fonction makeSTree suppose :
 * - le respect de l'ordre symtrique l < v < r
 * - les sous-arbres l et r tre forms
 *)

(* la funcin makeSTree supone :
 * - el respeto del orden simtrico l < v < r
 * - los subrboles l y r estar bien formados
 *)

let rec makeSTree = fun l k v r ->
  let hl = height l and hr = height r in
    if hl > hr + 2 then
      (match l with
	 | E -> invalid "makeSTree" UnexpectedPattern
	 | N l -> makeBTree l.l l.k l.v (makeSTree l.r k v r)
      )
    else if hr > hl + 2 then
      (match r with
	 | E -> invalid "makeSTree" UnexpectedPattern
	 | N r -> makeBTree (makeSTree l k v r.l) r.k r.v r.r
      )
    else makeBTree l k v r

(* [append] suposes that :
 * - the symetric order is respected l < r
 * - the two subtrees are well formed
 * - the height difference between l and r is not greater than 2 
 *)

(* la fonction [append] suppose :
 * - le respect de l'ordre symtrique l < r
 * - les sous-arbres l et r tre bien forms
 * - la diffrence de hauteur entre l et r ne pas exder 2
 *)

(* la funcin [append] supone :
 * - el respeto del orden simtrico l < r
 * - los subrboles l y r estar bien formados
 * - la diferencia de altura entre l y r no valer ms de 2
 *)

let rec append = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N a, N b) -> 
	makeBTree a.l a.k a.v (makeBTree (append a.r b.l) b.k b.v b.r)

(* [appendS] suposes that :
 * - the symetric order is respected l < r
 * - the two subtrees are well formed
 *)

(* la fonction [appendS] suppose :
 * - le respect de l'ordre symtrique l < r
 * - les sous-arbres l et r tre bien forms
 *)

(* la funcin [appendS] supone :
 * - el respeto del orden simtrico l < r
 * - los subrboles l y r estar bien formados
 *)

let rec appendS = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N a, N b) -> 
	makeSTree a.l a.k a.v (makeSTree (appendS a.r b.l) b.k b.v b.r)


(** Oprations catgorielles **)
(** Categorial operations **)
(** Operaciones categoriales **)


let injection = fun k v -> makeTree E k v E

let rec insert_raise k v = function
  | E -> injection k v
  | N t ->
      match compare k t.k with
	| n when n < 0 -> makeBTree (insert_raise k v t.l) t.k t.v t.r
	| n when n > 0 -> makeBTree t.l t.k t.v (insert_raise k v t.r)
	| _ -> raise Found

let insert k v = function map ->
  try 
    insert_raise k v map 
  with Found -> map

let rec from_list_rec result = function
  | [] -> result
  | (k, v) :: tail -> from_list_rec (insert k v result) tail

let from_list = function list -> from_list_rec E list

let rec to_list_rec result = function
  | E -> result
  | N t -> to_list_rec ((t.k, t.v) :: to_list_rec result t.r) t.l

let to_list = function set -> to_list_rec [] set

let map f = function set -> from_list (List.map f (to_list set))

(* Union algorithm :
 * Explained in the research report (avaible on the web)
 * 
 * Implementing Sets Efficiently in a Functional Language
 * Stephen Adams
 * CSTR 92-01
 * University of Southampton
 * 
 * We first give a version obtained by  divide and conquer 
 *)

(* Algorithme d'union :
 * Son fonctionnement est expliqu dans le rapport de recherche 
 * (disponible sur la toile)
 *
 * Implementing Sets Efficiently in a Functional Language
 * Stephen Adams
 * CSTR 92-01
 * University of Southampton
 *
 * Nous en rappelons ici le principe :
 * On construit une premire version via la mthode  diviser pour rgner 
 *)

(* Algoritmo de unin :
 * Su funcionamiento est explicado en el informe (disponible en la red)
 * 
 * Implementing Sets Efficiently in a Functional Language
 * Stephen Adams
 * CSTR 92-01
 * University of Southampton
 *
 * Recordamos rapidamente su principio :
 * Uno comienza construyendo una primera versin con el mtodo  dividir
 * para reinar 
 *)

let rec filter_lt x = function
  | E -> E
  | N t ->
      match compare x t.k with
	| n when n < 0 -> filter_lt x t.l
	| n when n > 0 -> makeSTree t.l t.k t.v (filter_lt x t.r)
	| _ -> t.l

let rec filter_gt x = function
  | E -> E
  | N t ->
      match compare x t.k with
	| n when n < 0 -> makeSTree (filter_gt x t.l) t.k t.v t.r
	| n when n > 0 -> filter_gt x t.r
	| _ -> t.r

(* We could have used a unique partition function like [partition_inf_sup]
 * but the to filter functions will be useful later in [plus]
 *)

(* Nous aurions pu utiliser une unique fonction de partition telle
 * [partition_inf_sup] mais les filtres dfinis nous seront utiles sparment 
 * dans la version optimise de [plus]
 *)
	    
(* Ubisemos podido utilizar una nica funcin de particin [partition_inf_sup]
 * pero los dos filtros que hemos definido nos sern tiles en la versin
 * optimizada de la funcin [plus]
 *)
	    
let rec old_plus = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N t, N _) ->
	let 
	  lt_v = filter_lt t.k b and 
	  gt_v = filter_gt t.k b
	in
	  makeSTree (old_plus t.l lt_v) t.k t.v (old_plus t.r gt_v)
	    
(* One can notice that there is no need to construct a tree (lt_v and gt_v)
 * which will be destructed just after by the recursive call to [old_plus]. 
 * That is why we introduce a triple containing a tree and two bounds. 
 * These bounds can be easely moved then.
 *)

(* On constate qu' rien ne sert de construire un arbre (lt_v et gt_v) qui sera
 * dcoup juste aprs par l'appel rcursif de la fonction [old_plus], raison 
 * pour laquelle on introduit un triplet contenant un arbre et deux bornes 
 * qui pourront tre modifies  moindre cot
 *)

(* Uno se da cuenta que de nada sirve construir un rbol (lt_v y gt_v) que ser
 * destruido justo despus al llamar recursivamente la funcin [old_plus].
 * Se introduce por esa razn un triplete formado por un rbol y dos 
 * limites, lmites que se podrn modificar lteriormente a poco costo 
 *)

(* [cut] constructs a triple, making sure that the root of the tree lies 
 * between the two bounds. If it is not the case, the functions calls itself
 * recusively on the corresponding subtree (right or left)
 *)

(* [cut] construit un triplet de faon  ce que la racine se trouve entre
 * les deux bornes. Si ce n'est pas le cas, la fonction restreint rcursivement
 * l'arbre  son sous-arbre gauche ou droit selon le cas de figure.
 *) 

(* [cut] constuye un triplete de manera tal a que la raz del rbol se 
 * encuentre entre los dos lmites. Si este no es el caso, la funcin
 * toma recursivamente el subrbol derecho o izquierdo segn convenga
 *)

let rec cut = fun lo tree hi ->
  match tree with
    | N t when compare t.k lo <= 0 -> cut lo t.r hi
    | N t when compare t.k hi >= 0 -> cut lo t.l hi
    | _ -> (lo, tree, hi)

(* [cut_lo] and [cut_hi] are specialized versions with an infinite 
   hight (respectively low) bound *)

(* [cut_lo] et [cut_hi] sont des versions spcialises avec la borne infrieure
 * (respectivement suprieure) ayant pour valeur plus l'infini (respectivement 
 * moins l'infini)
 *)

(* [cut_lo] y [cut_hi] son versiones especializadas con un limite inferior
 * (respectivamente superior) igual a ms (respectivamente menos) el infinito
 *)

let rec cut_hi = fun lo tree ->
  match tree with
    | N t when compare t.k lo <= 0 -> cut_hi lo t.r
    | _ -> (lo, tree)

let rec cut_lo = fun tree hi ->
  match tree with
    | N t when compare t.k hi >= 0 -> cut_lo t.l hi
    | _ -> (tree, hi)

(* [uni] uses the left tree to control the reconstruction :
 * - the right tree will be physically cut only if it lies between to 
 * successive values of the control tree (which guarantees that there will not
 * be any new cut after the reconstruction)
 * - if the right tree is empty, the whole control tree can be inserted
 *)

(* la fonction [uni] se sert de l'arbre de gauche pour contrler la
 * reconstruction : 
 * - l'arbre de droite ne sera dcoup physiquement que si l'on se trouve 
 * entre deux valeurs successives de l'arbre de contrle (ce qui garantit 
 * qu'il n'y aura pas ne nouvelle dcoupe aprs la reconstruction)
 * - si l'arbre de droite est vide, on peut insrer tout l'arbre de contrle
 *)

(* la funcin [uni] se sirve del rbol izquierdo para controlar la 
 * reconstruccin :
 * - el rbol derecho solo sera recortado fsicamente si se encuentra
 * entre dos valores succesivos del rbol de control (lo que garantiza
 * que no habr ninguna otra operacin de corte despues de la reconstruccin)
 * - si el rbol derecho est vaco, se puede insertar el rbol de control
 * entero
 *)

let rec uni = fun a (lo, b, hi) ->
  match (a, b) with
    | (_, E) -> a
    | (E, N t) -> makeSTree (filter_gt lo t.l) t.k t.v (filter_lt hi t.r)
    | (N t, N _) ->
	makeSTree (uni t.l (cut lo b t.k)) t.k t.v (uni t.r (cut t.k b hi))  

(* two specialized versions are given *)

(* on en donne galement deux versions spcialises *)

(* se dan tambin dos versiones especializadas *)

let rec uni_lo = fun a (b, hi) ->
  match (a, b) with
    | (_, E) -> a
    | (E, N t) -> makeSTree t.l t.k t.v (filter_lt hi t.r)
    | (N t, N _) -> 
	makeSTree (uni_lo t.l (cut_lo b t.k)) t.k t.v (uni t.r (cut t.k b hi))

let rec uni_hi = fun a (lo, b) ->
  match (a, b) with
    | (_, E) -> a
    | (E, N t) -> makeSTree (filter_gt lo t.l) t.k t.v t.r
    | (N t, N _) ->
	makeSTree (uni t.l (cut lo b t.k)) t.k t.v (uni_hi t.r (cut_hi t.k b)) 


(* We just put all together *)
	
(* On assemble le tout *)

(* Se juntan todos los pedazos *)

let plus = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N t, N _) ->
	makeSTree (uni_lo t.l (cut_lo b t.k)) t.k t.v 
	(uni_hi t.r (cut_hi t.k b)) 

let zero = function () -> E

let rec join_rec result = function
  | E -> result
  | N t -> join_rec (join_rec (plus t.k result) t.l) t.r

let join = function tree -> join_rec E tree


(** Oprations binaires **)
(** Binary operaations **)
(** Opraciones binrias **)


let rec partition_inf_sup pivot = function
  | E -> (E, None, E)
  | N t ->
      match compare pivot t.k with
	| n when n < 0 ->
	    let (inf, value, sup) = partition_inf_sup pivot t.l in
	      (inf, value, makeSTree sup t.k t.v t.r)
	| n when n > 0 ->
	    let (inf, value, sup) = partition_inf_sup pivot t.r in
	      (makeSTree t.l t.k t.v inf, value, sup)
	| _ -> (t.l, Some (t.k, t.v), t.r)

let union = plus

let rec intersection = fun a b ->
  match (a, b) with
    | (E, _) | (_, E) -> E
    | (N t, N _) ->
	match partition_inf_sup t.k b with
	  | (l', None, r') -> 
	      appendS (intersection t.l l') (intersection t.r r')
	  | (l', Some _, r') -> 
	      makeSTree (intersection t.l l') t.k t.v (intersection t.r r')

let rec difference = fun a b ->
  match (a, b) with
    | (E, _) -> E
    | (_, E) -> a
    | (N t, N _) ->
	match partition_inf_sup t.k b with
	  | (l', None, r') -> 
	      makeSTree (difference t.l l') t.k t.v (difference t.r r')
	  | (l', Some _, r') -> appendS (difference t.l l') (difference t.r r')
	      
let rec symetric_difference = fun a b ->
  match (a, b) with
    | (E, s) | (s, E) -> s
    | (N t, N _) ->
	match partition_inf_sup t.k b with
	  | (l', None, r') -> makeSTree (symetric_difference t.l l') t.k t.v
	      (symetric_difference t.r r')
	  | (l', Some _, r') ->
	      appendS (symetric_difference t.l l') (symetric_difference t.r r')

let rec subset = fun a b ->
  match (a, b) with
    | (E, _) -> true
    | (_, E) -> false
    | (N l, N r) ->
        match compare l.k r.k with
	  | n when n < 0 -> 
	      subset (makeBTree l.l l.k l.v E) r.l && subset l.r b
	  | n when n > 0 -> 
	      subset (makeBTree E l.k l.v l.r) r.r && subset l.l b
	  | _ -> subset l.l r.l && subset l.r r.r

let rec compare_rec = fun a b ->
  match (a, b) with
    | ([], []) -> 0
    | ([], _) -> -1
    | (_, []) -> 1
    | (E :: t1, E :: t2) -> compare_rec t1 t2
    | (N { l = E ; k = k1 ; v = v1 ; r = r1 } :: t1, 
       N { l = E ; k = k2 ; v = v2 ; r = r2 } :: t2) ->
	(match compare k1 k2 with
	   | n when n <> 0 -> n
	   | _ ->
	       match compare v1 v2 with
		 | 0 -> compare_rec (r1 :: t1) (r2 :: t2)
		 | other -> other
	)
    | (N { l = a ; k = k1 ; v = v1 ; r = r1 } :: t1, t2) -> compare_rec 
	(a :: N { l = E ; k = k1 ; v = v1 ; r = r1 ; h = 0 } :: t1) t2
    | (t1, N { l = b ; k = k2 ; v = v2 ; r = r2 } :: t2) -> compare_rec t1 
	(b :: N { l = E ; k = k2 ; v = v2 ; r = r2 ; h = 0} :: t2)

let equal = fun a b -> compare_rec [a] [b] = 0

(* let compare = fun a b -> compare_rec [a] [b]
 *
 * renvoy  la fin du fichier pour ne pas interfrer avec Pervasives.compare 
 * 
 * see the end of the file
 * 
 * ver el final del fichero
 *)


(** Accs **)
(** Acces **)
(** Acceso **)


(* [insert'] :
 * Variant of the [insert] function, it does only one comparasion
 * by branch instead of 3/2 but has to walk down the whole tree
 * 
 * If the inserted element is not already in the tree, the [insert] function
 * has anyway to go down the tree until the leafs, so in this particular
 * point the lost is not very important. The reason why this function
 * is not always faster than [insert] is that one more argument
 * must be transmitted in each recursive call
 *)

(* [insert'] :
 * Variante de la fonction [insert] ne faisant qu'une comparaison par
 * branche (au lieu de 3/2 en moyenne pour la fonction habituelle) mais
 * parcourt l'arbre jusqu' profondeur maximale
 *
 * Hormis le cas o l'lment insr se trouve dj dans l'arbre, la fonction
 * [insert] effectue un parcours sur toute la profondeur. La perte est donc
 * minime sur ce point, mais un argument supplmentaire doit tre transmis, 
 * raison pour laquelle cette fonction n'est pas toujours plus rapide que 
 * la version la plus simple
 *)

(* [insert'] :
 * Una variante de la funcin [insert] que hace una sola comparacin por
 * nivel (en vez de 3/2 en promedio para la version comn) pero recorre la 
 * totalidad del rbol, hasta las hojas.
 * 
 * Excepto en el caso en el cual el lemento que se est insertando se 
 * encuentra ya en el rbol, la funcin [insert] recorre de todas maneras
 * el rbol hasta las hojas, por lo cual la perdida en ese aspecto es
 * mnima. No obstante, un argumento suplementrio debe ser transmitido y
 * por ello esta funcin no es todo el tiempo ms rpida que la version
 * la mas simple
 *)


let rec insert_min_bound b k v = function
  | E -> 
      if compare b k = 0 then raise Found
      else injection k v
  | N t ->
      match compare k t.k with
	| n when n < 0 -> makeBTree (insert_min_bound b k v t.l) t.k t.v t.r
	| _ -> makeBTree t.l t.k t.v (insert_min_bound t.k k v t.r)

let rec insert_raise' k v = function
  | E -> injection k v
  | N t ->
      match compare k t.k with
	| n when n < 0 -> makeBTree (insert_raise' k v t.l) t.k t.v t.r
	| _ -> makeBTree t.l t.k t.v (insert_min_bound k v t.k t.r)

let insert' k v = function map ->
  try 
    insert_raise' k v map 
  with Found -> map

let insert_list = fun list map -> plus (from_list list) map

let rec delete_min = function
  | E -> E
  | N ({ l = E } as t) -> t.r
  | N t -> makeBTree (delete_min t.r) t.k t.v t.l

let rec delete_max = function
  | E -> E
  | N ({ r = E } as t) -> t.l
  | N t -> makeBTree t.l t.k t.v (delete_max t.r)

let rec delete_raise k = function
  | E -> raise Not_found
  | N t ->
      match compare k t.k with
	| n when n < 0 -> makeBTree (delete_raise k t.l) t.k t.v t.r
	| n when n > 0 -> makeBTree t.l t.k t.v (delete_raise k t.r)
	| _ -> append t.l t.r

let delete k = function map -> 
  try 
    delete_raise k map 
  with Not_found -> map

let rec delete_list = fun list map ->
  match list with
    | [] -> map
    | head :: tail -> delete_list tail (delete head map)

let rec filter_rec f result = function
  | E -> result
  | N t ->
      filter_rec f (
	filter_rec f (
	  if f t.k t.v then insert t.k t.v result else result) t.l) 
      t.r

let filter f = function map -> filter_rec f E map

let rec filter_le k = function
  | E -> E
  | N t ->
      match compare k t.k with
	| n when n < 0 -> filter_le k t.l
	| _ -> makeSTree t.l t.k t.v (filter_le k t.r)

let rec filter_ge k = function
  | E -> E
  | N t ->
      match compare k t.k with
	| n when n <= 0 -> makeSTree (filter_ge k t.l) t.k t.v t.r
	| _ -> filter_ge k t.r

let rec range min max = function
  | E -> E
  | N t ->
      if compare t.k min < 0 then range min max t.r
      else if compare t.k max > 0 then range min max t.l
      else makeSTree (filter_ge min t.l) t.k t.v (filter_le max t.r)


(** Vues **)
(** Views **)
(** Vistas **)


let rec min_view = function
  | E -> raise Empty
  | N ({ l = E } as t) -> ((t.k, t.v), t.r)
  | N t -> let (min, left) = min_view t.l in (min, makeBTree left t.k t.v t.r)

let rec max_view = function
  | E -> raise Empty
  | N ({ r = E } as t) -> (t.l, (t.k, t.v))
  | N t -> let (right, max) = max_view t.l in 
      (makeBTree t.l t.k t.v right, max)

let rec partition_rec f (trueS, falseS) = function
  | E -> (trueS, falseS)
  | N t -> 
      if f t.k t.v 
      then partition_rec f 
	(partition_rec f (insert t.k t.v trueS, falseS) t.l) t.r
      else partition_rec f 
	(partition_rec f (trueS, insert t.k t.v falseS) t.l) t.r

let partition = fun f map -> partition_rec f (E, E) map


(** Observers **)
(** Observateurs **)
(** Observadores **)


let is_empty = function
  | E -> true
  | N _ -> false

let rec size_rec result = function
  | E -> result
  | N t -> size_rec (size_rec (1 + result) t.r) t.l

let size = function tree -> size_rec 0 tree


(* [member'] : does (d + 1) comparations instead of 3/2 * d
 * following an idea of Arne Andersson, but walks down the whole height of 
 * the tree and an extra argument has to be transmitted in each call
 * 
 * It may not always be faster than [member]
 *)

(* [member'] : effectue en moyenne (d + 1) comparaisons au lieu de 3/2 * d
 * selon une ide d'Arne Andersson mais parcourt systmatiquement l'arbre sur
 * toute sa hauteur et doit transmettre un paramtre supplmentaire  chaque
 * appel
 * 
 * Il est possible qu'elle ne soit pas toujours plus rapide que [member]
 *) 

(* [member'] : hace (d + 1) comparaciones en vez de 3/2 * d en promedio, 
 * segn una idea de Arne Andersson, pero recorre toda la altura del rbol
 * y requiere un parmetro suplementrio en cada llamado
 * 
 * Es posible que no sea siempre ms veloz que [member]
 *)

let rec member_min_bound b x = function
  | E -> compare b x = 0
  | N t ->
      match compare x t.k with
	| k when k < 0 -> member_min_bound b x t.l
	| _ -> member_min_bound t.k x t.r
	    
let rec member' x = function
  | E -> false
  | N t ->
      match compare x t.k with
	| n when n < 0 -> member' x t.l
	| _ -> member_min_bound t.k x t.r

(* [member] : usual member function
 * some algorithms use the properties of the usual [member] function 
 * (stop of a positive search in the middle of the tree). It is the case
 * of optimal tree construction for example (Knuthm Garcia-Walsh, confer
 * The art of computer programming vol III)
 *)

(* [member] : fonction member habituelle 
 * un certain nombre d'algorithmes utilisent les proprits de la fonction
 * [member] habituelle (arrt d'une recherche positive  l'intrieur de 
 * l'arbre) par exemple les algorithmes de construction d'arbres optimaux
 * (Knuth, Garcia-Walsh, confer The art of computer programming vol III)
 *)

(* [member] : funcin member comn
 * ciertos algortmos usan las propriedades de la funcin [member] comn
 * (interupcin de una busqueda positiva en el medio del rbol) tal
 * los algortmos de construccin de rboles optimales (Knuth, Garcia-Walsh,
 * confer The art of computer programming vol III)
 *)

let rec member x = function
  | E -> false
  | N t ->
      match compare x t.k with
	| n when n < 0 -> member x t.l
	| n when n > 0 -> member x t.r
	| _ -> true

let rec look_for f = function
  | E -> raise Not_found
  | N t when f t.k t.v -> (t.k, t.v)
  | N t -> (try look_for f t.l with Not_found -> look_for f t.r)

let choose = function
  | E -> raise Empty
  | N t -> (t.k, t.v)

let rec min_element = function
  | E -> raise Not_found
  | N ({ l = E } as t) -> (t.k, t.v)
  | N t -> min_element t.l

let rec max_element = function
  | E -> raise Not_found
  | N ({ r = E } as t) -> (t.k, t.v)
  | N t -> max_element t.r

(* Could also be done with ['a option] *)

(* Auraient aussi pu tre codes au moyen de ['a option], nous avons
 * prfr illustrer les interruptions de calcul au moyens d'exceptions
 *)

(* stas operaciones tambin se hubiesen podido escribir con el typo 
 * ['a option] pero hemos querido aqu mostrar el uso de las excepciones para 
 * ruptura de clculo
 *)

let rec next k = function
  | E -> raise Not_found
  | N t ->
      match compare k t.k with
	| n when n < 0 -> (try next k t.l with Not_found -> (t.k, t.v))
	| n when n > 0 -> next k t.r
	| _ -> min_element t.r (* raises [Not_found] *)

let rec previous k = function
  | E -> raise Not_found
  | N t ->
      match compare k t.k with
	| n when n < 0 -> previous k t.l
	| n when n > 0 -> (try previous k t.r with Not_found -> (t.k, t.v))
	| _ -> max_element t.l (* raises [Not_found] *)
	    
let rec ceil k = function
  | E -> raise Not_found
  | N t ->
      match compare k t.k with
	| n when n < 0 -> (try ceil k t.l with Not_found -> (t.k, t.v))
	| n when n > 0 -> ceil k t.r
	| _ -> (k, t.v)

let rec floor k = function
  | E -> raise Not_found
  | N t ->
      match compare k t.k with
	| n when n < 0 -> floor k t.l
	| n when n > 0 -> (try floor k t.r with Not_found -> (t.k, t.v))
	| _ -> (k, t.v)

(* les arbres binaires de recherche sont indexs suivant l'ordre symtrique *)

(* binary search trees are indexed following the symetric order *)

(* los rboles de busqueda estn indizados segn el orden simtrico *)

let rec rank x = function
  | E -> raise Not_found
  | N t ->
      match compare x t.k with
	| n when n < 0 -> rank x t.l
	| n when n > 0 -> 1 + size t.l + rank x t.r
	| _ -> size t.l

let rec get n = function
  | E -> raise (Out_of_bounds n)
  | N t ->
      let k = size t.l in
	if n < k then get n t.l
	else if n > k then get (n - k - 1) t.r
	else (t.k, t.v)

let rec set_rec k v m = function
  | E -> raise (Out_of_bounds m)
  | N t ->
      let n = size t.l in
	if m < n then makeTree (set_rec k v m t.l) t.k t.v t.r
	else if m > n then makeTree t.l t.k t.v (set_rec k v (m - n - 1) t.r)
	else makeTree t.l k v t.r
	  
let unsafe_set = fun k v n tree -> 
  try 
    set_rec k v n tree 
  with Found -> tree

let rec for_all f = function
  | E -> true
  | N t -> f t.k t.v && for_all f t.l && for_all f t.r

let rec exists f = function
  | E -> false
  | N t -> f t.k t.v || exists f t.l || exists f t.r


(** Traversals **)
(** Parcours **)
(** Recorridos **)


let rec iter f = function
  | E -> ()
  | N t -> iter f t.l ; f t.k t.v ; iter f t.r

let rec fold_left f e = function
  | E -> e
  | N t -> fold_left f (f (fold_left f e t.l) t.k) t.r

let rec fold_right = fun f set e ->
  match set with
    | E -> e 
    | N t -> fold_right f t.l (f t.k (fold_right f t.r e))

let fold = fold_right


(** Miscellaneous **)
(** Divers **)
(** Miscelanios **)


let bind = fun set f -> join (map f set)

let concat_map = fun f set -> join (map f set)

let rec unsafe_insert_min k v = function
  | E -> injection k v
  | N t -> makeTree (unsafe_insert_min k v t.l) t.k t.v t.r 

let rec unsafe_insert_max k v = function
  | E -> injection k v
  | N t -> makeTree t.l t.k t.v (unsafe_insert_max k v t.r)


(** Compatibilit avec la librairie standard de Caml **)
(** Caml standard library compatibility **)
(** Compatibilidad con la librera estandar de Caml **)


let empty = E
let mem = member
let add = insert
let singleton = injection
let remove = delete
let inter = intersection
let diff = difference
let cardinal = size
let elements = to_list
let min_elt = min_element    
let max_elt = max_element


(** Relaxed operations **)
(** Oprations relaxes **)
(** Operaciones relajadas **)


(* Instable, utilisation non recommande *)
(* Unstable, not recommended for use *)
(* Inestable, reservado a usuarios avanzados *)

let makeRTree = makeTree

let rec relaxed_insert_raise k v = function
  | E -> injection k v
  | N t ->
      match compare k t.k with
	| n when n < 0 -> makeRTree (relaxed_insert_raise k v t.l) t.k t.v t.r
	| n when n > 0 -> makeRTree t.l t.k t.v (relaxed_insert_raise k v t.r)
	| _ -> raise Found

let relaxed_insert = fun k v tree ->
  try
    relaxed_insert_raise k v tree
  with Found -> tree

let rec relaxed_append = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N a, N b) -> 
	makeRTree a.l a.k a.v (makeRTree (relaxed_append a.r b.l) b.k b.v b.r)

let rec relaxed_delete_rec x = function
  | E -> raise Not_found
  | N t ->
      match compare x t.k with
	| n when n < 0 -> makeRTree (relaxed_delete_rec x t.l) t.k t.v t.r
	| n when n > 0 -> makeRTree t.l t.k t.v (relaxed_delete_rec x t.r)
	| _ -> relaxed_append t.l t.r

let relaxed_delete x = function tree -> 
  try 
    relaxed_delete_rec x tree 
  with Not_found -> tree

let rec rebuild_simple = function tree ->
  match tree with
    | E -> tree
    | N t -> makeSTree (rebuild_simple t.l) t.k t.v (rebuild_simple t.r)

let rec rebuild_rec = function (_, tree, _) ->
  match tree with
    | E -> (0, tree, true)
    | N t  ->
	let 
	  (hl, l', unchanged_l) = rebuild_rec (0, t.l, false) and
	  (hr, r', unchanged_r) = rebuild_rec (0, t.r, false)
	in
	let real_h = min (1 + max hl hr) (2 + min hl hr) in
	  if t.h = real_h && unchanged_l && unchanged_r then (t.h, tree, true)
	  else (real_h, makeSTree l' t.k t.v r', false)
	    
let rebuild = function tree ->
  let (_, new_tree, _) = rebuild_rec (0, tree, false) in new_tree

let rec relaxed_from_list_rec result = function
  | [] -> result
  | (k, v) :: tail -> relaxed_from_list_rec (relaxed_insert k v result) tail

let relaxed_from_list = function list -> relaxed_from_list_rec E list

let rec relaxed_delete_list = fun list set ->
  match list with
    | [] -> set
    | head :: tail -> relaxed_delete_list tail (relaxed_delete head set)

let rec relaxed_insert_list = fun list set ->
  match list with
    | [] -> set
    | (k, v) :: tail -> relaxed_insert_list tail (relaxed_insert k v set)


(** Objets **)
(** Ojects **)
(** Objetos **)


class ['a, 'b] o_map = fun set ->

object (self : 'c)
  
  val contents  = set

  method to_map = contents

  (* Categorial operations *)
  (* Oprations categorielles *)
  (* Operaciones categoriales *)

  method map f = {< contents = map f contents >}

  method plus = function (set : 'b) -> 
    {< contents = plus contents set#to_map >} 

  method to_list = to_list contents
		     
  (* Binary operations *)
  (* Oprations binaires *)
  (* Operaciones binrias *)
		     
  method union = function (set : 'c) -> 
    {< contents = plus contents set#to_map >}
    
  method intersection = function (set : 'c) -> 
    {< contents = intersection contents set#to_map >}
    
  method difference = function (set : 'c) -> 
    {< contents = difference contents set#to_map >}
    
  method symetric_difference = function (set : 'c) -> 
    {< contents = symetric_difference contents set#to_map >}
    
  method subset = function (set : 'c) -> subset contents set#to_map
  method equal = function (set : 'c)-> equal contents set#to_map
  method compare = function (set : 'c) -> compare_rec [contents] [set#to_map]
    
  (* Acces *)
  (* Accs *)
  (* Acceso *)

  method insert = fun k v -> 
    try 
      {< contents = insert_raise k v contents >} 
    with Found -> self

  method insert_raise = fun k v -> 
    {< contents = insert_raise k v contents >}

  method insert' = fun k v -> 
    try
      {< contents = insert_raise' k v contents >}
    with Found -> self

  method insert_raise' = fun k v -> 
    {< contents = insert_raise' k v contents >}
    
  method insert_list = function list -> 
    {< contents = insert_list list contents >}

  method delete = function x -> 
    try 
      {< contents = delete_raise x contents >} 
    with Not_found -> self

  method delete_raise = function x -> {< contents = delete_raise x contents >}

  method delete_list = function list -> 
    {< contents = delete_list list contents >}
    
  method delete_min = {< contents = delete_min contents >}
  method delete_max = {< contents = delete_max contents >}

  method filter = function f -> {< contents = filter f contents >}
  method filter_lt = function x -> {< contents = filter_lt x contents >}
  method filter_gt = function x -> {< contents = filter_gt x contents >}
  method filter_le = function x -> {< contents = filter_le x contents >}
  method filter_ge = function x -> {< contents = filter_ge x contents >}

  method range = fun min max -> {< contents = range min max contents >}

  (* Views *)
  (* Vues *)
  (* Vistas *)

  method min_view =
    let (min, other) = min_view contents in
      (min, {< contents = other >})

  method max_view =
    let (other, max) = max_view contents in
      ({< contents = other >}, max)

  method partition = function f -> 
    let (trueSet, falseSet) = partition f contents in
      ({< contents = trueSet >}, {< contents = falseSet >})

  method partition_inf_sup = function x ->
    let (inf, value, sup) = partition_inf_sup x contents in
      ({< contents = inf >}, value, {< contents = sup >})

  (* Observators *)
  (* Observateurs *)
  (* Observadores *)
      
  method is_empty = contents = E
  method size = size contents

  method member = function x -> member x contents
  method member' = function x -> member' x contents
  method look_for = function f -> look_for f contents

  method choose =
    match contents with
      | E -> raise Not_found
      | N t -> (t.k, t.v)

  method min_element = min_element contents
  method max_element = max_element contents

  method previous = function x -> previous x contents
  method next = function x -> next x contents
  method floor = function x -> floor x contents
  method ceil = function x -> ceil x contents

  method rank = function x -> rank x contents
  method get = function n -> get n contents

  method exists = function f -> exists f contents
  method for_all = function f -> for_all f contents

  (* Traversals *)
  (* Parcours *)
  (* Recorridos *)

  method iter : ('a -> 'b -> unit) -> unit = function f -> iter f contents

  method fold : 'd. ('a -> 'd -> 'd) -> 'd -> 'd = 
    fun f e -> fold_right f contents e

  method fold_left : 'd. ('d -> 'a -> 'd) -> 'd -> 'd =
    fun f e -> fold_left f e contents

  method fold_right : 'd. ('a -> 'd -> 'd) -> 'd -> 'd =
    fun f e -> fold_right f contents e

end

(* Constructors *)
(* Constructeurs *)
(* Constructores *)

let o_injection = fun k v -> new o_map (injection k v)
let o_map f = function a -> new o_map (map f a#to_map)
let o_plus = fun a b -> new o_map (plus a#to_map b#to_map)
let o_zero = function () -> new o_map E
let o_from_list = function list -> new o_map (from_list list)
let o_from_set = function set -> new o_map set

(* Binary operations *)
(* Oprations binaires *)
(* Operaciones binrias *)
  
let o_union = fun a b -> new o_map (plus a#to_map b#to_map)
let o_intersection = fun a b -> new o_map (intersection a#to_map b#to_map)
let o_difference = fun a b -> new o_map (difference a#to_map b#to_map)
let o_symetric_difference = fun a b -> 
  new o_map (symetric_difference a#to_map b#to_map)
  
let o_subset = fun a b -> subset a#to_map b#to_map
let o_equal = fun a b -> equal a#to_map b#to_map
let o_compare = fun a b -> compare_rec [a#to_map] [b#to_map]


module type OrderedType = sig
  type t
  val compare : t -> t -> int
end

module type Set = 
sig

  type t
  type elt

  val injection : elt -> t
  val map : (elt -> elt) -> t -> t
  val plus : t -> t -> t
  val zero : unit -> t
  val from_list : elt list -> t
  val to_list : t -> elt list
    
  val union : t -> t -> t
  val intersection : t -> t -> t
  val difference   : t -> t -> t
  val symetric_difference : t -> t -> t
  val subset  : t -> t -> bool
  val equal   : t -> t -> bool
  val compare : t -> t -> int

  val insert  : elt -> t -> t
  val insert' : elt -> t -> t
  val insert_raise  : elt -> t -> t (* [Found] *)
  val insert_raise' : elt -> t -> t (* [Found] *)
  val insert_list : elt list -> t -> t

  val delete_min : t -> t
  val delete_max : t -> t
  val delete : elt -> t -> t
  val delete_raise : elt -> t -> t (* [Not_found] *)
  val delete_list : elt list -> t -> t

  val filter : (elt -> bool) -> t -> t
  val filter_lt : elt -> t -> t
  val filter_gt : elt -> t -> t
  val filter_le : elt -> t -> t
  val filter_ge : elt -> t -> t

  val range : elt -> elt -> t -> t

  val min_view : t -> elt * t
  val max_view : t -> t * elt
  val partition : (elt -> bool) -> t -> t * t
  val partition_inf_sup : elt -> t -> t * elt option * t

  val is_empty : t -> bool
  val size : t -> int

  val member  : elt -> t -> bool
  val member' : elt -> t -> bool
  val look_for : (elt -> bool) -> t -> elt

  val choose : t -> elt (* [Empty] *)
  val min_element : t -> elt (* [Not_found] *)
  val max_element : t -> elt (* [Not_found] *)

  val next     : elt -> t -> elt (* [Not_found] *) 
  val previous : elt -> t -> elt (* [Not_found] *) 
  val ceil     : elt -> t -> elt (* [Not_found] *) 
  val floor    : elt -> t -> elt (* [Not_found] *) 

  val rank : elt -> t -> int  (* [Not_found] *) 
  val get  : int -> t -> elt  (* [Out_of_bounds] *)

  val for_all : (elt -> bool) -> t -> bool
  val exists  : (elt -> bool) -> t -> bool

  val iter : (elt -> 'a) -> t -> unit
  val fold_left  : ('a -> elt -> 'a) -> 'a -> t -> 'a
  val fold_right : (elt -> 'a -> 'a) -> t -> 'a -> 'a
  val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a

  val unsafe_inc_map : (elt -> elt) -> t -> t
  val unsafe_dec_map : (elt -> elt) -> t -> t
  val unsafe_set : elt -> int -> t -> t
  val unsafe_insert_min : elt -> t -> t
  val unsafe_insert_max : elt -> t -> t

  val old_plus : t -> t -> t

  val empty : t
  val mem : elt -> t -> bool
  val add : elt -> t -> t
  val singleton : elt -> t
  val remove : elt -> t -> t
  val inter : t -> t -> t
  val diff : t -> t -> t
  val cardinal : t -> int
  val elements : t -> elt list
  val min_elt : t -> elt (* [Not_found] *)
  val max_elt : t -> elt (* [Not_found] *)

  val stat : t -> int * int * int * int

end


