(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Maps represended by ordered lists
 * Dictionnaires reprsents par des listes ordonnes
 * Diccionrios representados por listas ordenadas
 *)

open BairePervasives

let moduleName = "OrderedListMap"


(** Utilitaires **)
(** Utils **)
(** tiles **)


let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type


(** Oprations catgorielles **)
(** Categorial operations **)
(** Operaciones categoriales **)


let injection = fun k v -> [(k, v)]

let rec map f = function
  | [] -> []
  | head :: tail -> (f head) :: map f tail

let rec plus = fun a b ->
  match (a, b) with
    | ([], _) -> b
    | (_, []) -> a
    | ((k, _) as a_head :: a_tail, ((k', _) as b_head) :: b_tail) ->
	match compare k k' with
	  | n when n < 0 -> a_head :: plus a_tail b
	  | 0 -> b_head :: plus a_tail b_tail
	  | _ -> b_head :: plus a b_tail

let zero = function () -> []

let to_list = id

let from_list = id

(*  implmenter construction d'arbre parfait *)

let inj = function x -> Node (Leaf, x, Leaf)

let rec to_tree = function
  | x :: y :: tail -> Node (inj x, y, to_tree tail)
  | x :: tail -> Node (Leaf, x, Leaf)
  | [] -> Leaf

let rec from_tree_rec result = function
  | Leaf -> result
  | Node (l, v, r) -> from_tree_rec (v :: (from_tree_rec result l)) r

let from_tree = function tree -> from_tree_rec [] tree


(** Oprations binaires **)
(** Binairy operations **)
(** Operaciones binrias **)


let union = plus

let rec intersection = fun a b ->
  match (a, b) with
    | ([], _) | (_, []) -> []
    | ((k, _) as a_head :: a_tail, ((k', _) as b_head) :: b_tail) ->
	match compare k k' with
	  | 0 -> a_head :: intersection a_tail b_tail
	  | _ -> intersection a_tail b_tail

let rec difference = fun a b ->
  match (a, b) with
    | ([], _) -> []
    | (_, []) -> a
    | ((k, _) as a_head :: a_tail, ((k', _) as b_head) :: b_tail) ->
	match compare k k' with
	  | n when n < 0 -> a_head :: difference a_tail b
	  | n when n > 0 -> a_head :: difference a_tail b_tail
	  | _ -> difference a_tail b_tail

let rec symetric_difference = fun a b ->
  match (a, b) with
    | ([], s) | (s, []) -> s
    | ((k, _) as a_head :: a_tail, ((k', _) as b_head) :: b_tail) ->
	match compare k k' with
	  | n when n < 0 -> a_head :: symetric_difference a_tail b
	  | n when n > 0 -> b_head :: symetric_difference a b_tail
	  | _ -> symetric_difference a_tail b_tail


(** Acces **)
(** Accs **)
(** Acceso **)


let rec insert k v = function
  | [] -> [(k, v)]
  | (k', v') as head :: tail ->
      match compare k k' with
	| n when n < 0 -> head :: insert k v tail
	| n when n > 0 -> (k', v') :: (k, v) :: tail
	| _ -> (k, v) :: tail

let rec delete k = function
  | [] -> []
  | (k', v') as head :: tail ->
      match compare k k' with
	| n when n < 0 -> head :: delete k tail
	| n when n > 0 -> head :: tail
	| _ -> tail

let delete_min = function
  | [] -> []
  | _ :: tail -> tail

let rec delete_max = function
  | [] -> []
  | [_] -> []
  | head :: tail -> head :: delete_max tail

let rec filter f = function
  | [] -> []
  | (k, v) as head :: tail when f k v -> head :: filter f tail
  | _ :: tail -> filter f tail

let rec filter_lt x = function
  | [] -> []
  | (k, v) as head :: tail ->
      match compare x k with
	| n when n < 0 -> []
	| _ -> head :: filter_lt x tail

let rec filter_le x = function
  | [] -> []
  | (k, v) as head :: tail ->
      match compare x k with
	| n when n <= 0 -> []
	| _ -> head :: filter_lt x tail

let rec filter_gt x = function
  | [] -> []
  | (k, _) :: tail as list ->
      match compare x k with
	| n when n <= 0 -> list 
	| _ -> filter_gt x tail

let rec filter_ge x = function
  | [] -> []
  | (k, _) :: tail as list ->
      match compare x k with
	| n when n < 0 -> list
	| _ -> filter_gt x tail

let rec range min max = function
  | [] -> []
  | (k, _) :: tail as list ->
      if k < min then range min max tail
      else filter_lt max list


(** Vues **)
(** Views **)
(** Vistas **)


let min_view = function
  | [] -> raise Empty
  | head :: tail -> (head, tail)

let rec max_view_rec head_list = function
  | [] -> invalid "max_view_rec" UnexpectedPattern
  | [max] -> (List.rev head_list, max)
  | head :: tail -> max_view_rec (head :: head_list) tail

let rec partition_rec f trueL falseL = function
  | [] -> (List.rev trueL, List.rev falseL)
  | head :: tail when (f head) -> partition_rec f (head :: trueL) falseL tail
  | head :: tail -> partition_rec f trueL (head :: falseL) tail

let partition f = function list -> partition_rec f [] [] list

let rec partition_inf_sup_rec head_list pivot = function
  | [] -> ([], None, [])
  | (k, _) as head :: tail ->
      match compare pivot k with
	| n when n < 0 -> partition_inf_sup_rec (head :: head_list) pivot tail
	| n when n > 0 -> (List.rev (head :: head_list), None, tail)
	| _ -> (List.rev (head :: head_list), Some head, tail)

let partition_inf_sup = fun p map -> partition_inf_sup_rec [] p map


(** Observateurs **)
(** Observers **)
(** Observadores **)


let is_empty = function
  | [] -> true
  | _ :: _ -> false

let rec size_rec k = function
  | [] -> k
  | _ :: tail -> size_rec (k + 1) tail

let size = function list -> size_rec 0 list

let rec keys = function
  | [] -> []
  | (k, _) :: tail -> k :: keys tail

let rec images = function
  | [] -> []
  | (_, v) :: tail -> v :: images tail

let rec member k = function
  | [] -> false
  | (k', _) :: tail ->
      match compare k k' with
	| n when n < 0 -> member k tail
	| n when n > 0 -> false
	| _ -> true

let rec lookup k = function
  | [] -> raise Not_found
  | (k', v') :: tail ->
      match compare k k' with
	| n when n < 0 -> lookup k tail
	| n when n > 0 -> raise Not_found
	| _ -> v'

let rec look_for f = function
  | [] -> raise Not_found
  | (k, v) :: _ when f k v -> (k, v)
  | _ :: tail -> look_for f tail

let choose = function
  | [] -> raise Not_found
  | head :: _ -> head

let min_element = function
  | [] -> raise Not_found
  | head :: _ -> head

let rec max_element = function
  | [] -> raise Not_found
  | [max] -> max
  | _ :: tail -> max_element tail

let rec previous x = function
  | [] -> raise Not_found
  | (k, _) as head :: tail ->
      match compare x k with
	| n when n < 0 -> (try previous x tail with Not_found -> head)
	| _ -> raise Not_found

let rec next x = function
  | [] -> raise Not_found
  | (k, _) as head :: tail ->
      match compare x k with
	| n when n <= 0 -> next x tail
	| 0 -> min_element tail
	| _ -> head

let rec floor x = function
  | [] -> raise Not_found
  | (k, _) as head :: tail ->
      match compare x k with
	| n when n <= 0 -> (try floor x tail with Not_found -> head)
	| _ -> raise Not_found

let rec ceil x = function
  | [] -> raise Not_found
  | (k, _) as head :: tail ->
      match compare x k with
	| n when n <= 0 -> next x tail
	| _ -> head

let rec rank_rec x total = function
  | [] -> raise Not_found
  | (k, _) :: _ when k = x -> total
  | _ :: tail -> rank_rec x (total + 1) tail

let rank x = function list -> rank_rec x 0 list

let rec get = fun list n ->
  match list with
    | [] -> raise (Out_of_bounds n)
    | head :: _  when (n = 0) -> head
    | _ :: tail -> get tail (n - 1)

let rec for_all f = function
  | [] -> true
  | (k, v) :: tail -> f k v && for_all f tail

let rec exists f = function 
  | [] -> false
  | (k, v) :: tail -> f k v || exists f tail


(** Traversals **)
(** Parcours **)
(** Recorridos **)


let rec iter f = function
  | [] -> ()
  | (k, v) :: tail -> f k v ; iter f tail

let rec fold_right = fun f map e ->
  match map with
  | [] -> e
  | (k, v) :: tail -> f k v (fold_right f tail e)

let rec fold_left f e = function
  | [] -> e
  | (k, v) :: tail -> fold_left f (f e k v) tail

let fold = fold_right


(** Miscellaneous **)
(** Divers **)
(** Micelaneos **)


let rec insert_with f k v = function
  | [] -> [(k, v)]
  | (k', v') as head :: tail ->
      match compare k k' with
	| n when n < 0 -> head :: insert k v tail
	| n when n > 0 -> (k', v') :: (k, v) :: tail
	| _ -> (k, f v v') :: tail

let rec union_with f = fun a b ->
  match (a, b) with
    | ([], _) -> b
    | (_, []) -> a
    | ((k, v) as a_head :: a_tail, ((k', v') as b_head) :: b_tail) ->
	match compare k k' with
	  | n when n < 0 -> a_head :: union_with f a_tail b
	  | n when n > 0 -> b_head :: union_with f a b_tail
	  | _ -> (k, f v v') :: union_with f a_tail b_tail

let rec intersection_with f = fun a b ->
  match (a, b) with
    | ([], _) | (_, []) -> []
    | ((k, v) as a_head :: a_tail, ((k', v') as b_head) :: b_tail) ->
	match compare k k' with
	  | 0 -> (k, f v v') :: intersection_with f a_tail b_tail
	  | _ -> intersection_with f a_tail b_tail

let rec map_images f = function
  | [] -> []
  | (k, v) :: tail -> (k, f v) :: map_images f tail

let rec map_images_with_index f = function
  | [] -> []
  | (k, v) :: tail -> (k, f k v) :: map_images_with_index f tail


(** Compatibilit avec la librairie standard de Caml **)
(** Caml standard library compatibility **)
(** Compatibilidad con la librera estandar de Caml **)


let empty = []
let mem = member
let add = insert
let singleton = injection
let remove = delete
let inter = intersection
let diff = difference
let cardinal = size
let elements = id
let min_elt = min_element    
let max_elt = max_element
let find = lookup
let mapi = map_images_with_index
