(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Two dimensional range maps (lexicographical order)
 * Dictionnaires bidimensionnels (ordre lexicographique)
 * Diccionarios bidimensionales (rden lexicogrfico)
 *)

open BairePervasives

type ('a, 'b) two = 
  | ETwo 
  | NTwo of ('a, 'b) two * 'a * 'b * ('a, 'b) two

type ('a, 'b) set = 
  | E 
  | N of ('a, 'b) set * 'a * ('a, 'b) set * ('a, 'b) two


(** Utils **)
(** Utilitaires **)
(** tiles **)


let moduleName = "Range2DMap"

let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type

let makeTree = fun l k r s -> N (l, k, r, s)
let makeTwo = fun l k v r -> NTwo (l, k, v, r)

let rec append_two = fun a b ->
  match (a, b) with
    | (ETwo, _) -> b
    | (_, ETwo) -> a
    | (NTwo (l1, k1, v1, r1), NTwo (l2, k2, v2, r2)) ->
	makeTwo l1 k1 v1 (makeTwo (append_two r1 l2) k2 v2 r2)

let rec append = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l1, k1, r1, s1), N (l2, k2, r2, s2)) ->
	match compare k1 k2 with
	  | 0 -> makeTree l1 k1 r2 (append_two s1 s2)
	  | _ -> makeTree l1 k1 (makeTree (append r1 l2) k2 r2 s2) s1


(** Categorial operations **)
(** Oprations catgorielles **)
(** Operaciones categoriales **)


let injection = fun x y v -> makeTree E x E (makeTwo ETwo y v ETwo)

let zero = function () -> E

let rec to_list_two x result = function
  | ETwo -> result
  | NTwo (l, y, v, r) -> to_list_two x ((x, y, v) :: to_list_two x result r) l

let rec to_list_one result = function
  | E -> result
  | N (l, k, r, s) -> 
      to_list_one (to_list_two k (to_list_one result r) s) l 

let to_list = function tree -> to_list_one [] tree

let rec insert_two y v = function
  | ETwo -> makeTwo ETwo y v ETwo
  | NTwo (l, k', v', r) ->
      match compare y k' with
	| n when n < 0 -> makeTwo (insert_two y v l) k' v' r
	| n when n > 0 -> makeTwo l k' v' (insert_two y v r)
	| _ -> raise Found

let rec insert_raise x y v = function 
  | E -> injection x y v
  | N (l, k, r, s) ->
      match compare x k with
	 | n when n < 0 -> makeTree (insert_raise x y v l) k r s
	 | n when n > 0 -> makeTree l k (insert_raise x y v r) s
	 | _ -> makeTree l k r (insert_two y v s)

let insert = fun x y v tree ->
  try
    insert_raise x y v tree
  with Found -> tree

let rec from_list_rec result = function
  | [] -> result
  | (x, y, v) :: tail -> from_list_rec (insert x y v result) tail

let from_list = function list -> from_list_rec E list

let map f = function tree -> from_list (List.map f (to_list tree)) 

let rec partition_one pivot = function
  | E -> (E, None, E)
  | N (l, x, r, s) as node ->
      match compare pivot x with
	| n when n < 0 ->
	    let (inf, value, sup) = partition_one pivot l in
	      (inf, value, makeTree sup x r s)
	| n when n > 0 ->
	    let (inf, value, sup) = partition_one pivot r in
	      (makeTree l x inf s, value, sup)
	| _ -> (l, Some (x, s), r)

let rec partition_two pivot = function
  | ETwo -> (ETwo, None, ETwo)
  | NTwo (l, y, v, r) as node ->
      match compare pivot y with
	| n when n < 0 ->
	    let (inf, value, sup) = partition_two pivot l in
	      (inf, value, makeTwo sup y v r)
	| n when n > 0 ->
	    let (inf, value, sup) = partition_two pivot r in
	      (makeTwo l y v inf, value, sup)
	| _ -> (l, Some (y, v), r)

let rec plus_two = fun a b -> 
  match (a, b) with
    | (ETwo, _) -> b
    | (_, ETwo) -> a
    | (NTwo (l, x, v, r), _) -> let (c, _, d) = partition_two x b in 
	makeTwo (plus_two l c) x v (plus_two r d)

let rec plus = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l, k, r, s), N _) ->
	match partition_one k b with
	  | (left, None, right) -> makeTree (plus l left) k (plus r right) s
	  | (left, Some (_, s'), right) -> 
	      makeTree (plus l left) k (plus r right) (plus_two s s')


(** Binary operations **)
(** Oprations binaires **)
(** Operaciones binrias **)


let union = plus

let rec intersection_two = fun a b ->
  match (a, b) with
    | (ETwo, _) | (_, ETwo) -> ETwo
    | (NTwo (l, k, v, r), NTwo _) ->
	match partition_two k b with
	  | (l', None, r') -> 
	      append_two (intersection_two l l') (intersection_two r r')
	  | (l', Some _, r') -> 
	      makeTwo (intersection_two l l') k v (intersection_two r r')

let rec intersection = fun a b ->
  match (a, b) with
    | (E, _) | (_, E) -> E
    | (N (l, k, r, s), N _) ->
	match partition_one k b with
	  | (l', None, r') -> append (intersection l l') (intersection r r')
	  | (l', Some (_, s'), r') ->
	      match intersection_two s s' with
		| ETwo -> append (intersection l l') (intersection r r')
		| s'' -> makeTree (intersection l l') k (intersection r r') s''

let rec difference_two = fun a b ->
  match (a, b) with
    | (ETwo, _) -> ETwo
    | (_, ETwo) -> a
    | (NTwo (l, k, v, r), NTwo _) -> 
	match partition_two k b with
	  | (l', None, r') -> 
	      makeTwo (difference_two l l') k v (difference_two r r')
	  | (l', Some _, r') -> 
	      append_two (difference_two l l') (difference_two r r')
	      
let rec difference = fun a b ->  
  match (a, b) with
    | (E, _) -> E
    | (_, E) -> a
    | (N (l, k, r, s) , N _) ->
	match partition_one k b with
	  | (l', None, r') -> makeTree (difference l l') k (difference r r') s
	  | (l', Some (_, s'), r') ->
	      match difference_two s s' with
		| ETwo -> append (difference l l') (difference r r')
		| s'' -> makeTree (difference l l') k (difference r r') s''

let rec symetric_difference_two = fun a b ->
  match (a, b) with
    | (ETwo, s) | (s, ETwo) -> s
    | (NTwo (l, k, v, r), NTwo _) ->
	match partition_two k b with
	  | (l', None, r') -> makeTwo (symetric_difference_two l l') k v 
	      (symetric_difference_two r r')
	  | (l', Some _, r') -> append_two (symetric_difference_two l l') 
	      (symetric_difference_two r r')

let rec symetric_difference = fun a b ->  
  match (a, b) with
    | (E, s) | (s, E) -> s
    | (N (l, k, r, s) , N _) ->
	match partition_one k b with
	  | (l', None, r') -> makeTree (symetric_difference l l') k 
	      (symetric_difference r r') s
	  | (l', Some (_, s'), r') ->
	      match symetric_difference_two s s' with
		| ETwo -> append (symetric_difference l l') 
		    (symetric_difference r r')
		| s'' -> makeTree (symetric_difference l l') k 
		    (symetric_difference r r') s''

let rec subset_two = fun a b ->
  match (a, b) with
    | (ETwo, _) -> true
    | (_, ETwo) -> false
    | (NTwo (l1, k1, v1, r1), NTwo (l2, k2, v2, r2)) ->
        match compare k1 k2 with
	  | n when n < 0 -> 
	      subset_two (makeTwo l1 k1 v1 ETwo) l2 && subset_two r1 b
	  | n when n > 0 -> 
	      subset_two (makeTwo ETwo k1 v1 r1) r2 && subset_two l1 b
	  | _ -> v1 = v2 && subset_two l1 l2 && subset_two r1 r2

let rec subset = fun a b ->
  match (a, b) with
    | (E, _) -> true
    | (_, E) -> false
    | (N (l1, k1, r1, s1), N (l2, k2, r2, s2)) ->
        match compare k1 k2 with
	  | n when n < 0 -> subset (makeTree l1 k1 E s1) l2 && subset r1 b
	  | n when n > 0 -> subset (makeTree E k1 r1 s1) r2 && subset l1 b
	  | _ -> subset_two s1 s2 && subset l1 l2 && subset r1 r2

let rec compare_two = fun a b ->
  match (a, b) with
    | ([], []) -> 0
    | ([], _) -> -1
    | (_, []) -> 1
    | (ETwo :: t1, ETwo :: t2) -> compare_two t1 t2
    | (NTwo (ETwo, k1, v1, r1) :: t1, NTwo (ETwo, k2, v2, r2) :: t2) ->
	(match compare k1 k2 with
	  | n when n <> 0 -> n
	  | _ -> 
	      match compare v1 v2 with
		| 0 ->compare_two (r1 :: t1) (r2 :: t2)
		| other -> other
	)
    | (NTwo (a, k1, v1, r1) :: t1, t2) ->
        compare_two (a :: NTwo (ETwo, k1, v1, r1) :: t1) t2
    | (t1, NTwo (b, k2, v2, r2) :: t2) ->
	compare_two t1 (b :: NTwo (ETwo, k2, v2, r2) :: t2)

let rec compare_rec = fun a b ->
  match (a, b) with
    | ([], []) -> 0
    | ([], _) -> -1
    | (_, []) -> 1
    | (E :: t1, E :: t2) -> compare_rec t1 t2
    | (N (E, k1, r1, s1) :: t1, N (E, k2, r2, s2) :: t2) ->
	(match compare k1 k2 with
	   | n when n <> 0 -> n
	   | _ -> match compare_two [s1] [s2] with
	       | 0 -> compare_rec (r1 :: t1) (r2 :: t2)
	       | other -> other
	)
    | (N (a, k1, r1, s1) :: t1, t2) ->
        compare_rec (a :: N (E, k1, r1, s1) :: t1) t2
    | (t1, N (b, k2, r2, s2) :: t2) ->
	compare_rec t1 (b :: N (E, k2, r2, s2) :: t2)

let equal = fun a b -> compare_rec [a] [b] = 0


(** Acces **)
(** Accs **)
(** Acceso **)


let rec delete_two y = function
  | ETwo -> raise Not_found
  | NTwo (l, k, v, r) ->
      match compare y k with
	| n when n < 0 -> makeTwo (delete_two y l) k v r
	| n when n > 0 -> makeTwo l k v (delete_two y r)
	| _ -> append_two l r

let rec delete_raise x y = function
  | E -> raise Not_found
  | N (l, k, r, s) ->
      match compare x k with
	| n when n < 0 -> makeTree (delete_raise x y l) k r s
	| n when n > 0 -> makeTree l k (delete_raise x y r) s
	| _ -> 
	    match delete_two y s with
	      | ETwo -> append l r
	      | s' -> makeTree l k r s'

let delete = fun x y tree ->
  try 
    delete_raise x y tree
  with Not_found -> tree

let rec delete_list = fun list set ->
  match list with
    | [] -> set
    | (x, y) :: tail -> delete_list tail (delete x y set)

let rec delete_min_two = function
  | ETwo -> ETwo
  | NTwo (ETwo, _, _, r) -> r
  | NTwo (l, k, v, r) -> makeTwo (delete_min_two l) k v r

let rec delete_min = function
  | E -> raise Empty
  | N (E, k, r, s) -> 
      (match delete_min_two s with
	 | ETwo -> r
	 | s' -> makeTree E k r s' 
      )
  | N (l, k, r, s) -> makeTree (delete_min l) k r s

let rec delete_max_two = function
  | ETwo -> ETwo
  | NTwo (l, _, _, ETwo) -> l
  | NTwo (l, k, v, r) -> makeTwo l k v (delete_max_two r)

let rec delete_max = function
  | E -> raise Empty
  | N (l, k, E, s) ->
      (match delete_max_two s with
	 | ETwo -> l
	 | s'-> makeTree l k E s'
      )
  | N (l, k, r, s) -> makeTree l k (delete_max r) s

let rec filter_two f result = function
  | ETwo -> result
  | NTwo (l, k, v, r) ->
      filter_two f 
      (filter_two f 
	  (if f k v then insert_two k v result 
	   else result) 
	  l) 
      r

let rec filter f = function
  | E -> E
  | N (l, k, r, s) ->
      match filter_two f ETwo s with
	| ETwo -> append (filter f l) (filter f r)
	| s' -> makeTree (filter f l) k (filter f r) s'

let rec filter_lt_two y = function
  | ETwo -> ETwo
  | NTwo (a, k, v, b) ->
      match compare y k with
	| n when n < 0 -> filter_lt_two y a
	| n when n > 0 -> makeTwo a k v (filter_lt_two y b)
	| _ -> a

let rec filter_gt_two y = function
  | ETwo -> ETwo
  | NTwo (a, k, v, b) ->
      match compare y k with
	| n when n < 0 -> makeTwo (filter_gt_two y a) k v b
	| n when n > 0 -> filter_gt_two y b
	| _ -> b

let rec filter_le_two x = function
  | ETwo -> ETwo
  | NTwo (l, k, v, r) ->
      match compare x k with
	| n when n < 0 -> filter_le_two x l
	| _ -> makeTwo l k v (filter_le_two x r)

let rec filter_ge_two x = function
  | ETwo -> ETwo
  | NTwo (l, k, v, r) ->
      match compare x k with
	| n when n <= 0 -> makeTwo (filter_ge_two x l) k v r
	| _ -> filter_ge_two x r

let rec filter_lt x y = function
  | E -> E
  | N (l, k, r, s) ->
      match compare x k with
	| n when n < 0 -> filter_lt x y l
	| n when n > 0 -> makeTree l k (filter_lt x y r) s
	| _ -> 
	    match filter_lt_two y s with
	      | ETwo -> l
	      | s' -> makeTree l k E s'

let rec filter_gt x y = function
  | E -> E
  | N (l, k, r, s) ->
      match compare x k with
	| n when n < 0 -> makeTree (filter_gt x y l) k r s
	| n when n > 0 -> filter_gt x y r
	| _ ->
	    match filter_gt_two y s with
	      | ETwo -> r
	      | s' -> makeTree E k r s'

let rec filter_le x y = function
  | E -> E
  | N (l, k, r, s) ->
      match compare x k with
	| n when n < 0 -> filter_le x y l
	| _ -> 
	    match filter_le_two y s with
	      | ETwo -> l
	      | s' -> makeTree l k E s'

let rec filter_ge x y = function
  | E -> E
  | N (l, k, r, s) ->
      match compare x k with
	| n when n <= 0 -> makeTree (filter_ge x y l) k r s
	| _ ->
	    match filter_ge_two y s with
	      | ETwo -> r
	      | s' -> makeTree E k r s'

let rec range_two min max = function
  | ETwo -> ETwo
  | NTwo (l, k, v, r) ->
      if compare k min < 0 then range_two min max r
      else if compare k max > 0 then range_two min max l
      else makeTwo (filter_ge_two min l) k v (filter_le_two max r)

let rec map_over_two f = function
  | E -> E
  | N (l, k, r, s) -> 
      match f s with
	| ETwo -> append (map_over_two f l) (map_over_two f r)
	| s' -> makeTree (map_over_two f l) k (map_over_two f r) s'

let rec filter_range_le f x = function
  | E -> E
  | N (l, k, r, s) ->
      match compare x k with
	| n when n < 0 -> filter_range_le f x l
	| n when n > 0 -> makeTree (map_over_two f l) k 
	    (filter_range_le f x r) s
	| _ ->
	    match f s with
	      | ETwo -> map_over_two f l
	      | s' -> makeTree (map_over_two f l) k E s'

let rec filter_range_ge f x = function
  | E -> E
  | N (l, k, r, s) ->
      match compare x k with
	| n when n < 0 -> makeTree (filter_range_ge f x l) k 
	    (map_over_two f r) s
	| n when n > 0 -> filter_range_le f x r
	| _ ->
	    match f s with
	      | ETwo -> map_over_two f r
	      | s' -> makeTree E k (map_over_two f r) s'

let rec range_one f min_x max_x = function
  | E -> E
  | N (l, k, r, s) ->
      if compare k min_x < 0 then range_one f min_x max_x r
      else if compare k max_x > 0 then range_one f min_x max_x l
      else match f s with
	| ETwo -> append (filter_range_ge f min_x l) 
	    (filter_range_le f max_x r)
	| s' -> makeTree (filter_range_ge f min_x l) k 
	    (filter_range_le f max_x r) s'

let range = fun min_x max_x min_y max_y tree ->
  range_one (range_two min_y max_y) min_x max_x tree

let is_empty = function
  | E -> true
  | N _ -> false

let rec size_two total = function
  | ETwo -> total
  | NTwo (l, _, _, r) -> size_two (size_two (1 + total) l) r

let rec size_one total = function
  | E -> total
  | N (l, _, r, s) -> size_one (size_one (size_two total s) l) r

let size = function tree -> size_one 0 tree

let rec member_two y = function
  | ETwo -> false
  | NTwo (l, k, _, r) ->
      match compare y k with
	| n when n < 0 -> member_two y l
	| n when n > 0 -> member_two y r
	| _ -> true

let rec member x y = function
  | E -> false
  | N (l, k, r, s) ->
      match compare x k with
	 | n when n < 0 -> member x y l
	 | n when n > 0 -> member x y r
	 | _ -> member_two y s

let rec look_for_two f = function
  | ETwo -> raise Not_found
  | NTwo (_, k, v, _) when f k v -> (k, v)
  | NTwo (l, _, _, r) -> 
      (try look_for_two f l with Not_found -> look_for_two f r)

let rec look_for f = function
  | E -> raise Not_found
  | N (l, k, r, s) ->
      try look_for_two (f k) s with
	  Not_found ->
	    try look_for f l with 
		Not_found -> look_for f r
	      
let choose = function
  | E -> raise Empty
  | N (_, x, _, NTwo (_, y, v, _)) -> (x, y, v)
  | _ -> invalid "choose" UnexpectedPattern

let rec min_element_two = function
  | ETwo -> raise Not_found
  | NTwo (ETwo, k, v, _) -> (k, v)
  | NTwo (l, _, _, _) -> min_element_two l

let rec min_element = function
  | E -> raise Not_found
  | N (E, x, _, s) -> let (y, v) = min_element_two s in (x, y, v)
  | N (l, _, _, _) -> min_element l

let rec max_element_two = function
  | ETwo -> raise Not_found
  | NTwo (_, k, v, ETwo) -> (k, v)
  | NTwo (_, _, _, r) -> max_element_two r

let rec max_element = function
  | E -> raise Not_found
  | N (_, x, E, s) -> let (y, v) = max_element_two s in (x, y, v)
  | N (_, _, l, _) -> max_element l

let rec next_two y = function
  | ETwo -> raise Not_found
  | NTwo (l, k, v, r) -> 
      match compare y k with
	| n when n < 0 -> (try next_two y l with Not_found -> (k, v))
	| n when n > 0 -> next_two y r
	| _ -> min_element_two r (* [Not_found] *)

let rec next x y = function
  | E -> raise Not_found
  | N (l, k, r, s) -> 
      match compare x k with
	| n when n < 0 -> 
	    (try next x y l 
	     with Not_found -> let (k', v) = min_element_two s in (k, k', v))
	| n when n > 0 -> next x y r
	| _ ->
	    try let (k', v) = next_two y s in (k, k', v) 
	    with Not_found -> min_element r

let rec previous_two y = function
  | ETwo -> raise Not_found
  | NTwo (l, k, v, r) ->
      match compare y k with
	| n when n < 0 -> previous_two y l
	| n when n > 0 -> (try previous_two y r with Not_found -> (k, v))
	| _ -> max_element_two l (* [Not_found] *)

let rec previous x y = function
  | E -> raise Not_found
  | N (l, k, r, s) ->
      match compare x k with
	| n when n < 0 -> previous x y l
	| n when n > 0 -> 
	    (try previous x y r 
	     with Not_found -> let (k', v) = max_element_two s in (k, k', v))
	| _ -> 
	    try let (k', v) = previous_two y s in (k, k', v)
	    with Not_found -> max_element l
	
let rec ceil_two y = function
  | ETwo -> raise Not_found
  | NTwo (l, k, v, r) -> 
      match compare y k with
	| n when n < 0 -> (try ceil_two y l with Not_found -> (k, v))
	| n when n > 0 -> ceil_two y r
	| _ -> (k, v)

let rec ceil x y = function
  | E -> raise Not_found
  | N (l, k, r, s) -> 
      match compare x k with
	| n when n < 0 -> 
	    (try ceil x y l 
	     with Not_found -> let (k', v) = min_element_two s in (k, k', v))
	| n when n > 0 -> ceil x y r
	| _ -> 
	    try let (k', v) = ceil_two y s in (k, k', v) 
	    with Not_found -> min_element r

let rec floor_two y = function
  | ETwo -> raise Not_found
  | NTwo (l, k, v, r) ->
      match compare y k with
	| n when n < 0 -> floor_two y l
	| n when n > 0 -> (try floor_two y r with Not_found -> (k, v))
	| _ -> (k, v)

let rec floor x y = function
  | E -> raise Not_found
  | N (l, k, r, s) ->
      match compare x k with
	| n when n < 0 -> floor x y l
	| n when n > 0 -> 
	    (try floor x y r 
	     with Not_found -> let (k', v) = max_element_two s in (k, k', v))
	| _ -> 
	    try let (k', v) = floor_two y s in (k, k', v) 
	    with Not_found -> max_element l

let rec rank_two y total = function
  | ETwo -> total
  | NTwo (a, k, _, b) ->
      match compare y k with
	| n when n < 0 -> rank_two y total a
	| n when n > 0 -> rank_two y (size_two (1 + total) a) b
	| _ -> size_two (1 + total) a

let rec rank_rec x y total = function
  | E -> total
  | N (a, k, b, s) ->
      match compare x k with
	| n when n < 0 -> rank_rec x y total a
	| n when n > 0 -> rank_rec x y (size_one (size_two total s) a) b
	| _ -> size_one (rank_two y total s) a

let rank = fun x y tree -> rank_rec x y 0 tree

let rec get_two n = function
  | ETwo -> raise (Out_of_bounds n)
  | NTwo (a, x, v, b) ->
      let k = size_two 0 a in
	if n < k then get_two n a
	else if n > k then get_two (n - k - 1) b
	else (x, v)

let rec exists_two f = function
  | ETwo -> false
  | NTwo (l, y, v, r) -> f y v || exists_two f l || exists_two f r

let rec exists f = function
  | E -> false
  | N (l, x, r, s) -> exists_two (f x) s || exists f l || exists f r

let rec for_all_two f = function
  | ETwo -> true
  | NTwo (l, y, v, r) -> f y v && for_all_two f l && for_all_two f r

let rec for_all f = function
  | E -> true
  | N (l, x, r, s) -> for_all_two (f x) s && for_all f l && for_all f r

let rec iter_two f = function
  | ETwo -> ()
  | NTwo (l, y, v, r) -> iter_two f l ; f y v ; iter_two f r

let rec iter f = function
  | E -> ()
  | N (l, x, r, s) -> iter f l ; iter_two (f x) s ; iter f r

let rec fold_left_two f e = function
  | ETwo -> e
  | NTwo (l, k, v, r) -> fold_left_two f (f (fold_left_two f e l) k v) r

let rec fold_left f e = function
  | E -> e
  | N (l, x, r, s) ->
      fold_left f (fold_left_two (fun a y -> f a x y) (fold_left f e l) s) r

let rec fold_right_two = fun f set e ->
  match set with
    | ETwo -> e 
    | NTwo (l, k, v, r) -> fold_right_two f l (f k v (fold_right_two f r e))

let rec fold_right = fun f set e ->
  match set with
    | E -> e
    | N (l, x, r, s) -> 
	fold_right f l (fold_right_two (f x) s (fold_right f r e))

let fold = fold_right

(** Compatibilit akec la librairie standard de Caml **)
(** Caml standard library compatibility **)
(** Compatibilidad con la librera estandar de Caml **)


let empty = E
let mem = member
let add = insert
let singleton = injection
let remoke = delete
let inter = intersection
let diff = difference
let cardinal = size
let elements = to_list
let min_elt = min_element    
let max_elt = max_element
