(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Red-black map
 * Dictionnaire bicolore
 * Diccionario bicromtico
 *
 *)

open BairePervasives

(* RedBlack trees are binary search trees in which the balance information
 * is encoded by the height of the tree.
 * The total height of the tree is then inferior to 2 log n
 * 
 * RedBlack trees can be seen as a specialisation of cartesian trees with
 * the given height function (cf cartesianTree.ml and localRedBlackSet.ml)
 * 
 * let rec height = function
 *   | E -> 0
 *   | N (a, x, b) -> 1 + min (height a) (height b)
 *)

(* Les arbres bicolores sont des arbres dont l'information de rquilibrage
 * encode la hauteur de l'arbre.
 * La hauteur de l'arbre est donc 2 log n
 *
 * Les arbres bicolores peuvent tre obtenus comme une spcialisation des 
 * arbres cartsiens avec pour cl (ou hauteur) associe 
 * (voir cartesianTree.ml et localRedBlackSet.ml)
 * 
 * let rec hauteur = function 
 *   | E -> 0 
 *   | N (a, x, b) -> 1 + min (hauteur a) (hauteur b)
 *)

(* Los rboles bicromticos son rboles binarios de busqueda cuya informacin 
 * de equilbrio se ha codificado en la altura del rbol.
 * La altura total del rbol no es mayor que 2 log n
 * 
 * Los rboles bicromticos pueden ser obtenidos especializando los rboles 
 * cartesianos con la funcin de altura siguiente 
 * (referirse a cartesianTree.ml y a localRedBlackSet.ml)
 * 
 * let rec altura = function
 *   | E -> 0
 *   | N (a, x, b) -> 1 + min (altura a) (altura b)
 *)

type ('a, 'b) map =
  | E
  | N of ('a, 'b) map * 'a * 'b * ('a, 'b) map * int


(** Utilitaires **)
(** Utils **)
(** tiles **)


let moduleName = "RedBlackMap"

let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type

(* The aim of this  smart constructor  is to hide the tree decorations 
 * and allow manipulation of trees as if they were usual (unbalanced)
 * binary search trees
 *)

(* Ce contructeur est destin  occulter les dcorations de l'arbre afin
 * de pouvoir manipuler les donnes comme des arbres binaires classiques
 *)

(* El propsito de ste constructor es ocultar las decoraciones del rbol
 * afin de poder manipular los datos como si fuesen un rboles binarios comnes
 *)

let makeTree = fun l k v r ->
  let h = match (l, r) with
    | (E, _) | (_, E) -> 1
    | (N (_, _, _, _, m), N (_, _, _, _, n)) -> 1 + min m n
  in
    N (l, k, v, r, h)

(* Note on the implementation of rotations :
 * The symetric order is conserved by rotations
 * 
 * examples :
 *  rotate_L : a x (ba y bb) -> (a x ba) y bb
 *  rotate_R : (aa y ab) x b -> aa y (ab x b)
 *  rotate_RL : a x ((baa y bab) z bb) -> (a x baa) y (bab z bb)
 *  rotate_LR : (aa x (aba y abb)) z b -> (aa x aba) y (abb z b) 
 *
 * This property allows a fast code correction verification
 *)

(* Note sur l'implmentation des rotations :
 * La notation infixe est drive de l'invariant des arbres de recherche 
 * donc l'ordre des termes (dit ordre symtrique) se conserve lors des 
 * rotations
 *
 * exemples : 
 *  rotate_L : a x (ba y bb) -> (a x ba) y bb
 *  rotate_R : (aa y ab) x b -> aa y (ab x b)
 *  rotate_RL : a x ((baa y bab) z bb) -> (a x baa) y (bab z bb)
 *  rotate_LR : (aa x (aba y abb)) z b -> (aa x aba) y (abb z b) 
 *
 * Cette proprit permet de vrifier immdiatement la correction
 * des algorithmes
 *)

(* Nota sobre la implementacin de las rotaciones :
 * El rden simtrico es conservado por las rotaciones
 * 
 * ejemplos :
 *  rotate_L : a x (ba y bb) -> (a x ba) y bb
 *  rotate_R : (aa y ab) x b -> aa y (ab x b)
 *  rotate_RL : a x ((baa y bab) z bb) -> (a x baa) y (bab z bb)
 *  rotate_LR : (aa x (aba y abb)) z b -> (aa x aba) y (abb z b) 
 *
 * sta propriedad permite verificar de manera imediata la correccin
 * de los algortmos 
 *)

let rotate_L = fun l k v r ->
  match r with
    | N (rl, rk, rv, rr, _) -> makeTree (makeTree l k v rl) rk rv rr
    | _ -> makeTree l k v r
	
let rotate_R = fun l k v r ->
  match l with
    | N (ll, lk, lv, lr, _) -> makeTree ll lk lv (makeTree lr k v r)
    | _ -> makeTree l k v r

let rotate_RL = fun l k v r ->
  match r with
    | N (N (rll, rlk, rlv, rlr, _), rk, rv, rr, _) -> 
	makeTree (makeTree l k v rll) rlk rlv (makeTree rlr rk rv rr)
    | _ -> makeTree l k v r

let rotate_LR = fun l k v r ->
  match l with
    | N (ll, lk, lv, N (lrl, lrk, lrv, lrr, _), _) ->
	makeTree (makeTree ll lk lv lrl) lrk lrv (makeTree lrr k v r)
    | _ -> makeTree l k v r

let height = function
  | E -> 0
  | N (_, _, _, _, n) -> n

(* [makeBTree] suposes that :
 * - the symetric order is respected l < v < r
 * - the two subtrees are well formed
 * - the height difference between l and r is not greater than 1
 *)

(* la fonction [makeBTree] suppose :
 * - le respect de l'ordre symtrique l < v < r
 * - les sous-arbres l et r tre bien forms
 * - la diffrence de hauteur entre l et r ne pas excder 1
 *)

(* la funcin [makeBTree] supone :
 * - el respeto del orden simtrico l < v < r
 * - los subrboles l y r estar bien formados
 * - la diferencia de altura entre l y r no valer ms de 1
 *)

let rec makeBTree = fun l k v r ->
  let hl = height l and hr = height r in
    match compare hl hr with
      | n when n < 0 ->
	  (match r with
	     | E -> makeTree l k v r
	     | N (rl, _, _, rr, _) ->
		 if height rr = hr then rotate_L l k v r
		 else if height rl = hr then rotate_RL l k v r
		 else makeTree l k v r
	  )
      | n when n > 0 ->
	  (match l with
	     | E -> makeTree l k v r
	     | N (ll, _, _, lr, _) ->
		 if height ll = hl then rotate_R l k v r
		 else if height lr = hl then rotate_LR l k v r
		 else makeTree l k v r
	  )
      | _ -> makeTree l k v r

(* makeSTree function suposes that :
 * - the symetric order is respected l < v < r
 * - the two subtrees are well formed
 *)

(* la fonction makeSTree suppose :
 * - le respect de l'ordre symtrique l < v < r
 * - les sous-arbres l et r tre forms
 *)

(* la funcin makeSTree supone :
 * - el respeto del orden simtrico l < v < r
 * - los subrboles l y r estar bien formados
 *)

let rec makeSTree = fun l k v r ->
  let hl = height l and hr = height r in
    if hl > hr + 1 then
      (match l with
	 | E -> makeTree l k v r
	 | N (ll, lk, lv, lr, _) -> makeBTree ll lk lv (makeSTree lr k v r)
      )
    else if hr > hl + 1 then
      (match r with
	 | E -> makeTree l k v r
	 | N (rl, rk, rv, rr, _) -> makeBTree (makeSTree l k v rl) rk rv rr
      )
    else makeBTree l k v r

(* [append] suposes that :
 * - the symetric order is respected l < r
 * - the two subtrees are well formed
 * - the height difference between l and r is not greater than 1
 *)

(* la fonction [append] suppose :
 * - le respect de l'ordre symtrique l < r
 * - les sous-arbres l et r tre bien forms
 * - la diffrence de hauteur entre l et r ne pas exder 1
 *)

(* la funcin [append] supone :
 * - el respeto del orden simtrico l < r
 * - los subrboles l y r estar bien formados
 * - la diferencia de altura entre l y r no valer ms de 1
 *)

let rec append = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l1, k1, v1, r1, _), N (l2, k2, v2, r2, _)) ->
        makeBTree l1 k1 v1 (makeBTree (append r1 l2) k2 v2 r2)

(* [appendS] suposes that :
 * - the symetric order is respected l < r
 * - the two subtrees are well formed
 *)

(* la fonction [appendS] suppose :
 * - le respect de l'ordre symtrique l < r
 * - les sous-arbres l et r tre bien forms
 *)

(* la funcin [appendS] supone :
 * - el respeto del orden simtrico l < r
 * - los subrboles l y r estar bien formados
 *)

let rec appendS = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l1, k1, v1, r1, _), N (l2, k2, v2, r2, _)) ->
        makeSTree l1 k1 v1 (makeSTree (appendS r1 l2) k2 v2 r2)


(** Oprations catgorielles **)
(** Categorial operations **)
(** Operaciones categoriales **)


let injection = fun k v -> makeTree E k v E

let rec insert_raise k v = function
  | E -> injection k v
  | N (l, k', v', r, _) ->
      match compare k k' with
	| n when n < 0 -> makeBTree (insert_raise k v l) k' v' r
	| n when n > 0 -> makeBTree l k' v' (insert_raise k v r)
	| _ when v <> v' -> makeTree l k v r
	| _ -> raise Found

let rec insert k v = function tree ->
  try 
    insert_raise k v tree 
  with Found -> tree

let rec from_list_rec result = function
  | [] -> result
  | (value, key) :: tail -> from_list_rec (insert value key result) tail

let from_list = function list -> from_list_rec E list

let rec to_list_rec result = function
  | E -> result
  | N (a, k, v, b, _) -> to_list_rec ((k, v) :: to_list_rec result b) a

let to_list = function map -> to_list_rec [] map

let map f = function set -> from_list (List.map f (to_list set))

(* Union algorithm :
 * Explained in the research report (avaible on the web)
 * 
 * Implementing Sets Efficiently in a Functional Language
 * Stephen Adams
 * CSTR 92-01
 * University of Southampton
 * 
 * We first give a version obtained by  divide and conquer 
 *)

(* Algorithme d'union :
 * Son fonctionnement est expliqu dans le rapport de recherche 
 * (disponible sur la toile)
 *
 * Implementing Sets Efficiently in a Functional Language
 * Stephen Adams
 * CSTR 92-01
 * University of Southampton
 *
 * Nous en rappelons ici le principe :
 * On construit une premire version via la mthode  diviser pour rgner 
 *)

(* Algoritmo de unin :
 * Su funcionamiento est explicado en el informe (disponible en la red)
 * 
 * Implementing Sets Efficiently in a Functional Language
 * Stephen Adams
 * CSTR 92-01
 * University of Southampton
 *
 * Recordamos rapidamente su principio :
 * Uno comienza construyendo una primera versin con el mtodo  dividir
 * para reinar 
 *)

let rec filter_lt x = function
  | E -> E
  | N (l, k, v, r, _) ->
      match compare x k with
	| n when n < 0 -> filter_lt x l
	| n when n > 0 -> makeSTree l k v (filter_lt x r)
	| _ -> l

let rec filter_gt x = function
  | E -> E
  | N (l, k, v, r, _) ->
      match compare x k with
	| n when n < 0 -> makeSTree (filter_gt x l) k v r
	| n when n > 0 -> filter_gt x r
	| _ -> r

(* We could have used a unique partition function like [partition_inf_sup]
 * but the to filter functions will be useful later in [plus]
 *)

(* Nous aurions pu utiliser une unique fonction de partition telle
 * [partition_inf_sup] mais les filtres dfinis nous seront utiles sparment 
 * dans la version optimise de [plus]
 *)
	
(* Ubisemos podido utilizar una nica funcin de particin [partition_inf_sup]
 * pero los dos filtros que hemos definido nos sern tiles en la versin
 * optimizada de la funcin [plus]
 *)

let rec old_plus = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l, k, v, r, _), N _) -> 
	let 
	  lt_k = filter_lt k b and 
	  gt_k = filter_gt k b 
	in
	  makeSTree (old_plus l lt_k) k v (old_plus r gt_k)
	    
(* One can notice that there is no need to construct a tree (lt_k and gt_k)
 * which will be destructed just after by the recursive call to [old_plus]. 
 * That is why we introduce a triple containing a tree and two bounds. 
 * These bounds can be easely moved then.
 *)

(* On constate qu' rien ne sert de construire un arbre (lt_k et gt_k) qui sera
 * dcoup juste aprs par l'appel rcursif de la fonction [old_plus], raison 
 * pour laquelle on introduit un triplet contenant un arbre et deux bornes 
 * qui pourront tre modifies  moindre cot
 *)

(* Uno se da cuenta que de nada sirve construir un rbol (lt_k y gt_k) que ser
 * destruido justo despus al llamar recursivamente la funcin [old_plus].
 * Se introduce por esa razn un triplete formado por un rbol y dos 
 * limites, lmites que se podrn modificar lteriormente a poco costo 
 *)

(* [cut] constructs a triple, making sure that the root of the tree lies 
 * between the two bounds. If it is not the case, the functions calls itself
 * recusively on the corresponding subtree (right or left)
 *)

(* [cut] construit un triplet de faon  ce que la racine se trouve entre
 * les deux bornes. Si ce n'est pas le cas, la fonction restreint rcursivement
 * l'arbre  son sous-arbre gauche ou droit selon le cas de figure.
 *) 

(* [cut] constuye un triplete de manera tal a que la raz del rbol se 
 * encuentre entre los dos lmites. Si este no es el caso, la funcin
 * toma recursivamente el subrbol derecho o izquierdo segn convenga
 *)

let rec cut = fun lo tree hi ->
  match tree with
    | N (_, k, _, r, _) when compare k lo <= 0 -> cut lo r hi
    | N (l, k, _, _, _) when compare k hi >= 0 -> cut lo l hi
    | _ -> (lo, tree, hi)

(* [cut_lo] and [cut_hi] are specialized versions with an infinite 
 * hight (respectively low) bound 
 *)

(* [cut_lo] et [cut_hi] sont des versions spcialises avec la borne infrieure
 * (respectivement suprieure) ayant pour valeur plus l'infini (respectivement 
 * moins l'infini)
 *)

(* [cut_lo] y [cut_hi] son versiones especializadas con un limite inferior
 * (respectivamente superior) igual a ms (respectivamente menos) el infinito
 *)

let rec cut_lo = fun tree hi ->
  match tree with
    | N (l, k, _, _, _) when compare k hi >= 0 -> cut_lo l hi
    | _ -> (tree, hi)

let rec cut_hi = fun lo tree ->
  match tree with
    | N (_, k, _, r, _) when compare k lo <= 0 -> cut_hi lo r
    | _ -> (lo, tree)

(* [uni] uses the left tree to control the reconstruction :
 * - the right tree will be physically cut only if it lies between to 
 * successive values of the control tree (which guarantees that there will not
 * be any new cut after the reconstruction)
 * - if the right tree is empty, the whole control tree can be inserted
 *)

(* la fonction [uni] se sert de l'arbre de gauche pour contrler la
 * reconstruction : 
 * - l'arbre de droite ne sera dcoup physiquement que si l'on se trouve 
 * entre deux valeurs successives de l'arbre de contrle (ce qui garantit 
 * qu'il n'y aura pas ne nouvelle dcoupe aprs la reconstruction)
 * - si l'arbre de droite est vide, on peut insrer tout l'arbre de contrle
 *)

(* la funcin [uni] se sirve del rbol izquierdo para controlar la 
 * reconstruccin :
 * - el rbol derecho solo sera recortado fsicamente si se encuentra
 * entre dos valores succesivos del rbol de control (lo que garantiza
 * que no habr ninguna otra operacin de corte despues de la reconstruccin)
 * - si el rbol derecho est vaco, se puede insertar el rbol de control
 * entero
 *)

let rec uni = fun a (lo, b, hi) ->
  match (a, b) with
    | (_, E) -> a
    | (E, N (l, k, v, r, _)) -> makeSTree (filter_gt lo l) k v (filter_lt hi r)
    | (N (l, k, v, r, _), N _) ->
	makeSTree (uni l (cut lo b k)) k v (uni r (cut k b hi))  

(* two specialized versions are given *)

(* on en donne galement deux versions spcialises *)

(* se dan tambin dos versiones especializadas *)

let rec uni_lo = fun a (b, hi) ->
  match (a, b) with
    | (_, E) -> a
    | (E, N (l, k, v, r, _)) -> makeSTree l k v (filter_lt hi r)
    | (N (l, k, v, r, _), N _) ->
	makeSTree (uni_lo l (cut_lo b k)) k v (uni r (cut k b hi))

let rec uni_hi = fun a (lo, b) ->
  match (a, b) with
    | (_, E) -> a
    | (E, N (l, k, v, r, _)) -> makeSTree (filter_gt lo l) k v r
    | (N (l, k, v, r, _), N _) -> 
	makeSTree (uni l (cut lo b k)) k v (uni_hi r (cut_hi k b)) 

(* We just put all together *)
	
(* On assemble le tout *)

(* Se juntan todos los pedazos *)

let plus = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l, k, v, r, _), N _) ->
	makeSTree (uni_lo l (cut_lo b k)) k v (uni_hi r (cut_hi k b)) 

let zero = function () -> E

let rec from_tree = function
  | Leaf -> E
  | Node (l, (k, v), r) -> makeSTree (from_tree l) k v (from_tree r)

let rec to_tree = function
  | E -> Leaf
  | N (l, k, v, r, _) -> Node (to_tree l, (k, v), to_tree r)


(** Oprations binaires **)
(** Binary operations **)
(** Opraciones binrias **)


let rec partition_inf_sup pivot = function
  | E -> (E, None, E)
  | N (l, k, v, r, _) as node ->
      match compare pivot k with
	| n when n < 0 ->
	    let (inf, value, sup) = partition_inf_sup pivot l in
	      (inf, value, makeSTree sup k v r)
	| n when n > 0 ->
	    let (inf, value, sup) = partition_inf_sup pivot r in
	      (makeSTree l k v inf, value, sup)
	| _ -> (l, Some (k, v), r)

let union = plus

let rec intersection = fun a b ->
  match (a, b) with
    | (E, _) | (_, E) -> E
    | (N (l, k, v, r, _), N _) ->
	match partition_inf_sup k b with
	  | (l', None, r') -> appendS (intersection l l') (intersection r r')
	  | (l', Some _, r') -> 
	      makeSTree (intersection l l') k v (intersection r r')

let rec difference = fun a b ->
  match (a, b) with
    | (E, _) -> E
    | (_, E) -> a
    | (N (l, k, v, r, _), N _) -> 
	match partition_inf_sup k b with
	  | (l', None, r') -> makeSTree (difference l l') k v (difference r r')
	  | (l', Some _, r') -> appendS (difference l l') (difference r r')

let rec symetric_difference = fun a b ->
  match (a, b) with
    | (E, s) | (s, E) -> s
    | (N (l, k, v, r, _), N _) ->
	match partition_inf_sup k b with
	  | (l', None, r') -> makeSTree (symetric_difference l l') k v
	      (symetric_difference r r')
	  | (l', Some _, r') -> 
	      appendS (symetric_difference l l') (symetric_difference r r')

let rec subset = fun a b ->
  match (a, b) with
    | (E, _) -> true
    | (_, E) -> false
    | (N (l1, k1, v1, r1, _), N (l2, k2, v2, r2, _)) ->
        match compare k1 k2 with
	  | n when n < 0 -> subset (makeBTree l1 k1 v1 E) l2 && subset r1 b
	  | n when n > 0 -> subset (makeBTree E k1 v1 r1) r2 && subset l1 b
	  | _ -> v1 = v2 && subset l1 l2 && subset r1 r2

let rec compare_rec = fun a b ->
  match (a, b) with
    | ([], []) -> 0
    | ([], _) -> -1
    | (_, []) -> 1
    | (E :: t1, E :: t2) -> compare_rec t1 t2
    | (N (E, v1, k1, r1, _) :: t1, N (E, v2, k2, r2, _) :: t2) ->
	(match compare v1 v2 with
	   | 0 when k1 = k2 -> compare_rec (r1 :: t1) (r2 :: t2)
	   | other -> other
	)
    | (N (a, v1, k1, r1, _) :: t1, t2) ->
        compare_rec (a :: N (E, v1, k1, r1, 0) :: t1) t2
    | (t1, N (b, v2, k2, r2, _) :: t2) ->
	compare_rec t1 (b :: N (E, v2, k2, r2, 0) :: t2)

let equal = fun a b -> compare_rec [a] [b] = 0

(* let compare = fun a b -> compare_rec [a] [b]
 *
 * renvoy  la fin du fichier pour ne pas interfrer avec Pervasives.compare 
 * 
 * see the end of the file
 * 
 * ver el final del fichero
 *)


(** Accs **)
(** Acces **)
(** Acceso **)


(* [insert'] :
 * Variant of the [insert] function, it does only one comparasion
 * by branch instead of 3/2 but has to walk down the whole tree
 * 
 * If the inserted element is not already in the tree, the [insert] function
 * has anyway to go down the tree until the leafs, so in this particular
 * point the lost is not very important. The reason why this function
 * is not always faster than [insert] is that one more argument
 * must be transmitted in each recursive call
 *)

(* [insert'] :
 * Variante de la fonction [insert] ne faisant qu'une comparaison par
 * branche (au lieu de 3/2 en moyenne pour la fonction habituelle) mais
 * parcourt l'arbre jusqu' profondeur maximale
 *
 * Hormis le cas o l'lment insr se trouve dj dans l'arbre, la fonction
 * [insert] effectue un parcours sur toute la profondeur. La perte est donc
 * minime sur ce point, mais un argument supplmentaire doit tre transmis, 
 * raison pour laquelle cette fonction n'est pas toujours plus rapide que 
 * la version la plus simple
 *)

(* [insert'] :
 * Una variante de la funcin [insert] que hace una sola comparacin por
 * nivel (en vez de 3/2 en promedio para la version comn) pero recorre la 
 * totalidad del rbol, hasta las hojas.
 * 
 * Excepto en el caso en el cual el lemento que se est insertando se 
 * encuentra ya en el rbol, la funcin [insert] recorre de todas maneras
 * el rbol hasta las hojas, por lo cual la perdida en ese aspecto es
 * mnima. No obstante, un argumento suplementrio debe ser transmitido y
 * por ello esta funcin no es todo el tiempo ms rpida que la version
 * la mas simple
 *)

let rec insert_min_bound b k v = function
  | E -> 
      if compare b k = 0 then raise Found
      else injection k v
  | N (l, k', v', r, _) ->
      match compare k k' with
	| n when n < 0 -> makeBTree (insert_min_bound b k v r) k' v' l
	| _ -> makeBTree l k' v' (insert_min_bound k' k v r)

let rec insert_raise' k v = function
  | E -> injection k v
  | N (l, k', v', r, _) ->
      match compare k k' with
	| n when n < 0 -> makeBTree (insert_raise' k v l) k' v' r
	| _ -> makeBTree l k' v' (insert_min_bound k' k v r)

let insert' k v = function tree ->
  try 
    insert_raise' k v tree 
  with Found -> tree

let insert_list = fun list map -> plus (from_list list) map

let rec delete_min = function
  | E -> E
  | N (E, _, _, r, _) -> r
  | N (l, k, v, r, _) -> makeBTree (delete_min l) k v r
      
let rec delete_max = function
  | E -> E
  | N (l, _, _, E, _) -> l
  | N (l, k, v, r, _) -> makeBTree l k v (delete_max r)

let rec delete_raise x = function
  | E -> raise Not_found
  | N (l, k, v, r, _) ->
      match compare x k with
	| n when n < 0 -> makeBTree (delete_raise x l) k v r
	| n when n > 0 -> makeBTree l k v (delete_raise x r)
	| _ -> append l r

let delete x = function map -> 
  try 
    delete_raise x map 
  with Not_found -> map

let rec delete_list = fun list map ->
  match list with
    | [] -> map
    | head :: tail -> delete_list tail (delete head map)

let rec filter_rec f result = function
  | E -> result
  | N (l, k, v, r, _) ->
      filter_rec f 
      (filter_rec f 
	 (if f k v then insert k v result else result) 
	 l) 
      r

let filter f = function map -> filter_rec f E map

let rec filter_le x = function
  | E -> E
  | N (l, k, v, r, _) ->
      match compare x k with
	| n when n < 0 -> filter_le x l
	| _ -> makeSTree l k v (filter_le x r)

let rec filter_ge x = function
  | E -> E
  | N (l, k, v, r, _) ->
      match compare x k with
	| n when n <= 0 -> makeSTree (filter_ge x l) k v r
	| _ -> filter_ge x r

let rec range min max = function
  | E -> E
  | N (l, k, v, r, _) ->
      if compare k min < 0 then range min max r
      else if compare k max > 0 then range min max l
      else makeSTree (filter_ge min l) k v (filter_le max r)


(** Vues **)
(** Views **)
(** Vistas **)


let rec min_view = function
  | E -> raise Empty
  | N (E, k, v, r, _) -> ((k, v), r)
  | N (l, k, v, r, _) -> 
      let (min, left) = min_view l in (min, makeBTree left k v r)
      
let rec max_view = function
  | E -> raise Empty
  | N (l, k, v, E, _) -> (l, (k, v))
  | N (l, k, v, r, _) -> let (right, max) = max_view r in 
      (makeBTree l k v right, max)

let rec partition_rec f (trueS, falseS) = function
  | E -> (trueS, falseS)
  | N (l, k, v, r, _) -> 
      if f k v
      then partition_rec f (partition_rec f (insert k v trueS, falseS) l) r
      else partition_rec f (partition_rec f (trueS, insert k v falseS) l) r

let partition = fun f map -> partition_rec f (E, E) map


(** Observers **)
(** Observateurs **)
(** Observadores **)


let is_empty = function
  | E -> true
  | N _ -> false

let rec size_rec result = function
  | E -> result
  | N (l, _, _, r, _) -> size_rec (size_rec (1 + result) r) l

let size = function map -> size_rec 0 map

let rec keys_rec result = function
  | E -> result
  | N (l, k, _, r, _) -> keys_rec (k :: keys_rec result r) l

let keys = function map -> keys_rec [] map

let rec images_rec result = function
  | E -> result
  | N (l, _, v, r, _) -> images_rec (v :: images_rec result r) l

let images = function map -> images_rec [] map

(* [member'] : does (d + 1) comparations instead of 3/2 * d
 * following an idea of Arne Andersson, but walks down the whole height of 
 * the tree and an extra argument has to be transmitted in each call
 * 
 * It may not always be faster than [member]
 *)

(* [member'] : effectue en moyenne (d + 1) comparaisons au lieu de 3/2 * d
 * selon une ide d'Arne Andersson mais parcourt systmatiquement l'arbre sur
 * toute sa hauteur et doit transmettre un paramtre supplmentaire  chaque
 * appel
 * 
 * Il est possible qu'elle ne soit pas toujours plus rapide que [member]
 *) 

(* [member'] : hace (d + 1) comparaciones en vez de 3/2 * d en promedio, 
 * segn una idea de Arne Andersson, pero recorre toda la altura del rbol
 * y requiere un parmetro suplementrio en cada llamado
 * 
 * Es posible que no sea siempre ms veloz que [member]
 *)

let rec member_min_bound b k = function
  | E -> compare b k = 0
  | N (l, k', _, r, _) ->
      match compare k k' with
	| v when v < 0 -> member_min_bound b k l
	| _ -> member_min_bound k' k r

let rec member' k = function
  | E -> false
  | N (l, k', _, r, _) ->
      match compare k k' with
	| n when n < 0 -> member' k l
	| _ -> member_min_bound k' k r

(* [member] : usual member function
 * some algorithms use the properties of the usual [member] function 
 * (stop of a positive search in the middle of the tree). It is the case
 * of optimal tree construction for example (Knuthm Garcia-Walsh, confer
 * The art of computer programming vol III)
 *)

(* [member] : fonction member habituelle 
 * un certain nombre d'algorithmes utilisent les proprits de la fonction
 * [member] habituelle (arrt d'une recherche positive  l'intrieur de 
 * l'arbre) par exemple les algorithmes de construction d'arbres optimaux
 * (Knuth, Garcia-Walsh, confer The art of computer programming vol III)
 *)

(* [member] : funcin member comn
 * ciertos algortmos usan las propriedades de la funcin [member] comn
 * (interupcin de una busqueda positiva en el medio del rbol) tal
 * los algortmos de construccin de rboles optimales (Knuth, Garcia-Walsh,
 * confer The art of computer programming vol III)
 *)

let rec member k = function
  | E -> false
  | N (l, k', _, r, _) ->
      match compare k k' with
	| n when n < 0 -> member k l
	| n when n > 0 -> member k r
	| _ -> true

let rec lookup k = function
  | E -> raise Not_found
  | N (l, k', v', r, _) ->
      match compare k k' with
	| n when n < 0 -> lookup k l
	| n when n > 0 -> lookup k r
	| _ -> v'

let rec look_for f = function
  | E -> raise Not_found
  | N (_, k, v, _, _) when f k v -> (k, v)
  | N (l, _, _, r, _) -> (try look_for f l with Not_found -> look_for f r)

let choose = function
  | E -> raise Empty
  | N (_, k, v, _, _) -> (k, v)

let rec min_element = function
  | E -> raise Not_found
  | N (E, k, v, _, _) -> (k, v)
  | N (l, _, _, _, _) -> min_element l

let rec max_element = function
  | E -> raise Not_found
  | N (_, k, v, E, _) -> (k, v)
  | N (_, _, _, r, _) -> max_element r

(* Could also be done with ['a option] *)

(* Auraient aussi pu tre codes au moyen de ['a option], nous avons
 * prfr illustrer les interruptions de calcul au moyens d'exceptions
 *)

(* stas operaciones tambin se hubiesen podido escribir con el typo 
 * ['a option] pero hemos querido aqu mostrar el uso de las excepciones para 
 * ruptura de clculo
 *)

let rec next x = function
  | E -> raise Not_found
  | N (l, k, v, r, _) ->
      match compare x k with
	| n when n < 0 -> (try next x l with Not_found -> (k, v))
	| n when n > 0 -> next x r
	| _ -> min_element r (* raises [Not_found] *)

let rec previous x = function
  | E -> raise Not_found
  | N (l, k, v, r, _) ->
      match compare x k with
	| n when n < 0 -> previous x l
	| n when n > 0 -> (try previous x r with Not_found -> (k, v))
	| _ -> max_element l (* raises [Not_found] *)

let rec ceil x = function
  | E -> raise Not_found
  | N (l, k, v, r, _) ->
      match compare x k with
	| n when n < 0 -> (try ceil x l with Not_found -> (k, v))
	| n when n > 0 -> ceil x r
	| _ -> (k, v)

let rec floor x = function
  | E -> raise Not_found
  | N (l, k, v, r, _) ->
      match compare x k with
	| n when n < 0 -> floor x l
	| n when n > 0 -> (try floor x r with Not_found -> (k, v))
	| _ -> (k, v)

(* les arbres binaires de recherche sont indexs suivant l'ordre symtrique *)

(* binary search trees are indexed following the symetric order *)

(* los rboles de busqueda estn indizados segn el orden simtrico *)

let rec rank x = function
  | E -> raise Not_found
  | N (l, k, _, r, _) ->
      match compare x k with
	| n when n < 0 -> rank x l
	| n when n > 0 -> 1 + size l + rank x r
	| _ -> size l

let rec get n = function
  | E -> raise (Out_of_bounds n)
  | N (l, k, v, r, _) ->
      let s = size l in
	if n < s then get n l
	else if n > s then get (n - s - 1) r
	else (k, v)

let rec for_all f = function
  | E -> true
  | N (l, k, v, r, _) -> f k v && for_all f l && for_all f r

let rec exists f = function
  | E -> false
  | N (l, k, v, r, _) -> f k v || exists f l || exists f r


(** Traversals **)
(** Parcours **)
(** Recorridos **)


let rec iter f = function
  | E -> ()
  | N (l, k, v, r, _) -> iter f l ; f k v ; iter f r

let rec fold_left f e = function
  | E -> e
  | N (l, k, v, r, _) -> fold_left f (f (fold_left f e l) k v) r

let rec fold_right = fun f set e ->
  match set with
    | E -> e 
    | N (l, k, v, r, _) -> fold_right f l (f k v (fold_right f r e))

let fold = fold_right


(** Miscellaneous **)
(** Divers **)
(** Miscelanios **)


let rec insert_with f k v = function
  | E -> injection k v
  | N (l, k', v', r, _) ->
      match compare k k' with
	| n when n < 0 -> makeBTree (insert_with f k v l) k' v' r
	| n when n > 0 -> makeBTree l k' v' (insert_with f k v r)
	| _ -> makeBTree l k (f v v') r

let rec union_with f = fun a b ->
  match (a, b) with
    | (E, _) -> a
    | (_, E) -> b
    | (N (l, k, v, r, _), N _) ->
	match partition_inf_sup k b with
	  | (l', None, r') -> 
	      makeSTree (union_with f l l') k v (union_with f r r')
	  | (l', Some (_, v'), r') ->
	      makeSTree (union_with f l l') k (f v v') (union_with f r r')

let rec intersection_with f = fun a b ->
  match (a, b) with
    | (E, _) | (_, E) -> E
    | (N (l, k, v, r, _), N _) ->
	match partition_inf_sup k b with
	  | (l', None, r') -> 
	      appendS (intersection_with f l l') (intersection_with f r r')
	  | (l', Some (_, v'), r') ->
	      makeSTree (intersection_with f l l') k (f v v')
	      (intersection_with f r r')

let rec map_images_with_index f = function 
  | E -> E
  | N (l, k, v, r, h) -> 
      N (map_images_with_index f l, k, f k v, map_images_with_index f r, h)

let rec map_images f = function
  | E -> E
  | N (l, k, v, r, h) -> N (map_images f l, k, f v, map_images f r, h)

let rec unsafe_insert_min k v = function
  | E -> injection k v
  | N (l, k', v', r, _) -> makeTree (unsafe_insert_min k v l) k' v' r 

let rec unsafe_insert_max k v = function
  | E -> injection k v
  | N (l, k', v', r, _) -> makeTree l k' v' (unsafe_insert_max k v r)


(** Compatibilit avec la librairie standard de Caml **)
(** Caml standard library compatibility **)
(** Compatibilidad con la librera estandar de Caml **)


let empty = E
let mem = member
let add = insert
let singleton = injection
let remove = delete
let inter = intersection
let diff = difference
let cardinal = size
let elements = to_list
let min_elt = min_element    
let max_elt = max_element
let find = lookup
let mapi = map_images_with_index

(* compatibility problems with 
 * - map : more general in Baire library
 *)

(* problmes de compatibilit avec
 * - map : la signature de [map] dans Baire est plus gnrale
 *)

(* problemas de compatibilidad con
 * - map : el typo de [map] est ms general en Baire
 *)

let std_map = map_images


(** Objets **)
(** Ojects **)
(** Objetos **)


class ['a, 'b] o_map = fun initial_map ->

object (self : 'c)

  val contents  = initial_map

  method to_map = contents

  (* Categorial operations *)
  (* Oprations categorielles *)
  (* Operaciones categoriales *)
		    
  method map f = {< contents = map f contents >}

  method plus = function (set : 'c) -> 
    {< contents = plus contents set#to_map >} 

  method to_list = to_list contents
		     
  (* Binary operations *)
  (* Oprations binaires *)
  (* Operaciones binrias *)

  method union = function (set : 'c) -> 
    {< contents = plus contents set#to_map >}
  
  method intersection = function (set : 'c) -> 
    {< contents = intersection contents set#to_map >}
  
  method difference = function (set : 'c) -> 
    {< contents = difference contents set#to_map >}
  
  method symetric_difference = function (set : 'c) -> 
    {< contents = symetric_difference contents set#to_map >}
    
  method subset = function (set : 'c) -> subset contents set#to_map
  method equal = function (set : 'c)-> equal contents set#to_map
  method compare = function (set : 'c) -> compare_rec [contents] [set#to_map]

  (* Acces *)
  (* Accs *)
  (* Acceso *)
    
  method insert = fun k v -> 
    try 
      {< contents = insert_raise k v contents >} 
    with Found -> self

  method insert_raise = fun k v -> {< contents = insert_raise k v contents >}

  method insert' = fun k v -> 
    try
      {< contents = insert_raise' k v contents >}
    with Found -> self

  method insert_raise' = fun k v -> {< contents = insert_raise' k v contents >}
      
  method insert_list = function list -> 
    {< contents = insert_list list contents >}

  method delete = function k -> 
    try 
      {< contents = delete_raise k contents >} 
    with Not_found -> self

  method delete_raise = function k -> {< contents = delete_raise k contents >}

  method delete_list = function list -> 
    {< contents = delete_list list contents >}
      
  method delete_min = {< contents = delete_min contents >}
  method delete_max = {< contents = delete_max contents >}

  method filter = function f -> {< contents = filter f contents >}
  method filter_lt = function x -> {< contents = filter_lt x contents >}
  method filter_gt = function x -> {< contents = filter_gt x contents >}
  method filter_le = function x -> {< contents = filter_le x contents >}
  method filter_ge = function x -> {< contents = filter_ge x contents >}

  method range = fun min max -> {< contents = range min max contents >}

  (* Views *)
  (* Vues *)
  (* Vistas *)

  method min_view =
    let (min, other) = min_view contents in
      (min, {< contents = other >})

  method max_view =
    let (other, max) = max_view contents in
      ({< contents = other >}, max)

  method partition = function f -> 
    let (trueSet, falseSet) = partition f contents in
      ({< contents = trueSet >}, {< contents = falseSet >})

  method partition_inf_sup = function x ->
    let (inf, value, sup) = partition_inf_sup x contents in
      ({< contents = inf >}, value, {< contents = sup >})

  (* Observators *)
  (* Observateurs *)
  (* Observadores *)
      
  method is_empty = contents = E
  method size = size contents

  method keys = keys contents
  method images = images contents

  method member = function x -> member x contents
  method member' = function x -> member' x contents

  method lookup = function x -> lookup x contents
  method look_for = function f -> look_for f contents

  method choose =
    match contents with
      | E -> raise Not_found
      | N (_, k, v, _, _) -> (k, v)

  method min_element = min_element contents
  method max_element = max_element contents

  method previous = function x -> previous x contents
  method next = function x -> next x contents
  method floor = function x -> floor x contents
  method ceil = function x -> ceil x contents

  method rank = function x -> rank x contents
  method get = function n -> get n contents

  method exists = function f -> exists f contents
  method for_all = function f -> for_all f contents

  (* Traversals *)
  (* Parcours *)
  (* Recorridos *)

  method iter :  ('a -> 'b -> unit) -> unit = function f -> iter f contents

  method fold : 'd. ('a -> 'b -> 'd -> 'd) -> 'd -> 'd = 
    fun f e -> fold_right f contents e

  method fold_left : 'd. ('d -> 'a -> 'b -> 'd) -> 'd -> 'd =
    fun f e -> fold_left f e contents

  method fold_right : 'd. ('a -> 'b -> 'd -> 'd) -> 'd -> 'd =
    fun f e -> fold_right f contents e

  (* Miscellaneous *)
  (* Divers *)
  (* Micelaneos *)

  method insert_with = fun f k v -> {< contents = insert_with f k v contents >}

  method union_with f = fun (map : 'c) -> 
    {< contents = union_with f contents map#to_map >}

  method intersection_with f = fun (map : 'c) ->
    {< contents = intersection_with f contents map#to_map >}

end

(* Constructors *)
(* Constructeurs *)
(* Constructores *)

let o_injection = fun k v -> new o_map (injection k v)
let o_map = fun f a -> new o_map (map f a#to_map)
let o_plus = fun a b -> new o_map (plus a#to_map b#to_map)
let o_zero = function () -> new o_map E
let o_from_list = function list -> new o_map (from_list list)
let o_from_map = function map -> new o_map map

(* Binary operations *)
(* Oprations binaires *)
(* Operaciones binrias *)
  
let o_union = fun a b -> new o_map (plus a#to_map b#to_map)
let o_intersection = fun a b -> new o_map (intersection a#to_map b#to_map)
let o_difference = fun a b -> new o_map (difference a#to_map b#to_map)
let o_symetric_difference = fun a b -> 
  new o_map (symetric_difference a#to_map b#to_map)
  
let o_subset = fun a b -> subset a#to_map b#to_map
let o_equal = fun a b -> equal a#to_map b#to_map
let o_compare = fun a b -> compare_rec [a#to_map] [b#to_map]

let o_union_with f = fun a b -> new o_map (union_with f a#to_map b#to_map)

let o_intersection_with f = fun a b -> 
  new o_map (intersection_with f a#to_map b#to_map)


module type Map =
sig
  
  type key
  type 'a t

  val injection : key -> 'a -> 'a t
  val map : (key * 'a -> key * 'b) -> 'a t -> 'b t
  val plus : 'a t -> 'a t -> 'a t
  val zero : unit -> 'a t
  val from_list : (key * 'a) list -> 'a t
  val to_list : 'a t -> (key * 'a) list
    
  val union        : 'a t -> 'a t -> 'a t
  val intersection : 'a t -> 'a t -> 'a t
  val difference   : 'a t -> 'a t -> 'a t

  val subset  : 'a t -> 'a t -> bool
  val equal   : 'a t -> 'a t -> bool
  val compare : 'a t -> 'a t -> int

  val insert  : key -> 'a -> 'a t -> 'a t
  val insert' : key -> 'a -> 'a t -> 'a t
  val insert_list : (key * 'a) list -> 'a t -> 'a t

  val insert_raise  : key -> 'a -> 'a t -> 'a t (* [Found] *)
  val insert_raise' : key -> 'a -> 'a t -> 'a t (* [Found] *)

  val delete : key -> 'a t -> 'a t
  val delete_min : 'a t -> 'a t
  val delete_max : 'a t -> 'a t
  val delete_list : key list -> 'a t -> 'a t

  val delete_raise : key -> 'a t -> 'a t (* [Not_found] *)

  val filter : (key -> 'a -> bool) -> 'a t -> 'a t
  val filter_lt : key -> 'a t -> 'a t
  val filter_gt : key -> 'a t -> 'a t
  val filter_le : key -> 'a t -> 'a t
  val filter_ge : key -> 'a t -> 'a t

  val range : key -> key -> 'a t -> 'a t

  val min_view : 'a t -> (key * 'a) * 'a t (* [Empty] *)
  val max_view : 'a t -> 'a t * (key * 'a) (* [Empty] *)

  val partition : (key -> 'a -> bool) -> 'a t -> 'a t * 'a t 
  val partition_inf_sup : key -> 'a t -> 'a t * (key * 'a) option * 'a t

  val is_empty : 'a t -> bool
  val size : 'a t -> int

  val keys   : 'a t -> key list
  val images : 'a t -> 'a list

  val member  : key -> 'a t -> bool
  val member' : key -> 'a t -> bool

  val lookup : key -> 'a t -> 'a (* [Not_found] *)
  val look_for : (key -> 'a -> bool) -> 'a t -> key * 'a (* [Not_found] *)

  val choose      : 'a t -> key * 'a (* [Empty] *) 
  val min_element : 'a t -> key * 'a (* [Empty] *)
  val max_element : 'a t -> key * 'a (* [Empty] *)

  val previous : key -> 'a t -> (key * 'a) (* [Not_found] *)
  val next     : key -> 'a t -> (key * 'a) (* [Not_found] *)
  val floor    : key -> 'a t -> (key * 'a) (* [Not_found] *)
  val ceil     : key -> 'a t -> (key * 'a) (* [Not_found] *)

  val rank : key -> 'a t -> int      (* [Not_found] *)
  val get  : int -> 'a t -> key * 'a (* [Out_of_bounds] *)

  val for_all : (key -> 'a -> bool) -> 'a t -> bool
  val exists  : (key -> 'a -> bool) -> 'a t -> bool

  val iter : (key -> 'a -> unit) -> 'a t -> unit

  val fold : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b
  val fold_left  : ('b -> key -> 'a -> 'b) -> 'b -> 'a t -> 'b 
  val fold_right : (key -> 'a -> 'b -> 'b) -> 'a t -> 'b -> 'b

  val insert_with : ('a -> 'a -> 'a) -> key -> 'a -> 'a t -> 'a t
  val union_with  : ('a -> 'a -> 'a) -> 'a t -> 'a t -> 'a t
  val intersection_with : ('a -> 'a -> 'a) -> 'a t -> 'a t -> 'a t

  val empty : 'a t
  val mem : key -> 'a t -> bool
  val add : key -> 'a -> 'a t -> 'a t
  val singleton : key -> 'a -> 'a t
  val remove : key -> 'a t -> 'a t
  val inter : 'a t -> 'a t -> 'a t
  val diff : 'a t -> 'a t -> 'a t
  val cardinal : 'a t -> int
  val elements : 'a t -> (key * 'a) list
  val min_elt : 'a t -> key * 'a (* [Not_found] *)
  val max_elt : 'a t -> key * 'a (* [Not_found] *)
  val find : key -> 'a t -> 'a (* [Not_found] *)
  val mapi : (key -> 'a -> 'b) -> 'a t -> 'b t

  val std_map : ('a -> 'b) -> 'a t -> 'b t
    
  class type ['a] o_t =
  object ('b)
    
    method map : (key * 'a -> key * 'a) -> 'b
    method plus : 'b -> 'b
    method to_list : (key * 'a) list
    method to_map : 'a t

    method union : 'b -> 'b
    method intersection : 'b -> 'b
    method difference : 'b -> 'b
    method symetric_difference : 'b -> 'b

    method subset  : 'b -> bool
    method equal   : 'b -> bool
    method compare : 'b -> int

    method insert  : key -> 'a -> 'b
    method insert' : key -> 'a -> 'b
    method insert_list : (key * 'a) list -> 'b

    method insert_raise  : key -> 'a -> 'b               (* [Found] *)
    method insert_raise' : key -> 'a -> 'b               (* [Found] *)

    method delete : key -> 'b
    method delete_list : key list -> 'b
    method delete_max : 'b
    method delete_min : 'b
    method delete_raise : key -> 'b                      (* [Not_found] *)
  
    method filter : (key -> 'a -> bool) -> 'b
    method filter_ge : key -> 'b
    method filter_gt : key -> 'b
    method filter_le : key -> 'b
    method filter_lt : key -> 'b

    method range : key -> key -> 'b

    method min_view : (key * 'a) * 'b                    (* [Empty] *)
    method max_view : 'b * (key * 'a)                    (* [Empty] *)

    method partition : (key -> 'a -> bool) -> 'b * 'b
    method partition_inf_sup : key -> 'b * (key * 'a) option * 'b

    method is_empty : bool
    method size : int

    method images : 'a list
    method keys : key list

    method member  : key -> bool
    method member' : key -> bool
      
    method lookup : key -> 'a                            (* [Not_found] *)
    method look_for : (key -> 'a -> bool) -> key * 'a    (* [Not_found] *)

    method choose : key * 'a                             (* [Empty] *)
    method max_element : key * 'a                        (* [Not_found] *)
    method min_element : key * 'a                        (* [Not_found] *)

    method next     : key -> key * 'a                    (* [Not_found] *)
    method previous : key -> key * 'a                    (* [Not_found] *)
    method ceil     : key -> key * 'a                    (* [Not_found] *)
    method floor    : key -> key * 'a                    (* [Not_found] *)

    method get : int -> key * 'a                         (* [Out_of_bounds] *)
    method rank : key -> int                             (* [Not_found] *)

    method exists : (key -> 'a -> bool) -> bool
    method for_all : (key -> 'a -> bool) -> bool

    method iter : (key -> 'a -> unit) -> unit
    method fold : (key -> 'a -> 'd -> 'd) -> 'd -> 'd
    method fold_left : ('d -> key -> 'a -> 'd) -> 'd -> 'd
    method fold_right : (key -> 'a -> 'd -> 'd) -> 'd -> 'd

    method insert_with : ('a -> 'a -> 'a) -> key -> 'a -> 'b
    method union_with  : ('a -> 'a -> 'a) -> 'b -> 'b
    method intersection_with : ('a -> 'a -> 'a) -> 'b -> 'b

  end

val o_injection : key -> 'a -> 'a o_t
val o_map : (key * 'a -> key * 'b) -> 'a o_t -> 'b o_t
val o_plus : 'a o_t -> 'a o_t -> 'a o_t
val o_zero : unit -> 'a  o_t
val o_from_list : (key * 'a) list -> 'a o_t
val o_from_map : 'a t -> 'a o_t

val o_union        : 'a o_t -> 'a o_t -> 'a o_t
val o_intersection : 'a o_t -> 'a o_t -> 'a o_t
val o_difference   : 'a o_t -> 'a o_t -> 'a o_t
val o_symetric_difference : 'a o_t -> 'a o_t -> 'a o_t

val o_subset  : 'a o_t -> 'a o_t -> bool
val o_equal   : 'a o_t -> 'a o_t -> bool
val o_compare : 'a o_t -> 'a o_t -> int

val o_union_with : ('a -> 'a -> 'a) -> 'a o_t -> 'a o_t -> 'a o_t
val o_intersection_with : ('a -> 'a -> 'a) -> 'a o_t -> 'a o_t -> 'a o_t

end


module type OrderedType = sig
  type t
  val compare : t -> t -> int
end

module Make = functor (Ord : OrderedType) ->
struct
  
  type key = Ord.t
  type 'a t = E | N of 'a t * key * 'a * 'a t * int

  let makeTree = fun l k v r ->
    let h = match (l, r) with
      | (E, _) | (_, E) -> 1
      | (N (_, _, _, _, m), N (_, _, _, _, n)) -> 1 + min m n
    in
      N (l, k, v, r, h)

  let rotate_L = fun l k v r ->
    match r with
      | N (rl, rk, rv, rr, _) -> makeTree (makeTree l k v rl) rk rv rr
      | _ -> makeTree l k v r
	  
  let rotate_R = fun l k v r ->
    match l with
      | N (ll, lk, lv, lr, _) -> makeTree ll lk lv (makeTree lr k v r)
      | _ -> makeTree l k v r
	  
  let rotate_RL = fun l k v r ->
    match r with
      | N (N (rll, rlk, rlv, rlr, _), rk, rv, rr, _) -> 
	  makeTree (makeTree l k v rll) rlk rlv (makeTree rlr rk rv rr)
      | _ -> makeTree l k v r
	  
  let rotate_LR = fun l k v r ->
    match l with
      | N (ll, lk, lv, N (lrl, lrk, lrv, lrr, _), _) ->
	  makeTree (makeTree ll lk lv lrl) lrk lrv (makeTree lrr k v r)
      | _ -> makeTree l k v r
	  
  let height = function
    | E -> 0
    | N (_, _, _, _, n) -> n
	
  let rec makeBTree = fun l k v r ->
    let hl = height l and hr = height r in
      match compare hl hr with
	| n when n < 0 ->
	    (match r with
	     | E -> makeTree l k v r
	     | N (rl, _, _, rr, _) ->
		 if height rr = hr then rotate_L l k v r
		 else if height rl = hr then rotate_RL l k v r
		 else makeTree l k v r
	    )
	| n when n > 0 ->
	    (match l with
	       | E -> makeTree l k v r
	       | N (ll, _, _, lr, _) ->
		   if height ll = hl then rotate_R l k v r
		   else if height lr = hl then rotate_LR l k v r
		   else makeTree l k v r
	    )
	| _ -> makeTree l k v r
	    
  let rec makeSTree = fun l k v r ->
    let hl = height l and hr = height r in
      if hl > hr + 1 then
	(match l with
	   | E -> makeTree l k v r
	   | N (ll, lk, lv, lr, _) -> makeBTree ll lk lv (makeSTree lr k v r)
	)
      else if hr > hl + 1 then
	(match r with
	   | E -> makeTree l k v r
	   | N (rl, rk, rv, rr, _) -> makeBTree (makeSTree l k v rl) rk rv rr
	)
      else makeBTree l k v r

  let rec append = fun a b ->
    match (a, b) with
      | (E, _) -> b
      | (_, E) -> a
      | (N (l1, k1, v1, r1, _), N (l2, k2, v2, r2, _)) ->
          makeBTree l1 k1 v1 (makeBTree (append r1 l2) k2 v2 r2)

  let rec appendS = fun a b ->
    match (a, b) with
      | (E, _) -> b
      | (_, E) -> a
      | (N (l1, k1, v1, r1, _), N (l2, k2, v2, r2, _)) ->
          makeSTree l1 k1 v1 (makeSTree (appendS r1 l2) k2 v2 r2)

  let injection = fun k v -> makeTree E k v E

  let rec insert_raise k v = function
    | E -> injection k v
    | N (l, k', v', r, _) ->
	match Ord.compare k k' with
	  | n when n < 0 -> makeBTree (insert_raise k v l) k' v' r
	  | n when n > 0 -> makeBTree l k' v' (insert_raise k v r)
	  | _ when v <> v' -> makeTree l k v r
	  | _ -> raise Found

  let rec insert k v = function tree ->
    try 
      insert_raise k v tree 
    with Found -> tree

  let rec from_list_rec result = function
    | [] -> result
    | (value, key) :: tail -> from_list_rec (insert value key result) tail

  let from_list = function list -> from_list_rec E list

  let rec to_list_rec result = function
    | E -> result
    | N (a, k, v, b, _) -> to_list_rec ((k, v) :: to_list_rec result b) a

  let to_list = function map -> to_list_rec [] map

  let map f = function set -> from_list (List.map f (to_list set))

  let rec filter_lt x = function
    | E -> E
    | N (l, k, v, r, _) ->
	match Ord.compare x k with
	  | n when n < 0 -> filter_lt x l
	  | n when n > 0 -> makeSTree l k v (filter_lt x r)
	  | _ -> l

  let rec filter_gt x = function
    | E -> E
    | N (l, k, v, r, _) ->
	match Ord.compare x k with
	  | n when n < 0 -> makeSTree (filter_gt x l) k v r
	  | n when n > 0 -> filter_gt x r
	  | _ -> r

  let rec old_plus = fun a b ->
    match (a, b) with
      | (E, _) -> b
      | (_, E) -> a
      | (N (l, k, v, r, _), N _) -> 
	  let 
	    lt_k = filter_lt k b and 
	    gt_k = filter_gt k b 
	  in
	    makeSTree (old_plus l lt_k) k v (old_plus r gt_k)

  let rec cut = fun lo tree hi ->
    match tree with
      | N (_, k, _, r, _) when Ord.compare k lo <= 0 -> cut lo r hi
      | N (l, k, _, _, _) when Ord.compare k hi >= 0 -> cut lo l hi
      | _ -> (lo, tree, hi)

  let rec cut_lo = fun tree hi ->
    match tree with
      | N (l, k, _, _, _) when Ord.compare k hi >= 0 -> cut_lo l hi
      | _ -> (tree, hi)

  let rec cut_hi = fun lo tree ->
    match tree with
      | N (_, k, _, r, _) when Ord.compare k lo <= 0 -> cut_hi lo r
      | _ -> (lo, tree)

  let rec uni = fun a (lo, b, hi) ->
    match (a, b) with
      | (_, E) -> a
      | (E, N (l, k, v, r, _)) -> 
	  makeSTree (filter_gt lo l) k v (filter_lt hi r)
      | (N (l, k, v, r, _), N _) ->
	  makeSTree (uni l (cut lo b k)) k v (uni r (cut k b hi))  

  let rec uni_lo = fun a (b, hi) ->
    match (a, b) with
      | (_, E) -> a
      | (E, N (l, k, v, r, _)) -> makeSTree l k v (filter_lt hi r)
      | (N (l, k, v, r, _), N _) ->
	  makeSTree (uni_lo l (cut_lo b k)) k v (uni r (cut k b hi))

  let rec uni_hi = fun a (lo, b) ->
    match (a, b) with
      | (_, E) -> a
      | (E, N (l, k, v, r, _)) -> makeSTree (filter_gt lo l) k v r
      | (N (l, k, v, r, _), N _) -> 
	  makeSTree (uni l (cut lo b k)) k v (uni_hi r (cut_hi k b)) 

  let plus = fun a b ->
    match (a, b) with
      | (E, _) -> b
      | (_, E) -> a
      | (N (l, k, v, r, _), N _) ->
	  makeSTree (uni_lo l (cut_lo b k)) k v (uni_hi r (cut_hi k b)) 

  let zero = function () -> E

  let rec partition_inf_sup pivot = function
    | E -> (E, None, E)
    | N (l, k, v, r, _) as node ->
	match Ord.compare pivot k with
	  | n when n < 0 ->
	      let (inf, value, sup) = partition_inf_sup pivot l in
		(inf, value, makeSTree sup k v r)
	  | n when n > 0 ->
	      let (inf, value, sup) = partition_inf_sup pivot r in
		(makeSTree l k v inf, value, sup)
	  | _ -> (l, Some (k, v), r)

  let union = plus

  let rec intersection = fun a b ->
    match (a, b) with
      | (E, _) | (_, E) -> E
      | (N (l, k, v, r, _), N _) ->
	  match partition_inf_sup k b with
	    | (l', None, r') -> appendS (intersection l l') (intersection r r')
	    | (l', Some _, r') -> 
		makeSTree (intersection l l') k v (intersection r r')

  let rec difference = fun a b ->
    match (a, b) with
      | (E, _) -> E
      | (_, E) -> a
      | (N (l, k, v, r, _), N _) -> 
	  match partition_inf_sup k b with
	    | (l', None, r') -> 
		makeSTree (difference l l') k v (difference r r')
	    | (l', Some _, r') -> appendS (difference l l') (difference r r')

  let rec symetric_difference = fun a b ->
    match (a, b) with
      | (E, s) | (s, E) -> s
      | (N (l, k, v, r, _), N _) ->
	  match partition_inf_sup k b with
	    | (l', None, r') -> makeSTree (symetric_difference l l') k v
		(symetric_difference r r')
	    | (l', Some _, r') -> 
		appendS (symetric_difference l l') (symetric_difference r r')

  let rec subset = fun a b ->
    match (a, b) with
      | (E, _) -> true
      | (_, E) -> false
      | (N (l1, k1, v1, r1, _), N (l2, k2, v2, r2, _)) ->
          match Ord.compare k1 k2 with
	    | n when n < 0 -> subset (makeBTree l1 k1 v1 E) l2 && subset r1 b
	    | n when n > 0 -> subset (makeBTree E k1 v1 r1) r2 && subset l1 b
	    | _ -> v1 = v2 && subset l1 l2 && subset r1 r1 
  let rec compare_rec = fun a b ->
    match (a, b) with
      | ([], []) -> 0
      | ([], _) -> -1
      | (_, []) -> 1
      | (E :: t1, E :: t2) -> compare_rec t1 t2
      | (N (E, v1, k1, r1, _) :: t1, N (E, v2, k2, r2, _) :: t2) ->
	  (match Ord.compare v1 v2 with
	     | 0 when k1 = k2 -> compare_rec (r1 :: t1) (r2 :: t2)
	     | other -> other
	  )
      | (N (a, v1, k1, r1, _) :: t1, t2) ->
          compare_rec (a :: N (E, v1, k1, r1, 0) :: t1) t2
      | (t1, N (b, v2, k2, r2, _) :: t2) ->
	  compare_rec t1 (b :: N (E, v2, k2, r2, 0) :: t2)

  let equal = fun a b -> compare_rec [a] [b] = 0

  let rec insert_min_bound b k v = function
    | E -> 
	if Ord.compare b k = 0 then raise Found
	else injection k v
    | N (l, k', v', r, _) ->
	match Ord.compare k k' with
	  | n when n < 0 -> makeBTree (insert_min_bound b k v r) k' v' l
	  | _ -> makeBTree l k' v' (insert_min_bound k' k v r)

  let rec insert_raise' k v = function
    | E -> injection k v
    | N (l, k', v', r, _) ->
	match Ord.compare k k' with
	  | n when n < 0 -> makeBTree (insert_raise' k v l) k' v' r
	  | _ -> makeBTree l k' v' (insert_min_bound k' k v r)

  let insert' k v = function tree ->
    try 
      insert_raise' k v tree 
    with Found -> tree

  let insert_list = fun list map -> plus (from_list list) map

  let rec delete_min = function
    | E -> E
    | N (E, _, _, r, _) -> r
    | N (l, k, v, r, _) -> makeBTree (delete_min l) k v r
	
  let rec delete_max = function
    | E -> E
    | N (l, _, _, E, _) -> l
    | N (l, k, v, r, _) -> makeBTree l k v (delete_max r)

  let rec delete_raise x = function
    | E -> raise Not_found
    | N (l, k, v, r, _) ->
	match Ord.compare x k with
	  | n when n < 0 -> makeBTree (delete_raise x l) k v r
	  | n when n > 0 -> makeBTree l k v (delete_raise x r)
	  | _ -> append l r

  let delete x = function map -> 
    try 
      delete_raise x map 
    with Not_found -> map

  let rec delete_list = fun list map ->
    match list with
      | [] -> map
      | head :: tail -> delete_list tail (delete head map)

  let rec filter_rec f result = function
    | E -> result
    | N (l, k, v, r, _) ->
	filter_rec f 
	(filter_rec f 
	   (if f k v then insert k v result else result) 
	   l) 
	r

  let filter f = function map -> filter_rec f E map

  let rec filter_le x = function
    | E -> E
    | N (l, k, v, r, _) ->
	match Ord.compare x k with
	  | n when n < 0 -> filter_le x l
	  | _ -> makeSTree l k v (filter_le x r)

  let rec filter_ge x = function
    | E -> E
    | N (l, k, v, r, _) ->
	match Ord.compare x k with
	  | n when n <= 0 -> makeSTree (filter_ge x l) k v r
	  | _ -> filter_ge x r

  let rec range min max = function
    | E -> E
    | N (l, k, v, r, _) ->
	if Ord.compare k min < 0 then range min max r
	else if Ord.compare k max > 0 then range min max l
	else makeSTree (filter_ge min l) k v (filter_le max r)

  let rec min_view = function
    | E -> raise Empty
    | N (E, k, v, r, _) -> ((k, v), r)
    | N (l, k, v, r, _) -> 
	let (min, left) = min_view l in (min, makeBTree left k v r)
					  
  let rec max_view = function
    | E -> raise Empty
    | N (l, k, v, E, _) -> (l, (k, v))
    | N (l, k, v, r, _) -> let (right, max) = max_view r in 
	(makeBTree l k v right, max)

  let rec partition_rec f (trueS, falseS) = function
    | E -> (trueS, falseS)
    | N (l, k, v, r, _) -> 
	if f k v
	then partition_rec f (partition_rec f (insert k v trueS, falseS) l) r
	else partition_rec f (partition_rec f (trueS, insert k v falseS) l) r

  let partition = fun f map -> partition_rec f (E, E) map

  let is_empty = function
    | E -> true
    | N _ -> false

  let rec size_rec result = function
    | E -> result
    | N (l, _, _, r, _) -> size_rec (size_rec (1 + result) r) l

  let size = function map -> size_rec 0 map

  let rec keys_rec result = function
    | E -> result
    | N (l, k, _, r, _) -> keys_rec (k :: keys_rec result r) l

  let keys = function map -> keys_rec [] map

  let rec images_rec result = function
    | E -> result
    | N (l, _, v, r, _) -> images_rec (v :: images_rec result r) l

  let images = function map -> images_rec [] map

  let rec member_min_bound b k = function
    | E -> Ord.compare b k = 0
    | N (l, k', _, r, _) ->
	match Ord.compare k k' with
	  | v when v < 0 -> member_min_bound b k l
	  | _ -> member_min_bound k' k r

  let rec member' k = function
    | E -> false
    | N (l, k', _, r, _) ->
	match Ord.compare k k' with
	  | n when n < 0 -> member' k l
	  | _ -> member_min_bound k' k r

  let rec member k = function
    | E -> false
    | N (l, k', _, r, _) ->
	match Ord.compare k k' with
	  | n when n < 0 -> member k l
	  | n when n > 0 -> member k r
	  | _ -> true

  let rec lookup k = function
    | E -> raise Not_found
    | N (l, k', v', r, _) ->
	match Ord.compare k k' with
	  | n when n < 0 -> lookup k l
	  | n when n > 0 -> lookup k r
	  | _ -> v'

  let rec look_for f = function
    | E -> raise Not_found
    | N (_, k, v, _, _) when f k v -> (k, v)
    | N (l, _, _, r, _) -> (try look_for f l with Not_found -> look_for f r)

  let choose = function
    | E -> raise Empty
    | N (_, k, v, _, _) -> (k, v)

  let rec min_element = function
    | E -> raise Not_found
    | N (E, k, v, _, _) -> (k, v)
    | N (l, _, _, _, _) -> min_element l

  let rec max_element = function
    | E -> raise Not_found
    | N (_, k, v, E, _) -> (k, v)
    | N (_, _, _, r, _) -> max_element r

  let rec next x = function
    | E -> raise Not_found
    | N (l, k, v, r, _) ->
	match Ord.compare x k with
	  | n when n < 0 -> (try next x l with Not_found -> (k, v))
	  | n when n > 0 -> next x r
	  | _ -> min_element r (* raises [Not_found] *)

  let rec previous x = function
    | E -> raise Not_found
    | N (l, k, v, r, _) ->
	match Ord.compare x k with
	  | n when n < 0 -> previous x l
	  | n when n > 0 -> (try previous x r with Not_found -> (k, v))
	  | _ -> max_element l (* raises [Not_found] *)

  let rec ceil x = function
    | E -> raise Not_found
    | N (l, k, v, r, _) ->
	match Ord.compare x k with
	  | n when n < 0 -> (try ceil x l with Not_found -> (k, v))
	  | n when n > 0 -> ceil x r
	  | _ -> (k, v)

  let rec floor x = function
    | E -> raise Not_found
    | N (l, k, v, r, _) ->
	match Ord.compare x k with
	  | n when n < 0 -> floor x l
	  | n when n > 0 -> (try floor x r with Not_found -> (k, v))
	  | _ -> (k, v)

  let rec rank x = function
    | E -> raise Not_found
    | N (l, k, _, r, _) ->
	match Ord.compare x k with
	  | n when n < 0 -> rank x l
	  | n when n > 0 -> 1 + size l + rank x r
	  | _ -> size l

  let rec get n = function
    | E -> raise (Out_of_bounds n)
    | N (l, k, v, r, _) ->
	let s = size l in
	  if n < s then get n l
	  else if n > s then get (n - s - 1) r
	  else (k, v)

  let rec for_all f = function
    | E -> true
    | N (l, k, v, r, _) -> f k v && for_all f l && for_all f r

  let rec exists f = function
    | E -> false
    | N (l, k, v, r, _) -> f k v || exists f l || exists f r

  let rec iter f = function
    | E -> ()
    | N (l, k, v, r, _) -> iter f l ; f k v ; iter f r

  let rec fold_left f e = function
    | E -> e
    | N (l, k, v, r, _) -> fold_left f (f (fold_left f e l) k v) r

  let rec fold_right = fun f set e ->
    match set with
      | E -> e 
      | N (l, k, v, r, _) -> fold_right f l (f k v (fold_right f r e))

  let fold = fold_right

  let rec insert_with f k v = function
    | E -> injection k v
    | N (l, k', v', r, _) ->
	match Ord.compare k k' with
	  | n when n < 0 -> makeBTree (insert_with f k v l) k' v' r
	  | n when n > 0 -> makeBTree l k' v' (insert_with f k v r)
	  | _ -> makeBTree l k (f v v') r

  let rec union_with f = fun a b ->
    match (a, b) with
      | (E, _) -> a
      | (_, E) -> b
      | (N (l, k, v, r, _), N _) ->
	  match partition_inf_sup k b with
	    | (l', None, r') -> 
		makeSTree (union_with f l l') k v (union_with f r r')
	    | (l', Some (_, v'), r') ->
		makeSTree (union_with f l l') k (f v v') (union_with f r r')

  let rec intersection_with f = fun a b ->
    match (a, b) with
      | (E, _) | (_, E) -> E
      | (N (l, k, v, r, _), N _) ->
	  match partition_inf_sup k b with
	    | (l', None, r') -> 
		appendS (intersection_with f l l') (intersection_with f r r')
	    | (l', Some (_, v'), r') ->
		makeSTree (intersection_with f l l') k (f v v')
		(intersection_with f r r')

  let rec map_images_with_index f = function 
    | E -> E
    | N (l, k, v, r, h) -> 
	N (map_images_with_index f l, k, f k v, map_images_with_index f r, h)

  let rec map_images f = function
    | E -> E
    | N (l, k, v, r, h) -> N (map_images f l, k, f v, map_images f r, h)

  let rec unsafe_insert_min k v = function
    | E -> injection k v
    | N (l, k', v', r, _) -> makeTree (unsafe_insert_min k v l) k' v' r 

  let rec unsafe_insert_max k v = function
    | E -> injection k v
    | N (l, k', v', r, _) -> makeTree l k' v' (unsafe_insert_max k v r)

  let empty = E
  let mem = member
  let add = insert
  let singleton = injection
  let remove = delete
  let inter = intersection
  let diff = difference
  let cardinal = size
  let elements = to_list
  let min_elt = min_element    
  let max_elt = max_element
  let find = lookup
  let mapi = map_images_with_index

  let std_map = map_images

  class ['a] o_t = fun initial_map ->

  object (self : 'b)

    val contents  = initial_map

    method to_map = contents
		      
    method map f = {< contents = map f contents >}

    method plus = function (set : 'b) -> 
      {< contents = plus contents set#to_map >} 

    method to_list = to_list contents

    method union = function (set : 'b) -> 
      {< contents = plus contents set#to_map >}
      
    method intersection = function (set : 'b) -> 
      {< contents = intersection contents set#to_map >}
      
    method difference = function (set : 'b) -> 
      {< contents = difference contents set#to_map >}
      
    method symetric_difference = function (set : 'b) -> 
      {< contents = symetric_difference contents set#to_map >}
      
    method subset = function (set : 'b) -> subset contents set#to_map
    method equal = function (set : 'b)-> equal contents set#to_map
    method compare = function (set : 'b) -> compare_rec [contents] [set#to_map]
      
    method insert = fun k v -> 
      try 
	{< contents = insert_raise k v contents >} 
      with Found -> self

    method insert_raise = fun k v -> {< contents = insert_raise k v contents >}

    method insert' = fun k v -> 
      try
	{< contents = insert_raise' k v contents >}
      with Found -> self

    method insert_raise' = fun k v -> 
      {< contents = insert_raise' k v contents >}
      
    method insert_list = function list -> 
      {< contents = insert_list list contents >}

    method delete = function k -> 
      try 
	{< contents = delete_raise k contents >} 
      with Not_found -> self

    method delete_raise = function k -> 
      {< contents = delete_raise k contents >}

    method delete_list = function list -> 
      {< contents = delete_list list contents >}
      
    method delete_min = {< contents = delete_min contents >}
    method delete_max = {< contents = delete_max contents >}

    method filter = function f -> {< contents = filter f contents >}
    method filter_lt = function x -> {< contents = filter_lt x contents >}
    method filter_gt = function x -> {< contents = filter_gt x contents >}
    method filter_le = function x -> {< contents = filter_le x contents >}
    method filter_ge = function x -> {< contents = filter_ge x contents >}

    method range = fun min max -> {< contents = range min max contents >}

    method min_view =
      let (min, other) = min_view contents in
	(min, {< contents = other >})

    method max_view =
      let (other, max) = max_view contents in
	({< contents = other >}, max)

    method partition = function f -> 
      let (trueSet, falseSet) = partition f contents in
	({< contents = trueSet >}, {< contents = falseSet >})

    method partition_inf_sup = function x ->
      let (inf, value, sup) = partition_inf_sup x contents in
	({< contents = inf >}, value, {< contents = sup >})

    method is_empty = contents = E
    method size = size contents

    method keys = keys contents
    method images = images contents

    method member = function x -> member x contents
    method member' = function x -> member' x contents

    method lookup = function x -> lookup x contents
    method look_for = function f -> look_for f contents

    method choose =
      match contents with
	| E -> raise Not_found
	| N (_, k, v, _, _) -> (k, v)

    method min_element = min_element contents
    method max_element = max_element contents

    method previous = function x -> previous x contents
    method next = function x -> next x contents
    method floor = function x -> floor x contents
    method ceil = function x -> ceil x contents

    method rank = function x -> rank x contents
    method get = function n -> get n contents

    method exists = function f -> exists f contents
    method for_all = function f -> for_all f contents

    method iter :  (key -> 'a -> unit) -> unit = function f -> iter f contents

    method fold : 'c. (key -> 'a -> 'c -> 'c) -> 'c -> 'c = 
      fun f e -> fold_right f contents e

    method fold_left : 'c. ('c -> key -> 'a -> 'c) -> 'c -> 'c =
      fun f e -> fold_left f e contents

    method fold_right : 'c. (key -> 'a -> 'c -> 'c) -> 'c -> 'c =
      fun f e -> fold_right f contents e

    method insert_with = fun f k v -> 
      {< contents = insert_with f k v contents >}

    method union_with f = fun (map : 'b) -> 
      {< contents = union_with f contents map#to_map >}

    method intersection_with f = fun (map : 'b) ->
      {< contents = intersection_with f contents map#to_map >}

  end

  let o_injection = fun k v -> new o_t (injection k v)
  let o_map = fun f a -> new o_t (map f a#to_map)
  let o_plus = fun a b -> new o_t (plus a#to_map b#to_map)
  let o_zero = function () -> new o_t E
  let o_from_list = function list -> new o_t (from_list list)
  let o_from_map = function map -> new o_t map
    
  let o_union = fun a b -> new o_t (plus a#to_map b#to_map)
  let o_intersection = fun a b -> new o_t (intersection a#to_map b#to_map)
  let o_difference = fun a b -> new o_t (difference a#to_map b#to_map)
  let o_symetric_difference = fun a b -> 
    new o_t (symetric_difference a#to_map b#to_map)
    
  let o_subset = fun a b -> subset a#to_map b#to_map
  let o_equal = fun a b -> equal a#to_map b#to_map
  let o_compare = fun a b -> compare_rec [a#to_map] [b#to_map]

  let o_union_with f = fun a b -> new o_t (union_with f a#to_map b#to_map)

  let o_intersection_with f = fun a b -> 
    new o_t (intersection_with f a#to_map b#to_map)

  let compare = fun a b -> compare_rec [a] [b]

  let rec mn = function
    | E -> 0
    | N (a, _, _, b, _) -> 1 + min (mn a) (mn b)

  let rec mx = function
    | E -> 0
    | N (a, _, _, b, _) -> 1 + max (mx a) (mx b)

  let rec med total = function
    | E -> [total]
    | N (a, _, _, b, _) -> med (total + 1) a @ med (total + 1) b

  let stat = function tree -> 
    let height_list = med 0 tree in
    let med_h = (List.fold_left (+) 0 height_list) / 
		(List.length height_list) 
    in
      (size tree, mn tree, mx tree, med_h)

end

module type RelaxationInformation = sig
  val relaxation_coefficient : int
end

module MakeRelaxed (Inf : RelaxationInformation) (Ord : OrderedType) =
struct
  
  type key = Ord.t
  type 'a t = E | N of 'a t * key * 'a * 'a t * int

  let bound = 
    if Inf.relaxation_coefficient <= 0 
    then invalid "RelaxationInformation" Negative
    else Inf.relaxation_coefficient

  let makeTree = fun l k v r ->
    let h = match (l, r) with
      | (E, _) | (_, E) -> 1
      | (N (_, _, _, _, m), N (_, _, _, _, n)) -> 1 + min m n
    in
      N (l, k, v, r, h)

  let rotate_L = fun l k v r ->
    match r with
      | N (rl, rk, rv, rr, _) -> makeTree (makeTree l k v rl) rk rv rr
      | _ -> makeTree l k v r
	  
  let rotate_R = fun l k v r ->
    match l with
      | N (ll, lk, lv, lr, _) -> makeTree ll lk lv (makeTree lr k v r)
      | _ -> makeTree l k v r
	  
  let rotate_RL = fun l k v r ->
    match r with
      | N (N (rll, rlk, rlv, rlr, _), rk, rv, rr, _) -> 
	  makeTree (makeTree l k v rll) rlk rlv (makeTree rlr rk rv rr)
      | _ -> makeTree l k v r
	  
  let rotate_LR = fun l k v r ->
    match l with
      | N (ll, lk, lv, N (lrl, lrk, lrv, lrr, _), _) ->
	  makeTree (makeTree ll lk lv lrl) lrk lrv (makeTree lrr k v r)
      | _ -> makeTree l k v r
	  
  let height = function
    | E -> 0
    | N (_, _, _, _, n) -> n
	
  let rec makeBTree = fun l k v r ->
    let hl = height l and hr = height r in
      if hl > hr + bound then 
	(match r with
	   | E -> makeTree l k v r
	   | N (rl, _, _, rr, _) ->
	       if height rr = hr then rotate_L l k v r
	       else if height rl = hr then rotate_RL l k v r
	       else makeTree l k v r
	)
      else if hr > hl + bound then
	(match l with
	   | E -> makeTree l k v r
	   | N (ll, _, _, lr, _) ->
	       if height ll = hl then rotate_R l k v r
	       else if height lr = hl then rotate_LR l k v r
	       else makeTree l k v r
	)
      else makeTree l k v r
	    
  let rec makeSTree = fun l k v r ->
    let hl = height l and hr = height r in
      if hl > hr + bound + 1 then
	(match l with
	   | E -> makeTree l k v r
	   | N (ll, lk, lv, lr, _) -> makeBTree ll lk lv (makeSTree lr k v r)
	)
      else if hr > hl + bound + 1 then
	(match r with
	   | E -> makeTree l k v r
	   | N (rl, rk, rv, rr, _) -> makeBTree (makeSTree l k v rl) rk rv rr
	)
      else makeBTree l k v r

  let rec append = fun a b ->
    match (a, b) with
      | (E, _) -> b
      | (_, E) -> a
      | (N (l1, k1, v1, r1, _), N (l2, k2, v2, r2, _)) ->
          makeBTree l1 k1 v1 (makeBTree (append r1 l2) k2 v2 r2)

  let rec appendS = fun a b ->
    match (a, b) with
      | (E, _) -> b
      | (_, E) -> a
      | (N (l1, k1, v1, r1, _), N (l2, k2, v2, r2, _)) ->
          makeSTree l1 k1 v1 (makeSTree (appendS r1 l2) k2 v2 r2)

  let injection = fun k v -> makeTree E k v E

  let rec insert_raise k v = function
    | E -> injection k v
    | N (l, k', v', r, _) ->
	match Ord.compare k k' with
	  | n when n < 0 -> makeBTree (insert_raise k v l) k' v' r
	  | n when n > 0 -> makeBTree l k' v' (insert_raise k v r)
	  | _ when v <> v' -> makeTree l k v r
	  | _ -> raise Found

  let rec insert k v = function tree ->
    try 
      insert_raise k v tree 
    with Found -> tree

  let rec from_list_rec result = function
    | [] -> result
    | (value, key) :: tail -> from_list_rec (insert value key result) tail

  let from_list = function list -> from_list_rec E list

  let rec to_list_rec result = function
    | E -> result
    | N (a, k, v, b, _) -> to_list_rec ((k, v) :: to_list_rec result b) a

  let to_list = function map -> to_list_rec [] map

  let map f = function set -> from_list (List.map f (to_list set))

  let rec filter_lt x = function
    | E -> E
    | N (l, k, v, r, _) ->
	match Ord.compare x k with
	  | n when n < 0 -> filter_lt x l
	  | n when n > 0 -> makeSTree l k v (filter_lt x r)
	  | _ -> l

  let rec filter_gt x = function
    | E -> E
    | N (l, k, v, r, _) ->
	match Ord.compare x k with
	  | n when n < 0 -> makeSTree (filter_gt x l) k v r
	  | n when n > 0 -> filter_gt x r
	  | _ -> r

  let rec old_plus = fun a b ->
    match (a, b) with
      | (E, _) -> b
      | (_, E) -> a
      | (N (l, k, v, r, _), N _) -> 
	  let 
	    lt_k = filter_lt k b and 
	    gt_k = filter_gt k b 
	  in
	    makeSTree (old_plus l lt_k) k v (old_plus r gt_k)

  let rec cut = fun lo tree hi ->
    match tree with
      | N (_, k, _, r, _) when Ord.compare k lo <= 0 -> cut lo r hi
      | N (l, k, _, _, _) when Ord.compare k hi >= 0 -> cut lo l hi
      | _ -> (lo, tree, hi)

  let rec cut_lo = fun tree hi ->
    match tree with
      | N (l, k, _, _, _) when Ord.compare k hi >= 0 -> cut_lo l hi
      | _ -> (tree, hi)

  let rec cut_hi = fun lo tree ->
    match tree with
      | N (_, k, _, r, _) when Ord.compare k lo <= 0 -> cut_hi lo r
      | _ -> (lo, tree)

  let rec uni = fun a (lo, b, hi) ->
    match (a, b) with
      | (_, E) -> a
      | (E, N (l, k, v, r, _)) -> 
	  makeSTree (filter_gt lo l) k v (filter_lt hi r)
      | (N (l, k, v, r, _), N _) ->
	  makeSTree (uni l (cut lo b k)) k v (uni r (cut k b hi))  

  let rec uni_lo = fun a (b, hi) ->
    match (a, b) with
      | (_, E) -> a
      | (E, N (l, k, v, r, _)) -> makeSTree l k v (filter_lt hi r)
      | (N (l, k, v, r, _), N _) ->
	  makeSTree (uni_lo l (cut_lo b k)) k v (uni r (cut k b hi))

  let rec uni_hi = fun a (lo, b) ->
    match (a, b) with
      | (_, E) -> a
      | (E, N (l, k, v, r, _)) -> makeSTree (filter_gt lo l) k v r
      | (N (l, k, v, r, _), N _) -> 
	  makeSTree (uni l (cut lo b k)) k v (uni_hi r (cut_hi k b)) 

  let plus = fun a b ->
    match (a, b) with
      | (E, _) -> b
      | (_, E) -> a
      | (N (l, k, v, r, _), N _) ->
	  makeSTree (uni_lo l (cut_lo b k)) k v (uni_hi r (cut_hi k b)) 

  let zero = function () -> E

  let rec partition_inf_sup pivot = function
    | E -> (E, None, E)
    | N (l, k, v, r, _) as node ->
	match Ord.compare pivot k with
	  | n when n < 0 ->
	      let (inf, value, sup) = partition_inf_sup pivot l in
		(inf, value, makeSTree sup k v r)
	  | n when n > 0 ->
	      let (inf, value, sup) = partition_inf_sup pivot r in
		(makeSTree l k v inf, value, sup)
	  | _ -> (l, Some (k, v), r)

  let union = plus

  let rec intersection = fun a b ->
    match (a, b) with
      | (E, _) | (_, E) -> E
      | (N (l, k, v, r, _), N _) ->
	  match partition_inf_sup k b with
	    | (l', None, r') -> appendS (intersection l l') (intersection r r')
	    | (l', Some _, r') -> 
		makeSTree (intersection l l') k v (intersection r r')

  let rec difference = fun a b ->
    match (a, b) with
      | (E, _) -> E
      | (_, E) -> a
      | (N (l, k, v, r, _), N _) -> 
	  match partition_inf_sup k b with
	    | (l', None, r') -> 
		makeSTree (difference l l') k v (difference r r')
	    | (l', Some _, r') -> appendS (difference l l') (difference r r')

  let rec symetric_difference = fun a b ->
    match (a, b) with
      | (E, s) | (s, E) -> s
      | (N (l, k, v, r, _), N _) ->
	  match partition_inf_sup k b with
	    | (l', None, r') -> makeSTree (symetric_difference l l') k v
		(symetric_difference r r')
	    | (l', Some _, r') -> 
		appendS (symetric_difference l l') (symetric_difference r r')

  let rec subset = fun a b ->
    match (a, b) with
      | (E, _) -> true
      | (_, E) -> false
      | (N (l1, k1, v1, r1, _), N (l2, k2, v2, r2, _)) ->
          match Ord.compare k1 k2 with
	    | n when n < 0 -> subset (makeBTree l1 k1 v1 E) l2 && subset r1 b
	    | n when n > 0 -> subset (makeBTree E k1 v1 r1) r2 && subset l1 b
	    | _ -> v1 = v2 && subset l1 l2 && subset r1 r1 
  let rec compare_rec = fun a b ->
    match (a, b) with
      | ([], []) -> 0
      | ([], _) -> -1
      | (_, []) -> 1
      | (E :: t1, E :: t2) -> compare_rec t1 t2
      | (N (E, v1, k1, r1, _) :: t1, N (E, v2, k2, r2, _) :: t2) ->
	  (match Ord.compare v1 v2 with
	     | 0 when k1 = k2 -> compare_rec (r1 :: t1) (r2 :: t2)
	     | other -> other
	  )
      | (N (a, v1, k1, r1, _) :: t1, t2) ->
          compare_rec (a :: N (E, v1, k1, r1, 0) :: t1) t2
      | (t1, N (b, v2, k2, r2, _) :: t2) ->
	  compare_rec t1 (b :: N (E, v2, k2, r2, 0) :: t2)

  let equal = fun a b -> compare_rec [a] [b] = 0

  let rec insert_min_bound b k v = function
    | E -> 
	if Ord.compare b k = 0 then raise Found
	else injection k v
    | N (l, k', v', r, _) ->
	match Ord.compare k k' with
	  | n when n < 0 -> makeBTree (insert_min_bound b k v r) k' v' l
	  | _ -> makeBTree l k' v' (insert_min_bound k' k v r)

  let rec insert_raise' k v = function
    | E -> injection k v
    | N (l, k', v', r, _) ->
	match Ord.compare k k' with
	  | n when n < 0 -> makeBTree (insert_raise' k v l) k' v' r
	  | _ -> makeBTree l k' v' (insert_min_bound k' k v r)

  let insert' k v = function tree ->
    try 
      insert_raise' k v tree 
    with Found -> tree

  let insert_list = fun list map -> plus (from_list list) map

  let rec delete_min = function
    | E -> E
    | N (E, _, _, r, _) -> r
    | N (l, k, v, r, _) -> makeBTree (delete_min l) k v r
	
  let rec delete_max = function
    | E -> E
    | N (l, _, _, E, _) -> l
    | N (l, k, v, r, _) -> makeBTree l k v (delete_max r)

  let rec delete_raise x = function
    | E -> raise Not_found
    | N (l, k, v, r, _) ->
	match Ord.compare x k with
	  | n when n < 0 -> makeBTree (delete_raise x l) k v r
	  | n when n > 0 -> makeBTree l k v (delete_raise x r)
	  | _ -> append l r

  let delete x = function map -> 
    try 
      delete_raise x map 
    with Not_found -> map

  let rec delete_list = fun list map ->
    match list with
      | [] -> map
      | head :: tail -> delete_list tail (delete head map)

  let rec filter_rec f result = function
    | E -> result
    | N (l, k, v, r, _) ->
	filter_rec f 
	(filter_rec f 
	   (if f k v then insert k v result else result) 
	   l) 
	r

  let filter f = function map -> filter_rec f E map

  let rec filter_le x = function
    | E -> E
    | N (l, k, v, r, _) ->
	match Ord.compare x k with
	  | n when n < 0 -> filter_le x l
	  | _ -> makeSTree l k v (filter_le x r)

  let rec filter_ge x = function
    | E -> E
    | N (l, k, v, r, _) ->
	match Ord.compare x k with
	  | n when n <= 0 -> makeSTree (filter_ge x l) k v r
	  | _ -> filter_ge x r

  let rec range min max = function
    | E -> E
    | N (l, k, v, r, _) ->
	if Ord.compare k min < 0 then range min max r
	else if Ord.compare k max > 0 then range min max l
	else makeSTree (filter_ge min l) k v (filter_le max r)

  let rec min_view = function
    | E -> raise Empty
    | N (E, k, v, r, _) -> ((k, v), r)
    | N (l, k, v, r, _) -> 
	let (min, left) = min_view l in (min, makeBTree left k v r)
					  
  let rec max_view = function
    | E -> raise Empty
    | N (l, k, v, E, _) -> (l, (k, v))
    | N (l, k, v, r, _) -> let (right, max) = max_view r in 
	(makeBTree l k v right, max)

  let rec partition_rec f (trueS, falseS) = function
    | E -> (trueS, falseS)
    | N (l, k, v, r, _) -> 
	if f k v
	then partition_rec f (partition_rec f (insert k v trueS, falseS) l) r
	else partition_rec f (partition_rec f (trueS, insert k v falseS) l) r

  let partition = fun f map -> partition_rec f (E, E) map

  let is_empty = function
    | E -> true
    | N _ -> false

  let rec size_rec result = function
    | E -> result
    | N (l, _, _, r, _) -> size_rec (size_rec (1 + result) r) l

  let size = function map -> size_rec 0 map

  let rec keys_rec result = function
    | E -> result
    | N (l, k, _, r, _) -> keys_rec (k :: keys_rec result r) l

  let keys = function map -> keys_rec [] map

  let rec images_rec result = function
    | E -> result
    | N (l, _, v, r, _) -> images_rec (v :: images_rec result r) l

  let images = function map -> images_rec [] map

  let rec member_min_bound b k = function
    | E -> Ord.compare b k = 0
    | N (l, k', _, r, _) ->
	match Ord.compare k k' with
	  | v when v < 0 -> member_min_bound b k l
	  | _ -> member_min_bound k' k r

  let rec member' k = function
    | E -> false
    | N (l, k', _, r, _) ->
	match Ord.compare k k' with
	  | n when n < 0 -> member' k l
	  | _ -> member_min_bound k' k r

  let rec member k = function
    | E -> false
    | N (l, k', _, r, _) ->
	match Ord.compare k k' with
	  | n when n < 0 -> member k l
	  | n when n > 0 -> member k r
	  | _ -> true

  let rec lookup k = function
    | E -> raise Not_found
    | N (l, k', v', r, _) ->
	match Ord.compare k k' with
	  | n when n < 0 -> lookup k l
	  | n when n > 0 -> lookup k r
	  | _ -> v'

  let rec look_for f = function
    | E -> raise Not_found
    | N (_, k, v, _, _) when f k v -> (k, v)
    | N (l, _, _, r, _) -> (try look_for f l with Not_found -> look_for f r)

  let choose = function
    | E -> raise Empty
    | N (_, k, v, _, _) -> (k, v)

  let rec min_element = function
    | E -> raise Not_found
    | N (E, k, v, _, _) -> (k, v)
    | N (l, _, _, _, _) -> min_element l

  let rec max_element = function
    | E -> raise Not_found
    | N (_, k, v, E, _) -> (k, v)
    | N (_, _, _, r, _) -> max_element r

  let rec next x = function
    | E -> raise Not_found
    | N (l, k, v, r, _) ->
	match Ord.compare x k with
	  | n when n < 0 -> (try next x l with Not_found -> (k, v))
	  | n when n > 0 -> next x r
	  | _ -> min_element r (* raises [Not_found] *)

  let rec previous x = function
    | E -> raise Not_found
    | N (l, k, v, r, _) ->
	match Ord.compare x k with
	  | n when n < 0 -> previous x l
	  | n when n > 0 -> (try previous x r with Not_found -> (k, v))
	  | _ -> max_element l (* raises [Not_found] *)

  let rec ceil x = function
    | E -> raise Not_found
    | N (l, k, v, r, _) ->
	match Ord.compare x k with
	  | n when n < 0 -> (try ceil x l with Not_found -> (k, v))
	  | n when n > 0 -> ceil x r
	  | _ -> (k, v)

  let rec floor x = function
    | E -> raise Not_found
    | N (l, k, v, r, _) ->
	match Ord.compare x k with
	  | n when n < 0 -> floor x l
	  | n when n > 0 -> (try floor x r with Not_found -> (k, v))
	  | _ -> (k, v)

  let rec rank x = function
    | E -> raise Not_found
    | N (l, k, _, r, _) ->
	match Ord.compare x k with
	  | n when n < 0 -> rank x l
	  | n when n > 0 -> 1 + size l + rank x r
	  | _ -> size l

  let rec get n = function
    | E -> raise (Out_of_bounds n)
    | N (l, k, v, r, _) ->
	let s = size l in
	  if n < s then get n l
	  else if n > s then get (n - s - 1) r
	  else (k, v)

  let rec for_all f = function
    | E -> true
    | N (l, k, v, r, _) -> f k v && for_all f l && for_all f r

  let rec exists f = function
    | E -> false
    | N (l, k, v, r, _) -> f k v || exists f l || exists f r

  let rec iter f = function
    | E -> ()
    | N (l, k, v, r, _) -> iter f l ; f k v ; iter f r

  let rec fold_left f e = function
    | E -> e
    | N (l, k, v, r, _) -> fold_left f (f (fold_left f e l) k v) r

  let rec fold_right = fun f set e ->
    match set with
      | E -> e 
      | N (l, k, v, r, _) -> fold_right f l (f k v (fold_right f r e))

  let fold = fold_right

  let rec insert_with f k v = function
    | E -> injection k v
    | N (l, k', v', r, _) ->
	match Ord.compare k k' with
	  | n when n < 0 -> makeBTree (insert_with f k v l) k' v' r
	  | n when n > 0 -> makeBTree l k' v' (insert_with f k v r)
	  | _ -> makeBTree l k (f v v') r

  let rec union_with f = fun a b ->
    match (a, b) with
      | (E, _) -> a
      | (_, E) -> b
      | (N (l, k, v, r, _), N _) ->
	  match partition_inf_sup k b with
	    | (l', None, r') -> 
		makeSTree (union_with f l l') k v (union_with f r r')
	    | (l', Some (_, v'), r') ->
		makeSTree (union_with f l l') k (f v v') (union_with f r r')

  let rec intersection_with f = fun a b ->
    match (a, b) with
      | (E, _) | (_, E) -> E
      | (N (l, k, v, r, _), N _) ->
	  match partition_inf_sup k b with
	    | (l', None, r') -> 
		appendS (intersection_with f l l') (intersection_with f r r')
	    | (l', Some (_, v'), r') ->
		makeSTree (intersection_with f l l') k (f v v')
		(intersection_with f r r')

  let rec map_images_with_index f = function 
    | E -> E
    | N (l, k, v, r, h) -> 
	N (map_images_with_index f l, k, f k v, map_images_with_index f r, h)

  let rec map_images f = function
    | E -> E
    | N (l, k, v, r, h) -> N (map_images f l, k, f v, map_images f r, h)

  let rec unsafe_insert_min k v = function
    | E -> injection k v
    | N (l, k', v', r, _) -> makeTree (unsafe_insert_min k v l) k' v' r 

  let rec unsafe_insert_max k v = function
    | E -> injection k v
    | N (l, k', v', r, _) -> makeTree l k' v' (unsafe_insert_max k v r)

  let empty = E
  let mem = member
  let add = insert
  let singleton = injection
  let remove = delete
  let inter = intersection
  let diff = difference
  let cardinal = size
  let elements = to_list
  let min_elt = min_element    
  let max_elt = max_element
  let find = lookup
  let mapi = map_images_with_index

  let std_map = map_images

  class ['a] o_t = fun initial_map ->

  object (self : 'b)

    val contents  = initial_map

    method to_map = contents
		      
    method map f = {< contents = map f contents >}

    method plus = function (set : 'b) -> 
      {< contents = plus contents set#to_map >} 

    method to_list = to_list contents

    method union = function (set : 'b) -> 
      {< contents = plus contents set#to_map >}
      
    method intersection = function (set : 'b) -> 
      {< contents = intersection contents set#to_map >}
      
    method difference = function (set : 'b) -> 
      {< contents = difference contents set#to_map >}
      
    method symetric_difference = function (set : 'b) -> 
      {< contents = symetric_difference contents set#to_map >}
      
    method subset = function (set : 'b) -> subset contents set#to_map
    method equal = function (set : 'b)-> equal contents set#to_map
    method compare = function (set : 'b) -> compare_rec [contents] [set#to_map]
      
    method insert = fun k v -> 
      try 
	{< contents = insert_raise k v contents >} 
      with Found -> self

    method insert_raise = fun k v -> {< contents = insert_raise k v contents >}

    method insert' = fun k v -> 
      try
	{< contents = insert_raise' k v contents >}
      with Found -> self

    method insert_raise' = fun k v -> 
      {< contents = insert_raise' k v contents >}
      
    method insert_list = function list -> 
      {< contents = insert_list list contents >}

    method delete = function k -> 
      try 
	{< contents = delete_raise k contents >} 
      with Not_found -> self

    method delete_raise = function k -> 
      {< contents = delete_raise k contents >}

    method delete_list = function list -> 
      {< contents = delete_list list contents >}
      
    method delete_min = {< contents = delete_min contents >}
    method delete_max = {< contents = delete_max contents >}

    method filter = function f -> {< contents = filter f contents >}
    method filter_lt = function x -> {< contents = filter_lt x contents >}
    method filter_gt = function x -> {< contents = filter_gt x contents >}
    method filter_le = function x -> {< contents = filter_le x contents >}
    method filter_ge = function x -> {< contents = filter_ge x contents >}

    method range = fun min max -> {< contents = range min max contents >}

    method min_view =
      let (min, other) = min_view contents in
	(min, {< contents = other >})

    method max_view =
      let (other, max) = max_view contents in
	({< contents = other >}, max)

    method partition = function f -> 
      let (trueSet, falseSet) = partition f contents in
	({< contents = trueSet >}, {< contents = falseSet >})

    method partition_inf_sup = function x ->
      let (inf, value, sup) = partition_inf_sup x contents in
	({< contents = inf >}, value, {< contents = sup >})

    method is_empty = contents = E
    method size = size contents

    method keys = keys contents
    method images = images contents

    method member = function x -> member x contents
    method member' = function x -> member' x contents

    method lookup = function x -> lookup x contents
    method look_for = function f -> look_for f contents

    method choose =
      match contents with
	| E -> raise Not_found
	| N (_, k, v, _, _) -> (k, v)

    method min_element = min_element contents
    method max_element = max_element contents

    method previous = function x -> previous x contents
    method next = function x -> next x contents
    method floor = function x -> floor x contents
    method ceil = function x -> ceil x contents

    method rank = function x -> rank x contents
    method get = function n -> get n contents

    method exists = function f -> exists f contents
    method for_all = function f -> for_all f contents

    method iter :  (key -> 'a -> unit) -> unit = function f -> iter f contents

    method fold : 'c. (key -> 'a -> 'c -> 'c) -> 'c -> 'c = 
      fun f e -> fold_right f contents e

    method fold_left : 'c. ('c -> key -> 'a -> 'c) -> 'c -> 'c =
      fun f e -> fold_left f e contents

    method fold_right : 'c. (key -> 'a -> 'c -> 'c) -> 'c -> 'c =
      fun f e -> fold_right f contents e

    method insert_with = fun f k v -> 
      {< contents = insert_with f k v contents >}

    method union_with f = fun (map : 'b) -> 
      {< contents = union_with f contents map#to_map >}

    method intersection_with f = fun (map : 'b) ->
      {< contents = intersection_with f contents map#to_map >}

  end

  let o_injection = fun k v -> new o_t (injection k v)
  let o_map = fun f a -> new o_t (map f a#to_map)
  let o_plus = fun a b -> new o_t (plus a#to_map b#to_map)
  let o_zero = function () -> new o_t E
  let o_from_list = function list -> new o_t (from_list list)
  let o_from_map = function map -> new o_t map
    
  let o_union = fun a b -> new o_t (plus a#to_map b#to_map)
  let o_intersection = fun a b -> new o_t (intersection a#to_map b#to_map)
  let o_difference = fun a b -> new o_t (difference a#to_map b#to_map)
  let o_symetric_difference = fun a b -> 
    new o_t (symetric_difference a#to_map b#to_map)
    
  let o_subset = fun a b -> subset a#to_map b#to_map
  let o_equal = fun a b -> equal a#to_map b#to_map
  let o_compare = fun a b -> compare_rec [a#to_map] [b#to_map]

  let o_union_with f = fun a b -> new o_t (union_with f a#to_map b#to_map)

  let o_intersection_with f = fun a b -> 
    new o_t (intersection_with f a#to_map b#to_map)

  let compare = fun a b -> compare_rec [a] [b]

  let rec mn = function
    | E -> 0
    | N (a, _, _, b, _) -> 1 + min (mn a) (mn b)

  let rec mx = function
    | E -> 0
    | N (a, _, _, b, _) -> 1 + max (mx a) (mx b)

  let rec med total = function
    | E -> [total]
    | N (a, _, _, b, _) -> med (total + 1) a @ med (total + 1) b

  let stat = function tree -> 
    let height_list = med 0 tree in
    let med_h = (List.fold_left (+) 0 height_list) / 
		(List.length height_list) 
    in
      (size tree, mn tree, mx tree, med_h)

end

(** Debug **)

let existe h = function
  | E -> h = 0
  | N (_, x, _, _, h') -> h = h'

let rec 
  conflit_rouge_rec h = function
    | E -> false
    | N (a, x, _, b, h') -> 
	conflit_rouge a || conflit_rouge b || 
	(h = h' && (existe h a || existe h b))
and	
  conflit_rouge = function
    | E -> false
    | N (a, _, _, b, h) -> conflit_rouge_rec h a || conflit_rouge_rec h b

let rec hauteur_noire = function
  | E -> true
  | N (a, x, _, b, h) ->
      let 
	ha = height a and 
	hb = height b 
      in
	hauteur_noire a &&
	hauteur_noire b &&
	(ha = h || ha = h - 1) && 
	(hb = h || hb = h - 1)

let rec mn = function
  | E -> 0
  | N (a, _, _, b, _) -> 1 + min (mn a) (mn b)

let rec mx = function
  | E -> 0
  | N (a, _, _, b, _) -> 1 + max (mx a) (mx b)

let rec med total = function
  | E -> [total]
  | N (a, _, _, b, _) -> med (total + 1) a @ med (total + 1) b

let stat = function tree -> 
  let height_list = med 0 tree in
  let med_h = (List.fold_left (+) 0 height_list) / 
	      (List.length height_list) 
  in
    (size tree, mn tree, mx tree, med_h)

(* To avoid interference with [Pervasives.compare] *)
(* Pour viter toute interfrence avec [Pervasives.compare] *)
(* Para evitar toda interferencia con [Pervasives.compare] *)

let compare = fun a b -> compare_rec [a] [b]
