(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Cartesian trees
 * Arbres cartsiens
 * rboles cartesianos
 *
 *)

open BairePervasives

let moduleName = "CartesianTree"

(* Plan :
 *
 * 1. Short explanation of cartesian trees
 * 2. Polymorphic code with some comments mostly on algorithms
 * 3. Short modules (height functions) and explanations of how and why
 *    balanced schemes can be obtained from cartesian tree specialization
 * 4. MakeSet functor
 * 5. Make functor (same as polymophic version)
 *)

(* Plan :
 *
 * 1. Prsentation sommaire des arbres cartsiens
 * 2. Code polymorphe comment notamment concernant les algorithmes
 * 3. Quelques modules (fonction de hauteur) et explications sur comment
 *    et pourquoi les schmas d'quilibrage usuels peuvent s'obtenir par
 *    spcialisation des arbres cartsiens
 * 4. foncteur MakeSet
 * 5. foncteur Make (semblable  la version polymorphe)
 *)

(* Plan :
 *
 * 1. Presentacin rpida de los rboles cartesianos
 * 2. Codigo polimrfico y unos cuantos comentrios sobre los algoritmos
 * 3. Modulos (funcines de altura) y explicaciones sobre cmo y porqu
 *    los esquemas de balance ms comunes se pueden obtener especializando
 *    los rboles cartesianos
 * 4. functor MakeSet
 * 5. functor Make (similar a la versin polimrfica)
 *) 

(* Cartesian trees were introduced in 1980 by Vuillemin. 
 *
 * They are the heart of the set representation in Baire library, since many 
 * binary search trees (AVL, red and black, treaps, splay, weight balanced) 
 * can be seen as particular cases of cartesian trees.
 *
 * Cartesian trees are binary trees in which every node has two keys.
 * The horizontal projection is a search tree, the vertical projection a heap
 * (see figure)
 * 
 *                               vertical projection (second key)
 * 
 *         (2, 5)                         5
 *          /  \                        /   \
 *      (1, 3) (6, 4)       ->        3       4
 *              /  \                         / \
 *          (5, 2) (7, 3)                   2    3
 * 
 *             |
 *             \/
 * 
 *    horizontal projection (first key)
 * 
 *        2
 *      /   \
 *     1     6
 *          / \
 *         5   7
 * 
 *     1 2 5 6 7
 * 
 * We will always suppose that first keys are all different.
 * The second key is not restricted, we will often call it  priority 
 * 
 * Cartesian trees were given that name because if all the couples are 
 * different, then the tree structure is unique. That is to say, every
 * couple can be seen as the cartesian coordinates of the node.
 *)

(* Les arbres cartsiens ont t introduits par Vuillemin en 1980. 
 *
 * Ils sont l'lment fondamental dans la construction des arbres binaires 
 * de Baire car de trs nombreux arbres quilibrs (AVL, bicolores, pondrs,
 * dploys,  treaps ) peuvent tre vus comme des cas particuliers d'arbres
 * cartsiens.
 *
 * Les arbres cartsiens sont des arbres binaires dont les noeuds contiennent 
 * deux valeurs, de faon  ce que la projection horizontale vrifie
 * l'ordre symtrique et la projection verticale l'ordre de tas
 * (voir la figure sommaire)
 * 
 *                               projection verticale (seconde composante)
 * 
 *         (2, 5)                         5
 *          /  \                        /   \
 *      (1, 3) (6, 4)       ->        3       4
 *              /  \                         / \
 *          (5, 2) (7, 3)                   2    3
 * 
 *             |
 *             \/
 * 
 *    projection horizontale (premire composante)
 * 
 *        2
 *      /   \
 *     1     6
 *          / \
 *         5   7
 * 
 *     1 2 5 6 7
 * 
 * Nous prendrons toutes les premires composantes diffrentes de manire
 *  ce que la premire projection soit toujours un arbre de recherche
 * binaire. Les secondes composantes (nommes dans la suite  priorit )
 * peuvent tre quelconques.
 * 
 * Les arbres cartsiens tirent leur nom du fait que si les couples sont
 * tous distincts, la structure de l'arbre est unique, autrement dit les
 * couples peuvent tre vus comme les coordonnes cartsiennes du noeud.
 *)

(* Los rboles cartesianos fueron introducidos por Vuillemin en 1980.
 *
 * Son el fundamento des los rboles en la librera Baire ya que muchos
 * rboles equilibrados (AVL, bicromaticos, ponderados, splay, treaps)
 * pueden ser obtenidos como casos particulares de rboles cartesianos.
 *
 * En los rboles cartesianos cada nudo contiene dos elementos, de manera
 * tal que la proyeccin horizontal verifique el rden simtrico (en otras
 * palabras que sea un rbol binario de bsqueda) y la proyeccin vertical
 * verifique el orden de montn (heap en ingls). Ver la ilustracin
 * 
 *                               proyeccin vertical (segunda componente)
 * 
 *         (2, 5)                         5
 *          /  \                        /   \
 *      (1, 3) (6, 4)       ->        3       4
 *              /  \                         / \
 *          (5, 2) (7, 3)                   2    3
 * 
 *             |
 *             \/
 * 
 *    proyeccin vertical (primera componente)
 * 
 *        2
 *      /   \
 *     1     6
 *          / \
 *         5   7
 * 
 *     1 2 5 6 7
 * 
 * Consideraremos siempre que las primeras componentes son todas distintas
 * (de manera que la proyeccin vertical represente un conjunto). La
 * segunda proyeccin no lleva ninguna restriccin, la llamaremos
 * a menudo  prioridad 
 * 
 * Los rboles cartesianos llevan ese nombre porque si todos los pares
 * de valores son distintos, entoces la estructura del rbol es nica.
 * En otras palabras, los pares pueden verse como la coordenada cartesiana
 * de cada nudo.
 *)

type ('a, 'b) cartesian =
  | E
  | N of ('a, 'b) cartesian * 'a * 'b * ('a, 'b) cartesian


(** Utilitaires **)
(** Utils **)
(** tiles **)


let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type

let makeTree = fun l x p r -> N (l, x, p, r)

let rotate_L = fun l v p r ->
  match r with
    | N (rl, rv, rp, rr) -> makeTree (makeTree l v p rl) rv rp rr
    | _ -> makeTree l v p r
	
let rotate_R = fun l v p r ->
  match l with
    | N (ll, lv, lp, lr) -> makeTree ll lv lp (makeTree lr v p r)
    | _ -> makeTree l v p r

(* [makeSTree] suposes that :
 * - the symetric order is respected l < r
 * - the two subtrees are well formed (cartesian trees)
 *)

(* la fonction [makeSTree] suppose :
 * - le respect de l'ordre symtrique l < r
 * - les sous-arbres l et r tre bien forms (arbres cartesiens)
 *)

(* la funcin [makeSTree] supone :
 * - el respeto del orden simtrico l < r
 * - los subrboles l y r estar bien formados (rboles cartesianos)
 *)

let rec makeSTree = fun l v p r ->
  match (l, r) with
    | (E, E) -> makeTree l v p r
    | (E, N (rl, rv, rp, rr)) ->
	if rp > p then makeTree (makeSTree l v p rl) rv rp r
	else makeTree l v p r
    | (N (ll, lv, lp, lr), E) ->
	if lp > p then makeTree ll lv lp (makeSTree lr v p r)
	else makeTree l v p r
    | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	if lp > rp then
	  (if lp > p then makeTree ll lv lp (makeSTree lr v p r)
	   else makeTree l v p r)
	else
	  (if rp > p then makeTree (makeSTree l v p rl) rv rp rr
	   else makeTree l v p r)

(* [appendS] suposes that :
 * - the symetric order is respected l < r
 * - the two subtrees are well formed (cartesian trees)
 *)

(* la fonction [appendS] suppose :
 * - le respect de l'ordre symtrique l < r
 * - les sous-arbres l et r tre bien forms (arbres cartesiens)
 *)

(* la funcin [appendS] supone :
 * - el respeto del orden simtrico l < r
 * - los subrboles l y r estar bien formados (rboles cartesianos)
 *)

let rec appendS = fun l r ->
  match (l, r) with
    | (E, s) | (s, E) -> s
    | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	match compare lp rp with
	  | n when n < 0 -> makeTree (appendS l rl) rv rp rr
	  | _ -> makeTree ll lv lp (appendS lr r)


(** Oprations catgorielles **)
(** Categorial operations **)
(** Operaciones categoriales **)


let injection = fun x p -> N (E, x, p, E)

let rec partition_inf_sup pivot = function
  | E -> (E, None, E)
  | N (l, v, p, r) ->
      match compare pivot v with
	| n when n < 0 ->
	    let (inf, value, sup) = partition_inf_sup pivot l in
	      (inf, value, makeTree sup v p r)
	| n when n > 0 ->
	    let (inf, value, sup) = partition_inf_sup pivot r in
	      (makeTree l v p inf, value, sup)
	| _ -> (l, Some (v, p), r)

let rec old_plus = fun l r ->
  match (l, r) with
    | (E, s) | (s, E) -> s
    | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	match compare lp rp with
	  | n when n > 0 ->
	      let (inf, _, sup) = partition_inf_sup lv r in
		makeTree (old_plus inf ll) lv lp (old_plus sup lr)
	  | _ ->
	      let (inf, _, sup) = partition_inf_sup rv l in
		makeTree (old_plus inf rl) rv rp (old_plus sup rr)

(* The union algorithm is derived from Stephen Adams' one, which is
 * explained in the research report (avaible on the web)
 * 
 * Implementing Sets Efficiently in a Functional Language
 * Stephen Adams
 * CSTR 92-01
 * University of Southampton
 *)

(* L'algorithme d'union est driv de celui d'union rapide dont
 * le fonctionnement est expliqu dans le rapport de recherche 
 * (disponible sur la toile)
 *
 * Implementing Sets Efficiently in a Functional Language
 * Stephen Adams
 * CSTR 92-01
 * University of Southampton
 *)

(* Algoritmo de unin est derivado del algoritmo de Stephen Adams cuyo 
 * funcionamiento est explicado en el informe (disponible en la red)
 * 
 * Implementing Sets Efficiently in a Functional Language
 * Stephen Adams
 * CSTR 92-01
 * University of Southampton
 *)

let rec filter_lt x = function
  | E -> E
  | N (l, v, p, r) ->
      match compare x v with
	| n when n < 0 -> filter_lt x l
	| n when n > 0 -> makeTree l v p (filter_lt x r)
	| _ -> l

let rec filter_gt x = function
  | E -> E
  | N (l, v, p, r) ->
      match compare x v with
	| n when n < 0 -> makeTree (filter_gt x l) v p r
	| n when n > 0 -> filter_gt x l
	| _ -> r

(* [cut] makes sure that the root of the tree lies between the two bounds. 
 * If this is not the case, the functions calls itself recusively on the 
 * corresponding subtree (right or left)
 *)

(* [cut] s'assure que la racine se trouve entre les deux bornes. 
 * Si ce n'est pas le cas, la fonction restreint rcursivement l'arbre  son 
 * sous-arbre gauche ou droit selon le cas de figure.
 *) 

(* [cut] se asegura que la raz del rbol se encuentre entre los dos lmites. 
 * Si este no es el caso, la funcin toma recursivamente el subrbol derecho 
 * o izquierdo segn convenga
 *)

let rec cut = fun lo tree hi ->
  match tree with
    | N (_, v, _, r) when compare v lo <= 0 -> cut lo r hi
    | N (l, v, _, _) when compare v hi >= 0 -> cut lo l hi
    | _ -> tree

(* [cut_lo] and [cut_hi] are specialized versions with an infinite 
   hight (respectively low) bound *)

(* [cut_lo] et [cut_hi] sont des versions spcialises avec la borne infrieure
 * (respectivement suprieure) ayant pour valeur plus l'infini (respectivement 
 * moins l'infini)
 *)

(* [cut_lo] y [cut_hi] son versiones especializadas con un limite inferior
 * (respectivamente superior) igual a ms (respectivamente menos) el infinito
 *)

let rec cut_hi = fun lo tree ->
  match tree with
    | N (_, v, _, r) when compare v lo <= 0 -> cut_hi lo r
    | _ -> tree

let rec cut_lo = fun tree hi ->
  match tree with
    | N (l, v, _, _) when compare v hi >= 0 -> cut_lo l hi
    | _ -> tree

(* [uni] uses the left tree to control the reconstruction :
 *
 * - the right tree will be physically cut only if it lies between to 
 * successive values of the control tree (which guarantees that there will not
 * be any new cut after the reconstruction)
 * 
 * - if the right tree is empty, the whole control tree cannot be inserted
 * because we do not know were are the bounds from. We then have to do 
 * one more cut
 *)

(* la fonction [uni] se sert de l'arbre dont la racine a la priorit maximale
 * pour contrler la reconstruction : 
 * 
 * - l'autre arbre ne sera dcoup que si l'on se trouve entre deux 
 * valeurs successives de l'arbre de contrle (ce qui garantit qu'il n'y 
 * aura pas ne nouvelle dcoupe aprs la reconstruction)
 * 
 * - contrairement  la version  contrle fixe, si l'autre arbre est vide, 
 * on ne peut pas insrer tout l'arbre de contrle, il faut rajouter une coupe 
 * supplmentaire car on ignore de quel arbre proviennent les bornes
 * 
 * Peut-tre pourrait-on gagner quelque vitesse en spcialisant uni selon
 * l'origine des bornes ?
 *)

(* la funcin [uni] se sirve del rbol izquierdo para controlar la 
 * reconstruccin :
 * 
 * - el rbol derecho solo sera recortado fsicamente si se encuentra
 * entre dos valores succesivos del rbol de control (lo que garantiza
 * que no habr ninguna otra operacin de corte despues de la reconstruccin)
 * 
 * - si el rbol derecho est vaco, contrariamente a la version con control
 * fijo no se puede insertar el rbol de control entero ya que se ignora
 * de que rbol provienen los lmites. Por ello hay que hacer un corte
 * suplementario
 *)

let rec uni = fun lo a b hi ->
  match (a, b) with
    | (E, E) -> E
    | (N _ as node, E) | (E, (N _ as node)) -> 
	(match cut lo node hi with
	   | E -> E
	   | N (l, v, p, r) -> makeTree (filter_gt lo l) v p (filter_lt hi r)
	)
    | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	match compare lp rp with
	  | n when n > 0 -> makeTree 
	      (uni lo ll (cut lo b lv) lv) lv lp (uni lv lr (cut lv b hi) hi)
	  | _ -> makeTree 
	      (uni lo rl (cut lo a rv) rv) rv rp (uni rv rr (cut rv a hi) hi)

(* two specialized versions are given *)

(* on en donne galement deux versions spcialises *)

(* se dan tambin dos versiones especializadas *)

let rec uni_lo = fun a b hi ->
  match (a, b) with
    | (E, E) -> E
    | (E, (N _ as node)) | (N _ as node, E) ->
	(match cut_lo node hi with
	   | E -> E
	   | N (l, v, p, r) -> makeTree l v p (filter_lt hi r)
	)
    | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	match compare lp rp with
	  | n when n > 0 -> makeTree 
	      (uni_lo ll (cut_lo b lv) lv) lv lp (uni lv lr (cut lv b hi) hi)
	  | _ -> makeTree 
	      (uni_lo rl (cut_lo a rv) rv) rv rp (uni rv rr (cut rv a hi) hi)

let rec uni_hi = fun lo a b ->
  match (a, b) with
    | (E, E) -> E
    | (E, (N _ as node)) | (N _ as node, E) ->
	(match cut_hi lo node with
	   | E -> E
	   | N (l, x, p, r) -> makeTree (filter_gt lo l) x p r
	)
    | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	match compare lp rp with
	  | n when n > 0 -> makeTree 
	      (uni lo ll (cut lo b lv) lv) lv lp (uni_hi lv lr (cut_hi lv b))
	  | _ -> makeTree 
	      (uni lo rl (cut lo a rv) rv) rv rp (uni_hi rv rr (cut_hi rv a))

(* We just put all together *)
	      
(* On assemble le tout *)

(* Se juntan todos los pedazos *)

let plus = fun a b ->
  match (a, b) with
    | (_, E) -> a
    | (E, _) -> b
    | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	match compare lp rp with
	  | n when n > 0 -> makeTree
	      (uni_lo ll (cut_lo b lv) lv) lv lp (uni_hi lv lr (cut_hi lv b))
	  | _ -> makeTree 
	      (uni_lo rl (cut_lo a rv) rv) rv rp (uni_hi rv rr (cut_hi rv a))

let zero = function () -> E

let rec join_rec result = function
  | E -> result
  | N (l, v, p, r) -> join_rec (join_rec (plus v result) l) r 

let join = function tree -> join_rec E tree

(* There are two ways to perform an insertion, the simplest has two cases :
 *
 * (i)  the priority of the inserted node is superior to the one of the root
 *      there is just to partition the tree following the inserted value
 * 
 * (ii) the priority of the inserted node is inferior to the one of the root
 *      we go down the tree to the left or the right according to the symetric
 *      order and call the insert function again
 * 
 * note : eventually an element of same key but different priority
 * is replaced
 *)

(* Il y a deux methodes possibles pour l'insertion, 
 * la plus simple se divise en deux cas :
 * 
 * (i)  la priorit du noeud  insrer est suprieure  celle de la racine
 *      alors on dcoupe l'arbre selon la valeur de l'lement et on runit 
 *      le tout.
 * 
 * (ii) la priorit du noeud  insrer est infrieure  la racine
 *      alors en fonction du rsultat de la comparaison entre l'lment et la
 *      racine on se dplace vers le sous arbre gauche ou droit et on ritre
 *      la procdure
 * 
 * note : on remplace ventuellement une entre de mme cl mais de 
 * priorit diffrente
 *)

(* Hay dos maneras de proceder a una insersin, la ms simple se divide
 * en dos casos
 * 
 * (i)  la prioridad del nudo que se desea inserir es superior a la de la
 *      raz, en cuyo caso se recorta el rbol segn el valor dado y se
 *      reunen las tres partes
 * 
 * (ii) la prioridad del nudo que se desea inserir es inferior a la de la
 *      raz, en funcin del resultado de la comparacion del elemento con
 *      la raz se deciende en el subrbol derecho o izquierdo y se
 *      llama recursivamente la funcin
 * 
 * nota : se podr reemplazar una entrada de mismo valor pero de prioridad
 * diferente
 *)

let rec insert x priority = function
  | E -> injection x priority
  | N (l, v, p, r) as node ->
      match compare priority p with
	| n when n > 0 -> 
	    let (inf, _, sup) = partition_inf_sup x node in 
	      makeTree inf x priority sup
	| _ ->
	    match compare x v with
	      | n when n < 0 -> makeTree (insert x priority l) v p r
	      | n when n > 0 -> makeTree l v p (insert x priority r)
	      | _ -> makeSTree l x priority r

(* The second way first inserts the element as if the tree was a common
 * binary search tree, then rebalances the tree until the heap order
 * is restablished on priorities
 *)

(* La seconde methode insre le nouvel lment comme dans un arbre binaire
 * de recherche puis rebalance au moyen de rotations afin de rtablir
 * l'ordre des priorits
 *)

(* El segundo mtodo insiere el nuevo elemento como se hara en un rbol
 * binario comn, luego corrige el balance de manera a reestablecer el
 * rden de las prioridades
 *)

let rec insert' x priority = function
  | E -> injection x priority
  | N (l, v, p, r) ->
      match compare x v with
	| n when n < 0 ->
	    if priority > p then rotate_R (insert' x priority l) v p r
	    else makeTree (insert' x priority l) v p r
	| n when n > 0 ->
	    if priority > p then rotate_L l v p (insert' x priority r)
	    else makeTree l v p (insert' x priority r)
	| _ -> makeSTree l x priority r

let rec from_list_rec result = function
  | [] -> result
  | (v, p) :: tail -> from_list_rec (insert v p result) tail

let from_list = function list -> from_list_rec E list

let rec to_list = function
  | E -> []
  | N (l, v, p, r) -> to_list l @ (v, p) :: to_list r

let rec to_list_by_priority = function
  | E -> []
  | N (l, v, p, r) -> (v, p) :: to_list_by_priority (plus l r)

let map f = function set -> 
  from_list (List.map (fun (x, y) -> (f x, y)) (to_list set))


(** Oprations binaires **)
(** Binary operaations **)
(** Opraciones binrias **)


let union = plus

let rec intersection = fun a b ->
  match (a, b) with
    | (E, _) | (_, E) -> E
    | (N (l, v, p, r), _) ->
	match partition_inf_sup v b with
	  | (l', None, r') -> appendS (intersection l l') (intersection r r')
	  | (l', Some (_, p'), r') -> 
	      makeSTree (intersection l l') v (max p p') (intersection r r')

let rec difference = fun a b ->
  match (a, b) with
    | (E, _) -> E
    | (_, E) -> a
    | (N (l, v, p, r), _) -> 
	match partition_inf_sup v b with
	  | (l', None, r') -> makeSTree (difference l l') v p (difference r r')
	  | (l', Some _, r') -> appendS (difference l l') (difference r r')
	      
let rec symetric_difference = fun a b ->
  match (a, b) with
    | (E, s) | (s, E) -> s
    | (N (l, v, p, r), _) ->
	match partition_inf_sup v b with
	  | (l', None, r') -> makeSTree (symetric_difference l l') v p
	      (symetric_difference r r')
	  | (l', Some _, r') -> 
	      appendS (symetric_difference l l') (symetric_difference r r')


(** Accs **)
(** Acces **)
(** Acceso **)


let rec delete_raise x = function
  | E -> raise Not_found
  | N (l, v, p, r) ->
      match compare x v with
	| n when n < 0 -> makeTree (delete_raise x l) v p r
	| n when n > 0 -> makeTree l v p (delete_raise x r)
	| _ -> plus l r

let delete = fun x tree ->
  try
    delete_raise x tree
  with Not_found -> tree

let rec delete_min = function
  | E -> E
  | N (E, _, _, r) -> r
  | N (l, v, p, r) -> makeTree (delete_min l) v p r

let rec delete_max = function
  | E -> E
  | N (l, _, _, E) -> l
  | N (l, v, p, r) -> makeTree l v p (delete_max r)

let rec delete_max_priority = function
  | E -> E
  | N (l, _, _, r) -> appendS l r

let rec min_priority_rec ((_, pm) as min) = function
  | E -> min
  | N (E, v, p, E) -> 
      (match compare p pm with
	 | n when n < 0 -> (v, p)
	 | _ -> min
      )
  | N (l, _, _, r) -> min_priority_rec (min_priority_rec min l) r

let min_priority = function
  | E -> raise Not_found
  | N (l, v, p, r) -> min_priority_rec (min_priority_rec (v, p) l) r

(* elements (first key) are unique *)
(* les lments (premire composante) sont uniques *)
(* los elementos (primera componente) son nicas *)

let delete_min_priority = function tree ->
  let (x, _) = min_priority tree in 
    delete x tree

let rec delete_list = fun list set ->
  match list with
    | [] -> set
    | head :: tail -> delete_list tail (delete head set)

let rec filter_rec f result = function
  | E -> result
  | N (l, v, p, r) ->
      filter_rec f (filter_rec f (
		      if f v p then insert v p result else result) l) r

let filter f = function set -> filter_rec f E set

let rec filter_le x = function
  | E -> E
  | N (l, v, p, r) ->
      match compare x v with
	| n when n < 0 -> filter_le x l
	| _ -> makeTree l v p (filter_le x r)

let rec filter_ge x = function
  | E -> E
  | N (l, v, p, r) ->
      match compare x v with
	| n when n <= 0 -> makeTree (filter_ge x l) v p r
	| _ -> filter_ge x r

let rec range min max = function
  | E -> E
  | N (l, v, p, r) ->
      if compare v min < 0 then range min max r
      else if compare max v < 0 then range min max l
      else makeSTree (filter_ge min l) v p (filter_le max r)


(** Vues **)
(** Views **)
(** Vistas **)


let rec min_view = function
  | E -> raise Empty
  | N (E, v, p, r) -> ((v, p), r)
  | N (l, v, p, r) -> let (min, left) = min_view l in 
      (min, makeTree left v p r)
      
let rec max_view = function
  | E -> raise Empty
  | N (l, v, p, E) -> (l, (v, p))
  | N (l, v, p, r) -> let (right, max) = max_view r in 
      (makeTree l v p right, max)

let max_priority_view = function
  | E -> raise Empty
  | N (l, v, p, r) -> ((v, p), plus l r)

let rec min_priority_view = function
  | E -> raise Empty
  | N (E, v, p, _) -> (E, (v, p))
  | N (l, v, p, E) -> 
      let (l', v') = min_priority_view l in (makeTree l' v p E, v')
  | N (l, v, p, r) ->
      let (_, lp) = min_priority l and (_, rp) = min_priority r in
	match compare lp rp with
	  | n when n > 0 ->
	      let (r', min) = min_priority_view r in (makeTree l v p r', min)
	  | _ ->
	      let (l', min) = min_priority_view l in (makeTree l' v p r, min)

let rec partition_rec f (trueS, falseS) = function
  | E -> (trueS, falseS)
  | N (l, v, p, r) -> 
      if f v p
      then partition_rec f (partition_rec f (insert v p trueS, falseS) l) r
      else partition_rec f (partition_rec f (trueS, insert v p falseS) l) r

let partition = fun f set -> partition_rec f (E, E) set


(** Observateurs **)
(** Observers **)
(** Observadores **)


let is_empty = function
  | E -> true
  | N _ -> false

let rec size_rec result = function
  | E -> result
  | N (l, _, _, r) -> size_rec (size_rec (1 + result) r) l

let size = function tree -> size_rec 0 tree

(* [member'] : does (d + 1) comparations instead of 3/2 * d
 * following an idea of Arne Andersson, but walks down the whole height of 
 * the tree and an extra argument has to be transmitted in each call
 * 
 * It may not always be faster than [member]
 *)

(* [member'] : effectue en moyenne (d + 1) comparaisons au lieu de 3/2 * d
 * selon une ide d'Arne Andersson mais parcourt systmatiquement l'arbre sur
 * toute sa hauteur et doit transmettre un paramtre supplmentaire  chaque
 * appel
 * 
 * Il est possible qu'elle ne soit pas toujours plus rapide que [member]
 *) 

(* [member'] : hace (d + 1) comparaciones en vez de 3/2 * d en promedio, 
 * segn una idea de Arne Andersson, pero recorre toda la altura del rbol
 * y requiere un parmetro suplementrio en cada llamado
 * 
 * Es posible que no sea siempre ms veloz que [member]
 *)

let rec member_min_bound b x = function
  | E -> compare b x = 0
  | N (l, v, _, r) ->
      match compare x v with
	| k when k < 0 -> member_min_bound b x l
	| _ -> member_min_bound v x r
	    
let rec member' x = function
  | E -> false
  | N (l, v, _, r) ->
      match compare x v with
	| n when n < 0 -> member' x l
	| _ -> member_min_bound v x r

(* [member] : usual member function
 * some algorithms use the properties of the usual [member] function 
 * (stop of a positive search in the middle of the tree). It is the case
 * of optimal tree construction for example (Knuthm Garcia-Walsh, confer
 * The art of computer programming vol III)
 *)

(* [member] : fonction member habituelle 
 * un certain nombre d'algorithmes utilisent les proprits de la fonction
 * [member] habituelle (arrt d'une recherche positive  l'intrieur de 
 * l'arbre) par exemple les algorithmes de construction d'arbres optimaux
 * (Knuth, Garcia-Walsh, confer The art of computer programming vol III)
 *)

(* [member] : funcin member comn
 * ciertos algortmos usan las propriedades de la funcin [member] comn
 * (interupcin de una busqueda positiva en el medio del rbol) tal
 * los algortmos de construccin de rboles optimales (Knuth, Garcia-Walsh,
 * confer The art of computer programming vol III)
 *)

let rec member x = function
  | E -> false
  | N (l, v, _, r) ->
      match compare x v with
	| n when n < 0 -> member x l
	| n when n > 0 -> member x r
	| _ -> true

let rec find f = function
  | E -> raise Not_found
  | N (_, v, _, _) when f v -> v
  | N (l, _, _, r) -> (try find f l with Not_found -> find f r)

let choose = function
  | E -> raise Empty
  | N (_, v, _, _) -> v

let rec min_element = function
  | E -> raise Not_found
  | N (E, v, p, _) -> (v, p)
  | N (l, _, _, _) -> min_element l

let rec max_element = function
  | E -> raise Not_found
  | N (_, v, p, E) -> (v, p)
  | N (_, _, _, r) -> max_element r

let rec max_priority = function
  | E -> raise Empty
  | N (_, v, p, _) -> (v, p)

let rec next x = function
  | E -> raise Not_found
  | N (l, v, p, r) ->
      match compare x v with
	| n when n < 0 -> (try next x l with Not_found -> (v, p))
	| n when n > 0 -> next x r
	| _ -> min_element r (* raises [Not_found] *)

let rec previous x = function
  | E -> raise Not_found
  | N (l, v, p, r) ->
      match compare x v with
	| n when n < 0 -> previous x l
	| n when n > 0 -> (try previous x r with Not_found -> (v, p))
	| _ -> max_element l (* raises [Not_found] *)
	    
let rec ceil x = function
  | E -> raise Not_found
  | N (l, v, p, r) ->
      match compare x v with
	| n when n < 0 -> (try ceil x l with Not_found -> (v, p))
	| n when n > 0 -> ceil x r
	| _ -> (x, p)

let rec floor x = function
  | E -> raise Not_found
  | N (l, v, p, r) ->
      match compare x v with
	| n when n < 0 -> floor x l
	| n when n > 0 -> (try floor x r with Not_found -> (v, p))
	| _ -> (x, p)

(* les arbres binaires de recherche sont indexs suivant l'ordre symtrique *)

(* binary search trees are indexed following the symetric order *)

(* los rboles de busqueda estn indizados segn el orden simtrico *)

let rec rank x = function
  | E -> raise Not_found
  | N (l, v, _, r) ->
      match compare x v with
	| n when n < 0 -> rank x l
	| n when n > 0 -> 1 + size l + rank x r
	| _ -> size l

let rec get n = function
  | E -> raise (Out_of_bounds n)
  | N (l, v, p, r) -> 
      let k = size l in
	if n < k then get n l
	else if n > k then get (n - k - 1) r
	else (v, p)

let rec unsafe_set value priority m = function
  | E -> raise (Out_of_bounds m)
  | N (l, v, p, r) ->
      let k = size l in
	if m < k then makeTree (unsafe_set value priority m l) v p r
	else if m > k then 
	  makeTree l v p (unsafe_set value priority (m - k - 1) r)
	else makeTree l value priority r

let rec for_all f = function
  | E -> true
  | N (l, v, p, r) -> f v p && for_all f l && for_all f r

let rec exists f = function
  | E -> false
  | N (l, v, p, r) -> f v p || exists f l || exists f r


(** Traversals **)
(** Parcours **)
(** Recorridos **)


let rec iter f = function
  | E -> ()
  | N (l, v, p, r) -> iter f l ; f v p ; iter f r

let rec fold_left f e = function
  | E -> e
  | N (l, v, p, r) -> fold_left f (f (fold_left f e l) v p) r

let rec fold_right = fun f set e ->
  match set with
    | E -> e 
    | N (l, v, p, r) -> fold_right f l (f v p (fold_right f r e))

let fold = fold_right


(** Miscellaneous **)
(** Divers **)
(** Miscelanios **)


let bind = fun set f -> join (map f set)

let concat_map = fun f set -> join (map f set)


(** Compatibilit avec la librairie standard de Caml **)
(** Caml standard library compatibility **)
(** Compatibilidad con la librera estandar de Caml **)

let empty = E
let mem = member
let add = insert
let singleton = injection
let remove = delete
let inter = intersection
let diff = difference
let cardinal = size
let elements = to_list
let min_elt = min_element    
let max_elt = max_element


module type HeightFunction = sig
  val height : int -> int -> int -> int
  val first_priority : unit -> int
end

module type OrderedType = sig
  type t
  val compare : t -> t -> int
end

(* This modules are given for the user to play with the MakeSet functor
 * and check that the usual balancing schemes may be obtained that way
 *)

(* Ces modules sont fournis afin que l'utilisateur puisse exprimenter avec
 * le foncteur MakeSet et constater que l'on obtient bien les schmas 
 * d'quilibrage usuels
 *)

(* El propsito de stos mdulos es que el utilisador pueda experimentar
 * con el funtor MakeSet y verifique que se obtienen de esa manera los
 * esquemas de equilibraje ms comunes
 *)

module Splay = struct

  let height = fun p l r -> p

  let count = ref 0

  let first_priority = function () -> incr count ; !count

end

module Treap = struct
  
  let height = fun p l r -> p
    
  let _ = Random.self_init ()

  let first_priority = function () -> Random.bits ()

end

module AVL = struct

  let height = fun p l r ->
    if abs (l - r) < 2 then 1 + max l r
    else 1 + min l r

  let first_priority = function () -> 0

end

(* We can see that the height function of :
 * - splay trees and treaps does not depend on the other nodes
 * - AVL trees depend on the two child nodes
 * 
 * To obtain a similar function for red and black trees we would
 * need to adapt the MakeSet functor to handle deeper dependencies
 * ( red-red conflicts  have to be solved)
 * 
 * Then the height function would look like the following
 *)

(* Nous constatons que la fonction de hauteur :
 * - des arbres Splay et treap ne dpend pas des autres noeuds
 * - des arbres AVL dpend des deux fils
 * 
 * Pour obtenir une fonction quivalente pour les arbres bicolores
 * il faudrait adapter le foncteur afin de prendre en compte les
 * noeuds de profondeur deux (pour rsoudre les  conflits rouges 
 * dans la nomeclature bicolore)
 * 
 * La fonction de hauteur aurait alors la forme suivante
 *)

(* Ntese que la funcin de altura de los rboles :
 * - splay y treap no depende de los tros nudos
 * - AVL depende de los dos nudos descendientes
 * 
 * Para obtener una funcin similar en el caso de los rboles bicromticos
 * tendriamos que adaptar el functor MakeSTree de manera tal que
 * accepte las dependencias ms profundas (hay que resolver los
 *  conflictos rojos  en la terminologia de los rboles bicromticos)
 * 
 * Entonces, la funcion de altura tendra la forma siguiente
 *)

let redBlackHeight = fun p ll l lr rl r rr ->
  if l > r then 
    if ll = l || lr = r (* red-red conflit *) then r
    else 1 + min l r
  else
    if rl = r || rr = r (* red-red conflit *) then l
    else 1 + min l r

(* Why does this functor work ?
 *
 * - AVL trees have a  natural  height function, that is the lenght
 * of the longest path from a given node to the leafs. It is not long
 * to check that an AVL tree in which every node is given a priority
 * equal to that lenght is a cartesian tree.
 * 
 * One way to calculate this length is the following
 * 
 * let rec height = function
 *   | E -> 0
 *   | N (l, v, r, _) -> 1 + max (height l) (height r)
 * 
 * That is exactly (1 + min l r) the function we use in the functor 
 * specialisation
 * 
 * - red and black trees also have a height function even if it is not
 * so well known. Theoricians call-il the  black height  because it
 * counts the number of black nodes between a given node and the leafs
 * (which is a constant because or red-black tree properties)
 * One can check that one also obtain a cartesian tree
 * 
 * The black height may be computed by the following function
 * 
 * let rec height = function
 *    | E -> 0
 *    | N (l, v, r, _) -> 1 + min (height l) (height r)
 * 
 * That is exactly the function we use (1 + min l r)
 * 
 * - in the same way one can show that many balancing schemes can be
 * associated to  height functions  (e. g. splay, treaps, weight balanced)
 *)

(* Pourquoi ce foncteur fonctionne-t-il ?
 *
 * - les arbres AVL ont une fonction  naturelle  de hauteur, il s'agit
 * de la longueur du plus long chemin d'un noeud aux feuilles. Il 
 * n'est pas difficile de vrifier qu'un arbre AVL dont les noeuds sont
 * pondrs par une priorit gale  leur hauteur vrifie les proprits
 * des arbres cartsiens
 * 
 * C'est justement cette hauteur que calcule la fonction 
 * 
 * let rec height = function
 *    | E -> 0
 *    | N (l, v, r, _) -> 1 + max (height l) (height r)
 * 
 * autrement dit la fonction 1 + max l r utilise plus haut
 * 
 * - les arbres bicolores ont galement une fonction de hauteur mais un peu
 * moins connue, il s'agit de ce que la thorie nomme la  hauteur noire 
 * car elle dsigne le nombre de noeuds noirs entre un noeud donn et
 * les feuilles (ce nombre est constant en raison des proprits des
 * arbres bicolores). Il n'est pas difficile de vrifier que l'on obtient 
 * aussi un arbre cartsien
 * 
 * Or cette hauteur noire se calcule au moyen de la fonction
 * 
 * let rec height = function
 *    | E -> 0
 *    | N (l, v, r, _) -> 1 + min (height l) (height r)
 * 
 * C'est exactement ce que nous faisons avec la fonction 1 + min l r
 * 
 * - on exhibe de mme des fonction de  hauteur  adaptes  divers
 * autres schmas d'quilibrage (dploy, pondr, etc.)
 *)

(* Porqu funciona ste funtor ?
 *
 * - los rboles AVL tienen una nocin natural de altura, se trata de
 * la longitud del camino ms largo entre un nudo dado y las hojas.
 * No es difcil verificar que si se toma un arbol AVL y a cada nudo
 * se le asigna una prioridad igual a esa logitud, se obtiene un rbol
 * cartesiano
 * 
 * la mayr logitud se puede calcular con la funcin
 * 
 * let height = function
 *    | E -> 0
 *    | N (l, v, r, _) -> 1 + max (height l) (height r)
 * 
 * Pues es sta justamente la funcin que usamos en nuestro funtor
 * 
 * - los rboles bicromticos tambien tienen una funcin de altura, es
 * verdad un poco menos conocida. La teora la llama  altura negra .
 * No es difcil verificar que tambin se obtiene un rbol cartesiano.
 * 
 * sta altura se calcula con con la funcin siguiente
 * 
 * let rec height = funciton
 *    | E -> 0
 *    | N (l, v, r, _) -> 1 + min (height l) (height r)
 * 
 * Pues una vez mas es sta la funcin que usamos en el funtor
 * 
 * - se pueden tambien construir funciones similares para varios esquemas
 * de balance (splay, rboles ponderados, etc.)
 *)

(* The files avlSet.ml and redBlackSet.ml contain an direct specialisation
 * of the MakeSet functor
 * 
 * the function Avl.Height has been splitted in two :
 * - the | l - r | >= 2 case is handled by rotations in makeTree
 * - the | l = r | < 2 case is directly included in makeTree (notice
 * the expression 1 + max l r)
 * 
 * the redBlackHeight function has been splitted in two :
 * - red conflicts (ll = l, lr = l, rl = r or rr = r) are handled
 * by rotations in makeBTree
 * - the other cases are directly included in makeTree (notice the expression
 * 1 + min l r)
 *)

(* Les modules avlSet.ml, redBalancedSet.ml contiennent une implmentation
 * directe de cette spcialisation des arbres cartsiens.
 * 
 * la fonction Avl.Height a t divise en deux parties :
 * - la partie | l - r | >= 2 est rgle par rotations dans makeBTree
 * - la partie | l - r | < 2 se trouve directement dans le constructeur 
 * makeTree (on reconnat le fragment 1 + max l r)
 * 
 * la fonction redBlackHeight a t divise en deux parties :
 * - les conflits rouges (ll = l, lr = l, rl = r ou rr = r) sont rgls
 * par rotations dans makeBTree
 * - le reste se trouve dans le constructeur makeTree (on reconnat le 
 * fragment 1 + min l r)
 *)

(* Los modulos avlSet.ml y redBlackSet.ml contienen una implementacin
 * directa de sta especializacin de los rboles cartesianos
 * 
 * La funcin Avl.Height ha sido dividida en dos partes :
 * - la parte | l - r | >= 2 se resuelve con rotaciones dentro de makeBTree
 * - la perte | l - r | < 2 est directamente incluida dentro del constructor
 * makeTree (se reconoce el fragmento 1 + max l r)
 * 
 * la funcin redBlackHeight ha sido dividida en dos partes :
 * - los conflictos rojos son resueltos con rotaciones dentro de makeBTree
 * - el resto se encuentra directamente en el contructor makeTree (se
 * reconoce el fragmento 1 + min l r)
 *)

(* localAvlSet.ml and localRedBlackSet.ml contain an optimisation
 * of AVL and red-black trees obtained by the  bichromatic formalism 
 * of Sedgewick and Guibas.
 * 
 * These are what people usually understand by AVL and red-black trees
 * (the way they are usually thought in textbooks)
 * 
 * These optimisations make
 * - local operations faster (insertions, deletions)
 * - global operations slower (union, intersection)
 * 
 * Confer directly to those files for more information
 *)

(* Les fichiers localAvlSet.ml et localRedBlackSet.ml contiennent une 
 * optimisation plus pousse des arbres AVL et bicolores, obtenue via
 * le  formalisme bicolore  de Sedgewick et Guibas
 * 
 * Il s'agit l de ce que l'on entend plus traditionnelement par arbres
 * bicolores ou AVL (tels que prsents dans les livres par exemple).
 * 
 * Ces optimisations
 * - acclrent les oprations locales (insertions, suppressions)
 * - ralentissent les oprations globales (union, intersection)
 *
 * Se rfrer  ces deux fichiers pour de plus amples informations
 *) 

(* Los ficheros localAvlSet.ml y localRedBlackSet.ml presentan una
 * optimizacin mayor de los rboles AVL y bicromticos, obtenida via
 * el  formalismo bicromtico  de Sedgewick y Guibas
 * 
 * Encontrara rboles AVL y bicromticos presentados de manera ms estandar
 * (tal como se encuentra en los libros de clase)
 * 
 * Estas optimizaciones
 * - acceleran las operaciones locales (inserciones, supreciones)
 * - frenan las operaciones globales (unin, interseccin)
 *)


module MakeSet (Inf : HeightFunction) (Ord : OrderedType) =
struct
  
  type elt = Ord.t
  type t = E | N of t * elt * int * t

  let height = Inf.height
  let first_priority = Inf.first_priority

  let h = function
    | E -> 0
    | N (_, _, h, _) -> h

  (* Modifications with respect to cartesian trees :
   *
   * - makeTree (priority set by height function)
   * 
   * - makeSTree replaces makeTree in several functions (insert, 
   * partition_inf_sup, filter, ... ) since priorities may not be constant
   * any more
   * 
   * - insert, insert' (priority set by first_priority function)
   * 
   * - from_list doesn't ask for priorities any more
   *)

  (* Modifications par rapport aux arbres cartsiens :
   *
   * - makeTree (priorit determine par la fonction height)
   * 
   * - makeSTree remplace makeTree dans plusieurs fonctions (insert,
   * partition_inf_sup, filter, ...) dans la mesure o les priorits
   * peuvent ne plus tre constantes
   * 
   * - insert, insert' (priorit dtermine par la fonction first_priority)
   * 
   * - from_list ne requiert plus d'indiquer les priorits
   *)

  (* Modificaciones con respecto a los rbols cartesianos :
   *
   * - makeTree (prioridad determinada por la funcin height)
   * 
   * - makeSTree reemplaza makeTree en varias funciones (insert, 
   * partition_inf_sup, filter, ...) pues las prioridades pueden no ser
   * constantes
   * 
   * - insert, insert' (prioridad determinada por la funcin first_priority)
   * 
   * - from_list no pide que se le den las prioridades
   *)

  let makeTree = fun l v p r -> N (l, v, height p (h l) (h r), r)

  let rotate_L = fun l v p r ->
    match r with
      | N (rl, rv, rp, rr) -> makeTree (makeTree l v p rl) rv rp rr
      | _ -> makeTree l v p r
	  
  let rotate_R = fun l v p r ->
    match l with
      | N (ll, lv, lp, lr) -> makeTree ll lv lp (makeTree lr v p r)
      | _ -> makeTree l v p r

  let rec makeSTree = fun l v p r ->
    let p = height p (h l) (h r) in
      match (l, r) with
	| (E, E) -> makeTree l v p r
	| (E, N (rl, rv, rp, rr)) ->
	    if rp > p then makeTree (makeSTree l v p rl) rv rp rr
	    else makeTree l v p r
	| (N (ll, lv, lp, lr), E) ->
	    if lp > p then makeTree ll lv lp (makeSTree lr v p r)
	    else makeTree l v p r
	| (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	    if lp > rp then
	      (if lp > p then makeTree ll lv lp (makeSTree lr v p r)
	       else makeTree l v p r)
	    else
	      (if rp > p then makeTree (makeSTree l v p rl) rv rp rr
	       else makeTree l v p r)

  let rec appendS = fun l r ->
    match (l, r) with
      | (E, s) | (s, E) -> s
      | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	  match compare lp rp with
	    | n when n < 0 -> makeSTree (appendS l rl) rv rp rr
	    | _ -> makeSTree ll lv lp (appendS lr r)

  let injection = fun x p -> makeTree E x p E

  let rec partition_inf_sup pivot = function
    | E -> (E, None, E)
    | N (a, x, p, b) ->
	match Ord.compare pivot x with
	  | n when n < 0 ->
	      let (inf, value, sup) = partition_inf_sup pivot a in
		(inf, value, makeSTree sup x p b)
	  | n when n > 0 ->
	      let (inf, value, sup) = partition_inf_sup pivot b in
		(makeSTree a x p inf, value, sup)
	  | _ -> (a, Some (x, p), b)

  let rec old_plus = fun l r ->
    match (l, r) with
      | (E, s) | (s, E) -> s
      | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	  match compare lp rp with
	    | n when n > 0 ->
		let (inf, _, sup) = partition_inf_sup lv r in
		  makeSTree (old_plus inf ll) lv lp (old_plus sup lr)
	    | _ ->
		let (inf, _, sup) = partition_inf_sup rv l in
		  makeSTree (old_plus inf rl) rv rp (old_plus sup rr)

  let rec filter_lt x = function
    | E -> E
    | N (a, y, p, b) ->
	match Ord.compare x y with
	  | n when n < 0 -> filter_lt x a
	  | n when n > 0 -> makeSTree a y p (filter_lt x b)
	  | _ -> a

  let rec filter_gt x = function
    | E -> E
    | N (a, y, p, b) ->
	match Ord.compare x y with
	  | n when n < 0 -> makeSTree (filter_gt x a) y p b
	  | n when n > 0 -> filter_gt x a
	  | _ -> b

  let rec cut = fun lo tree hi ->
    match tree with
      | N (_, x, _, b) when Ord.compare x lo <= 0 -> cut lo b hi
      | N (a, x, _, _) when Ord.compare x hi >= 0 -> cut lo a hi
      | _ -> tree

  let rec cut_hi = fun lo tree ->
    match tree with
      | N (_, x, _, b) when Ord.compare x lo <= 0 -> cut_hi lo b
      | _ -> tree

  let rec cut_lo = fun tree hi ->
    match tree with
      | N (a, x, _, _) when Ord.compare x hi >= 0 -> cut_lo a hi
      | _ -> tree

  let rec uni = fun lo a b hi ->
    match (a, b) with
      | (E, E) -> E
      | (N _ as node, E) | (E, (N _ as node)) -> 
	  (match cut lo node hi with
	     | E -> E
	     | N (l, x, p, r) -> 
		 makeSTree (filter_gt lo l) x p (filter_lt hi r)
	  )
      | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	  match compare lp rp with
	    | n when n > 0 -> makeSTree 
		(uni lo ll (cut lo b lv) lv) lv lp (uni lv lr (cut lv b hi) hi)
	    | _ -> makeSTree 
		(uni lo rl (cut lo a rv) rv) rv rp (uni rv rr (cut rv a hi) hi)

  let rec uni_lo = fun a b hi ->
    match (a, b) with
      | (E, E) -> E
      | (E, (N _ as node)) | (N _ as node, E) ->
	  (match cut_lo node hi with
	     | E -> E
	     | N (l, x, p, r) -> makeSTree l x p (filter_lt hi r)
	  )
      | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	  match compare lp rp with
	    | n when n > 0 -> makeSTree 
		(uni_lo ll (cut_lo b lv) lv) lv lp (uni lv lr (cut lv b hi) hi)
	    | _ -> makeSTree 
		(uni_lo rl (cut_lo a rv) rv) rv rp (uni rv rr (cut rv a hi) hi)

  let rec uni_hi = fun lo a b ->
    match (a, b) with
      | (E, E) -> E
      | (E, (N _ as node)) | (N _ as node, E) ->
	  (match cut_hi lo node with
	     | E -> E
	     | N (l, x, p, r) -> makeSTree (filter_gt lo l) x p r
	  )
      | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	  match compare lp rp with
	    | n when n > 0 -> makeSTree 
		(uni lo ll (cut lo b lv) lv) lv lp (uni_hi lv lr (cut_hi lv b))
	    | _ -> makeSTree 
		(uni lo rl (cut lo a rv) rv) rv rp (uni_hi rv rr (cut_hi rv a))

  let plus = fun a b ->
    match (a, b) with
      | (_, E) -> a
      | (E, _) -> b
      | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	  match compare lp rp with
	    | n when n > 0 -> makeSTree
		(uni_lo ll (cut_lo b lv) lv) lv lp (uni_hi lv lr (cut_hi lv b))
	    | _ -> makeSTree 
		(uni_lo rl (cut_lo a rv) rv) rv rp (uni_hi rv rr (cut_hi rv a))

  let zero = function () -> E

  let rec insert_rec x priority = function
    | E -> injection x priority
    | N (a, y, p, b) as node ->
	match compare priority p with
	  | n when n > 0 -> 
	      let (inf, _, sup) = partition_inf_sup x node in 
		makeSTree inf x priority sup
	  | _ ->
	      match Ord.compare x y with
		| n when n < 0 -> makeSTree (insert_rec x priority a) y p b
		| n when n > 0 -> makeSTree a y p (insert_rec x priority b)
		| _ -> makeSTree a x priority b
		    
  let insert = fun x p tree -> insert_rec x (first_priority ()) tree

  let rec insert_rec' x priority = function
    | E -> injection x priority
    | N (a, y, p, b) ->
	match Ord.compare x y with
	  | n when n < 0 ->
	      if priority > p then rotate_R (insert_rec' x priority a) y p b
	      else makeSTree (insert_rec' x priority a) y p b
	  | n when n > 0 ->
	      if priority > p then rotate_L a y p (insert_rec' x priority b)
	      else makeSTree a y p (insert_rec' x priority b)
	  | _ -> makeSTree a x priority b

  let insert'= fun x p tree -> insert_rec' x (first_priority ()) tree
    
  let rec from_list_rec result = function
    | [] -> result
    | x :: tail -> from_list_rec (insert x 0 result) tail

  let from_list = function list -> from_list_rec E list

  let rec to_list = function
    | E -> []
    | N (a, x, p, b) -> to_list a @ (x, p) :: to_list b

  let rec to_list_by_priority = function
    | E -> []
    | N (a, x, p, b) -> (x, p) :: to_list_by_priority (plus a b)

  let map f = function set -> 
    from_list (List.map (fun (x, y) -> (f x)) (to_list set))

  let union = plus

  let rec intersection = fun a b ->
    match (a, b) with
      | (E, _) | (_, E) -> E
      | (N (l, x, p, r), _) ->
	  match partition_inf_sup x b with
	    | (l', None, r') -> appendS (intersection l l') (intersection r r')
	    | (l', Some (_, p'), r') -> 
		makeSTree (intersection l l') x (max p p') (intersection r r')

  let rec difference = fun a b ->
    match (a, b) with
      | (E, _) -> E
      | (_, E) -> a
      | (N (l, x, p, r), _) -> 
	  match partition_inf_sup x b with
	    | (l', None, r') -> 
		makeSTree (difference l l') x p (difference r r')
	    | (l', Some _, r') -> appendS (difference l l') (difference r r')
		
  let rec symetric_difference = fun a b ->
    match (a, b) with
      | (E, s) | (s, E) -> s
      | (N (l, x, p, r), _) ->
	  match partition_inf_sup x b with
	    | (l', None, r') -> makeSTree (symetric_difference l l') x p
		(symetric_difference r r')
	    | (l', Some _, r') -> 
		appendS (symetric_difference l l') (symetric_difference r r')

  let rec delete_raise x = function
    | E -> raise Not_found
    | N (a, y, p, b) ->
	match Ord.compare x y with
	  | n when n < 0 -> makeSTree (delete_raise x a) y p b
	  | n when n > 0 -> makeSTree a y p (delete_raise x b)
	  | _ -> plus a b

  let delete = fun x tree ->
    try
      delete_raise x tree
    with Not_found -> tree

  let rec delete_min = function
    | E -> E
    | N (E, _, _, r) -> r
    | N (l, v, p, r) -> makeTree (delete_min l) v p r
	
  let rec delete_max = function
    | E -> E
    | N (l, _, _, E) -> l
    | N (l, v, p, r) -> makeTree l v p (delete_max r)

  let rec delete_max_priority = function
    | E -> E
    | N (a, _, _, b) -> appendS a b

  let rec min_priority_rec ((_, pm) as min) = function
    | E -> min
    | N (E, v, p, E) -> 
	(match compare p pm with
	   | n when n < 0 -> (v, p)
	   | _ -> min
	)
    | N (l, _, _, r) -> min_priority_rec (min_priority_rec min l) r
	
  let min_priority = function
    | E -> raise Not_found
    | N (l, v, p, r) -> min_priority_rec (min_priority_rec (v, p) l) r

  let delete_min_priority = function tree ->
    let (x, _) = min_priority tree in 
      delete x tree

  let rec delete_list = fun list set ->
    match list with
      | [] -> set
      | head :: tail -> delete_list tail (delete head set)

  let rec filter_rec f result = function
    | E -> result
    | N (a, x, p, b) ->
	filter_rec f (filter_rec f (
			if f x p then insert x p result else result) a) b

  let filter f = function set -> filter_rec f E set

  let rec filter_le x = function
    | E -> E
    | N (l, v, p, r) ->
	match Ord.compare x v with
	  | n when n < 0 -> filter_le x l
	  | _ -> makeTree l v p (filter_le x r)
	      
  let rec filter_ge x = function
    | E -> E
    | N (l, v, p, r) ->
	match Ord.compare x v with
	  | n when n <= 0 -> makeTree (filter_ge x l) v p r
	  | _ -> filter_ge x r

  let rec range min max = function
    | E -> E
    | N (l, v, p, r) ->
	if Ord.compare v min < 0 then range min max r
	else if Ord.compare max v < 0 then range min max l
	else makeSTree (filter_ge min l) v p (filter_le max r)

  let rec min_view = function
    | N (E, x, p, b) -> ((x, p), b)
    | N (a, x, p, b) -> let (min, left) = min_view a in 
	(min, makeSTree left x p b)
    | _ -> raise Empty
	
  let rec max_view = function
    | N (a, x, p, E) -> (a, (x, p))
    | N (a, x, p, b) -> let (right, max) = max_view b in 
	(makeSTree a x p right, max)
    | _ -> raise Empty

  let max_priority_view = function
    | N (a, x, p, b) -> ((x, p), plus a b)
    | _ -> raise Empty

  let rec min_priority_view = function
    | E -> raise Empty
    | N (E, x, p, _) -> (E, (x, p))
    | N (a, x, p, E) -> 
	let (a', y) = min_priority_view a in (makeTree a' x p E, y)
    | N (a, x, p, b) ->
	let (_, pa) = min_priority a and (_, pb) = min_priority b in
	  match compare pa pb with
	    | n when n > 0 ->
		let (b', min) = min_priority_view b in (makeTree a x p b', min)
	    | _ ->
		let (a', min) = min_priority_view a in (makeTree a' x p b, min)

  let rec partition_rec f (trueS, falseS) = function
    | E -> (trueS, falseS)
    | N (a, x, p, b) -> 
	if f x p
	then partition_rec f (partition_rec f (insert x p trueS, falseS) a) b
	else partition_rec f (partition_rec f (trueS, insert x p falseS) a) b

  let partition = fun f set -> partition_rec f (E, E) set

  let is_empty = function
    | E -> true
    | N _ -> false

  let rec size_rec result = function
    | E -> result
    | N (a, _, _, b) -> size_rec (size_rec (1 + result) b) a

  let size = function tree -> size_rec 0 tree

  let rec member_min_bound x y = function
    | E -> Ord.compare x y = 0
    | N (a, z, _, b) ->
	match Ord.compare x z with
	  | k when k < 0 -> member_min_bound x y a
	  | _ -> member_min_bound x z b
	      
  let rec member' x = function
    | E -> false
    | N (a, y, _, b) ->
	match Ord.compare x y with
	  | n when n < 0 -> member' x a
	  | _ -> member_min_bound x y b
	      
  let member' = fun x set -> member' x set

  let rec member x = function
    | E -> false
    | N (a, y, _, b) ->
	match Ord.compare x y with
	  | n when n < 0 -> member x a
	  | n when n > 0 -> member x b
	  | _ -> true

  let rec find f = function
    | E -> raise Not_found
    | N (_, x, _, _) when f x -> x
    | N (a, _, _, b) -> (try find f a with Not_found -> find f b)

  let choose = function
    | E -> raise Empty
    | N (_, x, _, _) -> x

  let rec min_element = function
    | E -> raise Not_found
    | N (E, x, p, _) -> (x, p)
    | N (a, _, _, _) -> min_element a

  let rec max_element = function
    | E -> raise Not_found
    | N (_, x, p, E) -> (x, p)
    | N (_, _, _, b) -> max_element b

  let rec max_priority = function
    | E -> raise Empty
    | N (_, x, p, _) -> (x, p)

  let rec next x = function
    | E -> raise Not_found
    | N (l, v, p, r) ->
	match Ord.compare x v with
	  | n when n < 0 -> (try next x l with Not_found -> (v, p))
	  | n when n > 0 -> next x r
	  | _ -> min_element r

  let rec previous x = function
    | E -> raise Not_found
    | N (l, v, p, r) ->
	match Ord.compare x v with
	  | n when n < 0 -> previous x l
	  | n when n > 0 -> (try previous x r with Not_found -> (v, p))
	  | _ -> max_element l
	      
  let rec ceil x = function
    | E -> raise Not_found
    | N (l, v, p, r) ->
	match Ord.compare x v with
	  | n when n < 0 -> (try ceil x l with Not_found -> (v, p))
	  | n when n > 0 -> ceil x r
	  | _ -> (x, p)

  let rec floor x = function
    | E -> raise Not_found
    | N (l, v, p, r) ->
	match Ord.compare x v with
	  | n when n < 0 -> floor x l
	  | n when n > 0 -> (try floor x r with Not_found -> (v, p))
	  | _ -> (x, p)

  let rec rank x = function
    | E -> raise Not_found
    | N (a, y, _, b) ->
	match Ord.compare x y with
	  | n when n < 0 -> rank x a
	  | n when n > 0 -> 1 + size a + rank x b
	  | _ -> size a

  let rec get n = function
    | E -> raise (Out_of_bounds n)
    | N (a, x, p, b) -> 
	let k = size a in
	  if n < k then get n a
	  else if n > k then get (n - k - 1) b
	  else (x, p)

  let rec unsafe_set value priority m = function
    | E -> raise (Out_of_bounds m)
    | N (a, x, p, b) ->
	let k = size a in
	  if m < k then makeTree (unsafe_set value priority m a) x p b
	  else if m > k then 
	    makeTree a x p (unsafe_set value priority (m - k - 1) b)
	  else makeTree a value priority b	  

  let rec for_all f = function
    | E -> true
    | N (a, x, p, b) -> f x p && for_all f a && for_all f b

  let rec exists f = function
    | E -> false
    | N (a, x, p, b) -> f x p || exists f a || exists f b

  let rec iter f = function
    | E -> ()
    | N (a, x, p, b) -> iter f a ; f x p ; iter f b

  let rec fold_left f e = function
    | E -> e
    | N (a, x, p, b) -> fold_left f (f (fold_left f e a) x p) b

  let rec fold_right = fun f set e ->
    match set with
      | E -> e 
      | N (a, x, p, b) -> fold_right f a (f x p (fold_right f b e))

  let fold = fold_right

  let empty = E
  let mem = member
  let add = insert
  let singleton = injection
  let remove = delete
  let inter = intersection
  let diff = difference
  let cardinal = size
  let elements = to_list
  let min_elt = min_element    
  let max_elt = max_element

end


module Make (Key : OrderedType) (Priority : OrderedType) =
struct

  type t =  E | N of t * Key.t * Priority.t * t


  let makeTree = fun l x p r -> N (l, x, p, r)

  let rotate_L = fun l v p r ->
    match r with
      | N (rl, rv, rp, rr) -> makeTree (makeTree l v p rl) rv rp rr
      | _ -> makeTree l v p r
	  
  let rotate_R = fun l v p r ->
    match l with
      | N (ll, lv, lp, lr) -> makeTree ll lv lp (makeTree lr v p r)
      | _ -> makeTree l v p r

  let rec makeSTree = fun l v p r ->
    match (l, r) with
      | (E, E) -> makeTree l v p r
      | (E, N (rl, rv, rp, rr)) ->
	  (match Priority.compare rp p with
	     | n when n > 0 -> makeTree (makeSTree l v p rl) rv rp r
	     | _ -> makeTree l v p r
	  )
      | (N (ll, lv, lp, lr), E) ->
	  (match Priority.compare lp p with
	     | n when n > 0 -> makeTree ll lv lp (makeSTree lr v p r)
	     | _ -> makeTree l v p r
	  )
      | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	  match Priority.compare lp rp with
	    | n when n > 0 ->
		(match Priority.compare lp p with
		   | n when n > 0 -> makeTree ll lv lp (makeSTree lr v p r)
		   | _ -> makeTree l v p r
		)
	    | _ ->
		match Priority.compare rp p with
		  | n when n > 0 -> makeTree (makeSTree l v p rl) rv rp rr
		  | _ -> makeTree l v p r

  let rec appendS = fun l r ->
    match (l, r) with
      | (E, s) | (s, E) -> s
      | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	  match Priority.compare lp rp with
	    | n when n < 0 -> makeTree (appendS l rl) rv rp rr
	    | _ -> makeTree ll lv lp (appendS lr r)

  let injection = fun x p -> N (E, x, p, E)

  let rec partition_inf_sup pivot = function
    | E -> (E, None, E)
    | N (a, x, p, b) ->
	match Key.compare pivot x with
	  | n when n < 0 ->
	      let (inf, value, sup) = partition_inf_sup pivot a in
		(inf, value, makeTree sup x p b)
	  | n when n > 0 ->
	      let (inf, value, sup) = partition_inf_sup pivot b in
		(makeTree a x p inf, value, sup)
	  | _ -> (a, Some (x, p), b)

  let rec old_plus = fun l r ->
    match (l, r) with
      | (E, s) | (s, E) -> s
      | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	  match Priority.compare lp rp with
	    | n when n > 0 ->
		let (inf, _, sup) = partition_inf_sup lv r in
		  makeTree (old_plus inf ll) lv lp (old_plus sup lr)
	    | _ ->
		let (inf, _, sup) = partition_inf_sup rv l in
		  makeTree (old_plus inf rl) rv rp (old_plus sup rr)

  let rec filter_lt x = function
    | E -> E
    | N (a, y, p, b) ->
	match Key.compare x y with
	  | n when n < 0 -> filter_lt x a
	  | n when n > 0 -> makeTree a y p (filter_lt x b)
	  | _ -> a

  let rec filter_gt x = function
    | E -> E
    | N (a, y, p, b) ->
	match Key.compare x y with
	  | n when n < 0 -> makeTree (filter_gt x a) y p b
	  | n when n > 0 -> filter_gt x a
	  | _ -> b

  let rec cut = fun lo tree hi ->
    match tree with
      | N (_, x, _, b) when Key.compare x lo <= 0 -> cut lo b hi
      | N (a, x, _, _) when Key.compare x hi >= 0 -> cut lo a hi
      | _ -> tree

  let rec cut_hi = fun lo tree ->
    match tree with
      | N (_, x, _, b) when Key.compare x lo <= 0 -> cut_hi lo b
      | _ -> tree

  let rec cut_lo = fun tree hi ->
    match tree with
      | N (a, x, _, _) when Key.compare x hi >= 0 -> cut_lo a hi
      | _ -> tree

  let rec uni = fun lo a b hi ->
    match (a, b) with
      | (E, E) -> E
      | (N _ as node, E) | (E, (N _ as node)) -> 
	  (match cut lo node hi with
	     | E -> E
	     | N (l, x, p, r) -> makeTree (filter_gt lo l) x p (filter_lt hi r)
	  )
      | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	  match Priority.compare lp rp with
	    | n when n > 0 -> makeTree 
		(uni lo ll (cut lo b lv) lv) lv lp (uni lv lr (cut lv b hi) hi)
	    | _ -> makeTree 
		(uni lo rl (cut lo a rv) rv) rv rp (uni rv rr (cut rv a hi) hi)

  let rec uni_lo = fun a b hi ->
    match (a, b) with
      | (E, E) -> E
      | (E, (N _ as node)) | (N _ as node, E) ->
	  (match cut_lo node hi with
	     | E -> E
	     | N (l, x, p, r) -> makeTree l x p (filter_lt hi r)
	  )
      | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	  match Priority.compare lp rp with
	    | n when n > 0 -> makeTree 
		(uni_lo ll (cut_lo b lv) lv) lv lp (uni lv lr (cut lv b hi) hi)
	    | _ -> makeTree 
		(uni_lo rl (cut_lo a rv) rv) rv rp (uni rv rr (cut rv a hi) hi)

  let rec uni_hi = fun lo a b ->
    match (a, b) with
      | (E, E) -> E
      | (E, (N _ as node)) | (N _ as node, E) ->
	  (match cut_hi lo node with
	     | E -> E
	     | N (l, x, p, r) -> makeTree (filter_gt lo l) x p r
	  )
      | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	  match Priority.compare lp rp with
	    | n when n > 0 -> makeTree 
		(uni lo ll (cut lo b lv) lv) lv lp (uni_hi lv lr (cut_hi lv b))
	    | _ -> makeTree 
		(uni lo rl (cut lo a rv) rv) rv rp (uni_hi rv rr (cut_hi rv a))

  let plus = fun a b ->
    match (a, b) with
      | (_, E) -> a
      | (E, _) -> b
      | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	  match Priority.compare lp rp with
	    | n when n > 0 -> makeTree
		(uni_lo ll (cut_lo b lv) lv) lv lp (uni_hi lv lr (cut_hi lv b))
	    | _ -> makeTree 
		(uni_lo rl (cut_lo a rv) rv) rv rp (uni_hi rv rr (cut_hi rv a))

  let zero = function () -> E

  let rec insert x priority = function
    | E -> N (E, x, priority, E)
    | N (a, y, p, b) as node ->
	match Priority.compare priority p with
	  | n when n > 0 -> 
	      let (inf, _, sup) = partition_inf_sup x node in 
		makeTree inf x priority sup
	  | _ ->
	      match Key.compare x y with
		| n when n < 0 -> makeTree (insert x priority a) y p b
		| n when n > 0 -> makeTree a y p (insert x priority b)
		| _ -> makeSTree a x priority b

  let rec insert' x priority = function
    | E -> N (E, x, priority, E)
    | N (a, y, p, b) ->
	match Key.compare x y with
	  | n when n < 0 ->
	      if Priority.compare priority p > 0 
	      then rotate_R (insert' x priority a) y p b
	      else makeTree (insert' x priority a) y p b
	  | n when n > 0 ->
	      if Priority.compare priority p > 0 
	      then rotate_L a y p (insert' x priority b)
	      else makeTree a y p (insert' x priority b)
	  | _ -> makeSTree a x priority b

  let rec from_list_rec result = function
    | [] -> result
    | (x, p) :: tail -> from_list_rec (insert x p result) tail

  let from_list = function list -> from_list_rec E list

  let rec to_list = function
    | E -> []
    | N (a, x, p, b) -> to_list a @ (x, p) :: to_list b

  let rec to_list_by_priority = function
    | E -> []
    | N (a, x, p, b) -> (x, p) :: to_list_by_priority (plus a b)

  let map f = function set -> 
    from_list (List.map (fun (x, y) -> (f x, y)) (to_list set))

  let union = plus

  let rec intersection = fun a b ->
    match (a, b) with
      | (E, _) | (_, E) -> E
      | (N (l, x, p, r), _) ->
	  match partition_inf_sup x b with
	    | (l', None, r') -> appendS (intersection l l') (intersection r r')
	    | (l', Some (_, p'), r') -> 
		makeSTree (intersection l l') x (max p p') (intersection r r')

  let rec difference = fun a b ->
    match (a, b) with
      | (E, _) -> E
      | (_, E) -> a
      | (N (l, x, p, r), _) -> 
	  match partition_inf_sup x b with
	    | (l', None, r') -> 
		makeSTree (difference l l') x p (difference r r')
	    | (l', Some _, r') -> appendS (difference l l') (difference r r')
		
  let rec symetric_difference = fun a b ->
    match (a, b) with
      | (E, s) | (s, E) -> s
      | (N (l, x, p, r), _) ->
	  match partition_inf_sup x b with
	    | (l', None, r') -> makeSTree (symetric_difference l l') x p
		(symetric_difference r r')
	    | (l', Some _, r') -> 
		appendS (symetric_difference l l') (symetric_difference r r')

  let rec delete_raise x = function
    | E -> raise Not_found
    | N (a, y, p, b) ->
	match Key.compare x y with
	  | n when n < 0 -> makeTree (delete_raise x a) y p b
	  | n when n > 0 -> makeTree a y p (delete_raise x b)
	  | _ -> plus a b

  let delete = fun x tree ->
    try
      delete_raise x tree
    with Not_found -> tree

  let rec delete_min = function
    | E -> E
    | N (E, _, _, b) -> b
    | N (a, _, _, _) -> delete_min a

  let rec delete_max = function
    | E -> E
    | N (a, _, _, E) -> a
    | N (_, _, _, b) -> delete_min b

  let rec delete_max_priority = function
    | E -> E
    | N (a, _, _, b) -> appendS a b

  let rec find_min_priority ((_, pm) as min) = function
    | E -> min
    | N (E, x, p, E) -> 
	(match Priority.compare p pm with
	   | n when n < 0 -> (x, p)
	   | _ -> min
	)
    | N (a, x, p, b) -> find_min_priority (find_min_priority min a) b

  let rec find_min_priority_leaf = function
    | E -> raise Empty
    | N (E, x, p, E) -> (x, p)
    | N (E, _, _, b) -> find_min_priority_leaf b
    | N (a, _, _, b) -> find_min_priority (find_min_priority_leaf a) b

  let min_priority = function
    | E -> raise Empty
    | N (E, x, p, E) -> (x, p)
    | N (E, _, _, b) -> find_min_priority_leaf b
    | N (a, _, _, b) -> find_min_priority (find_min_priority_leaf a) b

  let delete_min_priority = function tree ->
    let (x, _) = min_priority tree in 
      delete x tree

  let rec delete_list = fun list set ->
    match list with
      | [] -> set
      | head :: tail -> delete_list tail (delete head set)

  let rec filter_rec f result = function
    | E -> result
    | N (a, x, p, b) ->
	filter_rec f (filter_rec f (
			if f x p then insert x p result else result) a) b

  let filter f = function set -> filter_rec f E set

  let rec range min max = function
    | E -> E
    | N (l, v, p, r) ->
	if Key.compare v min < 0 then range min max r
	else if Key.compare max v < 0 then range min max l
	else makeSTree (filter_gt min l) v p (filter_lt max r)

  let rec min_view = function
    | N (E, x, p, b) -> ((x, p), b)
    | N (a, x, p, b) -> let (min, left) = min_view a in 
	(min, makeTree left x p b)
    | _ -> raise Empty
	
  let rec max_view = function
    | N (a, x, p, E) -> (a, (x, p))
    | N (a, x, p, b) -> let (right, max) = max_view b in 
	(makeTree a x p right, max)
    | _ -> raise Empty

  let max_priority_view = function
    | N (a, x, p, b) -> ((x, p), plus a b)
    | _ -> raise Empty

  let rec min_priority_view = function
    | E -> raise Empty
    | N (E, x, p, _) -> (E, (x, p))
    | N (a, x, p, E) -> 
	let (a', y) = min_priority_view a in (makeTree a' x p E, y)
    | N (a, x, p, b) ->
	let (_, pa) = min_priority a and (_, pb) = min_priority b in
	  match Priority.compare pa pb with
	    | n when n > 0 ->
		let (b', min) = min_priority_view b in (makeTree a x p b', min)
	    | _ ->
		let (a', min) = min_priority_view a in (makeTree a' x p b, min)

  let rec partition_rec f (trueS, falseS) = function
    | E -> (trueS, falseS)
    | N (a, x, p, b) -> 
	if f x p
	then partition_rec f (partition_rec f (insert x p trueS, falseS) a) b
	else partition_rec f (partition_rec f (trueS, insert x p falseS) a) b

  let partition = fun f set -> partition_rec f (E, E) set

  let is_empty = function
    | E -> true
    | N _ -> false

  let rec size_rec result = function
    | E -> result
    | N (a, _, _, b) -> size_rec (size_rec (1 + result) b) a

  let size = function tree -> size_rec 0 tree

  let rec member_min_bound x y = function
    | E -> Key.compare x y = 0
    | N (a, z, _, b) ->
	match Key.compare x z with
	  | k when k < 0 -> member_min_bound x y a
	  | _ -> member_min_bound x z b
	      
  let rec member' x = function
    | E -> false
    | N (a, y, _, b) ->
	match Key.compare x y with
	  | n when n < 0 -> member' x a
	  | _ -> member_min_bound x y b
	      
  let member' = fun x set -> member' x set

  let rec member x = function
    | E -> false
    | N (a, y, _, b) ->
	match Key.compare x y with
	  | n when n < 0 -> member x a
	  | n when n > 0 -> member x b
	  | _ -> true

  let rec find f = function
    | E -> raise Not_found
    | N (_, x, _, _) when f x -> x
    | N (a, _, _, b) -> (try find f a with Not_found -> find f b)

  let choose = function
    | E -> raise Empty
    | N (_, x, _, _) -> x

  let rec min_element = function
    | E -> raise Not_found
    | N (E, x, p, _) -> (x, p)
    | N (a, _, _, _) -> min_element a

  let rec max_element = function
    | E -> raise Not_found
    | N (_, x, p, E) -> (x, p)
    | N (_, _, _, b) -> max_element b

  let rec max_priority = function
    | E -> raise Empty
    | N (_, x, p, _) -> (x, p)

  let rec next x = function
    | E -> raise Not_found
    | N (l, v, p, r) ->
	match Key.compare x v with
	  | n when n < 0 -> (try next x l with Not_found -> (v, p))
	  | n when n > 0 -> next x r
	  | _ -> min_element r (* raises [Not_found] *)

  let rec previous x = function
    | E -> raise Not_found
    | N (l, v, p, r) ->
	match Key.compare x v with
	  | n when n < 0 -> previous x l
	  | n when n > 0 -> (try previous x r with Not_found -> (v, p))
	  | _ -> max_element l (* raises [Not_found] *)
	      
  let rec ceil x = function
    | E -> raise Not_found
    | N (l, v, p, r) ->
	match Key.compare x v with
	  | n when n < 0 -> (try ceil x l with Not_found -> (v, p))
	  | n when n > 0 -> ceil x r
	  | _ -> (x, p)

  let rec floor x = function
    | E -> raise Not_found
    | N (l, v, p, r) ->
	match Key.compare x v with
	  | n when n < 0 -> floor x l
	  | n when n > 0 -> (try floor x r with Not_found -> (v, p))
	  | _ -> (x, p)

  let rec rank x = function
    | E -> raise Not_found
    | N (a, y, _, b) ->
	match Key.compare x y with
	  | n when n < 0 -> rank x a
	  | n when n > 0 -> 1 + size a + rank x b
	  | _ -> size a

  let rec get n = function
    | E -> raise (Out_of_bounds n)
    | N (a, x, p, b) -> 
	let k = size a in
	  if n < k then get n a
	  else if n > k then get (n - k - 1) b
	  else (x, p)

  let rec unsafe_set value priority m = function
    | E -> raise (Out_of_bounds m)
    | N (a, x, p, b) ->
	let k = size a in
	  if m < k then makeTree (unsafe_set value priority m a) x p b
	  else if m > k then 
	    makeTree a x p (unsafe_set value priority (m - k - 1) b)
	  else makeTree a value priority b	  

  let rec for_all f = function
    | E -> true
    | N (a, x, p, b) -> f x p && for_all f a && for_all f b

  let rec exists f = function
    | E -> false
    | N (a, x, p, b) -> f x p || exists f a || exists f b

  let rec iter f = function
    | E -> ()
    | N (a, x, p, b) -> iter f a ; f x p ; iter f b

  let rec fold_left f e = function
    | E -> e
    | N (a, x, p, b) -> fold_left f (f (fold_left f e a) x p) b

  let rec fold_right = fun f set e ->
    match set with
      | E -> e 
      | N (a, x, p, b) -> fold_right f a (f x p (fold_right f b e))

  let fold = fold_right

  let empty = E
  let mem = member
  let add = insert
  let singleton = injection
  let remove = delete
  let inter = intersection
  let diff = difference
  let cardinal = size
  let elements = to_list
  let min_elt = min_element    
  let max_elt = max_element

end
