(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Priority queue over a binomial heap 
 * binary trees
 * dense number representation
 * 
 * File de priority bases sur les tas binomiaux 
 * arbres binaires
 * systme numrique dense
 * 
 * Filas de prioridad a partir de montones binomiales 
 * rboles binrios
 * sistema numerico denso
 * 
 *)

let moduleName = "BinaryBinomialHeapPQ"

open BairePervasives

type ('a, 'b) tree =  E | N of ('a, 'b) tree * 'a * 'b * ('a, 'b) tree
type ('a, 'b) pennant = Pennant of 'a * 'b * ('a, 'b) tree
type ('a, 'b) digit = Zero | One of ('a, 'b) pennant
type ('a, 'b) priorityQueue = ('a, 'b) digit list


(** Utilitaires **)
(** Utils **)
(** tiles **)


let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type

let injection_pennant = fun x v -> Pennant (x, v, E)

let link = fun (Pennant (x1, v1, t1)) (Pennant (x2, v2, t2)) ->
  match compare x1 x2 with
    | n when n < 0 -> Pennant (x1, v1, N (t2, x2, v2, t1))
    | _ -> Pennant (x2, v2, N (t1, x1, v1, t2))

let rec insert_pennant t = function
  | [] -> [One t]
  | Zero :: tail -> One t :: tail
  | One t' :: tail -> Zero :: insert_pennant (link t t') tail

let root = function Pennant (x, _, _) -> x

let key = function Pennant (_, k, _) -> k

let rec remove_min_pennant = function
  | [] -> raise Empty
  | [One t] -> (t, [])
  | Zero :: tail -> let (t', r) = remove_min_pennant tail in (t', Zero :: r)
  | One t :: tail -> let (t', r) = remove_min_pennant tail in
      match compare (root t) (root t') with
	| n when n > 0 -> (t', One t :: r)
	| _ -> (t, Zero :: tail)

let rec dismantle_tree = function
  | E -> []
  | N (a, x, v, b) -> Pennant (x, v, a) :: dismantle_tree b


(** Categorial operations **)
(** Oprations catgorielles **)
(** Operaciones categoriales **)


let injection = fun x v -> [One (injection_pennant x v)]

let rec plus = fun a b ->
  match (a, b) with
    | (_, []) -> a
    | ([], _) -> b
    | (Zero :: succ1, t :: succ2) -> t :: plus succ1 succ2
    | (t :: succ1, Zero :: succ2) -> t :: plus succ1 succ2
    | (One t1 :: succ1, One t2 :: succ2) -> 
	Zero :: insert_pennant (link t1 t2) (plus succ1 succ2)

let zero = function () -> []

let rec from_list_rec result = function
  | x :: y :: tail -> from_list_rec (plus x y :: result) tail
  | [x] -> from_list_rec (x :: result) []
  | [] -> 
      match result with
	| [x] -> x
	| _ -> from_list_rec [] result

let from_list = function liste -> from_list_rec [] 
  (List.map (fun (x, y) -> injection x y) liste)

let min_view = function heap ->
  let (Pennant (x, v, t), plist) = remove_min_pennant heap in
    (x, v, plus (List.rev_map (fun p -> One p) (dismantle_tree t)) plist) 

let rec to_list = function
  | [] -> []
  | heap -> let (x, v, h) = min_view heap in (x, v) :: to_list h

let map f = function heap -> from_list (List.map f (to_list heap))


(** Acces **)
(** Accs **)
(** Acceso **)


let insert x v = function heap -> insert_pennant (injection_pennant x v) heap

let rec insert_list_rec result = function
  | [] -> result
  | (x, v) :: tail -> insert_list_rec (insert x v result) tail

let insert_list = function list -> insert_list_rec [] list

let delete_min = function heap ->
  let (Pennant (_, _, t), plist) = remove_min_pennant heap in
    plus (List.rev_map (fun p -> One p) (dismantle_tree t)) plist


(** View **)
(** Vues **)
(** Vistas **)


let partition f = function heap ->
  let (a, b) = List.partition (fun (x, y) -> f x y) (to_list heap) in
    (from_list a, from_list b)


(** Observers **)
(** Observateurs **)
(** Observadores **)


let is_empty = function
  | [] -> true
  | _ -> false

let rec size_rec k total = function
  | [] -> total
  | Zero :: tail -> size_rec (2 * k) total tail
  | One _ :: tail -> size_rec (2 * k) (total + k) tail

let size = function heap -> size_rec 1 0 heap

(* a node is always inferior to its left sub_tree *)
(* un noeud est toujours infrieur  son arbre gauche *)
(* un nudo es siempre inferior a su subarbol izquierdo *)

let rec member_priority_tree x = function 
  | E -> false
  | N (a, y, _, b) ->
      match compare x y with
	| n when n < 0 -> member_priority_tree x b
	| 0 -> true
	| _ -> member_priority_tree x a || member_priority_tree x b

let rec member_priority x = function
  | [] -> false
  | Zero :: tail -> member_priority x tail
  | One (Pennant (y, _, t)) :: tail -> 
      match compare x y with
	| 0 -> true
	| _ -> member_priority_tree x t || member_priority x tail

let rec count_priority_tree x total = function 
  | E -> total
  | N (a, y, _, b) ->
      match compare x y with
	| n when n < 0 -> count_priority_tree x total b
	| 0 -> count_priority_tree x (count_priority_tree x (total + 1) a) b
	| _ -> count_priority_tree x (count_priority_tree x total a) b

let rec count_priority_rec x total = function
  | [] -> total
  | Zero :: tail -> count_priority_rec x total tail
  | One (Pennant (y, _, t)) :: tail -> 
      match compare x y with
	| 0 -> count_priority_rec x (count_priority_tree x (1 + total) t) tail
	| _ -> count_priority_rec x (count_priority_tree x total t) tail

let count_priority = fun x heap -> count_priority_rec x 0 heap

let rec min_rec min v = function
  | [] -> (min, v)
  | Zero :: tail -> min_rec min v tail
  | One t :: tail ->
      match compare (root t) min with
	| n when n < 0 -> min_rec (root t) (key t) tail
	| _ -> min_rec min v tail

let rec min_element = function
  | [] -> raise Empty
  | Zero :: tail -> min_element tail
  | One t :: tail -> min_rec (root t) (key t) tail

let rec exists_tree f = function
  | E -> false
  | N (a, x, v, b) -> f x v || exists_tree f a || exists_tree f b

let rec exists f = function
  | [] -> false
  | Zero :: tail -> exists f tail
  | One (Pennant (x, v, t)) :: tail -> 
      f x v || exists_tree f t || exists f tail

let rec for_all_tree f = function
  | E -> true
  | N (a, x, v, b) -> f x v && for_all_tree f a && for_all_tree f b

let rec for_all f = function
  | [] -> true
  | Zero :: tail -> for_all f tail
  | One (Pennant (x, v, t)) :: tail -> 
      f x v && for_all_tree f t && for_all f tail

let filter f = function heap -> 
  from_list (List.filter (fun (x, y) -> f x y) (to_list heap))


(** Traversals **)
(** Parcours **)
(** Recorridos **)


let rec iter_tree f = function
  | E -> ()
  | N (a, x, v, b) -> iter_tree f a ; f x v ; iter_tree f b

let rec iter f = function
  | [] -> ()
  | Zero :: tail -> iter f tail
  | One (Pennant (x, v, t)) :: tail -> f x v ; iter_tree f t ; iter f tail

let rec fold_left_tree f e = function
  | E -> e
  | N (a, x, v, b) -> fold_left_tree f (f (fold_left_tree f e a) x v) b

let rec fold_left f e = function
  | [] -> e
  | Zero :: tail -> fold_left f e tail
  | One (Pennant (x, v, t)) :: tail -> 
      fold_left f (fold_left_tree f (f e x v) t) tail

let rec fold_right_tree = fun f heap e ->
  match heap with
    | E -> e 
    | N (a, x, v, b) -> fold_right_tree f a (f x v (fold_right_tree f b e))

let rec fold_right = fun f heap e ->
  match heap with
    | [] -> e
    | Zero :: tail -> fold_right f tail e
    | One (Pennant (x, v, t)) :: tail -> 
	f x v (fold_right_tree f t (fold_right f tail e))

let fold = fold_right

