(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Priority queues over binomial heaps
 * multiway trees
 * dense number representation
 *
 * Files de priorit bases sur les tas binomiaux
 * arbres n-aires
 * systme numrique dense
 * 
 * Filas de prioridad a partir de montones binomiales
 * rboles n-arios
 * sistema numrico denso
 * 
 *)

open BairePervasives

let moduleName = "BinomialHeapPQ"

type ('a, 'b) tree = Node of 'a * 'b * ('a, 'b) tree list
type ('a, 'b) digit = Zero | One of ('a, 'b) tree
type ('a, 'b) priorityQueue = ('a, 'b) digit list


(** Utilitaires **)
(** Utils **)
(** tiles **)


let injection_tree = fun x v -> Node (x, v, [])

let link = fun (Node (x1, v1, c1) as t1) (Node (x2, v2, c2) as t2) ->
  match compare x1 x2 with
    | n when n > 0 -> Node (x2, v2, t1 :: c2)
    | _ -> Node (x1, v1, t2 :: c1)

let rec insert_tree t = function
  | [] -> [One t]
  | Zero :: tail -> One t :: tail
  | One t' :: tail -> Zero :: insert_tree (link t t') tail

let root = function Node (x, _, _) -> x

let key = function Node (_, v, _) -> v

let rec remove_min_tree = function
  | [] -> raise Empty
  | [One t] -> (t, [])
  | Zero :: tail -> let (t', r) = remove_min_tree tail in (t', Zero :: r)
  | One t :: tail -> let (t', r) = remove_min_tree tail in
      match compare (root t) (root t') with
	| n when n > 0 -> (t', One t :: r)
	| _ -> (t, Zero :: tail)


(** Categorial operations **)
(** Oprations catgorielles **)
(** Operaciones categoriales **)


let injection = fun x v -> [One (injection_tree x v)]

let rec plus = fun a b ->
  match (a, b) with
    | (_, []) -> a
    | ([], _) -> b
    | (Zero :: succ1, t :: succ2) -> t :: plus succ1 succ2
    | (t :: succ1, Zero :: succ2) -> t :: plus succ1 succ2
    | (One t1 :: succ1, One t2 :: succ2) -> 
	Zero :: insert_tree (link t1 t2) (plus succ1 succ2)

let zero = function () -> []

let rec from_list_rec result = function
  | x :: y :: tail -> from_list_rec (plus x y :: result) tail
  | [x] -> from_list_rec (x :: result) []
  | [] -> 
      match result with
	| [x] -> x
	| _ -> from_list_rec [] result

let from_list = function liste -> 
  from_list_rec [] (List.map (fun (x, v) -> injection x v) liste)

let min_view = function heap -> 
  let (Node (x, v, c), h) = remove_min_tree heap in 
    (x, v, plus h (List.map (function t -> One t) (List.rev c)))

let rec to_list = function
  | [] -> []
  | heap -> let (x, v, h) = min_view heap in (x, v) :: to_list h

let map f = function heap -> from_list (List.map f (to_list heap))


(** Acces **)
(** Accs **)
(** Acceso **)


let insert x v = function heap -> insert_tree (injection_tree x v) heap

let rec insert_list_rec result = function
  | [] -> result
  | (x, v) :: tail -> insert_list_rec (insert x v result) tail

let insert_list = function list -> insert_list_rec [] list

let delete_min = function heap ->
  let (Node (_, _, c), r) = remove_min_tree heap in 
    plus r (List.map (function t -> One t) (List.rev c))

let filter f = function heap -> 
  from_list (List.filter (fun (x, y) -> f x y) (to_list heap))


(** View **)
(** Vues **)
(** Vistas **)


let partition f = function heap ->
  let (a, b) = List.partition (fun (x, v) -> f x v) (to_list heap) in
    (from_list a, from_list b)


(** Observers **)
(** Observateurs **)
(** Observadores **)


let is_empty = function
  | [] -> true
  | _ -> false

let rec size_rec k total = function
  | [] -> total
  | Zero :: tail -> size_rec (2 * k) total tail
  | One _ :: tail -> size_rec (2 * k) (total + k) tail

let size = function heap -> size_rec 1 0 heap

let rec member_priority_tree x = function Node (y, v, succ) ->
  match compare x y with
    | n when n < 0 -> false
    | 0 -> true
    | _ -> List.exists (member_priority_tree x) succ

let rec member_priority x = function
  | [] -> false
  | Zero :: tail -> member_priority x tail
  | One t :: tail -> member_priority_tree x t || member_priority x tail

let rec count_priority_tree x total = function Node (y, _, succ) ->
  match compare x y with
    | n when n < 0 -> total
    | 0 -> List.fold_left (count_priority_tree x) (1 + total) succ
    | _ -> List.fold_left (count_priority_tree x) total succ

let rec count_priority_rec x total = function
  | [] -> total
  | Zero :: tail -> count_priority_rec x total tail
  | One t :: tail -> count_priority_rec x (count_priority_tree x total t) tail

let count_priority = fun x heap -> count_priority_rec x 0 heap
	    
let rec min_rec min v = function
  | [] -> (min, v)
  | Zero :: tail -> min_rec min v tail
  | One t :: tail ->
      match compare (root t) min with
	| n when n < 0 -> min_rec (root t) (key t) tail
	| _ -> min_rec min v tail

let rec min_element = function
  | [] -> raise Empty
  | Zero :: tail -> min_element tail
  | One t :: tail -> min_rec (root t) (key t) tail

let rec exists_tree f = function Node (x, v, succ) ->
  f x v || List.exists (exists_tree f) succ

let rec exists f = function
  | [] -> false
  | Zero :: tail -> exists f tail
  | One t :: tail -> exists_tree f t || exists f tail

let rec for_all_tree f = function Node (x, v, succ) ->
  f x v && List.for_all (for_all_tree f) succ

let rec for_all f = function
  | [] -> true
  | Zero :: tail -> for_all f tail
  | One t :: tail -> for_all_tree f t && for_all f tail


(** Traversals **)
(** Parcours **)
(** Recorridos **)


let rec iter_tree f = function Node (x, v, succ) -> 
  f x v ; List.iter (iter_tree f) succ

let rec iter f = function
  | [] -> ()
  | Zero :: tail -> iter f tail
  | One t :: tail -> iter_tree f t ; iter f tail

let rec fold_right_tree = fun f (Node (x, v, succ)) e ->
  f x v (List.fold_right (fold_right_tree f) succ e)

let rec fold_right = fun f tree e ->
  match tree with
    | [] -> e
    | Zero :: tail -> fold_right f tail e
    | One t :: tail -> fold_right_tree f t (fold_right f tail e)

let rec fold_left_tree f e = function Node (x, v, succ) -> 
  List.fold_left (fold_left_tree f) (f e x v) succ

let rec fold_left f e = function
  | [] -> e
  | Zero :: tail -> fold_left f e tail
  | One t :: tail -> fold_left f (fold_left_tree f e t) tail

let fold = fold_right

