(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Cartesian (min) heap
 * Tas cartsiens (min)
 * Montones cartesianos (min)
 *
 *)

open BairePervasives

type 'a cartesian = E | N of 'a cartesian * 'a * int * 'a  cartesian

(* Nouvelles oprations :
 *
 * val sub_one : 'a -> 'a cartesian -> 'a cartesian
 * 
 * soustrait une unit  la priorit associe  l'lment donn en paramtre
 * et rordonne l'arbre en consquence
 *)

(** Utilitaires **)
(** Utils **)
(** tiles **)


let moduleName = "CartesianHeap"

let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type

let makeTree = fun l x p r -> N (l, x, p, r)

let rotate_L = fun l v p r ->
  match r with
    | N (rl, rv, rp, rr) -> makeTree (makeTree l v p rl) rv rp rr
    | _ -> makeTree l v p r
	
let rotate_R = fun l v p r ->
  match l with
    | N (ll, lv, lp, lr) -> makeTree ll lv lp (makeTree lr v p r)
    | _ -> makeTree l v p r

let rec makeSTree = fun l v p r ->
  match (l, r) with
    | (E, E) -> makeTree E v p E
    | (E, N (rl, rv, rp, rr)) ->
	if rp < p then makeTree (makeSTree l v p rl) rv rp r
	else makeTree l v p r
    | (N (ll, lv, lp, lr), E) ->
	if lp < p then makeTree ll lv lp (makeSTree lr v p r)
	else makeTree l v p r
    | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	if lp < rp then
	  (if lp < p then makeTree ll lv lp (makeSTree lr v p r)
	   else makeTree l v p r)
	else
	  (if rp < p then makeTree (makeSTree l v p rl) rv rp rr
	   else makeTree l v p r)

let rec appendS = fun l r ->
  match (l, r) with
    | (E, s) | (s, E) -> s
    | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	match compare lp rp with
	  | n when n > 0 -> makeTree (appendS l rl) rv rp rr
	  | _ -> makeTree ll lv lp (appendS lr r)


(** Oprations catgorielles **)
(** Categorial operations **)
(** Operaciones categoriales **)


let injection = fun x p -> N (E, x, p, E)

let rec partition_inf_sup pivot = function
  | E -> (E, None, E)
  | N (l, v, p, r) ->
      match compare pivot v with
	| n when n < 0 ->
	    let (inf, value, sup) = partition_inf_sup pivot l in
	      (inf, value, makeTree sup v p r)
	| n when n > 0 ->
	    let (inf, value, sup) = partition_inf_sup pivot r in
	      (makeTree l v p inf, value, sup)
	| _ -> (l, Some (v, p), r)

let rec old_plus = fun l r ->
  match (l, r) with
    | (E, s) | (s, E) -> s
    | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	match compare lp rp with
	  | n when n > 0 ->
	      let (inf, _, sup) = partition_inf_sup lv r in
		makeTree (old_plus inf ll) lv lp (old_plus sup lr)
	  | _ ->
	      let (inf, _, sup) = partition_inf_sup rv l in
		makeTree (old_plus inf rl) rv rp (old_plus sup rr)

let rec filter_lt x = function
  | E -> E
  | N (l, v, p, r) ->
      match compare x v with
	| n when n < 0 -> filter_lt x l
	| n when n > 0 -> makeTree l v p (filter_lt x r)
	| _ -> l

let rec filter_gt x = function
  | E -> E
  | N (l, v, p, r) ->
      match compare x v with
	| n when n < 0 -> makeTree (filter_gt x l) v p r
	| n when n > 0 -> filter_gt x l
	| _ -> r

let rec cut = fun lo tree hi ->
  match tree with
    | N (_, v, _, r) when compare v lo <= 0 -> cut lo r hi
    | N (l, v, _, _) when compare v hi >= 0 -> cut lo l hi
    | _ -> tree

let rec cut_hi = fun lo tree ->
  match tree with
    | N (_, v, _, r) when compare v lo <= 0 -> cut_hi lo r
    | _ -> tree

let rec cut_lo = fun tree hi ->
  match tree with
    | N (l, v, _, _) when compare v hi >= 0 -> cut_lo l hi
    | _ -> tree

let rec uni = fun lo a b hi ->
  match (a, b) with
    | (E, E) -> E
    | (N _ as node, E) | (E, (N _ as node)) -> 
	(match cut lo node hi with
	   | E -> E
	   | N (l, v, p, r) -> makeTree (filter_gt lo l) v p (filter_lt hi r)
	)
    | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	match compare lp rp with
	  | n when n < 0 -> makeTree 
	      (uni lo ll (cut lo b lv) lv) lv lp (uni lv lr (cut lv b hi) hi)
	  | _ -> makeTree 
	      (uni lo rl (cut lo a rv) rv) rv rp (uni rv rr (cut rv a hi) hi)

let rec uni_lo = fun a b hi ->
  match (a, b) with
    | (E, E) -> E
    | (E, (N _ as node)) | (N _ as node, E) ->
	(match cut_lo node hi with
	   | E -> E
	   | N (l, v, p, r) -> makeTree l v p (filter_lt hi r)
	)
    | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	match compare lp rp with
	  | n when n < 0 -> makeTree 
	      (uni_lo ll (cut_lo b lv) lv) lv lp (uni lv lr (cut lv b hi) hi)
	  | _ -> makeTree 
	      (uni_lo rl (cut_lo a rv) rv) rv rp (uni rv rr (cut rv a hi) hi)

let rec uni_hi = fun lo a b ->
  match (a, b) with
    | (E, E) -> E
    | (E, (N _ as node)) | (N _ as node, E) ->
	(match cut_hi lo node with
	   | E -> E
	   | N (l, x, p, r) -> makeTree (filter_gt lo l) x p r
	)
    | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	match compare lp rp with
	  | n when n < 0 -> makeTree 
	      (uni lo ll (cut lo b lv) lv) lv lp (uni_hi lv lr (cut_hi lv b))
	  | _ -> makeTree 
	      (uni lo rl (cut lo a rv) rv) rv rp (uni_hi rv rr (cut_hi rv a))

let plus = fun a b ->
  match (a, b) with
    | (_, E) -> a
    | (E, _) -> b
    | (N (ll, lv, lp, lr), N (rl, rv, rp, rr)) ->
	match compare lp rp with
	  | n when n < 0 -> makeTree
	      (uni_lo ll (cut_lo b lv) lv) lv lp (uni_hi lv lr (cut_hi lv b))
	  | _ -> makeTree 
	      (uni_lo rl (cut_lo a rv) rv) rv rp (uni_hi rv rr (cut_hi rv a))

let zero = function () -> E

let rec join_rec result = function
  | E -> result
  | N (l, v, p, r) -> join_rec (join_rec (plus v result) l) r 

let join = function tree -> join_rec E tree

let rec insert x priority = function
  | E -> injection x priority
  | N (l, v, p, r) as node ->
      match compare priority p with
	| n when n < 0 -> 
	    let (inf, _, sup) = partition_inf_sup x node in 
	      makeTree inf x priority sup
	| _ ->
	    match compare x v with
	      | n when n < 0 -> makeTree (insert x priority l) v p r
	      | n when n > 0 -> makeTree l v p (insert x priority r)
	      | _ -> makeSTree l x priority r

let rec insert' x priority = function
  | E -> injection x priority
  | N (l, v, p, r) ->
      match compare x v with
	| n when n < 0 ->
	    if priority < p then rotate_R (insert' x priority l) v p r
	    else makeTree (insert' x priority l) v p r
	| n when n > 0 ->
	    if priority < p then rotate_L l v p (insert' x priority r)
	    else makeTree l v p (insert' x priority r)
	| _ -> makeSTree l x priority r

let rec from_list_rec result = function
  | [] -> result
  | (v, p) :: tail -> from_list_rec (insert v p result) tail

let from_list = function list -> from_list_rec E list

let rec to_list = function
  | E -> []
  | N (l, v, p, r) -> to_list l @ (v, p) :: to_list r

let rec to_list_by_priority = function
  | E -> []
  | N (l, v, p, r) -> (v, p) :: to_list_by_priority (plus l r)

let map f = function set -> 
  from_list (List.map (fun (x, y) -> (f x, y)) (to_list set))


(** Oprations binaires **)
(** Binary operaations **)
(** Opraciones binrias **)


let union = plus

let rec intersection = fun a b ->
  match (a, b) with
    | (E, _) | (_, E) -> E
    | (N (l, v, p, r), _) ->
	match partition_inf_sup v b with
	  | (l', None, r') -> appendS (intersection l l') (intersection r r')
	  | (l', Some (_, p'), r') -> 
	      makeSTree (intersection l l') v (max p p') (intersection r r')

let rec difference = fun a b ->
  match (a, b) with
    | (E, _) -> E
    | (_, E) -> a
    | (N (l, v, p, r), _) -> 
	match partition_inf_sup v b with
	  | (l', None, r') -> makeSTree (difference l l') v p (difference r r')
	  | (l', Some _, r') -> appendS (difference l l') (difference r r')
	      
let rec symetric_difference = fun a b ->
  match (a, b) with
    | (E, s) | (s, E) -> s
    | (N (l, v, p, r), _) ->
	match partition_inf_sup v b with
	  | (l', None, r') -> makeSTree (symetric_difference l l') v p
	      (symetric_difference r r')
	  | (l', Some _, r') -> 
	      appendS (symetric_difference l l') (symetric_difference r r')


(** Accs **)
(** Acces **)
(** Acceso **)


let rec delete_raise x = function
  | E -> raise Not_found
  | N (l, v, p, r) ->
      match compare x v with
	| n when n < 0 -> makeTree (delete_raise x l) v p r
	| n when n > 0 -> makeTree l v p (delete_raise x r)
	| _ -> plus l r

let delete = fun x tree ->
  try
    delete_raise x tree
  with Not_found -> tree

let rec delete_min = function
  | E -> E
  | N (E, _, _, r) -> r
  | N (l, v, p, r) -> makeTree (delete_min l) v p r

let rec delete_max = function
  | E -> E
  | N (l, _, _, E) -> l
  | N (l, v, p, r) -> makeTree l v p (delete_max r)

let rec delete_min_priority = function
  | E -> E
  | N (l, _, _, r) -> appendS l r

let rec max_priority_rec ((_, pm) as max) = function
  | E -> max
  | N (E, v, p, E) -> 
      (match compare p pm with
	 | n when n < 0 -> (v, p)
	 | _ -> max
      )
  | N (l, _, _, r) -> max_priority_rec (max_priority_rec max l) r

let max_priority = function
  | E -> raise Not_found
  | N (l, v, p, r) -> max_priority_rec (max_priority_rec (v, p) l) r

(* elements (first key) are unique *)
(* les lments (premire composante) sont uniques *)
(* los elementos (primera componente) son nicas *)

let delete_max_priority = function tree ->
  let (x, _) = max_priority tree in 
    delete x tree

let rec delete_list = fun list set ->
  match list with
    | [] -> set
    | head :: tail -> delete_list tail (delete head set)

let rec filter_rec f result = function
  | E -> result
  | N (l, v, p, r) ->
      filter_rec f (filter_rec f (
		      if f v p then insert v p result else result) l) r

let filter f = function set -> filter_rec f E set

let rec filter_le x = function
  | E -> E
  | N (l, v, p, r) ->
      match compare x v with
	| n when n < 0 -> filter_le x l
	| _ -> makeTree l v p (filter_le x r)

let rec filter_ge x = function
  | E -> E
  | N (l, v, p, r) ->
      match compare x v with
	| n when n <= 0 -> makeTree (filter_ge x l) v p r
	| _ -> filter_ge x r

let rec range min max = function
  | E -> E
  | N (l, v, p, r) ->
      if compare v min < 0 then range min max r
      else if compare max v < 0 then range min max l
      else makeSTree (filter_ge min l) v p (filter_le max r)


(** Vues **)
(** Views **)
(** Vistas **)


let rec min_view = function
  | E -> raise Empty
  | N (E, v, p, r) -> ((v, p), r)
  | N (l, v, p, r) -> let (min, left) = min_view l in 
      (min, makeTree left v p r)
      
let rec max_view = function
  | E -> raise Empty
  | N (l, v, p, E) -> (l, (v, p))
  | N (l, v, p, r) -> let (right, max) = max_view r in 
      (makeTree l v p right, max)

let min_priority_view = function
  | E -> raise Empty
  | N (l, v, p, r) -> ((v, p), plus l r)

let rec max_priority_view = function
  | E -> raise Empty
  | N (E, v, p, _) -> (E, (v, p))
  | N (l, v, p, E) -> 
      let (l', v') = max_priority_view l in (makeTree l' v p E, v')
  | N (l, v, p, r) ->
      let (_, lp) = max_priority l and (_, rp) = max_priority r in
	match compare lp rp with
	  | n when n < 0 ->
	      let (r', min) = max_priority_view r in (makeTree l v p r', min)
	  | _ ->
	      let (l', min) = max_priority_view l in (makeTree l' v p r, min)

let rec partition_rec f (trueS, falseS) = function
  | E -> (trueS, falseS)
  | N (l, v, p, r) -> 
      if f v p
      then partition_rec f (partition_rec f (insert v p trueS, falseS) l) r
      else partition_rec f (partition_rec f (trueS, insert v p falseS) l) r

let partition = fun f set -> partition_rec f (E, E) set


(** Observateurs **)
(** Observers **)
(** Observadores **)


let is_empty = function
  | E -> true
  | N _ -> false

let rec size_rec result = function
  | E -> result
  | N (l, _, _, r) -> size_rec (size_rec (1 + result) r) l

let size = function tree -> size_rec 0 tree

let rec member_min_bound b x = function
  | E -> compare b x = 0
  | N (l, v, _, r) ->
      match compare x v with
	| k when k < 0 -> member_min_bound b x l
	| _ -> member_min_bound v x r
	    
let rec member' x = function
  | E -> false
  | N (l, v, _, r) ->
      match compare x v with
	| n when n < 0 -> member' x l
	| _ -> member_min_bound v x r

let rec member x = function
  | E -> false
  | N (l, v, _, r) ->
      match compare x v with
	| n when n < 0 -> member x l
	| n when n > 0 -> member x r
	| _ -> true

let rec find f = function
  | E -> raise Not_found
  | N (_, v, _, _) when f v -> v
  | N (l, _, _, r) -> (try find f l with Not_found -> find f r)

let choose = function
  | E -> raise Empty
  | N (_, v, _, _) -> v

let rec min_element = function
  | E -> raise Not_found
  | N (E, v, p, _) -> (v, p)
  | N (l, _, _, _) -> min_element l

let rec max_element = function
  | E -> raise Not_found
  | N (_, v, p, E) -> (v, p)
  | N (_, _, _, r) -> max_element r

let rec min_priority = function
  | E -> raise Empty
  | N (_, v, p, _) -> (v, p)

let rec next x = function
  | E -> raise Not_found
  | N (l, v, p, r) ->
      match compare x v with
	| n when n < 0 -> (try next x l with Not_found -> (v, p))
	| n when n > 0 -> next x r
	| _ -> min_element r (* raises [Not_found] *)

let rec previous x = function
  | E -> raise Not_found
  | N (l, v, p, r) ->
      match compare x v with
	| n when n < 0 -> previous x l
	| n when n > 0 -> (try previous x r with Not_found -> (v, p))
	| _ -> max_element l (* raises [Not_found] *)
	    
let rec ceil x = function
  | E -> raise Not_found
  | N (l, v, p, r) ->
      match compare x v with
	| n when n < 0 -> (try ceil x l with Not_found -> (v, p))
	| n when n > 0 -> ceil x r
	| _ -> (x, p)

let rec floor x = function
  | E -> raise Not_found
  | N (l, v, p, r) ->
      match compare x v with
	| n when n < 0 -> floor x l
	| n when n > 0 -> (try floor x r with Not_found -> (v, p))
	| _ -> (x, p)

let rec rank x = function
  | E -> raise Not_found
  | N (l, v, _, r) ->
      match compare x v with
	| n when n < 0 -> rank x l
	| n when n > 0 -> 1 + size l + rank x r
	| _ -> size l

let rec get n = function
  | E -> raise (Out_of_bounds n)
  | N (l, v, p, r) -> 
      let k = size l in
	if n < k then get n l
	else if n > k then get (n - k - 1) r
	else (v, p)

let rec unsafe_set value priority m = function
  | E -> raise (Out_of_bounds m)
  | N (l, v, p, r) ->
      let k = size l in
	if m < k then makeTree (unsafe_set value priority m l) v p r
	else if m > k then 
	  makeTree l v p (unsafe_set value priority (m - k - 1) r)
	else makeTree l value priority r

let rec for_all f = function
  | E -> true
  | N (l, v, p, r) -> f v p && for_all f l && for_all f r

let rec exists f = function
  | E -> false
  | N (l, v, p, r) -> f v p || exists f l || exists f r


(** Traversals **)
(** Parcours **)
(** Recorridos **)


let rec iter f = function
  | E -> ()
  | N (l, v, p, r) -> iter f l ; f v p ; iter f r

let rec fold_left f e = function
  | E -> e
  | N (l, v, p, r) -> fold_left f (f (fold_left f e l) v p) r

let rec fold_right = fun f set e ->
  match set with
    | E -> e 
    | N (l, v, p, r) -> fold_right f l (f v p (fold_right f r e))

let fold = fold_right


(** Miscellaneous **)
(** Divers **)
(** Miscelanios **)


let bind = fun set f -> join (map f set)

let concat_map = fun f set -> join (map f set)

let root_priority = function
  | E -> raise Not_found
  | N (_, _, p, _) -> p

(* substracts 1 to the priority of a key and rebalances the tree *)

let rec substract_one value = function
  | E -> raise Not_found
  | N (l, v, p, r) ->
      match compare value v with
	| n when n < 0 ->
	    let left = substract_one value l in
	      if root_priority left < p then rotate_R left v p r
	      else makeTree left v p r
	| n when n > 0 ->
	    let right = substract_one value r in
	      if root_priority right < p then rotate_L l v p right
	      else makeTree l v p right
	| _ -> makeTree l v (p - 1) r

(* insert or replace if new_priority < old_priority *)

let rec reduce_rec x priority = function
  | E -> injection x priority
  | N (l, v, p, r) ->
      match compare x v with
	| n when n < 0 ->
	    if priority < p then rotate_R (reduce_rec x priority l) v p r
	    else makeTree (reduce_rec x priority l) v p r
	| n when n > 0 ->
	    if priority < p then rotate_L l v p (reduce_rec x priority r)
	    else makeTree l v p (reduce_rec x priority r)
	| _ -> 
	    if priority < p then makeSTree l x priority r
	    else raise Exit

let reduce = fun x priority heap ->
  try
    reduce_rec x priority heap
  with Exit -> heap


(** Compatibilit avec la librairie standard de Caml **)
(** Caml standard library compatibility **)
(** Compatibilidad con la librera estandar de Caml **)


let empty = E
let mem = member
let add = insert
let singleton = injection
let remove = delete
let inter = intersection
let diff = difference
let cardinal = size
let elements = to_list
let min_elt = min_element    
let max_elt = max_element
