(*
 * Priority search queues over unbalanced trees
 * 
 *)

open BairePervasives

(* Ce code est bas sur le rapport technique de Ralph Hinze
 * 
 * A simple implementation technique for priority search queues
 * Technical report UU-CS 2001 09
 * Universiteit Utrecht
 * 
 * L'auteur en a galement prsent une version plus rduite qui n'aborde
 * pas la questions des requtes par intervalles de priorit
 * 
 * Nous recommandons galement, afin de se familiariser avec la reprsentation
 * des tas qu'utilise ce code, la lecture de l'article
 * 
 * Explaining binomial heaps
 * Ralph Hinze
 * Functional Pearls 
 * 
 * Tous ces documents sont disponibles sur la toile
 *)

type 
  'a tree = 
  | E 
  | L of 'a * int * 'a tree * 'a * 'a tree
  | R of 'a * int * 'a tree * 'a * 'a tree 
and
  'a prioritySQ =
  | EmptyQ
  | Pennant of 'a * int * 'a tree * 'a

(* la cl supplmentaire dite de contrle dans l'arbre binaire de recherche 
 * sert  maintenir l'ordre symtrique pour les recherches en log n
 * (quatrime argument des constructeurs [L] et [R])
 * la cl supplmentaire dite maximale dans le pennant sert  maintenir
 * la structure de semi-tas (quatrime argument du constructeur [Pennant])
 *)

(* Invariants de semi-tas
 *   
 *  (i) Toutes les priorits doivent tre suprieures  celle de la tte 
 *      du pennant
 * 
 * (ii) La priorit de tout noeud doit tre infrieure  celle de la branche
 *      dont il est originaire, un lment est originaire de la branche gauche
 *      si sa cl est infrieure  la cl de contrle, de la branche droite
 *      autrement (conservation de la structure d'arbre binaire de recherche)
 *      Pour viter d'avoir  faire systmatiquement cette comparaison,
 *      l'origine du noeud est encode dans le constructeur
 * 
 * Invariants d'arbre de recherche binaire
 * 
 * (i)  Les lments de la branche gauche ont des cls infrieures  la cl de
 *      contrle
 * 
 * (ii) Les lments de la branche droite ont des cls suprieures  la cl de
 *      contrle
 * 
 * Invariants de cohrence 
 * 
 * les cls maximales (dans le pennant) et les cls de contrle (dans l'arbre
 * binaire de recherche) doivent correspondre  des lments contenus dans la
 * file de recherche par priorit
 * 
 * Invariant d'unicit :
 * 
 * Il ne peut y avoir deux lments avec la mme cl (en revanche des priorits
 * identiques sont parfaitement autorises)
 *)

(* la fonction [link] unit deux pennants tels que les cls de l'argument 
 * de droite soient infrieures  celle de l'argument de gauche 
 * (quivalent de [append] dans les arbres binaires de recherche)
 *)

let link = fun a b ->
  match (a, b) with
    | (EmptyQ, _) -> b
    | (_, EmptyQ) -> a
    | (Pennant (x, p, t, m), Pennant (x', p', t', m')) ->
	match compare p p' with
	  | n when n <= 0 -> Pennant (x, p, R (x', p', t, m, t'), m')
	  | _ -> Pennant (x', p', L (x, p, t, m, t'), m')

let unit = fun x p -> Pennant (x, p, E, x)

let max_element = function
  | EmptyQ -> raise Empty
  | Pennant (_, _, _, m) -> m

let rec min_tail_rec k = function
  | E -> EmptyQ
  | L (x, p, a, y, b) ->
      link (Pennant (x, p, a, k)) (min_tail_rec y b)
  | R (x, p, a, y, b) ->
      link (min_tail_rec k a) (Pennant (x, p, b, y))

let min_tail = function
  | EmptyQ -> raise Empty
  | Pennant (_, _, t, m) -> min_tail_rec m t

let min_view = function
  | EmptyQ -> raise Empty
  | Pennant (x, p, t, m) -> ((x, p), min_tail_rec m t)

let rec insert x p = function
  | EmptyQ -> Pennant (x, p, E, x)
  | Pennant (x', p', E, _) ->
      (match compare x x' with
	 | n when n < 0 -> link (unit x p) (unit x' p')
	 | 0 -> unit x p
	 | _ -> link (unit x' p') (unit x p)
      )
  | Pennant (rx, rp, L (lx, lp, lt, lm, rt), rm) ->
      (match compare x lm with
	 | n when n <= 0 -> 
	     link 
	     (insert x p (Pennant (lx, lp, lt, lm)))
	     (Pennant (rx, rp, rt, rm))
	 | _ -> 
	     link
	     (Pennant (lx, lp, lt, lm))
	     (insert x p (Pennant (rx, rp, rt, rm)))
      )
  | Pennant (lx, lp, R (rx, rp, lt, lm, rt), rm) ->
      match compare x lm with
	| n when n <= 0 -> 
	    link 
	    (insert x p (Pennant (lx, lp, lt, lm)))
	    (Pennant (rx, rp, rt, rm))
	| _ -> 
	    link
	    (Pennant (lx, lp, lt, lm))
	    (insert x p (Pennant (rx, rp, rt, rm)))

let rec delete x = function
  | EmptyQ -> EmptyQ 
  | Pennant (x', _, E, _) as psq ->
      (match compare x x' with
	 | 0 -> EmptyQ
	 | _ -> psq
      )
  | Pennant (rx, rp, L (lx, lp, lt, lm, rt), rm) ->
      (match compare x lm with
	 | n when n <= 0 -> 
	     link 
	     (delete x (Pennant (lx, lp, lt, lm)))
	     (Pennant (rx, rp, rt, rm))
	 | _ -> 
	     link
	     (Pennant (lx, lp, lt, lm))
	     (delete x (Pennant (rx, rp, rt, rm)))
      )
  | Pennant (lx, lp, R (rx, rp, lt, lm, rt), rm) ->
      match compare x lm with
	| n when n <= 0 -> 
	    link 
	    (delete x (Pennant (lx, lp, lt, lm)))
	    (Pennant (rx, rp, rt, rm))
	| _ -> 
	    link
	    (Pennant (lx, lp, lt, lm))
	    (delete x (Pennant (rx, rp, rt, rm)))

let rec to_list = function
  | EmptyQ -> []
  | Pennant (x, p, E, _) -> [(x , p)]
  | Pennant (rx, rp, L (lx, lp, lt, lm, rt), rm) ->
      to_list (Pennant (lx, lp, lt, lm)) @ to_list (Pennant (rx, rp, rt, rm))
  | Pennant (lx, lp, R (rx, rp, lt, lm, rt), rm) ->
      to_list (Pennant (lx, lp, lt, lm)) @ to_list (Pennant (rx, rp, rt, rm))
    
let rec from_list_rec result = function
  | [] -> result
  | (x, p) :: tail -> from_list_rec (insert x p result) tail

let from_list = function list -> from_list_rec EmptyQ list

let rec lookup x = function
  | EmptyQ -> raise Not_found
  | Pennant (x', p', E, _) ->
      (match compare x x' with
	 | 0 -> p'
	 | _ -> raise Not_found
      )
  | Pennant (rx, rp, L (lx, lp, lt, lm, rt), rm) ->
      (match compare x lm with
	 | n when n <= 0 -> lookup x (Pennant (lx, lp, lt, lm))
	 | _ -> lookup x (Pennant (rx, rp, rt, rm))
      )
  | Pennant (lx, lp, R (rx, rp, lt, lm, rt), rm) ->
      match compare x lm with
	| n when n <= 0 -> lookup x (Pennant (lx, lp, lt, lm))
	| _ -> lookup x (Pennant (rx, rp, rt, rm))

let zero = function () -> EmptyQ

let is_empty = function
  | EmptyQ -> true
  | _ -> false
