(*
 * Amortized Queue
 *
 *)

open BairePervasives
open StreamHeader

type 'a seq = int * 'a stream * 'a stream * int
		
(*
 * Invariant : 
 * l'avant doit tre au moins aussi long que l'arrire
 *)

		
(** Fonctions utiles **)

let makeQueue = fun i front rear j ->
  if i < j then (i + j, front ++ Stream.reverse rear, Nil, 0) 
  else (i, front, rear, j)

    
(** Signature monadique **)


let unit = function x -> (1, Cons (x, lazy Nil), Nil, 0)

let map f = function (i, front, rear, j) ->
  (i, Stream.map f front, Stream.map f rear, j)

let rec join_front = function
  | Nil -> Nil
  | Cons ((_, fr, re, _), tail) -> 
      fr ++ Stream.reverse re ++ join_front (force tail)

let rec join_rear = function
  | Nil -> Nil
  | Cons ((_, fr, re, _), tail) -> 
      re ++ Stream.reverse fr ++ join_rear (force tail)

let join = function (_, front, rear, _) ->
  let 
    front_j = join_front front and
    rear_j = join_rear rear
  in
     makeQueue (Stream.size front_j) front_j rear_j (Stream.size rear_j)
      
let bind = fun queue f -> join (map f queue)

let plus = fun (i1, front1, rear1, j1) (i2, front2, rear2, j2) ->
  (i1 + j1 + i2, front1 ++ Stream.reverse front2 ++ rear1, rear2, j2)

let zero = function () -> (0, Nil, Nil, 0)

let make n = function element ->
  if (n < 0) then invalid_arg "AmortizedQueue.make : argument ngatif"
  else (n, Stream.make n element, Nil, 0)
      
let get = fun queue n ->
  match queue with
    | (i, _, _, j) when (n > i + j - 1) -> raise (Out_of_bounds n)
    | (i, front, _, _) when (n < i) -> Stream.get front n
    | (i, _, rear, j) -> Stream.get rear (i + j - n - 1) 

let set = fun queue n x ->
  match queue with
    | (i, _, _, j) when (n > i + j - 1) -> raise (Out_of_bounds n)
    | (i, front, rear, j) when (n < i) -> (i, Stream.set front n x, rear, j)
    | (i, front, rear, j) -> (i, front, Stream.set rear (i + j - n - 1) x, j)
	
let compare = fun (_, fr1, re1, _) (_, fr2, re2, _) ->
  match Stream.compare fr1 fr2 with
    | 0 -> Stream.compare re1 re2
    | other  -> other


(** Constructeurs **)


let cons = fun x (i, front, rear, j) -> (i + 1, Cons (x, lazy front), rear, j)

let snoc = fun (i, front, rear, j) x  -> (i, front, Cons (x, lazy rear), j + 1)

let from_list = function list -> 
  (List.size list, Stream.from_list list, Nil, 0)

let init n = function f -> 
  if n < 0 then invalid_arg "AmortizedQueue.init : argument ngatif" 
  else (n, Stream.init n f, Nil, 0)


(** Destructeurs **)


let left_view = function
  | (i, Cons (head, tail), rear, j) -> 
      (head, makeQueue (i - 1) (force tail) rear j)
  | _ -> raise Empty

let left_head = function
  | (_, Cons (head, _), _, _) -> head
  | _ -> raise Empty

let left_tail = function
  | (i, Cons (_, tail), rear, j) -> makeQueue (i - 1) (force tail) rear j
  | _ -> raise Empty

let right_view = function
  | (i, front, Cons (head, tail), j) -> ((i, front, force tail, j - 1), head)
  | (i, front, Nil, _) -> let (tail, head) = Stream.right_view front in
      ((i - 1, tail, Nil, 0), head)
      
let right_head = function
  | (_, _, Cons (head, _), _) -> head
  | (_, front, _, _) -> Stream.right_head front
      
let right_tail = function
  | (i, front, Cons (head, tail), j) -> (i, front, force tail, j - 1)
  | (i, front, _, _) -> (i - 1, Stream.right_tail front, Nil, 0)
      
      
(** Observateurs **)
      
      
let is_empty = function (i, _, _, _) -> i = 0

let size = function (i, _, _, j) -> i + j

let to_list = function (_, front, rear, j) -> 
  match j with
    | 0 -> Stream.to_list front
    | _ -> Stream.to_list front @ List.reverse (Stream.to_list rear)


(** Renversements **)


let reverse = function (i, front, rear, j) -> makeQueue j rear front i

let reverse_onto = fun (i1, fr1, ar1, j1) (i2, fr2, ar2, j2) ->
  (i1 + j1 + i2, fr2 ++ Stream.reverse_append ar2 fr1, ar2, j2) 
  
let reverse_append = fun (i1, fr1, ar1, j1) (i2, fr2, ar2, j2) ->
  (i1 + j1 + i2, fr1 ++ Stream.reverse_append ar1 fr2, ar2, j2)
  

(** Maps et folds **)


let concat_map = fun f queue -> join (map f queue)
  
let fold_right f e = function (_, front, rear, _) ->
  Stream.fold_right f (Stream.fold_left (sym f) e rear) front
  
let fold_left f e = function (_, front, rear, _) ->
  Stream.fold_right (sym f) (Stream.fold_left f e front) rear
    
let rec iter_forwards_rec f = function
  | Nil  -> ()
  | Cons (head, tail) -> f head ; iter_forwards_rec f (force tail)
  
let rec iter_backwards_rec f = function
  | Nil -> ()
  | Cons (head, tail) -> iter_backwards_rec f (force tail) ; f head

let iter f = function (_, front, rear, _) ->
  iter_forwards_rec f front ;
  iter_backwards_rec f rear


(** Sous-squences **)


let take n = function
  | (i, front, _, _) when n < i -> makeQueue n (Stream.take n front) Nil 0
  | (i, front, rear, j) when n <= i + j -> 
      makeQueue i front (Stream.drop (i + j - n) rear) (n - i)
  | _ -> raise (Out_of_bounds n)     
 
let drop n = function
  | (i, front, rear, j) when n < i -> 
      makeQueue (i - n) (Stream.drop n front) rear j
  | (i, front, rear, j) when n <= i + j -> 
      makeQueue 0 Nil (Stream.take (i + j - n) rear) (i + j - n)
  | _ -> raise (Out_of_bounds n)
      
let split_at n = function
  | (i, front, rear, j) when (n < i) -> 
      let (x, y) = Stream.split_at n front in
	(makeQueue n x Nil 0, makeQueue (i - n) y rear j)
  | (i, front, rear, j) when n <= i + j ->
      let (x, y) = Stream.split_at (i + j - n) rear in
	(makeQueue i front y (n - i), 
	 makeQueue 0 Nil x (i +  j - n))
  | _ -> raise (Out_of_bounds n)

let sub dbut fin = function
  | (i, front, _, _) when dbut < i & fin < i ->
      makeQueue (fin - dbut + 1) (Stream.sub dbut fin front) Nil 0
  | (i, front, rear, j) when dbut < i & fin >= i ->
      makeQueue (i - dbut) (Stream.drop dbut front) 
      (Stream.drop (i + j - fin - 1) rear) (fin - i + 1)
  | (i, _, rear, j) -> makeQueue 0 Nil
      (Stream.sub (i + j - fin) (i + j - dbut) rear) (fin - dbut)

	
(** Prdicats **)


let exists f = function (_, front, rear, _) -> 
  Stream.exists f front || Stream.exists f rear
      
let for_all f = function (_, front, rear, _ ) -> 
  Stream.for_all f front && Stream.for_all f rear

let find f = function (_, front, rear, _) ->
  try Stream.find f front
  with Not_found -> Stream.find f (Stream.reverse rear)
  
let filter f = function (i, front, rear, j) ->
  let 
    front_f= Stream.filter f front and
    rear_f = Stream.filter f rear
  in
    makeQueue (Stream.size front_f) front_f rear_f (Stream.size rear_f)

let partition f = function queue ->
  let (a, b) = List.partition f (to_list queue) in (from_list a, from_list b)

let take_while f = function queue ->
  let list = List.take_while f (to_list queue) in from_list list
      
let drop_while f = function queue ->
  let list = List.drop_while f (to_list queue) in from_list list

let split_while f = function queue ->
  let (a, b) = List.split_while f (to_list queue) in 
    (from_list a, from_list b)

      
(** Oprations indexes **)


let map_with_index = fun f (i, front, rear, j) ->
  let rec   
    mapi_inc f k = function
      | Nil -> Nil
      | Cons (head, tail) -> 
	  Cons (f k head, lazy (mapi_inc f (k + 1) (force tail)))
  and    
    mapi_dec f k = function
      | Nil -> Nil
      | Cons (head, tail) -> 
	  Cons (f k head, lazy (mapi_dec f (k - 1) (force tail)))
  in
    (i, mapi_inc f 0 front, mapi_dec f (i + j - 1) rear, j)

let iter_with_index f = function queue ->
  List.iter_with_index f (to_list queue)

let find_with_index f = function queue ->
  List.find_with_index f (to_list queue)

let filter_with_index f = function queue ->
  from_list (List.filter_with_index f (to_list queue))

let fold_right_with_index f e = function queue ->
  List.fold_right_with_index f e (to_list queue)

let fold_left_with_index f e = function queue ->
  List.fold_left_with_index f e (to_list queue)

let fill queue dbut fin motif = 
  from_list (List.fill (to_list queue) dbut fin motif)

let blit queueA dbutA queueB dbutB taille =
  from_list (List.blit (to_list queueA) dbutA (to_list queueB) dbutB taille)


(** Zips and unzips **)


let zip = fun queueA queueB ->
  from_list (List.zip (to_list queueA) (to_list queueB))


let unzip = function (i, front, rear, j) ->
  let 
    (fr1, fr2) = Stream.unzip front and
    (re1, re2) = Stream.unzip rear
  in
    (makeQueue i fr1 re1 j, makeQueue i fr2 re2 j)


(** Aliases **)
    

let empty = zero
let append = plus



module Make = functor (Seq : Signatures.FunctionalSequence) ->
struct

  type 'a seq = int * 'a Seq.seq * 'a Seq.seq * int
		
  (** utils **)

  let (++) = Seq.plus
	       
  let makeQueue = fun i front rear j ->
    if i < j then (i + j, front ++ Seq.reverse rear, Seq.empty (), 0) 
    else (i, front, rear, j)
      
      
  (** Signature monadique **)


  let unit = function x -> (1, Seq.cons x (Seq.empty ()), Seq.empty (), 0)

  let map f = function (i, front, rear, j) ->
    (i, Seq.map f front, Seq.map f rear, j)

  let plus = fun (i1, front1, rear1, j1) (i2, front2, rear2, j2) ->
    (i1 + j1 + i2, front1 ++ Seq.reverse front2 ++ rear1, rear2, j2)

  let join = function (_, front, rear, _) -> raise (Not_working "join")
  let bind = fun queue f -> raise (Not_working "bind")
  let concat_map = fun f queue -> raise (Not_working "concat_map")

  let zero = function () -> (0, Seq.empty (), Seq.empty (), 0)

  let make n = function element ->
    if (n < 0) then invalid_arg "AmortizedQueue.make : argument ngatif"
    else (n, Seq.make n element, Seq.empty (), 0)
      
  let get = fun queue n ->
    match queue with
      | (i, _, _, j) when (n > i + j - 1) -> raise (Out_of_bounds n)
      | (i, front, _, _) when (n < i) -> Seq.get front n
      | (i, _, rear, j) -> Seq.get rear (i + j - n - 1) 

  let set = fun queue n x ->
    match queue with
      | (i, _, _, j) when (n > i + j - 1) -> raise (Out_of_bounds n)
      | (i, front, rear, j) when (n < i) -> (i, Seq.set front n x, rear, j)
      | (i, front, rear, j) -> (i, front, Seq.set rear (i + j - n - 1) x, j)

  let compare = fun (_, fr1, re1, _) (_, fr2, re2, _) ->
    match Stream.compare fr1 fr2 with
      | 0 -> Stream.compare re1 re2
      | other  -> other
	  

  (** Constructeurs **)


  let cons = fun x (i, front, rear, j) -> (i + 1, Seq.cons x front, rear, j)

  let snoc = fun (i, front, rear, j) x  -> (i, front, Seq.cons x rear, j + 1)

  let from_list = function list ->
    (List.size list, Seq.from_list list, Seq.empty (), 0)

  let init n = function f -> 
    if n < 0 then invalid_arg "AmortizedQueue.Make.init : argument ngatif"
    else (n, Seq.init n f, Seq.empty (), 0)


  (** Destructeurs **)


  let left_view = function (i, front, rear, j) -> 
    let (head, tail) = Seq.left_view front in 
      (head, makeQueue (i - 1) tail rear j)

  let left_head = function (_, front, _, _) -> Seq.left_head front

  let left_tail = function  (i, front, rear, j) -> 
    let tail = Seq.left_tail front in makeQueue (i - 1) tail rear j

  let right_view = function (i, front, rear, j) -> 
    if Seq.is_empty rear then let (tail, head) = Seq.right_view front in
      ((i - 1, tail, Seq.empty (), 0), head)    
    else let (head, tail) = Seq.left_view rear in 
      ((i, front, tail, j - 1), head)

  let right_head = function (_, front, rear, _) ->
    if Seq.is_empty rear then Seq.right_head front else Seq.left_head rear
      
  let right_tail = function (i, front, rear, j) ->
    if Seq.is_empty rear then (i - 1, Seq.right_tail front, Seq.empty (), 0)
    else (i, front, Seq.left_tail rear, j - 1)
      
      
  (** Observateurs **)
      
      
  let is_empty = function (i, _, _, _) -> i = 0

  let size = function (i, _, _, j) -> i + j

  let to_list = function (_, front, rear, j) -> 
    match j with
      | 0 -> Seq.to_list front
      | _ -> Seq.to_list front @ List.reverse (Seq.to_list rear)


  (** Renversements **)
	  
	  
  let reverse = function (i, front, rear, j) -> makeQueue j rear front i
    
  let reverse_onto = fun (i1, fr1, ar1, j1) (i2, fr2, ar2, j2) ->
    (i1 + j1 + i2, fr2 ++ Seq.reverse_append ar2 fr1, ar2, j2) 
    
  let reverse_append = fun (i1, fr1, ar1, j1) (i2, fr2, ar2, j2) ->
    (i1 + j1 + i2, fr1 ++ Seq.reverse_append ar1 fr2, ar2, j2)
  
    
  (** Maps et folds **)

  let fold_right f e = function (_, front, rear, _) ->
    Seq.fold_right f (Seq.fold_left (sym f) e rear) front
  
  let fold_left f e = function (_, front, rear, _) ->
    Seq.fold_right (sym f) (Seq.fold_left f e front) rear
    
  let rec iter_backwards_rec f = function seq ->
    if Seq.is_empty seq then () 
    else let (head, tail) = Seq.left_view seq in 
      iter_backwards_rec f tail ; f head
  
  let iter f = function (_, front, rear, _) ->
    Seq.iter f front ;
    iter_backwards_rec f rear
      

  (** Sous-squences **)
      
      
  let take n = function
    | (i, front, _, _) when n < i -> 
	makeQueue n (Seq.take n front) (Seq.empty ()) 0
    | (i, front, rear, j) when n <= i + j -> 
	makeQueue i front (Seq.drop (i + j - n) rear) (n - i)
    | _ -> raise (Out_of_bounds n)     
	
  let drop n = function
    | (i, front, rear, j) when n < i -> 
	makeQueue (i - n) (Seq.drop n front) rear j
    | (i, front, rear, j) when n <= i + j -> 
	makeQueue 0 (Seq.empty ()) (Seq.take (i + j - n) rear) (i + j - n)
    | _ -> raise (Out_of_bounds n)
	
  let split_at n = function
    | (i, front, rear, j) when (n < i) -> 
	let (x, y) = Seq.split_at n front in
	  (makeQueue n x (Seq.empty ()) 0, makeQueue (i - n) y rear j)
    | (i, front, rear, j) when n <= i + j ->
	let (x, y) = Seq.split_at (i + j - n) rear in
	  (makeQueue i front y (n - i), 
	   makeQueue 0 (Seq.empty ()) x (i +  j - n))
    | _ -> raise (Out_of_bounds n)
	
  let sub dbut fin = function
    | (i, front, _, _) when dbut < i & fin < i ->
	makeQueue (fin - dbut + 1) (Seq.sub dbut fin front) (Seq.empty ()) 0
    | (i, front, rear, j) when dbut < i & fin >= i ->
	makeQueue (i - dbut) (Seq.drop dbut front) 
	(Seq.drop (i + j - fin - 1) rear) (fin - i + 1)
    | (i, _, rear, j) -> makeQueue 0 (Seq.empty ())
	(Seq.sub (i + j - fin) (i + j - dbut) rear) (fin - dbut)
	
	
  (** Prdicats **)
	
	
  let exists f = function (_, front, rear, _) -> 
    Seq.exists f front || Seq.exists f rear
    
  let for_all f = function (_, front, rear, _ ) -> 
    Seq.for_all f front && Seq.for_all f rear
    
  let find f = function (_, front, rear, _) ->
    try Seq.find f front
    with Not_found -> Seq.find f (Seq.reverse rear)
  
  let filter f = function (i, front, rear, j) ->
    let 
      front_f = Seq.filter f front and
      rear_f = Seq.filter f rear
    in
      makeQueue (Seq.size front_f) front_f rear_f (Seq.size rear_f)

  let partition f = function queue ->
    let (a, b) = List.partition f (to_list queue) in (from_list a, from_list b)
						       
  let take_while f = function queue ->
    let list = List.take_while f (to_list queue) in from_list list
						      
  let drop_while f = function queue ->
    let list = List.drop_while f (to_list queue) in from_list list
						      
  let split_while f = function queue ->
    let (a, b) = List.split_while f (to_list queue) in 
      (from_list a, from_list b)
      
      
  (** Oprations indexes **)
      
      
  let map_with_index = fun f (i, front, rear, j) ->
    let rec   
      mapi_inc f k = function seq ->
	if Seq.is_empty seq then Seq.empty ()
	else let (head, tail) = Seq.left_view seq in
	  Seq.cons (f k head) (mapi_inc f (k + 1) tail)
    and
      mapi_dec f k = function seq ->
	if Seq.is_empty seq then Seq.empty ()
	else let (head, tail) = Seq.left_view seq in
	Seq.cons (f k head) (mapi_dec f (k - 1) tail)
    in
      (i, mapi_inc f 0 front, mapi_dec f (i + j - 1) rear, j)
      
  let iter_with_index f = function queue ->
    List.iter_with_index f (to_list queue)
    
  let find_with_index f = function queue ->
    List.find_with_index f (to_list queue)
    
  let filter_with_index f = function queue ->
    from_list (List.filter_with_index f (to_list queue))
    
  let fold_right_with_index f e = function queue ->
    List.fold_right_with_index f e (to_list queue)
    
  let fold_left_with_index f e = function queue ->
    List.fold_left_with_index f e (to_list queue)
    
  let fill queue dbut fin motif = 
    from_list (List.fill (to_list queue) dbut fin motif)
      
  let blit queueA dbutA queueB dbutB taille =
    from_list (List.blit (to_list queueA) dbutA (to_list queueB) dbutB taille)


  (** Zips and unzips **)
      
      
  let zip = fun queueA queueB ->
    from_list (List.zip (to_list queueA) (to_list queueB))
    
    
  let unzip = function (i, front, rear, j) ->
    let 
      (fr1, fr2) = Seq.unzip front and
      (re1, re2) = Seq.unzip rear
    in
      (makeQueue i fr1 re1 j, makeQueue i fr2 re2 j)
      
      
  (** Aliases **)
      

  let empty = zero
  let append = plus

end
