(*
 * Batched Queue
 *
 *)

open BairePervasives

type 'a seq = 'a list * 'a list

(* 
 * invariant :
 * l'avant est vide seulement si l'arrire est vide aussi
 *)


let makeQueue = fun a b ->
  match a with
    | [] -> (List.reverse b, [])
    | _ -> (a, b)

let from_list = function liste -> (liste, []) 
	
let to_list = function (a, b) -> a @ List.reverse b

let unit = function x -> ([x], [])

let map f = function (a, b) -> (List.reverse_map_append f (List.map f a) b, [])

(* join ne bascule pas immdiatement la liste de queue en raison de 
 * l'appel non rcursif terminal  @ sur des listes de taille importante
 * que cela provoquerait
 *)

let join = function (a, b) ->
  let rec
    join_forwards = function
      | [] -> []
      | (a, b) :: tail ->
          a @ List.reverse b @ join_forwards tail
  and
    join_backwards = function
      | [] -> []
      | (a, b) :: tail -> b @ List.reverse a @ join_forwards tail 
  in
    makeQueue (join_forwards a) (join_backwards b)

let bind = fun liste f -> join (map f liste)

let plus = fun (a, b) (c, d) -> makeQueue (a @ List.reverse_append b c) d

let zero = function () -> ([], [])

let make = fun n x -> 
  if n < 0 then raise (Invalid_argument "make : argument ngatif")
  else (List.make n x, [])

let get = fun queue n -> List.get (to_list queue) n

let set = fun queue n x -> from_list (List.set (to_list queue) n x)

let cons x = function (a, b) -> (x :: a, b)

let snoc = fun (a, b) x ->
  match a with
    | [] -> unit x
    | _ -> (a, x :: b)

let init = fun f n -> from_list (List.init f n)

let left_view = function
  | ([x], a) -> (x, (List.reverse a, []))
  | (head :: tail, a) -> (head, (tail, a))
  | ([], _) -> raise Empty

let left_head = function
  | (head :: _, _) -> head
  | ([], _) -> raise Empty

let left_tail = function
  | ([x], a) -> (List.reverse a, [])
  | (head :: tail, a) -> (tail, a)
  | _ -> raise Empty

let right_view = function
  | (a, head :: tail) -> ((a, tail), head)
  | (a, _) -> let (tail, x) = List.right_view a in ((tail, []), x)

let right_head = function
  | (_, head :: _) -> head
  | (a, _) -> List.right_head a

let right_tail = function
  | (a, _ :: tail) -> (a, tail)
  | (a, _) -> (List.right_tail a, [])

let is_empty = function
  | ([], _) -> true
  | _ -> false

let size = function (a, b) -> List.size a + List.size b

let reverse = function (a, b) -> makeQueue b a

let reverse_append = fun (a, b) (c, d) -> (b @ List.reverse_append a c, d)

let reverse_onto = fun (a, b) (c, d) -> (c @ List.reverse_append d a, b)

let concat_map = fun f liste -> join (map f liste)

let fold_right f e = function (a, b) ->
  List.fold_right f (List.fold_left (sym f) e b) a

let fold_left f e = function (a, b) ->
  List.fold_right (sym f) (List.fold_left f e a) b

let rec iter_backwards f = function
  | [] -> ()
  | head :: tail -> iter_backwards f tail ; f head

let iter f = function
  | ([], _) -> ()
  | (a, b) -> List.iter f a ; iter_backwards f b

let take n = function liste -> from_list (List.take n (to_list liste))

let drop n = function liste -> from_list (List.drop n (to_list liste))

let split_at n = function liste ->
  let (listeA, listeB) = List.split_at n (to_list liste) in
    (from_list listeA, from_list listeB)

let sub = fun dbut fin liste -> from_list (List.sub dbut fin (to_list liste))

let for_all f = function (a, b) -> List.for_all f a && List.for_all f b

let exists f = function (a, b) -> List.exists f a || List.exists f b

let find f = function (a, b) ->
  try List.find f a
  with Not_found -> List.find f (List.reverse b)

let filter f = function (a, b) -> makeQueue (List.filter f a) (List.filter f b)

let partition f = function (a, b) ->
  let
    (trueA, falseA) = List.partition f a and
    (trueB, falseB) = List.partition f b

  in
    (makeQueue trueA trueB, makeQueue falseA falseB)

let take_while f = function queue -> 
  from_list (List.take_while f (to_list queue))

let drop_while f = function queue -> 
  from_list (List.drop_while f (to_list queue))

let split_while f = function queue ->
  let (a, b) = List.split_while f (to_list queue) in (from_list a, from_list b)

let map_with_index f = function queue ->
  from_list (List.map_with_index f (to_list queue))
  
let iter_with_index f = function queue -> 
  List.iter_with_index f (to_list queue)

let find_with_index f = function queue ->
  List.find_with_index f (to_list queue)

let filter_with_index f = function queue ->
  from_list (List.filter_with_index f (to_list queue))

let fold_right_with_index f e = function queue ->
  (List.fold_right_with_index f e (to_list queue))

let fold_left_with_index f e = function queue ->
  (List.fold_left_with_index f e (to_list queue))

let fill = fun queue dbut fin motif ->
  from_list (List.fill  (to_list queue) dbut fin motif)

let blit = fun queueA indiceA queueB indiceB taille ->
  from_list (
    List.blit (to_list queueA) indiceA (to_list queueB) indiceB taille
  )
  
let zip = fun (a, b) (c, d) -> (List.zip a c, List.zip b d)

let unzip = fun (a, b) -> 
  let 
    (frontA, frontB) = List.unzip a and
    (rearA, rearB) = List.unzip b
  in
    ((frontA, rearA), (frontB, rearB))

(** Aliases **)

let empty = zero
let append = plus
