(*
 * Scheduled Queue
 *
 *)

open BairePervasives
open StreamHeader

type 'a seq = 'a stream * 'a list * 'a stream

(* Invariant (front, rear, schedule) : 
 *
 *  size schedule = size front - size rear
 *
 * en particulier size front >= size rear
 * cet invariant garantit que les clotures sont excutes  temps
 *)
		  
(* Utils *)

(* Rotate : d'aprs l'invariant la taille de l'avant est toujours
 * suprieure  celle de l'arrire d'o la simplification de la
 * fonction rotate par rapport au cas gnral.
 *
 * En ralit Okasaki n'envisage que le cas rotate (x, y, a) avec
 *
 *  size x + 1 = size y
 *
 * let rec rotate = function
 *  | (Nil, y :: _, a) -> Cons (y, lazy a)
 *  | (Cons (x, xs), y :: ys, a) -> 
 *       Cons (x, lazy (rotate (force xs, ys, Cons (y, lazy a))))
 *  | _ -> failwith "unexpected error : realTimeQueue.rotate"
 * 
 * Nous en donnons une version beaucoup plus gnrale et flexible
 * drive de Stream.append_reverse
 *)

let rec reverse_append = fun a b ->
  match a with
  | [] -> b
  | head :: tail -> Cons (head, lazy (reverse_append tail b))

let rec rotate = function
  | (Nil, list, a) -> reverse_append list a
  | (stream, [], a) -> Stream.plus stream a
  | (Cons (x, xs), y :: ys, a) -> 
      Cons (x, lazy (rotate (force xs, ys, Cons (y, lazy a))))

(* Exec : force la dernire fermeture 
 * l'invariant n'est conserv que parce que les fonctions appelantes
 * (left_tail, snoc, etc.) s'en assurent explicitement
 * Une implmentations plus facilement maintenable est sans doute souhaitable
 *)

let exec = function
  | (front, rear, Cons (_, shedule)) -> (front, rear, force shedule)
  | (front, rear, Nil) -> let f' = rotate (front, rear, Nil) in (f', [], f')


(** Monadic operations **)


let unit = function x -> let front = Cons (x, lazy Nil) in (front, [], front)

let map f = function (front, rear, _) ->
  let f' = Stream.map f front in (f', List.map f rear, f')

let plus = fun (f1, r1, _) (f2, r2, _) ->
  let front = rotate (f1, r1, rotate (f2, r2, Nil)) in
    (front, [], front)

let zero = function () -> (Nil, [], Nil)

let make = fun n x -> 
  let front = Stream.make n x in (front, [], front)

let rec get_rec k rear = function
  | Nil -> let n = List.size rear - k - 1 in
      if n < 0 then raise Empty 
      else List.get rear n
  | Cons (x, _) when k = 0 -> x
  | Cons (_, tail) -> get_rec (k - 1) rear (force tail)

let get = fun (front, rear, _) n -> 
  if n < 0
  then raise (Invalid_argument "ScheduledQueue.get : argument ngatif")
  else get_rec n rear front

(** Constructors **)


let cons x = function (front, rear, schedule) ->
  (Cons (x, lazy front), rear, Cons (x, lazy schedule))

let snoc (f, r, s) x = exec (f, x :: r, s)

let from_list = function list -> 
  let front = Stream.from_list list in (front, [], front)

let init = fun n f ->
  let front = Stream.init n f in (front, [], front)


(** Destructors **)


let left_view = function (front, rear, schedule) ->
  match front with
    | Nil -> raise Empty
    | Cons (x, tail) -> (x, exec (force tail, rear, schedule))

let left_head = function (front, _, _) ->
  match front with
    | Nil -> raise Empty
    | Cons (x, _) -> x

let left_tail = function
  | (Nil, _, _) -> raise Empty
  | (Cons (_, front), rear, schedule) -> exec (force front, rear, schedule)

let right_head = function
  | (stream, [], _) -> Stream.right_head stream
  | (_, head :: _, _) -> head

let right_tail = function
  | (stream, [], _) -> let f' = Stream.right_tail stream in (f', [], f')
  | (stream, _ :: tail, schedule) -> exec (stream, tail, schedule)

let right_view = function
  | (stream, head :: tail, schedule) -> (exec (stream, tail, schedule), head)
  | (stream, [], _) -> let (f', head) = Stream.right_view stream in 
      ((f', [], f'), head)


(** Observers **)


let is_empty = function 
  | (Nil, _, _) -> true 
  | _ -> false
  
let size = function (front, rear, _) -> Stream.size front + List.size rear

(* confer Stream.append_reverse_rec *)

let rec to_list_rec result = fun stream list ->
  match stream with
    | Nil -> List.reverse_onto result list
    | Cons (x, next_stream) ->
	match list with
	  | [] -> Stream.to_list_append result stream
	  | head :: tail ->
	      x :: to_list_rec (head :: result) (force next_stream) tail

let to_list = function (front, rear, _) -> to_list_rec [] front rear


(** Reverse **)


let reverse = function (front, rear, _) ->
  let f'= Stream.append_reverse (Stream.from_list rear) front in
    (f', [], f')

let reverse_append = fun a b -> plus (reverse a) b
let reverse_onto = fun a b ->  plus (reverse b) a
let append_reverse = fun a b -> plus a (reverse b)


(** Folds and maps **)


let fold_right f e = function (front, rear, _) ->
  Stream.fold_right f (List.fold_left (sym f) e rear) front 

let fold_left f e = function (front, rear, _) ->
  List.fold_right (sym f) (Stream.fold_left f e front) rear

let rec reverse_iter f = function
  | [] -> ()
  | head :: tail -> reverse_iter f tail ; f head

let iter f = function (front, rear, _) ->
  Stream.iter f front ;
  reverse_iter f rear


(** Subsequences **)

let take = fun n queue -> from_list (List.take n (to_list queue))

let drop = fun n queue -> from_list (List.drop n (to_list queue))

let split_at = fun n queue -> 
  let (a, b) = List.split_at n (to_list queue) in
    (from_list a, from_list b)

let sub = fun st en queue -> from_list (List.sub st en (to_list queue))


(** Predicate operations **)


let exists f = function (front, rear, _) ->
  Stream.exists f front || List.exists f rear

let for_all f = function (front, rear, _) ->
  Stream.for_all f front && List.for_all f rear

let find f = function queue -> List.find f (to_list queue)

let filter f = function (front, rear, _) ->
  let f' = rotate (Stream.filter f front, List.filter f rear, Nil) in
    (f', [], f')



(** Index based operations **)

(** Aliases **)

let append = plus
let empty = zero
