(*
 * Sized Queue
 *
 *)

open BairePervasives

type 'a seq = int * 'a list * 'a list * int
		
(*
 * Invariant : 
 * l'avant doit tre au moins aussi long que l'arrire
 *)

		
(** Fonctions utiles **)

let makeQueue = fun i front rear j ->
  match i with
    | 0 -> (j, List.reverse rear, [], 0)
    | _ -> (i, front, rear, j)

    
(** Signature monadique **)


let unit = function x -> (1, [x], [], 0)

let map f = function (i, front, rear, j) ->
  (i, List.map f front, List.map f rear, j)

let rec join_front = function
  | [] -> []
  | (_, fr, re, _) :: tail -> fr @ List.reverse re @ join_front tail

let rec join_rear = function
  | [] -> []
  | (_, fr, re, _) :: tail -> re @ List.reverse fr @ join_rear tail

let join = function (_, front, rear, _) ->
  let 
    front_j = join_front front and
    rear_j = join_rear rear
  in
     makeQueue (List.size front_j) front_j rear_j (List.size rear_j)
      
let bind = fun queue f -> join (map f queue)

let plus = fun (i1, front1, rear1, j1) (i2, front2, rear2, j2) ->
  (i1 + j1 + i2, front1 @ List.reverse_onto front2 rear1, rear2, j2)

let zero = (0, [], [], 0)

let make n = function element ->
  if (n < 0) then raise (Invalid_argument "make : argument ngatif")
  else (n, List.make n element, [], 0)
      
let get = fun queue n ->
  match queue with
    | (i, _, _, j) when (n > i + j - 1) -> raise (Out_of_bounds n)
    | (i, front, _, _) when (n < i) -> List.get front n
    | (i, _, rear, j) -> List.get rear (i + j - n - 1) 

let set = fun queue n x ->
  match queue with
    | (i, _, _, j) when (n > i + j - 1) -> raise (Out_of_bounds n)
    | (i, front, rear, j) when (n < i) -> (i, List.set front n x, rear, j)
    | (i, front, rear, j) -> (i, front, List.set rear (i + j - n - 1) x, j)
	

(** Constructeurs **)


let cons = fun x (i, front, rear, j) -> (i + 1, x :: front, rear, j)

let snoc = fun (i, front, rear, j) x  -> (i, front, x :: rear, j + 1)

let from_list = function list -> (List.size list, list, [], 0)

let init n = function f -> 
  if n < 0 then raise (Invalid_argument "init : argument ngatif") 
  else (n, List.init n f, [], 0)


(** Destructeurs **)


let left_view = function
  | (i, head :: tail, rear, j) -> (head, makeQueue (i - 1) tail rear j)
  | _ -> raise Empty

let left_head = function
  | (_, head :: _, _, _) -> head
  | _ -> raise Empty

let left_tail = function
  | (i, head :: tail, rear, j) -> makeQueue (i - 1) tail rear j
  | _ -> raise Empty

let right_view = function
  | (i, front, head :: tail, j) -> ((i, front, tail, j - 1), head)
  | (i, front, [], _) -> let (tail, head) = List.right_view front in
      ((i - 1, tail, [], 0), head)

let right_head = function
  | (_, _, head :: _, _) -> head
  | (_, front, _, _) -> List.right_head front

let right_tail = function
  | (i, front, head :: tail, j) -> (i, front, tail, j - 1)
  | (i, front, _, _) -> (i - 1, List.right_tail front, [], 0)

      
(** Observateurs **)


let is_empty = function (i, _, _, _) -> i = 0

let size = function (i, _, _, j) -> i + j

let to_list = function (_, front, rear, j) -> 
  match j with
    | 0 -> front
    | _ -> front @ List.reverse rear


(** Renversements **)


let reverse = function (i, front, rear, j) -> makeQueue j rear front i

let reverse_onto = fun (i1, fr1, ar1, j1) (i2, fr2, ar2, j2) ->
  (i1 + j1 + i2, fr2 @ List.reverse_append ar2 fr1, ar2, j2) 
  
let reverse_append = fun (i1, fr1, ar1, j1) (i2, fr2, ar2, j2) ->
  (i1 + j1 + i2, fr1 @ List.reverse_append ar1 fr2, ar2, j2)
  

(** Maps et folds **)


let concat_map = fun x y -> raise (Not_working "concat_map")

let fold_right f e = function (_, front, rear, _) ->
  List.fold_right f (List.fold_left (sym f) e rear) front

let fold_left f e = function (_, front, rear, _) ->
  List.fold_right (sym f) (List.fold_left f e front) rear

let rec iter_forwards_rec f = function
  | [] -> ()
  | head :: tail -> f head ; iter_forwards_rec f tail
  
let rec iter_backwards_rec f = function
  | [] -> ()
  | head :: tail -> iter_backwards_rec f tail ; f head

let iter f = function (_, front, rear, _) ->
  iter_forwards_rec f front ;
  iter_backwards_rec f rear


(** Sous-squences **)


let take n = function
  | _ when n < 0 -> 
      raise (Invalid_argument "SizedQueue.take : argument ngatif")
  | (i, front, _, _) when n <= i -> makeQueue n (List.take n front) [] 0
  | (i, front, rear, j) when n <= i + j -> 
      makeQueue i front (List.drop (i + j - n) rear) (n - i)
  | _ -> raise (Out_of_bounds n)     
 
let drop n = function
  | _ when n < 0 -> 
      raise (Invalid_argument "SizedQueue.drop : argument ngatif")
  | (i, front, rear, j) when n <= i -> 
      makeQueue (i - n) (List.drop n front) rear j
  | (i, front, rear, j) when n <= i + j -> 
      makeQueue 0 [] (List.take (i + j - n) rear) (i + j - n)
  | _ -> raise (Out_of_bounds n)
      
let split_at n = function
  | _ when n < 0 ->
      raise (Invalid_argument "SizedQueue.split_at : argument ngatif")
  | (i, front, rear, j) when (n < i) ->
      let (x, y) = List.split_at n front in
	(makeQueue n x [] 0, makeQueue (i - n) y rear j)
  | (i, front, rear, j) when n <= i + j ->
      let (x, y) = List.split_at (i + j - n) rear in
	(makeQueue i front y (n - i), 
	 makeQueue 0 [] x (i +  j - n))
  | _ -> raise (Out_of_bounds n)

let sub dbut fin = function
  | (i, front, _, _) when dbut < i & fin < i ->
      makeQueue (fin - dbut + 1) (List.sub dbut fin front) [] 0
  | (i, front, rear, j) when dbut < i & fin >= i ->
      makeQueue (i - dbut) (List.drop dbut front) 
      (List.drop (i + j - fin - 1) rear) (fin - i + 1)
  | (i, _, rear, j) -> makeQueue 0 [] 
      (List.sub (i + j - fin) (i + j - dbut) rear) (fin - dbut)

	
(** Prdicats **)


let exists f = function (_, front, rear, _) -> 
  List.exists f front || List.exists f rear
      
let for_all f = function (_, front, rear, _ ) -> 
  List.for_all f front && List.for_all f rear

let find f = function (_, front, rear, _) ->
  try List.find f front
  with Not_found -> List.find f (List.reverse rear)
  
let filter f = function (i, front, rear, j) ->
  let 
    front_f= List.filter f front and
    rear_f = List.filter f rear
  in
    makeQueue (List.size front_f) front_f rear_f (List.size rear_f)

let partition f = function queue ->
  let (a, b) = List.partition f (to_list queue) in (from_list a, from_list b)

let take_while f = function queue ->
  let list = List.take_while f (to_list queue) in from_list list
      
let drop_while f = function queue ->
  let list = List.drop_while f (to_list queue) in from_list list

let split_while f = function queue ->
  let (a, b) = List.split_while f (to_list queue) in 
    (from_list a, from_list b)

      
(** Oprations indexes **)


let map_with_index = fun f (i, front, rear, j) ->
  let rec   
    mapi_inc f k = function
      | [] -> []
      | head :: tail -> f k head :: mapi_inc f (k + 1) tail
  and    
    mapi_dec f k = function
      | [] -> []
      | head :: tail -> f k head :: mapi_dec f (k - 1) tail
  in
    (i, mapi_inc f 0 front, mapi_dec f (i + j - 1) rear, j)

let iter_with_index f = function queue ->
  List.iter_with_index f (to_list queue)

let find_with_index f = function queue ->
  List.find_with_index f (to_list queue)

let filter_with_index f = function queue ->
  from_list (List.filter_with_index f (to_list queue))

let fold_right_with_index f e = function queue ->
  List.fold_right_with_index f e (to_list queue)

let fold_left_with_index f e = function queue ->
  List.fold_left_with_index f e (to_list queue)

let fill queue dbut fin motif = 
  from_list (List.fill (to_list queue) dbut fin motif)

let blit queueA dbutA queueB dbutB taille =
  from_list (List.blit (to_list queueA) dbutA (to_list queueB) dbutB taille)


(** Zips and unzips **)


let zip = fun queueA queueB ->
  from_list (List.zip (to_list queueA) (to_list queueB))

let unzip = function (i, front, rear, j) ->
  let 
    (fr1, fr2) = List.unzip front and
    (re1, re2) = List.unzip rear
  in
    (makeQueue i fr1 re1 j, makeQueue i fr2 re2 j)


(** Aliases **)
    

let empty = zero
let append = plus
