(*
 * Sized and Scheduled Queue
 *)

type 'a seq = int * 'a stream * 'a list * int * 'a stream


(** Utils **)

let rec reverse_append = fun a b ->
  match a with
  | [] -> b
  | head :: tail -> Cons (head, lazy (reverse_append tail b))

let rec rotate = fun stream list astream ->
  match (stream, list) with
    | (Nil, _) -> reverse_append list astream
    | (_, []) -> Stream.plus stream astream
    | (Cons (x, xs), y :: ys) -> 
	Cons (x, lazy (rotate (force xs) ys (Cons (y, lazy astream))))

let exec = function (i, front, rear, j, schedule) as queue ->
  match (schedule, rear) with
    | (Cons (_, new_schedule), _) -> (i, front, rear, j, force new_schedule)
    | (_, []) -> queue
    | _ -> let f' = rotate front rear Nil in (i + j, f', [], 0, f')


let rec force_stream = function
  | Nil -> ()
  | Cons (_, tail) -> force_stream (force tail)

let force_front = function (_, front, _, _, _) -> force_stream front

let force_schedule = function (_, _, _, _, schedule) -> force_stream schedule

let count_forced_values = function (i, front, rear, j, sch) ->
  (i, Stream.count_forced front, List.size rear, j, Stream.count_forced sch)


(** Monadic operations **)


let unit = function x -> (1, Steam.force_stream (Stream.unit x), [], 0, Nil)

let map f = function (i, front, rear, j, _) ->
  let f' = Stream.map f front in (i, f', List.map f rear, j, f')

let plus = fun (i1, f1, r1, j1, _) (i2, f2, r2, j2, _) ->
  let f' = rotate f1 r1 f2 in (i1 + j1 + i2, f', r2, j2, f')

let zero = function () -> (0, Nil, [], 0, Nil)

let make = fun n x -> 
  let front = Stream.make n x in Stream.force_stream front ; 
    (n, front, [], 0, Nil)

let get (i, front, rear, j, _) = function
  | k when k < 0 ->
      raise (Invalid_argument "SizedScheduledList.get : argument ngatif")
  | k when k < i -> Stream.get front k
  | k when k < i + j -> List.get rear (j - k - 1)
  | k -> raise (Out_of_bounds k)

let set = fun (i, front, rear, j, schedule) k x ->
  match k with
  | _ when k < 0 ->
      raise (Invalid_argument "SizedScheduledList.set : argument ngatif")
  | _ when k <= i -> let f' = Stream.set front k x in (i, f', rear, j, f')
  | _ when k <= i + j -> (i, front, List.set rear (j - k - 1) x, j, schedule)
  | _ -> raise (Out_of_bounds k)
  

(** Constructors **)


let cons x = function (i, front, rear, j, schedule) ->
  (i + 1, Stream.cons x front, rear, j, schedule)

let snoc = fun (i, front, rear, j, schedule) x ->
  exec (i, front, x :: rear, j + 1, schedule)

let from_list = function list ->
  let front = Stream.from_list list in (List.size list, front, [], 0, Nil)

let init = fun n f ->
  let front = Stream.init n f in (n, front, [], 0, front)


(** Destructors **)


let left_view = function (i, front, rear, j, schedule) ->
  match front with
    | Nil -> raise Empty
    | Cons (x, tail) -> (x, exec (i - 1, force tail, rear, j, schedule))

let left_head = function
  | (_, Cons (x, _), _, _, _) -> x
  | _ -> raise Empty

let left_tail = function (i, front, rear, j, schedule) ->
  match front with
    | Nil -> raise Empty
    | Cons (_, tail) -> exec (i - 1, force tail, rear, j, schedule)

let right_head = function
  | (_, front, [], _, _) -> Stream.right_head front
  | (_, _, head :: _, _, _) -> head

let right_tail = function
  | (i, front, [], _, _) -> 
      let f' = Stream.right_tail front in (i - 1, f', [], 0, f')
  | (i, front, _ :: tail, j, sch) -> exec (i, front, tail, j - 1, sch)

let right_view = function
  | (i, front, [], _, _) -> 
      let (f', x) = Stream.right_view front in ((i - 1, f', [], 0, f'), x)
  | (i, front, head :: tail, j, sch) -> 
      (exec (i, front, tail, j - 1, sch), head)


(** Observers **)


let is_empty = function 
  | (0, _, _, _, _) -> true
  | _ -> false

let size = function (i, _, _, j, _) -> i + j

let rec to_list_rec result = fun stream list ->
  match stream with
    | Nil -> List.reverse_onto result list
    | Cons (x, next_stream) ->
	match list with
	  | [] -> Stream.to_list_append result stream
	  | head :: tail ->
	      x :: to_list_rec (head :: result) (force next_stream) tail
	      
let to_list = function (_, front, rear, _, _) -> to_list_rec [] front rear


(** Reverse **)


let reverse = function (i, front, rear, j, _) ->
  let f'= Stream.append_reverse (Stream.from_list rear) front in
    (i + j, f', [], 0, f')

let reverse_append = fun a b -> plus (reverse a) b
let reverse_onto = fun a b ->  plus (reverse b) a
let append_reverse = fun a b -> plus a (reverse b)


(** Folds and maps **)


let fold_right f e = function (_, front, rear, _, _) ->
  Stream.fold_right f (List.reverse_fold_right f e rear) front 

let fold_left f e = function (_, front, rear, _, _) ->
  List.reverse_fold_left f (Stream.fold_left f e front) rear

let rec reverse_iter f = function
  | [] -> ()
  | head :: tail -> reverse_iter f tail ; f head

let iter f = function (_, front, rear, _, _) ->
  Stream.iter f front ;
  reverse_iter f rear


(** Subsequences **)


let take = fun n (i, front, rear, j, schedule) ->
  match n with
    | _ when n < 0 -> 
	raise (Invalid_argument "SizedScheduledQueue.take : argument ngatif")
    | _ when n <= i -> let f' = Stream.take n front in (n, f', [], 0, f')
    | _ when n <= i + j -> let r' = List.drop (i + j - n) rear in
	(i, front, r', n - i, schedule)
    | _ -> raise (Out_of_bounds n)

let drop = fun n (i, front, rear, j, schedule) ->
  match n with
    | _ when n < 0 ->
	raise (Invalid_argument "SizedScheduledQueue.drop : argument ngatif")
    | _ when n <= i -> let f' = Stream.drop n front in
	(i - n, f', rear, j, f')
    | _ when n <= i + j ->
	exec (0, Nil, List.take (i + j - n) rear, i + j - n, Nil)
    | _ -> raise (Out_of_bounds n)

let split_at = fun n (i, front, rear, j, _) ->
  match n with
    | _ when n < 0 -> raise
	(Invalid_argument "SizedScheduledQueue.split_at : argument ngatif")
    | _ when n <= i -> let (f, f') = Stream.split_at n front in
	(exec (n, f, [], 0, f), exec (i - n, f', rear, j, f'))
    | _ when n <= i + j -> let (r', r) = List.split_at (i + j - n) rear in
	(exec (i, front, r, n - i, front), exec (0, Nil, r', i +  j - n, Nil))
    | _ -> raise (Out_of_bounds n)

let sub = fun st en queue -> from_list (List.sub st en (to_list queue))


