open BairePervasives

(* random acces lists in dense number representation *)
  
type 'a tree = 
  | L of 'a 
  | N of int * 'a tree * 'a tree

type 'a digit = Zero | One of 'a tree

type 'a seq = 'a digit list


(** Utils **)


let tree_size = function
  | L _ -> 1
  | N (n, _, _) -> n

let link = fun t1 t2 -> N (tree_size t1 + tree_size t2, t1, t2)

let rec insert_tree t = function
  | [] -> [One t]
  | Zero :: tail -> One t :: tail
  | One t' :: tail -> Zero :: insert_tree (link t t') tail

(* uncons renvoie toujours une feuille *)

let rec uncons = function
  | [] -> raise Empty
  | [One t] -> (t, [])
  | One t :: tail -> (t, Zero :: tail)
  | Zero :: tail ->
      match uncons tail with
        | (N (_, t1, t2), tail') -> (t1, One t2 :: tail')
        | _ -> raise (Unexpected_pattern "DenseRandomAccesList.uncons")


(** Monadic operations **)


let unit = function x -> [One (L x)]

let rec map_tree f = function
  | L x -> L (f x)
  | N (k, a, b) -> N (k, map_tree f a, map_tree f b)

let rec map f = function
  | [] -> []
  | Zero :: tail -> Zero :: map f tail
  | One t :: tail -> One (map_tree f t) :: map f tail

let rec plus = fun a b ->
  match (a, b) with
    | ([], _) -> b
    | (_, []) -> a
    | (Zero :: succ1, t :: succ2) -> t :: plus succ1 succ2
    | (t :: succ1, Zero :: succ2) -> t :: plus succ1 succ2
    | (One t1 :: succ1, One t2 :: succ2) ->
	Zero :: insert_tree (link t1 t2) (plus succ1 succ2)

let zero = function () -> []

let rec get_tree = fun n tree ->
  match tree with
    | L x when n = 0 -> x
    | L x -> raise (Out_of_bounds (n - 1))
    | N (k, t1, t2) ->
	if n < k / 2 then get_tree n t1
	else get_tree (n - k / 2) t2

let rec get n = function
  | [] -> raise (Out_of_bounds n)
  | Zero :: tail -> get n tail
  | One t :: tail ->
      if n < tree_size t then get_tree n t
      else get (n - tree_size t) tail
	  
let rec set_tree = fun n x tree -> 
  match tree with
    | L _ when n = 0 -> L x
    | L _ -> raise (Out_of_bounds (n - 1))
    | N (k, t1, t2) ->
	if n < k / 2 then N (k, set_tree n x t1, t2)
	else N (k, t1, set_tree (n - k / 2) x t2)

let rec set n x = function
  | [] -> raise (Out_of_bounds n)
  | Zero :: tail -> Zero :: set n x tail
  | One t :: tail ->
      if n < tree_size t then One (set_tree n x t) :: tail
      else One t :: set (n - tree_size t) x tail
  

(** Constructors **)


let cons = fun x rlist -> insert_tree (L x) rlist

let rec from_list_rec result = function
  | [] -> result
  | head :: tail -> from_list_rec (cons head result) tail

let from_list = function list -> from_list_rec [] (List.reverse list)


(** Destructors **)

let left_view = function rlist ->
  let (t, list) = uncons rlist in
    match t with
      | L x -> (x, list)
      | N _ -> raise (Unexpected_pattern "left_view")

let left_head = function rlist ->
  match uncons rlist with
    | (L x, _) -> x
    | _ -> raise Empty

let left_tail = function rlist -> let (_, tail) = uncons rlist in tail

let rec last_element_tree = function
  | L x -> x
  | N (_, _, a) -> last_element_tree a

let rec right_head = function
  | [] -> raise Empty
  | [One t] -> last_element_tree t
  | _ :: tail -> right_head tail

(*-- not working yet --*)

let rec rearrange result = function
  | L x -> cons x result
  | N (_, a, b) -> rearrange (rearrange result b) a

let rec right_tail_rec = function
  | L x -> ([], x)
  | N (_, a, L x) -> (rearrange [] a, x)
  | N (_, a, b) -> let (b', x) = right_tail_rec b in (rearrange b' a, x)

let rec right_tail = function
  | [] -> raise Empty
  | Zero :: tail -> right_tail tail
  | One t :: tail -> let (rtail, _) = right_tail_rec t in
      plus tail rtail


let is_empty = function
  | [] -> true
  | _ -> false

let rec size_rec result = function
  | [] -> result
  | Zero :: tail -> size_rec result tail
  | One t :: tail -> size_rec (result + tree_size t) tail

let size = function rlist -> size_rec 0 rlist

let rec to_list_tree result = function
  | L x -> x :: result
  | N (_, a, b) -> to_list_tree (to_list_tree result a) b

let rec to_list_rec result = function
  | [] -> List.reverse result
  | Zero :: tail -> to_list_rec result tail
  | One t :: tail -> to_list_rec (to_list_tree result t) tail

let to_list = function rlist -> to_list_rec [] rlist

