(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Two dimensional range sets (lexicographical order)
 * Ensembles bidimensionnels (ordre lexicographique)
 * Conjuntos bidimensionales (rden lexicogrfico)
 *)

open BairePervasives

type 'a two = ETwo | NTwo of 'a two * 'a * 'a two * int
type 'a set = E | N of 'a set * 'a * 'a set * 'a two * int


(** Utils **)
(** Utilitaires **)
(** tiles **)


let moduleName = "AvlRange2DSet"

let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type

let makeTree = fun l v r s -> 
  let h = match (l, r) with
    | (E, E) -> 1
    | (E, N (_, _, _, _, n)) | (N (_, _, _, _, n), E) -> 1 + n
    | (N (_, _, _, _, m), N (_, _, _, _, n)) -> 1 + max m n
  in
    N (l, v, r, s, h)

let rotate_L = fun l v r s ->
  match r with
    | N (rl, rv, rr, rs, _) -> makeTree (makeTree l v rl s) rv rr rs
    | _ -> makeTree l v r s
	
let rotate_R = fun l v r s ->
  match l with
    | N (ll, lv, lr, ls, _) -> makeTree ll lv (makeTree lr v r s) ls
    | _ -> makeTree l v r s

let rotate_RL = fun l v r s ->
  match r with
    | N (N (rll, rlv, rlr, rls, _), rv, rr, rs, _) -> 
	makeTree (makeTree l v rll s) rlv (makeTree rlr rv rr rs) rls
    | _ -> makeTree l v r s

let rotate_LR = fun l v r s ->
  match l with
    | N (ll, lv, N (lrl, lrv, lrr, lrs, _), ls, _) ->
	makeTree (makeTree ll lv lrl ls) lrv (makeTree lrr v r s) lrs
    | _ -> makeTree l v r s

let height = function
  | E -> 0
  | N (_, _, _, _, n) -> n

let makeBTree = fun l v r s ->
  let hl = height l and hr = height r in
    if hl > hr + 1 then
      (match l with
	 | E -> invalid "makeBTree" UnexpectedPattern
         | N (ll, _, lr, _, _) ->
             if height ll >= height lr then rotate_R l v r s
	     else rotate_LR l v r s
      )
    else if hr > hl + 1 then
      (match r with
	 | E -> invalid "makeBTree" UnexpectedPattern
         | N (rl, _, rr, _, _) ->
             if height rr >= height rl then rotate_L l v r s
	     else rotate_RL l v r s
      )
    else makeTree l v r s
      
let rec makeSTree = fun l v r s ->
  let hl = height l and hr = height r in
    if hl > hr + 2 then
      (match l with
	 | E -> invalid "makeSTree" UnexpectedPattern
	 | N (ll, lv, lr, ls, _) -> makeBTree ll lv (makeSTree lr v r s) ls
      )
    else if hr > hl + 2 then
      (match r with
	 | E -> invalid "makeSTree" UnexpectedPattern
	 | N (rl, rv, rr, rs, _) -> makeBTree (makeSTree l v rl s) rv rr rs
      )
    else makeBTree l v r s

let rec append = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l1, v1, r1, s1, _), N (l2, v2, r2, s2, _)) ->
        makeBTree l1 v1 (makeBTree (append r1 l2) v2 r2 s2) s1

let rec appendS = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l1, v1, r1, s1, _), N (l2, v2, r2, s2, _)) ->
        makeSTree l1 v1 (makeSTree (appendS r1 l2) v2 r2 s2) s1

let makeTwo = fun l v r -> 
  let h = match (l, r) with
    | (ETwo, ETwo) -> 1
    | (ETwo, NTwo (_, _, _, n)) | (NTwo (_, _, _, n), ETwo) -> 1 + n
    | (NTwo (_, _, _, m), NTwo (_, _, _, n)) -> 1 + max m n
  in
    NTwo (l, v, r, h)

let rotate_L_two = fun l v r ->
  match r with
    | NTwo (rl, rv, rr, _) -> makeTwo (makeTwo l v rl) rv rr
    | _ -> makeTwo l v r
	
let rotate_R_two = fun l v r ->
  match l with
    | NTwo (ll, lv, lr, _) -> makeTwo ll lv (makeTwo lr v r)
    | _ -> makeTwo l v r

let rotate_RL_two = fun l v r ->
  match r with
    | NTwo (NTwo (rll, rlv, rlr, _), rv, rr, _) -> 
	makeTwo (makeTwo l v rll) rlv (makeTwo rlr rv rr)
    | _ -> makeTwo l v r

let rotate_LR_two = fun l v r ->
  match l with
    | NTwo (ll, lv, NTwo (lrl, lrv, lrr, _), _) ->
	makeTwo (makeTwo ll lv lrl) lrv (makeTwo lrr v r)
    | _ -> makeTwo l v r

let height_two = function
  | ETwo -> 0
  | NTwo (_, _, _, n) -> n

let makeBTwo = fun l v r ->
  let hl = height_two l and hr = height_two r in
    if hl > hr + 1 then
      (match l with
	 | ETwo -> invalid "makeBTree" UnexpectedPattern
         | NTwo (ll, _, lr, _) ->
             if height_two ll >= height_two lr then rotate_R_two l v r 
	     else rotate_LR_two l v r
      )
    else if hr > hl + 1 then
      (match r with
	 | ETwo -> invalid "makeBTree" UnexpectedPattern
         | NTwo (rl, _, rr, _) ->
             if height_two rr >= height_two rl then rotate_L_two l v r 
	     else rotate_RL_two l v r
      )
    else makeTwo l v r
      
let rec makeSTwo = fun l v r ->
  let hl = height_two l and hr = height_two r in
    if hl > hr + 2 then
      (match l with
	 | ETwo -> invalid "makeSTree" UnexpectedPattern
	 | NTwo (ll, lv, lr, _) -> makeBTwo ll lv (makeSTwo lr v r)
      )
    else if hr > hl + 2 then
      (match r with
	 | ETwo -> invalid "makeSTree" UnexpectedPattern
	 | NTwo (rl, rv, rr, _) -> makeBTwo (makeSTwo l v rl) rv rr
      )
    else makeBTwo l v r

let rec append_two = fun a b ->
  match (a, b) with
    | (ETwo, _) -> b
    | (_, ETwo) -> a
    | (NTwo (l1, v1, r1, _), NTwo (l2, v2, r2, _)) ->
        makeBTwo l1 v1 (makeBTwo (append_two r1 l2) v2 r2)

let rec appendS_two = fun a b ->
  match (a, b) with
    | (ETwo, _) -> b
    | (_, ETwo) -> a
    | (NTwo (l1, v1, r1, _), NTwo (l2, v2, r2, _)) ->
        makeSTwo l1 v1 (makeSTwo (appendS_two r1 l2) v2 r2)


(** Categorial operations **)
(** Oprations catgorielles **)
(** Operaciones categoriales **)


let injection = fun x y -> makeTree E x E (makeTwo ETwo y ETwo)

let zero = function () -> E

let rec to_list_two x result = function
  | ETwo -> result
  | NTwo (l, y, r, _) -> to_list_two x ((x, y) :: to_list_two x result r) l

let rec to_list_one result = function
  | E -> result
  | N (l, v, r, s, _) -> 
      to_list_one (to_list_two v (to_list_one result r) s) l 

let to_list = function tree -> to_list_one [] tree

let rec insert_two y = function
  | ETwo -> makeTwo ETwo y ETwo
  | NTwo (l, v, r, _) ->
      match compare y v with
	| n when n < 0 -> makeBTwo (insert_two y l) v r
	| n when n > 0 -> makeBTwo l v (insert_two y r)
	| _ -> raise Found

let rec insert_raise x y = function 
  | E -> injection x y
  | N (l, v, r, s, _) ->
      match compare x v with
	 | n when n < 0 -> makeBTree (insert_raise x y l) v r s
	 | n when n > 0 -> makeBTree l v (insert_raise x y r) s
	 | _ -> makeTree l v r (insert_two y s)

let insert = fun x y tree ->
  try
    insert_raise x y tree
  with Found -> tree

let rec from_list_rec result = function
  | [] -> result
  | (x, y) :: tail -> from_list_rec (insert x y result) tail

let from_list = function list -> from_list_rec E list

let map f = function tree -> from_list (List.map f (to_list tree)) 

let rec partition_one pivot = function
  | E -> (E, None, E)
  | N (l, x, r, s, _) ->
      match compare pivot x with
	| n when n < 0 ->
	    let (inf, value, sup) = partition_one pivot l in
	      (inf, value, makeSTree sup x r s)
	| n when n > 0 ->
	    let (inf, value, sup) = partition_one pivot r in
	      (makeSTree l x inf s, value, sup)
	| _ -> (l, Some (x, s), r)

let rec partition_two pivot = function
  | ETwo -> (ETwo, None, ETwo)
  | NTwo (l, y, r, _) ->
      match compare pivot y with
	| n when n < 0 ->
	    let (inf, value, sup) = partition_two pivot l in
	      (inf, value, makeSTwo sup y r)
	| n when n > 0 ->
	    let (inf, value, sup) = partition_two pivot r in
	      (makeSTwo l y inf, value, sup)
	| _ -> (l, Some y, r)

let rec plus_two = fun a b -> 
  match (a, b) with
    | (ETwo, _) -> b
    | (_, ETwo) -> a
    | (NTwo (l, y, r, _), _) -> let (c, _, d) = partition_two y b in 
	makeSTwo (plus_two l c) y (plus_two r d)

let rec plus = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l, v, r, s, _), N _) ->
	match partition_one v b with
	  | (left, None, right) -> makeSTree (plus l left) v (plus r right) s
	  | (left, Some (_, s'), right) -> 
	      makeSTree (plus l left) v (plus r right) (plus_two s s')


(** Binary operations **)
(** Oprations binaires **)
(** Operaciones binrias **)


let union = plus

let rec intersection_two = fun a b ->
  match (a, b) with
    | (ETwo, _) | (_, ETwo) -> ETwo
    | (NTwo (l, v, r, _), NTwo _) ->
	match partition_two v b with
	  | (l', None, r') -> 
	      appendS_two (intersection_two l l') (intersection_two r r')
	  | (l', Some _, r') -> 
	      makeSTwo (intersection_two l l') v (intersection_two r r')

let rec intersection = fun a b ->
  match (a, b) with
    | (E, _) | (_, E) -> E
    | (N (l, v, r, s, _), N _) ->
	match partition_one v b with
	  | (l', None, r') -> append (intersection l l') (intersection r r')
	  | (l', Some (_, s'), r') ->
	      match intersection_two s s' with
		| ETwo -> appendS (intersection l l') (intersection r r')
		| s'' -> 
		    makeSTree (intersection l l') v (intersection r r') s''

let rec difference_two = fun a b ->
  match (a, b) with
    | (ETwo, _) -> ETwo
    | (_, ETwo) -> a
    | (NTwo (l, v, r, _), NTwo _) -> 
	match partition_two v b with
	  | (l', None, r') -> 
	      makeSTwo (difference_two l l') v (difference_two r r')
	  | (l', Some _, r') -> 
	      appendS_two (difference_two l l') (difference_two r r')
	      
let rec difference = fun a b ->  
  match (a, b) with
    | (E, _) -> E
    | (_, E) -> a
    | (N (l, v, r, s, _) , N _) ->
	match partition_one v b with
	  | (l', None, r') -> makeTree (difference l l') v (difference r r') s
	  | (l', Some (_, s'), r') ->
	      match difference_two s s' with
		| ETwo -> appendS (difference l l') (difference r r')
		| s'' -> makeSTree (difference l l') v (difference r r') s''

let rec symetric_difference_two = fun a b ->
  match (a, b) with
    | (ETwo, s) | (s, ETwo) -> s
    | (NTwo (l, v, r, _), NTwo _) ->
	match partition_two v b with
	  | (l', None, r') -> makeSTwo (symetric_difference_two l l') v 
	      (symetric_difference_two r r')
	  | (l', Some _, r') -> appendS_two (symetric_difference_two l l') 
	      (symetric_difference_two r r')

let rec symetric_difference = fun a b ->  
  match (a, b) with
    | (E, s) | (s, E) -> s
    | (N (l, v, r, s, _) , N _) ->
	match partition_one v b with
	  | (l', None, r') -> makeSTree (symetric_difference l l') v 
	      (symetric_difference r r') s
	  | (l', Some (_, s'), r') ->
	      match symetric_difference_two s s' with
		| ETwo -> appendS (symetric_difference l l') 
		    (symetric_difference r r')
		| s'' -> makeSTree (symetric_difference l l') v 
		    (symetric_difference r r') s''

let rec subset_two = fun a b ->
  match (a, b) with
    | (ETwo, _) -> true
    | (_, ETwo) -> false
    | (NTwo (l1, v1, r1, _), NTwo (l2, v2, r2, _)) ->
        match compare v1 v2 with
	  | n when n < 0 -> 
	      subset_two (makeTwo l1 v1 ETwo) l2 && subset_two r1 b
	  | n when n > 0 -> 
	      subset_two (makeTwo ETwo v1 r1) r2 && subset_two l1 b
	  | _ -> subset_two l1 l2 && subset_two r1 r2

let rec subset = fun a b ->
  match (a, b) with
    | (E, _) -> true
    | (_, E) -> false
    | (N (l1, v1, r1, s1, _), N (l2, v2, r2, s2, _)) ->
        match compare v1 v2 with
	  | n when n < 0 -> subset (makeTree l1 v1 E s1) l2 && subset r1 b
	  | n when n > 0 -> subset (makeTree E v1 r1 s1) r2 && subset l1 b
	  | _ -> subset_two s1 s2 && subset l1 l2 && subset r1 r2

let rec compare_two = fun a b ->
  match (a, b) with
    | ([], []) -> 0
    | ([], _) -> -1
    | (_, []) -> 1
    | (ETwo :: t1, ETwo :: t2) -> compare_two t1 t2
    | (NTwo (ETwo, v1, r1, _) :: t1, NTwo (ETwo, v2, r2, _) :: t2) ->
	(match compare v1 v2 with
	  | 0 -> compare_two (r1 :: t1) (r2 :: t2)
	  | other -> other
	)
    | (NTwo (a, v1, r1, _) :: t1, t2) ->
        compare_two (a :: NTwo (ETwo, v1, r1, 0) :: t1) t2
    | (t1, NTwo (b, v2, r2, _) :: t2) ->
	compare_two t1 (b :: NTwo (ETwo, v2, r2, 0) :: t2)

let rec compare_rec = fun a b ->
  match (a, b) with
    | ([], []) -> 0
    | ([], _) -> -1
    | (_, []) -> 1
    | (E :: t1, E :: t2) -> compare_rec t1 t2
    | (N (E, v1, r1, s1, _) :: t1, N (E, v2, r2, s2, _) :: t2) ->
	(match compare v1 v2 with
	   | n when n <> 0 -> n
	   | _ -> match compare_two [s1] [s2] with
	       | 0 -> compare_rec (r1 :: t1) (r2 :: t2)
	       | other -> other
	)
    | (N (a, v1, r1, s1, _) :: t1, t2) ->
        compare_rec (a :: N (E, v1, r1, s1, 0) :: t1) t2
    | (t1, N (b, v2, r2, s2, _) :: t2) ->
	compare_rec t1 (b :: N (E, v2, r2, s2, 0) :: t2)

let equal = fun a b -> compare_rec [a] [b] = 0


(** Acces **)
(** Accs **)
(** Acceso **)


let rec insert_list = fun list set ->
  match list with
    | [] -> set
    | (x, y) :: tail -> insert_list tail (insert x y set)

let rec insert_tree x t = function
  | E -> makeTree E x E t
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> makeTree (insert_tree x t l) v r s
	| n when n > 0 -> makeTree l v (insert_tree x t r) s
	| _ -> makeTree l v r (plus_two t s)

let rec delete_two y = function
  | ETwo -> raise Not_found
  | NTwo (l, v, r, _) ->
      match compare y v with
	| n when n < 0 -> makeBTwo (delete_two y l) v r
	| n when n > 0 -> makeBTwo l v (delete_two y r)
	| _ -> append_two l r

let rec delete_raise x y = function
  | E -> raise Not_found
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> makeBTree (delete_raise x y l) v r s
	| n when n > 0 -> makeBTree l v (delete_raise x y r) s
	| _ -> 
	    match delete_two y s with
	      | ETwo -> append l r
	      | s' -> makeTree l v r s'

let delete = fun x y tree ->
  try 
    delete_raise x y tree
  with Not_found -> tree

let rec delete_tree_raise x = function
  | E -> raise Not_found
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> makeBTree (delete_tree_raise x l) v r s
	| n when n > 0 -> makeBTree l v (delete_tree_raise x r) s
	| _ -> append l r

let rec delete_tree = fun t set ->
  try
    delete_tree_raise t set 
  with Not_found -> set

let rec delete_list = fun list set ->
  match list with
    | [] -> set
    | (x, y) :: tail -> delete_list tail (delete x y set)

let rec delete_min_two = function
  | ETwo -> ETwo
  | NTwo (ETwo, v, r, _) -> r
  | NTwo (l, v, r, _) -> makeBTwo (delete_min_two l) v r

let rec delete_min = function
  | E -> raise Empty
  | N (E, v, r, s, _) -> 
      (match delete_min_two s with
	 | ETwo -> r
	 | s' -> makeSTree E v r s' 
      )
  | N (l, v, r, s, _) -> makeBTree (delete_min l) v r s

let rec delete_max_two = function
  | ETwo -> ETwo
  | NTwo (l, v, ETwo, _) -> l
  | NTwo (l, v, r, _) -> makeBTwo l v (delete_max_two r)

let rec delete_max = function
  | E -> raise Empty
  | N (l, v, E, s, _) ->
      (match delete_max_two s with
	 | ETwo -> l
	 | s'-> makeBTree l v E s'
      )
  | N (l, v, r, s, _) -> makeBTree l v (delete_max r) s

let rec delete_min_tree = function
  | E -> raise Empty
  | N (E, _, r, _, _) -> r
  | N (l, v, r, s, _) -> makeBTree (delete_min_tree l) v r s

let rec delete_max_tree = function
  | E -> raise Empty
  | N (l, _, E, _, _) -> l
  | N (l, v, r, s, _) -> makeBTree l v (delete_max_tree r) s

let rec filter_two f result = function
  | ETwo -> result
  | NTwo (l, v, r, _) ->
      filter_two f 
      (filter_two f 
	  (if f v then insert_two v result 
	   else result) 
	  l) 
      r

let rec filter f = function
  | E -> E
  | N (l, v, r, s, _) ->
      match filter_two f ETwo s with
	| ETwo -> appendS (filter f l) (filter f r)
	| s' -> makeSTree (filter f l) v (filter f r) s'

let rec filter_lt_two y = function
  | ETwo -> ETwo
  | NTwo (l, v, r, _) ->
      match compare y v with
	| n when n < 0 -> filter_lt_two y l
	| n when n > 0 -> makeSTwo l v (filter_lt_two y r)
	| _ -> l

let rec filter_gt_two y = function
  | ETwo -> ETwo
  | NTwo (l, v, r, _) ->
      match compare y v with
	| n when n < 0 -> makeSTwo (filter_gt_two y l) v r
	| n when n > 0 -> filter_gt_two y r
	| _ -> r

let rec filter_le_two x = function
  | ETwo -> ETwo
  | NTwo (l, v, r, _) ->
      match compare x v with
	| n when n < 0 -> filter_le_two x l
	| _ -> makeSTwo l v (filter_le_two x r)

let rec filter_ge_two x = function
  | ETwo -> ETwo
  | NTwo (l, v, r, _) ->
      match compare x v with
 	| n when n <= 0 -> makeSTwo (filter_ge_two x l) v r
	| _ -> filter_ge_two x r

let rec filter_lt x y = function
  | E -> E
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> filter_lt x y l
	| n when n > 0 -> makeSTree l v (filter_lt x y r) s
	| _ -> 
	    match filter_lt_two y s with
	      | ETwo -> l
	      | s' -> makeSTree l v E s'

let rec filter_gt x y = function
  | E -> E
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> makeSTree (filter_gt x y l) v r s
	| n when n > 0 -> filter_gt x y r
	| _ ->
	    match filter_gt_two y s with
	      | ETwo -> r
	      | s' -> makeSTree E v r s'

let rec filter_le x y = function
  | E -> E
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> filter_le x y l
	| _ -> 
	    match filter_le_two y s with
	      | ETwo -> l
	      | s' -> makeSTree l v E s'

let rec filter_ge x y = function
  | E -> E
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n <= 0 -> makeSTree (filter_ge x y l) v r s
	| _ ->
	    match filter_ge_two y s with
	      | ETwo -> r
	      | s' -> makeSTree E v r s'

let rec range_two min max = function
  | ETwo -> ETwo
  | NTwo (l, v, r, _) ->
      if compare v min < 0 then range_two min max r
      else if compare v max > 0 then range_two min max l
      else makeSTwo (filter_ge_two min l) v (filter_le_two max r)

let rec map_over_two f = function
  | E -> E
  | N (l, v, r, s, _) -> 
      match f s with
	| ETwo -> append (map_over_two f l) (map_over_two f r)
	| s' -> makeTree (map_over_two f l) v (map_over_two f r) s'

let rec filter_range_le f x = function
  | E -> E
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> filter_range_le f x l
	| n when n > 0 -> makeSTree (map_over_two f l) v 
	    (filter_range_le f x r) s
	| _ ->
	    match f s with
	      | ETwo -> map_over_two f l
	      | s' -> makeSTree (map_over_two f l) v E s'

let rec filter_range_ge f x = function
  | E -> E
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> makeSTree (filter_range_ge f x l) v 
	    (map_over_two f r) s
	| n when n > 0 -> filter_range_le f x r
	| _ ->
	    match f s with
	      | ETwo -> map_over_two f r
	      | s' -> makeSTree E v (map_over_two f r) s'

let rec range_one f min_x max_x = function
  | E -> E
  | N (l, v, r, s, _) ->
      if compare v min_x < 0 then range_one f min_x max_x r
      else if compare v max_x > 0 then range_one f min_x max_x l
      else match f s with
	| ETwo -> appendS (filter_range_ge f min_x l) 
	    (filter_range_le f max_x r)
	| s' -> makeSTree (filter_range_ge f min_x l) v 
	    (filter_range_le f max_x r) s'

let range = fun min_x max_x min_y max_y tree ->
  range_one (range_two min_y max_y) min_x max_x tree


(** View **)
(** Vues **)
(** Vistas **)


let partition_inf_sup x y = function tree ->
  match partition_one x tree with
    | (_, None, _) as part -> part
    | (l, Some (v, s), r) ->
	let (ll, value, rr) = partition_two y s in
	let 
	  left = match ll with
	    | ETwo -> l
	    | ll -> makeSTree l v E ll
	and
	  right = match rr with
	    | ETwo -> r
	    | rr -> makeSTree E v r rr
	and
	  center = match value with
	    | None -> None
	    | Some v' -> Some (v, v')
	in
	  (left, center, right)	

let partition_inf_sup_tree = partition_one


(** Observers **)
(** Observateurs **)
(** Observadores **)


let is_empty = function
  | E -> true
  | N _ -> false

let rec size_two total = function
  | ETwo -> total
  | NTwo (l, _, r, _) -> size_two (size_two (1 + total) l) r

let rec size_one total = function
  | E -> total
  | N (l, _, r, s, _) -> size_one (size_one (size_two total s) l) r

let size = function tree -> size_one 0 tree

let rec member_two y = function
  | ETwo -> false
  | NTwo (l, v, r, _) ->
      match compare y v with
	| n when n < 0 -> member_two y l
	| n when n > 0 -> member_two y r
	| _ -> true

let rec member x y = function
  | E -> false
  | N (l, v, r, s, _) ->
      match compare x v with
	 | n when n < 0 -> member x y l
	 | n when n > 0 -> member x y r
	 | _ -> member_two y s

let rec look_for_two f = function
  | ETwo -> raise Not_found
  | NTwo (_, v, _, _) when f v -> v
  | NTwo (l, _, r, _) -> 
      try look_for_two f l 
      with Not_found -> look_for_two f r

let rec look_for f = function
  | E -> raise Not_found
  | N (l, v, r, s, _) ->
      try look_for_two (f v) s with
	  Not_found ->
	    try look_for f l with 
		Not_found -> look_for f r
	      
let choose = function
  | E -> raise Empty
  | N (_, x, _, NTwo (_, y, _, _), _) -> (x, y)
  | _ -> invalid "choose" UnexpectedPattern

let rec min_element_two = function
  | ETwo -> raise Not_found
  | NTwo (ETwo, v, _, _) -> v
  | NTwo (l, _, _, _) -> min_element_two l

let rec min_element = function
  | E -> raise Not_found
  | N (E, v, _, s, _) -> (v, min_element_two s)
  | N (l, _, _, _, _) -> min_element l

let rec max_element_two = function
  | ETwo -> raise Not_found
  | NTwo (_, v, ETwo, _) -> v
  | NTwo (_, _, r, _) -> max_element_two r

let rec max_element = function
  | E -> raise Not_found
  | N (_, v, E, s, _) -> (v, max_element_two s)
  | N (_, _, l, _, _) -> max_element l

let rec min_tree = function
  | E -> raise Not_found
  | N (E, v, _, s, _) -> (v, s)
  | N (l, _, _, _, _) -> min_tree l

let rec max_tree = function
  | E -> raise Not_found
  | N (_, v, E, s, _) -> (v, s)
  | N (_, _, r, _, _) -> max_tree r

let rec next_two y = function
  | ETwo -> raise Not_found
  | NTwo (l, v, r, _) -> 
      match compare y v with
	| n when n < 0 -> (try next_two y l with Not_found -> v)
	| n when n > 0 -> next_two y r
	| _ -> min_element_two r (* [Not_found] *)

let rec next x y = function
  | E -> raise Not_found
  | N (l, v, r, s, _) -> 
      match compare x v with
	| n when n < 0 -> 
	    (try next x y l with Not_found -> (v, min_element_two s))
	| n when n > 0 -> next x y r
	| _ -> try (v, next_two y s) with Not_found -> min_element r

let rec next_tree x = function
  | E -> raise Not_found
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> (try next_tree x l with Not_found -> (v, s))
	| n when n > 0 -> next_tree x r
	| _ -> min_tree r

let rec previous_tree x = function
  | E -> raise Not_found
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> previous_tree x l
	| n when n > 0 -> (try previous_tree x r with Not_found -> (v, s))
	| _ -> max_tree l (* raises [Not_found] *)
	      
let rec ceil_tree x = function
  | E -> raise Not_found
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> (try ceil_tree x l with Not_found -> (v, s))
	| n when n > 0 -> ceil_tree x r
	| _ -> (v, s)

let rec floor_tree x = function
  | E -> raise Not_found
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> floor_tree x l
	| n when n > 0 -> (try floor_tree x r with Not_found -> (v, s))
	| _ -> (v, s)

let rec previous_two y = function
  | ETwo -> raise Not_found
  | NTwo (l, v, r, _) ->
      match compare y v with
	| n when n < 0 -> previous_two y l
	| n when n > 0 -> (try previous_two y r with Not_found -> v)
	| _ -> max_element_two l (* [Not_found] *)

let rec previous x y = function
  | E -> raise Not_found
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> previous x y l
	| n when n > 0 -> 
	    (try previous x y r with Not_found -> (v, max_element_two s))
	| _ -> try (v, previous_two y s) with Not_found -> max_element l

let rec ceil_two y = function
  | ETwo -> raise Not_found
  | NTwo (l, v, r, _) -> 
      match compare y v with
	| n when n < 0 -> (try ceil_two y l with Not_found -> v)
	| n when n > 0 -> ceil_two y r
	| _ -> v

let rec ceil x y = function
  | E -> raise Not_found
  | N (l, v, r, s, _) -> 
      match compare x v with
	| n when n < 0 -> 
	    (try ceil x y l with Not_found -> (v, min_element_two s))
	| n when n > 0 -> ceil x y r
	| _ -> try (v, ceil_two y s) with Not_found -> min_element r

let rec floor_two y = function
  | ETwo -> raise Not_found
  | NTwo (l, v, r, _) ->
      match compare y v with
	| n when n < 0 -> floor_two y l
	| n when n > 0 -> (try floor_two y r with Not_found -> v)
	| _ -> v

let rec floor x y = function
  | E -> raise Not_found
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> floor x y l
	| n when n > 0 -> 
	    (try floor x y r with Not_found -> (v, max_element_two s))
	| _ -> try (v, floor_two y s) with Not_found -> max_element l

let rec rank_two y total = function
  | ETwo -> total
  | NTwo (l, v, r, _) ->
      match compare y v with
	| n when n < 0 -> rank_two y total l
	| n when n > 0 -> rank_two y (size_two ( 1 + total) l) r
	| _ -> size_two (1 + total) l

let rec rank_rec x y total = function
  | E -> total
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> rank_rec x y total l
	| n when n > 0 -> rank_rec x y (size_one (size_two total s) l) r
	| _ -> size_one (rank_two y total s) l

let rank = fun x y tree -> rank_rec x y 0 tree

let rec rank_tree_rec x total = function
  | E -> total
  | N (l, v, r, s, _) ->
      match compare x v with
	| n when n < 0 -> rank_tree_rec x total l
	| n when n > 0 -> rank_tree_rec x (size_one (1 + total) l) r
	| _ -> size_one total l

let rank_tree = fun x set -> rank_tree_rec x 0 set

let rec get_two n = function
  | ETwo -> raise (Out_of_bounds n)
  | NTwo (a, x, b, _) ->
      let k = size_two 0 a in
	if n < k then get_two n a
	else if n > k then get_two (n - k - 1) b
	else x

let rec exists_two f = function
  | ETwo -> false
  | NTwo (l, y, r, _) -> f y || exists_two f l || exists_two f r

let rec exists f = function
  | E -> false
  | N (l, x, r, s, _) -> exists_two (f x) s || exists f l || exists f r

let rec for_all_two f = function
  | ETwo -> true
  | NTwo (l, y, r, _) -> f y && for_all_two f l && for_all_two f r

let rec for_all f = function
  | E -> true
  | N (l, x, r, s, _) -> for_all_two (f x) s && for_all f l && for_all f r

let rec iter_two f = function
  | ETwo -> ()
  | NTwo (l, y, r, _) -> iter_two f l ; f y ; iter_two f r

let rec iter f = function
  | E -> ()
  | N (l, x, r, s, _) -> iter f l ; iter_two (f x) s ; iter f r

let rec fold_left_two f e = function
  | ETwo -> e
  | NTwo (l, v, r, _) -> fold_left_two f (f (fold_left_two f e l) v) r

let rec fold_left f e = function
  | E -> e
  | N (l, x, r, s, _) ->
      fold_left f (fold_left_two (fun a y -> f a x y) (fold_left f e l) s) r

let rec fold_right_two = fun f set e ->
  match set with
    | ETwo -> e 
    | NTwo (l, v, r, _) -> fold_right_two f l (f v (fold_right_two f r e))

let rec fold_right = fun f set e ->
  match set with
    | E -> e
    | N (l, x, r, s, _) -> 
	fold_right f l (fold_right_two (f x) s (fold_right f r e))

let fold = fold_right


(** Compatibilit avec la librairie standard de Caml **)
(** Caml standard library compatibility **)
(** Compatibilidad con la librera estandar de Caml **)


let empty = E
let mem = member
let add = insert
let singleton = injection
let remove = delete
let inter = intersection
let diff = difference
let cardinal = size
let elements = to_list
let min_elt = min_element    
let max_elt = max_element
