(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * (Local) red-black set
 * Ensemble bicolore (version locale)
 * Conjunto bicromtico (versin local)
 *
 *)

open BairePervasives

(* Plan :
 *
 * 1. Introduction to bichromatic framework and links between the global
 *    red-black tree expression (RedBlackSet module) and local expression
 *    (traditional presentation by Sedgewick and Guibas)
 * 2. Polymorphic code and comments explaining various algorithms
 * 3. Bibliography
 * 4. Make functor
 *)
 
(* Plan :
 *
 * 1. Introduction du formalisme bicolore, lien entre l'expression globale des
 *    arbres bicolores (module RedBlackSet) et l'expression locale 
 *    (prsentation traditionnelle par Sedgewick et Guibas)
 * 2. Code polymorphe avec commentaires expliquant les divers algorithmes
 * 3. Bibliographie
 * 4. Foncteur Make
 *)

(* Plan :
 *
 * 1. Introduccin al formalisme bicromtico, relacion entre la expresin 
 *    global de los rboles bicromticos (modulo RedBlackSet) y la expresin 
 *    local (presentacin tradicional por Sedgewick y Guibas)
 * 2. Cdico polimrfico y comentrios que explican varios algoritmos
 * 3. Bibliografa
 * 4. Functor Make
 *)

(* Let suppose we have a cartesian tree, we will first of all draw it by layers
 * (see figure)
 * 
 * 
 *         (2, 3)                       2        level 3
 *          /  \                       / \
 *      (1, 1) (6, 2)       ->        /   6      level 2
 *              /  \                 /   / \
 *          (7, 1) (9, 1)           1   7   9    level 1
 * 
 * Then we will color it in the following way :
 * - if a node is on the same level than his father, it is a RED node
 * - if a node is not on the same level than his father, it is a BLACK node
 * 
 * We will say that we have a  red conflict  when there are two successive
 * red nodes, a  black overweight  when there is an unoccupied level
 * 
 *  level 3              2            Black
 *                      / \
 *  level 2            /   6          Black
 *                    /     \ 
 *    black          /       8        Red  
 *    overweight    /       / \              red conflict
 *                 /       /   10     Red
 *                /       /   / \
 *  level 1     1        7   9  11    Black
 * 
 * 
 * We now take a red-black tree, that is a colored binary tree with the 
 * following properties (Sedgewick and Guibas):
 * - all leafs are black
 * - every path from a node to a leaf must contain the same number of 
 * black nodes
 * - there cannot be two successive red nodes
 * 
 * We will associate to every node his  black height  that is the number of
 * black nodes between the node and the leafs. This augmented tree can be
 * then seen as if it was a cartesian tree.
 *
 * Drawing this cartesian (red-black) tree in layers, one can translate 
 * red black properties on the following:
 * - terminal nodes lie in level 1
 * - no unoccupied level (no black overweight)
 * - no more than two succesive nodes by level (no red conflict)
 * 
 * If we colour back the cartesian tree, we do not always obtain the same 
 * red-black tree, not even a red-black tree. This will be explained
 * later
 *)

(* Considrons un arbre cartsien, la premire opration que nous allons
 * effectuer est de le dessiner en couches (voir figure)
 * 
 * 
 *         (2, 3)                       2        niveau 3
 *          /  \                       / \
 *      (1, 1) (6, 2)       ->        /   6      niveau 2
 *              /  \                 /   / \
 *          (7, 1) (9, 1)           1   7   9    niveau 1
 * 
 * Ensuite nous allons le colorier de la faon suivante
 * - si un noeud est sur le mme niveau que son pre il est ROUGE
 * - si un noeud n'est pas sur le mme niveau que son pre il est NOIR
 * 
 * Nous dirons qu'on a un  conflit rouge  si deux noeud rouges se succdent,
 * et qu'on a un  surpoids noir  si un niveau n'est pas occup
 * 
 *  niveau 3             2           Noir 
 *                      / \ 
 *  niveau 2           /   6         Noir 
 *                    /     \ 
 *   surpoids        /       8       Rouge  
 *    noir          /       / \              conflit rouge
 *                 /       /   10    Rouge
 *                /       /   / \
 *  niveau 1    1        7   9  11   Noir
 * 
 * 
 * Maintenant nous prenons un arbre rouge et noir, autrement dit vrifiant
 * - les feuilles sont noires
 * - entre tout noeud et les feuilles il y a le mme nombre de noeuds noirs
 * - il n'y a pas deux noeuds rouges successifs
 * 
 * Nous allons associer  chaque noeud son  hauteur noire  (le nombre de
 * noeud noirs entre lui et les feuilles) et le considrer alors comme
 * un arbre cartsien
 * 
 * Nous allons ensuite le dessiner en couches. La traduction des proprits
 * des arbres bicolores en termes de couches donne
 * - les noeuds terminaux sont de niveau 1
 * - il n'y a pas de niveau non occup (pas de surpoids noir)
 * - il n'y a pas de niveau contenant plus de 2 noeuds (pas de conflit rouge)
 * 
 * Quand on colore l'arbre cartsien ainsi obtenu, on peut ne pas retrouver
 * l'arbre bicolore initial, voir pas mme un arbre bicolore. Ce phnomne
 * sera expliqu ultrieurement.
 *)

(* Consideremos un rbol cartesiano, la primera transformacin que le vamos
 * a hacer es dibujarlo por capas (ver figura)
 * 
 * 
 *         (2, 3)                       2        nivel 3
 *          /  \                       / \
 *      (1, 1) (6, 2)       ->        /   6      nivel 2
 *              /  \                 /   / \
 *          (7, 1) (9, 1)           1   7   9    nivel 1
 * 
 * Despues lo vamos a colorear de la manera siguiente
 * - si un nudo se encuentra en el mismo nivel que su padre, es un nudo ROJO
 * - si un nodo no se encuentra en el mismo nivel que su padre, es NEGRO
 * 
 * Diremos que hay un  conflicto rojo  si dos nudos rojos se siguen, 
 * y diremos que hay un  sobrepeso negro  cuando entre un nudo y su padre 
 * hay uno o varios niveles vacos (ver el ejemplo)
 * 
 *  nivel 3              2          Negro
 *                      / \
 *  nivel 2            /   6        Negro
 *                    /     \ 
 *   sobrepeso       /       8      Rojo   
 *    negro         /       / \              conflicto rojo
 *                 /       /   10   Rojo
 *                /       /   / \
 *  nivel 1     1        7   9  11  Negro
 * 
 * 
 * Considerar ahora un rbolo bicrmatico es decir que verifique las 
 * propriedades :
 * - todas las hojas son negras
 * - hay el mismo nmero de nudos negros entre un nudo dado y las hojas
 * - dos nudos rojos no pueden seguirse
 * 
 * Vamos ahora a asociar a cada nudo su  altura negra  (es el numero de nudos
 * negros entre el nudo dado y las hojas) y considerar el rbol asi aumentado
 * como si fuese un rbol cartesiano
 * 
 * Luego, lo dibujaremos por capas. La traduccin de las propriedades de los
 * rboles bicromticos en terminos de capas es :
 * - los nudos terminales estan en en nivel 1
 * - no hay nivel que no est ocupado (no hay sobrepeso negro)
 * - no hay nivel conteniendo mas de dos nudos (no hay conflicto rojo)
 * 
 * Si uno colorea el rbol cartesiano, no siempre se obtiene el rbol 
 * bicromtico inicial, incluso puede no ser un rbol bicromtico. Explicaremos
 * ste fenmeno luego.
 *)

(* Examining the RedBlackSet module one can see that this is exactly what
 * the given constructors do, in other words the constructed tree is a 
 * red-black tree in the sense of Sedgewick and Guibas
 * 
 * the [makeTree] function makes sure that
 * - terminal nodes lie in level 1
 * - there are no unoccupied levels
 *)

(* Si l'on examine le module RedBlackSet on se rend compte que c'est
 * exactement les conditions que vrifient les constructeurs
 * donns (autrement dit l'arbre construit est bien un arbre rouge et noir
 * tel que dfini par Sedgewick et Guibas) :
 * 
 * la fonction [makeTree] s'assure que :
 * - les noeuds terminaux sont de niveau 1
 * - il n'y a pas de niveau non occup
 *)

(* Si uno examina el modulo RedBlackSet uno se da cuenta que estas son 
 * exactamente las condiciones que los constructores dados garantizan 
 * (en otras palabras los rboles construidos por ese mdulo son arboles rojos 
 * y negros en el sentido definido por Sedgewick y Guibas)
 * 
 * la funcin [makeTree] se asegura que :
 * - los nudos terminales 
 * - no hay nivel vaco
 *)

type 'a tree = E | N of 'a tree * 'a * 'a tree * int

let makeTree = fun l v r ->
  let h = match (l, r) with
    | (E, _) | (_, E) -> 1
    | (N (_, _, _, m), N (_, _, _, n)) -> 1 + min m n
  in
    N (l, v, r, h)

let height = function x -> 0 
let rotate_L = makeTree  (* for the system not to complain *)
let rotate_R = makeTree  (* pour que le typeur ne se plaigne pas *)
let rotate_RL = makeTree (* para que el motor de inferencia no se queje *)
let rotate_LR = makeTree

(* the [makeBTree] function makes sure that there is no red conflict,
 * doing if needed rotations to eliminate them
 *)

(* la fonction [makeBTree] s'assure qu'il n'y a pas de conflit rouge
 * oprant si besoin est  des rotations pour les liminer
 *)

(* la funcin [makeBTree] se asegura que no hay conflicto rojo, operando
 * rotacines si necesario para eliminarlos
 *)

let rec makeBTree = fun l v r ->
  let hl = height l and hr = height r in
    match compare hl hr with
      | n when n < 0 ->
	  (match r with
	     | E -> makeTree l v r
	     | N (rl, rv, rr, _) ->
		 if height rr = hr then rotate_L l v r (* red conflict *)
		 else if height rl = hr then rotate_RL l v r (* red conflict *)
		 else makeTree l v r
	  )
      | n when n > 0 ->
	  (match l with
	     | E -> makeTree l v r
	     | N (ll, lv, lr, _) ->
		 if height ll = hl then rotate_R l v r (* red conflict *)
		 else if height lr = hl then rotate_LR l v r (* red conflict *)
		 else makeTree l v r
	  )
      | _ -> makeTree l v r

(* Conversely, all operations we did int the  absolute representation  
 * can be done with the  local representation 
 * 
 * type of tree    global information         local information
 * 
 * AVL             height (integer)           height difference (-1, 0, 1)
 * Bichromatic     black height (integer)     colour (Red or Black)
 * 
 * - local operations are faster (insertions, deletions)
 * - global operations are slower (unions, intersection)
 * - most algorithms are more complex
 * 
 * Notice that in the bichromatic formalism, we loose the information we had
 * on  black overweight , and in most algorithms this information will have
 * to be propagated explicitely
 *)

(* Inversement, toutes les oprations que nous avons effectues avec
 * la  reprsentation absolue  des arbres bicolores, nous pouvons les
 * faire dans la  reprsentation locale  (disposant seulement des
 * couleurs des noeuds).
 * 
 * type d'arbre    information globale        information locale
 * 
 * AVL             hauteur (entier)           diffrence de hauteur (-1, 0, 1)
 * Bicolore        hauteur noire (entier)     couleur (Rouge ou Noir)
 *
 * - les oprations locales sont plus rapides (insertions, suppressions)
 * - les oprations globales son plus lentes (union, intersection)
 * - les algorithmes sont beaucoup plus compliqus
 * 
 * Notons tout particulirement que dans le formalisme bicolore nous perdons
 * l'information dont nous disposons sur le  surpoids noir , et qu'il
 * va souvent falloir dans les algorithmes propager explicitement cette 
 * information
 *)

(* Inversamente, todas las operaciones que hicimos en la  representacin 
 * absoluta  de los rboles bicromticos, las podemos hacer con la 
 *  representacin local 
 * 
 * typo de rbol     informacin global       informacin local
 * 
 * AVL               altura (entero)          diferencia de altura (-1, 0, 1)
 * Bicromtico       altura negra (entero)    color (Rojo o Negro)
 * 
 * - las operaciones locales son ms rpidas (inserciones, supresiones)
 * - las operaciones globales son ms lentas (uniones, intersecciones)
 * - los algoritmos son mas complicados
 * 
 * Notese que en el formalismo bicromtico se ha perdido la informacin sobre
 * el  sobrepeso negro , razn por la cual en varios algoritmos se habra de
 * propagar esa informacin de manera explicita
 *)

type colour = Red | Black

type 'a set =
  | E
  | N of colour * 'a set * 'a * 'a set

type overweight = Yes | No

(* We first present some classical algorithms.
 * Further we will show how [makeBTree] constructor can be adapted to local
 * red-black tree
 *)

(* Nous prsentons d'abord quelques algorithmes classiques.
 * Plus loin montrerons-nous comment adapter [makeBTree]  la formulation 
 * locale des arbres bicolores 
 *)

(* Presentaremos primero algunos algoritmos clasicos.
 * Ms adelante mostraremos cmo se puede adaptar el constructor [makeBTree]
 * a los rboles bicromticos en su formulacin local
 *)

(* This is a very simple insertion algorithm by Chris Okasaki. It has the
 * property of systematically propagationg the red conflict (that means
 * that it always correct the red conflicts by rotations even if one could
 * have done it just by changing the color of one or more nodes)
 *)

(* Voici un premier algorithme trs simple d'insertion, propos par
 * Chris Okasaki. Il a la caractristique de  propager systmatiquement
 * le conflit rouge vers le haut  (cela signifie qu'il corrige toujours 
 * les conflits rouges par rotations, mme dans les situations o on aurait 
 * pu y arriver juste en changeant les couleurs d'un ou plusieurs noeuds)
 *)

(* Este algoritmo simple fue propuesto por Chris Okasaki. Tiene la propriedad
 * de propagar sistemticamente el conflicto rojo (lo que significa que siempre
 * corrige los conflictos rojos haciendo rotaciones, incluso cuando se hubiese
 * podido resolver con solo cambiar el color de uno o varios nudos)
 *)

let red_conflict_simple = function
  | (Black, N (Red, N (Red, a, x, b), y, c), z, d)
  | (Black, a, x, N (Red, N (Red, b, y, c), z, d))
  | (Black, N (Red, a, x, N (Red, b, y, c)), z, d)
  | (Black, a, x, N (Red, b, y, N (Red, c, z, d))) ->
      N (Red, N (Black, a, x, b), y, N (Black, c, z, d))
  | (a, b, c, d) -> N (a, b, c, d)

let rec insert_I_rec x = function
  | E -> N (Red, E, x, E)
  | N (colour, a, y, b) ->
      match compare x y with
	| n when n < 0 -> 
	    red_conflict_simple (colour, insert_I_rec x a, y, b)
	| n when n > 0 -> 
	    red_conflict_simple (colour, a, y, insert_I_rec x b)
	| _ -> raise Found

let insert_I = fun x set ->
  try 
    (match insert_I_rec x set with
       | N (Red, a, y, b) -> N (Black, a, y, b)
       | other -> other
    )
  with Found -> set

(* This time the red conflict is only checked in the path to the root (on two
 * levels), that is why we obtain 4 cases
 *)

(* Cette fois-ci l'existence d'un ventuel conflit rouge n'est vrifie
 * que le long du chemin vers la racine sur deux niveaux,
 * d'o les 4 cas possibles
 *)

(* sta vez la presencia de un conflicto rojo slo es verificada a lo largo del
 * camino hacia la raz (sobre dos niveles), razn por la cual se obtienen
 * 4 casos
 *)

let red_conflict_LL = fun ll x lr y r ->
  match ll with
    | N (Red, lll, z, llr) -> 
	N (Red, N (Black, lll, z, llr), x, N (Black, lr, y, r))
    | _ -> N (Black, N (Red, ll, x, lr), y, r)

let red_conflict_LR = fun ll x lr y r ->
  match lr with
    | N (Red, lrl, z, lrr) ->
	N (Red, N (Black, ll, x, lrl), z, N (Black, lrr, y, r))
    | _ -> N (Black, N (Red, ll, x, lr), y, r)

let red_conflict_RL = fun l x rl y rr ->
  match rl with
    | N (Red, rll, z, rlr) ->
	N (Red, N (Black, l, x, rll), z, N (Black, rlr, y, rr))
    | _ -> N (Black, l, x, N (Red, rl, y, rr))

let red_conflict_RR = fun l x rl y rr ->
  match rr with
    | N (Red, rrl, z, rrr) -> 
	N (Red, N (Black, l, x, rl), y, N (Black, rrl, z, rrr))
    | _ -> N (Black, l, x, N (Red, rl, y, rr))

(* the adapted insertion function *)

(* la fonction d'insertion est adapte en consquence *)

(* la funcin de insercin adaptada en consecuencia *)

let rec insert_II_rec x = function
  | E -> N (Red, E, x, E)
  | N (_, l, y, r) ->
      match compare x y with
	| n when n < 0 ->
	    (match l with
	       | N (Red, ll, z, lr) ->
		   (match compare x z with
		      | n when n < 0 -> 
			  red_conflict_LL (insert_II_rec x ll) z lr y r
		      | n when n > 0 -> 
			  red_conflict_LR ll z (insert_II_rec x lr) y r
		      | _ -> raise Found
		   )
	       | _ -> N (Black, insert_II_rec x l, y, r)
	    )
	| n when n > 0 ->
	    (match r with
	       | N (Red, rl, z, rr) ->
		   (match compare x z with
		      | n when n < 0 -> 
			  red_conflict_RL l y (insert_II_rec x rl) z rr
		      | n when n > 0 -> 
			  red_conflict_RR l y rl z (insert_II_rec x rr)
		      | _ -> raise Found
		   )
	       | _ -> N (Black, l, y, insert_II_rec x r)
	    )
	| _ -> raise Found

let insert_II = fun x set ->
  try
    (match insert_II_rec x set with
       | N (Red, a, y, b) -> N (Black, a, y, b)
       | other -> other
    )
  with Found -> set

(* This second version presented by Okasaki tries not to propagate conflicts
 *
 * In functional languages not only it is not very important (since the path
 * to the root has anyway to be reconstructed) but the data structure is
 * some times slower, maybe because :
 * 
 * - the tree is less well balanced
 * - there are many conditionals
 *)

(* Voici une seconde version prsente par Chris Okasaki, mais s'efforant
 * de ne pas propager les conflits
 *
 * Dans les langages fonctionnels, non seulement cette proccupation est
 * secondaire (puisque il faut de toutes les faons reconstruire le
 * chemin  la racine car les arbres sont persistants) mais la procdure
 * se montre parfois plus lente, possiblement car :
 * 
 * - l'arbre est moins bien balanc
 * - il faut faire de nombreux branchements conditionnels 
 *)

(* Encuentrese aqu otra versin presentada por Chris Okasaki, pero
 * sta vez se intenta minimizar la propagacin
 * 
 * En los lenguajes funcionales non solo es secudaria sta preocupacin,
 * (ya que de todas maneras hay que reconstruir el camino hacia la raz) 
 * sino que no siempre se obtiene una estructura de datos ms veloz,
 * tal vez porque
 * 
 * - el rbol est menos bien balanceado
 * - hay que hacer verificar numerosas condiciones
 *)

let red_conflict_complex = function

  (* colour flips *)
  | (Black, N (Red, (N (Red, _, _, _) as a), x, b), y, N (Red, c, z, d))
  | (Black, N (Red, a, x, (N (Red, _, _, _) as b)), y, N (Red, c, z, d))
  | (Black, N (Red, a, x, b), y, N (Red, (N (Red, _, _, _) as c), z, d))
  | (Black, N (Red, a, x, b), y, N (Red, c, z, (N (Red, _, _, _) as d)))
    -> N (Red, N (Black, a, x, b), y, N (Black, c, z, d))

  (* single rotations *)
  | (Black, N (Red, (N (Red, _, _, _) as a), x, b), y, c) ->
      N (Black, a, x , N (Red, b, y, c))
  | (Black, a, x, N (Red, b, y, (N (Red, _, _, _) as c))) ->
      N (Black, N (Red, a, x, b), y, c)

  (* double rotations *)
  | (Black, N (Red, a, x, N (Red, b, y, c)), z, d)
  | (Black, a, x, N (Red, N (Red, b, y, c), z, d))
    -> N (Black, N (Red, a, x, b), y, N (Red, c, z, d))

  (* no balancing necessary *)
  | (colour, a, x, b) -> N (colour, a, x, b)

let rec insert_III_rec_rec x = function
  | E -> N (Red, E, x, E)
  | N (color, a, y, b) ->
      match compare x y with
	| n when n < 0 -> 
	    red_conflict_complex (color, insert_III_rec_rec x a, y, b)
	| n when n > 0 -> 
	    red_conflict_complex (color, a, y, insert_III_rec_rec x b)
	| _ -> raise Found

let insert_III = fun x set ->
  try 
    (match insert_III_rec_rec x set with
       | N (Red, a, y, b) -> N (Black, a, y, b)
       | other -> other
    )
  with Found -> set

(* The same function expanded on two levels. This is the classical insertion
 * funcion in red-black trees.
 *)

(* La mme fonction dveloppe le long du chemin de recherche sur une hauteur 
 * de deux niveaux : c'est la fonction classique d'insertion dans les arbres 
 * bicolores
 *)

(* La misma funcin desarrolada sobre dos niveles. Es la funcin de insercin 
 * clasica en los rboles bicromticos.
 *)

let red_conflict_complex_LL = fun ll x lr y r ->
  match ll with
    | N (Red, _, _, _) ->
	(match r with
	   | N (Red, rl, z, rr) -> 
	       N (Red, N (Black, ll, x, lr), y, N (Black, rl, z, rr))
	   | _ -> N (Black, ll, x, N (Red, lr, y, r))
	)
    | _ -> N (Black, N (Red, ll, x, lr), y, r)

let red_conflict_complex_LR = fun ll x lr y r ->
  match lr with
    | N (Red, lrl, lrv, lrr) -> 
	(match r with
	   | N (Black, rl, z, rr) -> 
	       N (Red, N (Black, ll, x, lr), y, N (Black, rl, z, rr))
	   | _ -> N (Black, N (Red, ll, x, lrl), lrv, N (Red, lrr, y, r))
	)
    | _ -> N (Black, N (Red, ll, x, lr), y, r)

let red_conflict_complex_RR = fun l x rl y rr ->
  match rr with
    | N (Red, _, _, _) -> 
	(match l with
	   | N (Red, ll, lv, lr) -> 
	       N (Red, N (Black, ll, lv, lr), x, N (Black, rl, y, rr))
	   | _ -> N (Black, N (Red, l, x, rl), y, rr)
	)
    | _ -> N  (Black, l, x, N (Red, rl, y, rr))

let red_conflict_complex_RL = fun l x rl y rr ->
  match rl with
    | N (Red, rll, rlv, rlr) ->
	(match l with
	   | N (Red, ll, lv, lr) -> 
	       N (Red, N (Black, ll, lv, lr), x, N (Black, rl, y, rr))
	   | _ -> N (Black, N (Red, l, x, rll), rlv, N (Red, rlr, y, rr))
	)
    | _ -> N (Black, l, x, N (Red, rl, y, rr))

let rec insert_IV_rec x = function
  | E -> N (Red, E, x, E)
  | N (_, l, y, r) ->
      match compare x y with
	| n when n < 0 ->
	    (match l with
	       | N (Red, ll, z, lr) ->
		   (match compare x z with
		      | n when n < 0 -> 
			  red_conflict_complex_LL 
			  (insert_IV_rec x ll) z lr y r
		      | n when n > 0 -> 
			  red_conflict_complex_LR ll z 
			  (insert_IV_rec x lr) y r
		      | _ -> raise Found
		   )
	       | _ -> N (Black, insert_IV_rec x l, y, r)
	    )
	| n when n > 0 ->
	    (match r with
	       | N (Red, rl, z, rr) ->
		   (match compare x z with
		      | n when n < 0 -> 
			  red_conflict_complex_RL l y 
			  (insert_IV_rec x rl) z rr
		      | n when n > 0 -> 
			  red_conflict_complex_RR l y rl z 
			  (insert_IV_rec x rr)
		      | _ -> raise Found
		   )
	       | _ -> N (Black, l, y, insert_IV_rec x r)
	    )
 	| _ -> raise Found
	    
let insert_IV = fun x set ->
  try 
    (match insert_IV_rec x set with
       | N (Red, a, y, b) -> N (Black, a, y, b)
       | other -> other
    )
  with Found -> set

(* Notice here the propagation of overweight information. This propagation is
 * needed because it is impossible to distinguish locally a black node that
 * follows his father and a black node that has one unoccupied level between
 * him and his father
 *)

(* On note la propagation de l'information de surpoids. Cette propagation
 * est ncessaire car il est impossible de distinguer localement
 * un noeud noir qui suit immdiatement son pre d'un noeud noir
 * qui laisse un ou plusieurs niveaux entre les deux
 *)

(* Notese la propagacin de la informacin de sobrepeso. sta propagacin
 * es necesaria ya que nada permite distinguir de manera local entre un nudo
 * negro que sigue imediatamente a su padre y un nudo negro que deja un nivel
 * vacio entre l y su padre
 *)

let black_overweight_left_simple = fun overweight tree ->
  match overweight with
    | No -> (tree, No)
    | Yes -> 
	match tree with
	  | N (Red, l, v, N (Black, rl, rv, rr)) ->
	      (red_conflict_simple (Black, l, v, N (Red, rl, rv, rr)), No)
	  | N (Black, l, v, N (Red, N (_, rll, rlv, rlr), rv, rr)) -> 
	      (N (Black, red_conflict_simple 
		    (Black, l, v, N (Red, rll, rlv, rlr)), rv, rr),
	       No)
	  | N (Black, l, v, N (Black, rl, rv, rr)) ->
	      (red_conflict_simple (Black, l, v, N (Red, rl, rv, rr)), Yes)
	  | other -> (other, No)

let black_overweight_right_simple = fun overweight tree ->
  match overweight with
    | No -> (tree, No)
    | Yes ->
	match tree with
	  | N (Red, N (Black, ll, lv, lr), v, r) ->
 	      (red_conflict_simple (Black, N (Red, ll, lv, lr), v, r), No)
	  | N (Black, N (Red, ll, lv, N (_, lrl, lrv, lrr)), v, r) ->
	      (N (Black, ll, lv,
		  red_conflict_simple
		    (Black, N (Red, lrl, lrv, lrr), v, r)),
	       No)
	  | N (Black, N (Black, ll, lv, lr), v, r) ->
	      (red_conflict_simple (Black, N (Red, ll, lv, lr), v, r), Yes)
	  | other -> (other, No)

let rec min_element = function
  | E -> raise Not_found
  | N (_, E, v, _) -> v
  | N (_, l, _, _) -> min_element l

let rec max_element = function
  | E -> raise Not_found
  | N (_, _, v, E) -> v
  | N (_, _, _, r) -> max_element r

let rec delete_I_rec x = fun (tree, _) ->
  match tree with
    | E -> raise Not_found
    | N (Red, E, y, E) ->
	(match compare x y with
	   | 0 -> (E, No)
	   | _ -> raise Not_found
	)
    | N (Black, E, y, E) ->
	(match compare x y with
	   | 0 -> (E, Yes)
	   | _ -> raise Not_found
	)
    | N (colour, a, y, b) ->
	match compare x y with
	  | n when n < 0 -> 
	      let (left, propagate) = delete_I_rec x (a, No) in
		black_overweight_left_simple propagate (N (colour, left, y, b))
	  | n when n > 0 -> 
	      let (right, propagate) = delete_I_rec x (b, No) in
		black_overweight_right_simple propagate 
		  (N (colour, a, y, right))
	  | _ -> 
	      match (a, b) with
		| (E, _) ->
		    let min = min_element b in
		    let (right, propagate) = delete_I_rec min (b, No) in
		      black_overweight_right_simple propagate 
			(N (colour, a, min, right))
		| _ ->
		    let max = max_element a in
		    let (left, propagate) = delete_I_rec max (a, No) in
		      black_overweight_left_simple propagate
			(N (colour, left, max, b))
   
let delete_I = fun x set ->
  try 
    let (new_set, _) = delete_I_rec x (set, Yes) in new_set
  with Not_found -> set

(* In this function, we just expand the red-conflict solving function inline,
 * and we make trivial simplifications
 * 
 * It is the usual deletion function in red-black trees
 *)

(* Dans cette fonction de requilibrage nous reprenons la fonction prcdente
 * en expandant en ligne le traitement des conflits rouge-rouge sur 2 niveaux
 * et en simplifiant les rsultats
 * 
 * C'est la fonction usuelle de suppression dans les arbres bicolores
 *)

(* En sta funcin, se ha simplemente expandido en linea la resolucin de
 * los conflictor rojos, y se ha simplificado cuando posible.
 * 
 * Es la funcin comn de supresin en los rboles bicromticos
 *)

let is_black = function
  | E | N (Black, _, _, _) -> true
  | N (Red, _, _, _) -> false

let blacking = function
  | N (Red, l, v, r) -> N (Black, l, v, r)
  | N (Black, _, _, _) | E as node -> node

let rec black_overweight_right_complex = fun overweight tree ->
  match overweight with
    | No -> (tree, No)
    | Yes -> 
	match tree with

	  | N (Black, N (Black, ll, lv, lr), v, r) ->
	      (match lr with
		 | N (Black, _, _, _) | E ->
		     if is_black ll then 
		       (N (Black, N (Red, ll, lv, lr), v, r), Yes)
		     else
		       (N (Black, blacking ll, lv, N (Black, lr, v, r)), No)
		 | N (Red, lrl, lrv, lrr) ->
		       (N (Black, N (Black, ll, lv, lrl), lrv, 
			   N (Black, lrr, v, r)), 
			No)
	      )
	      
	  | N (Red, N (Black, ll, lv, lr), v, r) ->
	      (match lr with
		 | N (Black, _, _, _) | E ->
		     if is_black ll then 
		       (N (Black, N (Red, ll, lv, lr), v, r), No)
		     else
		       (N (Black, ll, lv, N (Red, lr, v, r)), No)
		 | N (Red, lrl, lrv, lrr) ->
		     (N (Red, N (Black, ll, lv, lrl), lrv, 
			 N (Black, lrr, v, r)),
		      No)
	      )

	  | N (Black, N (Red, ll, lv, N (Black, lrl, lrv, lrr)), v, r) ->
	      (match lrr with
		 | N (Black, _, _, _) | E ->
		     if is_black lrl then
		       (N (Black, ll, lv, 
			   N (Black, N (Red, lrl, lrv, lrr), v, r)), 
			No)
		     else
		       (N (Black, ll, lv, 
			   N (Black, lrl, lrv, N (Red, lrr, v, r))), 
			No)
		 | N (Red, lrrl, lrrv, lrrr) ->
		     (N (Black, ll, lv, 
			 N (Red, N (Black, lrl, lrv, lrrl),
			    lrrv, N (Black, lrrr, v, r))), 
		      No)
	      )

	  | other -> (other, No)
		      

let rec black_overweight_left_complex = fun overweight tree ->
  match overweight with
    | No -> (tree, No)
    | Yes -> 
	match tree with

	  | N (Black, l, v, N (Black, rl, rv, rr)) ->
	      (match rl with
		 | N (Black, _, _, _) | E ->
		     if is_black rr then
		       (N (Black, l, v, N (Red, rl, rv, rr)), Yes)
		     else
		       (N (Black, N (Black, l, v, rl), rv, blacking rr), No)
		 | N (Red, rll, rlv, rlr) ->
		     (N (Black, N (Black, l, v, rll), rlv, 
			 N (Black, rlr, rv, rr)),
		      No)
	      )

	  | N (Red, l, v, N (Black, rl, rv, rr)) ->
	      (match rl with
		 | N (Black, _, _, _) | E ->
		     if is_black rr then
		       (N (Black, l, v, N (Red, rl, rv, rr)), No)
		     else
		       (N (Black, N (Red, l, v, rl), rv, rr), No)
		 | N (Red, rll, rlv, rlr) ->
		     (N (Red, N (Black, l, v, rll), rlv, 
			 N (Black, rlr, rv, rr)),
		      No)
	      )

	  | N (Black, l, v, N (Red, N (Black, rll, rlv, rlr), rv, rr)) ->
	      (match rll with
		 | N (Black, _, _, _) | E ->
		     if is_black rlr then
		       (N (Black, N (Black, l, v, N (Red, rll, rlv, rlr)), 
			   rv, rr), 
			No)
		     else
		       (N (Black, N (Black, N (Red, l, v, rll), rlv, rlr),
			   rv, rr),
			No)
		 | N (Red, rlll, rllv, rllr) ->
		     (N (Black, 
			 N (Red, N (Black, l, v, rlll), rllv, 
			    N (Black, rllr, rlv, rlr)), rv, rr),
		      No)
	      )

	  | other -> (other, No)
		 

let rec delete_II_rec x = fun (tree, _) ->
  match tree with
    | E -> raise Not_found
    | N (Red, E, y, E) ->
	(match compare x y with
	   | 0 -> (E, No)
	   | _ -> raise Not_found
	)
    | N (Black, E, y, E) ->
	(match compare x y with
	   | 0 -> (E, Yes)
	   | _ -> raise Not_found
	)
    | N (colour, a, y, b) ->
	match compare x y with
	  | n when n < 0 -> 
	      let (left, propagate) = delete_II_rec x (a, No) in
		black_overweight_left_complex propagate 
		  (N (colour, left, y, b))
	  | n when n > 0 -> 
	      let (right, propagate) = delete_II_rec x (b, No) in
		black_overweight_right_complex propagate 
		  (N (colour, a, y, right))
	  | _ -> 
	      match (a, b) with
		| (E, _) ->
		    let min = min_element b in
		    let (right, propagate) = delete_II_rec min (b, No) in
		      black_overweight_right_complex propagate 
			(N (colour, a, min, right))
		| _ ->
		    let max = max_element a in
		    let (left, propagate) = delete_II_rec max (a, No) in
		      black_overweight_left_complex propagate
			(N (colour, left, max, b))
   
let delete_II = fun x set ->
  try 
    let (new_set, _) = delete_II_rec x (set, Yes) in new_set
  with Not_found -> set

(* SMK deletion function does not explicitely propagate the black overweight 
 * information, and has a much simpler expression.
 * The counterpart is a lot of node rebuilding
 *)

(* Cette fonction de suppression par SMK ne propage pas explicitement 
 * l'information de surpoids noir et s'crit de manire simple et plus proche
 * des algorithmes globaux obtenus par ailleurs.
 * La contrepartie est qu'il faut reconstruire de trs nombreux noeuds
 *)

(* sta funcin de supresin por SMK no propaga de manera explicita la 
 * informacin de sobrepeso, ademas su expresion es relativamente simple
 * y ms cercana de la que se obtenia en el caso global.
 * La contraparte es que hay que reconstruir un gran nmero de nudos.
 *)

let balance = fun l v r ->
  match (l, v, r) with
    | (N (Red, a, x, b), y, N (Red, c, z, d))
    | (N (Red, N (Red, a, x, b), y, c), z, d)
    | (N (Red, a, x, N (Red, b, y, c)), z, d)
    | (a, x, N (Red, b, y, N (Red, c, z, d)))
    | (a, x, N (Red, N (Red, b, y, c), z, d)) -> 
	N (Red, N (Black, a, x, b), y, N (Black, c, z, d))
    | (a, x, b) -> N (Black, a, x, b)

let sub1 = function
  | N (Black, a, x, b) -> N (Red, a, x, b)
  | _ -> failwith "invariance violation"

let balleft = fun l v r ->
  match (l, v, r) with
    | (N (Red, a, x, b), y, c) -> N (Red, N (Black, a, x, b), y, c)
    | (bl, x, N (Black, a, y, b)) -> balance bl x (N (Red, a, y, b))
    | (bl, x, N (Red, N (Black, a, y, b), z, c)) ->
	N (Red, N (Black, bl, x, a), y, balance b z (sub1 c))
    | _ -> failwith "unexpected_case"

let balright = fun l v r ->
  match (l, v, r) with
    | (a, x, N (Red, b, y, c)) -> N (Red, a, x, N (Black, b, y, c))
    | (N (Black, a, x, b), y, bl) -> balance (N (Red, a, x, b)) y bl
    | (N (Red, a, x, N (Black, b, y, c)), z, bl) ->
	N (Red, balance (sub1 a) x b, y, N (Black, c, z, bl))
    | _ -> failwith "unexpected_case"

let rec app = fun l r ->
  match (l, r) with
    | (E, x) | (x, E) -> x
    | (N (Red, a, x, b), N (Red, c, y, d)) ->
	(match app b c with
	   | N (Red, b', z, c') -> 
	       N (Red, N (Red, a, x, b'), z, N (Red, c', y, d))
	   | bc -> N (Red, a, x, N (Red, bc, y, d))
	)
    | (N (Black, a, x, b), N (Black, c, y, d)) ->
	(match app b c with
	   | N (Red, b', z, c') -> 
	       N (Red, N (Black, a, x, b'), z, N (Black, c', y, d))
	   | bc -> balleft a x (N (Black, bc, y, d))
	)
    | (a, N (Red, b, x, c)) -> N (Red, app a b, x, c)
    | (N (Red, a, x, b), c) -> N (Red, a, x, (app b c)) 

let rec del x = function
  | E -> E
  | N (_, a, y, b) ->
      match compare x y with
	| n when n < 0 -> 
	    if is_black a then balleft (del x a) y b
	    else N (Red, del x a, y, b)
	| n when n > 0 ->
	    if is_black b then balright a y (del x b)
	    else N (Red, a, y, del x b)
	| _ -> app a b

let delete_III = fun x set ->
  match del x set with
    | E -> E
    | N (_, a, y, b) -> N (Black, a, y, b)

let make = fun colour l v r -> N (colour, l, v, r)
let makeBlack = fun l v r -> N (Black, l, v, r)
let makeRed = fun l v r -> N (Red, l, v, r)

let rotate_L = fun l v r ->
  match r with
    | N (_, rl, rv, rr) -> makeRed (makeBlack l v rl) rv (blacking rr)
    | _ -> makeBlack l v r
	
let rotate_R = fun l v r ->
  match l with
    | N (_, ll, lv, lr) -> makeRed (blacking ll) lv (makeBlack lr v r)
    | _ -> makeBlack l v r

let rotate_RL = fun l v r ->
  match r with
    | N (_, N (_, rll, rlv, rlr), rv, rr) -> 
	makeRed (makeBlack l v rll) rlv (makeBlack rlr rv rr)
    | _ -> makeBlack l v r

let rotate_LR = fun l v r ->
  match l with
    | N (_, ll, lv, N (_, lrl, lrv, lrr)) ->
	makeRed (makeBlack ll lv lrl) lrv (makeBlack lrr v r)
    | _ -> makeBlack l v r

let is_red = function
  | E | N (Black, _, _, _) -> false
  | N (Red, _, _, _) -> true

let makeLeftHigher = fun colour l v r ->
  match l with
    | E | N (Black, _, _, _) -> make colour l v r
    | N (Red, ll, lv, lr) ->
	if is_red ll then rotate_R l v r
	else if is_red lr then rotate_LR l v r
	else make colour l v r

let makeRightHigher = fun colour l v r ->
  match r with
    | E | N (Black, _, _, _) -> make colour l v r
    | N (Red, rl, rv, rr) ->
	if is_red rr then rotate_L l v r
	else if is_red rl then rotate_RL l v r
	else make colour l v r

let reding = function
  | N (Black, l, v, r) -> N (Red, l, v, r)
  | other -> other 

(* Computes the black height along the shortest path when posible *)
(* Calcule la hauteur noire sur la branche la plus courte quand possible *)
(* Calcula la altura negra en la rama la ms corta cuando posible *)

let hc = function
  | Red -> 0
  | Black -> 1

let rec black_height_rec total = function
  | E -> total
  | N (colour, l, v, r) ->
      match (l, r) with
	| (E, _) | (_, E) -> total + hc colour
	| (N (Black, _, _, _), _) -> black_height_rec (total + hc colour) l
	| _ -> black_height_rec (total + hc colour) r

let black_height = function tree -> black_height_rec 0 tree

let makeBTree = fun l v r ->
  let hl = black_height l and hr = black_height r in
    match compare hl hr with
      | n when n < 0 -> makeRightHigher Black l v r
      | n when n > 0 -> makeLeftHigher Black l v r
      | _ -> makeBlack l v r

let rec append = fun l r ->
  match (l, r) with
    | (E, E) -> E
    | (E, s) | (s, E) -> reding s
    | (N (colour1, l1, v1, r1), N (colour2, l2, v2, r2)) -> 
	makeRightHigher colour1 l1 v1 
	(makeLeftHigher colour2 (append r1 l2) v2 r2)

let rec insert_rec' = function
  | E -> makeRed E x E
  | N (_, l, v, r) ->
      match compare x v with
	| n when n < 0 -> makeBTree (insert_rec' x l) v r
	| n when n > 0 -> makeBTree l v (insert_rec' x r)
	| _ -> raise Found

let insert' = fun x set -> try blacking (insert_rec' x set) with Found -> set

let rec insert_rec x = function
  | E -> makeRed E x E
  | N (colour, l, v, r) ->
      match compare x v with
	| n when n < 0 -> makeLeftHigher colour (insert_rec x l) v r
	| n when n > 0 -> makeRightHigher colour l v (insert_rec x r)
	| _ -> raise Found

let insert = fun x set -> try blacking (insert_rec x set) with Found -> set

let rec delete_rec x = function
  | E -> raise Not_found
  | N (colour, l, v, r) ->
      match compare x v with
	| n when n < 0 -> makeRightHigher colour (delete_rec x l) v r
	| n when n > 0 -> makeLeftHigher colour l v (delete_rec x r)
	| _ -> append l r

let delete = fun x set -> try blacking (delete_rec x set) with Not_found -> set

let rec min_element = function
  | E -> raise Not_found
  | N (_, E, v, _) -> v
  | N (_, l, _, _) -> min_element l

let rec max_element = function
  | E -> raise Not_found
  | N (_, _, v, E) -> v
  | N (_, _, _, r) -> max_element r

let rec delete_min = function
  | E -> E
  | N (_, E, v, r) -> r
  | N (colour, l, v, r) -> makeRightHigher colour (delete_min l) v r

let rec delete_max = function
  | E -> E
  | N (_, l, v, E) -> l
  | N (colour, l, v, r) -> makeLeftHigher colour l v (delete_max r)

let injection = function x -> makeBlack E x E

let zero = function () -> E

let rec from_list_rec result = function
  | [] -> result
  | head :: tail -> from_list_rec (insert head result) tail

let from_list = function list -> from_list_rec E list

let rec to_list_rec result = function
  | E -> result
  | N (_, a, x, b) -> to_list_rec (x :: to_list_rec result b) a

let to_list = function tree -> to_list_rec [] tree

let map f = function set -> from_list (BaireList.map f (to_list set))

let rec max_element = function
  | E -> raise Not_found
  | N (_, _, x, E) -> x
  | N (_, _, _, b) -> max_element b

let rec min_element = function
  | E -> raise Not_found
  | N (_, E, x, _) -> x
  | N (_, a, _, _) -> min_element a

let rec member_some x y = function
  | E -> compare x y = 0
  | N (_, a, z, b) ->
      match compare x z with
	| n when n < 0 -> member_some x y a
	| _ -> member_some x z b

let rec member_none x = function
  | E -> false
  | N (_, a, y, b) ->
      match compare x y with
	| n when n < 0 -> member_none x a
	| _ -> member_some x y b

let member' = fun x tree -> member_none x tree

let rec member x = function
  | E -> false
  | N (_, a, y, b) ->
      match compare x y with
	| n when n < 0 -> member x a
	| n when n > 0 -> member x b
	| _ -> true

let rec size_rec result = function
  | E -> result
  | N (_, a, _, b) -> size_rec (size_rec (1 + result) b) a

let size = function tree -> size_rec 0 tree

let rec get_rec = fun tree n ->
  match tree with
    | E -> raise (Out_of_bounds n)
    | N (_, a, x, b) ->
	let k = size_rec 0 a in
	  if n < k then get_rec a n
	  else if n > k then get_rec b (n - k - 1)
	  else x

let rec for_all f = function
  | E -> true
  | N (_, a, x, b) -> f x && for_all f a && for_all f b

let rec exists f = function
  | E -> false
  | N (_, a, x, b) -> f x || exists f a || exists f b

let rec find f = function
  | E -> raise Not_found
  | N (_, _, x, _) when f x -> x
  | N (_, a, _, b) -> (try find f a with Not_found -> find f b)

let rec filter_rec f result = function
  | E -> result
  | N (_, a, x, b) ->
      filter_rec f (filter_rec f (if f x then insert x result else result) a) b

let rec filter f = function tree -> filter_rec f E tree
      
let rec partition_rec f (trueS, falseS) = function
  | E -> (trueS, falseS)
  | N (_, a, x, b) -> 
      if f x 
      then partition_rec f (partition_rec f (insert x trueS, falseS) a) b
      else partition_rec f (partition_rec f (trueS, insert x falseS) a) b

let partition = fun f set -> partition_rec f (E, E) set

let rec iter f = function
  | E -> ()
  | N (_, a, x, b) -> iter f a ; f x ; iter f b

let rec fold_left f e = function
  | E -> e
  | N (_, a, x, b) -> fold_left f (f (fold_left f e a) x) b

let rec fold_right f e = function
  | E -> e 
  | N (_, a, x, b) -> fold_right f (f x (fold_right f e b)) a

let fold f e = function tree -> fold_right f e tree

let rec depth = function
  | E -> 0
  | N (_, a, _, b) -> 1 + max (depth a) (depth b)

let rec black_height = function
  | E -> 0
  | N (Black, a, _, b) ->
      let ha = black_height a and hb = black_height b in
	if ha <> hb then failwith "black height problem"
	else 1 + ha
  | N (Red, a, _, b) ->
      let ha = black_height a and hb = black_height b in
	if ha <> hb then failwith "black height problem"
	else ha

let rec red_red = function
  | E -> false
  | N (Black, a, _, b) -> red_red a || red_red b
  | N (Red, a, _, b) ->
      (match a with N (Red, _, _, _) -> true | _ -> red_red a ) ||
      (match b with N (Red, _, _, _) -> true | _ -> red_red b)

let rec nb_red = function
  | E -> 0
  | N (Black, a, _, b) -> nb_red a + nb_red b
  | N (Red, a, _, b) -> 1 + nb_red a + nb_red b

let stat = function tree -> 
  (
    size tree, 
    depth tree, 
    black_height tree,
    red_red tree,
    nb_red tree
  )

exception Red_Erreur of int * int set * int list
exception Black_Erreur of int * int set * int list
