(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Maps represended by ordered lists
 * Dictionnaires reprsents par des listes ordonnes
 * Diccionrios representados por listas ordenadas
 *)

open BairePervasives

let moduleName = "OrderedListSet"

let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type


(** Oprations catgorielles **)
(** Categorial operations **)
(** Operaciones categoriales **)


let injection = fun v -> [v]

let rec map f = function
  | [] -> []
  | head :: tail -> (f head) :: map f tail

let rec plus = fun a b ->
  match (a, b) with
    | ([], _) -> b
    | (_, []) -> a
    | (v :: a_tail, v' :: b_tail) ->
	match compare v v' with
	  | n when n < 0 -> v :: plus a_tail b
	  | 0 -> v' :: plus a_tail b_tail
	  | _ -> v' :: plus a b_tail

let zero = function () -> []

let rec join_rec result = function
  | [] -> result
  | head :: tail -> join_rec (plus head result) tail

let join = function list -> join_rec [] list

let to_list = id

let from_list = id

(*  implmenter construction d'arbre parfait *)

let inj = function x -> Node (Leaf, x, Leaf)

let rec to_tree = function
  | x :: y :: tail -> Node (inj x, y, to_tree tail)
  | x :: tail -> Node (Leaf, x, Leaf)
  | [] -> Leaf

let rec from_tree_rec result = function
  | Leaf -> result
  | Node (l, v, r) -> from_tree_rec (v :: (from_tree_rec result l)) r

let from_tree = function tree -> from_tree_rec [] tree


(** Oprations binaires **)
(** Binairy operations **)
(** Operaciones binrias **)


let union = plus

let rec intersection = fun a b ->
  match (a, b) with
    | ([], _) | (_, []) -> []
    | (v :: a_tail, v' :: b_tail) ->
	match compare v v' with
	  | 0 -> v :: intersection a_tail b_tail
	  | _ -> intersection a_tail b_tail

let rec difference = fun a b ->
  match (a, b) with
    | ([], _) -> []
    | (_, []) -> a
    | (v :: a_tail, v' :: b_tail) ->
	match compare v v' with
	  | n when n < 0 -> v :: difference a_tail b
	  | n when n > 0 -> v :: difference a_tail b_tail
	  | _ -> difference a_tail b_tail

let rec symetric_difference = fun a b ->
  match (a, b) with
    | ([], s) | (s, []) -> s
    | (v :: a_tail, v' :: b_tail) ->
	match compare v v' with
	  | n when n < 0 -> v :: symetric_difference a_tail b
	  | n when n > 0 -> v' :: symetric_difference a b_tail
	  | _ -> symetric_difference a_tail b_tail


(** Acces **)
(** Accs **)
(** Acceso **)


let rec insert_raise v = function
  | [] -> [v]
  | v' :: tail ->
      match compare v v' with
	| n when n < 0 -> v' :: insert_raise v tail
	| n when n > 0 -> v' :: v :: tail
	| _ -> raise Found

let insert = fun v set -> try insert_raise v set with Found -> set

let rec delete_raise v = function
  | [] -> raise Not_found
  | v' :: tail ->
      match compare v v' with
	| n when n < 0 -> v' :: delete_raise v tail
	| n when n > 0 -> raise Not_found
	| _ -> tail

let delete = fun v set -> try delete_raise v set with Not_found -> set

let delete_min = function
  | [] -> []
  | _ :: tail -> tail

let rec delete_max = function
  | [] -> []
  | [_] -> []
  | head :: tail -> head :: delete_max tail

let rec filter f = function
  | [] -> []
  | v :: tail when f v -> v :: filter f tail
  | _ :: tail -> filter f tail

let rec filter_lt x = function
  | [] -> []
  | v :: tail ->
      match compare x v with
	| n when n < 0 -> []
	| _ -> v :: filter_lt x tail

let rec filter_le x = function
  | [] -> []
  | v :: tail ->
      match compare x v with
	| n when n <= 0 -> []
	| _ -> v :: filter_lt x tail

let rec filter_gt x = function
  | [] -> []
  | v :: tail as list ->
      match compare x v with
	| n when n <= 0 -> list 
	| _ -> filter_gt x tail

let rec filter_ge x = function
  | [] -> []
  | v :: tail as list ->
      match compare x v with
	| n when n < 0 -> list
	| _ -> filter_gt x tail

let rec range min max = function
  | [] -> []
  | v :: tail as list ->
      if v < min then range min max tail
      else filter_lt max list


(** Vues **)
(** Views **)
(** Vistas **)


let min_view = function
  | [] -> raise Empty
  | head :: tail -> (head, tail)

let rec max_view_rec head_list = function
  | [] -> invalid "max_view_rec" UnexpectedPattern
  | [max] -> (List.rev head_list, max)
  | head :: tail -> max_view_rec (head :: head_list) tail

let rec partition_rec f trueL falseL = function
  | [] -> (List.rev trueL, List.rev falseL)
  | head :: tail when (f head) -> partition_rec f (head :: trueL) falseL tail
  | head :: tail -> partition_rec f trueL (head :: falseL) tail

let partition f = function list -> partition_rec f [] [] list

let rec partition_inf_sup_rec head_list pivot = function
  | [] -> ([], None, [])
  | v :: tail ->
      match compare pivot v with
	| n when n < 0 -> partition_inf_sup_rec (v :: head_list) pivot tail
	| n when n > 0 -> (List.rev (v :: head_list), None, tail)
	| _ -> (List.rev (v :: head_list), Some v, tail)

let partition_inf_sup = fun p map -> partition_inf_sup_rec [] p map


(** Observateurs **)
(** Observers **)
(** Observadores **)


let is_empty = function
  | [] -> true
  | _ :: _ -> false

let rec size_rec total = function
  | [] -> total
  | _ :: tail -> size_rec (total + 1) tail

let size = function list -> size_rec 0 list

let rec member v = function
  | [] -> false
  | v' :: tail ->
      match compare v v' with
	| n when n < 0 -> member v tail
	| n when n > 0 -> false
	| _ -> true

let rec look_for f = function
  | [] -> raise Not_found
  | v :: _ when f v -> v
  | _ :: tail -> look_for f tail

let choose = function
  | [] -> raise Not_found
  | head :: _ -> head

let min_element = function
  | [] -> raise Not_found
  | head :: _ -> head

let rec max_element = function
  | [] -> raise Not_found
  | [max] -> max
  | _ :: tail -> max_element tail

let rec previous x = function
  | [] -> raise Not_found
  | v :: tail ->
      match compare x v with
	| n when n < 0 -> (try previous x tail with Not_found -> v)
	| _ -> raise Not_found

let rec next x = function
  | [] -> raise Not_found
  | v :: tail ->
      match compare x v with
	| n when n <= 0 -> next x tail
	| 0 -> min_element tail
	| _ -> v

let rec floor x = function
  | [] -> raise Not_found
  | v :: tail ->
      match compare x v with
	| n when n <= 0 -> (try floor x tail with Not_found -> v)
	| _ -> raise Not_found

let rec ceil x = function
  | [] -> raise Not_found
  | v :: tail ->
      match compare x v with
	| n when n <= 0 -> next x tail
	| _ -> v

let rec rank_rec x v = function
  | [] -> raise Not_found
  | head :: _ when head = x -> v
  | _ :: tail -> rank_rec x (v + 1) tail

let rank x = function list -> rank_rec x 0 list

let rec get = fun list n ->
  match list with
    | [] -> raise (Out_of_bounds n)
    | head :: _  when (n = 0) -> head
    | _ :: tail -> get tail (n - 1)

let rec for_all f = function
  | [] -> true
  | head :: tail -> f head && for_all f tail

let rec exists f = function 
  | [] -> false
  | head :: tail -> f head || exists f tail


(** Traversals **)
(** Parcours **)
(** Recorridos **)


let rec iter f = function
  | [] -> ()
  | v :: tail -> f v ; iter f tail

let rec fold_right = fun f map e ->
  match map with
  | [] -> e
  | v :: tail -> f v (fold_right f tail e)

let rec fold_left f e = function
  | [] -> e
  | v :: tail -> fold_left f (f e v) tail

let fold = fold_right


(** Compatibilit avec la librairie standard de Caml **)
(** Caml standard library compatibility **)
(** Compatibilidad con la librera estandar de Caml **)


let empty = []
let mem = member
let add = insert
let singleton = injection
let remove = delete
let inter = intersection
let diff = difference
let cardinal = size
let elements = id
let min_elt = min_element    
let max_elt = max_element


(** Objets **)
(** Ojects **)
(** Objetos **)


class ['a] o_set = fun set ->

object (self : 'b)
  
  val contents  = set

  method to_set = contents

  (* Categorial operations *)
  (* Oprations categorielles *)
  (* Operaciones categoriales *)

  method map f = {< contents = map f contents >}

  method plus = function (set : 'b) -> 
    {< contents = plus contents set#to_set >} 

  method to_list = to_list contents
		     
  (* Binary operations *)
  (* Oprations binaires *)
  (* Operaciones binrias *)
		     
  method union = function (set : 'b) -> 
    {< contents = plus contents set#to_set >}
  
  method intersection = function (set : 'b) -> 
    {< contents = intersection contents set#to_set >}
  
  method difference = function (set : 'b) -> 
    {< contents = difference contents set#to_set >}
  
  method symetric_difference = function (set : 'b) -> 
    {< contents = symetric_difference contents set#to_set >}
    
  (* Acces *)
  (* Accs *)
  (* Acceso *)

  method insert = function x -> 
    try 
      {< contents = insert_raise x contents >} 
    with Found -> self

  method insert_raise = function x -> {< contents = insert_raise x contents >}

  method delete = function x -> 
    try 
      {< contents = delete_raise x contents >} 
    with Not_found -> self

  method delete_raise = function x -> {< contents = delete_raise x contents >}
      
  method delete_min = {< contents = delete_min contents >}
  method delete_max = {< contents = delete_max contents >}

  method filter = function f -> {< contents = filter f contents >}
  method filter_lt = function x -> {< contents = filter_lt x contents >}
  method filter_gt = function x -> {< contents = filter_gt x contents >}
  method filter_le = function x -> {< contents = filter_le x contents >}
  method filter_ge = function x -> {< contents = filter_ge x contents >}

  method range = fun min max -> {< contents = range min max contents >}

  (* Views *)
  (* Vues *)
  (* Vistas *)

  method min_view =
    let (min, other) = min_view contents in
      (min, {< contents = other >})

  method partition = function f -> 
    let (trueSet, falseSet) = partition f contents in
      ({< contents = trueSet >}, {< contents = falseSet >})

  method partition_inf_sup = function x ->
    let (inf, value, sup) = partition_inf_sup x contents in
      ({< contents = inf >}, value, {< contents = sup >})

  (* Observators *)
  (* Observateurs *)
  (* Observadores *)
      
  method is_empty = contents = []
  method size = size contents

  method member = function x -> member x contents
  method look_for = function f -> look_for f contents

  method choose = 
    match contents with
      | [] -> raise Empty
      | head :: _ -> head

  method min_element = min_element contents
  method max_element = max_element contents

  method previous = function x -> previous x contents
  method next = function x -> next x contents
  method floor = function x -> floor x contents
  method ceil = function x -> ceil x contents

  method rank = function x -> rank x contents

  method exists = function f -> exists f contents
  method for_all = function f -> for_all f contents

  (* Traversals *)
  (* Parcours *)
  (* Recorridos *)

  method iter : ('a -> unit) -> unit = function f -> iter f contents

  method fold : 'c. ('a -> 'c -> 'c) -> 'c -> 'c = 
    fun f e -> fold_right f contents e

  method fold_left : 'c. ('c -> 'a -> 'c) -> 'c -> 'c =
    fun f e -> fold_left f e contents

  method fold_right : 'c. ('a -> 'c -> 'c) -> 'c -> 'c =
    fun f e -> fold_right f contents e

end

(* Constructors *)
(* Constructeurs *)
(* Constructores *)

let o_injection = function x -> new o_set (injection x)
let o_map f = function a -> new o_set (map f a#to_set)
let o_plus = fun a b -> new o_set (plus a#to_set b#to_set)
let o_zero = function () -> new o_set []
let o_from_list = function list -> new o_set (from_list list)
let o_from_set = function set -> new o_set set

(* Binary operations *)
(* Oprations binaires *)
(* Operaciones binrias *)
  
let o_union = fun a b -> new o_set (plus a#to_set b#to_set)
let o_intersection = fun a b -> new o_set (intersection a#to_set b#to_set)
let o_difference = fun a b -> new o_set (difference a#to_set b#to_set)
let o_symetric_difference = fun a b -> 
  new o_set (symetric_difference a#to_set b#to_set)


module type OrderedType = sig
  type t
  val compare : t -> t -> int
end

module type Set = 
sig

  type t
  type elt

  val injection : elt -> t
  val map : (elt -> elt) -> t -> t
  val plus : t -> t -> t
  val zero : unit -> t
  val from_list : elt list -> t
  val to_list : t -> elt list
    
  val union : t -> t -> t
  val intersection : t -> t -> t
  val difference   : t -> t -> t
  val symetric_difference : t -> t -> t
  val subset  : t -> t -> bool
  val equal   : t -> t -> bool
  val compare : t -> t -> int

  val insert  : elt -> t -> t
  val insert' : elt -> t -> t
  val insert_raise  : elt -> t -> t (* [Found] *)
  val insert_raise' : elt -> t -> t (* [Found] *)
  val insert_list : elt list -> t -> t

  val delete_min : t -> t
  val delete_max : t -> t
  val delete : elt -> t -> t
  val delete_raise : elt -> t -> t (* [Not_found] *)
  val delete_list : elt list -> t -> t

  val filter : (elt -> bool) -> t -> t
  val filter_lt : elt -> t -> t
  val filter_gt : elt -> t -> t
  val filter_le : elt -> t -> t
  val filter_ge : elt -> t -> t

  val range : elt -> elt -> t -> t

  val min_view : t -> elt * t
  val max_view : t -> t * elt
  val partition : (elt -> bool) -> t -> t * t
  val partition_inf_sup : elt -> t -> t * elt option * t

  val is_empty : t -> bool
  val size : t -> int

  val member  : elt -> t -> bool
  val member' : elt -> t -> bool
  val look_for : (elt -> bool) -> t -> elt

  val choose : t -> elt (* [Empty] *)
  val min_element : t -> elt (* [Not_found] *)
  val max_element : t -> elt (* [Not_found] *)

  val next     : elt -> t -> elt (* [Not_found] *) 
  val previous : elt -> t -> elt (* [Not_found] *) 
  val ceil     : elt -> t -> elt (* [Not_found] *) 
  val floor    : elt -> t -> elt (* [Not_found] *) 

  val rank : elt -> t -> int  (* [Not_found] *) 
  val get  : int -> t -> elt  (* [Out_of_bounds] *)

  val for_all : (elt -> bool) -> t -> bool
  val exists  : (elt -> bool) -> t -> bool

  val iter : (elt -> 'a) -> t -> unit
  val fold_left  : ('a -> elt -> 'a) -> 'a -> t -> 'a
  val fold_right : (elt -> 'a -> 'a) -> t -> 'a -> 'a
  val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a

  val unsafe_inc_map : (elt -> elt) -> t -> t
  val unsafe_dec_map : (elt -> elt) -> t -> t
  val unsafe_set : elt -> int -> t -> t
  val unsafe_insert_min : elt -> t -> t
  val unsafe_insert_max : elt -> t -> t

  val old_plus : t -> t -> t

  val empty : t
  val mem : elt -> t -> bool
  val add : elt -> t -> t
  val singleton : elt -> t
  val remove : elt -> t -> t
  val inter : t -> t -> t
  val diff : t -> t -> t
  val cardinal : t -> int
  val elements : t -> elt list
  val min_elt : t -> elt (* [Not_found] *)
  val max_elt : t -> elt (* [Not_found] *)

  val stat : t -> int * int * int * int

  class type o_t =
  object ('a)

    method map : (elt -> elt) -> 'a
    method plus : 'a -> 'a
    method to_list : elt list
    method to_set : t
      
    method union        : 'a -> 'a
    method intersection : 'a -> 'a
    method difference   : 'a -> 'a
    method symetric_difference : 'a -> 'a
      
    method subset  : 'a -> bool
    method equal   : 'a -> bool
    method compare : 'a -> int
      
    method insert  : elt -> 'a
    method insert' : elt -> 'a
    method insert_raise  : elt -> 'a               (* [Found] *)
    method insert_raise' : elt -> 'a               (* [Found] *)
    method insert_list : elt list -> 'a
      
    method delete : elt -> 'a
    method delete_raise : elt -> 'a                (* [Not_Found] *)
    method delete_list : elt list -> 'a
    method delete_max : 'a
    method delete_min : 'a
      
    method filter : (elt -> bool) -> 'a
    method filter_gt : elt -> 'a
    method filter_lt : elt -> 'a
    method filter_ge : elt -> 'a
    method filter_le : elt -> 'a
      
    method range : elt -> elt -> 'a
      
    method min_view : elt * 'a                     (* [Empty] *)
    method max_view : 'a * elt                     (* [Empty] *)
    method partition : (elt -> bool) -> 'a * 'a
    method partition_inf_sup : elt -> 'a * elt option * 'a
      
    method is_empty : bool
    method size : int
      
    method member  : elt -> bool
    method member' : elt -> bool
    method look_for : (elt -> bool) -> elt         (* [Not_found] *)
      
    method choose : elt                            (* [Empty] *)
    method min_element : elt                       (* [Not_found] *)
    method max_element : elt                       (* [Not_found] *)
      
    method next     : elt -> elt                   (* [Not_found] *)
    method previous : elt -> elt                   (* [Not_found] *)
    method floor    : elt -> elt                   (* [Not_found] *)
    method ceil     : elt -> elt                   (* [Not_found] *)
      
    method get  : int -> elt                       (* [Out_of_bounds] *)
    method rank : elt -> int                       (* [Not_found] *)
      
    method for_all : (elt -> bool) -> bool
    method exists  : (elt -> bool) -> bool
      
    method iter : (elt -> unit) -> unit
    method fold : (elt -> 'b -> 'b) -> 'b -> 'b
    method fold_left  : ('b -> elt -> 'b) -> 'b -> 'b
    method fold_right : (elt -> 'b -> 'b) -> 'b -> 'b

  end

  val o_injection : elt -> o_t
  val o_map : (elt -> elt) -> o_t -> o_t
  val o_plus : o_t -> o_t -> o_t
  val o_zero : unit -> o_t
  val o_from_list : elt list -> o_t
  val o_from_set  : t -> o_t
    
  val o_union : o_t -> o_t -> o_t
  val o_intersection : o_t -> o_t -> o_t
  val o_difference   : o_t -> o_t -> o_t
  val o_symetric_difference : o_t -> o_t -> o_t

  val o_subset  : o_t -> o_t -> bool
  val o_equal   : o_t -> o_t -> bool
  val o_compare : o_t -> o_t -> int

end
