(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Patricia trees
 * Arbres patricia
 * rboles patricia
 *)

open BairePervasives

let moduleName = "PatriciaLow"

let block_size = 32

type intSet =
  | E
  | ZLeaf of int
  | Leaf of int
  | ZNode of intSet * int * intSet
  | Node of intSet * int * intSet
  | Prefix of int * intSet


(** Utilitaires **)
(** Utils **)
(** tiles **)

let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type

(* Use lsl and lsr insted of * and / beacause they differ over negative
 * values
 *)

(* Attention, il faut utiliser les dcalages lsl et lsr et non la 
 * multiplication ou division entire * et / car elles diffrent sur les 
 * nombres ngatifs
 *)

(* Cuidado, hay que utilizar las operaciones binrias lsl y lsr en vez del
 * producto y la divisin entera * y / ya que difieren en los enteros
 * negativos
 *)

let mask = function n -> 1 lsl (n - 1)

let makeTree = fun l n r -> Node (l, n, r)
let makeZTree = fun l n r -> ZNode (l, n, r)

let log2 = function n -> int_of_float (log (float n) /. log 2.)

let offset = block_size - 1 - log2 block_size
let branch_mask = (block_size - 1) lsl offset

let branching_bit = fun n -> (n land branch_mask) lsr offset

let make_prefix = fun n p -> p lsl offset lor n

let first_bit = fun x y -> (-x) land y


(** Categorial operations **)
(** Oprations catgorielles **)
(** Operaciones categoriales **)


let rec injection = function
  | 0 -> ZLeaf 0
  | n when n < block_size -> Leaf (mask n)
  | n -> Prefix (make_prefix n (block_size - 1), ZLeaf 0)
  | n when n mod 2 = 0 -> makeTree (injection (n lsr 1)) 0 E
  | n -> makeTree E 0 (injection (n lsr 1))

let rec insert_raise = fun n trie ->
  if n = 0 then
    match trie with
      | E -> ZLeaf 0
      | Leaf m -> ZLeaf m
      | Node (l, m, r) -> ZNode (l, m, r)
      | _ -> raise Found
  else match trie with
    | E -> injection n
    | Leaf m ->
	if n < block_size then
	  if m land (mask n) = 0 then Leaf (m lor (mask n))
	  else raise Found
	else 
	  if n mod 2 = 0 then makeTree (injection (n lsr 1)) m E
	  else makeTree E m (injection (n lsr 1))
    | ZLeaf m ->
	if n < block_size then
	  if m land (mask n) = 0 then ZLeaf (m lor (mask n))
	  else raise Found
	else 
	  if n mod 2 = 0 then makeZTree (injection (n lsr 1)) m E
	  else makeZTree E m (injection (n lsr 1))
    | Node (l, m, r) ->
	if n < block_size then
	  if m land (mask n) = 0 then makeTree l (m lor (mask n)) r
	  else raise Found
	else
	  if n mod 2 = 0 then makeTree (insert_raise (n lsr 1) l) m r
	  else makeTree l m (insert_raise (n lsr 1) r)
    | ZNode (l, m, r) ->
	if n < block_size then
	  if m land (mask n) = 0 then makeZTree l (m lor (mask n)) r
	  else raise Found
	else
	  if n mod 2 = 0 then makeZTree (insert_raise (n lsr 1) l) m r
	  else makeZTree l m (insert_raise (n lsr 1) r)

let insert = fun n set ->
  try
    insert_raise n set
  with Found -> set

let rec from_list_rec result = function
  | [] -> result
  | n :: tail -> from_list_rec (insert n result) tail

let from_list = function list -> from_list_rec E list

let rec to_list_leaf k prefix level = function
  | 0 -> []
  | n -> 
      if n mod 2 = 0 then to_list_leaf (k + 1) prefix level (n lsr 1)
      else ((k lsl level) lor prefix) :: 
	to_list_leaf (k + 1) prefix level (n lsr 1)

let rec to_list_rec prefix level result = function
  | E -> result
  | ZLeaf m -> (to_list_leaf 1 prefix level m) @ prefix :: result
  | Leaf m -> (to_list_leaf 1 prefix level m) @ result
  | ZNode (l, m, r) ->
      to_list_rec (prefix lsl 1) (level + 1) (
	to_list_rec (1 + prefix lsl 1) (level + 1) (
	  to_list_leaf 1 prefix level m @ prefix :: result
	) r
      ) l
  | Node (l, m, r) ->
      to_list_rec (prefix lsl 1) (level + 1) (
	to_list_rec (1 + prefix lsl 1) (level + 1) (
	  to_list_leaf 1 prefix level m @ result
	) r
      ) l

let rec to_list = function trie -> to_list_rec 0 0 [] trie

let map = fun f trie -> from_list (List.map f (to_list trie))

let rec plus = fun a b ->
  match a with
    | E -> b
    | Leaf m ->
	(match b with
	   | E -> a
	   | Leaf n -> Leaf (m lor n)
	   | ZLeaf n -> ZLeaf (m lor n)
	   | Node (l, n, r) -> makeTree l (m lor n) r
	   | ZNode (l, n, r) -> makeZTree l (m lor n) r
	)
    | ZLeaf m ->
	(match b with
	   | E -> a
	   | Leaf n | ZLeaf n -> ZLeaf (m lor n)
	   | Node (l, n, r) | ZNode (l, n, r) -> makeZTree l (m lor n) r
	)
    | Node (l, m, r) ->
	(match b with
	   | E -> a
	   | Leaf n -> makeTree l (m lor n) r
	   | ZLeaf n -> makeZTree l (m lor n) r
	   | Node (l', m', r') -> makeTree (plus l l') (m lor m') (plus r r')
	   | ZNode (l', m', r') -> makeZTree (plus l l') (m lor m') (plus r r')
	)
    | ZNode (l, m, r) ->
	match b with
	  | E -> a
	  | Leaf n | ZLeaf n -> makeZTree l (m lor n) r
	  | Node (l', m', r') | ZNode (l', m', r') -> 
	      makeZTree (plus l l') (m lor m') (plus r r')

let zero = function () -> E

let rec from_tree_rec result = function
  | BairePervasives.Leaf -> result
  | BairePervasives.Node (l, v, r) ->
      from_tree_rec (
	from_tree_rec (
	  insert v result
	) l
      ) r

let from_tree = function tree -> from_tree_rec E tree


(** Binary operations **)
(** Oprations binaires **)
(** Operaciones binrias **)


let union = plus

let rec normalize_tree = function
  | Leaf 0 -> E
  | Node (E, 0, E) -> E
  | E | Leaf _ | ZLeaf _ as leaves -> leaves
  | Node (l, m, r) -> makeTree (normalize_tree l) m (normalize_tree r)
  | ZNode (l, m, r) -> makeZTree (normalize_tree l) m (normalize_tree r)

let normalize = function
  | Leaf 0 -> E
  | Node (E, 0, E) -> E
  | E | Leaf _ | ZLeaf _ as leaves -> leaves
  | Node (l, m, r) -> makeTree l m r
  | ZNode (l, m, r) -> makeZTree l m r

let makeNTree = fun l v r ->
  match (l, v, r) with
    | (E, 0, E) -> E
    | _ -> Node (l, v, r)

let makeNLeaf = function
  | 0 -> E
  | n -> Leaf n

let rec intersection = fun a b ->
  match a with
    | E -> E
    | Leaf m ->
	(match b with
	   | E -> E
	   | Leaf n | ZLeaf n | Node (_, n, _) | ZNode (_, n, _) -> 
	       makeNLeaf (m land n)
	)
    | ZLeaf m ->
	(match b with
	   | E -> E
	   | Leaf n | Node (_, n, _) -> makeNLeaf (m land n)
	   | ZLeaf n | ZNode (_, n, _) -> ZLeaf (m land n)
	)
    | Node (l, m, r) ->
	(match b with
	   | E -> E
	   | Leaf n | ZLeaf n -> makeNLeaf (m land n)
	   | Node (l', m', r') | ZNode (l', m', r') -> 
	       makeNTree (intersection l l') (m land m') (intersection r r')
	)
    | ZNode (l, m, r) ->
	match b with
	  | E -> E
	  | Leaf n -> makeNLeaf (m land n)
	  | ZLeaf n -> ZLeaf (m land n)
	  | Node (l', m', r') -> 
	      makeNTree (intersection l l') (m land m') (intersection r r')
	  | ZNode (l', m', r') -> 
	      makeZTree (intersection l l') (m land m') (intersection r r')

let rec difference = fun a b ->
  match a with
    | E -> E
    | Leaf m ->
	(match b with
	   | E -> a
	   | Leaf n | ZLeaf n | Node (_, n, _) | ZNode (_, n, _) -> 
	       makeNLeaf (m land (lnot n))
	)
    | ZLeaf m ->
	(match b with
	   | E -> a
	   | Leaf n | Node (_, n, _) -> ZLeaf (m land (lnot n))
	   | ZLeaf n | ZNode (_, n, _) -> makeNLeaf (m land (lnot n))
	)
    | Node (l, m, r) ->
	(match b with
	   | E -> a
	   | Leaf n | ZLeaf n -> makeNLeaf (m land (lnot n))
	   | Node (l', m', r') | ZNode (l', m', r') -> 
	       makeNTree (difference l l') (m land (lnot m')) (difference r r')
	)
    | ZNode (l, m, r) ->
	match b with
	  | E -> a
	  | Leaf n -> ZLeaf (m land (lnot n))
	  | ZLeaf n -> makeNLeaf (m land (lnot n))
	  | Node (l', m', r') -> 
	      makeZTree (difference l l') (m land (lnot m')) (difference r r')
	  | ZNode (l', m', r') -> 
	      makeNTree (difference l l') (m land (lnot m')) (difference r r')

let rec subset = fun a b ->
  match a with
    | E -> true
    | Leaf m ->
	(match b with
	   | E -> false
	   | Leaf n | ZLeaf n | Node (_, n, _) | ZNode (_, n, _) -> m lor n = n
	)
    | ZLeaf m ->
	(match b with
	   | E | Leaf _ | Node _ -> false
	   | ZLeaf n | ZNode (_, n, _) -> m lor n = n
	)
    | Node (l, m, r) ->
	(match b with
	   | E | Leaf _ | ZLeaf _ -> false
	   | Node (l', m', r') | ZNode (l', m', r') -> 
	       m lor m' = m' && subset l l' && subset r r'
	)
    | ZNode (l, m, r) ->
	match b with
	  | E | Leaf _ | ZLeaf _ | Node _ -> false
	  | ZNode (l', m', r') -> m lor m' = m' && subset l l' && subset r r'

let equal = fun a b -> Pervasives.compare a b = 0

let compare = Pervasives.compare


(** Acces **)
(** Accs **)
(** Acceso **)


let rec insert_list = fun list trie ->
  match list with
    | [] -> trie
    | x :: tail -> insert_list tail (insert x trie)

let rec delete_raise = fun n trie ->
  if n = 0 then
    match trie with
      | E | Leaf _ | Node _ -> raise Not_found
      | ZLeaf m -> Leaf m
      | ZNode (l, m, r) -> Node (l, m, r)
  else if n < block_size then
    match trie with
      | E -> raise Not_found
      | Leaf m ->
	  if m land (mask n) <> 0 then Leaf (m lxor (mask n))
	  else raise Not_found
      | ZLeaf m ->
	  if m land (mask n) <> 0 then ZLeaf (m lxor (mask n))
	  else raise Not_found
      | Node (l, m, r) ->
	  if m land (mask n) <> 0 then makeTree l (m lxor (mask n)) r
	  else raise Not_found
      | ZNode (l, m, r) ->
	  if m land (mask n) <> 0 then makeZTree l (m lxor (mask n)) r
	  else raise Not_found
  else
    match trie with
      | E | Leaf _ | ZLeaf _ -> raise Not_found
      | Node (l, m, r) ->
	  if n mod 2 = 0 then makeTree (delete_raise (n lsr 1) l) m r
	  else makeTree l m (delete_raise (n lsr 1) r)
      | ZNode (l, m, r) ->
	  if n mod 2 = 0 then makeZTree (delete_raise (n lsr 1) l) m r
	  else makeZTree l m (delete_raise (n lsr 1) r)

let delete = fun n trie ->
  try
    delete_raise n trie
  with Not_found -> trie

let rec delete_list = fun list trie ->
  match list with
    | [] -> trie
    | x :: tail -> delete_list tail (delete x trie)


(** Observers **)
(** Observateurs **)
(** Observadores **)


let is_empty = function
  | E -> true
  | _ -> false

let rec count total = function
  | 0 -> total
  | n -> count (if n mod 2 = 0 then total else 1 + total) (n lsr 1)

let rec size_rec total = function
  | E -> total
  | Leaf n -> count total n
  | ZLeaf n -> count (1 + total) n
  | Node (l, n, r) -> size_rec (size_rec (count total n) l) r
  | ZNode (l, n, r) -> size_rec (size_rec (count (1 + total) n) l) r

let size = function trie -> size_rec 0 trie

let rec member = fun n trie ->
  if n = 0 then 
    match trie with
      | E | Leaf _ | Node _ -> false
      | ZLeaf _ | ZNode _ -> true
  else match trie with
    | E -> false
    | Leaf m | ZLeaf m -> 
	if n < block_size then (mask n) land m <> 0 else false
    | Node (l, m, r) | ZNode (l, m, r) ->
	if n < block_size then (mask n) land m <> 0
	else if n mod 2 = 0 then member (n lsr 1) l
	else member (n lsr 1) r
    | Prefix (p, t) ->
	let b = branching_bit p in
	  if first_bit p x < b then false
	  else member (x lsr (1 lsl b)) t
 

(** Traversals **)
(** Parcours **)
(** Recorridos **)


let rec iter_leaf f k prefix level = function
  | 0 -> ()
  | n ->
      if n mod 2 = 0 then iter_leaf f (k + 1) prefix level (n lsr 1)
      else (
	f ((k lsl level) lor prefix) ; 
	iter_leaf f (k + 1) prefix level (n lsr 1)
      )
      
let rec iter_rec f prefix level = function
  | E -> ()
  | ZLeaf m -> iter_leaf f 1 prefix level m ; f prefix
  | Leaf m -> iter_leaf f 1 prefix level m
  | ZNode (l, m, r) ->
      iter_rec f (prefix lsl 1) (level + 1) l ;
      iter_leaf f 1 prefix level m ;
      f prefix ;
      iter_rec f (1 + prefix lsl 1) (level + 1) r
  | Node (l, m, r) ->
      iter_rec f (prefix lsl 1) (level + 1) l ;
      iter_leaf f 1 prefix level m ;
      iter_rec f (1 + prefix lsl 1) (level + 1) r

let rec iter = fun f trie -> iter_rec f 0 0 trie


(** Compatibilit avec la librairie standard de Caml **)
(** Caml standard library compatibility **)
(** Compatibilidad con la librera estandar de Caml **)


let empty = E
let mem = member
let add = insert
let singleton = injection
let remove = delete
let inter = intersection
let diff = difference
let cardinal = size
let elements = to_list


(** Debug **)

let rec mn = function
  | E | Leaf _ | ZLeaf _ -> 0
  | Node (l, _, r) | ZNode (l, _, r) -> 1 + min (mn l) (mn r)

let rec mx = function
  | E | Leaf _ | ZLeaf _ -> 0
  | Node (l, _, r) | ZNode (l, _, r) -> 1 + max (mx l) (mx r)

let rec med total = function
  | E | Leaf _ | ZLeaf _ -> [total]
  | Node (l, _, r) | ZNode (l, _, r) ->
      med (total + 1) l @ med (total + 1) r

let stat = function tree -> 
  let height_list = med 0 tree in
  let med_h = (List.fold_left (+) 0 height_list) / 
	      (List.length height_list) 
  in
    (size tree, mn tree, mx tree, med_h)


