(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Unbalanced discrete interval encoding tree (diet)
 * Arbre d'intervales discrets non quilibr
 * rbol de intervalos discretos no equilibrado
 * 
 *)

open BairePervasives

(* Discrete interval encoding trees are binary search trees in which all
 * sets of following values are replaced by an interval
 * 
 * eg :     5
 *         / \
 *        2   6       ->      5 7
 *       / \   \              /
 *      1   3   7           1 3
 *
 * The invariants of the data structure are
 * - symmetric order
 * - interval separation
 * 
 * What we mean by interval separation is that there must be at least one
 * value lying stricly between the given two intervals
 * 
 * [1, 5] and [7, 9] are separated by 6
 * [1, 5] and [6, 9] are not separated
 *)

(* Les arbres d'intervalles discrets sont des arbres binaires de recherche
 * dans lesquels tout ensemble de valeurs conscutives est remplace par
 * un intervalle n'indiquant que ses bornes
 * 
 * ex :     5
 *         / \
 *        2   6       ->      5 7
 *       / \   \              /
 *      1   3   7           1 3
 *
 * Les invariants  maintenir sont alors :
 * - l'ordre symtrique
 * - la sparation des intervalles
 * 
 * Par sparation des intervalles nous voulons dire qu'il doit y avoir au moins
 * une valeur strictement comprise entre chaque coupe intervalle successifs
 * 
 * [1, 5] et [7, 9] sont spars par la valeur 6
 * [1, 5] et [6, 9] sont disjoints mais non spars
 *)

(* Los rboles de intervalos discretos son rboles binarios de busqueda en
 * los cuales todo conjuto de valores succesivos se ha reemplzado con un
 * intervalo cuyos limites solo se dan
 * 
 * ej :     5
 *         / \
 *        2   6       ->      5 7
 *       / \   \              /
 *      1   3   7           1 3
 * 
 * Los invariantes de la estructura de datos son
 * - el orden simtrico
 * - la separacin de los intervalos
 * 
 * Lo que se entiende con separacin es que debe haber entre cada par de
 * intervalos por lo menos un valor entre cada par de intervalos succesivos
 * 
 * [1, 5] y [7, 9] estn separados por el valor 6
 * [1, 5] y [6, 9] no estn separados
 *)

type 'a set = E | N of 'a set * 'a * 'a * 'a set


(** Utils **)
(** Utilitaires **)
(** tiles **)


let moduleName = "UnbalancedDiet"

let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type

let successive = fun x y -> if x = y then false else false (* succ x = y *)

let rec max_interval = function
  | E -> raise Not_found
  | N (_, x, y, E) -> (x, y)
  | N (_, _, _, r) -> max_interval r

let rec min_interval = function
  | E -> raise Not_found
  | N (E, x, y, _) -> (x, y)
  | N (l, _, _, _) -> min_interval l

let makeTree = fun l x y r -> N (l, x, y, r)

let rec delete_min_interval = function
  | E -> E
  | N (E, x, y, r) -> r
  | N (l, x, y, r) -> makeTree (delete_min_interval l) x y r

let rec delete_max_interval = function
  | E -> E
  | N (l, x, y, E) -> l
  | N (l, x, y, r) -> makeTree l x y (delete_max_interval r)

(* Notice the use of exceptions to avoid useless code duplication *)

(* Noter l'utilisation du mcanisme d'exceptions pour viter la duplication
 * de code inutile
 *)

(* Notese el uso del mecanismo de excepciones para evitar toda duplicacin
 * inutil de cdigo
 *)

let makeFTree = fun l x y r ->
  let 
    (l', x') =
    (try let (min_l, max_l) = max_interval l in
       if successive max_l x then (delete_max_interval l, min_l)
       else (l, x)
     with Not_found -> (E, x)
    )
  and
    (y', r') =
    (try let (min_r, max_r) = min_interval r in
       if successive y min_r then (max_r, delete_min_interval r)
       else (y, r)
     with Not_found -> (y, E)
    )
  in
    makeTree l' x' y' r'


(** Oprations catgorielles **)
(** Categorial operations **)
(** Operaciones categoriales **)


let injection = function x -> makeTree E x x E

let rec insert_raise v = function
  | E -> injection v
  | N (l, x, y, r) when v < x -> 
      if successive v x then makeFTree l v y r 
      else makeTree (insert_raise v l) x y r
  | N (l, x, y, r) when v > y ->
      if successive y v then makeFTree l x v r 
      else makeTree l x y (insert_raise v r)
  | N _ -> raise Found

let insert = fun v tree ->
  try
    insert_raise v tree
  with Found -> tree

let rec from_list_rec result = function
  | [] -> result
  | head :: tail -> from_list_rec (insert head result) tail

let from_list = function list -> from_list_rec E list

let rec to_list_interval_rec result = function
  | E -> result
  | N (l, x, y, r) -> 
      to_list_interval_rec ((x, y) :: (to_list_interval_rec result r)) l

let to_list_interval = function set -> to_list_interval_rec [] set

let rec list_range result = fun x y ->
  if x = y then x :: result
  else list_range (y :: result) x (y - 1)

let rec to_list_rec result = function
  | E -> result
  | N (l, x, y, r) -> to_list_rec (list_range (to_list_rec result r) x y) l

let to_list = function tree -> to_list_rec [] tree

let map f = function tree -> from_list (List.map f (to_list tree))

let zero = function () -> E


(** Acces **)
(** Accs **)
(** Acceso **)


let rec delete_raise v = function
  | E -> raise Not_found
  | N (l, x, y, r) when v < x -> makeTree (delete_raise v l) x y r
  | N (l, x, y, r) when v > y -> makeTree l x y (delete_raise v r)
  | N (l, x, y, r) ->
      if x = v then makeTree l (succ x) y r
      else if y = v then makeTree l x (pred y) r
      else makeTree (makeTree l x (pred v) E) (succ v) y r
 
let delete = fun v tree ->
  try
    delete_raise v tree
  with Found -> tree

let rec delete_min = function
  | E -> E
  | N (E, x, y, r) -> if x = y then r else makeTree E (succ x) y r
  | N (l, x, y, r) -> makeTree (delete_min l) x y r

let rec delete_max = function
  | E -> E
  | N (l, x, y, E) -> if x = y then l else makeTree l x (pred y) E
  | N (l, x, y, r) -> makeTree l x y (delete_min r)



(*-- Working sources --*)

(*
let rec partition a b = function
  | E -> (E, None, E)
  | N (l, x, y, r) when y < a ->
      let (l', bounds, r') = partition a b r in
	(makeTree l x y l', bounds, r')
  | N (l, x, y, r) when b < x ->
      let (l', bounds, r') = partition a b l in
	(l', bounds, makeTree r' x y r)
  | N (l, x, y, r) ->
      let (l)

*)

let rec interval_list result = function
  | E -> result
  | N (l, x, y, r) -> interval_list ((x, y) :: (interval_list result l)) r

let rec inf_interval_rec result v = function
  | E -> (result, E)
  | N (l, x, y, r) when v > y -> 
      inf_interval_rec ((x, y) :: interval_list result l) v r
  | N (l, x, y, r) when v < x ->
      let (intervals, sup) = inf_interval_rec result v l in
	(intervals, makeTree sup x y r)
  | N (l, x, y, r) -> (((x, y) :: interval_list result l), r)

let rec partition_interval x y = function
  | E -> (E, [], E)
  | N (l, a, b, r) when b < x ->
      let (l', intersection, r') = partition_interval x y r in
	(makeTree l a b l', intersection, r')
  | N (l, a, b, r) when y < a ->
      let (l', intersection, r') = partition_interval x y l in
	(l', intersection, makeTree r' a b r)
  | N (l, a, b, r) ->
      if a <= x && y <= b then (l, [(a, b)], r)
      else if a <= y then
	let (_, intersection, r') = partition_interval x y r in
	  (l, (a, b) :: intersection, r')
      else if y <= b then
	let (l', intersection, _) = partition_interval x y l in
	  (l', (a, b) :: intersection, r)
      else
	let 
	  (l', intersection_l, _) = partition_interval x y l and
	  (_, intersection_r, r') = partition_interval x y r
	in
	  (l', intersection_l @ (a, b) :: intersection_r, r')



let rec min_view = function
  | E -> raise Empty
  | N (E, x, y, r) -> if x = y then (x, r) else (x, makeTree E (succ x) y r)
  | N (l, x, y, r) -> 
      let (min, left) = min_view l in (min, makeTree left x y r)

let rec max_view = function
  | E -> raise Empty
  | N (l, x, y, E) -> if x = y then (l, x) else (makeTree l x (pred y) E, y)
  | N (l, x, y, r) ->
      let (right, max) = max_view r in (makeTree l x y right, max)

let rec partition_inf_sup v = function tree ->
  match tree with
    | E -> (E, None, E)
    | N (l, x, y, r) when v < x ->
	let (inf, value, sup) = partition_inf_sup v l in
	  (inf, value, makeTree sup x y r)
    | N (l, x, y, r) when v > y ->
	let (inf, value, sup) = partition_inf_sup v r in
	  (makeTree l x y inf, value, sup)
    | N (l, x, y, r) ->
	if v = x then (l, Some v, makeTree E (succ x) y r)
	else if v = y then (makeTree l x (pred y) E, Some v, r)
	else (makeTree l x (pred v) E, Some v, makeTree E (succ v) y r)

let is_empty = function
  | E -> true
  | N _ -> false

let rec size_rec total = function
  | E -> total
  | N (l, x, y, r) -> size_rec (size_rec (total + y - x + 1) l) r

let size = function tree -> size_rec 0 tree

let rec member v = function
  | E -> false
  | N (l, x, _, _) when v < x -> member v l
  | N (_, _, y, r) when v > y -> member v r
  | N _ -> true

let rec min_element = function
  | E -> raise Empty
  | N (E, x, _, _) -> x
  | N (l, _, _, _) -> min_element l

let rec max_element = function
  | E -> raise Empty
  | N (_, _, y, E) -> y
  | N (_, _, _, r) -> max_element r

let choose = function
  | E -> raise Empty
  | N (_, x, _, _) -> x

let rec exists_range f = fun x y ->
  match compare x y with
    | 0 -> f y
    | _ -> f x || exists_range f (succ x) y

let rec exists f = function
  | E -> false
  | N (l, x, y, r) -> exists_range f x y || exists f l || exists f r

let rec for_all_range f = fun x y ->
  match compare x y with
    | 0 -> f y
    | _ -> f x && for_all_range f (succ x) y

let rec for_all f = function
  | E -> true
  | N (l, x, y, r) -> for_all_range f x y && for_all f l && for_all f r

let rec look_for_range f = fun x y ->
  if successive x y then raise Not_found
  else if f x then x
  else look_for_range f (succ x) y

let rec look_for f = function
  | E -> raise Not_found
  | N (l, x, y, r) ->
      try look_for f l with Not_found ->
	try look_for_range f x y with Not_found ->
	  look_for f r

let iter_range f = fun x y ->
  for i = x to y do
    f i ;
  done

let rec iter f = function
  | E -> ()
  | N (l, x, y, r) ->
      iter f l ;
      iter_range f x y ;
      iter f r

let rec fold_left_range f e = fun x y ->
  if x = y then f e x
  else fold_left_range f (f e x) (succ x) y

let rec fold_left f e = function
  | E -> e
  | N (l, x, y, r) -> fold_left f (fold_left_range f (fold_left f e l) x y) r

let rec fold_right_range f e = fun x y ->
  if x = y then f x e
  else fold_right_range f (f y e) x (pred y)

let rec fold_right f e = function
  | E -> e
  | N (l, x, y, r) -> 
      fold_right f (fold_right_range f (fold_right f e r) x y) l


