(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 *
 * Unbalanced Set
 * Ensemble non quilibr
 * Conjuto sin balance
 *
 *)

open BairePervasives

let moduleName = "UnbalancedSet"

(* The type definition type 'a tree = Leaf | Node of 'a tree * 'a * 'a tree
 * is in [BairePervasives]
 *)

(* La dfinition de type 'a tree = Leaf | Node of 'a tree * 'a * 'a tree
 * se trouve dans [BairePervasives] afin d'tre visible par tous les
 * modules de la librairie et pouvoir servir de format d'change
 *)

(* La definicin de tipo type 'a tree = Leaf | Node of 'a tree * 'a * 'a tree
 * se encuentra en [BairePervasives] para poder ser visible por todo mdulo
 * de la librera y servir de formato de intercambio
 *)

type 'a path = 
  | Root
  | Left of 'a * 'a tree * 'a path
  | Right of 'a tree * 'a * 'a path

type 'a zipper = 'a tree * 'a path


(** Utility **)
(** Utilitaires **)
(** Utilitarios **)


let invalid = fun function_name error_type ->
  invalid_argument moduleName function_name error_type

let makeTree = fun l v r -> Node (l, v, r)

let rec append = fun a b ->
  match (a, b) with
    | (Leaf, _) -> b
    | (_, Leaf) -> a
    | (Node (l1, v1, r1), Node (l2, v2, r2)) ->
	makeTree l1 v1 (makeTree (append r1 l2) v2 r2)

(* rebuilding is still experimental *)
(* les procedures de reconstruction sont encore sous test *)
(* las funciones de reconstruccin son experimentales *)

let rotate_L = fun l v r ->
  match r with
    | Node (rl, rv, rr) -> makeTree (makeTree l v rl) rv rr
    | _ -> makeTree l v r
	
let rotate_R = fun l v r ->
  match l with
    | Node (ll, lv, lr) -> makeTree ll lv (makeTree lr v r)
    | _ -> makeTree l v r

let rotate_RL = fun l v r ->
  match r with
    | Node (Node (rll, rlv, rlr), rv, rr) -> 
	makeTree (makeTree l v rll) rlv (makeTree rlr rv rr)
    | _ -> makeTree l v r

let rotate_LR = fun l v r ->
  match l with
    | Node (ll, lv, Node (lrl, lrv, lrr)) ->
	makeTree (makeTree ll lv lrl) lrv (makeTree lrr v r)
    | _ -> makeTree l v r

let rec height = function
  | Leaf -> 0
  | Node (l, v, r) -> 1 + max (height l) (height r)
  
(* Simple AVL rebuilding functions *)
(* Procdures de requilibrage AVL simples *)
(* Funciones de reconstruccin AVL simples *)

let makeBTree = fun l v r ->
  let hl = height l and hr = height r in
    if hl > hr + 1 then
      (match l with
	 | Leaf -> invalid "makeBTree" UnexpectedPattern
	 | Node (ll, _, lr) ->
	     if height ll >= height lr then rotate_R l v r 
	     else rotate_LR l v r
      )
    else if hr > hl + 1 then
      (match r with
	 | Leaf -> invalid "makeBTree" UnexpectedPattern
	 | Node (rl, _, rr) ->
	     if height rr >= height rl then rotate_L l v r 
	     else rotate_RL l v r
      )
    else makeTree l v r

let rec makeSTree = fun l v r ->
  let hl = height l and hr = height r in
    if hl > hr + 2 then
      (match l with
	 | Leaf -> invalid "makeSTree" UnexpectedPattern
	 | Node (ll, lv, lr) -> makeBTree ll lv (makeSTree lr v r)
      )
    else if hr > hl + 2 then
      (match r with
	 | Leaf -> invalid "makeSTree" UnexpectedPattern
	 | Node (rl, rv, rr) -> makeBTree (makeSTree l v rl) rv rr
      )
    else makeBTree l v r

let rec simple_rebuild_avl = function
  | Leaf -> Leaf
  | Node (l, v, r) -> makeSTree (simple_rebuild_avl l) v (simple_rebuild_avl r)

(* AVL rebuilding functions *)
(* Procedures de reconstruction AVL en limitant le recalcul *)
(* Funciones de reconstruccin AVL intentando limitar el clculo *)

let height_level_2 = function
  | Leaf -> (0, 0)
  | Node (l, _, r) -> (height l, height r)

let makeBTree = fun l v r hll hlr hrl hrr ->
  let hl = 1 + max hll hlr and hr = 1 + max hrl hrr in
    if hl > hr + 1 then
      (match l with
	 | Leaf -> invalid "makeBTree" UnexpectedPattern
	 | Node (ll, _, lr) ->
	     if hll >= hlr then (rotate_R l v r, hll, 1 + max hlr hr)
	     else let (hlrl, hlrr) = height_level_2 lr in
	       (rotate_LR l v r, 1 + max hll hlrl, 1 + max hlrr hr)
      )
    else if hr > hl + 1 then
      (match r with
	 | Leaf -> invalid "makeBTree" UnexpectedPattern
	 | Node (rl, _, rr) ->
	     if hrr >= hrl then (rotate_L l v r, 1 + max hl hrl, hrr)
	     else let (hrll, hrlr) = height_level_2 rl in
	       (rotate_RL l v r, 1 + max hl hrll, 1 + max hrlr hrr)
      )
    else (makeTree l v r, hl, hr)

let rec makeSTree = fun l v r hll hlr hrl hrr ->
  let hl = 1 + max hll hlr and hr = 1 + max hrl hrr in
    if hl > hr + 2 then
      (match l with
	 | Leaf -> invalid "makeSTree" UnexpectedPattern
	 | Node (ll, lv, lr) -> 
	     let 
	       (hlll, hllr) = height_level_2 ll and
	       (hlrl, hlrr) = height_level_2 lr
	     in
	     let (r', hrl, hrr) = (makeSTree lr v r hlrl hlrr hrl hrr) in
	       makeBTree ll lv r' hlll hllr hrl hrr
      )
    else if hr > hl + 2 then
      (match r with
	 | Leaf -> invalid "makeSTree" UnexpectedPattern
	 | Node (rl, rv, rr) -> 
	     let
	       (hrll, hrlr) = height_level_2 rl and
	       (hrrl, hrrr) = height_level_2 rr
	     in
	     let (l', hll, hlr) = makeSTree l v rl hll hlr hrll hrlr in
	       makeBTree l' rv rr hll hlr hrrl hrrr
      )
    else makeBTree l v r hll hlr hrl hrr

(* This function only computes again the total height on double rotations *)

(* Cette fonction ne recalcule la hauteur totale de l'arbre que lors des
 * doubles rotations. Autrement l'information ncessaire est transmise par les 
 * constructeurs
 *)

(* Esta funcin solo calcula la altura total de rbol cuando se hacen
 * rotaciones dobles. En todos los otros casos, la informacin necesaria
 * es transmitida por los constructores 
 *)

let rec rebuild_avl_rec = function
  | Leaf -> (Leaf, -1, -1) (* provoque moins de double rotations que 0, 0 *)
  | Node (l, v, r) ->
      let 
	(left, hll, hlr) = rebuild_avl_rec l and
	(right, hrl, hrr) = rebuild_avl_rec r
      in
	makeSTree left v right hll hlr hrl hrr

let rebuild_avl = function tree -> 
  let (new_tree, _, _) = rebuild_avl_rec tree in
    new_tree


(** Oprations catgorielles **)
(** Categorial operations **)
(** Operaciones categoriales **)


let injection = function x -> makeTree Leaf x Leaf

let rec insert_raise x = function
  | Leaf -> injection x
  | Node (l, v, r) ->
      match compare x v with
	| n when n < 0 -> makeTree (insert_raise x l) v r
	| n when n > 0 -> makeTree l v (insert_raise x r)
	| _ -> raise Found

let insert x = function set ->
  try 
    insert_raise x set 
  with Found -> set

let rec from_list_rec result = function
  | [] -> result
  | head :: tail -> from_list_rec (insert head result) tail

let from_list = function list -> from_list_rec Leaf list

let rec to_list_rec result = function
  | Leaf -> result
  | Node (l, v, r) -> to_list_rec (v :: to_list_rec result r) l

let to_list = function set -> to_list_rec [] set

let map f = function set -> from_list (List.map f (to_list set))

(* Union algorithm :
 * Explained in the research report (avaible on the web)
 * 
 * Implementing Sets Efficiently in a Functional Language
 * Stephen Adams
 * CSTR 92-01
 * University of Southampton
 * 
 * We first give a version obtained by  divide and conquer 
 *)

(* Algorithme d'union :
 * Son fonctionnement est expliqu dans le rapport de recherche 
 * (disponible sur la toile)
 *
 * Implementing Sets Efficiently in a Functional Language
 * Stephen Adams
 * CSTR 92-01
 * University of Southampton
 *
 * Nous en rappelons ici le principe :
 * On construit une premire version via la mthode  diviser pour rgner 
 *)

(* Algoritmo de unin :
 * Su funcionamiento est explicado en el informe (disponible en la red)
 * 
 * Implementing Sets Efficiently in a Functional Language
 * Stephen Adams
 * CSTR 92-01
 * University of Southampton
 *
 * Recordamos rapidamente su principio :
 * Uno comienza construyendo una primera versin con el mtodo  dividir
 * para reinar 
 *)

let rec filter_lt x = function
  | Leaf -> Leaf
  | Node (l, v, r) ->
      match compare x v with
	| n when n < 0 -> filter_lt x l
	| n when n > 0 -> makeTree l v (filter_lt x r)
	| _ -> l

let rec filter_gt x = function
  | Leaf -> Leaf
  | Node (l, v, r) ->
      match compare x v with
	| n when n < 0 -> makeTree (filter_gt x l) v r
	| n when n > 0 -> filter_gt x r
	| _ -> r

(* We could have used a unique partition function like [partition_inf_sup]
 * but the to filter functions will be useful later in [plus]
 *)

(* Nous aurions pu utiliser une unique fonction de partition telle
 * [partition_inf_sup] mais les filtres dfinis nous seront utiles sparment 
 * dans la version optimise de [plus]
 *)
	
(* Ubisemos podido utilizar una nica funcin de particin [partition_inf_sup]
 * pero los dos filtros que hemos definido nos sern tiles en la versin
 * optimizada de la funcin [plus]
 *)
    
let rec old_plus = fun a b ->
  match (a, b) with
    | (Leaf, _) -> b
    | (_, Leaf) -> a
    | (Node (l, v, r), Node _) -> 
	let 
	  lt_v = filter_lt v b and 
	  gt_v = filter_gt v b 
	in
	  makeTree (old_plus l lt_v) v (old_plus r gt_v)
	    

(* One can notice that there is no need to construct a tree (lt_v and gt_v)
 * which will be destructed just after by the recursive call to [old_plus]. 
 * That is why we introduce a triple containing a tree and two bounds. 
 * These bounds can be easely moved then.
 *)

(* On constate qu' rien ne sert de construire un arbre (lt_v et gt_v) qui sera
 * dcoup juste aprs par l'appel rcursif de la fonction [old_plus], raison 
 * pour laquelle on introduit un triplet contenant un arbre et deux bornes 
 * qui pourront tre modifies  moindre cot
 *)

(* Uno se da cuenta que de nada sirve construir un rbol (lt_v y gt_v) que ser
 * destruido justo despus al llamar recursivamente la funcin [old_plus].
 * Se introduce por esa razn un triplete formado por un rbol y dos 
 * limites, lmites que se podrn modificar lteriormente a poco costo 
 *)

(* [cut] constructs a triple, making sure that the root of the tree lies 
 * between the two bounds. If it is not the case, the functions calls itself
 * recusively on the corresponding subtree (right or left)
 *)

(* [cut] construit un triplet de faon  ce que la racine se trouve entre
 * les deux bornes. Si ce n'est pas le cas, la fonction restreint rcursivement
 * l'arbre  son sous-arbre gauche ou droit selon le cas de figure.
 *) 

(* [cut] constuye un triplete de manera tal a que la raz del rbol se 
 * encuentre entre los dos lmites. Si este no es el caso, la funcin
 * toma recursivamente el subrbol derecho o izquierdo segn convenga
 *)

let rec cut = fun lo tree hi ->
  match tree with
    | Node (_, v, r) when compare v lo <= 0 -> cut lo r hi
    | Node (l, v, _) when compare v hi >= 0 -> cut lo l hi
    | _ -> (lo, tree, hi)

(* [cut_lo] and [cut_hi] are specialized versions with an infinite 
 * hight (respectively low) bound 
 *)

(* [cut_lo] et [cut_hi] sont des versions spcialises avec la borne infrieure
 * (respectivement suprieure) ayant pour valeur plus l'infini (respectivement 
 * moins l'infini)
 *)

(* [cut_lo] y [cut_hi] son versiones especializadas con un limite inferior
 * (respectivamente superior) igual a ms (respectivamente menos) el infinito
 *)

let rec cut_hi = fun lo tree ->
  match tree with
    | Node (_, v, r) when compare v lo <= 0 -> cut_hi lo r
    | _ -> (lo, tree)

let rec cut_lo = fun tree hi ->
  match tree with
    | Node (l, v, _) when compare v hi >= 0 -> cut_lo l hi
    | _ -> (tree, hi)

(* [uni] uses the left tree to control the reconstruction :
 * - the right tree will be physically cut only if it lies between to 
 * successive values of the control tree (which guarantees that there will not
 * be any new cut after the reconstruction)
 * - if the right tree is empty, the whole control tree can be inserted
 *)

(* la fonction [uni] se sert de l'arbre de gauche pour contrler la
 * reconstruction : 
 * - l'arbre de droite ne sera dcoup physiquement que si l'on se trouve 
 * entre deux valeurs successives de l'arbre de contrle (ce qui garantit 
 * qu'il n'y aura pas ne nouvelle dcoupe aprs la reconstruction)
 * - si l'arbre de droite est vide, on peut insrer tout l'arbre de contrle
 *)

(* la funcin [uni] se sirve del rbol izquierdo para controlar la 
 * reconstruccin :
 * - el rbol derecho solo sera recortado fsicamente si se encuentra
 * entre dos valores succesivos del rbol de control (lo que garantiza
 * que no habr ninguna otra operacin de corte despues de la reconstruccin)
 * - si el rbol derecho est vaco, se puede insertar el rbol de control
 * entero
 *)

let rec uni = fun a (lo, b, hi) ->
  match (a, b) with
    | (_, Leaf) -> a
    | (Leaf, Node (l, v, r)) -> makeTree (filter_gt lo l) v (filter_lt hi r)
    | (Node (l, v, r), Node _) ->
	makeTree (uni l (cut lo b v)) v (uni r (cut v b hi))  

(* two specialized versions are given *)

(* on en donne galement deux versions spcialises *)

(* se dan tambin dos versiones especializadas *)

let rec uni_lo = fun a (b, hi) ->
  match (a, b) with
    | (_, Leaf) -> a
    | (Leaf, Node (l, v, r)) -> makeTree l v (filter_lt hi r)
    | (Node (l, v, r), Node _) -> 
	makeTree (uni_lo l (cut_lo b v)) v (uni r (cut v b hi))

let rec uni_hi = fun a (lo, b) ->
  match (a, b) with
    | (_, Leaf) -> a
    | (Leaf, Node (l, v, r)) -> makeTree (filter_gt lo l) v r
    | (Node (l, v, r), Node _) -> 
	makeTree (uni l (cut lo b v)) v (uni_hi r (cut_hi v b)) 

(* We just put all together *)
	
(* On assemble le tout *)

(* Se juntan todos los pedazos *)

let plus = fun a b ->
  match (a, b) with
    | (Leaf, _) -> b
    | (_, Leaf) -> a
    | (Node (l, v, r), Node _) ->
	makeTree (uni_lo l (cut_lo b v)) v (uni_hi r (cut_hi v b)) 

let zero = function () -> Leaf

let rec join_rec result = function
  | Leaf -> result
  | Node (l, v, r) -> join_rec (join_rec (plus v result) l) r 

let join = function tree -> join_rec Leaf tree


(** Binary operations **)
(** Oprations binaires **)
(** Opraciones binrias **)


let rec partition_inf_sup pivot = function
  | Leaf -> (Leaf, None, Leaf)
  | Node (l, v, r)  ->
      match compare pivot v with
	| n when n < 0 ->
	    let (inf, value, sup) = partition_inf_sup pivot l in
	      (inf, value, makeTree sup v r)
	| n when n > 0 ->
	    let (inf, value, sup) = partition_inf_sup pivot r in
	      (makeTree l v inf, value, sup)
	| _ -> (l, Some v, r)

let union = plus

let rec intersection = fun a b ->
  match (a, b) with
    | (Leaf, _) | (_, Leaf) -> Leaf
    | (Node (l, v, r), Node _) ->
	match partition_inf_sup v b with
	  | (l', None, r') -> append (intersection l l') (intersection r r')
	  | (l', Some _, r') -> 
	      makeTree (intersection l l') v (intersection r r')

let rec difference = fun a b ->
  match (a, b) with
    | (Leaf, _) -> Leaf
    | (_, Leaf) -> a
    | (Node (l, v, r), Node _) -> 
	match partition_inf_sup v b with
	  | (l', None, r') -> makeTree (difference l l') v (difference r r')
	  | (l', Some _, r') -> append (difference l l') (difference r r')
	      
let rec symetric_difference = fun a b ->
  match (a, b) with
    | (Leaf, s) | (s, Leaf) -> s
    | (Node (l, v, r), Node _) ->
	match partition_inf_sup v b with
	  | (l', None, r') -> makeTree (symetric_difference l l') v 
	      (symetric_difference r r')
	  | (l', Some _, r') -> 
	      append (symetric_difference l l') (symetric_difference r r')

let rec subset = fun a b ->
  match (a, b) with
    | (Leaf, _) -> true
    | (_, Leaf) -> false
    | (Node (l1, v1, r1), Node (l2, v2, r2)) ->
        match compare v1 v2 with
	  | n when n < 0 -> subset (makeTree l1 v1 Leaf) l2 && subset r1 b
	  | n when n > 0 -> subset (makeTree Leaf v1 r1) r2 && subset l1 b
	  | _ -> subset l1 l2 && subset r1 r2

let rec compare_rec = fun a b ->
  match (a, b) with
    | ([], []) -> 0
    | ([], _) -> -1
    | (_, []) -> 1
    | (Leaf :: t1, Leaf :: t2) -> compare_rec t1 t2
    | (Node (Leaf, v1, r1) :: t1, Node (Leaf, v2, r2) :: t2) ->
	(match compare v1 v2 with
	  | 0 -> compare_rec (r1 :: t1) (r2 :: t2)
	  | other -> other
	)
    | (Node (a, v1, r1) :: t1, t2) ->
        compare_rec (a :: Node (Leaf, v1, r1) :: t1) t2
    | (t1, Node (b, v2, r2) :: t2) ->
	compare_rec t1 (b :: Node (Leaf, v2, r2) :: t2)

let equal = fun a b -> compare_rec [a] [b] = 0

(* let compare = fun a b -> compare_rec [a] [b]
 *
 * renvoy  la fin du fichier pour ne pas interfrer avec Pervasives.compare 
 * 
 * see the end of the file
 * 
 * ver el final del fichero
 *)


(** Acces **)
(** Accs **)
(** Acceso **)


(* [insert'] :
 * Variant of the [insert] function, it does only one comparasion
 * by branch instead of 3/2 but has to walk down the whole tree
 * 
 * If the inserted element is not already in the tree, the [insert] function
 * has anyway to go down the tree until the leafs, so in this particular
 * point the lost is not very important. The reason why this function
 * is not always faster than [insert] is that one more argument
 * must be transmitted in each recursive call
 *)

(* [insert'] :
 * Variante de la fonction [insert] ne faisant qu'une comparaison par
 * branche (au lieu de 3/2 en moyenne pour la fonction habituelle) mais
 * parcourt l'arbre jusqu' profondeur maximale
 *
 * Hormis le cas o l'lment insr se trouve dj dans l'arbre, la fonction
 * [insert] effectue un parcours sur toute la profondeur. La perte est donc
 * minime sur ce point, mais un argument supplmentaire doit tre transmis, 
 * raison pour laquelle cette fonction n'est pas toujours plus rapide que 
 * la version la plus simple
 *)

(* [insert'] :
 * Una variante de la funcin [insert] que hace una sola comparacin por
 * nivel (en vez de 3/2 en promedio para la version comn) pero recorre la 
 * totalidad del rbol, hasta las hojas.
 * 
 * Excepto en el caso en el cual el lemento que se est insertando se 
 * encuentra ya en el rbol, la funcin [insert] recorre de todas maneras
 * el rbol hasta las hojas, por lo cual la perdida en ese aspecto es
 * mnima. No obstante, un argumento suplementrio debe ser transmitido y
 * por ello esta funcin no es todo el tiempo ms rpida que la version
 * la mas simple
 *)

let rec insert_min_bound b x = function
  | Leaf -> 
      if compare b x = 0 then raise Found
      else injection x
  | Node (l, v, r) ->
      match compare x v with
	| n when n < 0 -> makeTree (insert_min_bound b x l) v r
	| _ -> makeTree l v (insert_min_bound v x r)

let rec insert_raise' x = function
  | Leaf -> injection x
  | Node (l, v, r) ->
      match compare x v with
	| n when n < 0 -> makeTree (insert_raise' x l) v r
	| _ -> makeTree l v (insert_min_bound x v r)

let insert' x = function set ->
  try 
    insert_raise' x set 
  with Found -> set

let insert_list = fun list set -> plus (from_list list) set

let rec delete_min = function
  | Leaf -> Leaf
  | Node (Leaf, _, r) -> r
  | Node (l, v, r) -> makeTree (delete_min l) v r

let rec delete_max = function
  | Leaf -> Leaf
  | Node (l, _, Leaf) -> l
  | Node (l, v, r) -> makeTree l v (delete_max r)

let rec delete_raise x = function
  | Leaf -> raise Not_found
  | Node (l, v, r) ->
      match compare x v with
	| n when n < 0 -> makeTree (delete_raise x l) v r
	| n when n > 0 -> makeTree l v (delete_raise x r)
	| _ -> append l r

let delete x = function set -> 
  try 
    delete_raise x set 
  with Not_found -> set

 let rec delete_list = fun list set ->
  match list with
    | [] -> set
    | head :: tail -> delete_list tail (delete head set)

let rec filter_rec f result = function
  | Leaf -> result
  | Node (l, v, r) ->
      filter_rec f (filter_rec f (if f v then insert v result else result) l) r

let filter f = function set -> filter_rec f Leaf set

let rec filter_le x = function
  | Leaf -> Leaf
  | Node (l, v, r) ->
      match compare x v with
	| n when n < 0 -> filter_le x l
	| _ -> makeTree l v (filter_le x r)

let rec filter_ge x = function
  | Leaf -> Leaf
  | Node (l, v, r) ->
      match compare x v with
	| n when n <= 0 -> makeTree (filter_ge x l) v r
	| _ -> filter_ge x r

let rec range min max = function
  | Leaf -> Leaf
  | Node (l, v, r) ->
      if compare v min < 0 then range min max r
      else if compare v max > 0 then range min max l
      else makeTree (filter_ge min l) v (filter_le max r)


(** Views **)
(** Vues **)
(** Vistas **)


let rec min_view = function
  | Leaf -> raise Empty
  | Node (Leaf, v, r) -> (v, r)
  | Node (l, v, r) -> let (min, left) = min_view l in (min, makeTree left v r)

let rec max_view = function
  | Leaf -> raise Empty
  | Node (l, v, Leaf) -> (l, v)
  | Node (l, v, r) -> let (right, max) = max_view r in 
      (makeTree l v right, max)

let rec partition_rec f (trueS, falseS) = function
  | Leaf -> (trueS, falseS)
  | Node (l, v, r) -> 
      if f v 
      then partition_rec f (partition_rec f (insert v trueS, falseS) l) r
      else partition_rec f (partition_rec f (trueS, insert v falseS) l) r

let partition = fun f set -> partition_rec f (Leaf, Leaf) set


(** Observers **)
(** Observateurs **)
(** Observadores **)


let is_empty = function
  | Leaf -> true
  | Node _ -> false

let rec size_rec result = function
  | Leaf -> result
  | Node (l, _, r) -> size_rec (size_rec (1 + result) r) l

let size = function tree -> size_rec 0 tree


(* [member'] : does (d + 1) comparations instead of 3/2 * d
 * following an idea of Arne Andersson, but walks down the whole height of 
 * the tree and an extra argument has to be transmitted in each call
 * 
 * It may not always be faster than [member]
 *)

(* [member'] : effectue en moyenne (d + 1) comparaisons au lieu de 3/2 * d
 * selon une ide d'Arne Andersson mais parcourt systmatiquement l'arbre sur
 * toute sa hauteur et doit transmettre un paramtre supplmentaire  chaque
 * appel
 * 
 * Il est possible qu'elle ne soit pas toujours plus rapide que [member]
 *) 

(* [member'] : hace (d + 1) comparaciones en vez de 3/2 * d en promedio, 
 * segn una idea de Arne Andersson, pero recorre toda la altura del rbol
 * y requiere un parmetro suplementrio en cada llamado
 * 
 * Es posible que no sea siempre ms veloz que [member]
 *)

let rec member_min_bound b x = function
  | Leaf -> compare b x = 0
  | Node (l, v, r) ->
      match compare x v with
	| k when k < 0 -> member_min_bound b x l
	| _ -> member_min_bound v x r
	    
let rec member' x = function
  | Leaf -> false
  | Node (l, v, r) ->
      match compare x v with
	| n when n < 0 -> member' x l
	| _ -> member_min_bound v x r
	    
(* [member] : usual member function
 * some algorithms use the properties of the usual [member] function 
 * (stop of a positive search in the middle of the tree). It is the case
 * of optimal tree construction for example (Knuthm Garcia-Walsh, confer
 * The art of computer programming vol III)
 *)

(* [member] : fonction member habituelle 
 * un certain nombre d'algorithmes utilisent les proprits de la fonction
 * [member] habituelle (arrt d'une recherche positive  l'intrieur de 
 * l'arbre) par exemple les algorithmes de construction d'arbres optimaux
 * (Knuth, Garcia-Walsh, confer The art of computer programming vol III)
 *)

(* [member] : funcin member comn
 * ciertos algortmos usan las propriedades de la funcin [member] comn
 * (interupcin de una busqueda positiva en el medio del rbol) tal
 * los algortmos de construccin de rboles optimales (Knuth, Garcia-Walsh,
 * confer The art of computer programming vol III)
 *)

let rec member x = function
  | Leaf -> false
  | Node (l, v, r) ->
      match compare x v with
	| n when n < 0 -> member x l
	| n when n > 0 -> member x r
	| _ -> true

let rec look_for f = function
  | Leaf -> raise Not_found
  | Node (_, v, _) when f v -> v
  | Node (l, _, r) -> (try look_for f l with Not_found -> look_for f r)

let choose = function
  | Leaf -> raise Empty
  | Node (_, v, _) -> v

let rec min_element = function
  | Leaf -> raise Not_found
  | Node (Leaf, v, _) -> v
  | Node (l, _, _) -> min_element l

let rec max_element = function
  | Leaf -> raise Not_found
  | Node (_, v, Leaf) -> v
  | Node (_, _, r) -> max_element r

(* Could also be done with ['a option] *)

(* Auraient aussi pu tre codes au moyen de ['a option], nous avons
 * prfr illustrer les interruptions de calcul au moyens d'exceptions
 *)

(* stas operaciones tambin se hubiesen podido escribir con el typo 
 * ['a option] pero hemos querido aqu mostrar el uso de las excepciones para 
 * ruptura de clculo
 *)

let rec next x = function
  | Leaf -> raise Not_found
  | Node (l, v, r) ->
      match compare x v with
	| n when n < 0 -> (try next x l with Not_found -> v)
	| n when n > 0 -> next x r
	| _ -> min_element r (* raises [Not_found] *)

let rec previous x = function
  | Leaf -> raise Not_found
  | Node (l, v, r) ->
      match compare x v with
	| n when n < 0 -> previous x l
	| n when n > 0 -> (try previous x r with Not_found -> v)
	| _ -> max_element l (* raises [Not_found] *)
	      
let rec ceil x = function
  | Leaf -> raise Not_found
  | Node (l, v, r) ->
      match compare x v with
	| n when n < 0 -> (try ceil x l with Not_found -> v)
	| n when n > 0 -> ceil x r
	| _ -> x

let rec floor x = function
  | Leaf -> raise Not_found
  | Node (l, v, r) ->
      match compare x v with
	| n when n < 0 -> floor x l
	| n when n > 0 -> (try floor x r with Not_found -> v)
	| _ -> x

(* les arbres binaires de recherche sont indexs suivant l'ordre symtrique *)

(* binary search trees are indexed following the symetric order *)

(* los rboles de busqueda estn indizados segn el orden simtrico *)

let rec rank x = function
  | Leaf -> raise Not_found
  | Node (l, v, r) ->
      match compare x v with
	| n when n < 0 -> rank x l
	| n when n > 0 -> 1 + size l + rank x r
	| _ -> size l

let rec get n = function
  | Leaf -> raise (Out_of_bounds n)
  | Node (l, v, r) ->
      let k = size l in
	if n < k then get n l
	else if n > k then get (n - k - 1) r
	else v

let rec set_rec value m = function
  | Leaf -> raise (Out_of_bounds m)
  | Node (l, v, r) ->
      let k = size l in
	if m < k then makeTree (set_rec value m l) v r
	else if m > k then makeTree l v (set_rec value (m - k - 1) r)
	else if v = value then raise Found
	else makeTree l value r
	  
let unsafe_set = fun value n tree -> 
  try 
    set_rec value n tree 
  with Found -> tree

let rec for_all f = function
  | Leaf -> true
  | Node (l, v, r) -> f v && for_all f l && for_all f r

let rec exists f = function
  | Leaf -> false
  | Node (l, v, r) -> f v || exists f l || exists f r


(** Traversals **)
(** Parcours **)
(** Recorridos **)


let rec iter f = function
  | Leaf -> ()
  | Node (l, v, r) -> iter f l ; f v ; iter f r

let rec fold_left f e = function
  | Leaf -> e
  | Node (l, v, r) -> fold_left f (f (fold_left f e l) v) r

let rec fold_right = fun f set e ->
  match set with
    | Leaf -> e 
    | Node (l, v, r) -> fold_right f l (f v (fold_right f r e))

let fold = fold_right


(** Zipper operations **)
(** Operations sur les zippers **)
(** Operaciones sobre los zippers **)


(* Zippers are functional pointers and allow all pointer operations on a
 * full typed and secure (single threaded) maner :
 * 
 * pointer local movements O(1)
 * 
 * - move_left : moves the current pointer to the left child
 * - move_right : moves the current pointer to the right child
 * - move_up : moves the current pointer to the father
 * 
 * pointer acces O(1)
 * 
 * - current_tree : returns the current node (tree)
 * - current_value : returns the value of the current node
 * 
 * pointer global movement O(log n)
 * 
 * - move_to : moves the current pointer to the node having the given value
 * - move_top : moves the current pointer to the top of the tree
 * 
 * insertions
 * 
 * - zipper_insert : insertion of a value
 * - replace_tree : replaces the current node (tree) with the given node (tree)
 * 
 * constructors
 * 
 * - make_zipper : returns a zipper on the top of a given tree
 * - return_tree : returns the under? tree of a zipper
 * 
 * Confer unbalancedSet.mli for all avaible operations on zippers
 *)

(* Les zippers sont des pointeurs fonctionnels qui permettent toutes les
 * oprations usuelles sur les pointeurs de manire fortement type
 * et sre. Cependant un seul pointeur est permis et les oprations sur
 * les pointeurs allouent de la mmoire
 * 
 * mouvements locaux de pointeur O(1)
 * 
 * - move_left : dplace le pointeur courant vers le fils gauche
 * - move_right : dplace le pointeur courant vers le fils droit
 * - move_up : dplace le pointeur courant vers le pre
 * 
 * accs aux lments points O(1)
 * 
 * - current_tree : renvoie le noeud (arbre) point
 * - current_value : renvoie la valeur du noeud point
 * 
 * mouvements globaux de pointeur O(log n)
 * 
 * - move_to : dplace le pointeur courant vers le noeud de valeur spcife
 * - move_top : dplace le pointeur courant vers la racine
 * 
 * insertions
 * 
 * - zipper_insert : insre une valeur
 * - replace_tree : remplace le sous-arbre point
 * 
 * contructeurs
 * 
 * - make_zipper : renvoie un pointeur sur un arbre donn
 * - return_tree : renvoie l'arbre sous jacent (en entier)
 * 
 * Se reporter au fichier "unbalancedSet.mli" pour connatre toutes les 
 * oprations disponibles sur les zippers
 *)


exception Bottom
exception Top

let make_zipper = function tree -> (tree, Root)

let move_left = fun (tree, path) ->
  match tree with
    | Leaf -> raise Bottom
    | Node (l, v, r) -> (l, Left (v, r, path))

let move_right = fun (tree, path) ->
  match tree with
    | Leaf -> raise Bottom
    | Node (l, v, r) -> (r, Right (l, v, path))

let move_up = fun (tree, path) ->
  match path with
    | Root -> raise Top
    | Left (v, r, tail) -> (makeTree tree v r, tail)
    | Right (l, v, tail) -> (makeTree l v tree, tail)

let rec move_top = function ((tree, path) as pointer) ->
  match path with
    | Root -> pointer
    | Left (v, r, tail) -> move_top (makeTree tree v r, tail)
    | Right (l, v, tail) -> move_top (makeTree l v tree, tail)

let rec move_to x = function ((tree, path) as pointer) ->
  match tree with
    | Leaf -> 
	(match path with
	   | Right (_, rv, _) when x <= rv -> move_to x (move_up pointer)
	   | Left (lv, _, _) when x >= lv -> move_to x (move_up pointer)
	   | _ -> raise Not_found
	)
    | Node (_, v, _) ->
	match compare x v with
	  | n when n < 0 ->
	      (match path with
		 | Right (_, rv, _) when x < rv -> move_to x (move_up pointer)
		 | Right _ | Root | Left _ -> move_to x (move_left pointer)
	      )
	  | n when n > 0 ->
	      (match path with
		 | Left (lv, _, _) when x > lv -> move_to x (move_up pointer)
		 | Left _ | Root | Right _ -> move_to x (move_right pointer)
	      )
	  | _ -> pointer

let current_tree = function (tree, _) -> tree

let current_value = function (tree, _) ->
  match tree with
    | Leaf -> None
    | Node (_, v, _) -> Some v

let current_value' = function (tree, _) ->
  match tree with
    | Leaf -> raise Empty
    | Node (_, v, _) -> v

let rec zipper_insert x = function ((tree, path) as pointer) ->
  match tree with
    | Leaf -> 
	(match path with
	   | Right (_, rv, _) when x <= rv -> zipper_insert x (move_up pointer)
	   | Left (lv, _, _) when x >= lv -> zipper_insert x (move_up pointer)
	   | _ -> (makeTree Leaf x Leaf, path)
	)
    | Node (_, v, _) ->
	match compare x v with
	  | n when n < 0 ->
	      (match path with
		 | Right (_, rv, _) when x < rv -> 
		     zipper_insert x (move_up pointer)
		 | Right _ | Root | Left _ -> 
		     zipper_insert x (move_left pointer)
	      )
	  | n when n > 0 ->
	      (match path with
		 | Left (lv, _, _) when x > lv -> 
		     zipper_insert x (move_up pointer)
		 | Left _ | Root | Right _ -> 
		     zipper_insert x (move_right pointer)
	      )
	  | _ -> raise Found

let return_tree = function zipper -> let (tree, _) = move_top zipper in tree

let unsafe_replace_tree = fun tree (_, path) -> (tree, path)


(** Miscellaneous **)
(** Divers **)
(** Miscelanios **)


let bind = fun set f -> join (map f set)

let concat_map = fun f set -> join (map f set)

let rec unsafe_inc_map f = function
  | Leaf -> Leaf
  | Node (l, v, r) -> makeTree (unsafe_inc_map f l) (f v) (unsafe_inc_map f r)

let rec unsafe_dec_map f = function
  | Leaf -> Leaf
  | Node (l, v, r) -> makeTree (unsafe_dec_map f l) (f v) (unsafe_dec_map f r)

let rec unsafe_insert_min x = function
  | Leaf -> injection x
  | Node (l, v, r) -> makeTree (unsafe_insert_min x l) v r 

let rec unsafe_insert_max x = function
  | Leaf -> injection x
  | Node (l, v, r) -> makeTree l v (unsafe_insert_max x r)


(** Compatibilit avec la librairie standard de Caml **)
(** Caml standard library compatibility **)
(** Compatibilidad con la librera estandar de Caml **)


let empty = Leaf
let mem = member
let add = insert
let singleton = injection
let remove = delete
let inter = intersection
let diff = difference
let cardinal = size
let elements = to_list
let min_elt = min_element    
let max_elt = max_element


(** Objets **)
(** Ojects **)
(** Objetos **)


class ['a] o_set = fun set ->

object (self : 'b)
  
  val contents  = set

  method to_set = contents

  (* Categorial operations *)
  (* Oprations categorielles *)
  (* Operaciones categoriales *)

  method map f = {< contents = map f contents >}

  method plus = function (set : 'b) -> 
    {< contents = plus contents set#to_set >} 

  method to_list = to_list contents
		     
  (* Binary operations *)
  (* Oprations binaires *)
  (* Operaciones binrias *)
		     
  method union = function (set : 'b) -> 
    {< contents = plus contents set#to_set >}
  
  method intersection = function (set : 'b) -> 
    {< contents = intersection contents set#to_set >}
  
  method difference = function (set : 'b) -> 
    {< contents = difference contents set#to_set >}
  
  method symetric_difference = function (set : 'b) -> 
    {< contents = symetric_difference contents set#to_set >}
    
  method subset = function (set : 'b) -> subset contents set#to_set
  method equal = function (set : 'b)-> equal contents set#to_set
  method compare = function (set : 'b) -> compare_rec [contents] [set#to_set]
    
  (* Acces *)
  (* Accs *)
  (* Acceso *)

  method insert = function x -> 
    try 
      {< contents = insert_raise x contents >} 
    with Found -> self

  method insert_raise = function x -> {< contents = insert_raise x contents >}

  method insert' = function x -> 
    try
      {< contents = insert_raise' x contents >}
    with Found -> self

  method insert_raise' = function x -> 
    {< contents = insert_raise' x contents >}
      
  method insert_list = function list -> 
    {< contents = insert_list list contents >}

  method delete = function x -> 
    try 
      {< contents = delete_raise x contents >} 
    with Not_found -> self

  method delete_raise = function x -> {< contents = delete_raise x contents >}

  method delete_list = function list -> 
    {< contents = delete_list list contents >}
      
  method delete_min = {< contents = delete_min contents >}
  method delete_max = {< contents = delete_max contents >}

  method filter = function f -> {< contents = filter f contents >}
  method filter_lt = function x -> {< contents = filter_lt x contents >}
  method filter_gt = function x -> {< contents = filter_gt x contents >}
  method filter_le = function x -> {< contents = filter_le x contents >}
  method filter_ge = function x -> {< contents = filter_ge x contents >}

  method range = fun min max -> {< contents = range min max contents >}

  (* Views *)
  (* Vues *)
  (* Vistas *)

  method min_view =
    let (min, other) = min_view contents in
      (min, {< contents = other >})

  method max_view =
    let (other, max) = max_view contents in
      ({< contents = other >}, max)

  method partition = function f -> 
    let (trueSet, falseSet) = partition f contents in
      ({< contents = trueSet >}, {< contents = falseSet >})

  method partition_inf_sup = function x ->
    let (inf, value, sup) = partition_inf_sup x contents in
      ({< contents = inf >}, value, {< contents = sup >})

  (* Observators *)
  (* Observateurs *)
  (* Observadores *)
      
  method is_empty = contents = Leaf
  method size = size contents

  method member = function x -> member x contents
  method member' = function x -> member' x contents
  method look_for = function f -> look_for f contents

  method choose =
    match contents with
      | Leaf -> raise Not_found
      | Node (_, v, _) -> v

  method min_element = min_element contents
  method max_element = max_element contents

  method previous = function x -> previous x contents
  method next = function x -> next x contents
  method floor = function x -> floor x contents
  method ceil = function x -> ceil x contents

  method rank = function x -> rank x contents
  method get = function n -> get n contents

  method exists = function f -> exists f contents
  method for_all = function f -> for_all f contents

  (* Traversals *)
  (* Parcours *)
  (* Recorridos *)

  method iter : ('a -> unit) -> unit = function f -> iter f contents

  method fold : 'c. ('a -> 'c -> 'c) -> 'c -> 'c = 
    fun f e -> fold_right f contents e

  method fold_left : 'c. ('c -> 'a -> 'c) -> 'c -> 'c =
    fun f e -> fold_left f e contents

  method fold_right : 'c. ('a -> 'c -> 'c) -> 'c -> 'c =
    fun f e -> fold_right f contents e

end

(* Constructors *)
(* Constructeurs *)
(* Constructores *)

let o_injection = function x -> new o_set (injection x)
let o_map f = function a -> new o_set (map f a#to_set)
let o_plus = fun a b -> new o_set (plus a#to_set b#to_set)
let o_zero = function () -> new o_set Leaf
let o_from_list = function list -> new o_set (from_list list)
let o_from_set = function set -> new o_set set

let o_simple_rebuild_avl = function oset -> 
  new o_set (simple_rebuild_avl oset#to_set)

let o_rebuild_avl = function oset -> 
  new o_set (rebuild_avl oset#to_set)

(* Binary operations *)
(* Oprations binaires *)
(* Operaciones binrias *)
  
let o_union = fun a b -> new o_set (plus a#to_set b#to_set)
let o_intersection = fun a b -> new o_set (intersection a#to_set b#to_set)
let o_difference = fun a b -> new o_set (difference a#to_set b#to_set)
let o_symetric_difference = fun a b -> 
  new o_set (symetric_difference a#to_set b#to_set)
  
let o_subset = fun a b -> subset a#to_set b#to_set
let o_equal = fun a b -> equal a#to_set b#to_set
let o_compare = fun a b -> compare_rec [a#to_set] [b#to_set]


module type OrderedType = sig
  type t
  val compare : t -> t -> int
end

module type Set = 
sig

  type t
  type elt

  val injection : elt -> t
  val map : (elt -> elt) -> t -> t
  val plus : t -> t -> t
  val zero : unit -> t
  val from_list : elt list -> t
  val to_list : t -> elt list
    
  val union : t -> t -> t
  val intersection : t -> t -> t
  val difference   : t -> t -> t
  val symetric_difference : t -> t -> t
  val subset  : t -> t -> bool
  val equal   : t -> t -> bool
  val compare : t -> t -> int

  val insert  : elt -> t -> t
  val insert' : elt -> t -> t
  val insert_raise  : elt -> t -> t (* [Found] *)
  val insert_raise' : elt -> t -> t (* [Found] *)
  val insert_list : elt list -> t -> t

  val delete_min : t -> t
  val delete_max : t -> t
  val delete : elt -> t -> t
  val delete_raise : elt -> t -> t (* [Not_found] *)
  val delete_list : elt list -> t -> t

  val filter : (elt -> bool) -> t -> t
  val filter_lt : elt -> t -> t
  val filter_gt : elt -> t -> t
  val filter_le : elt -> t -> t
  val filter_ge : elt -> t -> t

  val range : elt -> elt -> t -> t

  val min_view : t -> elt * t
  val max_view : t -> t * elt
  val partition : (elt -> bool) -> t -> t * t
  val partition_inf_sup : elt -> t -> t * elt option * t

  val is_empty : t -> bool
  val size : t -> int

  val member  : elt -> t -> bool
  val member' : elt -> t -> bool
  val look_for : (elt -> bool) -> t -> elt

  val choose : t -> elt (* [Empty] *)
  val min_element : t -> elt (* [Not_found] *)
  val max_element : t -> elt (* [Not_found] *)

  val next     : elt -> t -> elt (* [Not_found] *) 
  val previous : elt -> t -> elt (* [Not_found] *) 
  val ceil     : elt -> t -> elt (* [Not_found] *) 
  val floor    : elt -> t -> elt (* [Not_found] *) 

  val rank : elt -> t -> int  (* [Not_found] *) 
  val get  : int -> t -> elt  (* [Out_of_bounds] *)

  val for_all : (elt -> bool) -> t -> bool
  val exists  : (elt -> bool) -> t -> bool

  val iter : (elt -> 'a) -> t -> unit
  val fold_left  : ('a -> elt -> 'a) -> 'a -> t -> 'a
  val fold_right : (elt -> 'a -> 'a) -> t -> 'a -> 'a
  val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a

  val unsafe_inc_map : (elt -> elt) -> t -> t
  val unsafe_dec_map : (elt -> elt) -> t -> t
  val unsafe_set : elt -> int -> t -> t
  val unsafe_insert_min : elt -> t -> t
  val unsafe_insert_max : elt -> t -> t

  val old_plus : t -> t -> t

  val empty : t
  val mem : elt -> t -> bool
  val add : elt -> t -> t
  val singleton : elt -> t
  val remove : elt -> t -> t
  val inter : t -> t -> t
  val diff : t -> t -> t
  val cardinal : t -> int
  val elements : t -> elt list
  val min_elt : t -> elt (* [Not_found] *)
  val max_elt : t -> elt (* [Not_found] *)

  val stat : t -> int * int * int * int

  class type o_t =
  object ('a)

    method map : (elt -> elt) -> 'a
    method plus : 'a -> 'a
    method to_list : elt list
    method to_set : t
      
    method union        : 'a -> 'a
    method intersection : 'a -> 'a
    method difference   : 'a -> 'a
    method symetric_difference : 'a -> 'a
      
    method subset  : 'a -> bool
    method equal   : 'a -> bool
    method compare : 'a -> int
      
    method insert  : elt -> 'a
    method insert' : elt -> 'a
    method insert_raise  : elt -> 'a               (* [Found] *)
    method insert_raise' : elt -> 'a               (* [Found] *)
    method insert_list : elt list -> 'a
      
    method delete : elt -> 'a
    method delete_raise : elt -> 'a                (* [Not_Found] *)
    method delete_list : elt list -> 'a
    method delete_max : 'a
    method delete_min : 'a
      
    method filter : (elt -> bool) -> 'a
    method filter_gt : elt -> 'a
    method filter_lt : elt -> 'a
    method filter_ge : elt -> 'a
    method filter_le : elt -> 'a
      
    method range : elt -> elt -> 'a
      
    method min_view : elt * 'a                     (* [Empty] *)
    method max_view : 'a * elt                     (* [Empty] *)
    method partition : (elt -> bool) -> 'a * 'a
    method partition_inf_sup : elt -> 'a * elt option * 'a
      
    method is_empty : bool
    method size : int
      
    method member  : elt -> bool
    method member' : elt -> bool
    method look_for : (elt -> bool) -> elt         (* [Not_found] *)
      
    method choose : elt                            (* [Empty] *)
    method min_element : elt                       (* [Not_found] *)
    method max_element : elt                       (* [Not_found] *)
      
    method next     : elt -> elt                   (* [Not_found] *)
    method previous : elt -> elt                   (* [Not_found] *)
    method floor    : elt -> elt                   (* [Not_found] *)
    method ceil     : elt -> elt                   (* [Not_found] *)
      
    method get  : int -> elt                       (* [Out_of_bounds] *)
    method rank : elt -> int                       (* [Not_found] *)
      
    method for_all : (elt -> bool) -> bool
    method exists  : (elt -> bool) -> bool
      
    method iter : (elt -> unit) -> unit
    method fold : (elt -> 'b -> 'b) -> 'b -> 'b
    method fold_left  : ('b -> elt -> 'b) -> 'b -> 'b
    method fold_right : (elt -> 'b -> 'b) -> 'b -> 'b

  end

  val o_injection : elt -> o_t
  val o_map : (elt -> elt) -> o_t -> o_t
  val o_plus : o_t -> o_t -> o_t
  val o_zero : unit -> o_t
  val o_from_list : elt list -> o_t
  val o_from_set  : t -> o_t
    
  val o_union : o_t -> o_t -> o_t
  val o_intersection : o_t -> o_t -> o_t
  val o_difference   : o_t -> o_t -> o_t
  val o_symetric_difference : o_t -> o_t -> o_t

  val o_subset  : o_t -> o_t -> bool
  val o_equal   : o_t -> o_t -> bool
  val o_compare : o_t -> o_t -> int

end

(** Debug **)

let rec avl = function
  | Leaf -> true
  | Node (l, _, r) ->
      let hl = height l and hr = height r in
	avl l && avl r && abs (hl - hr) <= 2

let rec mn = function
  | Leaf -> 0
  | Node (a, _, b) -> 1 + min (mn a) (mn b)

let rec mx = function
  | Leaf -> 0
  | Node (a, _, b) -> 1 + max (mx a) (mx b)

let rec med total = function
  | Leaf -> [total]
  | Node (a, _, b) -> med (total + 1) a @ med (total + 1) b

let stat = function tree -> 
  let height_list = med 0 tree in
  let med_h = (List.fold_left (+) 0 height_list) / 
	      (List.length height_list) 
  in
    (size tree, mn tree, mx tree, med_h)

(* To avoid interference with [Pervasives.compare] *)
(* Pour viter toute interfrence avec [Pervasives.compare] *)
(* Para evitar toda interferencia con [Pervasives.compare] *)

let compare = fun a b -> compare_rec [a] [b]


