(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Weight balanced set
 * Ensemble pondr
 * Conjunto ponderado
 *
 *)

open BairePervasives
 
(* Weight balanced trees are binary search trees in which the balance 
 * information is encoded by the height of the tree. The height of one of the 
 * subtrees mustn't be more than wb times the height of the other one.
 * 
 * Weight balanced trees can be seen as a specialisation of cartesian trees 
 * with the given height function
 * 
 * let rec height = function
 *   | E -> 0
 *   | N (a, x, b) -> 1 + height a + height b
 * 
 * We recommend the lecture of Stephen Adams resarch report, specially
 * for the choice of rebalancing constants
 * 
 * Implementing Sets Efficiently in a Functional Language
 * Stephen Adams
 * CSTR 92-01
 * University of Southampton
 *)

(* Les arbres pondrs sont des arbres dont l'information de rquilibrage
 * encode le nombre d'lment dans chaque sous arbre. Le rapport entre la 
 * diffrence de hauteur entre les sous arbres gauche et droit ne peut dpasser
 * une constante wb.
 *
 * Les arbres pondrs peuvent tre obtenus comme une spcialisation des arbres
 * cartsiens avec pour cl (ou hauteur) associe
 * 
 * let rec hauteur = function 
 *   | E -> 0 
 *   | N (a, x, b) -> 1 + hauteur a + hauteur b
 * 
 * Nous recommandons la lecture du rapport de Stephen Adams, notamment
 * concernant le choix de la constante de requilibrage
 * 
 * Implementing Sets Efficiently in a Functional Language
 * Stephen Adams
 * CSTR 92-01
 * University of Southampton
 *)

(* Los rboles ponderados son rboles binarios de busqueda cuya informacin de 
 * equilbrio se ha codificado en la altura del rbol. La diferencia de
 * altura entre los subrboles izquierdo y derecho de cada nudo no debe
 * ser de razn mayor a una constante wb dada.
 * 
 * Los rboles ponderados pueden ser obtenidos especializando los rboles 
 * cartesianos con la funcin de altura siguiente
 * 
 * let rec altura = function
 *   | E -> 0
 *   | N (a, x, b) -> 1 + altura a + altura b
 * 
 * Recomendamos la lectura del informe de Stephen Adams, particularmente
 * en lo que concierne la constante de balance
 * 
 * Implementing Sets Efficiently in a Functional Language
 * Stephen Adams
 * CSTR 92-01
 * University of Southampton
 *)

type 'a set =
  | E
  | N of 'a set * 'a * 'a set * int


(** Utilitaires **)
(** Utils **)
(** tiles **)


let moduleName = "WeightBalancedSet"

let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type

(* Fonction de poids : des foncteurs sont fournis pour choisir 
 * autrement le coefficient ou mme la fonction de poids
 *)

(* Weight function : several functors are given for the user to choose
 * his balance constant or even his own balancing function
 *)

(* Funcin de balance : varios functores permiten escojer otras constantes
 * e incluso su prpria funcin de balance 
 *)

let wt : int -> int = function i -> 3 * i


(* The aim of this  smart constructor  is to hide the tree decorations 
 * and allow manipulation of trees as if they were usual (unbalanced)
 * binary search trees
 *)

(* Ce contructeur est destin  occulter les dcorations de l'arbre afin
 * de pouvoir manipuler les donnes comme des arbres binaires classiques
 *)

(* El propsito de ste constructor es ocultar las decoraciones del rbol
 * afin de poder manipular los datos como si fuesen un rboles binarios comnes
 *)

let makeTree = fun l v r ->
  let n = match (l, r) with
    | (E, E) -> 1
    | (E, N (_, _, _, n)) | (N (_, _, _, n), E) -> n + 1
    | (N (_, _, _, m), N (_, _, _, n)) -> m + n + 1
  in
    N (l, v, r, n)

(* Note on the implementation of rotations :
 * The symetric order is conserved by rotations
 * 
 * examples :
 *  rotate_L : a x (ba y bb) -> (a x ba) y bb
 *  rotate_R : (aa y ab) x b -> aa y (ab x b)
 *  rotate_RL : a x ((baa y bab) z bb) -> (a x baa) y (bab z bb)
 *  rotate_LR : (aa x (aba y abb)) z b -> (aa x aba) y (abb z b) 
 *
 * This property allows a fast code correction verification
 *)

(* Note sur l'implmentation des rotations :
 * La notation infixe est drive de l'invariant des arbres de recherche 
 * donc l'ordre des termes (dit ordre symtrique) se conserve lors des 
 * rotations
 *
 * exemples : 
 *  rotate_L : a x (ba y bb) -> (a x ba) y bb
 *  rotate_R : (aa y ab) x b -> aa y (ab x b)
 *  rotate_RL : a x ((baa y bab) z bb) -> (a x baa) y (bab z bb)
 *  rotate_LR : (aa x (aba y abb)) z b -> (aa x aba) y (abb z b) 
 *
 * Cette proprit permet de vrifier immdiatement la correction
 * des algorithmes
 *)

(* Nota sobre la implementacin de las rotaciones :
 * El rden simtrico es conservado por las rotaciones
 * 
 * ejemplos :
 *  rotate_L : a x (ba y bb) -> (a x ba) y bb
 *  rotate_R : (aa y ab) x b -> aa y (ab x b)
 *  rotate_RL : a x ((baa y bab) z bb) -> (a x baa) y (bab z bb)
 *  rotate_LR : (aa x (aba y abb)) z b -> (aa x aba) y (abb z b) 
 *
 * sta propriedad permite verificar de manera imediata la correccin
 * de los algortmos 
 *)

let rotate_L = fun l v r ->
  match r with
    | N (rl, rv, rr, _) -> makeTree (makeTree l v rl) rv rr
    | _ -> makeTree l v r
	
let rotate_R = fun l v r ->
  match l with
    | N (ll, lv, lr, _) -> makeTree ll lv (makeTree lr v r)
    | _ -> makeTree l v r

let rotate_RL = fun l v r ->
  match r with
    | N (N (rll, rlv, rlr, _), rv, rr, _) -> 
	makeTree (makeTree l v rll) rlv (makeTree rlr rv rr)
    | _ -> makeTree l v r

let rotate_LR = fun l v r ->
  match l with
    | N (ll, lv, N (lrl, lrv, lrr, _), _) ->
	makeTree (makeTree ll lv lrl) lrv (makeTree lrr v r)
    | _ -> makeTree l v r

let size = function
  | E -> 0
  | N (_, _, _, n) -> n

(* [makeBTree] suposes that :
 * - the symetric order is respected l < v < r
 * - the two subtrees are well formed
 * - the height mean between l and r is not greater than wb
 *)

(* la fonction [makeBTree] suppose :
 * - le respect de l'ordre symtrique l < v < r
 * - les sous-arbres l et r tre bien forms
 * - la rapport de hauteur entre l et r ne pas excder wb
 *)

(* la funcin [makeBTree] supone :
 * - el respeto del orden simtrico l < v < r
 * - los subrboles l y r estar bien formados
 * - la razn de altura entre l y r no valer ms de wb
 *)

let makeBTree = fun l v r ->
  let ln = size l and rn = size r in
    if ln > wt rn then
      (match l with
	 | E -> makeTree l v r
	 | N (ll, _, lr, _) ->
	     if size ll > size lr then rotate_R l v r
	     else rotate_LR l v r
      )
    else if rn > wt ln then
      (match r with
	 | E -> makeTree l v r
	 | N (rl, _, rr, _) ->
	     if size rr > size rl then rotate_L l v r
	     else rotate_RL l v r
      )
    else makeTree l v r

(* makeSTree function suposes that :
 * - the symetric order is respected l < v < r
 * - the two subtrees are well formed
 *)

(* la fonction makeSTree suppose :
 * - le respect de l'ordre symtrique l < v < r
 * - les sous-arbres l et r tre forms
 *)

(* la funcin makeSTree supone :
 * - el respeto del orden simtrico l < v < r
 * - los subrboles l y r estar bien formados
 *)

let rec makeSTree = fun l v r ->
  let hl = size l and hr = size r in
    if hl > wt hr then
      (match l with
	 | E -> makeTree l v r
	 | N (ll, lv, lr, _) -> makeBTree ll lv (makeSTree lr v r)
      )
    else if hr > wt hl then
      (match r with
	 | E -> makeTree l v r
	 | N (rl, rv, rr, _) -> makeBTree (makeSTree l v rl) rv rr
      )
    else makeBTree l v r

(* [append] suposes that :
 * - the symetric order is respected l < r
 * - the two subtrees are well formed
 * - the height mean between l and r is not greater than wb
 *)

(* la fonction [append] suppose :
 * - le respect de l'ordre symtrique l < r
 * - les sous-arbres l et r tre bien forms
 * - le rapport de hauteur entre l et r ne pas exder wb
 *)

(* la funcin [append] supone :
 * - el respeto del orden simtrico l < r
 * - los subrboles l y r estar bien formados
 * - la razn de altura entre l y r no valer ms de wb
 *)

let rec append = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l1, v1, r1, _), N (l2, v2, r2, _)) ->
        makeBTree l1 v1 (makeBTree (append r1 l2) v2 r2)

(* [appendS] suposes that :
 * - the symetric order is respected l < r
 * - the two subtrees are well formed
 *)

(* la fonction [appendS] suppose :
 * - le respect de l'ordre symtrique l < r
 * - les sous-arbres l et r tre bien forms
 *)

(* la funcin [appendS] supone :
 * - el respeto del orden simtrico l < r
 * - los subrboles l y r estar bien formados
 *)

let rec appendS = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l1, v1, r1, _), N (l2, v2, r2, _)) ->
        makeSTree l1 v1 (makeSTree (appendS r1 l2) v2 r2)


(** Oprations catgorielles **)
(** Categorial operations **)
(** Operaciones categoriales **)


let injection = function x -> makeTree E x E

let rec insert_raise x = function
  | E -> injection x
  | N (l, v, r, _) ->
      match compare x v with
	| n when n < 0 -> makeBTree (insert_raise x l) v r
	| n when n > 0 -> makeBTree l v (insert_raise x r)
	| _ -> raise Found

let insert x = function set ->
  try 
    insert_raise x set 
  with Found -> set

let rec from_list_rec result = function
  | [] -> result
  | head :: tail -> from_list_rec (insert head result) tail

let from_list = function list -> from_list_rec E list

let rec to_list_rec result = function
  | E -> result
  | N (l, v, r, _) -> to_list_rec (v :: to_list_rec result r) l

let to_list = function set -> to_list_rec [] set

let map f = function set -> from_list (List.map f (to_list set))

(* Union algorithm :
 * Explained in the research report (avaible on the web)
 * 
 * Implementing Sets Efficiently in a Functional Language
 * Stephen Adams
 * CSTR 92-01
 * University of Southampton
 * 
 * We first give a version obtained by  divide and conquer 
 *)

(* Algorithme d'union :
 * Son fonctionnement est expliqu dans le rapport de recherche 
 * (disponible sur la toile)
 *
 * Implementing Sets Efficiently in a Functional Language
 * Stephen Adams
 * CSTR 92-01
 * University of Southampton
 *
 * Nous en rappelons ici le principe :
 * On construit une premire version via la mthode  diviser pour rgner 
 *)

(* Algoritmo de unin :
 * Su funcionamiento est explicado en el informe (disponible en la red)
 * 
 * Implementing Sets Efficiently in a Functional Language
 * Stephen Adams
 * CSTR 92-01
 * University of Southampton
 *
 * Recordamos rapidamente su principio :
 * Uno comienza construyendo una primera versin con el mtodo  dividir
 * para reinar 
 *)

let rec filter_lt x = function
  | E -> E
  | N (l, v, r, _) ->
      match compare x v with
	| n when n < 0 -> filter_lt x l
	| n when n > 0 -> makeSTree l v (filter_lt x r)
	| _ -> l

let rec filter_gt x = function
  | E -> E
  | N (l, v, r, _) ->
      match compare x v with
	| n when n < 0 -> makeSTree (filter_gt x l) v r
	| n when n > 0 -> filter_gt x r
	| _ -> r

(* We could have used a unique partition function like [partition_inf_sup]
 * but the to filter functions will be useful later in [plus]
 *)

(* Nous aurions pu utiliser une unique fonction de partition telle
 * [partition_inf_sup] mais les filtres dfinis nous seront utiles sparment 
 * dans la version optimise de [plus]
 *)
	
(* Ubisemos podido utilizar una nica funcin de particin [partition_inf_sup]
 * pero los dos filtros que hemos definido nos sern tiles en la versin
 * optimizada de la funcin [plus]
 *)
    
let rec old_plus = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (_, N (l, v, r, _)) -> 
	let 
	  lt_v = filter_lt v a and 
	  gt_v = filter_gt v a 
	in
	  makeSTree (old_plus lt_v l) v (old_plus gt_v r)
	    

(* One can notice that there is no need to construct a tree (lt_v and gt_v)
 * which will be destructed just after by the recursive call to [old_plus]. 
 * That is why we introduce a triple containing a tree and two bounds. 
 * These bounds can be easely moved then.
 *)

(* On constate qu' rien ne sert de construire un arbre (lt_v et gt_v) qui sera
 * dcoup juste aprs par l'appel rcursif de la fonction [old_plus], raison 
 * pour laquelle on introduit un triplet contenant un arbre et deux bornes 
 * qui pourront tre modifies  moindre cot
 *)

(* Uno se da cuenta que de nada sirve construir un rbol (lt_v y gt_v) que ser
 * destruido justo despus al llamar recursivamente la funcin [old_plus].
 * Se introduce por esa razn un triplete formado por un rbol y dos 
 * limites, lmites que se podrn modificar lteriormente a poco costo 
 *)

(* [cut] constructs a triple, making sure that the root of the tree lies 
 * between the two bounds. If it is not the case, the functions calls itself
 * recusively on the corresponding subtree (right or left)
 *)

(* [cut] construit un triplet de faon  ce que la racine se trouve entre
 * les deux bornes. Si ce n'est pas le cas, la fonction restreint rcursivement
 * l'arbre  son sous-arbre gauche ou droit selon le cas de figure.
 *) 

(* [cut] constuye un triplete de manera tal a que la raz del rbol se 
 * encuentre entre los dos lmites. Si este no es el caso, la funcin
 * toma recursivamente el subrbol derecho o izquierdo segn convenga
 *)

let rec cut = fun lo tree hi ->
  match tree with
    | N (_, v, r, _) when compare v lo <= 0 -> cut lo r hi
    | N (l, v, _, _) when compare v hi >= 0 -> cut lo l hi
    | _ -> (lo, tree, hi)

(* [cut_lo] and [cut_hi] are specialized versions with an infinite 
 * hight (respectively low) bound 
 *)

(* [cut_lo] et [cut_hi] sont des versions spcialises avec la borne infrieure
 * (respectivement suprieure) ayant pour valeur plus l'infini (respectivement 
 * moins l'infini)
 *)

(* [cut_lo] y [cut_hi] son versiones especializadas con un limite inferior
 * (respectivamente superior) igual a ms (respectivamente menos) el infinito
 *)

let rec cut_hi = fun lo tree ->
  match tree with
    | N (_, v, r, _) when compare v lo <= 0 -> cut_hi lo r
    | _ -> (lo, tree)

let rec cut_lo = fun tree hi ->
  match tree with
    | N (l, v, _, _) when compare v hi >= 0 -> cut_lo l hi
    | _ -> (tree, hi)

(* [uni] uses the left tree to control the reconstruction :
 * - the right tree will be physically cut only if it lies between to 
 * successive values of the control tree (which guarantees that there will not
 * be any new cut after the reconstruction)
 * - if the right tree is empty, the whole control tree can be inserted
 *)

(* la fonction [uni] se sert de l'arbre de gauche pour contrler la
 * reconstruction : 
 * - l'arbre de droite ne sera dcoup physiquement que si l'on se trouve 
 * entre deux valeurs successives de l'arbre de contrle (ce qui garantit 
 * qu'il n'y aura pas ne nouvelle dcoupe aprs la reconstruction)
 * - si l'arbre de droite est vide, on peut insrer tout l'arbre de contrle
 *)

(* la funcin [uni] se sirve del rbol izquierdo para controlar la 
 * reconstruccin :
 * - el rbol derecho solo sera recortado fsicamente si se encuentra
 * entre dos valores succesivos del rbol de control (lo que garantiza
 * que no habr ninguna otra operacin de corte despues de la reconstruccin)
 * - si el rbol derecho est vaco, se puede insertar el rbol de control
 * entero
 *)

let rec uni = fun a (lo, b, hi) ->
  match (a, b) with
    | (_, E) -> a
    | (E, N (l, v, r, _)) -> makeSTree (filter_gt lo l) v (filter_lt hi r)
    | (N (l, v, r, _), N _) ->
	makeSTree (uni l (cut lo b v)) v (uni r (cut v b hi))  

(* two specialized versions are given *)

(* on en donne galement deux versions spcialises *)

(* se dan tambin dos versiones especializadas *)

let rec uni_lo = fun a (b, hi) ->
  match (a, b) with
    | (_, E) -> a
    | (E, N (l, v, r, _)) -> makeSTree l v (filter_lt hi r)
    | (N (l, v, r, _), N _) -> 
	makeSTree (uni_lo l (cut_lo b v)) v (uni r (cut v b hi))

let rec uni_hi = fun a (lo, b) ->
  match (a, b) with
    | (_, E) -> a
    | (E, N (l, v, r, _)) -> makeSTree (filter_gt lo l) v r
    | (N (l, v, r, _), N _) -> 
	makeSTree (uni l (cut lo b v)) v (uni_hi r (cut_hi v b)) 

(* We just put all together *)
	
(* On assemble le tout *)

(* Se juntan todos los pedazos *)

let plus = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l, v, r, _), N _) ->
	makeSTree (uni_lo l (cut_lo b v)) v (uni_hi r (cut_hi v b)) 

let zero = function () -> E

let rec join_rec result = function
  | E -> result
  | N (l, v, r, _) -> join_rec (join_rec (plus v result) l) r 

let join = function tree -> join_rec E tree

let rec to_tree = function
  | E -> Leaf
  | N (l, v, r, _) -> Node (to_tree l, v, to_tree r)

let rec from_tree = function
  | Leaf -> E
  | Node (l, v, r) -> makeSTree (from_tree l) v (from_tree r)


(** Binary operaations **)
(** Oprations binaires **)
(** Opraciones binrias **)


let rec partition_inf_sup pivot = function
  | E -> (E, None, E)
  | N (l, v, r, _) as node ->
      match compare pivot v with
	| n when n < 0 ->
	    let (inf, value, sup) = partition_inf_sup pivot l in
	      (inf, value, makeSTree sup v r)
	| n when n > 0 ->
	    let (inf, value, sup) = partition_inf_sup pivot r in
	      (makeSTree l v inf, value, sup)
	| _ -> (l, Some v, r)

let union = plus

let rec intersection = fun a b ->
  match (a, b) with
    | (E, _) | (_, E) -> E
    | (N (l, v, r, _), N _) ->
	match partition_inf_sup v b with
	  | (l', None, r') -> appendS (intersection l l') (intersection r r')
	  | (l', Some _, r') -> 
	      makeSTree (intersection l l') v (intersection r r')

let rec difference = fun a b ->
  match (a, b) with
    | (E, _) -> E
    | (_, E) -> a
    | (N (l, v, r, _), N _) -> 
	match partition_inf_sup v b with
	  | (l', None, r') -> makeSTree (difference l l') v (difference r r')
	  | (l', Some _, r') -> appendS (difference l l') (difference r r')
	      
let rec symetric_difference = fun a b ->
  match (a, b) with
    | (E, s) | (s, E) -> s
    | (N (l, v, r, _), N _) ->
	match partition_inf_sup v b with
	  | (l', None, r') -> makeSTree (symetric_difference l l') v 
	      (symetric_difference r r')
	  | (l', Some _, r') -> 
	      appendS (symetric_difference l l') (symetric_difference r r')

let rec subset = fun a b ->
  match (a, b) with
    | (E, _) -> true
    | (_, E) -> false
    | (N (l1, v1, r1, _), N (l2, v2, r2, _)) ->
        match compare v1 v2 with
	  | n when n < 0 -> subset (makeBTree l1 v1 E) l2 && subset r1 b
	  | n when n > 0 -> subset (makeBTree E v1 r1) r2 && subset l1 b
	  | _ -> subset l1 l2 && subset r1 r2

let rec compare_rec = fun a b ->
  match (a, b) with
    | ([], []) -> 0
    | ([], _) -> -1
    | (_, []) -> 1
    | (E :: t1, E :: t2) -> compare_rec t1 t2
    | (N (E, v1, r1, _) :: t1, N (E, v2, r2, _) :: t2) ->
	(match compare v1 v2 with
	  | 0 -> compare_rec (r1 :: t1) (r2 :: t2)
	  | other -> other
	)
    | (N (a, v1, r1, _) :: t1, t2) ->
        compare_rec (a :: N (E, v1, r1, 0) :: t1) t2
    | (t1, N (b, v2, r2, _) :: t2) ->
	compare_rec t1 (b :: N (E, v2, r2, 0) :: t2)

let equal = fun a b -> compare_rec [a] [b] = 0

(* let compare = fun a b -> compare_rec [a] [b]
 *
 * renvoy  la fin du fichier pour ne pas interfrer avec Pervasives.compare 
 * 
 * see the end of the file
 * 
 * ver el final del fichero
 *)


(** Acces **)
(** Accs **)
(** Acceso **)


(* [insert'] :
 * Variant of the [insert] function, it does only one comparasion
 * by branch instead of 3/2 but has to walk down the whole tree
 * 
 * If the inserted element is not already in the tree, the [insert] function
 * has anyway to go down the tree until the leafs, so in this particular
 * point the lost is not very important. The reason why this function
 * is not always faster than [insert] is that one more argument
 * must be transmitted in each recursive call
 *)

(* [insert'] :
 * Variante de la fonction [insert] ne faisant qu'une comparaison par
 * branche (au lieu de 3/2 en moyenne pour la fonction habituelle) mais
 * parcourt l'arbre jusqu' profondeur maximale
 *
 * Hormis le cas o l'lment insr se trouve dj dans l'arbre, la fonction
 * [insert] effectue un parcours sur toute la profondeur. La perte est donc
 * minime sur ce point, mais un argument supplmentaire doit tre transmis, 
 * raison pour laquelle cette fonction n'est pas toujours plus rapide que 
 * la version la plus simple
 *)

(* [insert'] :
 * Una variante de la funcin [insert] que hace una sola comparacin por
 * nivel (en vez de 3/2 en promedio para la version comn) pero recorre la 
 * totalidad del rbol, hasta las hojas.
 * 
 * Excepto en el caso en el cual el lemento que se est insertando se 
 * encuentra ya en el rbol, la funcin [insert] recorre de todas maneras
 * el rbol hasta las hojas, por lo cual la perdida en ese aspecto es
 * mnima. No obstante, un argumento suplementrio debe ser transmitido y
 * por ello esta funcin no es todo el tiempo ms rpida que la version
 * la mas simple
 *)

let rec insert_min_bound b x = function
  | E -> 
      if compare b x = 0 then raise Found
      else injection x
  | N (l, v, r, _) ->
      match compare x v with
	| n when n < 0 -> makeBTree (insert_min_bound b x l) v r
	| _ -> makeBTree l v (insert_min_bound v x r)

let rec insert_raise' x = function
  | E -> injection x
  | N (l, v, r, _) ->
      match compare x v with
	| n when n < 0 -> makeBTree (insert_raise' x l) v r
	| _ -> makeBTree l v (insert_min_bound x v r)

let insert' x = function set ->
  try 
    insert_raise' x set 
  with Found -> set

let insert_list = fun list set -> plus (from_list list) set

let rec delete_min = function
  | E -> E
  | N (E, _, r, _) -> r
  | N (l, v, r, _) -> makeBTree (delete_min l) v r

let rec delete_max = function
  | E -> E
  | N (l, _, E, _) -> l
  | N (l, v, r, _) -> makeBTree l v (delete_max r)

let rec delete_raise x = function
  | E -> raise Not_found
  | N (l, v, r, _) ->
      match compare x v with
	| n when n < 0 -> makeBTree (delete_raise x l) v r
	| n when n > 0 -> makeBTree l v (delete_raise x r)
	| _ -> append l r

let delete x = function set -> 
  try 
    delete_raise x set 
  with Not_found -> set

 let rec delete_list = fun list set ->
  match list with
    | [] -> set
    | head :: tail -> delete_list tail (delete head set)

let rec filter_rec f result = function
  | E -> result
  | N (l, v, r, _) ->
      filter_rec f (filter_rec f (if f v then insert v result else result) l) r

let filter f = function set -> filter_rec f E set

let rec filter_le x = function
  | E -> E
  | N (l, v, r, _) ->
      match compare x v with
	| n when n < 0 -> filter_le x l
	| _ -> makeSTree l v (filter_le x r)

let rec filter_ge x = function
  | E -> E
  | N (l, v, r, _) ->
      match compare x v with
	| n when n <= 0 -> makeSTree (filter_ge x l) v r
	| _ -> filter_ge x r

let rec range min max = function
  | E -> E
  | N (l, v, r, _) ->
      if compare v min < 0 then range min max r
      else if compare v max > 0 then range min max l
      else makeSTree (filter_ge min l) v (filter_le max r)


(** Views **)
(** Vues **)
(** Vistas **)


let rec min_view = function
  | E -> raise Empty
  | N (E, v, r, _) -> (v, r)
  | N (l, v, r, _) -> let (min, left) = min_view l in (min, makeBTree left v r)

let rec max_view = function
  | E -> raise Empty
  | N (l, v, E, _) -> (l, v)
  | N (l, v, r, _) -> let (right, max) = max_view r in 
      (makeBTree l v right, max)

let rec partition_rec f (trueS, falseS) = function
  | E -> (trueS, falseS)
  | N (l, v, r, _) -> 
      if f v 
      then partition_rec f (partition_rec f (insert v trueS, falseS) l) r
      else partition_rec f (partition_rec f (trueS, insert v falseS) l) r

let partition = fun f set -> partition_rec f (E, E) set


(** Observers **)
(** Observateurs **)
(** Observadores **)


let is_empty = function
  | E -> true
  | N _ -> false

(* [member'] : does (d + 1) comparations instead of 3/2 * d
 * following an idea of Arne Andersson, but walks down the whole height of 
 * the tree and an extra argument has to be transmitted in each call
 * 
 * It may not always be faster than [member]
 *)

(* [member'] : effectue en moyenne (d + 1) comparaisons au lieu de 3/2 * d
 * selon une ide d'Arne Andersson mais parcourt systmatiquement l'arbre sur
 * toute sa hauteur et doit transmettre un paramtre supplmentaire  chaque
 * appel
 * 
 * Il est possible qu'elle ne soit pas toujours plus rapide que [member]
 *) 

(* [member'] : hace (d + 1) comparaciones en vez de 3/2 * d en promedio, 
 * segn una idea de Arne Andersson, pero recorre toda la altura del rbol
 * y requiere un parmetro suplementrio en cada llamado
 * 
 * Es posible que no sea siempre ms veloz que [member]
 *)

let rec member_min_bound b x = function
  | E -> compare b x = 0
  | N (l, v, r, _) ->
      match compare x v with
	| k when k < 0 -> member_min_bound b x l
	| _ -> member_min_bound v x r
	    
let rec member' x = function
  | E -> false
  | N (l, v, r, _) ->
      match compare x v with
	| n when n < 0 -> member' x l
	| _ -> member_min_bound v x r
	    
(* [member] : usual member function
 * some algorithms use the properties of the usual [member] function 
 * (stop of a positive search in the middle of the tree). It is the case
 * of optimal tree construction for example (Knuthm Garcia-Walsh, confer
 * The art of computer programming vol III)
 *)

(* [member] : fonction member habituelle 
 * un certain nombre d'algorithmes utilisent les proprits de la fonction
 * [member] habituelle (arrt d'une recherche positive  l'intrieur de 
 * l'arbre) par exemple les algorithmes de construction d'arbres optimaux
 * (Knuth, Garcia-Walsh, confer The art of computer programming vol III)
 *)

(* [member] : funcin member comn
 * ciertos algortmos usan las propriedades de la funcin [member] comn
 * (interupcin de una busqueda positiva en el medio del rbol) tal
 * los algortmos de construccin de rboles optimales (Knuth, Garcia-Walsh,
 * confer The art of computer programming vol III)
 *)

let rec member x = function
  | E -> false
  | N (l, v, r, _) ->
      match compare x v with
	| n when n < 0 -> member x l
	| n when n > 0 -> member x r
	| _ -> true

let rec look_for f = function
  | E -> raise Not_found
  | N (_, v, _, _) when f v -> v
  | N (l, _, r, _) -> (try look_for f l with Not_found -> look_for f r)

let choose = function
  | E -> raise Empty
  | N (_, v, _, _) -> v

let rec min_element = function
  | E -> raise Not_found
  | N (E, v, _, _) -> v
  | N (l, _, _, _) -> min_element l

let rec max_element = function
  | E -> raise Not_found
  | N (_, v, E, _) -> v
  | N (_, _, r, _) -> max_element r

(* Could also be done with ['a option] *)

(* Auraient aussi pu tre codes au moyen de ['a option], nous avons
 * prfr illustrer les interruptions de calcul au moyens d'exceptions
 *)

(* stas operaciones tambin se hubiesen podido escribir con el typo 
 * ['a option] pero hemos querido aqu mostrar el uso de las excepciones para 
 * ruptura de clculo
 *)

let rec next x = function
  | E -> raise Not_found
  | N (l, v, r, _) ->
      match compare x v with
	| n when n < 0 -> (try next x l with Not_found -> v)
	| n when n > 0 -> next x r
	| _ -> min_element r (* raises [Not_found] *)

let rec previous x = function
  | E -> raise Not_found
  | N (l, v, r, _) ->
      match compare x v with
	| n when n < 0 -> previous x l
	| n when n > 0 -> (try previous x r with Not_found -> v)
	| _ -> max_element l (* raises [Not_found] *)
	      
let rec ceil x = function
  | E -> raise Not_found
  | N (l, v, r, _) ->
      match compare x v with
	| n when n < 0 -> (try ceil x l with Not_found -> v)
	| n when n > 0 -> ceil x r
	| _ -> x

let rec floor x = function
  | E -> raise Not_found
  | N (l, v, r, _) ->
      match compare x v with
	| n when n < 0 -> floor x l
	| n when n > 0 -> (try floor x r with Not_found -> v)
	| _ -> x

(* les arbres binaires de recherche sont indexs suivant l'ordre symtrique *)

(* binary search trees are indexed following the symetric order *)

(* los rboles de busqueda estn indizados segn el orden simtrico *)

let rec rank x = function
  | E -> raise Not_found
  | N (l, v, r, _) ->
      match compare x v with
	| n when n < 0 -> rank x l
	| n when n > 0 -> 1 + size l + rank x r
	| _ -> size l

let rec get n = function
  | E -> raise (Out_of_bounds n)
  | N (l, v, r, _) ->
      let k = size l in
	if n < k then get n l
	else if n > k then get (n - k - 1) r
	else v

let rec set_rec value m = function
  | E -> raise (Out_of_bounds m)
  | N (l, v, r, h) ->
      let k = size l in
	if m < k then N (set_rec value m l, v, r, h)
	else if m > k then N (l, v, set_rec value (m - k - 1) r, h)
	else if v = value then raise Found
	else N (l, value, r, h)
	  
let unsafe_set = fun value n tree -> 
  try 
    set_rec value n tree 
  with Found -> tree

let rec for_all f = function
  | E -> true
  | N (l, v, r, _) -> f v && for_all f l && for_all f r

let rec exists f = function
  | E -> false
  | N (l, v, r, _) -> f v || exists f l || exists f r


(** Traversals **)
(** Parcours **)
(** Recorridos **)


let rec iter f = function
  | E -> ()
  | N (l, v, r, _) -> iter f l ; f v ; iter f r

let rec fold_left f e = function
  | E -> e
  | N (l, v, r, _) -> fold_left f (f (fold_left f e l) v) r

let rec fold_right = fun f set e ->
  match set with
    | E -> e 
    | N (l, v, r, _) -> fold_right f l (f v (fold_right f r e))

let fold = fold_right


(** Miscellaneous **)
(** Divers **)
(** Miscelanios **)


let bind = fun set f -> join (map f set)

let concat_map = fun f set -> join (map f set)

let rec unsafe_inc_map f = function
  | E -> E
  | N (l, v, r, _) -> makeTree (unsafe_inc_map f l) (f v) (unsafe_inc_map f r)

let rec unsafe_dec_map f = function
  | E -> E
  | N (l, v, r, _) -> makeTree (unsafe_dec_map f l) (f v) (unsafe_dec_map f r)

let rec unsafe_insert_min x = function
  | E -> injection x
  | N (l, v, r, _) -> makeTree (unsafe_insert_min x l) v r 

let rec unsafe_insert_max x = function
  | E -> injection x
  | N (l, v, r, _) -> makeTree l v (unsafe_insert_max x r)


(** Compatibilit avec la librairie standard de Caml **)
(** Caml standard library compatibility **)
(** Compatibilidad con la librera estandar de Caml **)


let empty = E
let mem = member
let add = insert
let singleton = injection
let remove = delete
let inter = intersection
let diff = difference
let cardinal = size
let elements = to_list
let min_elt = min_element    
let max_elt = max_element


(** Objets **)
(** Ojects **)
(** Objetos **)


class ['a] o_set = fun set ->

object (self : 'b)
  
  val contents  = set

  method to_set = contents

  (* Categorial operations *)
  (* Oprations categorielles *)
  (* Operaciones categoriales *)

  method map f = {< contents = map f contents >}

  method plus = function (set : 'b) -> 
    {< contents = plus contents set#to_set >} 
  method to_list = to_list contents
		     
  (* Binary operations *)
  (* Oprations binaires *)
  (* Operaciones binrias *)
		     
  method union = function (set : 'b) -> 
    {< contents = plus contents set#to_set >}
  
  method intersection = function (set : 'b) -> 
    {< contents = intersection contents set#to_set >}
  
  method difference = function (set : 'b) -> 
    {< contents = difference contents set#to_set >}
  
  method symetric_difference = function (set : 'b) -> 
    {< contents = symetric_difference contents set#to_set >}
    
  method subset = function (set : 'b) -> subset contents set#to_set
  method equal = function (set : 'b)-> equal contents set#to_set
  method compare = function (set : 'b) -> compare_rec [contents] [set#to_set]
    
  (* Acces *)
  (* Accs *)
  (* Acceso *)

  method insert = function x -> 
    try 
      {< contents = insert_raise x contents >} 
    with Found -> self

  method insert_raise = function x -> {< contents = insert_raise x contents >}

  method insert' = function x -> 
    try
      {< contents = insert_raise' x contents >}
    with Found -> self

  method insert_raise' = function x -> 
    {< contents = insert_raise' x contents >}
      
  method insert_list = function list -> 
    {< contents = insert_list list contents >}

  method delete = function x -> 
    try 
      {< contents = delete_raise x contents >} 
    with Not_found -> self

  method delete_raise = function x -> {< contents = delete_raise x contents >}

  method delete_list = function list -> 
    {< contents = delete_list list contents >}
      
  method delete_min = {< contents = delete_min contents >}
  method delete_max = {< contents = delete_max contents >}

  method filter = function f -> {< contents = filter f contents >}
  method filter_lt = function x -> {< contents = filter_lt x contents >}
  method filter_gt = function x -> {< contents = filter_gt x contents >}
  method filter_le = function x -> {< contents = filter_le x contents >}
  method filter_ge = function x -> {< contents = filter_ge x contents >}

  method range = fun min max -> {< contents = range min max contents >}

  (* Views *)
  (* Vues *)
  (* Vistas *)

  method min_view =
    let (min, other) = min_view contents in
      (min, {< contents = other >})

  method max_view =
    let (other, max) = max_view contents in
      ({< contents = other >}, max)

  method partition = function f -> 
    let (trueSet, falseSet) = partition f contents in
      ({< contents = trueSet >}, {< contents = falseSet >})

  method partition_inf_sup = function x ->
    let (inf, value, sup) = partition_inf_sup x contents in
      ({< contents = inf >}, value, {< contents = sup >})

  (* Observators *)
  (* Observateurs *)
  (* Observadores *)
      
  method is_empty = contents = E
  method size = 
    match contents with
      | E -> 0
      | N (_, _, _, n) -> n

  method member = function x -> member x contents
  method member' = function x -> member' x contents
  method look_for = function f -> look_for f contents

  method choose =
    match contents with
      | E -> raise Not_found
      | N (_, v, _, _) -> v

  method min_element = min_element contents
  method max_element = max_element contents

  method previous = function x -> previous x contents
  method next = function x -> next x contents
  method floor = function x -> floor x contents
  method ceil = function x -> ceil x contents

  method rank = function x -> rank x contents
  method get = function n -> get n contents

  method exists = function f -> exists f contents
  method for_all = function f -> for_all f contents

  (* Traversals *)
  (* Parcours *)
  (* Recorridos *)

  method iter : ('a -> unit) -> unit = function f -> iter f contents

  method fold : 'c. ('a -> 'c -> 'c) -> 'c -> 'c = 
    fun f e -> fold_right f contents e

  method fold_left : 'c. ('c -> 'a -> 'c) -> 'c -> 'c =
    fun f e -> fold_left f e contents

  method fold_right : 'c. ('a -> 'c -> 'c) -> 'c -> 'c =
    fun f e -> fold_right f contents e

end

(* Constructors *)
(* Constructeurs *)
(* Constructores *)

let o_injection = function x -> new o_set (injection x)
let o_plus = fun a b -> new o_set (plus a#to_set b#to_set)
let o_zero = function () -> new o_set E
let o_from_list = function list -> new o_set (from_list list)
let o_from_set = function set -> new o_set set

(* Binary operations *)
(* Oprations binaires *)
(* Operaciones binrias *)
  
let o_union = fun a b -> new o_set (plus a#to_set b#to_set)
let o_intersection = fun a b -> new o_set (intersection a#to_set b#to_set)
let o_difference = fun a b -> new o_set (difference a#to_set b#to_set)
let o_symetric_difference = fun a b -> 
  new o_set (symetric_difference a#to_set b#to_set)
  
let o_subset = fun a b -> subset a#to_set b#to_set
let o_equal = fun a b -> equal a#to_set b#to_set
let o_compare = fun a b -> compare_rec [a#to_set] [b#to_set]


module type OrderedType = sig
  type t
  val compare : t -> t -> int
end

module type Set = 
sig

  type t
  type elt

  val injection : elt -> t
  val map : (elt -> elt) -> t -> t
  val plus : t -> t -> t
  val zero : unit -> t
  val from_list : elt list -> t
  val to_list : t -> elt list
    
  val union : t -> t -> t
  val intersection : t -> t -> t
  val difference   : t -> t -> t
  val symetric_difference : t -> t -> t
  val subset  : t -> t -> bool
  val equal   : t -> t -> bool
  val compare : t -> t -> int

  val insert  : elt -> t -> t
  val insert' : elt -> t -> t
  val insert_raise  : elt -> t -> t (* [Found] *)
  val insert_raise' : elt -> t -> t (* [Found] *)
  val insert_list : elt list -> t -> t

  val delete_min : t -> t
  val delete_max : t -> t
  val delete : elt -> t -> t
  val delete_raise : elt -> t -> t (* [Not_found] *)
  val delete_list : elt list -> t -> t

  val filter : (elt -> bool) -> t -> t
  val filter_lt : elt -> t -> t
  val filter_gt : elt -> t -> t
  val filter_le : elt -> t -> t
  val filter_ge : elt -> t -> t

  val range : elt -> elt -> t -> t

  val min_view : t -> elt * t
  val max_view : t -> t * elt
  val partition : (elt -> bool) -> t -> t * t
  val partition_inf_sup : elt -> t -> t * elt option * t

  val is_empty : t -> bool
  val size : t -> int

  val member  : elt -> t -> bool
  val member' : elt -> t -> bool
  val look_for : (elt -> bool) -> t -> elt

  val choose : t -> elt (* [Empty] *)
  val min_element : t -> elt (* [Not_found] *)
  val max_element : t -> elt (* [Not_found] *)

  val next     : elt -> t -> elt (* [Not_found] *) 
  val previous : elt -> t -> elt (* [Not_found] *) 
  val ceil     : elt -> t -> elt (* [Not_found] *) 
  val floor    : elt -> t -> elt (* [Not_found] *) 

  val rank : elt -> t -> int  (* [Not_found] *) 
  val get  : int -> t -> elt  (* [Out_of_bounds] *)

  val for_all : (elt -> bool) -> t -> bool
  val exists  : (elt -> bool) -> t -> bool

  val iter : (elt -> 'a) -> t -> unit
  val fold_left  : ('a -> elt -> 'a) -> 'a -> t -> 'a
  val fold_right : (elt -> 'a -> 'a) -> t -> 'a -> 'a
  val fold : (elt -> 'a -> 'a) -> t -> 'a -> 'a

  val unsafe_inc_map : (elt -> elt) -> t -> t
  val unsafe_dec_map : (elt -> elt) -> t -> t
  val unsafe_set : elt -> int -> t -> t
  val unsafe_insert_min : elt -> t -> t
  val unsafe_insert_max : elt -> t -> t

  val old_plus : t -> t -> t

  val empty : t
  val mem : elt -> t -> bool
  val add : elt -> t -> t
  val singleton : elt -> t
  val remove : elt -> t -> t
  val inter : t -> t -> t
  val diff : t -> t -> t
  val cardinal : t -> int
  val elements : t -> elt list
  val min_elt : t -> elt (* [Not_found] *)
  val max_elt : t -> elt (* [Not_found] *)

  val stat : t -> int * int * int * int

  class type o_t =
  object ('a)

    method map : (elt -> elt) -> 'a
    method plus : 'a -> 'a
    method to_list : elt list
    method to_set : t
      
    method union        : 'a -> 'a
    method intersection : 'a -> 'a
    method difference   : 'a -> 'a
    method symetric_difference : 'a -> 'a
      
    method subset  : 'a -> bool
    method equal   : 'a -> bool
    method compare : 'a -> int
      
    method insert  : elt -> 'a
    method insert' : elt -> 'a
    method insert_raise  : elt -> 'a               (* [Found] *)
    method insert_raise' : elt -> 'a               (* [Found] *)
    method insert_list : elt list -> 'a
      
    method delete : elt -> 'a
    method delete_raise : elt -> 'a                (* [Not_Found] *)
    method delete_list : elt list -> 'a
    method delete_max : 'a
    method delete_min : 'a
      
    method filter : (elt -> bool) -> 'a
    method filter_gt : elt -> 'a
    method filter_lt : elt -> 'a
    method filter_ge : elt -> 'a
    method filter_le : elt -> 'a
      
    method range : elt -> elt -> 'a
      
    method min_view : elt * 'a                     (* [Empty] *)
    method max_view : 'a * elt                     (* [Empty] *)
    method partition : (elt -> bool) -> 'a * 'a
    method partition_inf_sup : elt -> 'a * elt option * 'a
      
    method is_empty : bool
    method size : int
      
    method member  : elt -> bool
    method member' : elt -> bool
    method look_for : (elt -> bool) -> elt         (* [Not_found] *)
      
    method choose : elt                            (* [Empty] *)
    method min_element : elt                       (* [Not_found] *)
    method max_element : elt                       (* [Not_found] *)
      
    method next     : elt -> elt                   (* [Not_found] *)
    method previous : elt -> elt                   (* [Not_found] *)
    method floor    : elt -> elt                   (* [Not_found] *)
    method ceil     : elt -> elt                   (* [Not_found] *)
      
    method get  : int -> elt                       (* [Out_of_bounds] *)
    method rank : elt -> int                       (* [Not_found] *)
      
    method for_all : (elt -> bool) -> bool
    method exists  : (elt -> bool) -> bool
      
    method iter : (elt -> unit) -> unit
    method fold : (elt -> 'b -> 'b) -> 'b -> 'b
    method fold_left  : ('b -> elt -> 'b) -> 'b -> 'b
    method fold_right : (elt -> 'b -> 'b) -> 'b -> 'b

  end

  val o_injection : elt -> o_t
  val o_plus : o_t -> o_t -> o_t
  val o_zero : unit -> o_t
  val o_from_list : elt list -> o_t
  val o_from_set  : t -> o_t
    
  val o_union : o_t -> o_t -> o_t
  val o_intersection : o_t -> o_t -> o_t
  val o_difference   : o_t -> o_t -> o_t
  val o_symetric_difference : o_t -> o_t -> o_t

  val o_subset  : o_t -> o_t -> bool
  val o_equal   : o_t -> o_t -> bool
  val o_compare : o_t -> o_t -> int

end


module Make = functor (Ord : OrderedType) ->
struct
  
  type elt = Ord.t
  type t = E | N of t * elt * t * int

  let makeTree = fun l v r -> 
    let h = match (l, r) with
      | (E, E) -> 1
      | (E, N (_, _, _, n)) | (N (_, _, _, n), E) -> 1 + n
      | (N (_, _, _, m), N (_, _, _, n)) -> 1 + m + n
    in
      N (l, v, r, h)

  let rotate_L = fun l v r ->
    match r with
      | N (rl, rv, rr, _) -> makeTree (makeTree l v rl) rv rr
      | _ -> makeTree l v r
	  
  let rotate_R = fun l v r ->
    match l with
      | N (ll, lv, lr, _) -> makeTree ll lv (makeTree lr v r)
      | _ -> makeTree l v r

  let rotate_RL = fun l v r ->
    match r with
      | N (N (rll, rlv, rlr, _), rv, rr, _) -> 
	  makeTree (makeTree l v rll) rlv (makeTree rlr rv rr)
      | _ -> makeTree l v r

  let rotate_LR = fun l v r ->
    match l with
      | N (ll, lv, N (lrl, lrv, lrr, _), _) ->
	  makeTree (makeTree ll lv lrl) lrv (makeTree lrr v r)
      | _ -> makeTree l v r

  let size = function
    | E -> 0
    | N (_, _, _, n) -> n

  let makeBTree = fun l v r ->
    let ln = size l and rn = size r in
      if ln > wt rn then
	(match l with
	   | E -> makeTree l v r
	   | N (ll, _, lr, _) ->
	       if size ll > size lr then rotate_R l v r
	       else rotate_LR l v r
	)
      else if rn > wt ln then
	(match r with
	   | E -> makeTree l v r
	   | N (rl, _, rr, _) ->
	       if size rr > size rl then rotate_L l v r
	       else rotate_RL l v r
	)
      else makeTree l v r
	
  let rec makeSTree = fun l v r ->
    let hl = size l and hr = size r in
      if hl > wt hr then
	(match l with
	   | E -> makeTree l v r
	   | N (ll, lv, lr, _) -> makeBTree ll lv (makeSTree lr v r)
	)
      else if hr > wt hl then
	(match r with
	   | E -> makeTree l v r
	   | N (rl, rv, rr, _) -> makeBTree (makeSTree l v rl) rv rr
      )
      else makeBTree l v r

  let rec append = fun a b ->
    match (a, b) with
      | (E, _) -> b
      | (_, E) -> a
      | (N (l1, v1, r1, _), N (l2, v2, r2, _)) ->
          makeBTree l1 v1 (makeBTree (append r1 l2) v2 r2)

  let rec appendS = fun a b ->
    match (a, b) with
      | (E, _) -> b
      | (_, E) -> a
      | (N (l1, v1, r1, _), N (l2, v2, r2, _)) ->
          makeSTree l1 v1 (makeSTree (appendS r1 l2) v2 r2)

  let injection = function x -> makeTree E x E

  let rec insert_raise x = function
    | E -> injection x
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> makeBTree (insert_raise x l) v r
	  | n when n > 0 -> makeBTree l v (insert_raise x r)
	  | _ -> raise Found

  let insert x = function set ->
    try 
      insert_raise x set 
    with Found -> set

  let rec from_list_rec result = function
    | [] -> result
    | head :: tail -> from_list_rec (insert head result) tail

  let from_list = function list -> from_list_rec E list

  let rec to_list_rec result = function
    | E -> result
    | N (l, v, r, _) -> to_list_rec (v :: to_list_rec result r) l

  let to_list = function set -> to_list_rec [] set

  let map f = function set -> from_list (List.map f (to_list set))

  let rec filter_lt x = function
    | E -> E
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> filter_lt x l
	  | n when n > 0 -> makeSTree l v (filter_lt x r)
	  | _ -> l

  let rec filter_gt x = function
    | E -> E
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> makeSTree (filter_gt x l) v r
	  | n when n > 0 -> filter_gt x r
	  | _ -> r
	      
  let rec old_plus = fun a b ->
    match (a, b) with
      | (E, _) -> b
      | (_, E) -> a
      | (_, N (l, v, r, _)) -> 
	  let 
	    lt_v = filter_lt v a and 
	    gt_v = filter_gt v a 
	  in
	    makeSTree (old_plus lt_v l) v (old_plus gt_v r)

  let rec cut = fun lo tree hi ->
    match tree with
      | N (_, v, r, _) when Ord.compare v lo <= 0 -> cut lo r hi
      | N (l, v, _, _) when Ord.compare v hi >= 0 -> cut lo l hi
      | _ -> (lo, tree, hi)

  let rec cut_hi = fun lo tree ->
    match tree with
      | N (_, v, r, _) when Ord.compare v lo <= 0 -> cut_hi lo r
      | _ -> (lo, tree)

  let rec cut_lo = fun tree hi ->
    match tree with
      | N (l, v, _, _) when Ord.compare v hi >= 0 -> cut_lo l hi
      | _ -> (tree, hi)

  let rec uni = fun a (lo, b, hi) ->
    match (a, b) with
      | (_, E) -> a
      | (E, N (l, v, r, _)) -> makeSTree (filter_gt lo l) v (filter_lt hi r)
      | (N (l, v, r, _), N _) ->
	  makeSTree (uni l (cut lo b v)) v (uni r (cut v b hi))  

  let rec uni_lo = fun a (b, hi) ->
    match (a, b) with
      | (_, E) -> a
      | (E, N (l, v, r, _)) -> makeSTree l v (filter_lt hi r)
      | (N (l, v, r, _), N _) -> 
	  makeSTree (uni_lo l (cut_lo b v)) v (uni r (cut v b hi))

  let rec uni_hi = fun a (lo, b) ->
    match (a, b) with
      | (_, E) -> a
      | (E, N (l, v, r, _)) -> makeSTree (filter_gt lo l) v r
      | (N (l, v, r, _), N _) -> 
	  makeSTree (uni l (cut lo b v)) v (uni_hi r (cut_hi v b)) 

  let plus = fun a b ->
    match (a, b) with
      | (E, _) -> b
      | (_, E) -> a
      | (N (l, v, r, _), N _) ->
	  makeSTree (uni_lo l (cut_lo b v)) v (uni_hi r (cut_hi v b)) 

  let zero = function () -> E

  let rec partition_inf_sup pivot = function
    | E -> (E, None, E)
    | N (l, v, r, _) as node ->
	match Ord.compare pivot v with
	  | n when n < 0 ->
	      let (inf, value, sup) = partition_inf_sup pivot l in
		(inf, value, makeSTree sup v r)
	  | n when n > 0 ->
	      let (inf, value, sup) = partition_inf_sup pivot r in
		(makeSTree l v inf, value, sup)
	  | _ -> (l, Some v, r)

  let union = plus

  let rec intersection = fun a b ->
    match (a, b) with
      | (E, _) | (_, E) -> E
      | (N (l, v, r, _), N _) ->
	  match partition_inf_sup v b with
	    | (l', None, r') -> appendS (intersection l l') (intersection r r')
	    | (l', Some _, r') -> 
		makeSTree (intersection l l') v (intersection r r')

  let rec difference = fun a b ->
    match (a, b) with
      | (E, _) -> E
      | (_, E) -> a
      | (N (l, v, r, _), N _) -> 
	  match partition_inf_sup v b with
	    | (l', None, r') -> makeSTree (difference l l') v (difference r r')
	    | (l', Some _, r') -> appendS (difference l l') (difference r r')
		
  let rec symetric_difference = fun a b ->
    match (a, b) with
      | (E, s) | (s, E) -> s
      | (N (l, v, r, _), N _) ->
	  match partition_inf_sup v b with
	    | (l', None, r') -> makeSTree (symetric_difference l l') v 
		(symetric_difference r r')
	    | (l', Some _, r') -> 
		appendS (symetric_difference l l') (symetric_difference r r')

  let rec subset = fun a b ->
    match (a, b) with
      | (E, _) -> true
      | (_, E) -> false
      | (N (l1, v1, r1, _), N (l2, v2, r2, _)) ->
          match Ord.compare v1 v2 with
	    | n when n < 0 -> subset (makeBTree l1 v1 E) l2 && subset r1 b
	    | n when n > 0 -> subset (makeBTree E v1 r1) r2 && subset l1 b
	    | _ -> subset l1 l2 && subset r1 r2

  let rec compare_rec = fun a b ->
    match (a, b) with
      | ([], []) -> 0
      | ([], _) -> -1
      | (_, []) -> 1
      | (E :: t1, E :: t2) -> compare_rec t1 t2
      | (N (E, v1, r1, _) :: t1, N (E, v2, r2, _) :: t2) ->
	  (match Ord.compare v1 v2 with
	     | 0 -> compare_rec (r1 :: t1) (r2 :: t2)
	     | other -> other
	  )
      | (N (a, v1, r1, _) :: t1, t2) ->
          compare_rec (a :: N (E, v1, r1, 0) :: t1) t2
      | (t1, N (b, v2, r2, _) :: t2) ->
	  compare_rec t1 (b :: N (E, v2, r2, 0) :: t2)

  let equal = fun a b -> compare_rec [a] [b] = 0

  let rec insert_min_bound b x = function
    | E -> 
	if Ord.compare b x = 0 then raise Found
	else injection x
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> makeBTree (insert_min_bound b x l) v r
	  | _ -> makeBTree l v (insert_min_bound v x r)

  let rec insert_raise' x = function
    | E -> injection x
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> makeBTree (insert_raise' x l) v r
	  | _ -> makeBTree l v (insert_min_bound x v r)

  let insert' x = function set ->
    try 
      insert_raise' x set 
    with Found -> set

  let insert_list = fun list set -> plus (from_list list) set

  let rec delete_min = function
    | E -> E
    | N (E, _, r, _) -> r
    | N (l, v, r, _) -> makeBTree (delete_min l) v r

  let rec delete_max = function
    | E -> E
    | N (l, _, E, _) -> l
    | N (l, v, r, _) -> makeBTree l v (delete_max r)

  let rec delete_raise x = function
    | E -> raise Not_found
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> makeBTree (delete_raise x l) v r
	  | n when n > 0 -> makeBTree l v (delete_raise x r)
	  | _ -> append l r

  let delete x = function set -> 
    try 
      delete_raise x set 
    with Not_found -> set

  let rec delete_list = fun list set ->
    match list with
      | [] -> set
      | head :: tail -> delete_list tail (delete head set)

  let rec filter_rec f result = function
    | E -> result
    | N (l, v, r, _) ->
	filter_rec f 
	(filter_rec f (if f v then insert v result else result) l) r

  let filter f = function set -> filter_rec f E set

  let rec filter_le x = function
    | E -> E
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> filter_le x l
	  | _ -> makeSTree l v (filter_le x r)

  let rec filter_ge x = function
    | E -> E
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n <= 0 -> makeSTree (filter_ge x l) v r
	  | _ -> filter_ge x r

  let rec range min max = function
    | E -> E
    | N (l, v, r, _) ->
	if Ord.compare v min < 0 then range min max r
	else if Ord.compare v max > 0 then range min max l
	else makeSTree (filter_ge min l) v (filter_le max r)

  let rec min_view = function
    | E -> raise Empty
    | N (E, v, r, _) -> (v, r)
    | N (l, v, r, _) -> let (min, left) = min_view l in 
	(min, makeBTree left v r)

  let rec max_view = function
    | E -> raise Empty
    | N (l, v, E, _) -> (l, v)
    | N (l, v, r, _) -> let (right, max) = max_view r in 
	(makeBTree l v right, max)

  let rec partition_rec f (trueS, falseS) = function
    | E -> (trueS, falseS)
    | N (l, v, r, _) -> 
	if f v 
	then partition_rec f (partition_rec f (insert v trueS, falseS) l) r
	else partition_rec f (partition_rec f (trueS, insert v falseS) l) r

  let partition = fun f set -> partition_rec f (E, E) set

  let is_empty = function
    | E -> true
    | N _ -> false

  let rec member_min_bound b x = function
    | E -> Ord.compare b x = 0
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | k when k < 0 -> member_min_bound b x l
	  | _ -> member_min_bound v x r
	      
  let rec member' x = function
    | E -> false
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> member' x l
	  | _ -> member_min_bound v x r
	      
  let rec member x = function
    | E -> false
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> member x l
	  | n when n > 0 -> member x r
	  | _ -> true

  let rec look_for f = function
    | E -> raise Not_found
    | N (_, v, _, _) when f v -> v
    | N (l, _, r, _) -> (try look_for f l with Not_found -> look_for f r)

  let choose = function
    | E -> raise Empty
    | N (_, v, _, _) -> v

  let rec min_element = function
    | E -> raise Not_found
    | N (E, v, _, _) -> v
    | N (l, _, _, _) -> min_element l

  let rec max_element = function
    | E -> raise Not_found
    | N (_, v, E, _) -> v
    | N (_, _, r, _) -> max_element r

  let rec next x = function
    | E -> raise Not_found
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> (try next x l with Not_found -> v)
	  | n when n > 0 -> next x r
	  | _ -> min_element r (* raises [Not_found] *)

  let rec previous x = function
    | E -> raise Not_found
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> previous x l
	  | n when n > 0 -> (try previous x r with Not_found -> v)
	  | _ -> max_element l (* raises [Not_found] *)
	      
  let rec ceil x = function
    | E -> raise Not_found
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> (try ceil x l with Not_found -> v)
	  | n when n > 0 -> ceil x r
	  | _ -> x

  let rec floor x = function
    | E -> raise Not_found
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> floor x l
	  | n when n > 0 -> (try floor x r with Not_found -> v)
	  | _ -> x

  let rec rank x = function
    | E -> raise Not_found
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> rank x l
	  | n when n > 0 -> 1 + size l + rank x r
	  | _ -> size l

  let rec get n = function
    | E -> raise (Out_of_bounds n)
    | N (l, v, r, _) ->
	let k = size l in
	  if n < k then get n l
	  else if n > k then get (n - k - 1) r
	  else v

  let rec set_rec value m = function
    | E -> raise (Out_of_bounds m)
    | N (l, v, r, h) ->
	let k = size l in
	  if m < k then N (set_rec value m l, v, r, h)
	  else if m > k then N (l, v, set_rec value (m - k - 1) r, h)
	  else if v = value then raise Found
	  else N (l, value, r, h)
	    
  let unsafe_set = fun value n tree -> 
    try 
      set_rec value n tree 
    with Found -> tree

  let rec for_all f = function
    | E -> true
    | N (l, v, r, _) -> f v && for_all f l && for_all f r

  let rec exists f = function
    | E -> false
    | N (l, v, r, _) -> f v || exists f l || exists f r

  let rec iter f = function
    | E -> ()
    | N (l, v, r, _) -> iter f l ; f v ; iter f r

  let rec fold_left f e = function
    | E -> e
    | N (l, v, r, _) -> fold_left f (f (fold_left f e l) v) r

  let rec fold_right = fun f set e ->
    match set with
      | E -> e 
      | N (l, v, r, _) -> fold_right f l (f v (fold_right f r e))

  let fold = fold_right

  let rec unsafe_inc_map f = function
    | E -> E
    | N (l, v, r, _) -> 
	makeTree (unsafe_inc_map f l) (f v) (unsafe_inc_map f r)

  let rec unsafe_dec_map f = function
    | E -> E
    | N (l, v, r, _) -> 
	makeTree (unsafe_dec_map f l) (f v) (unsafe_dec_map f r)

  let rec unsafe_insert_min x = function
    | E -> injection x
    | N (l, v, r, _) -> makeTree (unsafe_insert_min x l) v r 

  let rec unsafe_insert_max x = function
    | E -> injection x
    | N (l, v, r, _) -> makeTree l v (unsafe_insert_max x r)

  let empty = E
  let mem = member
  let add = insert
  let singleton = injection
  let remove = delete
  let inter = intersection
  let diff = difference
  let cardinal = size
  let elements = to_list
  let min_elt = min_element    
  let max_elt = max_element

  class o_t = fun set ->
  object (self : 'a)
    
    val contents  = set

    method to_set = contents

    method map f = {< contents = map f contents >}

    method plus = function (set : 'a) -> 
      {< contents = plus contents set#to_set >} 
    method to_list = to_list contents
		       
    method union = function (set : 'a) -> 
      {< contents = plus contents set#to_set >}
      
    method intersection = function (set : 'a) -> 
      {< contents = intersection contents set#to_set >}
      
    method difference = function (set : 'a) -> 
      {< contents = difference contents set#to_set >}
      
    method symetric_difference = function (set : 'a) -> 
      {< contents = symetric_difference contents set#to_set >}
      
    method subset = function (set : 'a) -> subset contents set#to_set
    method equal = function (set : 'a)-> equal contents set#to_set
    method compare = function (set : 'a) -> compare_rec [contents] [set#to_set]
      
    method insert = function x -> 
      try 
	{< contents = insert_raise x contents >} 
      with Found -> self

    method insert_raise = function x -> 
      {< contents = insert_raise x contents >}

    method insert' = function x -> 
      try
	{< contents = insert_raise' x contents >}
      with Found -> self

    method insert_raise' = function x -> 
      {< contents = insert_raise' x contents >}
      
    method insert_list = function list -> 
      {< contents = insert_list list contents >}

    method delete = function x -> 
      try 
	{< contents = delete_raise x contents >} 
      with Not_found -> self

    method delete_raise = function x -> 
      {< contents = delete_raise x contents >}

    method delete_list = function list -> 
      {< contents = delete_list list contents >}
      
    method delete_min = {< contents = delete_min contents >}
    method delete_max = {< contents = delete_max contents >}

    method filter = function f -> {< contents = filter f contents >}
    method filter_lt = function x -> {< contents = filter_lt x contents >}
    method filter_gt = function x -> {< contents = filter_gt x contents >}
    method filter_le = function x -> {< contents = filter_le x contents >}
    method filter_ge = function x -> {< contents = filter_ge x contents >}

    method range = fun min max -> {< contents = range min max contents >}

    method min_view =
      let (min, other) = min_view contents in
	(min, {< contents = other >})

    method max_view =
      let (other, max) = max_view contents in
	({< contents = other >}, max)

    method partition = function f -> 
      let (trueSet, falseSet) = partition f contents in
	({< contents = trueSet >}, {< contents = falseSet >})

    method partition_inf_sup = function x ->
      let (inf, value, sup) = partition_inf_sup x contents in
	({< contents = inf >}, value, {< contents = sup >})

    method is_empty = contents = E

    method size = 
      match contents with
	| E -> 0
	| N (_, _, _, n) -> n

    method member = function x -> member x contents
    method member' = function x -> member' x contents
    method look_for = function f -> look_for f contents

    method choose =
      match contents with
	| E -> raise Not_found
	| N (_, v, _, _) -> v

    method min_element = min_element contents
    method max_element = max_element contents

    method previous = function x -> previous x contents
    method next = function x -> next x contents
    method floor = function x -> floor x contents
    method ceil = function x -> ceil x contents

    method rank = function x -> rank x contents
    method get = function n -> get n contents

    method exists = function f -> exists f contents
    method for_all = function f -> for_all f contents

    method iter : (elt -> unit) -> unit = function f -> iter f contents

    method fold : 'b. (elt -> 'b -> 'b) -> 'b -> 'b = 
      fun f e -> fold_right f contents e

    method fold_left : 'b. ('b -> elt -> 'b) -> 'b -> 'b =
      fun f e -> fold_left f e contents

    method fold_right : 'b. (elt -> 'b -> 'b) -> 'b -> 'b =
      fun f e -> fold_right f contents e

  end

  let o_injection = function x -> new o_t (injection x)
  let o_plus = fun a b -> new o_t (plus a#to_set b#to_set)
  let o_zero = function () -> new o_t E
  let o_from_set = function set -> new o_t set
  let o_from_list = function list -> new o_t (from_list list)

  let o_union = fun a b -> new o_t (plus a#to_set b#to_set)
  let o_intersection = fun a b -> new o_t (intersection a#to_set b#to_set)
  let o_difference = fun a b -> new o_t (difference a#to_set b#to_set)
  let o_symetric_difference = fun a b -> 
    new o_t (symetric_difference a#to_set b#to_set)

  let o_subset = fun a b -> subset a#to_set b#to_set
  let o_equal = fun a b -> equal a#to_set b#to_set
  let o_compare = fun a b -> compare_rec [a#to_set] [b#to_set]

  let compare = fun a b -> compare_rec [a] [b]

  (** Debug **)
    
  let rec mn = function
    | E -> 0
    | N (a, _, b, _) -> 1 + min (mn a) (mn b)
	
  let rec mx = function
    | E -> 0
    | N (a, _, b, _) -> 1 + max (mx a) (mx b)
	
  let rec med total = function
    | E -> [total]
    | N (a, _, b, _) -> med (total + 1) a @ med (total + 1) b
	
  let stat = function tree -> 
    let height_list = med 0 tree in
    let med_h = (List.fold_left (+) 0 height_list) / 
		(List.length height_list) 
    in
      (size tree, mn tree, mx tree, med_h)
      
end


module type RelaxationInformation = sig
  val relaxation_coefficient : int
end

module MakeRelaxed (Inf : RelaxationInformation) (Ord : OrderedType) =
struct
  
  type elt = Ord.t
  type t = E | N of t * elt * t * int

  let bound = 
    if Inf.relaxation_coefficient <= 0 
    then invalid "RelaxationInformation" Negative
    else Inf.relaxation_coefficient

  let wt = function x -> bound * x

  let makeTree = fun l v r -> 
    let h = match (l, r) with
      | (E, E) -> 1
      | (E, N (_, _, _, n)) | (N (_, _, _, n), E) -> 1 + n
      | (N (_, _, _, m), N (_, _, _, n)) -> 1 + m + n
    in
      N (l, v, r, h)

  let rotate_L = fun l v r ->
    match r with
      | N (rl, rv, rr, _) -> makeTree (makeTree l v rl) rv rr
      | _ -> makeTree l v r
	  
  let rotate_R = fun l v r ->
    match l with
      | N (ll, lv, lr, _) -> makeTree ll lv (makeTree lr v r)
      | _ -> makeTree l v r

  let rotate_RL = fun l v r ->
    match r with
      | N (N (rll, rlv, rlr, _), rv, rr, _) -> 
	  makeTree (makeTree l v rll) rlv (makeTree rlr rv rr)
      | _ -> makeTree l v r

  let rotate_LR = fun l v r ->
    match l with
      | N (ll, lv, N (lrl, lrv, lrr, _), _) ->
	  makeTree (makeTree ll lv lrl) lrv (makeTree lrr v r)
      | _ -> makeTree l v r

  let size = function
    | E -> 0
    | N (_, _, _, n) -> n

  let makeBTree = fun l v r ->
    let ln = size l and rn = size r in
      if ln > wt rn then
	(match l with
	   | E -> makeTree l v r
	   | N (ll, _, lr, _) ->
	       if size ll > size lr then rotate_R l v r
	       else rotate_LR l v r
	)
      else if rn > wt ln then
	(match r with
	   | E -> makeTree l v r
	   | N (rl, _, rr, _) ->
	       if size rr > size rl then rotate_L l v r
	       else rotate_RL l v r
	)
      else makeTree l v r
	
  let rec makeSTree = fun l v r ->
    let hl = size l and hr = size r in
      if hl > wt hr then
	(match l with
	   | E -> makeTree l v r
	   | N (ll, lv, lr, _) -> makeBTree ll lv (makeSTree lr v r)
	)
      else if hr > wt hl then
	(match r with
	   | E -> makeTree l v r
	   | N (rl, rv, rr, _) -> makeBTree (makeSTree l v rl) rv rr
      )
      else makeBTree l v r

  let rec append = fun a b ->
    match (a, b) with
      | (E, _) -> b
      | (_, E) -> a
      | (N (l1, v1, r1, _), N (l2, v2, r2, _)) ->
          makeBTree l1 v1 (makeBTree (append r1 l2) v2 r2)

  let rec appendS = fun a b ->
    match (a, b) with
      | (E, _) -> b
      | (_, E) -> a
      | (N (l1, v1, r1, _), N (l2, v2, r2, _)) ->
          makeSTree l1 v1 (makeSTree (appendS r1 l2) v2 r2)

  let injection = function x -> makeTree E x E

  let rec insert_raise x = function
    | E -> injection x
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> makeBTree (insert_raise x l) v r
	  | n when n > 0 -> makeBTree l v (insert_raise x r)
	  | _ -> raise Found

  let insert x = function set ->
    try 
      insert_raise x set 
    with Found -> set

  let rec from_list_rec result = function
    | [] -> result
    | head :: tail -> from_list_rec (insert head result) tail

  let from_list = function list -> from_list_rec E list

  let rec to_list_rec result = function
    | E -> result
    | N (l, v, r, _) -> to_list_rec (v :: to_list_rec result r) l

  let to_list = function set -> to_list_rec [] set

  let map f = function set -> from_list (List.map f (to_list set))

  let rec filter_lt x = function
    | E -> E
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> filter_lt x l
	  | n when n > 0 -> makeSTree l v (filter_lt x r)
	  | _ -> l

  let rec filter_gt x = function
    | E -> E
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> makeSTree (filter_gt x l) v r
	  | n when n > 0 -> filter_gt x r
	  | _ -> r
	      
  let rec old_plus = fun a b ->
    match (a, b) with
      | (E, _) -> b
      | (_, E) -> a
      | (_, N (l, v, r, _)) -> 
	  let 
	    lt_v = filter_lt v a and 
	    gt_v = filter_gt v a 
	  in
	    makeSTree (old_plus lt_v l) v (old_plus gt_v r)

  let rec cut = fun lo tree hi ->
    match tree with
      | N (_, v, r, _) when Ord.compare v lo <= 0 -> cut lo r hi
      | N (l, v, _, _) when Ord.compare v hi >= 0 -> cut lo l hi
      | _ -> (lo, tree, hi)

  let rec cut_hi = fun lo tree ->
    match tree with
      | N (_, v, r, _) when Ord.compare v lo <= 0 -> cut_hi lo r
      | _ -> (lo, tree)

  let rec cut_lo = fun tree hi ->
    match tree with
      | N (l, v, _, _) when Ord.compare v hi >= 0 -> cut_lo l hi
      | _ -> (tree, hi)

  let rec uni = fun a (lo, b, hi) ->
    match (a, b) with
      | (_, E) -> a
      | (E, N (l, v, r, _)) -> makeSTree (filter_gt lo l) v (filter_lt hi r)
      | (N (l, v, r, _), N _) ->
	  makeSTree (uni l (cut lo b v)) v (uni r (cut v b hi))  

  let rec uni_lo = fun a (b, hi) ->
    match (a, b) with
      | (_, E) -> a
      | (E, N (l, v, r, _)) -> makeSTree l v (filter_lt hi r)
      | (N (l, v, r, _), N _) -> 
	  makeSTree (uni_lo l (cut_lo b v)) v (uni r (cut v b hi))

  let rec uni_hi = fun a (lo, b) ->
    match (a, b) with
      | (_, E) -> a
      | (E, N (l, v, r, _)) -> makeSTree (filter_gt lo l) v r
      | (N (l, v, r, _), N _) -> 
	  makeSTree (uni l (cut lo b v)) v (uni_hi r (cut_hi v b)) 

  let plus = fun a b ->
    match (a, b) with
      | (E, _) -> b
      | (_, E) -> a
      | (N (l, v, r, _), N _) ->
	  makeSTree (uni_lo l (cut_lo b v)) v (uni_hi r (cut_hi v b)) 

  let zero = function () -> E

  let rec partition_inf_sup pivot = function
    | E -> (E, None, E)
    | N (l, v, r, _) as node ->
	match Ord.compare pivot v with
	  | n when n < 0 ->
	      let (inf, value, sup) = partition_inf_sup pivot l in
		(inf, value, makeSTree sup v r)
	  | n when n > 0 ->
	      let (inf, value, sup) = partition_inf_sup pivot r in
		(makeSTree l v inf, value, sup)
	  | _ -> (l, Some v, r)

  let union = plus

  let rec intersection = fun a b ->
    match (a, b) with
      | (E, _) | (_, E) -> E
      | (N (l, v, r, _), N _) ->
	  match partition_inf_sup v b with
	    | (l', None, r') -> appendS (intersection l l') (intersection r r')
	    | (l', Some _, r') -> 
		makeSTree (intersection l l') v (intersection r r')

  let rec difference = fun a b ->
    match (a, b) with
      | (E, _) -> E
      | (_, E) -> a
      | (N (l, v, r, _), N _) -> 
	  match partition_inf_sup v b with
	    | (l', None, r') -> makeSTree (difference l l') v (difference r r')
	    | (l', Some _, r') -> appendS (difference l l') (difference r r')
		
  let rec symetric_difference = fun a b ->
    match (a, b) with
      | (E, s) | (s, E) -> s
      | (N (l, v, r, _), N _) ->
	  match partition_inf_sup v b with
	    | (l', None, r') -> makeSTree (symetric_difference l l') v 
		(symetric_difference r r')
	    | (l', Some _, r') -> 
		appendS (symetric_difference l l') (symetric_difference r r')

  let rec subset = fun a b ->
    match (a, b) with
      | (E, _) -> true
      | (_, E) -> false
      | (N (l1, v1, r1, _), N (l2, v2, r2, _)) ->
          match Ord.compare v1 v2 with
	    | n when n < 0 -> subset (makeBTree l1 v1 E) l2 && subset r1 b
	    | n when n > 0 -> subset (makeBTree E v1 r1) r2 && subset l1 b
	    | _ -> subset l1 l2 && subset r1 r2

  let rec compare_rec = fun a b ->
    match (a, b) with
      | ([], []) -> 0
      | ([], _) -> -1
      | (_, []) -> 1
      | (E :: t1, E :: t2) -> compare_rec t1 t2
      | (N (E, v1, r1, _) :: t1, N (E, v2, r2, _) :: t2) ->
	  (match Ord.compare v1 v2 with
	     | 0 -> compare_rec (r1 :: t1) (r2 :: t2)
	     | other -> other
	  )
      | (N (a, v1, r1, _) :: t1, t2) ->
          compare_rec (a :: N (E, v1, r1, 0) :: t1) t2
      | (t1, N (b, v2, r2, _) :: t2) ->
	  compare_rec t1 (b :: N (E, v2, r2, 0) :: t2)

  let equal = fun a b -> compare_rec [a] [b] = 0

  let rec insert_min_bound b x = function
    | E -> 
	if Ord.compare b x = 0 then raise Found
	else injection x
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> makeBTree (insert_min_bound b x l) v r
	  | _ -> makeBTree l v (insert_min_bound v x r)

  let rec insert_raise' x = function
    | E -> injection x
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> makeBTree (insert_raise' x l) v r
	  | _ -> makeBTree l v (insert_min_bound x v r)

  let insert' x = function set ->
    try 
      insert_raise' x set 
    with Found -> set

  let insert_list = fun list set -> plus (from_list list) set

  let rec delete_min = function
    | E -> E
    | N (E, _, r, _) -> r
    | N (l, v, r, _) -> makeBTree (delete_min l) v r

  let rec delete_max = function
    | E -> E
    | N (l, _, E, _) -> l
    | N (l, v, r, _) -> makeBTree l v (delete_max r)

  let rec delete_raise x = function
    | E -> raise Not_found
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> makeBTree (delete_raise x l) v r
	  | n when n > 0 -> makeBTree l v (delete_raise x r)
	  | _ -> append l r

  let delete x = function set -> 
    try 
      delete_raise x set 
    with Not_found -> set

  let rec delete_list = fun list set ->
    match list with
      | [] -> set
      | head :: tail -> delete_list tail (delete head set)

  let rec filter_rec f result = function
    | E -> result
    | N (l, v, r, _) ->
	filter_rec f 
	(filter_rec f (if f v then insert v result else result) l) r

  let filter f = function set -> filter_rec f E set

  let rec filter_le x = function
    | E -> E
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> filter_le x l
	  | _ -> makeSTree l v (filter_le x r)

  let rec filter_ge x = function
    | E -> E
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n <= 0 -> makeSTree (filter_ge x l) v r
	  | _ -> filter_ge x r

  let rec range min max = function
    | E -> E
    | N (l, v, r, _) ->
	if Ord.compare v min < 0 then range min max r
	else if Ord.compare v max > 0 then range min max l
	else makeSTree (filter_ge min l) v (filter_le max r)

  let rec min_view = function
    | E -> raise Empty
    | N (E, v, r, _) -> (v, r)
    | N (l, v, r, _) -> let (min, left) = min_view l in 
	(min, makeBTree left v r)

  let rec max_view = function
    | E -> raise Empty
    | N (l, v, E, _) -> (l, v)
    | N (l, v, r, _) -> let (right, max) = max_view r in 
	(makeBTree l v right, max)

  let rec partition_rec f (trueS, falseS) = function
    | E -> (trueS, falseS)
    | N (l, v, r, _) -> 
	if f v 
	then partition_rec f (partition_rec f (insert v trueS, falseS) l) r
	else partition_rec f (partition_rec f (trueS, insert v falseS) l) r

  let partition = fun f set -> partition_rec f (E, E) set

  let is_empty = function
    | E -> true
    | N _ -> false

  let rec member_min_bound b x = function
    | E -> Ord.compare b x = 0
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | k when k < 0 -> member_min_bound b x l
	  | _ -> member_min_bound v x r
	      
  let rec member' x = function
    | E -> false
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> member' x l
	  | _ -> member_min_bound v x r
	      
  let rec member x = function
    | E -> false
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> member x l
	  | n when n > 0 -> member x r
	  | _ -> true

  let rec look_for f = function
    | E -> raise Not_found
    | N (_, v, _, _) when f v -> v
    | N (l, _, r, _) -> (try look_for f l with Not_found -> look_for f r)

  let choose = function
    | E -> raise Empty
    | N (_, v, _, _) -> v

  let rec min_element = function
    | E -> raise Not_found
    | N (E, v, _, _) -> v
    | N (l, _, _, _) -> min_element l

  let rec max_element = function
    | E -> raise Not_found
    | N (_, v, E, _) -> v
    | N (_, _, r, _) -> max_element r

  let rec next x = function
    | E -> raise Not_found
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> (try next x l with Not_found -> v)
	  | n when n > 0 -> next x r
	  | _ -> min_element r (* raises [Not_found] *)

  let rec previous x = function
    | E -> raise Not_found
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> previous x l
	  | n when n > 0 -> (try previous x r with Not_found -> v)
	  | _ -> max_element l (* raises [Not_found] *)
	      
  let rec ceil x = function
    | E -> raise Not_found
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> (try ceil x l with Not_found -> v)
	  | n when n > 0 -> ceil x r
	  | _ -> x

  let rec floor x = function
    | E -> raise Not_found
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> floor x l
	  | n when n > 0 -> (try floor x r with Not_found -> v)
	  | _ -> x

  let rec rank x = function
    | E -> raise Not_found
    | N (l, v, r, _) ->
	match Ord.compare x v with
	  | n when n < 0 -> rank x l
	  | n when n > 0 -> 1 + size l + rank x r
	  | _ -> size l

  let rec get n = function
    | E -> raise (Out_of_bounds n)
    | N (l, v, r, _) ->
	let k = size l in
	  if n < k then get n l
	  else if n > k then get (n - k - 1) r
	  else v

  let rec set_rec value m = function
    | E -> raise (Out_of_bounds m)
    | N (l, v, r, h) ->
	let k = size l in
	  if m < k then N (set_rec value m l, v, r, h)
	  else if m > k then N (l, v, set_rec value (m - k - 1) r, h)
	  else if v = value then raise Found
	  else N (l, value, r, h)
	    
  let unsafe_set = fun value n tree -> 
    try 
      set_rec value n tree 
    with Found -> tree

  let rec for_all f = function
    | E -> true
    | N (l, v, r, _) -> f v && for_all f l && for_all f r

  let rec exists f = function
    | E -> false
    | N (l, v, r, _) -> f v || exists f l || exists f r

  let rec iter f = function
    | E -> ()
    | N (l, v, r, _) -> iter f l ; f v ; iter f r

  let rec fold_left f e = function
    | E -> e
    | N (l, v, r, _) -> fold_left f (f (fold_left f e l) v) r

  let rec fold_right = fun f set e ->
    match set with
      | E -> e 
      | N (l, v, r, _) -> fold_right f l (f v (fold_right f r e))

  let fold = fold_right

  let rec unsafe_inc_map f = function
    | E -> E
    | N (l, v, r, _) -> 
	makeTree (unsafe_inc_map f l) (f v) (unsafe_inc_map f r)

  let rec unsafe_dec_map f = function
    | E -> E
    | N (l, v, r, _) -> 
	makeTree (unsafe_dec_map f l) (f v) (unsafe_dec_map f r)

  let rec unsafe_insert_min x = function
    | E -> injection x
    | N (l, v, r, _) -> makeTree (unsafe_insert_min x l) v r 

  let rec unsafe_insert_max x = function
    | E -> injection x
    | N (l, v, r, _) -> makeTree l v (unsafe_insert_max x r)

  let empty = E
  let mem = member
  let add = insert
  let singleton = injection
  let remove = delete
  let inter = intersection
  let diff = difference
  let cardinal = size
  let elements = to_list
  let min_elt = min_element    
  let max_elt = max_element

  class o_t = fun set ->
  object (self : 'a)
    
    val contents  = set

    method to_set = contents

    method map f = {< contents = map f contents >}

    method plus = function (set : 'a) -> 
      {< contents = plus contents set#to_set >} 
    method to_list = to_list contents
		       
    method union = function (set : 'a) -> 
      {< contents = plus contents set#to_set >}
      
    method intersection = function (set : 'a) -> 
      {< contents = intersection contents set#to_set >}
      
    method difference = function (set : 'a) -> 
      {< contents = difference contents set#to_set >}
      
    method symetric_difference = function (set : 'a) -> 
      {< contents = symetric_difference contents set#to_set >}
      
    method subset = function (set : 'a) -> subset contents set#to_set
    method equal = function (set : 'a)-> equal contents set#to_set
    method compare = function (set : 'a) -> compare_rec [contents] [set#to_set]
      
    method insert = function x -> 
      try 
	{< contents = insert_raise x contents >} 
      with Found -> self

    method insert_raise = function x -> 
      {< contents = insert_raise x contents >}

    method insert' = function x -> 
      try
	{< contents = insert_raise' x contents >}
      with Found -> self

    method insert_raise' = function x -> 
      {< contents = insert_raise' x contents >}
      
    method insert_list = function list -> 
      {< contents = insert_list list contents >}

    method delete = function x -> 
      try 
	{< contents = delete_raise x contents >} 
      with Not_found -> self

    method delete_raise = function x -> 
      {< contents = delete_raise x contents >}

    method delete_list = function list -> 
      {< contents = delete_list list contents >}
      
    method delete_min = {< contents = delete_min contents >}
    method delete_max = {< contents = delete_max contents >}

    method filter = function f -> {< contents = filter f contents >}
    method filter_lt = function x -> {< contents = filter_lt x contents >}
    method filter_gt = function x -> {< contents = filter_gt x contents >}
    method filter_le = function x -> {< contents = filter_le x contents >}
    method filter_ge = function x -> {< contents = filter_ge x contents >}

    method range = fun min max -> {< contents = range min max contents >}

    method min_view =
      let (min, other) = min_view contents in
	(min, {< contents = other >})

    method max_view =
      let (other, max) = max_view contents in
	({< contents = other >}, max)

    method partition = function f -> 
      let (trueSet, falseSet) = partition f contents in
	({< contents = trueSet >}, {< contents = falseSet >})

    method partition_inf_sup = function x ->
      let (inf, value, sup) = partition_inf_sup x contents in
	({< contents = inf >}, value, {< contents = sup >})

    method is_empty = contents = E
    method size = size contents

    method member = function x -> member x contents
    method member' = function x -> member' x contents
    method look_for = function f -> look_for f contents

    method choose =
      match contents with
	| E -> raise Not_found
	| N (_, v, _, _) -> v

    method min_element = min_element contents
    method max_element = max_element contents

    method previous = function x -> previous x contents
    method next = function x -> next x contents
    method floor = function x -> floor x contents
    method ceil = function x -> ceil x contents

    method rank = function x -> rank x contents
    method get = function n -> get n contents

    method exists = function f -> exists f contents
    method for_all = function f -> for_all f contents

    method iter : (elt -> unit) -> unit = function f -> iter f contents

    method fold : 'b. (elt -> 'b -> 'b) -> 'b -> 'b = 
      fun f e -> fold_right f contents e

    method fold_left : 'b. ('b -> elt -> 'b) -> 'b -> 'b =
      fun f e -> fold_left f e contents

    method fold_right : 'b. (elt -> 'b -> 'b) -> 'b -> 'b =
      fun f e -> fold_right f contents e

  end

  let o_injection = function x -> new o_t (injection x)
  let o_plus = fun a b -> new o_t (plus a#to_set b#to_set)
  let o_zero = function () -> new o_t E
  let o_from_set = function set -> new o_t set
  let o_from_list = function list -> new o_t (from_list list)

  let o_union = fun a b -> new o_t (plus a#to_set b#to_set)
  let o_intersection = fun a b -> new o_t (intersection a#to_set b#to_set)
  let o_difference = fun a b -> new o_t (difference a#to_set b#to_set)
  let o_symetric_difference = fun a b -> 
    new o_t (symetric_difference a#to_set b#to_set)

  let o_subset = fun a b -> subset a#to_set b#to_set
  let o_equal = fun a b -> equal a#to_set b#to_set
  let o_compare = fun a b -> compare_rec [a#to_set] [b#to_set]

  let compare = fun a b -> compare_rec [a] [b]

  (** Debug **)
    
  let rec mn = function
    | E -> 0
    | N (a, _, b, _) -> 1 + min (mn a) (mn b)
	
  let rec mx = function
    | E -> 0
    | N (a, _, b, _) -> 1 + max (mx a) (mx b)
	
  let rec med total = function
    | E -> [total]
    | N (a, _, b, _) -> med (total + 1) a @ med (total + 1) b
	
  let stat = function tree -> 
    let height_list = med 0 tree in
    let med_h = (List.fold_left (+) 0 height_list) / 
		(List.length height_list) 
    in
      (size tree, mn tree, mx tree, med_h)
      
end

(** Debug **)

let rec mn = function
  | E -> 0
  | N (a, _, b, _) -> 1 + min (mn a) (mn b)

let rec mx = function
  | E -> 0
  | N (a, _, b, _) -> 1 + max (mx a) (mx b)

let rec med total = function
  | E -> [total]
  | N (a, _, b, _) -> med (total + 1) a @ med (total + 1) b

let stat = function tree -> 
  let height_list = med 0 tree in
  let med_h = (List.fold_left (+) 0 height_list) / 
	      (List.length height_list) 
  in
    (size tree, mn tree, mx tree, med_h)

(* To avoid interference with [Pervasives.compare] *)
(* Pour viter toute interfrence avec [Pervasives.compare] *)
(* Para evitar toda interferencia con [Pervasives.compare] *)

let compare = fun a b -> compare_rec [a] [b]


