(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Streams
 * Listes  valuation paresseuse
 * Listas peresosas
 * 
 *)

open BairePervasives

let moduleName = "StreamIV"

type
  'a stream = { mutable data : 'a streamCell }  and
  'a streamCell =
    | Nil
    | Cons of 'a * 'a stream
    | Append of 'a stream * 'a stream
    | Delayed of (unit -> 'a stream)
    | Generator of int * (int -> int) * (int -> 'a)

(* Nous introduisons  prsent un oprateur de concatnation et un
 * gnrateur de valeurs indices :
 *
 *  Generator (index, succ, gen)
 *
 * Le calcul de la valeur suivante se fait par gen (succ (index))
 * laissant le gnrateur dans l'tat (succ (index), succ, gen)
 * Cela permet d'offrir des fonctions [drop] et [take] relativement
 * efficaces
 *
 * - la fonction [gen] renvoie une valeur de type ['a] ou bien lve l'exception
 * [Empty] pour signaler la fin du flot 
 * - la fonction [succ] renvoie une valeur de type [int] ou bien lve 
 * l'exception [Empty] pour signaler la fin du flot
 *)


(** Utils *)

let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type

let makeStream = function x -> { data = x }

let rec last_element = function stream ->
  match stream.data with
    | Cons (_, tail) -> last_element tail
    | Append (a, b) -> last_element a
    | Delayed _ | Generator _ | Nil -> stream

let rec force = function stream ->
  (match stream.data with
    | Delayed f -> print_endline "force lazy" ; 
	let x = f () in stream.data <- x.data
    | Generator (index, succ, gen) ->
	print_endline "force generator" ;
	(try
	   let next_index = succ index in
	     stream.data <-
	     Cons (
	       gen (next_index), 
	       makeStream (Generator (next_index, succ, gen)))
	 with Empty -> stream.data<- Nil)
    | Append (a, b) ->
	let new_stream = force a in
	let last = last_element new_stream in
	  stream.data <- new_stream.data ;
	  last.data <- Append (makeStream last.data, b)
    | Cons _ | Nil -> ()
  ) ;
  stream

let o = fun f g -> (function x -> f (g x))


(** Monadic operations *)


let unit = function x -> makeStream (Cons (x, makeStream Nil))

let rec map f = function stream ->
  match stream.data with
    | Nil -> makeStream Nil
    | Cons (head, tail) -> makeStream (Cons (f head, map f tail))
    | Append (a, b) -> makeStream (Append (map f a, map f b))
    | Delayed _ -> makeStream (Delayed (function () -> map f (force stream)))
    | Generator (i, s, g) -> makeStream (Generator (i, s, o f g))

let rec plus = fun a b ->
  match a.data with
    | Nil -> b
    | Cons (head, tail) -> makeStream (Cons (head, plus tail b))
    | Append (c, d) -> plus (plus c d) b
    | Delayed _ -> makeStream (Delayed (function () -> plus (force a) b))
    | Generator _ -> makeStream (Append (a, b))

let zero = function () -> makeStream Nil

let rec join = function stream ->
  match stream.data with
    | Nil -> makeStream Nil
    | Cons (head, tail) -> plus head (join tail)
    | Append (a, b) -> plus (join a) (join b)
    | Delayed _ | Generator _ -> 
	makeStream (Delayed (function () -> join (force stream)))

let bind = fun stream f -> join (map f stream)

let rec make_rec = fun n x ->
  match n with
    | 0 -> makeStream Nil
    | _ -> makeStream (Cons (x, make_rec (n - 1) x))

let make = fun n x ->
  if n < 0 then invalid "make" Negative
  else make_rec n x

let rec get = fun stream n ->
  match stream.data with
    | Nil -> raise (Out_of_bounds n)
    | Cons (x, _) when n = 0 -> x
    | Cons (_, a) -> get a (n - 1)
    | Append (a, b) -> (try get a n with Out_of_bounds k -> get b k)
    | Delayed _ | Generator _  -> get (force stream) n

let rec set = fun stream n x ->
  match stream.data with
    | Nil when n = 0 -> makeStream (Cons (x, makeStream Nil))
    | Nil -> raise (Out_of_bounds n)
    | Cons (_, tail) when n = 0 -> makeStream (Cons (x, tail))
    | Cons (head, tail) -> makeStream (Cons (head, set tail (n - 1) x))
    | Append (a, _) -> set a n x
    | Delayed _ -> makeStream (Delayed (function () -> set (force stream) n x))
    | Generator _ -> set (force stream) n x

let rec from_list = function
  | [] -> makeStream Nil
  | head :: tail -> makeStream (Cons (head, from_list tail))

let from_function = function f -> makeStream (Generator (0, succ, f))

let from_generator = fun i s f -> makeStream (Generator (i, s, f))

let rec to_list_rec e = function stream ->
  match stream.data with
    | Nil -> e
    | Cons (head, tail) -> head :: to_list_rec e tail
    | Append (a, b) -> to_list_rec (to_list_rec e b) a
    | Delayed _ | Generator _ -> to_list_rec e (force stream)

let to_list = function stream -> to_list_rec [] stream

let rec reverse_onto result = function stream ->
  match stream.data with
    | Nil -> result
    | Cons (head, tail) -> 
	reverse_onto (makeStream (Cons (head, result))) tail
    | Append (a, b) -> reverse_onto (reverse_onto result a) b
    | Delayed _ | Generator _ -> reverse_onto result (force stream)
	
let reverse = function stream -> reverse_onto (makeStream Nil) stream

let reverse_append = fun a b -> reverse_onto b a

let append_reverse = fun a b -> plus a (reverse b)

let rec reverse_iter f = function stream ->
  match stream.data with
    | Nil -> ()
    | Cons (head, tail) -> reverse_iter f tail ; f head
    | Append (a, b) -> reverse_iter f b ; reverse_iter f a
    | Delayed _ | Generator _ -> reverse_iter f (force stream)

let cons x = function stream -> makeStream (Cons (x, stream))

let rec snoc = fun stream x ->
  match stream.data with
    | Nil -> makeStream (Cons (x, makeStream Nil))
    | Cons (head, tail) -> makeStream (Cons (head, snoc tail x))
    | Append (a, b) -> makeStream (Append (a, snoc b x))
    | Delayed _ | Generator _ -> 
	makeStream (Append (stream, makeStream (Cons (x, makeStream Nil))))

let init = fun n f ->
  let rec init_rec = function
    | k when k = n -> makeStream Nil
    | k -> makeStream (Cons (f k, init_rec (n + 1)))
  in
    init_rec 0

let rec left_view = function stream ->
  match stream.data with
    | Nil -> raise Empty
    | Cons (head, tail) -> (head, tail)
    | Append (a, b) -> let (ha, ta) = left_view a in (ha, plus ta b)
    | Delayed _ | Generator _ -> left_view (force stream)

let rec left_head = function stream ->
  match stream.data with
    | Nil -> raise Empty
    | Cons (head, _) -> head
    | Append (a, _) -> left_head a
    | Delayed _ | Generator _ -> left_head (force stream)

let rec left_tail = function stream ->
  match stream.data with
    | Nil -> raise Empty
    | Cons (_, tail) -> tail
    | Append (a, b) -> plus (left_tail a) b
    | Delayed _ | Generator _ -> left_tail (force stream)

let rec is_empty = function stream ->
  match stream.data with
    | Nil -> true
    | Cons _ -> false
    | Append (a, _) -> is_empty a
    | Delayed _ | Generator _ -> is_empty (force stream)

let rec size = function stream ->
  match stream.data with
    | Nil -> 0
    | Cons (_, tail) -> 1 + size tail
    | Append (a, b) -> size a + size b
    | Delayed _ | Generator _ -> size (force stream)

let concat_map = fun f stream -> join (map f stream)

let rec fold_right f e = function stream ->
  match stream.data with
    | Nil -> e
    | Cons (head, tail) -> f head (fold_right f e tail)      
    | Append (a, b) -> fold_right f (fold_right f e b) a 
    | Delayed _ | Generator _ -> fold_right f e (force stream)

let rec fold_left f e = function stream ->
  match stream.data with
    | Nil -> e
    | Cons (head, tail) -> fold_left f (f e head) tail
    | Append (a, b) -> fold_left f (fold_left f e a) b
    | Delayed _ | Generator _ -> fold_left f e (force stream)

let rec iter f = function stream ->
  match stream.data with
    | Nil -> ()
    | Cons (head, tail) -> f head ; iter f tail
    | Append (a, b) -> iter f a ; iter f b
    | Delayed _ | Generator _ -> iter f (force stream)

let rec take_rec k = function stream ->
  if k = 0 then makeStream Nil
  else match stream.data with
    | Nil -> raise (Out_of_bounds k)
    | Cons (head, tail) -> makeStream (Cons (head, take_rec (k - 1) tail))
    | Append (a, b) -> (try take_rec k a with Out_of_bounds k -> take_rec k b)
    | Delayed _ ->
	makeStream (Delayed (function () -> take_rec k (force stream)))
    | Generator (i, s, f) ->
	let count = ref k in
	let s' = function index ->
	  if !count = 0 then raise Empty
	  else decr count ; s index
	in
	  makeStream (Generator (i, s', f))

let take n = function stream ->
  if n < 0 then invalid_arg "take" Negative
  else take_rec n stream

let rec power f = function
  | 1 -> f
  | n -> function x -> f (power f (n - 1) x)

let rec drop_rec k = function stream ->
  match stream.data with
    | _ when k = 0 -> stream
    | Nil -> raise (Out_of_bounds k)
    | Cons (_, tail) -> drop_rec (k - 1) tail
    | Append (a, b) -> drop_rec k (plus a b)
    | Delayed _ -> 
	makeStream (Delayed (function () -> drop_rec k (force stream)))
    | Generator (i, s, f) -> let succ = power s k in 
	makeStream (Generator (succ i, s, f))

let drop n = function stream ->
  if n < 0 then invalid_arg "drop" Negative
  else drop_rec n stream

let rec exists f = function stream ->
  match stream.data with
    | Nil -> false
    | Cons (head, tail) -> f head || exists f tail
    | Append (a, b) -> exists f a || exists f b
    | Delayed _ | Generator _ -> exists f (force stream)

let rec for_all f = function stream ->
  match stream.data with
    | Nil -> true
    | Cons (head, tail) -> f head && for_all f tail
    | Append (a, b) -> for_all f a && for_all f b
    | Delayed _ | Generator _ -> for_all f (force stream)

let rec find f = function stream ->
  match stream.data with
    | Nil -> raise Not_found
    | Cons (head, _) when f head -> head
    | Cons (_, tail) -> find f tail
    | Append (a, b) -> (try find f a with Not_found -> find f b)
    | Delayed _ | Generator _ -> find f (force stream)

let rec filter f = function stream ->
  match stream.data with
    | Nil -> makeStream Nil
    | Cons (head, tail) when f head -> makeStream (Cons (head, filter f tail))
    | Cons (_, tail) -> filter f tail
    | Append (a, b) -> plus (filter f a) (filter f b)
    | Delayed _ | Generator _ -> 
	makeStream (Delayed (function () -> filter f (force stream)))

let rec partition_rec f trueS falseS = function stream ->
  match stream.data with
    | Nil -> (reverse trueS, reverse falseS)
    | Cons (head, tail) when f head -> 
	partition_rec f (cons head trueS) falseS tail
    | Cons (head, tail) ->
	partition_rec f trueS (cons head falseS) tail
    | Append (a, b) ->
	let (x, y) = partition_rec f (makeStream Nil) (makeStream Nil) a in
	  partition_rec f (reverse_append x trueS) (reverse_append y falseS) b
    | Delayed _ | Generator _ -> 
	(reverse_append trueS (filter f stream), 
	 reverse_append falseS (filter (fun x -> not (f x)) falseS))

let partition = fun f stream-> 
  partition_rec f (makeStream Nil) (makeStream Nil) stream
