(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * AVL Map over lists
 * Dictionnaire AVL sur les listes
 * Diccionario AVL sobre listas
 *
 *)

open BairePervasives

type ('a, 'b) lexicalMap = 
  | Epsilon of 'b * ('a, 'b) tree 
  | NoEpsilon of ('a, 'b) tree 
and 
  ('a, 'b) tree = 
  | E 
  | N of ('a, 'b) tree * 'a * ('a, 'b) tree * ('a, 'b) lexicalMap * int


(** Utils **)
(** Utilitaires **)
(** tiles **)


let moduleName = "LexicalMap"

let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type

let rec insert_all x = function
  | [] -> []
  | (list, v) :: tail -> (x :: list, v) :: insert_all x tail

let makeTree = fun l v r s -> 
  let h = match (l, r) with
    | (E, E) -> 1
    | (E, N (_, _, _ , _, n)) | (N (_, _, _, _, n), E) -> 1 + n
    | (N (_, _, _, _, m), N (_, _, _, _, n)) -> 1 + max m n
  in
    N (l, v, r, s, h)

let rotate_L = fun l v r s ->
  match r with
    | N (rl, rv, rr, rs, _) -> makeTree (makeTree l v rl s) rv rr rs
    | _ -> makeTree l v r s
	
let rotate_R = fun l v r s ->
  match l with
    | N (ll, lv, lr, ls, _) -> makeTree ll lv (makeTree lr v r s) ls
    | _ -> makeTree l v r s

let rotate_RL = fun l v r s ->
  match r with
    | N (N (rll, rlv, rlr, rls, _), rv, rr, rs, _) -> 
	makeTree (makeTree l v rll s) rlv (makeTree rlr rv rr rs) rls
    | _ -> makeTree l v r s

let rotate_LR = fun l v r s ->
  match l with
    | N (ll, lv, N (lrl, lrv, lrr, lrs, _), ls, _) ->
	makeTree (makeTree ll lv lrl ls) lrv (makeTree lrr v r s) lrs
    | _ -> makeTree l v r s

let height = function
  | E -> 0
  | N (_, _, _, _, n) -> n

let makeBTree = fun l v r s ->
  let hl = height l and hr = height r in
    if hl > hr + 1 then
      (match l with
	 | E -> invalid "makeBTree" UnexpectedPattern
         | N (ll, _, lr, _, _) ->
             if height ll >= height lr then rotate_R l v r s 
	     else rotate_LR l v r s
      )
    else if hr > hl + 1 then
      (match r with
	 | E -> invalid "makeBTree" UnexpectedPattern
         | N (rl, _, rr, _, _) ->
             if height rr >= height rl then rotate_L l v r s 
	     else rotate_RL l v r s
      )
    else makeTree l v r s

let rec makeSTree = fun l v r s ->
  let hl = height l and hr = height r in
    if hl > hr + 2 then
      (match l with
	 | E -> invalid "makeSTree" UnexpectedPattern
	 | N (ll, lv, lr, ls, _) -> makeBTree ll lv (makeSTree lr v r s) ls
      )
    else if hr > hl + 2 then
      (match r with
	 | E -> invalid "makeSTree" UnexpectedPattern
	 | N (rl, rv, rr, rs, _) -> makeBTree (makeSTree l v rl s) rv rr rs
      )
    else makeBTree l v r s

let rec append = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l1, v1, r1, s1, _), N (l2, v2, r2, s2, _)) ->
        makeBTree l1 v1 (makeBTree (append r1 l2) v2 r2 s2) s1

let rec appendS = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l1, v1, r1, s1, _), N (l2, v2, r2, s2, _)) ->
        makeSTree l1 v1 (makeSTree (appendS r1 l2) v2 r2 s2) s1


(** Categorial operations **)
(** Oprations catgorielles **)
(** Operaciones categoriales **)


let rec injection = fun list value ->
  match list with
    | [] -> Epsilon (value, E)
    | head :: tail -> NoEpsilon (makeBTree E head E (injection tail value))

let zero = function () -> NoEpsilon E

let rec to_list = function
  | Epsilon (v, s) -> ([], v) :: to_list_tree s
  | NoEpsilon s -> to_list_tree s

and to_list_tree = function
  | E -> []
  | N (l, v, r, s, _) -> to_list_tree l @ to_list_tree r @ 
      insert_all v (to_list s)


(** Binary operations **)
(** Oprations binaires **)
(** Operaciones binrias **)


(** Acces **)
(** Accs **)
(** Acceso **)


let rec insert_tree x value = function
    | E -> makeTree E (List.hd x) E (injection (List.tl x) value)
    | N (l, v, r, s, _) ->
	match compare (List.hd x) v with
	  | n when n < 0 -> makeBTree (insert_tree x value l) v r s
	  | n when n > 0 -> makeBTree l v (insert_tree x value r) s
	  | _ -> makeBTree l v r (insert_raise (List.tl x) value s)

and insert_raise = fun x v trie ->
  match (x, trie) with
    | ([], Epsilon (v', s')) -> 
	(match compare v v' with
	   | 0 -> raise Found
	   | _ -> Epsilon (v, s')
	)
      | ([], NoEpsilon s') -> Epsilon (v, s')
      | (_, NoEpsilon s') -> NoEpsilon (insert_tree x v s')
      | (_, Epsilon (v', s')) -> Epsilon (v', insert_tree x v s')
	  
let insert = fun x v trie ->
  try
    insert_raise x v trie
  with Found -> trie

let rec delete_tree x = function
  | E -> raise Not_found
  | N (l, v, r, s, _) ->
      match compare (List.hd x) v with
	| n when n < 0 -> makeBTree (delete_tree x l) v r s
	| n when n > 0 -> makeBTree l v (delete_tree x r) s
	| _ ->
	    match delete_raise (List.tl x) s with
	      | NoEpsilon E -> append l r
	      | s' -> makeBTree l v r s'

and delete_raise = fun x trie ->
  match (x, trie) with
    | ([], Epsilon (_, s')) -> NoEpsilon s'
    | ([], NoEpsilon s') -> raise Not_found
    | (_, NoEpsilon s') -> NoEpsilon (delete_tree x s')
    | (_, Epsilon (v, s')) -> Epsilon (v, delete_tree x s')

let delete = fun x trie ->
  try
    delete_raise x trie
  with Not_found -> trie


(** View **)
(** Vues **)
(** Vistas **)


(** Observers **)
(** Observateurs **)
(** Observadores **)


let is_empty = function
  | NoEpsilon E -> true
  | NoEpsilon (N _) | Epsilon _ -> false

let rec size_rec total = function
  | Epsilon (_, s) -> size_tree (total + 1) s
  | NoEpsilon s -> size_tree total s

and size_tree total = function
  | E -> total
  | N (l, _, r, s, _) ->
      size_tree (
	size_tree (
	  size_rec total s
	) l
      ) r

let size = function trie -> size_rec 0 trie

let rec images_rec result = function
  | Epsilon (v, s) -> images_tree (v :: result) s
  | NoEpsilon s -> images_tree result s

and images_tree result = function
  | E -> result
  | N (l, _, r, s, _) ->
      images_tree (
	images_tree (
	  images_rec result s
	) l
      ) r

let images = function trie -> images_rec [] trie

let rec member = fun x trie ->
  match (x, trie) with
    | ([], Epsilon _) -> true
    | ([], NoEpsilon _) -> false
    | (_, Epsilon (_, s)) | (_, NoEpsilon s) -> member_tree x s

and member_tree x = function
  | E -> false
  | N (l, v, r, s, _) ->
      match compare (List.hd x) v with
	| n when n < 0 -> member_tree x l
	| n when n > 0 -> member_tree x r
	| _ -> member (List.tl x) s

let rec lookup = fun x trie ->
  match (x, trie) with
    | ([], Epsilon (v, _)) -> v
    | ([], NoEpsilon _) -> raise Not_found
    | (_, Epsilon (_, s)) | (_, NoEpsilon s) -> lookup_tree x s

and lookup_tree x = function
  | E -> raise Not_found
  | N (l, v, r, s, _) ->
      match compare (List.hd x) v with
	| n when n < 0 -> lookup_tree x l
	| n when n > 0 -> lookup_tree x r
	| _ -> lookup (List.tl x) s


(** Traversals **)
(** Parcours **)
(** Recorridos **)


let rec iter_tree f path = function
  | E -> ()
  | N (l, v, r, s, _) ->
      iter_tree f path l ;
      iter_trie f (v :: path) s ;
      iter_tree f path r

and iter_trie f path = function
  | Epsilon (v, s') -> f path v ; iter_tree f path s'
  | NoEpsilon s' -> iter_tree f path s'

let iter = fun f trie -> iter_trie f [] trie

