(* Diego Olivier Fernandez Pons
 * Universit Pierre et Marie Curie, Paris VI
 * diego-olivier.fernandez-pons@cicrp.jussieu.fr
 * 
 * Set over lists
 * Ensemble de listes
 * Conjunto de listas
 *
 *)

open BairePervasives

type 
  'a trie = Epsilon of 'a tree | NoEpsilon of 'a tree and
  'a tree = E | N of 'a tree * 'a * 'a tree * 'a trie


(** Utils **)
(** Utilitaires **)
(** tiles **)


let moduleName = "LexicalSet"

let invalid = fun function_name error_type -> 
  invalid_argument moduleName function_name error_type

let rec insert_all x = function
  | [] -> []
  | head :: tail -> (x :: head) :: insert_all x tail

let makeTree = fun l v r s -> N (l, v, r, s)

let rec append_tree = fun a b ->
  match (a, b) with
    | (E, _) -> b
    | (_, E) -> a
    | (N (l1, v1, r1, s1), N (l2, v2, r2, s2)) ->
        makeTree l1 v1 (makeTree (append_tree r1 l2) v2 r2 s2) s1


(** Categorial operations **)
(** Oprations catgorielles **)
(** Operaciones categoriales **)


let rec injection = function
  | [] -> Epsilon E
  | head :: tail -> NoEpsilon (makeTree E head E (injection tail))

let zero = function () -> NoEpsilon E

let rec to_list = function
  | Epsilon s -> [] :: to_list_tree s
  | NoEpsilon s -> to_list_tree s

and to_list_tree = function
  | E -> []
  | N (l, v, r, s) -> to_list_tree l @ to_list_tree r @ 
      insert_all v (to_list s)


(** Binary operations **)
(** Oprations binaires **)
(** Operaciones binrias **)


(** Acces **)
(** Accs **)
(** Acceso **)


let rec insert_tree x = function
    | E -> makeTree E (List.hd x) E (injection (List.tl x))
    | N (l, v, r, s) ->
	match compare (List.hd x) v with
	  | n when n < 0 -> makeTree (insert_tree x l) v r s
	  | n when n > 0 -> makeTree l v (insert_tree x r) s
	  | _ -> makeTree l v r (insert_raise (List.tl x) s)

and insert_raise = fun x trie ->
  match (x, trie) with
    | ([], Epsilon s') -> raise Found
    | ([], NoEpsilon s') -> Epsilon s'
    | (_, NoEpsilon s') -> NoEpsilon (insert_tree x s')
    | (_, Epsilon s') -> Epsilon (insert_tree x s')
	  
let insert = fun x trie ->
  try
    insert_raise x trie
  with Found -> trie

let rec delete_tree x = function
  | E -> raise Not_found
  | N (l, v, r, s) ->
      match compare (List.hd x) v with
	| n when n < 0 -> makeTree (delete_tree x l) v r s
	| n when n > 0 -> makeTree l v (delete_tree x r) s
	| _ ->
	    match delete_raise (List.tl x) s with
	      | NoEpsilon E -> append_tree l r
	      | s' -> makeTree l v r s'

and delete_raise = fun x trie ->
  match (x, trie) with
    | ([], Epsilon s') -> NoEpsilon s'
    | ([], NoEpsilon s') -> raise Not_found
    | (_, NoEpsilon s') -> NoEpsilon (delete_tree x s')
    | (_, Epsilon s') -> Epsilon (delete_tree x s')

let delete = fun x trie ->
  try
    delete_raise x trie
  with Not_found -> trie


(** View **)
(** Vues **)
(** Vistas **)


(** Observers **)
(** Observateurs **)
(** Observadores **)


let is_empty = function
  | NoEpsilon E -> true
  | NoEpsilon (N _) | Epsilon _ -> false

let rec size_rec total = function
  | Epsilon s -> size_tree (total + 1) s
  | NoEpsilon s -> size_tree total s

and size_tree total = function
  | E -> total
  | N (l, _, r, s) ->
      size_tree (
	size_tree (
	  size_rec total s
	) l
      ) r

let size = function trie -> size_rec 0 trie

let rec member = fun x trie ->
  match (x, trie) with
    | ([], Epsilon _) -> true
    | ([], NoEpsilon _) -> false
    | (_, Epsilon s) | (_, NoEpsilon s) -> member_tree x s

and member_tree x = function
  | E -> false
  | N (l, v, r, s) ->
      match compare (List.hd x) v with
	| n when n < 0 -> member_tree x l
	| n when n > 0 -> member_tree x r
	| _ -> member (List.tl x) s


(** Traversals **)
(** Parcours **)
(** Recorridos **)


let rec iter_tree f path = function
  | E -> ()
  | N (l, v, r, s) ->
      iter_tree f path l ;
      iter_trie f (v :: path) s ;
      iter_tree f path r

and iter_trie f path = function
  | Epsilon s' -> f path ; iter_tree f path s'
  | NoEpsilon s' -> iter_tree f path s'

let iter = fun f trie -> iter_trie f [] trie

